

(in-package "CL-USER") 



(defpackage "OM-API"
  (:nicknames "OA")
  (:use "COMMON-LISP" "CL-USER" "CAPI" "LISPWORKS" "GP"))


(defpackage "OpenMusic"
   (:use "COMMON-LISP" "CL-USER" "OM-API" "HARLEQUIN-COMMON-LISP" "LISPWORKS" )
   (:import-from "CL-USER")
   (:nicknames "OM"))
 


(defvar *this-file* *load-truename*)

(defun compile&load (file &optional (verbose t))
   (let ((lisp-file 
          (if (pathname-type file) file (concatenate 'string (namestring file) ".lisp")))
         (fasl-file (compile-file-pathname file)))
     (when (or (not (probe-file fasl-file))
               (not (file-write-date lisp-file))
               (not (file-write-date fasl-file))
               (> (file-write-date lisp-file) (file-write-date fasl-file)))
       (compile-file file :verbose verbose))
     (load file :verbose verbose)))

(export 'compile&load :cl-user)

(defun set-cl-logical-path ()
   (let* ((root-path  (make-pathname :device (pathname-device *this-file*)
					:directory (butlast (butlast (pathname-directory *this-file*)))) )
	  (path-code  (make-pathname :device (pathname-device root-path)
				     :directory (append (pathname-directory root-path) (list "CODE"))))
	  (path-api  (make-pathname :device (pathname-device root-path)
                                    :directory (append (pathname-directory root-path) (list "CODE" "API" "OM-LW")))))
     (print (format () "~A**;*.*" (namestring root-path)))
     (setf (logical-pathname-translations "CL")
	   (list (list "**;*.*.*" (format () "~A**/*.*" (namestring root-path)))))
     (setf (logical-pathname-translations "code")
	   (list (list "**;*.*.*" (format () "~A**/*.*" (namestring path-code)))))
     (setf (logical-pathname-translations "api")
	   (list (list "**;*.*.*" (format () "~A**/*.*" (namestring path-api)))))
     (print (list  "--> logical path translation" (truename "cl:")))
     ))
 
(set-cl-logical-path)

;;; TEMP -- BUG LISPWORKS

; (trace (error :backtrace :bug-form :trace-output *terminal-io*))

;(editor:defcommand "Buffer List To File" (p)
;  ""
;  (with-open-file (out "~/LispWorks-Buffer-List.txt"
;                       :direction :output
;                       :if-exists :supersede)
;    (print editor::*buffer-list* out)))
;(editor:bind-key "Buffer List To File" #("control-c" "z"))
;;; END

(in-package :om)

(defvar *ml2-release* nil)

(defun load-om-api ()
  (load "code:api;om-LW;load-api.lisp"))

(defun load-om-kernel ()
  (load "code:kernel;kernelfiles.lisp")
  (mapc #'compile&load  om::*kernel-files*)
)

;;; (om::choise-directories (om::om-directory "cl:code;projects;*" :files nil :directories t)))
(defun load-om-projects (&optional projects)
  (let ((projects-list (if projects projects (oa::om-directory "cl:code;projects;*" :files nil :directories t))))
    (mapc #'(lambda (x) (load (om::string+ "cl:code;projects;" x ";" x ".lisp") :if-does-not-exist nil)) projects-list)
    ))

(defun clean-sources ()
  (load (make-pathname :directory (pathname-directory cl-user::*this-file*) :name "houseKeepingUtils"))
  (mapc #'(lambda (file) (delete-file file)) (p-directory "CL:**;*.xfasl" "CL:"))
  (mapc #'(lambda (file) (delete-file file)) (p-directory "CL:**;*.nfasl" "CL:"))
  )

; (clean-sources)

(eval-when (eval compile load)
  (load-om-api)
  (oa::load-om-libs '(:midi :audio :sdif))
  (load-om-kernel)
  (load-om-projects '("1-basicproject" "2-musicproject"))
)


; (showobjecteditor *current-workspace*)


(push :om *features*)

(defun launch-om ()
  ;; (setf om::*om-root* (truename "cl:"))
  ;; dans oa::om-root-init
  (om::load-modif-patches)
  (setf CLOS::*CHECK-MAKE-INSTANCE-INITARGS* nil)
  (oa::om-init-funcall)
  (om::show-workspaces-dialog))

;; (launch-om)

;;  (om-midi-startup)
;;  (init-scheduler)
;;  (midi-open)
;;  (om::play (make-instance 'om::note))
;;  (stop-player *general-player*)

;;  (ms::openplayer "OMPlayer")
;;  (om-init-funcall)