
;;;===============================
;;; OM-GRAPHIC OBJECTS
;;; DISPLAY AND ACTIONS CALLBACKS
;;;===============================

(export '(
          om-view-key-handler
          om-view-click-handler
          om-view-doubleclick-handler
          om-view-mouse-enter-handler
          om-view-mouse-leave-handler
          om-window-mouse-moved-handler
          
          om-init-motion
          om-click-motion-handler
          om-click-release-handler

           om-draw-contents
           om-component-border
           om-invalidate-view
           om-invalidate-corners
           om-with-delayed-redraw
           om-draw-view-outline
           om-redraw-view
           
           om-shift-key-p
          om-control-key-p
          om-command-key-p
          om-option-key-p
          
          om-wait-for-click
          om-mouse-position
          om-mouse-down-p
          ) :om-api)


(in-package :om-api)

;;;========================
;;; DISPLAY
;;;========================

(defmethod om-draw-contents ((self om-graphic-object)) t)

(defmethod om-component-border ((self om-graphic-object)) nil)

(defmethod om-draw-contents-callback ((self om-graphic-object) x y w h)
  ;; ??? pas la peine ?
  ;(when *motion-rep*
  ;     (capi::draw-pinboard-object self *motion-rep*))
  (om-draw-contents self))

(defmethod om-invalidate-view ((self om-graphic-object) &optional (erase t))
  (gp::invalidate-rectangle (om-get-view self)))

(defmethod om-invalidate-corners ((self om-graphic-object) topleft bottomright &optional (erase t))
  (gp::invalidate-rectangle (om-get-view self) (om-point-h topleft) (om-point-v topleft) 
                               (- (om-point-h bottomright) (om-point-h topleft)) 
                               (- (om-point-v bottomright) (om-point-v topleft))))


(defvar *delayed-redraw* nil)

(defmacro om-with-delayed-redraw (view &body body)
  `(progn   
     (gp::validate-rectangle (om-get-view ,view))
     (let ((*delayed-redraw* t)) ,@body)
     ;(maybe-call-update ,view)
     (update-for-subviews-changes ,view t)
     (capi:redraw-pinboard-layout (om-get-view ,view) 0 0 (om-width ,view) (om-height ,view))))

;;; nouveau.. 
(defmethod om-redraw-view ((self om-graphic-object))
  (capi:redraw-pinboard-layout (om-get-view self) 0 0 (om-width self) (om-height self) t))

(defmethod om-draw-view-outline ((self om-graphic-object) &optional (pensize 1))
  (om-with-focused-view self
                        (gp::with-graphics-state (*curstream* :thickness pensize)
                          (om-draw-rect 0 0 (- (om-width self)  1) (- (om-height self) 1)))))


;;;=======================
;;; MOUSE
;;;=======================


                          
(defmethod om-view-click-handler ((self om-abstract-window) position) (call-next-method))

(defmethod om-view-doubleclick-handler ((self om-abstract-window) pos) self)


;;; CLIC
;(defmethod om-clic-callback ((self om-graphic-object) x y)
;  (prog1 
;      (om-view-click-handler self (om-make-point x y))
;    (om-init-motion self x y)))

(defvar *clicked-view* nil)

(defmethod om-clic-callback ((self om-graphic-object) x y type)
  (set-meta-keys (eval type))
  (setf *clicked-view* (om-find-view-containing-point self (om-make-point x y)))
  (om-view-click-handler *clicked-view* (om-convert-coordinates (om-make-point x y) self *clicked-view*)))

(defmethod om-view-click-handler ((self om-graphic-object) position) self)

;(defmethod om-view-click-handler ((self om-graphic-object) position)
; (let ((clicked self)) 
;   (loop for item in (om-subviews self) 
;         while (equal clicked self) do
;         (when (om-view-contains-point-p item position)
;           (setf clicked (om-view-click-handler item (om-convert-coordinates position self item)))))
;   clicked))


;;; DOUBLE CLIC
(defmethod om-double-clic-callback ((self om-graphic-object) x y)
  (om-view-doubleclick-handler self (om-make-point x y)))

(defmethod om-view-doubleclick-handler ((self om-graphic-object) pos)
  (let ((clicked self)) 
    (loop for item in (om-subviews self) 
	  while (equal clicked self) do
	  (when (om-view-contains-point-p item pos)
	    (setf clicked (om-view-doubleclick-handler item (om-convert-coordinates pos self item)))))
    clicked))

;(defmethod  om-clic-motion-callback ((self om-graphic-object) x y)
;  (let ((clicked self)
;        (pos (om-make-point x y))) 
;    (loop for item in (om-subviews self) 
;	  while (equal clicked self) do
;	  (when (om-view-contains-point-p item pos)
;	    (setf clicked (om-clic-motion-callback item (om-point-h (om-convert-coordinates pos self item))
;                                                   (om-point-v (om-convert-coordinates pos self item))))))
;    clicked))

(defmethod om-init-motion (self x y))

(defmethod om-clic-motion-callback (self x y type) 
  (set-meta-keys (eval type))
  (when *clicked-view*
    (let ((newpoint (om-convert-coordinates (om-make-point x y) self *clicked-view*)))
      (om-click-motion-handler *clicked-view* newpoint))))

(defmethod om-click-motion-handler (self pos) t)


;;; RELEASE
(defmethod om-clic-release-callback ((self om-graphic-object) x y type) 
  (set-meta-keys (eval type))
  (om-click-release-handler self (om-make-point x y)))

(defmethod om-click-release-handler ((self om-graphic-object) pos) nil) 

;;; MOVE
(defvar *last-containing-view* nil)

(defmethod om-motion-callback ((self om-graphic-object) x y)
  (let ((view (om-find-view-containing-point self (om-make-point x y))))
    (unless (equal *last-containing-view* view)
      (om-view-mouse-enter-handler view)
      (when *last-containing-view*
        (om-view-mouse-leave-handler *last-containing-view*))
      (setf *last-containing-view* view))
    (update-view-cursor view)
  (om-window-mouse-moved-handler (om-view-window self) (om-make-point x y))
))

(defmethod om-window-mouse-moved-handler ((self om-abstract-window) position) 
  (declare (ignore self position)) t)

(defmethod om-view-mouse-enter-handler ((self om-graphic-object)) t)
(defmethod om-view-mouse-leave-handler ((self om-graphic-object)) t)

(defmethod update-view-cursor ((self om-view))
  (setf (capi::simple-pane-cursor self) (om-view-cursor self)))

(defmethod update-view-cursor ((self t)) nil)

(defmethod om-view-cursor ((self om-graphic-object)) nil)


;;;===============
;;; TOOLS
;;;===============

(setf *wait-for-click* nil)

(defun om-wait-for-click (win &optional (cursor t))
 "pas sur s'il faut le faire" )
     
(defun om-mouse-position (view)
 (let ((point (multiple-value-list (if view (capi::current-pointer-position :relative-to view :pane-relative-p t)
                                     (capi::current-pointer-position)))))
   (om-make-point (car point) (second point))))

(defun om-mouse-down-p () nil)

 
;;;=======================
;;; KEYBOARD
;;;=======================
        

(defun get-om-character (c)
  (case (char-code c)
    (63276  :om-key-pageup)
    (63277 :om-key-pagedown)
    (63275  :om-key-end)
    (63273  :om-key-home)
    (63234  :om-key-left)
    (63232  :om-key-up)
    (63235 :om-key-right)
    (63233  :om-key-down)
    (127  :om-key-delete)
    (3  :om-key-enter)
    (13  :om-key-return)
    (27  :om-key-esc)
    (9  :om-key-tab)
    (otherwise c)))

                      
(defmethod om-char-callback ((self om-abstract-window) x y c)
  (om-view-key-handler self (get-om-character c)))

(defmethod om-char-callback ((self om-view) x y c)
  (om-char-callback (om-view-window self) x y c))

(defmethod om-view-key-handler ((self t) key)
  (declare (ignore key)) nil)


(defvar *om-shift-key-p* nil)
(defvar *om-control-key-p*  nil)
(defvar *om-command-key-p*  nil)
(defvar *om-option-key-p*  nil)


(defun set-meta-keys (list)
  (setf *om-shift-key-p* (first list))
  (setf *om-command-key-p*  (second list) )
  (setf *om-option-key-p*   (third list)))

(defun release-meta-keys ()
  (setf *om-shift-key-p* nil)
  (setf *om-command-key-p*  nil)
  (setf *om-option-key-p*  nil))



(defun om-shift-key-p () *om-shift-key-p* )
(defun om-command-key-p () *om-command-key-p* )
(defun om-option-key-p ()  *om-option-key-p* )

(defun om-control-key-p () nil)



