;;===========================================================================
;OM API 
;Multiplatform API for OpenMusic
;Macintosh version (LispWorks)
;
;Copyright (C) 2004 IRCAM-Centre Georges Pompidou, Paris, France.
; 
;This program is free software; you can redistribute it and/or
;modify it under the terms of the GNU General Public License
;as published by the Free Software Foundation; either version 2
;of the License, or (at your option) any later version.
;
;See file LICENSE for further informations on licensing terms.
;
;This program is distributed in the hope that it will be useful,
;but WITHOUT ANY WARRANTY; without even the implied warranty of
;MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;GNU General Public License for more details.
;
;You should have received a copy of the GNU General Public License
;along with this program; if not, write to the Free Software
;Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
;
;Authors: Jean Bresson and Augusto Agon
;;===========================================================================

;;===========================================================================
;DocFile
; DRAG AND DROP
;;===========================================================================




(in-package :om-api)
;;;==========
;;; export :
;;;==========
(export '(
          om-drag-view
          om-drop-view
          
          om-drag-start
          om-drag-receive
          om-get-current-drag-initpos
          
          om-drag-create-item-bounds
          om-drag-selection-p
          
          om-drag-view-cursor
          om-drag-enter-view
          om-drag-leave-view
          om-view-drag-hilite-p
          
          ) :om-api)


;;;==================
;;;Class definitions
;;;==================
(defclass om-drag-session ()
   ((drag-on :accessor drag-on :initarg :drag-on :initform nil)
    (drag-init-pos :accessor drag-init-pos :initarg :drag-init-pos :initform nil)
    (dragged-view :accessor dragged-view :initarg :dragged-view :initform nil)
    (flavor :accessor flavor :initarg :flavor :initform nil)
    (drop-view :accessor drop-view :initarg :drop-view :initform nil)
    (value :accessor value :initarg :value :initform nil)
    (drag-region :accessor drag-region :initarg :drag-region :initform nil)
    (drag-ph :accessor drag-ph :initarg :drag-ph :initform nil)))

(defvar *om-drag-session* nil)

(defun om-init-drag-and-drop ()
   (setf *om-drag-session* (make-instance 'om-drag-session)))

(om-add-init-func 'om-init-drag-and-drop)


;;; DRAG/DROP VIEW

(defclass om-drag-view () ()
  (:default-initargs 
   :drag-allow-move-p t))

(defmethod om-drag-view-p ((self t)) nil)
(defmethod om-drag-view-p ((self om-drag-view)) t)

(defclass om-drop-view () ()
  (:default-initargs 
   :drag-accepted-flavor-list (list :|OMDR|)
   :drop-callback 'finish-pinboard-drop-callback
))

(defmethod om-drop-view-p ((self t)) nil)
(defmethod om-drop-view-p ((self om-drop-view)) t)



;-------tools

;;; all operations are on panes
(defmethod om-item-pane ((self t)) self)
(defmethod om-item-pane ((self om-standard-dialog-item)) (om-view-container self))

(defun region2image (region)
  (when (and region (elements region))
    (let* ((bounds (om-region-bounds region))
           (ofsetx (* -1 (first bounds)))
           (ofsety (* -1 (second bounds))) rep)
      (om-offset-region region ofsetx ofsety)
      (setf rep
            (capi:with-internal-metafile  
                (port :pane nil  :bounds (list 0 0 (+ (third bounds) ofsetx) (+ (fourth bounds) ofsety)))
              (loop for elt in (elements region) do
                    (when (rect-p elt)
                      (gp:draw-rectangle port (rx elt) (ry elt) (rw elt) (rh elt) 
                                         :filled t :foreground (color::make-rgb 0.6s0 0.6s0 0.6s0 0.5)
                                         ) ))
              port))
      (list rep ofsetx ofsety) )))

(defmethod built-d&d-phantome ((arg t))
  (let ((rep (drag-ph *om-drag-session*)))
    (values (draw-metafile-to-image (om-item-pane arg) (car rep) :alpha 0.1)
            (+ (second rep) (om-point-h (drag-init-pos *om-drag-session*)))
            (+ (third rep) (om-point-v (drag-init-pos *om-drag-session*))))))


;---------Methods

(defmethod om-drag-selection-p ((view om-drag-view) local-mouse-position)
  (declare (ignore view local-mouse-position)) nil)

(defmethod om-drag-start ((self om-drag-view)) nil)

;-----Callbacks

(defvar *d-d-first-click* nil)
(defvar *motion-call-next-method* nil)

(defmethod om-view-click-handler :before ((self om-drag-view) position)
  (setf *d-d-first-click* position)
  (setf (drag-on *om-drag-session*) nil)
  (setf *motion-call-next-method* nil))

(defmethod  om-click-motion-handler :before ((view om-drag-view) pos) 
  (unless  *motion-call-next-method*
    (unless (drag-on *om-drag-session*)
      (let ((start-drag (om-drag-selection-p view *d-d-first-click*)))
        (setf *motion-call-next-method* (not start-drag))
        (when start-drag
          (setf (drag-on *om-drag-session*) t)
          (setf (dragged-view *om-drag-session*) view)
          (setf (drag-init-pos *om-drag-session*)  *d-d-first-click*)
          (setf start-drag (om-drag-start view))
          (setf *motion-call-next-method* (not start-drag))
          (when start-drag
            (setf (drag-ph *om-drag-session*) (region2image (drag-region *om-drag-session*)))
            (capi:drag-pane-object (om-item-pane view) "OM" :string "OM" :image-function 'built-d&d-phantome)
            ))))))

;; :operations '(:move :copy)

(defun finish-pinboard-drop-callback (self drop-object stage)
 (case  stage
    (:formats
      (capi:set-drop-object-supported-formats drop-object '(:string)))
    (:enter
     (when (and (capi:drop-object-provides-format drop-object :string)
                (capi:drop-object-allows-drop-effect-p drop-object :copy))
       (setf (capi:drop-object-drop-effect drop-object) :copy))
     (om-drag-enter-view self))
    (:leave
     (when (and (capi:drop-object-provides-format drop-object :string)
                (capi:drop-object-allows-drop-effect-p drop-object :copy))
       (setf (capi:drop-object-drop-effect drop-object) :copy))
     (om-drag-leave-view self))
    (:drag
    (when (and (capi:drop-object-provides-format drop-object :string)
               (capi:drop-object-allows-drop-effect-p drop-object :copy))
      (setf (capi:drop-object-drop-effect drop-object) :copy)))
   (:drop
    (when (drag-on *om-drag-session*)
     (setf (drop-view *om-drag-session*) self)
     (when (om-drag-receive self (dragged-view *om-drag-session*))
     (setf (capi:drop-object-drop-effect drop-object) :copy)))))
 )


(defmethod om-drag-receive ((view om-drop-view) 
                            (dragged-view t))
  (declare (ignore view dragged-view)) nil)

(defmethod om-drag-receive ((view t) 
                             (dragged-view t))
  (declare (ignore flavor dragged-view value))
  (format t "~&Drag Canceled.~%") nil)

;------om-get-current-drag-initpos

(defmethod om-get-current-drag-initpos ((view t))
  (if (equal view (dragged-view *om-drag-session*))
       (drag-init-pos *om-drag-session*)
     (om-convert-coordinates (drag-init-pos *om-drag-session*) (dragged-view *om-drag-session*) view)))

;------om-drag-create-item-bounds

(defun om-drag-create-item-bounds (view region)
   (declare (ignore view))
   (setf (drag-region *om-drag-session*) (om-copy-region region)))

;------om-drag-view-cursor

(defmethod om-drag-view-cursor ((self om-drag-view)) t)

;------om-view-drag-hilite-p

(defmethod om-view-drag-hilite-p ((self t)) t)

;------om-drag-enter-view

(defmethod om-drag-enter-view ((self t))
  (when (om-view-drag-hilite-p self)
   (om-with-focused-view self
     (om-draw-hilite-rect 0 0 (om-width self) (om-height self))))) 

;------om-drag-leave-view
(defmethod om-drag-leave-view ((self t))
  (om-redraw-view self))


