;; ============================================================================================
;; The LibAudioStream Library is Copyright (c) Grame, Computer Music Research Laboratory 03-05
;;
;; Grame : Computer Music Research Laboratory
;; Web : http://www.grame.fr/Research
;; ============================================================================================

;; This file contains definitions for entry points of the LibAudioStream library
;; It must be used with the LibAudioStream.Framework located in /System/Library/Frameworks

(in-package :cl-user)

(cl:defpackage "Audio"
  (:nicknames "AU")
   (:use common-lisp))

(in-package :au)

(defvar *liblibaudio* "/Library/Frameworks/LibAudioStream.framework/LibAudioStream")

(defvar *libaudiostream* nil)

(defun libaudiostream-framework ()
  (or *libaudiostream*
     (setq *libaudiostream*
           (if (probe-file *liblibaudio*)
               (progn (cffi:load-foreign-library *liblibaudio*)
                 ;(hcl::add-special-free-action 'DeleteSound) 
                 t)))))



;;;  (libaudiostream-framework)

;; libsndfile types

(defparameter SF_FORMAT_WAV     #x010000)
(defparameter SF_FORMAT_AIFF	#x020000)
(defparameter SF_FORMAT_AU	#x030000)
(defparameter SF_FORMAT_RAW	#x040000)
(defparameter SF_FORMAT_PAF	#x050000)
(defparameter SF_FORMAT_SVX	#x060000)
(defparameter SF_FORMAT_NIST	#x070000)
(defparameter SF_FORMAT_VOC	#x080000)
(defparameter SF_FORMAT_IRCAM	#x0A0000)
(defparameter SF_FORMAT_W64	#x0B0000)
(defparameter SF_FORMAT_MAT4	#x0C0000)
(defparameter SF_FORMAT_MAT5	#x0D0000)	
(defparameter SF_FORMAT_PCM_S8	#x0001)
(defparameter SF_FORMAT_PCM_16	#x0002)
(defparameter SF_FORMAT_PCM_24	#x0003)
(defparameter SF_FORMAT_PCM_32	#x0004)
(defparameter SF_FORMAT_PCM_U8	#x0005)


;;---------------------------------------------------------------------------------
;;---------------------------------------------------------------------------------
;;
;; 				Player Data Structures
;;
;;---------------------------------------------------------------------------------
;;---------------------------------------------------------------------------------

(defparameter kPortAudioRenderer 0)
(defparameter kJackRenderer 1)

(defparameter no_err 0)
(defparameter open_err -1)
(defparameter load_err -2)
(defparameter file_not_found_err -3)
(defparameter state_err -4)

(cffi:defcstruct TChannelInfo 
  (fStatus :long)
  (fCurFrame :long)
  (fVol :float)
  (fPanLeft :float)
  (fPanRight :float)
  (fLeftOut :long)
  (fRightOut :long))

(cffi:define-foreign-type audio-name () ':pointer)
 
;;................................................................................: status
(defmacro status (e)
 `(cffi:foreign-slot-value ,e 'TChannelInfo 'fStatus))

;;................................................................................: vol
(defmacro vol (e)
 `(cffi:foreign-slot-value ,e 'TChannelInfo 'fVol))

;;................................................................................: pan
(defmacro panLeft (e)
 `(cffi:foreign-slot-value ,e 'TChannelInfo 'fPanLeft))

(defmacro panRight (e)
 `(cffi:foreign-slot-value ,e 'TChannelInfo 'fPanRight))

;;................................................................................: left-out
(defmacro left-out (e)
 `(cffi:foreign-slot-value ,e 'TChannelInfo 'fLeftOut))

;;................................................................................: right-out
(defmacro right-out (e)
 `(cffi:foreign-slot-value ,e 'TChannelInfo 'fRightOut))



;entrypoints
;................................................................................: LibAudioStream
(cffi:defcfun ("LibVersion" lib-version) :short)

(defun LibVersion () 
 (lib-version))
     

;;; Build sound

;................................................................................: MakeNullSound
(cffi:defcfun ("MakeNullSoundPtr" make-null-sound-ptr) :pointer (length :long))

(defun MakeNullSound (length)
 (let ((sound (make-null-sound-ptr length)))
   ; (hcl::flag-special-free-action sound)
   sound))
 
;................................................................................: MakeReadSound
(cffi:defcfun  ("MakeReadSoundPtr" make-read-sound-ptr) :pointer (s audio-name))

(defun MakeReadSound (name)
  (cffi:with-foreign-string (s name)
    (let ((sound (make-read-sound-ptr s)))
      ; (hcl::flag-special-free-action sound)
    sound)))

;................................................................................: MakeRegionSound
(cffi:defcfun  ("MakeRegionSoundPtr" make-region-sound-ptr) :pointer (s audio-name) (begin :long) (end :long))

(defun MakeRegionSound (name begin end)
  (cffi:with-foreign-string (s name)
    (let ((sound (make-region-sound-ptr s begin end)))
    ; (hcl::flag-special-free-action sound)
    sound)))


;................................................................................: MakStereoSound
(cffi:defcfun ( "MakeStereoSoundPtr" make-stereo-sound) :pointer (sound :pointer))

(defun MakeStereoSound (sound)
   (let ((son (make-stereo-sound sound)))
     ; (hcl::flag-special-free-action son)
    son))

;................................................................................: MakeFadeSound
(cffi:defcfun ("MakeFadeSoundPtr" make-fade-sound-ptr) :pointer (sound :pointer) (fadein :long) (fadeout :long))

(defun MakeFadeSound (sound fadein fadeout)
   (let ((son (make-fade-sound-ptr sound fadein fadeout)))
     ; (hcl::flag-special-free-action son)
    son))

;................................................................................: MakeLoopSound
(cffi:defcfun ("MakeLoopSoundPtr" make-loop-sound-ptr) :pointer (sound :pointer) (len :long))

(defun MakeLoopSound (sound len)
   (let ((son (make-loop-sound-ptr sound len)))
     ; (hcl::flag-special-free-action son)
    son))
                   

;................................................................................: MakeCutSound
(cffi:defcfun  ("MakeCutSoundPtr" make-cut-sound-ptr) :pointer (sound :pointer) (begin :long) (end :long))

(defun MakeCutSound (sound begin end)
   (let ((son (make-cut-sound-ptr sound begin end)))
     ; (hcl::flag-special-free-action son)
    son))

;................................................................................: MakeSeqSound
(cffi:defcfun  ("MakeSeqSoundPtr" make-seq-sound-ptr) :pointer (s1 :pointer) (s2 :pointer) (crossfade :long))

(defun MakeSeqSound (s1 s2 crossfade)
   (let ((son (make-seq-sound-ptr s1 s2 crossfade)))
     ; (hcl::flag-special-free-action son)
    son))


;................................................................................: MakeMixSound
(cffi:defcfun  ("MakeMixSoundPtr" make-mix-sound-ptr) :pointer (s1 :pointer) (s2 :pointer))

(defun MakeMixSound (s1 s2 )
   (let ((son (make-mix-sound-ptr s1 s2 )))
     ; (hcl::flag-special-free-action son)
    son))


;................................................................................: MakeTransformSound
(cffi:defcfun  ( "MakeTransformSoundPtr" make-transform-sound-ptr) :pointer (sound :pointer) (effect :pointer) (fadein :long) (fadeout :long))

(defun MakeTransformSound (sound effect fadein fadeout )
   (let ((son (make-transform-sound-ptr sound effect fadein fadeout )))
     ; (hcl::flag-special-free-action son)
    son))

;................................................................................: MakeRegionSound
(cffi:defcfun  ( "MakeWriteSoundPtr" make-write-sound-ptr) :pointer (s audio-name) (sound :pointer) (format :long))

(defun MakeWriteSound (name sound format)
   (cffi:with-foreign-string (s name)
     (let ((son (make-write-sound-ptr s sound format )))
       ; (hcl::flag-special-free-action son)
       son)))
 

;................................................................................: MakeInputSound
(cffi:defcfun  ( "MakeInputSoundPtr" make-input-sound-ptr) :pointer)

(defun MakeInputSound ()
   (let ((son (make-input-sound-ptr)))
     ; (hcl::flag-special-free-action son)
       son))
 

;................................................................................: MakeRendererSound
(cffi:defcfun  ( "MakeRendererSoundPtr" make-renderer-sound-ptr) :pointer (sound :pointer))

(defun MakeRendererSound (sound)
  (let ((son (make-renderer-sound-ptr sound)))
    ; (hcl::flag-special-free-action son)
      son))

;................................................................................: GetLengthSound
(cffi:defcfun  ( "GetLengthSoundPtr" get-lenght-sound-ptr) :long (sound :pointer))

(defun GetLengthSound (sound)
  (get-lenght-sound-ptr sound))

;................................................................................: GetChannelsSound
(cffi:defcfun  ( "GetChannelsSoundPtr" get-channel-sound-ptr) :long (sound :pointer))

(defun GetChannelsSound (sound)
  (get-channel-sound-ptr sound))

;................................................................................: ResetSound
(cffi:defcfun  ( "ResetSoundPtr" reset-sound-ptr) :void (sound :pointer))

(defun ResetSound (sound)
  (reset-sound-ptr sound))

;................................................................................: ReadSound
(cffi:defcfun  ( "ReadSoundPtr" read-sound-ptr) :long (sound :pointer) (buffer :pointer) (buffer-size :long) (channels :long))

(defun ReadSound (sound buffer buffer_size channels)
  (read-sound-ptr sound buffer buffer_size channels))

;................................................................................: DeleteSound
(cffi:defcfun  ( "DeleteSoundPtr" delete-sound) :void (sound :pointer))

(defun DeleteSound (sound)
  (delete-sound sound))


;................................................................................: OpenAudioPlayer
(cffi:defcfun  ( "OpenAudioPlayer" open-audio-player) :pointer
                     (inchan :long)
                     (outchan :long)
                     (channels :long)
                     (sr :long)
                     (bs :long)
                     (sbs :long)
                     (rtbs :long)
                     (renderer :long)
                     (thread_num :long))

(defun OpenAudioPlayer (inchan outchan channels sr bs sbs rtbs renderer thread_num)
  (open-audio-player inchan outchan channels sr bs sbs rtbs renderer thread_num))

;;................................................................................: CloseAudioPlayer
(cffi:defcfun  ( "CloseAudioPlayer" close-audio-player) :void (player :pointer))

(defun CloseAudioPlayer (player)
  (close-audio-player player))

;; Channels
;................................................................................: LoadChannel


(cffi:defcfun  ("LoadChannelPtr" load-channel-ptr) :long
   (player :pointer)
   (sound :pointer)
   (chan :long)
   (vol :float)
   (panLeft :float)
   (panRight :float))


(defun LoadChannel (player sound chan vol panLeft panRight)
  (load-channel-ptr player sound chan vol panLeft panRight))


;................................................................................: GetInfoChannel
(cffi:defcfun  ("GetInfoChannel" get-info-channel) :void (player :pointer) (chan :long) (info :pointer))


(defun GetInfoChannel (player chan info) 
  (get-info-channel player chan info))

;; Transport
;................................................................................: StartAudioPlayer
(cffi:defcfun  ("StartAudioPlayer" start-audio-player) :void (player :pointer) )

(defun StartAudioPlayer (player)
  (start-audio-player player))

;................................................................................: StartAudioPlayer
(cffi:defcfun  ("StopAudioPlayer" stop-audio-player) :void (player :pointer) )

(defun StopAudioPlayer (player)
  (stop-audio-player player))

;................................................................................: StartSound
(cffi:defcfun  ("StartChannel" start-channel) :void (player :pointer) (chan :long) )

(defun StartChannel (player chan)
  (start-channel player chan))

;................................................................................: ContSound
(cffi:defcfun  ("ContChannel" cont-channel) :void (player :pointer) (chan :long) )

(defun ContChannel (player chan)
  (cont-channel player chan))

;................................................................................: StopSound
(cffi:defcfun  ("StopChannel" stop-channel) :void (player :pointer) (chan :long) )

(defun StopChannel (player chan)
  (stop-channel player chan))

;; Params

;................................................................................: SetVolChannel
(cffi:defcfun  ("SetVolChannel" set-vol-channel) :void (player :pointer) (chan :long) (vol :float) )

(defun SetVolChannel (player chan vol)
  (set-vol-channel player chan vol))

;................................................................................: SetPanChannel
(cffi:defcfun  ("SetPanChannel" set-pan-channel) :void (player :pointer) (chan :long) (panleft :float) (panright :float))

(defun SetPanChannel (player chan panLeft panRight)
  (set-pan-channel player chan panLeft panRight))

;................................................................................: SetEffectListChannel
(cffi:defcfun  ("SetEffectListChannel" set-effect-list-channel) :void (player :pointer) (chan :long) (effect_list :pointer) (fadein :long) (fadeout :long))

(defun SetEffectListChannel (player chan effect_list fadein fadeout)
  (set-effect-list-channel (player chan effect_list fadein fadeout)))

;; Master

;................................................................................: SetVolAudioPlayer
(cffi:defcfun  ("SetVolAudioPlayer" set-vol-audio-player) :void (player :pointer)  (vol :float))

(defun SetVolAudioPlayer (player vol)
 (set-vol-audio-player player vol))
;................................................................................: SetPanSound
(cffi:defcfun  ("SetPanAudioPlayer" set-pan-audio-player) :void (player :pointer)  (vol :float) (panl :float) (panr :float))

(defun SetPanAudioPlayer (player panLeft panRight)
  (set-pan-audio-player player panLeft panRight))

;................................................................................: SetEffectAudioPlayer
(cffi:defcfun  ("SetEffectListAudioPlayer" set-effect-list-audio-player) :void (player :pointer)  (effectlist :pointer) (fadein :long) (fadeout :long))

(defun SetEffectListAudioPlayer (player effect_list fadein fadeout)
  (set-effect-list-audio-player player effect_list fadein fadeout))

;;;========================== EFFECTS 

(cffi:defcfun  ("DeleteEffectListPtr" delete-effect-list-ptr) :void  (effectlist :pointer) )

(defun DeleteEffectList (effect_list)
 (delete-effect-list-ptr effect_list))

(cffi:defcfun  ("DeleteEffectPtr" delete-effect-ptr) :void  (effect :pointer) )

(defun DeleteEffect (effect)
  (delete-effect-ptr effect))

(cffi:defcfun  ("MakeAudioEffectListPtr" make-audio-effect-list) :pointer  )

(defun MakeAudioEffectList ()
  (let ((effect_list (make-audio-effect-list) ))
     effect_list)) 

(cffi:defcfun  ("AddAudioEffectPtr" add-audio-effect-ptr) :pointer  (effectlist :pointer) (effect :pointer))
     
(defun AddAudioEffect (effect-list effect)
  (add-audio-effect-ptr effect-list effect))

(cffi:defcfun  ("RemoveAudioEffect" remove-audio-effect-ptr) :pointer  (effectlist :pointer) (effect :pointer))

(defun RemoveAudioEffect (effect-list effect)
  (remove-audio-effect-ptr effect-list effect))

(cffi:defcfun  ("MakeVolAudioEffectPtr" make-vol-audio-effect-ptr) :pointer  (gain :float))

(defun MakeVolAudioEffect (gain)
  (let ((effect (make-vol-audio-effect-ptr gain)))
     
     effect))

(cffi:defcfun  ("MakeMonoPanAudioEffectPtr" make-mono-pan-audio-effect-ptr) :pointer  (pan :float))

(defun MakeMonoPanAudioEffect (pan)
  (let ((effect (make-mono-pan-audio-effect-ptr pan)))
     effect))

(cffi:defcfun  ("MakeStereoPanAudioEffectPtr" make-stereo-pan-audio-effect-ptr) :pointer  (panl :float) (panr :float))

(defun MakeStereoPanAudioEffect (panLeft panRight)
  (let ((effect (make-stereo-pan-audio-effect-ptr panLeft panRight)))
     effect))

(cffi:defcfun  ("MakeFaustAudioEffectPtr" make-faust-audio-effect-ptr) :pointer  (s audio-name))

(defun MakeFaustAudioEffect (name)
  (cffi:with-foreign-string (s name)
    (let ((effect (make-faust-audio-effect-ptr s)))
       effect)))


(cffi:defcfun  ("GetControlCountPtr" get-control-count-ptr) :long  (effect :pointer))

(defun GetControlCount (effect)
 (get-control-count-ptr effect))


(cffi:defcfun  ("GetControlParamPtr" get-control-param-ptr) :void  (effect :pointer) (control :long) (name audio-name) (min :pointer) (max :pointer) (init :pointer))

(defun GetControlParam (effect control)
  (let ((name (cffi::%foreign-alloc 64))
        (min (cffi::%foreign-alloc 4))
        (max (cffi::%foreign-alloc 4))
        (init (cffi::%foreign-alloc 4) ) str minrep maxrep initrep)
 (get-control-param-ptr effect control name min max init)
 (setf str (cffi::foreign-string-to-lisp name))
 (setf minrep (cffi::mem-ref min :float))
 (setf maxrep (cffi::mem-ref max :float))
 (setf initrep (cffi::mem-ref init :float))
 (cffi::foreign-free name)
 (cffi::foreign-free min)
 (cffi::foreign-free max)
 (cffi::foreign-free init)
 (values str minrep maxrep initrep)))


(cffi:defcfun  ("SetControlValuePtr" set-control-value-ptr) :void  (effect :pointer) (control :long) (value :float))

(defun SetControlValue (effect control value)
 (set-control-value-ptr effect control value))

(cffi:defcfun  ("GetControlValuePtr" get-control-value-ptr) :float  (effect :pointer) (control :long))
(defun GetControlValue (effect control)
 (get-control-value-ptr effect control))

