;;===========================================================================
;OM API 
;Multiplatform API for OpenMusic
;Macintosh version (Digitool Macintosh Common Lisp - MCL)
;
;Copyright (C) 2004 IRCAM-Centre Georges Pompidou, Paris, France.
; 
;This program is free software; you can redistribute it and/or
;modify it under the terms of the GNU General Public License
;as published by the Free Software Foundation; either version 2
;of the License, or (at your option) any later version.
;
;See file LICENSE for further informations on licensing terms.
;
;This program is distributed in the hope that it will be useful,
;but WITHOUT ANY WARRANTY; without even the implied warranty of
;MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;GNU General Public License for more details.
;
;You should have received a copy of the GNU General Public License
;along with this program; if not, write to the Free Software
;Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
;
;Authors: Augusto Agon and Jean Bresson
;;===========================================================================

;;===========================================================================
;DocFile
; AUDIO CLASSES AND FUNCTIONS
;;===========================================================================


(in-package :om-api)



(export '(
          om-start-audio
          
          om-open-audio-player
          om-close-audio-player
          om-add-sound-to-player

          om-start-player
          om-pause-player
          om-continue-player
          om-stop-player
          om-reset-player
          
          om-set-track-volume
          om-set-track-pan
          
          om-play-one-sound
          ) :om-api)



(defvar *tracks-list* nil)
(defvar *init-settings-table* nil)

(defun om-start-audio ()
  (setf *tracks-list* (make-hash-table))
  (setf *init-settings-table* (make-hash-table))
  (loop for i from 1 to 16 do (setf (gethash i *init-settings-table*) (list 1.0 0.5)))
  (setf au::*libaudiostream* nil)
  (au::libaudiostream-framework)
 t)

(defvar *player-sample-rate* 44100)

(defun om-open-audio-player ()
  (au::OpenAudioPlayer 2 2 32 44100 1024 65536 65536 au::kPortAudioRenderer 1)
  )

(defun om-close-audio-player (player)
  (when player
    (au::CloseAudioPlayer player)))


;;; sound
(defmethod om-read-sound ((self om-sound) &optional start end)
  (let ((s (if (and start end)
             (au::MakeRegionSound (om-path2cmdpath (filename self)) (round (* start (sample-rate self))) (round (* end (sample-rate self)))) 
             (au::MakeReadSound (om-path2cmdpath (filename self))))))
    (if (cffi:null-pointer-p s) nil s)))

;;; stream
(defmethod om-read-sound ((self t) &optional start end)
  (if (and start end)
    (let ((s (au::MakeCutSound self (round (* start *player-sample-rate*)) (round (* end *player-sample-rate*)))))
      (if (cffi:null-pointer-p s) nil s))
    self))



(defun om-add-sound-to-player (player sound at &optional start end (tracknum 1) (vol nil) (pan nil))
  (let ((snd nil))
      ;;; si intervalle
    (if (and start end)
      (unless (> at end) ;;; on joue pas
        (if (>= start at)  ;;; on prend une region du son
          (setq snd (om-read-sound sound
                                   (round (- start at) 1000) 
                                   (round (- end at) 1000)))
          ;;; on ajoute un blanc avant et on prend une region plus petite
          (let ((s (om-read-sound sound
                                0 
                                (round (- end at) 1000))))
                (if s (setq snd 
                            (au::MakeSeqSound 
                             (au::MakeNullSound (round (* (- at start) *player-sample-rate*) 1000))
                             s
                             0))  
                    (print "error read sound")
                    ))))
      
      ;;; sinon on joue le son (eventuellement shift de AT ms)
      (progn   
        (setq snd (om-read-sound sound))
        (if snd 
          (when (not (zerop at))  
            (let ((position (round (* at *player-sample-rate*) 1000)))
              (setq snd (au::MakeSeqSound (au::MakeNullSound position) snd 0))) 
            )
          (print "error read sound")
          ))
      
      )  ;;; endif

    (when (and snd (or vol pan))
      (setq effectlist (au::MakeAudioEffectList))
      (when pan
        (setq effectlist (au::AddAudioEffect effectlist (au::MakeMonoPanAudioEffect pan)))
        )
      (when vol 
        (setq effectlist (au::AddAudioEffect effectlist (au::MakeVolAudioEffect vol)))
        )
      (setq snd (au::MakeTransformSound snd effectlist 50 50))
      )
          
      ;;; mix le son dans son channel
      (when snd
        (let ((trk (gethash tracknum *tracks-list*)))
          (if trk
            (setf (gethash tracknum *tracks-list*) (au::MakeMixSound trk snd))
            (setf (gethash tracknum *tracks-list*) snd))
          ))  
      
  ))


(defun om-reset-player (player)
  (clrhash *tracks-list*))

(defvar *player-state* nil)

(defun pan2panpan (pan)
  (let ((val (- (* pan 2.0) 1)))
    (if (> val 0)
      (list 1.0 val)
      (list (+ val 1) 0.0)))) 

(defun om-start-player (player)
  (if *player-state* 
    ;;; loop : restart sans recharger les canaux
    (progn 
      (au::StopAudioPlayer player)
      (maphash #'(lambda (num snd)
                  ; (rlet ((i :TChannelInfo))
                  ;   (au::getinfochannel player num i)
                  ;   (when (= 2 (au::status i))
                       (au::StopChannel player num)
                  ;     ))
                   (au::StartChannel player num)
                   ) *tracks-list*))
    (maphash #'(lambda (num snd)
                 (let ((pan2 (if (= 1 (au::GetChannelsSound snd))
                               (list (cadr (gethash num *init-settings-table*)) (cadr (gethash num *init-settings-table*)))
                               (pan2panpan (cadr (gethash num *init-settings-table*))))))
                   (au::LoadChannel player snd num (car (gethash num *init-settings-table*)) 
                                    (car pan2) (cadr pan2))
                   ;(rlet ((i :TChannelInfo))
                   ;  (au::getinfochannel player num i)
                   ; (when (= 2 (au::status i))
                   (au::StopChannel player num)
                   ;    ))
                   (au::StartChannel player num)
                   )
                   ) *tracks-list*)
             )
    (au::StartAudioPlayer player)
    (setf *player-state* t)
    )

#|
(setf player (au::OpenAudioPlayer 0 2 32 44100 1024 65536 65536 au::kPortAudioRenderer 1))
(rlet ((i :TChannelInfo))
  (au::getinfochannel player 0 i)
  (print (au::status i)))
|#

(defun om-pause-player (player)
  (maphash #'(lambda (num snd)
               (au::StopChannel player num))
           *tracks-list*))


(defun om-continue-player (player)
  (maphash #'(lambda (num snd)
               (au::ContChannel player num))
           *tracks-list*)  
  (au::StartAudioPlayer player))


(defun om-stop-player (player)
  (maphash #'(lambda (num snd)
               ;(rlet ((i :TChannelInfo))
               ;  (au::getinfochannel player num i)
               ;  (when (= 2 (au::status i))
                   (au::StopChannel player num)
               ;    ))
               ) *tracks-list*)
  (clrhash *tracks-list*)
  (au::StopAudioPlayer player)
  (setf *player-state* nil)
  ;;; pour liberer des fichiers...
  ;(gc)
  )


(defun om-set-track-volume (player tracknum vol)
  (au::SetVolChannel player tracknum vol)
  (setf (car (gethash tracknum *init-settings-table*)) vol)  
  )


;;; !!!!
(defun om-set-track-pan (player tracknum pan)
  (let ((pan2 (pan2panpan pan)))
    (au::SetPanChannel player tracknum (car pan2) (cadr pan2))
    (setf (cadr (gethash tracknum *init-settings-table*)) pan)
  ))



;;;; ---------------------
;;; TEST PLAY
(defun om-play-one-sound (snd player)
  (om-reset-player player)
  (let ((trk (gethash 16 oa::*tracks-list*)))
    (if trk
        (setf (gethash 16 oa::*tracks-list*) (au::MakeMixSound trk snd))
      (setf (gethash 16 oa::*tracks-list*) snd))
    )
  (om-start-player player))


