(in-package :om-api)


(export '(
                om-sound
                om-make-sound
                
                om-sound-file-name
                om-sound-n-samples
                om-sound-sample-rate
                om-sound-sample-size
                om-sound-n-channels
                om-sound-data-pos
                om-sound-format

                om-sound-get-pict
                om-read-sound-data
                
                ) :om-api)


;;; crotale.aiff 372958
;;; b24.aiff 88200

;;;===================
;;; SOUND CLASS
;;;===================
(defclass om-sound ()  
   ((filename :accessor filename :initarg :filename :initform nil)
    (audio-format :accessor audio-format :initarg :audio-format :initform nil)
    (device :accessor device :initarg :device :initform nil)
    (number-of-samples :accessor number-of-samples :initarg :number-of-samples)
    (sample-rate  :accessor sample-rate :initarg :sample-rate)
    (number-of-channels :accessor number-of-channels :initarg :number-of-channels)
    (sample-size :accessor sample-size :initarg :sample-size)
    (data-position :accessor data-position :initarg :data-position)
    )
   )

(defmethod om-sound-file-name ((self om-sound))
   (filename self))

(defmethod om-sound-sample-rate ((self om-sound))
   (sample-rate self))

(defmethod om-sound-sample-size ((self om-sound))
   (sample-size self))

(defmethod om-sound-n-samples ((self om-sound))
   (number-of-samples self))

(defmethod om-sound-n-channels ((self om-sound))
   (number-of-channels self))

(defmethod om-sound-data-pos ((self om-sound))
   (data-position self))

(defmethod om-sound-format ((self om-sound))
   (audio-format self))


;;;=====================================================


;;;=============
;;; AIFF
;;;=============

(defun read-short (s)
  (let ((a (read-byte s))
        (b (read-byte s)))
    (logior (ash a 8) b)))

(defun read-24 (s)
  (let ((a (read-byte s))
        (b (read-byte s))
        (c (read-byte s)))
    (logior (ash a 16) (ash b 8) c)))

(defun read-long (s)
  (let ((a (read-byte s))
        (b (read-byte s))
        (c (read-byte s))
        (d (read-byte s)))
    (logior (ash a 24) (ash b 16) (ash c 8) d)))

(defun read-ostype (s)
  (let ((a (read-byte s))
        (b (read-byte s))
        (c (read-byte s))
        (d (read-byte s)))
    (coerce (list (code-char a) (code-char b) (code-char c) (code-char d)) 'string)))


;;;==================
;;; !!!!!!!!
(defun x80tod (ptr)
  (let* ((expon (logior (ash (logand (nth 0 ptr) #x7F) 8) (nth 1 ptr)))
         (hiMant (logior (ash (nth 2 ptr) 24) (ash (nth 3 ptr) 16) (ash (nth 4 ptr) 8) (nth 5 ptr)))
         (loMant (logior (ash (nth 6 ptr) 24) (ash (nth 7 ptr) 16) (ash (nth 8 ptr) 8) (nth 9 ptr)))
         (signe (logand (nth 0 ptr) #x80))
         rep)
    (if (and (= 0 expon) (= 0 hiMant) (= 0 loMant))
        (setf rep 0)
      (progn 
        (decf expon 16383)
        (setf rep (* (float himant) (expt 2 (decf expon 31))))
        (setf rep (+ rep (* (float himant) (expt 2 (decf expon 32)))))
        ))
    (unless (= 0 signe) (setf rep (- rep)))
    rep))

(defun read-extended (s)
  (let (10bytes rep)
    (setf 10bytes (loop for i from 1 to 10 collect (read-byte  s)))
    (setf rep (x80tod 10bytes))
    rep))




;;; read sample value (i bytes) in file
(defun read-aiff-sample (file i)
  (case i
    (1 (read-byte file)) 
    (2 (read-short file))
    (3 (read-24 file))))


(defun aiff-file-p (pathname)
  (let ((in (open pathname :element-type 'unsigned-byte)))
    (read-ostype in)
    (read-ostype in)
    (string-equal "AIFF" (read-ostype in))
    (close in)))


(defun aiff-stream-init (file)
  (file-position file 12))

(defun aiff-look-for-chunck (s ckname)
  (aiff-stream-init s)
  (loop while (and (not (stream-eofp s)) (ignore-errors (not (string-equal ckname (print (read-ostype s)))))) do
        (let ((sizeck (read-long s)))
          (loop for i from 1 to sizeck do
                (read-byte s))))
  (if (stream-eofp s)
    (om::om-beep-msg (format nil "File has not a ~D chunck~%" ckname))
    (read-long s)))

(defmethod om-sound-get-info-aiff ((self om-sound))
  (let ((in (open (filename self) :element-type 'unsigned-byte)))
    (read-ostype in)
    (read-ostype in)
    (cond
     ((not (string-equal "AIFF" (read-ostype in)))
      (om-message-dialog "Sorry this file is not an AIFF file")
      (setf (audio-format self) 'ERR)
      nil)
     (t
      (aiff-look-for-chunck in "COMM")
      (setf (number-of-channels self) (read-short in))
      (setf (number-of-samples  self) (read-long in))
      (setf (sample-size  self) (read-short in))
      (cond
       ((not (or (= (sample-size self) 24) (= (sample-size  self) 16) (= (sample-size  self) 8)))
           (om-message-dialog "Sorry i can read AIFF in 8, 16 or 24 bits only") 
           nil)
       (t 
        (setf (sample-rate  self) (read-extended in))
        (aiff-look-for-chunck in "SSND")
        (read-long in)
        (read-long in)
        (setf (data-position  self) (file-position in))
        ))))
    (close in)))


;;;===========
;;; WAVE 
;;;===========

(defun read-wav-sample (file i)
  (let* ((base 256)
        (rep (read-byte file)) 
        (list (list rep)))
    (loop for k from 2 to i do
          (let ((byte (read-byte file)))
            (setf rep (+ rep (* base byte)))
            (setf base (* base 256))))
    rep))


(defun wave-file-p (pathname)
   (let ((str (make-sequence 'string 4))
         (rep t))
   (with-open-file (in pathname :direction :input)
       (file-position in 0)
       (read-sequence str in)
       (setf rep (equal str "RIFF"))
       (when rep 
           (file-position in 8)
           (read-sequence str in)
           (setf rep (equal str "WAVE"))
           )
       )
   rep))


(defmethod om-sound-get-info-wave ((self om-sound))
   (let ((in (open (filename self) :element-type 'unsigned-byte)))
         (file-position in 22)
         (setf (number-of-channels self) (read-wav-sample in 2))
         (setf (sample-rate  self) (read-wav-sample in 4))
         (file-position in 34)
         (setf (sample-size  self) (read-wav-sample in 2))
         (file-position in 40)
       (setf (number-of-samples self) (round (read-wav-sample in 4) (* (/ (sample-size  self) 8) (number-of-channels self))))
       (setf (data-position self) 44)
       (close in)
       ))




;;;============================================
;;; GENERAL

(defun audio-file-type (pathname)
  (cond 
   ((wave-file-p pathname) 'WAVE)
   ((aiff-file-p pathname) 'AIFF)
   (t nil)))

(defun om-make-sound (class name)
  (let ((format (audio-file-type name))
        (snd nil))
    (when format
      (setf snd (make-instance class 
                  :filename name
                  :audio-format format))
      (om-sound-get-info snd))
    snd))

(defmethod om-sound-get-info ((self om-sound))
  (cond 
   ((equal (audio-format self) 'WAVE) (om-sound-get-info-wave self))
   ((equal (audio-format self) 'AIFF) (om-sound-get-info-aiff self))
   (t nil))
  self)


;;; !!! c'est pas du tout efficace. 
;;; en refaire une avec le fichier deja ouvert.
(defmethod om-read-sound-data ((self om-sound) position nbytes)
   (let ((in (open (filename self) :element-type 'unsigned-byte))
         (data 0))
       (file-position in position)
       (setf data (cond 
                   ((equal (audio-format self) 'WAVE) (read-wav-sample in nbytes))
                   ((equal (audio-format self) 'AIFF) (read-aiff-sample in nbytes))
                   (t 0)))
       (close in)
       data))



;;;===================================
;;; PICTURE A PARTIR DU SOUND FILE
;;;===================================

(defun wave-data (seq i nbytes)
  (cond
   ((= nbytes 1) (aref seq i))
   ((= nbytes 2) (logior (ash (aref seq i) 8) (aref seq (+ i 1))))
   ((= nbytes 3) (logior (ash (aref seq i) 16) (ash (aref seq (+ i 1)) 8) (aref seq (+ i 2))))
   (t 0)))

;;; returns max sample value in a window of size size
;;; of an input buffer 
;;; (pas utilisee)
(defun max-in-window (self input-stream size chanstep)
  (loop for i from 1 to size 
        maximize 
        (let ((val (if (= (sample-size self) 8) (read-byte input-stream) (read-short input-stream))))
          (file-position input-stream (+ (file-position input-stream) chanstep))
          val)))

;;;
(defun max-in-win (seq n numchan sizedata)
   (let* ((nbytes (round sizedata 8))
           (step (* numchan nbytes))
           (maxpos (- (length seq) 1)))
     (loop for i = (* n nbytes) then (+ i step) 
               while (< i maxpos)
            maximize (wave-data seq i nbytes))))


(defun read-wave-window (in winsize numchans samplesize)
     (let ((seq (make-array (* winsize numchans samplesize) :element-type 'unsigned-byte)))
       (read-sequence seq in :start 0 :end (* winsize numchans samplesize))
       seq
     ))
   
(defmethod cons-wave-pict ((self om-sound))
  (let* ((pixnum 10000)
         (sizedata (sample-size self))
         (numdat (number-of-samples  self))
         (numchan (number-of-channels  self))
         (step (floor numdat pixnum))
         (bytesizedata (round sizedata 8))
         (winsize (min step 20))
         
         (positivesize (round 128 numchan))
         (axes (loop for i from 1 to numchan
                     collect (+ positivesize (* 2 (- i 1) positivesize))))
         (channelstep (* (- numchan 1) bytesizedata))
         (deltax (round 256  numchan))
         firstpict win-seq 
         (last-pos (data-position self)))
    (om-with-cursor *om-wait-cursor*
      (setf firstpict 
            (om-record-pict *om-default-font2* (om-make-point pixnum 256) 
              (loop for i from 0 to (- numchan 1) do 
                    (om-draw-line 0 (+ (* i deltax) (round deltax 2)) pixnum (+ (* i deltax) (round deltax 2))))                            
              (let ((in (open (filename self) :element-type 'unsigned-byte)))
                (file-position in (data-position self))                             
                (loop for i from 0 to (- pixnum 1)
                      do (when (stream::stream-listen in)
                           (setf win-seq (read-wave-window in winsize numchan bytesizedata))
                           (setf last-pos (+ last-pos (* bytesizedata numchan step)))
                           (file-position in last-pos)
                           (loop for k from 1 to numchan do
                                 (let (new-point pixelpoint)
                                   (setf new-point (round (max-in-win win-seq (- k 1) numchan sizedata) (ash 1 (- sizedata 8))))
                                   ;(cond
                                   ; ((= sizedata 8) (setf new-point (max-in-win win-seq (- k 1) numchan sizedata )))
                                   ; ((= sizedata 16) (setf new-point (round (max-in-win win-seq (- k 1) numchan sizedata ) 256)))
                                   ; )
                                   (when (> new-point 128) (setf new-point (- new-point 256)))
                                   (setf pixelpoint (- 128 new-point))
                                   (om-draw-line i (+ (* (- k 1) deltax) (round pixelpoint numchan)) i  
                                                 (+ (* (- k 1) deltax) (round (+ pixelpoint (* 2 new-point)) numchan)))                                          
                                   ))))
                (close in))))
      firstpict)))


(defmethod om-sound-get-pict ((self om-sound))
  (cons-wave-pict self))





