;;;-*- Mode: Lisp; Package: ccl -*-
;*********************************************************************
;*                                                                   *
;*        Apple GL library   AGL                                     *
;*                                                                   *
;*********************************************************************
   ;* Author    : Alexander Repenning (alexander@agentsheets.com)    *
   ;*             http://www.agentsheets.com                         *
   ;* Copyright : (c) 1996-2002, AgentSheets Inc.                    *
   ;* Filename  : OpenGL for MCL.lisp                                *
   ;* Updated   : 12/06/02                                           *
   ;* Version   :                                                    *
   ;*   1.0     : 11/29/02 factored out from MacOpenGL               *
   ;*   1.1     : 12/04/02 decribe renderer functions                *
   ;* SW/HW     : PowerPC G4, MCL 5                                  *
   ;* Abstract  : AGL wrapper for OpenGL for MCL                     *
   ;*                                                                *
   ;******************************************************************

;; http://developer.apple.com/techpubs/macosx/Carbon/graphics/OpenGL/AGL_OpenGL/

(in-package :ccl)

(use-package :gl)

(eval-when (load compile eval)
  (add-to-shared-library-search-path "Apple;Carbon;Multimedia" :check))

;;*******************************
;; AGL low-level interface      *
;;*******************************

;; No need to export most of AGL since it should not be accessed directly

;; Types

(def-mactype :GLboolean (find-mactype :boolean))

(def-mactype :AGLDevice (find-mactype :GDhandle))

(def-mactype :GLint (find-mactype :long))

(def-mactype :GLenum (find-mactype :long))

(def-mactype :GLsizei (find-mactype :long))

(def-mactype :GLfloat (find-mactype :long))

(def-mactype :AGLPixelFormat (find-mactype :pointer))

(def-mactype :AGLContext (find-mactype :pointer))

(def-mactype :AGLDrawable (find-mactype :cGrafPtr))


;; Constants

(defconstant AGL_NONE 0)
(defconstant AGL_USE_GL 1   "support AGL rendering")
(defconstant AGL_BUFFER_SIZE 2   "depth of the color buffer")
(defconstant AGL_LEVEL 3   "level in plane stacking")
(defconstant AGL_RGBA 4 "RGB format")
(defconstant AGL_DOUBLEBUFFER 5 "double buffering supported")
(defconstant AGL_STEREO 6   "stereo buffering supported")
(defconstant AGL_AUX_BUFFERS 7   "number of aux buffers")
(defconstant AGL_RED_SIZE 8   "number of red component bits")
(defconstant AGL_GREEN_SIZE 9   "number of green component bits")
(defconstant AGL_BLUE_SIZE 10  "number of blue component bits")
(defconstant AGL_ALPHA_SIZE 11 "number of alpha bits")
(defconstant AGL_DEPTH_SIZE 12 "number of depth bits")
(defconstant AGL_STENCIL_SIZE 13  "number of stencil bits")
(defconstant AGL_ACCUM_RED_SIZE 14  "number of red accum bits")
(defconstant AGL_ACCUM_GREEN_SIZE 15  "number of green accum bits")
(defconstant AGL_ACCUM_BLUE_SIZE 16  "number of blue accum bits")
(defconstant AGL_ACCUM_ALPHA_SIZE 17  "number of alpha accum bits")

;** Extended attributes

(defconstant AGL_PIXEL_SIZE 50  "frame buffer bits per pixel")
(defconstant AGL_MINIMUM_POLICY 51  "never choose smaller buffers than requested")
(defconstant AGL_MAXIMUM_POLICY 52  "choose largest buffers of type requested")
(defconstant AGL_OFFSCREEN 53  "choose an off-screen capable renderer")
(defconstant AGL_FULLSCREEN 54  "choose a full-screen capable renderer")

;** Renderer management

(defconstant AGL_RENDERER_ID 70  "request renderer by ID")
(defconstant AGL_SINGLE_RENDERER 71  "choose a single renderer for all screens")
(defconstant AGL_NO_RECOVERY 72  "disable all failure recovery systems")
(defconstant AGL_ACCELERATED 73  "choose a hardware accelerated renderer")
(defconstant AGL_CLOSEST_POLICY 74  "choose the closest color buffer to request")
(defconstant AGL_ROBUST 75  "renderer does not need failure recovery")
(defconstant AGL_BACKING_STORE 76  "back buffer contents are valid after swap")
(defconstant AGL_MP_SAFE 78  "renderer is multi-processor safe")

;** Only for aglDescribePixelFormat

(defconstant AGL_WINDOW 80  "can be used to render to an onscreen window")
(defconstant AGL_MULTISCREEN 81  "single window can span multiple screens")
(defconstant AGL_VIRTUAL_SCREEN 82  "virtual screen number")
(defconstant AGL_COMPLIANT 83  "renderer is opengl compliant")

;** Hardware specific contstants

(defconstant ATI_FSAA_LEVEL 510 "ATI Radeon Full Screen Anti Aliasing")

;** Integer parameter names

(defconstant AGL_SWAP_RECT 200  "Enable or set the swap rectangle")
(defconstant AGL_BUFFER_RECT 202  "Enable or set the buffer rectangle")
(defconstant AGL_COLORMAP_TRACKING 210  "Enable or disable colormap tracking")
(defconstant AGL_COLORMAP_ENTRY 212  "Set a colormap entry to {index, r, g, b}")
(defconstant AGL_RASTERIZATION 220  "Enable or disable all rasterization")
(defconstant AGL_SWAP_INTERVAL 222  "0 -> Don't sync, n -> Sync every n retrace")
(defconstant AGL_STATE_VALIDATION 230  "Validate state for multi-screen functionality")



;; Functions

(define-entry-point "aglGetVersion" ((Major (:pointer :long)) (Minor (:pointer :long))) 
  :GLboolean)

(define-entry-point "aglChoosePixelFormat" ((gdev :AGLDevice) (ndev :GLint) (atttributes :pointer))
  :AGLPixelFormat)

(define-entry-point "aglDescribePixelFormat" ((Pix :AGLPixelFormat) (Attribute :GLint) (Value (:pointer :GLint))) 
   :GLBoolean)

(define-entry-point "aglDestroyPixelFormat" ((Pix :AGLPixelFormat)) nil)


(define-entry-point "aglCreateContext" ((pix :AGLPixelFormat) (share :AGLDevice)) 
  :AGLContext)

(define-entry-point "aglDestroyContext" ((context :AGLContext))
  :GLboolean)


(define-entry-point "aglDescribeRenderer" ((&rendererInfo :pointer) (prop :GLint) (&value (:pointer :GLint)))
  :GLboolean)

(define-entry-point "aglQueryRendererInfo" ((&gdev (:pointer :GLint)) (ndev :GLint))
  :pointer)


(define-entry-point "aglDestroyRendererInfo" ((:pointer :GLint)))

(define-entry-point "aglUpdateContext" ((context :AGLContext)) 
  :GLboolean)

(define-entry-point "aglSetDrawable" ((context :AGLContext) (draw :AGLDrawable)) 
  :GLBoolean)

(define-entry-point "aglGetDrawable" ((context :AGLContext)) 
  :AGLDrawable)

(define-entry-point "aglSetFullScreen" ((context :AGLContext) (width :GLsizei) (height :GLsizei) 
                                        (frequency :GLsizei) (device :GLint))
  :GLBoolean)

(define-entry-point "aglSetCurrentContext" ((context :AGLContext))
  :GLBoolean)


(define-entry-point "aglGetCurrentContext" () :AGLContext)

(define-entry-point "aglSwapBuffers" ((context :AGLContext)) nil)

(define-entry-point "aglUseFont" ((context :AGLContext) (fontID :GLint) (face :GLint)
  (size :GLint) (first :GLint) (count :GLint) (base :GLint))
  :GLBoolean)


(define-entry-point "aglGetError" () :glint)

(define-entry-point "aglErrorString" ((Code :GLint)) :string)


(define-entry-point "aglEnable" ((context :AGLContext) (pname :GLenum)) :GLboolean)


(define-entry-point "aglDisable" ((context :AGLContext) (pname :GLenum)) :GLboolean)


(define-entry-point "aglSetInteger" ((context :AGLContext) (pname :GLenum) (*params (:pointer :GLint))) 
  :GLboolean)

(define-entry-point "aglGetInteger" ((context :AGLContext) (pname :GLenum) (*params (:pointer :GLint))) 
  :GLboolean)


;;_____________________
;; AGL Error Handling  |
;;____________________/

(defconstant AGL_NO_ERROR 0 "No error")
(defconstant AGL_BAD_ATTRIBUTE 10000 "unknown attribute to get/set")
(defconstant AGL_BAD_PROPERTY 10001 "invalid renderer property")
(defconstant AGL_BAD_PIXELFMT 10002 "invalid pixel format")
(defconstant AGL_BAD_CONTEXT 10004 "invalid context specified")
(defconstant AGL_BAD_DRAWABLE 10005 "invalid drawable specified")
(defconstant AGL_INVALID_GDEV 10006 "invalid GDevice for attribute")
;;;(defconstant AGL_INVALID_OPERATION 10005 "invalid operation")
;;; (defconstant AGL_GL_ERROR 10006 "GL error occurred")

(defun AGL-ERROR-CHECK (&optional (Note "")) "
  Check if any AGL error has ocurred. Raise exception if needed."
  (let ((Code (aglGetError)))
    (unless (= Code AGL_NO_ERROR)
      (let ((Message (%get-cstring (aglerrorstring Code))))
        (error "AGL Error ~A: ~A ~A" Code Message Note)))))

;;_____________________
;; Gestalt Functions   |
;;____________________/

(defun OPENGL-VERSION () "
  out: Is-OpenGL-Available {boolean}.
       Mayor-Version {fixnum}.
       Minor-Version {fixnum}.
  Return the OpenGL version number if OpenGL is available." 
  (rlet ((&mayor :long)
         (&minor :long))
    (values
     (aglgetversion &mayor &minor)
     (%get-long &mayor)
     (%get-long &minor))))

;;_____________________
;; Helper Functions    |
;;____________________/

(defun INIT-ARRAY (&Array Type &rest Values)
  (declare (dynamic-extent Values))
  (let ((Index 0)
        (Size (record-field-length Type)))
    (dolist (Value Values &Array)
      (ecase Type
        (:long (%put-long &Array Value Index))
        (:single-float (%put-single-float &Array Value Index)))
      (incf Index Size))))


(defmacro WITH-RGBA-VECTOR (Vector (Red Green Blue Alpha) &body Forms)
  `(rlet ((,Vector (array :single-float 4)))
     (init-array ,Vector :single-float ,Red ,Green ,Blue ,Alpha)
     ,@Forms))


(defun MAKE-VECTOR (&rest Values) "
  in: &rest Values {list of {:long or :single-float}}.
  Create a vector initialized with <Values>.
  Vector is not automatically deallocated."
  (let* ((Index 0)
         (Size 0)
         (&Vector nil))
  (etypecase (first Values)
    (fixnum 
     (setq Size 4)
     (setq &Vector (#_NewPtr (* Size (length Values))))
     (dolist (Value Values &Vector)
       (%put-long &Vector Value Index)
       (incf Index Size)))
    ((or short-float double-float)
     (setq Size #.(record-field-length :single-float))
     (setq &Vector (#_NewPtr (* Size (length Values))))
     (dolist (Value Values &Vector)
       (%put-single-float &Vector Value Index)
       (incf Index Size))))))


(defun READ-NUMBER (Stream)
  (read-from-string
   (with-output-to-string (Out)
     (loop
       (let ((Char (read-char Stream nil nil)))
         (case Char
          ((#\} #\Space)
           (unread-char Char Stream)
           (return))
          (t (write-char Char Out))))))))


(defun VECTOR-READER (Stream Char)
  (declare (ignore Char))
  (let ((Numbers nil))
    (loop
      (let ((Char (read-char Stream nil #\])))
        (case Char
          (#\} (return `(make-vector ,@(reverse Numbers))))
          ((#\Space  #\Newline))
          (t (unread-char Char Stream)
             (push (read-number Stream) Numbers)))))))


(set-macro-character #\{ #'vector-reader)

;;; example: (defvar *Vector1* {3 4 5 5 77})


 
(defun COMPUTE-NORMAL (v1x v1y v1z v2x v2y v2z v3x v3y v3z) "
  in: v1x v1y v1z v2x v2y v2z v3x v3y v3z {float}.
  Compute and set the normal vector of a polygon defined by vertices v1, v2, v3
  The vector computed is NOT normalized."
  ;;; this code needs to be optimized since it will create a lot of garbage
  (let ((u1x (- v1x v2x))
        (u1y (- v1y v2y))
        (u1z (- v1z v2z))
        (u2x (- v2x v3x))
        (u2y (- v2y v3y))
        (u2z (- v2z v3z)))
    (glnormal3f
     (- (* u1y u2z) (* u2y u1z))
     (- (* u2x u1z) (* u1x u2z))
     (- (* u1x u2y) (* u2x u1y)))))

;_______________________________
; Full Screen Anti Aliasing     |
;______________________________/

(defun ATI-RADEON-ENABLE-FULL-SCREEN-ANTI-ALIASING (&optional (Level 4)) "
  in: &optional Level {int} default 4.
  Enable full screen anti aliasing for Radeon boards
  Keep level between 1 (no aliasing) and 4 (max aliasing)."
  ;; ATI Radeon Extension: Enable full screen anti aliasing
  ;; Great stuff but this will degrade performance considerably
  (rlet ((&myFSAALevel :long Level))
    (aglSetInteger (aglGetCurrentContext) ATI_FSAA_LEVEL &myFSAALevel)
    ;; This will return an error even on a Radeon board
    ;; but it still works
    (aglGetError)))

;; (ati-radeon-enable-full-screen-anti-aliasing)

#| Examples:

(dotimes (i 10)
  (dotimes (Level 4)
    (ati-radeon-enable-full-screen-anti-aliasing (1+ Level))
    (view-draw-contents *Window*)
    (format t "~%level=~A, frame rate=~A [fps]" (1+ Level) (frame-rate *Window*))))

|#



