;;===========================================================================
;OM API 
;Multiplatform API for OpenMusic
;Macintosh version (Digitool Macintosh Common Lisp - MCL)
;
;Copyright (C) 2004 IRCAM-Centre Georges Pompidou, Paris, France.
; 
;This program is free software; you can redistribute it and/or
;modify it under the terms of the GNU General Public License
;as published by the Free Software Foundation; either version 2
;of the License, or (at your option) any later version.
;
;See file LICENSE for further informations on licensing terms.
;
;This program is distributed in the hope that it will be useful,
;but WITHOUT ANY WARRANTY; without even the implied warranty of
;MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;GNU General Public License for more details.
;
;You should have received a copy of the GNU General Public License
;along with this program; if not, write to the Free Software
;Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
;
;Authors: Jean Bresson and Augusto Agon
;;===========================================================================

;;===========================================================================
;DocFile
; CONNECTION SDIFEDIT USING APPLEEVENTS
;;===========================================================================

(in-package :ccl)

;;;======================================================
;;; APPLE EVENTS AVEC SDIFEDIT

(defun my-launch-se (filename command-line assoc-sound sbox)
   (if (find-process :|OMSE|)
     (send-cl-to-se command-line assoc-sound sbox)
     (rlet ((fsspec :FSSpec))
       (rlet ((pb :launchParamBlockRec
                  :launchBlockID #$extendedBlock
                  :launchEPBLength #$extendedBlockLen
                  :launchControlFlags (+ #$launchContinue #$launchNoFileFlags #$launchUseMinimum)
                  :launchAppSpec fsspec
                  :launchAppParameters (%null-ptr)))
         (with-pstrs ((name (mac-namestring (probe-file filename))))
           (#_FSMakeFSSpec 0 0 name fsspec))
         (without-interrupts             ; don't let the app start up until we've sent the AppleEvent
             (eql 0 (#_LaunchApplication pb)))
         (send-cl-to-se command-line assoc-sound sbox)
         )))
   t)


(defun create-se-cl (the-desc the-target string assoc-sound &rest create-keywords)
   (declare (dynamic-extent create-keywords))
   (apply 'create-appleevent the-desc :|SeOm| (if assoc-sound :|OASE| :|OPSE|) the-target
          create-keywords)
   (ae-put-parameter-char the-desc #$keyDirectObject string)
   (when assoc-sound
     (ae-put-parameter-char the-desc :|AAAF| assoc-sound)))

(defun send-cl-to-se (command-string assoc-sound sbox)
   (let (id)
     (with-aedescs (appleevent reply target)
       (create-signature-target target :|OMSE|)
       (create-se-cl appleevent target command-string assoc-sound)
       (send-appleevent appleevent reply :reply-mode :wait-reply)
       (setf id  (ae-get-parameter-char reply #$keyDirectObject nil))
       (setf (gethash sbox oa::*se-ht*) id))))

(defun send-window-select-to-se (str)
   (with-aedescs (appleevent reply target)
     (create-signature-target target :|OMSE|)
     (create-appleevent appleevent :|SeOm| :|WSSE| target)
     (ae-put-parameter-char appleevent #$keyDirectObject str)
     (send-appleevent appleevent reply :reply-mode :no-reply)))

(defun send-close-app-to-se (str)
   (with-aedescs (appleevent reply target)
     (create-signature-target target :|OMSE|)
     (create-appleevent appleevent :|SeOm| :|QTSE| target)
     ;(ae-put-parameter-char appleevent #$keyDirectObject str)
     (send-appleevent appleevent reply :reply-mode :no-reply)))

(defun launch-SE-with-an-event (sdifedit sdiffile assoc-sound sbox)
   (let ((seRunning (find-process :|OMSE|)))
     (if seRunning (send-cl-to-se sdiffile assoc-sound sbox)
         (let ((file (mac-namestring sdifedit)))
           (clrhash oa::*se-ht*)
           (my-launch-se file sdiffile assoc-sound sbox)))))
  

;;;======================================================

(in-package :oa)


(export '(
          om-run-sdifedit
          om-sdifedit-open-p
          om-select-sdifedit-window
          ) :om-api)


(defvar *sdif-edit* nil)
(setf *sdif-edit* "cl:image;MacOS;Externs;SDIF-Edit")



(defvar *se-ht* (make-hash-table))
;;;(print-hash-table *se-ht*)
;;;(clrhash oa::*se-ht*)

(defun reset-sdifedit ()
  (clrhash oa::*se-ht*))

(defun om-run-sdifedit (ref filepath &optional audio)
   (if *sdif-edit*
     (ccl::launch-SE-with-an-event *sdif-edit* filepath audio ref)
     (om-message-dialog "SDIF-Edit not available !"))
   )

;;;(defun om-close-sdifedit (objref)
;;;   (let ((pid (gethash objref *se-ht*)))
;;;     (when pid
;;;         (mp::process-kill pid)
;;;         )))

(defun om-sdifedit-open-p (objref)
   (gethash objref *se-ht*))

(defun om-select-sdifedit-window (objref)
  (if (ccl::find-process :|OMSE|)
   (ccl::send-window-select-to-se (gethash objref *se-ht*))
   (om-run-sdifedit objref (om::filepathname (om::value objref)))
   ))




;=======================
;AE handlers

(defmethod ccl::close-sdif-editor ((a application) theAppleEvent reply handlerRefcon)
  (declare (ignore handlerRefcon reply ))
  (let ((string (ccl::ae-get-parameter-char theAppleEvent #$keyDirectObject nil)))
    (maphash #'(lambda (key val)
                 (if (string-equal val string)
                   (REMHASH key oa::*se-ht*))) oa::*se-ht*)
))

;;; SDIFEDIT ENVOIE QUAND IL SE FERME
(install-appleevent-handler :|SeOm| :|CLED| #'ccl::close-sdif-editor)