(in-package :om-api)

(export '(om-init-motion-functions
          om-selection-rectangle
          om-movable-line
          om-movable-rectangle

          om-new-movable-object
          om-update-movable-object
          om-erase-movable-object
          om-get-rect-movable-object
          om-init-point-movable-object

          om-cursor-line
          om-view-cursor-play
          om-new-movable-cursor
          om-update-movable-cursor 
          om-erase-movable-cursor

          ) :om-api)
;=======================
;utilities
;=======================
(defmethod pane-geometry ((pane pinboard-object))
  (multiple-value-bind (x y) (pinboard-pane-position pane)
     (multiple-value-bind (w h) (pinboard-pane-size pane)
       (values x y w h))))

(defun rectangle-intersection-p (rect1 rect2)
  (destructuring-bind (x y w h) rect1
    (destructuring-bind (x1 y1 w1 h1) rect2
      (and (< x (+ x1 w1))
           (< x1 (+ x w))
           (< y (+ y1 h1))
           (< y1 (+ y h))))))


(defun valid-image-p (image) ;is the cocoa pointer for the image exists ?
  (and image
       (when-let (representation (gp::image-representation image))
         (slot-value representation 'image))))




;=======================
;the basic class of the movable object
;=======================
(defclass movable-object (pinboard-object)
  ((restore-image :initform nil :accessor restore-image) ;the image
   (geometry :initform '(0 0)))) ;used by the system

;the generic for drawing the object : the basic method does nothing, it must be specialized

(defmethod set-motion-object-position ((obj movable-object) x y)
  (setf (capi::pinboard-pane-position obj) (values x  y))
  )

(defmethod get-motion-object-position ((obj movable-object))
  (capi::pinboard-pane-position obj) 
  )

(defmethod get-motion-object-size ((obj movable-object))
  (capi::pinboard-pane-size obj)
  )

(defmethod set-motion-object-size ((obj movable-object) x y)
  (setf (capi::pinboard-pane-size obj) (values x  y))
)



(defgeneric draw-movable-object (pinboard object &optional x y w h))

(defmethod draw-movable-object (pinboard (object movable-object) &optional x y w h) (declare (ignore pinboard object x y w h)) nil)


;the four steps in the capi method specialized for the class of "movable-object". NOrmally this is called into the "mouse-motion-callback" method

(defmethod draw-pinboard-object (pane (self movable-object) &key x y width height erase)
  (multiple-value-bind (ox oy ow oh) (pane-geometry self)
    (when (or erase (or (not (and x y width height)) (rectangle-intersection-p (list x y width height) (list ox oy ow oh))))
      (when (slot-value self 'restore-image) ;step 1 and 2 are naturally conditionned by the fact an image already exists
        (draw-movable-object-image pane self) ;STEP 1
        (free-movable-object-image pane self)) ;STEP 2
      
      (unless erase ;in an erase process STEP 3 & 4 are not used
        (when (and (> ow 0) (> oh 0))
          (make-movable-object-image pane self ox oy ow oh) ;STEP 3om-create-motion-obj
          (draw-movable-object pane self x y width height) ;STEP 4
          )))))

;onely usefull in some very specific situations

(defmethod erase-movable-object (pane (self movable-object))
  (draw-pinboard-object pane self :erase t))

;=======================
;-----MAIN PROCEDURE
;=======================

;-----MAIN PROCEDURE

;STEP 1 : draw the stored image and restore the port. If the type of the object is "selection" it works with four images rather than one

(defmethod draw-movable-object-image (pane obj) 
 (with-slots (restore-image geometry) obj
   (with-geometry pane
        (let* ((w (port-width pane))
               (h (port-height pane))
               (left (if (capi::simple-pane-horizontal-scroll pane) (or %scroll-x% 0) 0))
               (top (if (capi::simple-pane-vertical-scroll pane) (or %scroll-y% 0) 0)))
          (flet ((draw (img a b c d)
                   (when (and a b c d 
                              (valid-image-p img)
                              (< a (+ left w)) (>= c left) (< b (+ top h)) (>= d top))
                     (draw-image pane img a b))))
           (destructuring-bind (x y w h) geometry
                   (draw restore-image x y (+ x w) (+ y h)) ))))))





;STEP 2 : destroy the stored image. If the type of the object is "selection" it works with four images rather than one

(defun free-movable-object-image (pane obj &optional (restore-Image nil value-p))
  (when-let (restore-image (or restore-image (slot-value obj 'restore-image)))
    (if (listp restore-image)
      (loop for image in restore-image when (valid-image-p image) do (free-Image pane image))
      (when (valid-image-p restore-image) (free-Image pane restore-image)))
    (unless value-p (setf (slot-value obj 'restore-image) nil))))


;STEP 3 : make a new image of the port region before drawing the object itself. If the type of the object is "selection" it works with four images rather than one

(defmethod make-movable-object-image (pane obj x y w h)  ;la nouvelle gomtrie de l'objet
  (with-slots (restore-image geometry ) obj
    (with-geometry pane
      (let* ((width (port-width pane))
             (height (port-height pane))
             (left (if (capi::simple-pane-horizontal-scroll pane) (or %scroll-x% 0) 0))
             (top (if (capi::simple-pane-vertical-scroll pane) (or %scroll-y% 0) 0))
             (right (+ left width))
             (bottom (+ top height))
             (x4 (+ x w))
             (y4 (+ y h))
             images geometries)
        (flet ((test (a b c d) (and (< a right) (> b left) (< c bottom) (> d top)))
               (make-image (a b c d)
                 (let (ox oy ow oh) ;l'ancienne gomtrie de l'objet
                   (setf ox (max left a) oy (max top c) ow (- (min right b) ox) oh (- (min bottom d) oy))
                   (values (when (and (> ow 0) (> oh 0))
                             (make-image-from-port pane ox  oy ow oh))
                           (list ox oy ow oh)))))
         (when (test x x4 y y4) 
           (multiple-value-bind (image geom) (make-image x x4 y y4)
             (setf restore-image image 
                   geometry geom))))))))

;================
;API
;================
;THE PANE 

(defclass om-movable-view () 
  ((current-movable-object :initform nil :accessor current-movable-object)
   (initial-position :initform nil :accessor initial-position)
   (last-position :initform nil :accessor last-position)
   (click-motion-action :initform nil :accessor click-motion-action)
   (release-motion-action :initform nil :accessor release-motion-action)))


(defmethod om-init-motion-functions ((self om-movable-view) click-motion release)
  (setf (click-motion-action self) click-motion)
  (setf (release-motion-action self) release))

(defmethod om-click-release-handler :after ((self om-movable-view) pos)
 (when (release-motion-action self)
    (funcall (release-motion-action self) self pos))
  (setf (release-motion-action self) nil)
  (setf (click-motion-action self) nil))

(defmethod om-click-motion-handler :after ((self om-movable-view) pos)
  (when (click-motion-action self)
      (funcall (click-motion-action self) self pos)))

(defmethod om-editor-has-movable-p ((self om-movable-view))
  (current-movable-object self))


(defmethod om-new-movable-object ((self om-movable-view) x y w h class)
  (with-slots (current-movable-object  initial-position) self
    (setf current-movable-object (make-instance class)
          initial-position (list x y))
    (set-motion-object-position current-movable-object x y)
    (set-motion-object-size current-movable-object w h)))     

(defmethod om-update-movable-object ((self om-movable-view)  x y w h)
  (with-slots (current-movable-object movable-object-type initial-position ) self
    (when current-movable-object
      (set-motion-object-size current-movable-object  (max 4 w) (max 4 h))
      (set-motion-object-position current-movable-object  x y)
      (capi::draw-pinboard-object self current-movable-object x y w h))))


(defmethod om-erase-movable-object ((self om-movable-view))
  (with-slots (current-movable-object) self
    (when current-movable-object
    (when (oa::restore-image current-movable-object)
      (oa::draw-movable-object-image self current-movable-object)
      (oa::free-movable-object-image self current-movable-object)
      (setf current-movable-object nil)))))


(defmethod om-init-point-movable-object ((self om-movable-view))
  (with-slots (current-movable-object initial-position) self
    (when current-movable-object 
      (destructuring-bind (ox oy) initial-position
        (list ox oy)))))

(defmethod om-get-rect-movable-object ((self om-movable-view) x y)
   (with-slots (current-movable-object initial-position) self
     (when current-movable-object 
      (multiple-value-bind (w h) (get-motion-object-size current-movable-object)
        (destructuring-bind (ox oy) initial-position
           (let* ((rect (om-pts-to-rect (om-make-point ox oy) (om-make-point x y)))
                  (rx  (om-rect-left rect))
                  (ry (om-rect-top rect))
                  (rw (om-rect-w rect))  
                  (rh (om-rect-h rect)))
             (list rx ry rw rh)))))))


;SELECTION RECTANGLE

(defclass om-selection-rectangle (movable-object) ())

(defmethod draw-movable-object (self (object om-selection-rectangle) &optional x y w h)
  (declare (ignore x y w h))
  (multiple-value-bind (x y w h) (pane-geometry object)
    ;(draw-rectangle self (1+ x) (1+ y) (- w 3) (- h 3) :filled nil :thickness 2 :dashed t :dash '(2 2))
    (draw-rectangle self (1+ x) (1+ y) (- w 3) (- h 3) :filled t :foreground *om-select-color-alpha*)
   (draw-rectangle self (1+ x) (1+ y) (- w 3) (- h 3) :filled nil :foreground *om-gray-color* :thickness 1)
    ))

; RECTANGLE

(defclass om-movable-rectangle (movable-object) ())

(defmethod draw-movable-object (self (object om-movable-rectangle) &optional x y w h)
  (declare (ignore x y w h))
  (multiple-value-bind (x y w h) (pane-geometry object)
    (draw-rectangle self x y (- w 1) (- h 1) :filled nil :thickness 1 :dashed t :dash '(1 1))
    ))

; LINE
(defclass om-movable-line (movable-object) ())

(defmethod draw-movable-object (self (object om-movable-line) &optional x y w h)
  (declare (ignore x y w h))
  (multiple-value-bind (x y w h) (pane-geometry object)
    (draw-line self (car (initial-position self)) (second (initial-position self))
                (if (> (car (initial-position self)) x)
                    (- (car (initial-position self)) w -1)
                    (+ w  x -1))
                (if (> (second (initial-position self)) y)
                    (- (second (initial-position self)) h -1)
                    (+ y h -1))  :thickness 1)))


;play cursor


(defclass om-view-cursor-play () 
   ((movable-cursor :initform nil :accessor movable-cursor)))

(defmethod om-new-movable-cursor ((self om-view-cursor-play) x y w h class)
  (with-slots (movable-cursor  initial-position) self
    (setf movable-cursor (make-instance class)
          initial-position (list x y))
    (set-motion-object-position movable-cursor x y)
    (set-motion-object-size movable-cursor w h)))    

(defmethod om-update-movable-cursor ((self om-view-cursor-play)  x y w h)
  (capi:apply-in-pane-process self 
                              #'(lambda ()
                                  (with-slots (movable-cursor ) self
                                    (when movable-cursor
                                      (set-motion-object-size movable-cursor  (max 4 w) (max 4 h))
                                      (set-motion-object-position movable-cursor  x y)
                                      (capi::draw-pinboard-object self movable-cursor x y w h))))))


(defmethod om-erase-movable-cursor ((self om-view-cursor-play))
  (with-slots (movable-cursor) self
    (when movable-cursor
    (when (oa::restore-image movable-cursor)
      (oa::draw-movable-object-image self movable-cursor)
      (oa::free-movable-object-image self movable-cursor)
      (setf movable-cursor nil)))))


(defclass om-cursor-line (movable-object) ())

(defmethod draw-movable-object (self (object om-cursor-line) &optional x y w h)
  (declare (ignore x y w h))
  (multiple-value-bind (x y w h) (pane-geometry object)
    (draw-line self x y (+ w  x -1) (+ h y -1) :foreground :red :thickness 1 :dashed t :dash '(1 1))))

