;;===========================================================================
;OM API 
;Multiplatform API for OpenMusic
;Macintosh version (Digitool Macintosh Common Lisp - MCL)
;
;Copyright (C) 2004 IRCAM-Centre Georges Pompidou, Paris, France.
; 
;This program is free software; you can redistribute it and/or
;modify it under the terms of the GNU General Public License
;as published by the Free Software Foundation; either version 2
;of the License, or (at your option) any later version.
;
;See file LICENSE for further informations on licensing terms.
;
;This program is distributed in the hope that it will be useful,
;but WITHOUT ANY WARRANTY; without even the implied warranty of
;MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;GNU General Public License for more details.
;
;You should have received a copy of the GNU General Public License
;along with this program; if not, write to the Free Software
;Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
;
;Authors: Jean Bresson and Augusto Agon
;;===========================================================================

;;===========================================================================
;DocFile
; SPECIAL OPENMUSIC FUNCTIONS
;;===========================================================================


(in-package :om-api)



(export '( 
          om-show-documentation
          om-get-workspace-initsize
          om-get-workspace-initpos
          ) :om-api)

(defvar *doc-buffer* nil)
(defvar *doc-window* nil)

(defclass doc-win (capi::interface) ())

(defun show-doc-window ()
   (unless *doc-window* 
     (setf *doc-window* (make-instance 'doc-win 
                                       :title "Documentation Window"
                                       :name "OM Doc Window"
                                       :best-width 400
                                       :best-height 300
                                       :layout
                                       (make-instance 'simple-layout
                                                      :description 
                                                      (list (make-instance 'capi::collector-pane 
                                                                           :buffer (buffer *doc-buffer*)
                                                                           :font (gp::make-font-description :name "Arial" :size 13)
                                                                           :width 300
                                                                           )))
                                       :destroy-callback #'(lambda (interface) (setf *doc-window* nil))))
     (display *doc-window*))
   (capi::find-interface 'doc-win :name "OM Doc Window")
   )

(defun om-show-documentation (symbol) 
  (unless *doc-buffer* 
    (setf *doc-buffer* (om-make-buffer)))
  (om-buffer-delete *doc-buffer*)
  (if (not (symbolp symbol))
       (om-beep)
    (let ((*print-case* :downcase)
          (doc-list (list symbol)))
      (om-buffer-insert *doc-buffer* (string-upcase (format nil "~%~S~%" (intern symbol))))
      (if (fboundp symbol)
          ;;; FUNCTION OR GEN-FUN
          (let ((fun (fdefinition symbol))
                (args ()))
            (om-buffer-insert *doc-buffer* "    Arguments: ")
            (dolist (arg (arglist symbol))
              ;;; ce serait bien de mettre les keywords en italic
              (if (memq arg lambda-list-keywords) 
                  (setf args (append args (list arg)))
                (setf args (append args (list arg)))))
            (setf doc-list (append doc-list (list args)))
            (if (null args)
                (om-buffer-insert *doc-buffer* "()")
              (loop for arg in (second doc-list) do (om-buffer-insert *doc-buffer* (format nil " ~a" arg))))
            
            (om-buffer-insert *doc-buffer* (format nil "~%"))
            (if (typep fun 'generic-function)
                (setf doc-list (append doc-list (list 'GENERIC-FUNCTION)))
              (setf doc-list (append doc-list (list 'FUNCTION)))
              )
            (om-buffer-insert *doc-buffer* (format nil "    [~S]~%~%" (third doc-list)))
            (setf doc-list (append doc-list (list (system::function-documentation symbol))))
            (om-buffer-insert *doc-buffer* (fourth doc-list))
            )
        (let ((class (find-class symbol nil)))
          (if class 
              (progn 
               (setf doc-list (append doc-list (list nil 'type (system::class-documentation class))))
               (om-buffer-insert *doc-buffer* (fourth doc-list)))
            (om-buffer-insert *doc-buffer* (format nil "Function or class ~S could not be found" symbol)))
          ))
            
      (show-doc-window)
      (setf (capi::interface-title *doc-window*) 
            (concatenate 'string "Documentation for " (string-upcase (string (intern symbol)))))
      )))

; (system::class-documentation (find-class 'om::note))
; (om-show-documentation 'om::om+)



;=======================
; find and edit source code 
;=======================


(defun show-definitions-dialog (symb deflist)
  (let ((win (om-make-window 'om-dialog :size (om-make-point 400 300)
                             :window-title (concatenate 'string "Definitions of " (string-upcase (string symb))))))
    (om-add-subviews win
                     (om-make-dialog-item 'om-single-item-list
                                          (om-make-point 10 10)
                                          (om-make-point 380 280)
                                          ""
                                          :callback-type '(:collection)
                                          :test-function 'string-equal
                                          :range (loop for item in deflist collect (format nil "~A" (car item)))
                                          :action-callback #'(lambda (list)
                                                               (let ((file (cadr (nth (capi::choice-selection list) deflist))))
                                                                 (if (and (pathnamep file) (probe-file file))
                                                                     (om-open-new-text-editor file)
                                                                   (progn (om-beep) 
                                                                     (print (format nil "File ~A not found" file)))
                                                                   )))
                                          ))
    (om-select-window win)))


(defun om-edit-definition (symbol)
  (let ((definitions 
         (dspec:name-definition-locations dspec:*dspec-classes* symbol)))
    (cond 
     ((null definitions)
      (om-beep)
      (print (concatenate 'string "No definition found for " (string-upcase (string symbol)))))
     ((= (length definitions) 1)
      (let ((file (cadr (car definitions))))
        (if (and (pathnamep file) (probe-file file))
            (om-open-new-text-editor file)
          (progn (om-beep) (print (format nil "File ~A not found" file)))
          )))
     (definitions
      (show-definitions-dialog symbol definitions))
     )))

; (om-edit-definition 'om::om+)
; (setf dspec::*active-finders* (list :internal dspec::*active-finders*))
; (dspec:find-name-locations dspec:*dspec-classes* 'om::om+)
; (dspec:name-definition-locations dspec:*dspec-classes* 'om::om+)
; (*active-finders*)
; save-tags-database

(defun om-get-workspace-initpos ()
  (om-make-point 5 42))

(defun om-get-workspace-initsize ()
  (let ((screen (capi:convert-to-screen) ))
    (om-make-point (round (* 1/3 (capi::screen-width screen)))
                   (round (* 5/6 (capi::screen-height screen))))))



