

(unless (find-package :OM-API)
  (defpackage "OM-API"
    (:nicknames "OA")
    (:use "COMMON-LISP" "CL-USER" "CAPI" "LISPWORKS" "GP"))
  )

(in-package :oa)

;;;================= 
;;; TEXT BUFFER ====
;;;=================

(export '(
          om-make-buffer
          om-copy-buffer
          om-buffer-delete
          om-kill-buffer
          om-buffer-insert
          om-buffer-insert-newline
          om-buffer-text
          om-buffer-char
          om-buffer-size
          om-buffer-substring
          om-lines-in-buffer
          om-buffer-line-start   
          om-buffer-line-end
          om-buffer-skip-forward
          om-buffer-insert-file
          om-buffer-write-file
          ) :om-api)


;;; buffer : un editor::buffer
;;; window : un editeur sur laquelle le buffer est utilis (ne pas effacer)
;;; killed? (plus utilise) un flag pour voir si le buffer est deja detruit (t)
;;;         ou en attente si il est encore sur un editeur (:flagged)
(defclass ombuffer () 
  ((buffer :accessor buffer :initarg :buffer :initform nil)
   (window :accessor window :initform nil)
   (killed? :accessor killed? :initform nil)))

(defun om-make-buffer ()
   ; (print (list "create buffer" (buffer b)))
  (make-instance 'ombuffer :buffer (editor::make-buffer (string (gensym)))))
   

; PLUS UTILISE...
; En declarant les objets :
;    (hcl::flag-special-free-action b)
; ...
; on peut leur appliquer une action au moment du GC
;
;(defun kill-buffers (object)
;  (when (equal (type-of object) 'ombuffer)
;    (print (list "buffer garbage" (buffer object)))
;    (editor::delete-buffer (buffer object))
;    ))
;
; (hcl::add-special-free-action 'kill-buffers)

(defun om-kill-buffer (ombuffer)
  (print (list "kill buffer explicit" (buffer ombuffer)))
  (editor::delete-buffer (buffer ombuffer))
  (editor::kill-buffer-no-confirm (buffer ombuffer)))

  ;(if (window ombuffer) (setf (killed? ombuffer) :flagged)
  ;  (progn (editor::delete-buffer (buffer ombuffer))
  ;    (setf (killed? ombuffer) t)))

(defun om-copy-buffer (ombuffer)
  (let ((newbuffer (om-make-buffer)))
    (when (and ombuffer (buffer ombuffer))
      (editor::set-buffer-contents (buffer newbuffer)
                                   (editor::points-to-string (editor::buffers-start (buffer ombuffer))
                                                             (editor::buffers-end (buffer ombuffer)))))
    newbuffer))

(defun om-buffer-insert (ombuffer string &optional (position nil))
  (let ((textbuffer (buffer ombuffer)))
    (editor::use-buffer textbuffer
      (setf (editor::face-italic-p editor::*current-font*) t)
      (if position
          (editor::with-point ((p (editor:buffers-start textbuffer)))
            (editor::character-offset p position)
            (editor::insert-string p string))
        (editor::insert-string (editor:buffer-point textbuffer) string)))))

(defun om-buffer-insert-newline (ombuffer &optional (position nil))
    (let ((textbuffer (buffer ombuffer)))
      (editor::use-buffer textbuffer
        (if position
            (editor::with-point ((p (editor:buffers-start textbuffer)))
              (editor::character-offset p position)
              (editor::newline p))
          (editor::newline (editor:buffer-point textbuffer))))))

(defun om-buffer-text (ombuffer)
      (let ((textbuffer (buffer ombuffer)))
        (editor::use-buffer textbuffer
          (editor:points-to-string (editor:buffers-start textbuffer) 
                                   (editor:buffers-end textbuffer)))))
      
(defun om-buffer-substring (ombuffer from &optional to)  
      (let ((textbuffer (buffer ombuffer)))
        (editor::use-buffer textbuffer
          (editor::with-point ((p1 (editor:buffers-start textbuffer))
                               (p2 (if to (editor:buffers-start textbuffer)
                                       (editor:buffers-end textbuffer))))
            (editor::character-offset p1 from)
            (when to (editor::character-offset p2 to))
            (editor:points-to-string p1 p2)))))

(defun om-buffer-char (ombuffer &optional pos)
  (let ((textbuffer (buffer ombuffer)))
    (editor::use-buffer textbuffer
      (editor::with-point ((p (if pos (editor:buffers-start textbuffer)
                                (editor:buffer-point textbuffer))))
        (editor::character-at p pos)))))

(defun om-buffer-size (ombuffer)
  (let ((textbuffer (buffer ombuffer)))
    (editor::use-buffer textbuffer
      (editor::count-characters (editor::buffers-start textbuffer)
                                (editor::buffers-end textbuffer)))))
  
(defun om-lines-in-buffer (ombuffer)
  (let ((textbuffer (buffer ombuffer)))
    (editor::use-buffer textbuffer
      (+ 1 (editor::count-lines (editor::buffers-start textbuffer)
                                (editor::buffers-end textbuffer))))))
  

(defun om-buffer-delete (ombuffer &optional start end)
  (let ((textbuffer (buffer ombuffer)))
    (if start
        (editor::use-buffer textbuffer
          (editor::with-point ((p1 (editor::buffers-start textbuffer)) 
                               (p2 (if end (editor::buffers-start textbuffer)
                                     (editor::buffers-end textbuffer))))
            (editor::character-offset p1 start)
            (when end (editor::character-offset p2 end))
            (editor::delete-between-points p1 p2)))
      (editor::clear-buffer textbuffer))))
    


;;; position de la fin de la ligne courante (ou au point start)
(defun om-buffer-line-end (ombuffer &optional pos)
  (let ((textbuffer (buffer ombuffer)))
    (editor::use-buffer textbuffer
      (editor::with-point ((p (if pos (editor::buffers-start textbuffer)
                                (editor:buffer-point textbuffer))))
        (when pos (editor::character-offset p pos))
        (editor::next-newline p)
        ))))

;;; position du debut de la ligne courante (ou au point start)
(defun om-buffer-line-start (ombuffer &optional pos)
  (let ((textbuffer (buffer ombuffer)))
    (editor::use-buffer textbuffer
      (editor::with-point ((p (if pos (editor::buffers-start textbuffer)
                                (editor:buffer-point textbuffer))))
        (when pos (editor::character-offset p pos))
        (if (editor::same-line-p p (editor::buffers-start textbuffer))
            0
          (if (equal (editor::character-at p 0) #\Newline) pos
            (+ 1 (editor::find-point-offset textbuffer (editor::find-previous-character p #\Newline))))
          )))))

(defun om-buffer-skip-forward (ombuffer &optional start end)
  (let ((textbuffer (buffer ombuffer)))
    (editor::use-buffer textbuffer
      (editor::with-point ((p (if start (editor::buffers-start textbuffer)
                                (editor:buffer-point textbuffer)))
                           (limit (if end (editor::buffers-start textbuffer)
                                    (editor::buffers-end textbuffer))))
        (when start (editor::character-offset p start))
        (when end (editor::character-offset limit end))
        (loop while (and (editor::blank-line-p p) (editor::point< p limit)) do
                 (editor::move-point-to-offset p (editor::next-newline p)))
        (editor::find-point-offset textbuffer p)))
    ))
;; pour sauter aussi les comments : SKIP-LISP-READER-WHITESPACE

;; tests
;(setf buf (om-make-buffer))
;(om-buffer-insert buf "123456789")
;(om-buffer-insert buf "  ")
;(om-buffer-insert-newline buf)
;(om-buffer-text buf)
;(om-buffer-char buf 4)
;(om-buffer-delete buf)
;(om-buffer-size buf)
;(om-lines-in-buffer buf)
;(setf b2 (om-copy-buffer buf))
;(om-buffer-substring buf 6 6)
;(editor::kill-buffer-no-confirm (buffer buf))
;;(editor::delete-buffer (buffer buf))

(defun test-list-of-lines (buffer)
  (when buffer
    (let ((numlines (om-lines-in-buffer buffer))
          (pos 0) rep)
      (loop for i from 1 to numlines do
            (let ((start (om-buffer-line-start buffer pos))
                  (end (om-buffer-line-end buffer pos)))
              (push (om-buffer-substring buffer start end) rep)
              (print (list pos start end))
              (setf pos (+ 1 end))))
      (reverse rep))))

; (test-list-of-lines buf)

;;; DOC fonctions du package EDITOR
; (editor::buffers-start buf)  --> renvoie un point correspondant au debut du buffer
; (editor::current-point buf) --> (verif) renvoir lepoint courant (tous buffers confondus)
; (editor::buffer-point buf) --> renvoie le point courant sur un buffer
; (editor::buffer-end point) -> deplace point a la fin du buffer
; (editor::buffer-start point) -> deplace point au debut du debut... buffer
; (editor::delete-buffer buf &optional force-reset) --> efface un buffer
; (editor::clear-buffer) --> ?
; (editor::delete-between-points start end) --> efface entre les deux points
; (editor::delete-characters point &optional (n 1) --> efface n caracteres apres point
; (editor::count-lines beg end) --> compte le nombre de lignes dans le region
; (editor::line-end point) (editor::line-start point) --> ramene point au debut ou a la fin de la ligne
; (editor::list-lines point num) --> ?
; (editor::newline point &optional cound) --> ?
; (editor::start-line-p point) (editor::end-line-p point) --> T si debut ou fin de ligne
; (editor::move-point-to-offset point offset) --> voir a la place de faire characte-offset, start-point, etc. 
 
;;; ecrit le contenu d'un fichier dans le buffer
(defun om-buffer-insert-file (ombuffer path &optional position)
  (let ((buffer (buffer ombuffer)))
    (let ((filebuf (editor::find-file-buffer path)))
      (if position
          (editor::use-buffer buffer 
            (editor::with-point ((p (editor:buffers-start buffer)))
              (editor::character-offset p position)
              (editor::insert-string p (editor::points-to-string (editor::buffers-start filebuf)
                                                                 (editor::buffers-end filebuf)))))
        (editor::set-buffer-contents buffer 
                                     (editor::points-to-string (editor::buffers-start filebuf)
                                                               (editor::buffers-end filebuf)))
        )
      (editor::kill-buffer-no-confirm filebuf))))

;;; ecrit le contenu du buffer dans un fichier 
(defun om-buffer-write-file (ombuffer path &key (if-exists :supersede))
  (let ((buffer (buffer ombuffer)))
    (with-open-file (s path :direction :output :if-exists if-exists)
      (write-string (editor::points-to-string 
                     (editor::buffers-start buffer) 
                     (editor::buffers-end buffer)) 
                    s))))



;;;======================================
;;; OM TEXT EDITOR GENERAL
;;;======================================
;;; THIS INTERFACE IS INSPIRED BY ANVITA EDITOR
;;; by S. Ball

;; ERRORS HANDLING from S. Ball
;; this causes editor errors to be reported in echo area
;;(setf editor::*dont-catch-errors* nil)

(defvar *ae-error-log* 
  (make-pathname :name "OM Text Editor" 
                 :type "log" 
                 :defaults (sys:get-folder-path :local-appdata)))

;; this causes errors in the interface process to be displayed in a message box
(defun report-ae-error (cc)
   (display-message "Error ~S in OM Text Editor Process" cc)
   (with-open-file (log *ae-error-log* 
                        :direction :output 
                        :if-exists :append 
                        :if-does-not-exist :create)
     (dbg:output-backtrace :verbose :stream log))
   (abort))

;; just to distinguish editor-error
(defun report-ae-editor-error (cc)
   (display-message "Editor Error ~S in OM Text Editor Process" cc)
   (with-open-file (log *ae-error-log* 
                        :direction :output 
                        :if-exists :append 
                        :if-does-not-exist :create)
     (dbg:output-backtrace :verbose :stream log))
   (abort))

(defun handle-ae-error (func interface)
  (declare (ignore interface))
  (handler-bind 
      ((editor:editor-error #'report-ae-editor-error)
       (error #'report-ae-error))
    (funcall func)))

(defun echo-string (interface string)
  (with-slots (ep) interface
    (editor:process-character 
     (list 'editor:message string) 
     (editor-window ep))))
  
(defun report-file-operation (interface op file)
  (when op
    (echo-string interface (format nil "~A file ~A." op file))))

(defun report-failed-file-operation (interface op file)
  (declare (ignore interface))
  (display-message "Failed to ~A file ~A." op file))

(defmacro with-safe-file-operation (interface path present past &body forms)
 `(restart-case
      (prog1 ,@forms (report-file-operation ,interface ,past ,path))
    (abort ()
           (report-failed-file-operation ,interface ,present ,path))))

;;;======================================
;;; OM Text Editor peut editer un BUFFER ou un FICHIER (TEXTE ou LISP)
;;;======================================

(defvar *om-text-editor-initial-xy*)
(setq *om-text-editor-initial-xy* #(100 100))
(setq *om-text-editor-count* 0)

(define-interface om-text-editor ()
  ((editor-buffer :initform nil :accessor editor-buffer :initarg :editor-buffer)
   (editor-file :initform nil :accessor editor-file :initarg :editor-file)
   (lisp-editor? :initform nil :accessor lisp-editor? :initarg :lisp-editor?))
  ;;; editor-buffer = ombuffer
  ;;; editor-file = pathname (optional)
  ;;; lisp-editor? = t or nil (if t : syntax color, load, etc.)
  
  (:panes
   (ep editor-pane :buffer-name "Main" :echo-area t))
  (:layouts 
   (ed-layout simple-layout '(ep)))
  (:menu-bar file edit)
  (:menus
   (file
    "File"
    ((:component
      (("New..." :callback 'open-new-text-editor :accelerator #\n :enabled-function 'file-operations-enabled)
       ("Open..." :callback 'open-new-text-file :accelerator #\o :enabled-function 'file-operations-enabled))
      :callback-type :interface)
     (:component
      (("Import From..." :callback 'import-text-from-file)
       ("Revert to Saved" :callback 'revert-text-file :accelerator nil :enabled-function 'save-operation-enabled))
      :callback-type :interface)
     (:component
      (("Save" :callback 'save-text-file :accelerator #\s :enabled-function 'save-operation-enabled)
       ("Save As..." :callback 'save-as-text-file :accelerator nil))
      :callback-type :interface)
     (:component
      (("Close" :callback 'close-text-editor-window :accelerator #\w))
      :callback-type :interface)
     (:component
      (("Eval" :callback 'eval-lisp-buffer :accelerator #\y :enabled-function 'load-operation-enabled))
      :callback-type :interface)))
   (edit
    "Edit" 
    ((:component
      (("Undo" :callback 'text-edit-undo :accelerator #\z))
      :callback-type :interface)
     (:component
      (("Copy" :callback 'text-edit-copy :accelerator #\c)
       ("Cut" :callback 'text-edit-cut :accelerator #\x)
       ("Paste" :callback 'text-edit-paste :accelerator #\v))
      :callback-type :interface)
     (:component
      (("Select All" :callback 'text-select-all :accelerator #\a))
      :callback-type :interface)
     (:component
      (("Find..." :callback 'find-in-file :accelerator #\f :enabled-function 'disabled)
       ("Replace..." :callback 'replace-in-file :accelerator #\r :enabled-function 'disabled))
      :callback-type :interface)
     (:component
      (("Search..." :callback 'search-files :accelerator nil :enabled-function 'disabled))
      :callback-type :interface))
    ))

  (:default-initargs
   :best-x (+ (aref *om-text-editor-initial-xy* 0) (* 15 (mod (incf *om-text-editor-count*) 10)) (* 200 (floor (/ *om-text-editor-count* 10))))
   :best-y (+ (aref *om-text-editor-initial-xy* 1) (* 15 (mod *om-text-editor-count* 10)))
   :best-height 400
   :best-width 500
   :layout 'ed-layout
   :message-area t
   :create-callback 'init-text-editor
   :destroy-callback 'destroy-text-editor
   :confirm-destroy-function 'check-save-buffer
   ;; :top-level-hook 'handle-ae-error
   :title "Text Editor"))

(defun disabled (win) nil)

(defmethod destroy-text-editor ((self om-text-editor))
  (om-destroy-callback self))

(defmethod buffer-modified-p ((self om-text-editor))
  (when (editor-buffer self)
    (editor:buffer-modified (buffer (editor-buffer self)))))

(defmethod check-save-buffer ((self om-text-editor))
  (if (save-operation-enabled self)
      (multiple-value-bind (answer successp)
          (capi:prompt-for-confirmation
           (format nil "Do you want to save changes on ~A ?" 
                   (if (editor-file self) (pathname-name (editor-file self)) "this text buffer"))
           :cancel-button t)
    (if answer (save-before-quit self) successp))
    t))
;;;
;;; SET A BUFFER TO EDIT
;;; text file
(defun display-text-file (interface buffer path)
  (with-slots (ep) interface
    (setf (editor-pane-buffer ep) buffer))
  (echo-string interface ""))
;;; buffer only
(defun display-buffer (interface buffer)
  (with-slots (ep) interface
    (setf (editor-pane-buffer ep) buffer))
  (echo-string interface ""))

;;; OPEN A BUFFER given or created
;;; Previous buffer is not killed
(defun open-new-buffer (interface &optional buffer)
  (with-slots (ep) interface
    (let ((new-buffer (or buffer (editor::make-buffer (string (gensym))))))
      (display-buffer interface new-buffer)
      (if (lisp-editor? interface) 
          (call-editor ep (list 'editor:lisp-mode-command (editor-pane-buffer ep))))
      (echo-string interface "welcome to OM lisp editor")
      )))

;;; OPEN A FILE given or prompted
;;; Previous buffer is not killed
(defun open-new-file (interface &optional path)
  (with-slots (ep) interface
    (when-let* ((path (or path 
                          (prompt-for-file "Open File:" 
                                           :if-does-not-exist :error 
                                         :filter "*.*" 
                                         :filters '("Lisp Files" "*.lisp" "Text files" "*.txt" "All Files" "*.*"))))
              (buffer (with-safe-file-operation interface path "open" nil (editor:find-file-buffer path))))
      (display-text-file interface buffer path)
    (if (lisp-editor? interface) 
        (call-editor ep (list 'editor:lisp-mode-command (editor-pane-buffer ep))))
    (echo-string interface "welcome to OM lisp editor"))))
  
;;;================================
;;;================================
;;;================================

;;; INIT FUNC : depends on the attroibutes of the window
;;; if file -> open file
;;; if buffer --> edit buffer
;;; else --> create and edit new buffer
(defmethod init-text-editor ((self om-text-editor))
  (with-slots (ep) self
    (cond ((editor-file self) 
           (open-new-file self (editor-file self))
           (setf (editor-buffer self) (make-instance 'ombuffer :buffer (capi::editor-pane-buffer ep)))
           (setf (window (editor-buffer self)) self))
          ((editor-buffer self)
           (open-new-buffer self (buffer (editor-buffer self)))
           (setf (window (editor-buffer self)) self))
          (t (open-new-buffer self)
             (setf (editor-buffer self) (make-instance 'ombuffer :buffer (capi::editor-pane-buffer ep)))
             (setf (window (editor-buffer self)) self))
          )))

;;; CHANGE EDITOR-BUFFER DELETE SETS THE WINDOW ATTRIBUTE FOR BUFFER
;;; Previous buffer is not killed
(defmethod (setf editor-buffer) :before ((buffer ombuffer) (self om-text-editor))
  (when (editor-buffer self)
    (setf (window (editor-buffer self)) nil)))

(defmethod (setf editor-buffer) :after ((buffer ombuffer) (self om-text-editor))
  (setf (window (editor-buffer self)) self))

(defmethod (setf editor-file) :after (path (self om-text-editor))
  (if path
      (setf (capi::interface-title self) (namestring path))
    (setf (capi::interface-title self) "New Buffer")))

;;;==================
;;; MENU ACTIONS

;;; IMPORT : imports text from a file
;;; keeps the same ombuffers and change its internal buffer attribute 
;;; (previous is killed)
(defmethod import-text-from-file ((self om-text-editor) &optional path)
  (with-slots (ep) self
    (when-let* ((path (or path (prompt-for-file "Open File:" 
                                                :if-does-not-exist :error 
                                                :filter "*.*"
                                                :filters '("Lisp Files" "*.lisp" "Text files" "*.txt" "All Files" "*.*"))))
                (current (editor-pane-buffer ep))
                (newbuffer (with-safe-file-operation self path "open" nil 
                             (editor:find-file-buffer path)))
                (newtext (editor::use-buffer newbuffer
                           (editor:points-to-string 
                            (editor:buffers-start newbuffer) 
                            (editor:buffers-end newbuffer)))))
      (editor::use-buffer (buffer (editor-buffer self))
        (editor::clear-buffer (buffer (editor-buffer self)))
        (editor::insert-string (editor::buffers-start (buffer (editor-buffer self))) newtext))
      (setf (editor:buffer-modified (buffer (editor-buffer self))) t)
      (editor::kill-buffer-no-confirm newbuffer))))

;;; OPEN file : changes the buffer accordingly
;;; and updates editor-file
;;; (supposed to be enabled when editor-file is non-NIL)
(defmethod open-new-text-file ((self om-text-editor) &optional path)
   (with-slots (ep) self
     (let ((current (capi::editor-pane-buffer ep)))
       (check-save-buffer self)
       (open-new-file self path)
       (setf (buffer (editor-buffer self)) (capi::editor-pane-buffer ep))
       (setf (editor-file self) (editor:buffer-pathname (capi::editor-pane-buffer ep)))
       (editor::kill-buffer-no-confirm current))))
   
;;; SAVE the current buffer to the attached pathname
;;; or save as + change if no pathname
(defmethod save-text-file ((self om-text-editor))
  (with-slots (ep) self
    (let ((current (editor-pane-buffer ep)))
      (if (print (editor-file self))
          ;;; SAVE
          (with-safe-file-operation self (editor:buffer-pathname current) "save" "Saved"
            (call-editor ep (list 'editor:save-file-command current)))
        ;;; SAVE AS
        (when-let (path (prompt-for-file "Save file as:"
                                       :pathname (editor-file self)
                                       :if-does-not-exist :ok
                                       :filter "*.*"
                                       :if-exists :prompt
                                       :operation :save))
          (with-safe-file-operation self path "save" "Saved"
            (with-open-file (ss path :direction :output :if-exists :supersede)
              (write-string (editor:points-to-string 
                             (editor:buffers-start current) 
                             (editor:buffers-end current)) 
                            ss))
            t)
          (open-new-file self path)
          (setf (buffer (editor-buffer self)) (editor-pane-buffer ep)) 
          (editor::kill-buffer-no-confirm current)
          (setf (editor-file self) path)
          )))))

;;; SAVE AS : the current buffer to a new file
;;; any buffer ok, always works
;;; do not change current buffer
(defmethod save-as-text-file ((self om-text-editor))
  (with-slots (ep) self
    (let ((current (editor-pane-buffer ep)))
      (when-let (path (prompt-for-file "Save file as:"
                                       :pathname (editor-file self)
                                       :if-does-not-exist :ok
                                       :filter "*.*"
                                       :if-exists :prompt
                                       :operation :save))
        (with-safe-file-operation self path "save" "Saved"
          (with-open-file (ss path :direction :output :if-exists :supersede)
            (write-string (editor:points-to-string 
                           (editor:buffers-start current) 
                           (editor:buffers-end current)) 
                          ss))
          t
          )))))


(defmethod save-before-quit ((self om-text-editor))
  (with-slots (ep) self
    (let ((current (editor-pane-buffer ep)))
      (if (editor-file self)
          ;;; SAVE
          (with-safe-file-operation self (editor:buffer-pathname current) "save" "Saved"
            (call-editor ep (list 'editor:save-file-command current)))
        ;;; SAVE AS
        (when-let (path (prompt-for-file "Save file as:"
                                       :pathname (editor-file self)
                                       :if-does-not-exist :ok
                                       :filter "*.*"
                                       :if-exists :prompt
                                       :operation :save))
          (with-safe-file-operation  self path "save" "Saved"
            (with-open-file (ss path :direction :output :if-exists :supersede)
              (write-string (editor:points-to-string 
                             (editor:buffers-start current) 
                             (editor:buffers-end current)) 
                            ss)))
          ))))
  t)

;;; REVERT : put the file contents back in the editor
(defmethod revert-text-file ((self om-text-editor))
  (if (editor-file self)
    (with-slots (ep) self
      (let* ((current (editor-pane-buffer ep))
             (path (editor:buffer-pathname current)))
        (with-safe-file-operation self path "revert" "Reverted"
          (call-editor ep (list 'editor:revert-buffer-command nil current nil)))
        ))
    (error "No file is attached to this editor")))

;;; CLOSE
(defmethod close-text-editor-window ((self om-text-editor))
  (check-save-buffer self)
  (destroy self))

;;; LOAD THE LISP FILE ATTACHED...
(defmethod load-lisp-file ((self om-text-editor))
  (with-slots (ep) self
    (let ((path (editor:buffer-pathname (editor-pane-buffer ep))))
      (if (probe-file path)
          (with-safe-file-operation self path "load" "Loaded"
            (load path :print nil :verbose nil))
        (error "File ~A not found." path)))))

;;; EVAL the buffer...
(defmethod eval-lisp-buffer ((self om-text-editor))
  (with-slots (ep) self
    (let ((buffer (editor-pane-buffer ep)))
      (call-editor ep (list 'editor::evaluate-buffer-command buffer)))))

;;; NEW editor window
(defun open-new-text-editor (window &optional path)
  (let ((buffer (if path (editor:find-file-buffer path)
                  (editor::make-buffer (string (gensym))))))
    (capi::display (make-instance 'om-text-editor
                                  :x 200 :y 200 :width 800 :height 800
                                  :parent (capi:convert-to-screen)
                                  :internal-border 0 :external-border 0
                                  :display-state :normal
                                  :internal-min-height 200 :internal-min-width 300
                                  ;:font (om-make-font "courier" 20)
                                  :title (if path (namestring path) "New Buffer") 
                                  :editor-buffer (make-instance 'ombuffer :buffer buffer)
                                  :editor-file path
                                  :lisp-editor? t
                                  ))))

(defmethod text-edit-copy ((self om-text-editor))
  (with-slots (ep) self
    (let ((buffer (editor-pane-buffer ep)))
      (call-editor ep (list 'editor::copy-to-cut-buffer-command buffer)))))

(defmethod text-edit-cut ((self om-text-editor))
  (with-slots (ep) self
    (let ((buffer (editor-pane-buffer ep)))
      (call-editor ep (list 'editor::copy-to-cut-buffer-command buffer))
      (call-editor ep (list 'editor::kill-region-command buffer)))))

(defmethod text-edit-paste ((self om-text-editor))
  (with-slots (ep) self
    (let ((buffer (editor-pane-buffer ep)))
      (call-editor ep (list 'editor::insert-cut-buffer-command buffer)))))

(defmethod text-edit-undo ((self om-text-editor))
  (with-slots (ep) self
    (let ((buffer (editor-pane-buffer ep)))
      (call-editor ep (list 'editor::undo-command buffer)))))

;;; SELECT ALL BUFFER
;;; to do ge back to initial position...
(defmethod text-select-all ((self om-text-editor))
    (with-slots (ep) self
      (let ((buffer (editor-pane-buffer ep)))
        (editor::use-buffer buffer
          (editor::with-point ((p (editor::buffer-point buffer)))
            (call-editor ep (list 'editor::beginning-of-buffer-cancelling-selection-command buffer))
            (call-editor ep (list 'editor::end-of-buffer-cancelling-selection-command buffer))
            ;(editor::goto-line buffer 4)
            ;(call-editor ep (list 'editor::goto-point-command buffer p))
            ;(editor::move-point (editor::current-point) p)
            ;(editor::set-current-mark (editor::current-point))
            (call-editor ep (list 'editor::goto-point-command p 3))
            )))))


;;; Utile pour la suite...
;;;
;;; EDITOR::APROPOS-COMMAND --> symbol browser
;;; EDITOR::BACKUP-FILE-COMMAND --> cree un backup sans changer le buffer
;;; EDITOR::COMMENT-REGION-COMMAND
;;; EDITOR::EVALUATE-REGION-COMMAND

;;;======================
;;; OM GENERAL LISP EDITOR
;;;======================

(export '(
          om-open-new-text-editor
          om-open-new-text-file
          ) :om-api)

;;; ouvre un fichier ou un new buffer
(defun om-open-new-text-editor (&optional path)
  (open-new-text-editor nil path))

;;; ouvre un fichier only
(defmethod om-open-new-text-file (&optional path)
  (if (not path)
      (multiple-value-bind
          (a b)
          (capi:with-dialog-results (new-path ok)
              (capi:prompt-for-file "Choose a File")
            (when ok
              (open-new-text-editor nil new-path))))
    (open-new-text-editor nil path)))


(defmethod load-operation-enabled ((self t)) nil)
(defmethod load-operation-enabled ((self om-text-editor)) (lisp-editor? self))

(defmethod file-operations-enabled ((self t)) t)
(defmethod file-operations-enabled ((self om-text-editor)) t)

(defmethod save-operation-enabled ((self om-text-editor)) (buffer-modified-p self))

(defmethod om-destroy-callback  ((self om-text-editor))
  (editor::kill-buffer-no-confirm (buffer (editor-buffer self)))
  (setf (window (editor-buffer self)) nil))

; (om-open-new-text-editor)



;;;======================
;;; OM TEXT EDIT WINDOW
;;; Editor Window pour OM
;;;==============================
;;; & FONCTIONS EXPORTEES DANS OM
;;;==============================

(export '(
          om-text-edit-window
          om-set-text
          om-set-text-buffer
          om-get-lisp-expression
          ) :om-api)
       

(defclass om-text-edit-window (om-text-editor) ())

;;; celui-la soit il a un fichier soit non mais c'est fixe
(defmethod file-operations-enabled ((self om-text-edit-window)) (editor-file self))
(defmethod save-operation-enabled ((self om-text-edit-window)) (and (file-operations-enabled self) (buffer-modified-p self)))

(defmethod make-window-layout ((self om-text-edit-window) &optional color) nil)

;;; DOES NOT KILL THE BUFFER 
(defmethod om-destroy-callback  ((self om-text-edit-window))
  (setf (window (editor-buffer self)) nil)
  (om-window-close-event self))

(defmethod om-select-window ((self om-text-edit-window)) 
    (capi::find-interface (type-of self) :name (capi::capi-object-name self)))

(defmethod om-view-window ((self om-text-edit-window)) self)

;;; met un nouveau buffer
(defmethod om-set-text-buffer ((self om-text-edit-window) newbuffer &optional kill) 
  (let ((rec (buffer (editor-buffer self))))
    (setf (editor-buffer self) newbuffer)
    (init-text-editor self)
     ;;; detruit le buffer precedent
    (when kill (editor::kill-buffer-no-confirm rec))))

;;; garde le mme buffer mais change le contenu
(defmethod om-set-text ((self om-text-edit-window) (text string)) 
  (om-delete-buffer (editor-buffer self))
  (om-buffer-insert (editor-buffer self) text))

(defmethod om-get-lisp-expression ((self om-text-edit-window)) t)


;;;=============================================
;;; ADDITIONAL EDITOR FEATURES
;;;=============================================


;;;=============================================
;;; ARGLIST (from LW examples)
;;;=============================================
(in-package "CL-USER")

(defvar *arglist-delay* 1)
(defvar *arglist-timer* nil)

(defvar *setf-names-p* nil "when true, show the argument list for setf names too.")

(editor:defcommand "Insert Space and Show Arglist" (p)
     "Display the argument list in the echo area a while after inserting Space to the right of the function."
     "Display the argument list."
  (editor:self-insert-command p #\Space)
  (let* ((x (if *setf-names-p*
                (editor:with-point ((temp1 (editor:current-point))
                                    (temp2 (editor:current-point)))
                  (when (editor:form-offset temp1 -1)
                    (ignore-errors
                      (let ((*package* (editor::buffer-package-to-use temp1)))
                        (read-from-string 
                         (editor:points-to-string temp1 temp2))))))
              (editor:buffer-symbol-at-point (editor:current-buffer))))
         (window (editor:current-window))
         (function (find-function-for-arglist x)))
    (when (fboundp function)
      (show-arglist function 
                    (capi:top-level-interface 
                     (editor:window-text-pane window)) 
                    window))))

(defun show-arglist (function interface editor-window)
  (let ((lambdalist (function-lambda-list  function)))
    (setq *arglist-timer* 
          (mp:make-timer 'capi:execute-with-interface interface 
                         'editor:process-character 
                         (list 'editor:message (format nil "ARGS: ~A" lambdalist))
                         editor-window))
    (mp:schedule-timer-relative *arglist-timer* *arglist-delay*)
    ))

(defun find-function-for-arglist (x)
  (typecase x
    (symbol x)
    (list (unless (dotted-list-p x)
            (if (eq (length x) 1)
                (find-function-for-arglist (car x))
              (case (car x)
                ((quote function) (find-function-for-arglist (cdr x)))
                (setf (and (= (length x) 2)
                           (symbolp (second x))
                           x)))))))) 

(editor:bind-key "Insert Space and Show Arglist" #\Space :mode "Lisp") 
(editor:bind-key "Insert Space and Show Arglist" #\Space :mode "Execute") 