;;===========================================================================
;OM API 
;Multiplatform API for OpenMusic
;Macintosh version (Digitool Macintosh Common Lisp - MCL)
;
;Copyright (C) 2004 IRCAM-Centre Georges Pompidou, Paris, France.
; 
;This program is free software; you can redistribute it and/or
;modify it under the terms of the GNU General Public License
;as published by the Free Software Foundation; either version 2
;of the License, or (at your option) any later version.
;
;See file LICENSE for further informations on licensing terms.
;
;This program is distributed in the hope that it will be useful,
;but WITHOUT ANY WARRANTY; without even the implied warranty of
;MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;GNU General Public License for more details.
;
;You should have received a copy of the GNU General Public License
;along with this program; if not, write to the Free Software
;Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
;
;Authors: Augusto Agon and Jean Bresson
;;===========================================================================

;;===========================================================================
;DocFile
; Misc. TOOLS (buffers, pointers, print, etc...)
;;===========================================================================

(in-package :om-api)
;;;========
;;; export :
;;;========
(export '(
                quoted-form-p
               ; special-form-p
                lambda-expression-p
                om-report-condition
                om-read-list-from-string
          
                om-make-pointer
                om-free-pointer
                om-read-ptr
                om-write-ptr
                memq
                while
                integer-to-string
                
                om-copy-command
                om-paste-command
                om-cut-command
                
                om-random-value
                
                om-show-documentation
                om-edit-definition

                om-page-setup
                om-print-window
                
                om-closure-function
                ) :om-api)

;----------------------
; LISP TOOLS FEATURES
;----------------------

(defun memq (item list)
       (member item list :test #'eq))

(defmacro %car (x)
  `(car (the cons ,x)))

(defmacro %cdr (x)
  `(cdr (the cons ,x)))

(defun quoted-form-p (form)
  (and (consp form)
        (eq (%car form) 'quote)
        (consp (%cdr form))
        (null (%cdr (%cdr form)))))

(defun lambda-expression-p (form)
   (and (consp form)
       (eq (%car form) 'lambda)
       (consp (%cdr form))
       (listp (%cadr form))))

(defmacro while (test &body body)
  (let ((testlab (gensym))
        (toplab (gensym)))
    `(tagbody
       (go ,testlab)
      ,toplab
      (progn ,@body)
      ,testlab
      (when ,test (go ,toplab)))))


(defun integer-to-string (int)
   (format nil "~D" int))

(defmethod om-report-condition ((c condition))
  "Condition report.. ?")

;(defmethod special-form-p (s)
;  (lispworks::special-form-p s))



(defun om-read-list-from-string (string)
  (multiple-value-bind (val pos) (ignore-errors 
				  (read-from-string string nil :eof :start 0))
    (if (eql  val :eof)
	nil
	(cons val (om-read-list-from-string string pos)))))

;=================
; copy/paste 

;;; appel au clipboard du systeme
(defmethod om-copy-command ((self t))
   (print "Nothing to copy to clipboard"))

(defmethod om-paste-command ((self t))
  (print "Nothing to paste from clipboard"))

(defmethod om-cut-command ((self t))
   (print "Nothing to cut and copy to clipboard"))

;=================
;safe random
(defun om-random-value (num)
  (if (< num 0)
    (- (random (- num)))
    (random num)))


;;;==== POINTER ======


(defun om-make-pointer (size &optional (clear nil))
  t)



(defun om-free-pointer (ptr) 
  t)

;;; a completer pour les autres types
(defun om-write-ptr (ptr pos type value)
  t)

(defun om-read-ptr (ptr pos type)
  t)


;;;==================
;;; PRINT
;;;==================

(defun om-page-setup ()
  (print "no printing..."))

(defun om-print-window (win)
  (declare (ignore win))
  (print "no printing..."))

;======for situation

(defun om-closure-function (fun)
  t)