;;===========================================================================
;OM API 
;Multiplatform API for OpenMusic
;Macintosh version (Digitool Macintosh Common Lisp - MCL)
;
;Copyright (C) 2004 IRCAM-Centre Georges Pompidou, Paris, France.
; 
;This program is free software; you can redistribute it and/or
;modify it under the terms of the GNU General Public License
;as published by the Free Software Foundation; either version 2
;of the License, or (at your option) any later version.
;
;See file LICENSE for further informations on licensing terms.
;
;This program is distributed in the hope that it will be useful,
;but WITHOUT ANY WARRANTY; without even the implied warranty of
;MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;GNU General Public License for more details.
;
;You should have received a copy of the GNU General Public License
;along with this program; if not, write to the Free Software
;Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
;
;Authors: Augusto Agon and Jean Bresson
;;===========================================================================

;;===========================================================================
;DocFile
; PREDEFINED DIALOGS AND MESSAGES 
;;===========================================================================



(in-package :om-api)

;;;==========
;;; export :
;;;==========
(export '(
                om-get-user-string
                om-choose-directory-dialog
                om-choose-new-directory-dialog
                om-choose-file-dialog
                om-choose-new-file-dialog
                om-y-or-n-dialog
                om-choose-color-dialog
                om-choose-font-dialog
                om-message-dialog
                om-beep
                ) :om-api)


;;;==========



;;; ASK USER FOR STRING
;(om-get-user-string "dame" :initial-string "yo")
(defun om-get-user-string (prompt &key (initial-string "") (window-title "get string") owner 
                                  (size (om-make-point 365 100))
                                  (position (om-make-point 200 140)))
  (prompt-for-string prompt :initial-value initial-string))
     

;;; CHOOSE A FILE
;(om-choose-file-dialog :prompt "escoja" :directory "cl:")
(defun om-choose-file-dialog (&key (prompt "Choose a file") (directory nil) (button-string "OK"))
  (prompt-for-file prompt :filter "*.*" :owner (capi::convert-to-screen) :pathname directory))

;;; CHOOSE A NEW FILE
(defun om-choose-new-file-dialog (&key (prompt "Choose a new file") (directory nil) (name "") (button-string "OK") (types nil))
   (declare (ignore name))
   (prompt-for-file prompt :filter "*.*" :owner (capi::convert-to-screen) :pathname directory :operation :save))

;;; CHOOSE A DIRECTORY
(defun om-choose-directory-dialog (&key (directory nil))
   (prompt-for-directory "Choose a directory" :owner (capi::convert-to-screen) :pathname directory))

;; (prompt-for-file "HEllo" :filter nil :filters nil :pathname (make-pathname :directory oa::*api-directory*) :operation :save)
;; (setf ppp (om-choose-new-directory-dialog :directory (make-pathname :directory oa::*api-directory*)))


;;; CHOOSE A NEW DIRECTORY
(defun om-choose-new-directory-dialog (&key (directory nil))
  (let ((path (prompt-for-file "Choose location and name for New Directory" :owner (capi::convert-to-screen) :filter nil :filters nil :pathname directory :operation :save)))
    (when path
      (make-pathname :device (pathname-device path) :directory (append (pathname-directory path) (list (pathname-name path))))
      )))


;;; YES OR NO DIALOG
(defun om-y-or-n-dialog (message &key (size (om-make-point 300 150)))
   (prompt-for-confirmation message))

;;; CHOOSE A COLOR
(defun om-choose-color-dialog (&key color)
  (prompt-for-color "Choose a color" :color (or color (om-make-color 0.5 0 0))))


;;;MESSAGE
(defun om-message-dialog (message &key (window-title "Warning") owner (size (om-make-point 335 100)) 
                                  (position (om-make-point 200 140)))
   (declare (ignore owner))
   (display-message message))

;;; SYSTEM BEEP
(defun om-beep ()
   (beep-pane nil))

(defun om-choose-font-dialog (&key (font *om-default-font2*))
   (prompt-for-font "Choose a font" :font font :owner nil))

; (om-choose-font-dialog)



