;OpenMusic
;
;Copyright (C) 1997, 1998, 1999, 2000 by IRCAM-Centre Georges Pompidou, Paris, France.
; 
;This program is free software; you can redistribute it and/or
;modify it under the terms of the GNU General Public License
;as published by the Free Software Foundation; either version 2
;of the License, or (at your option) any later version.
;
;See file LICENSE for further informations on licensing terms.
;
;This program is distributed in the hope that it will be useful,
;but WITHOUT ANY WARRANTY; without even the implied warranty of
;MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;GNU General Public License for more details.
;
;You should have received a copy of the GNU General Public License
;along with this program; if not, write to the Free Software
;Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
;
;Authors: Gerard Assayag and Augusto Agon


;DocFile
;Classes for omloop are defined in this file.
;Last Modifications :
;18/10/97 first date.
;DocFile
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package :om)

;==========================================
;LOOP
;==========================================

        
;===========================================================================
;PATCH

(defclass patchForLoop (patchForBox) ()
   (:documentation "This class implements the special patches for the omloop boxes.  #enddoc#
#seealso# (OMPatch omloop-box) #seealso#"))

(defmethod get-editor-class ((self patchForLoop)) 'LoopEditor)

(defmethod OpenEditorframe ((self patchForLoop))
   (or (editorframe self)
       (panel (open-new-RelationFrame  self  (name self) (get-elements self)))))


(defmethod compile-patch ((self patchForLoop))
   "Code generates by Loop patches is generate by this method."
   (let* ((boxes (boxes self))
          (oldletlist *let-list*)
          (*let-list* nil)
          (do-box (car (find-class-boxes boxes 'OMLoopDo)))
          (final-box (car (find-class-boxes boxes 'OMFinalDo)))
          (in-boxes (sort (find-class-boxes boxes 'OMin) '< :key 'indice))
          (symbols (mapcar #'(lambda (thein) (setf (in-symbol thein) (gensym))) in-boxes))
          (loop-boxes (find-loop-boxes boxes))
          (loop-code (mapcan #'(lambda (theout)
                                 (loop-gen-code theout 0)) loop-boxes))
          (acum-boxes (find-acum-boxes boxes))
          (acum-declaration (mapcar #'(lambda (acumm)
                                        (declare-closure acumm)) acum-boxes))
          (acum-inits (mapcar #'(lambda (acumm)
                                  (gen-code acumm -1)) acum-boxes))
          body)
     (setf body (gen-code do-box 0))
     (eval `(defun ,(intern (string (first (code self))) :om)  (,.symbols) 
              (let (,.acum-declaration) ,.acum-inits
                   (loop ,.loop-code 
                         finally (return (values ,.(loop for i from 0 to (- (length (inputs final-box)) 1)
                                                         collect  (gen-code final-box i)))) do
                         (let* ,*let-list* ,body)))))
     (setf *let-list* oldletlist)))


;----------------
;EDITOR
;----------------
(omg-defclass loopEditor (boxpatchEditor) ())

(defmethod get-editor-panel-class ((self loopEditor))  'loopPanel)

;----------------
;PANEL
;----------------
(omg-defclass loopPanel (boxpatchPanel) ()
   (:documentation "This is the class for the editor panel which define a OMloop box.  #enddoc#
#seealso# (omLoop-Box box-with-patch-frame) #seealso#
#patch# This slot keeps the patch associated to the box. #patch#"))

(defmethod omg-remove-element ((self loopPanel) (box inFrame))
   "The inputs of a loop patch can be removed only from the omloop box."
   (om-beep-msg "delete inputs in the loop box"))

(defmethod omg-remove-element ((self loopPanel) (box t))
   "Finally and eachTime boxes can not be removed."
   (if (or (equal (class-name (class-of (object box))) 'OMLoopDo)
           (equal (class-name (class-of (object box))) 'OMfinalDo))
     (om-beep-msg "These boxes can not be erased")
     (call-next-method)))

(defmethod add-window-buttons  ((self loopPanel))
   "Add iterator and accumulator buttons to the patch."
   (om-add-subviews self (om-make-view 'button-icon
                        :iconID 151
                        :position (om-make-point 45 5)
                        :size (om-make-point 24 24)
                        :help-spec "For i from exp to exp"
                        :action
                        #'(lambda(item) (declare (ignore item)) 
                           (omG-add-element self
                                            (make-frame-from-callobj 
                                             (omNG-make-new-boxcall (fdefinition 'forloop)
                                                                    (om-make-point 5 80)
                                                                    (mk-unique-name self "for"))))))
                 (om-make-view 'button-icon
                   :iconID 152
                   :position (om-make-point 75 5)
                   :size (om-make-point 24 24)
                   :help-spec "While exp"
                   :action
                   #'(lambda(item)  (declare (ignore item))
                      (omG-add-element self
                                       (make-frame-from-callobj 
                                        (omNG-make-new-boxcall (fdefinition 'whileloop)
                                                               (om-make-point 5 80)
                                                               (mk-unique-name self "while"))))))
                 (om-make-view 'button-icon
                   :iconID 153
                   :position (om-make-point 105 5)
                   :size (om-make-point 24 24)
                   :help-spec "for element in list"
                   :action
                   #'(lambda(item) (declare (ignore item)) 
                      (omG-add-element self
                                       (make-frame-from-callobj 
                                        (omNG-make-new-boxcall (fdefinition 'listloop)
                                                               (om-make-point 5 80)
                                                               (mk-unique-name self "inlist"))))))
                 (om-make-view 'button-icon
                   :iconID 158
                   :position (om-make-point 135 5)
                   :size (om-make-point 24 24)
                   :help-spec "for element on list"
                   :action
                   #'(lambda(item) (declare (ignore item)) 
                      (omG-add-element self
                                       (make-frame-from-callobj 
                                        (omNG-make-new-boxcall (fdefinition 'onlistloop)
                                                               (om-make-point 5 80)
                                                               (mk-unique-name self "onlist"))))))
                 
                 (om-make-view 'button-icon
                   :iconID 175
                   :position (om-make-point 190 5)
                   :size (om-make-point 24 24)
                   :action
                   #'(lambda(item) (declare (ignore item)) 
                      (omG-add-element self
                                       (make-frame-from-callobj 
                                        (omNG-make-new-boxcall (fdefinition 'counter)
                                                               (om-make-point 5 80)
                                                               (mk-unique-name self "count"))))))
                 (om-make-view 'button-icon
                   :iconID 172
                   :position (om-make-point 215 5)
                   :size (om-make-point 24 24)
                   :action
                   #'(lambda(item)  (declare (ignore item))
                      (omG-add-element self
                                       (make-frame-from-callobj 
                                        (omNG-make-new-boxcall (fdefinition 'sum)
                                                               (om-make-point 5 80)
                                                               (mk-unique-name self "sum"))))))
                 (om-make-view 'button-icon
                   :iconID 174
                   :position (om-make-point 240 5)
                   :size (om-make-point 24 24)
                   :action
                   #'(lambda(item)  (declare (ignore item))
                      (omG-add-element self
                                       (make-frame-from-callobj 
                                        (omNG-make-new-boxcall (fdefinition 'minim)
                                                               (om-make-point 5 80)
                                                               (mk-unique-name self "min"))))))
                 (om-make-view 'button-icon
                   :iconID 173
                   :position (om-make-point 265 5)
                   :size (om-make-point 24 24)
                   :action
                   #'(lambda(item)  (declare (ignore item))
                      (omG-add-element self
                                       (make-frame-from-callobj 
                                        (omNG-make-new-boxcall (fdefinition 'maxi)
                                                               (om-make-point 5 80)
                                                               (mk-unique-name self "max"))))))
                 (om-make-view 'button-icon
                   :iconID 206
                   :position (om-make-point 290 5)
                   :size (om-make-point 24 24)
                   :action
                   #'(lambda(item)  (declare (ignore item))
                      (omG-add-element self
                                       (make-frame-from-callobj 
                                        (omNG-make-new-boxcall (fdefinition 'listing)
                                                               (om-make-point 5 80)
                                                               (mk-unique-name self "collect"))))))
                 (om-make-view 'button-icon
                   :iconID 171
                   :position (om-make-point 315 5)
                   :size (om-make-point 24 24)
                   :help-spec "General accumulator"
                   :action
                   #'(lambda(item)  (declare (ignore item))
                      (omG-add-element self
                                       (make-frame-from-callobj 
                                        (omNG-make-new-boxcall (fdefinition 'accumulator)
                                                               (om-make-point 5 80)
                                                               (mk-unique-name self "acum"))))))))


               
;--------------------

(defclass OMLoopDo (box-seq-call) ()
   (:documentation "This is the class for the each time boxes in a loop.#enddoc#
#seealso# (omloop-box loopIterators acumboxes) #seealso#"))

(defmethod* loopDo  ((patch t) &rest oppatch) :numouts 0 
  :initvals '(nil) :indoc '("patch") :doc "eval in ordre" :icon 208
  (declare (ignore oppatch)) nil)

(defmethod numouts ((self OMLoopDo)) 0)
(defmethod get-frame-name ((self OMLoopDo)) (name self))
(defmethod get-boxcallclass-fun ((self (eql 'loopDo))) 'OMLoopDo)
(defmethod allow-lock-button ((self OMLoopDo)) 
   "each time boxes do not allow a lock button." nil)
(defmethod no-allow-copy-p ((self OMLoopDo)) 
   "'Eachtime' boxes can not be directly copied"
   "'Eachtime' loop boxes.")

;--------------------
(defclass OMfinalDo (box-seq-call) ()
   (:documentation "This is the class for the finally boxes in a loop.#enddoc#
#seealso# (omloop-box loopIterators acumboxes) #seealso#"))

(defmethod allow-lock-button ((self OMfinalDo))
   "Finally boxes do not allow a lock button." nil)

(defmethod no-allow-copy-p ((self OMfinalDo))
   "Finally boxes can not be directly copied" "Finally loop boxes.")

(defmethod* finalDo  ((patch t) &rest oppatch) :numouts 0 
  :initvals '(nil) :indoc '("patch") :doc "Connect in this box the result of the loop evaluation you can add multiple outputs" :icon 216
  (declare (ignore oppatch)))

(defmethod numouts ((self OMfinalDo)) 0)
(defmethod get-frame-name ((self OMfinalDo)) (name self))
(defmethod get-boxcallclass-fun ((self (eql 'finalDo))) 'OMfinalDo)


(defmethod do-delete-one-input-extra ((self OMfinalDo))
   "When you erase one input from the finally box you must delete one input from the loop box also."
   (call-next-method)
   (let ((box (box (object (om-view-container (car (frames self)))))))
     (setf (numouts box) (- (numouts box) 1))
     (redraw-frame (car (frames box))) t))

(defmethod do-add-one-input-extra ((self OMfinalDo))
   "When you add one input to the finally box you must add one input to the loop box also."
   (call-next-method)
   (let ((box (box (object (om-view-container (car (frames self)))))))
     (setf (numouts box) (+ (numouts box) 1))
     (redraw-frame (car (frames box)))
     t))


;==================THE BOX LOOP CLASS=======================
(defclass omloop-box (box-with-patch) 
   ((numouts :initform 1 :accessor numouts))
   (:documentation "This is the class for the omloop boxes.#enddoc#
#seealso# (box-with-editor loopIterators acumboxes) #seealso#"))

(defmethod omNG-copy ((self omloop-box))
  `(let* ((copy ,(call-next-method)))
     (setf (numouts copy) ,(numouts self))
     copy))

(defmethod initialize-instance :after ((self omloop-box) &key controls)
  (declare (ignore controls))
  (let ((thedo (omNG-make-new-boxcall (fdefinition 'loopDo) (om-make-point 5 170) "eachTime"))
        (thefinal (omNG-make-new-boxcall (fdefinition 'finalDo) (om-make-point 150 170) "finally")))
    (setf (code (patch self)) (list (gensym)))
    (omNG-add-element (patch self) thedo)
    (omNG-add-element (patch self) thefinal)))

(defmethod openeditorframe ((self omloop-box))
   (openobjecteditor (patch self)) nil)

(defmethod get-patch-editor-class ((self omloop-box))
   "'patchForLoop' is the class of the Patch associated to 'self'."
   'patchForLoop)

(defmethod call-gen-code ((self omloop-box) numout)
   (let ((in-list (mapcar #'(lambda (thein) (gen-code thein 0)) (inputs self))))
     (if (zerop numout) 
       `(,(first (code (patch self))) ,.in-list)
       `(nth ,numout (multiple-value-list (,(first (code (patch self))) ,.in-list))))))

;screamer
(defmethod call-gen-code ((self omloop-box) numout)
   (let ((in-list (mapcar #'(lambda (thein) (gen-code thein 0)) (inputs self))))
     (if (zerop numout) 
       `(,(intern (string (first (code (patch self)))) :om) ,.in-list)
       `(nth ,numout (multiple-value-list (,(intern (string (first (code (patch self)))) :om) ,.in-list))))))

(defmethod gen-code-call ((self omloop-box))
   (let ((in-list (mapcar #'(lambda (thein) (gen-code thein 0)) (inputs self))))
     `(,(first (code (patch self))) ,.in-list)))

;screamer
(defmethod gen-code-call ((self omloop-box))
   (let ((in-list (mapcar #'(lambda (thein) (gen-code thein 0)) (inputs self))))
     `(,(intern (string (first (code (patch self)))) :om) ,.in-list)))

(defmethod special-lambda-value ((self omloop-box) symbol)
   "Eval a loop box in lambda mode."
   (let* ((nesymbs nil)
          (args  (mapcar #'(lambda (input)
                             (if (connected? input)
                               `',(omNG-box-value  input)
                               (let ((newsymbol (gensym)))
                                 (push newsymbol nesymbs)
                                 newsymbol))) (inputs self))))
     (if (null nesymbs)
       symbol
       (eval `#'(lambda ,(reverse nesymbs)
                  (apply (fdefinition ',(first (code (patch self)))) (list ,.args)))))))

;screamer
(defmethod special-lambda-value ((self omloop-box) symbol)
   "Eval a loop box in lambda mode."
   (let* ((nesymbs nil)
          (args  (mapcar #'(lambda (input)
                             (if (connected? input)
                               `',(omNG-box-value  input)
                               (let ((newsymbol (gensym)))
                                 (push newsymbol nesymbs)
                                 newsymbol))) (inputs self))))
     (if (null nesymbs)
       symbol
       (eval `#'(lambda ,(reverse nesymbs)
                  (apply (fdefinition ',(intern (string (first (code (patch self)))) :om)) (list ,.args)))))))

(defmethod curry-lambda-code ((self omloop-box) symbol)
   "Lisp code generation for a loop box in lambda mode."
  (let* ((nesymbs nil)
         (args  (mapcar #'(lambda (input)
                            (if (connected? input)
                              (gen-code input 0)
                              (let ((newsymbol (gensym)))
                                (push newsymbol nesymbs)
                                newsymbol))) (inputs self))))
    (if (null nesymbs)
      symbol
      `#'(lambda ,(reverse nesymbs)
                 (apply (fdefinition ',(first (code (patch self)))) (list ,.args))))))

;screamer

(defmethod curry-lambda-code ((self omloop-box) symbol)
   "Lisp code generation for a loop box in lambda mode."
  (let* ((nesymbs nil)
         (args  (mapcar #'(lambda (input)
                            (if (connected? input)
                              (gen-code input 0)
                              (let ((newsymbol (gensym)))
                                (push newsymbol nesymbs)
                                newsymbol))) (inputs self))))
    (if (null nesymbs)
      symbol
      `#'(lambda ,(reverse nesymbs)
                 (apply (fdefinition ',(intern (string (first (code (patch self)))) :om)) (list ,.args))))))


(defmethod special-value ((self  omloop-box) &optional (args nil))
    (unless (compiled? (patch self)) 
      (compile-patch (patch self))
      (setf (compiled? (patch self)) t))
    (apply (fdefinition (intern (string (first (code (patch self)))) :om)) args))

(defmethod get-boxcallclass-fun ((self (eql 'omloop))) 'omloop-box)

;----frame

(omg-defclass loopboxframe (boxframe) ()
   (:documentation "Simple frame for OMLOOP boxes. #enddoc#
#seealso# (OMloop) #seealso#"))


(defmethod get-frame-class ((self omloop-box)) 'loopboxframe)

(defmethod omG-rename :after ((self loopboxframe) new-name)
   (declare (ignore new-name))
   (let ((thepatch (patch (object self))))
     (when (EditorFrame thepatch)
       (omG-rename (EditorFrame thepatch) (frame-name (object self))))
     (setf (name thepatch)  (frame-name (object self)))
     ))


(defmethod update-from-reference ((self omloop-box) &optional (udt? t))
   "Called when occurs changes in the patch reference of 'self'."
   (declare (ignore udt?))
   (let* ((new-inputs (mapcar #'(lambda (input) 
                                  (make-instance 'input-funbox
                                    :name (or (frame-name input) (name input))
                                    :value (eval (defval input))
                                    :doc-string (docu input))) 
                              (get-patch-inputs (patch self))))
          conec-to-me)
     (when *input-to-erase*
       (setf (inputs self) 
             (delete (nth *input-to-erase* (inputs self)) (inputs self) :test 'equal)))
     (mapc #'(lambda (oldin newin) 
               (setf (connected? newin) (connected? oldin))
               (setf (value newin) (value oldin))) 
           (inputs self) new-inputs)
     (setf (inputs self) new-inputs)
     (when *output-to-delete*
       (setf conec-to-me (get-conect-to-me self))
       (loop for item in conec-to-me do
             (erase-out-connec self item *output-to-delete*)))
     (when (frames self)
       (when *output-to-delete*
         (loop for source in conec-to-me do
               (box-draw-connections (car (frames source)) nil)
               (redraw-frame (car (frames source)))))
       (box-draw-connections (car (frames self)) nil)
       (redraw-frame (car (frames self))))))


(defmethod update-doc ((box loopboxframe))
   (update-from-reference (object box)))

(defmethod close-frame ((box loopboxframe))
   "If miniview show a picture we must kill it."
   (when (EditorFrame (patch (object box)))
     (om-close-window (window (EditorFrame (patch (object box))))))
   (setf (frames (object box)) nil))

;-------------


(defmethod* omloop  (&rest oplist) :numouts 1 
   :initvals '(nil) :indoc '("aditional input")
   :doc "General OM loop.
Add as many optional inputs as needed. Then dbl-click on the omloop icon.
Program your loop in the loop editor." 
   :icon 181 
   (declare (ignore oplist)) nil)

;==================ITERATORS=======================
(defclass loopIterators (OMBoxcall) 
   ((code :initform nil :accessor code))
   (:documentation "This is the class for iterator boxes in a loop.
There are specific subclasses for each type of iterator.#enddoc#
#seealso# (OMloop-box forloopcall whileloopcall listloopcall onlistloopcall)  #seealso#
#code# This slot store the specific Lisp code associated to the iteratore.#code#"))

(defmethod initialize-instance :after ((self loopIterators) &key controls)
   (declare (ignore controls))
   (setf (code self) (gensym)))

(defmethod allow-lock-button ((self loopIterators))
   "Iterators do not allow a lock button." nil)
  
(defmethod gen-code ((self loopIterators) numout)
   (declare (ignore numout))
   (code self))

(defmethod loopbox? ((self t)) nil)
(defmethod loopbox? ((self loopIterators)) t)

(defmethod* omNG-box-value ((self  loopIterators) &optional (numout 0))
   (declare (ignore numout))
   (dialog-message "You can not evaluate this box")
   (om-abort)
   )

(defun find-loop-boxes (lis)
  (let* (rep )
    (mapc #'(lambda (box)
              (when (loopbox? box)
                (push box rep))) lis)
   (sort-loop-boxes  rep)))

(defun sort-loop-boxes (lis)
  (let (cheked)
    (loop while lis do
          (loop for item in lis do
                (when (not (member-closure item lis))
                  (push item cheked)
                  (setf lis (remove item lis)))))
    (reverse cheked)))

(defun member-closure (item lis)
  (let (rep)
    (loop for box in lis 
          while (not rep) do
          (when (member item (get-closure box) :test 'equal)
            (setf rep t)))
    rep))

(defun get-closure (box)
  (let ((firstboxes (get-conect-to-me box)))
    (remove-duplicates (append firstboxes (loop for item in firstboxes
                                                append (get-closure item))) :test 'equal)))

;-------------------------------------
(defclass forloopcall (loopIterators) ()
   (:documentation "This is the class for the FOR 'i' FROM 'n' TO 'm' BY 'step' iterator.#enddoc#
#seealso# (whileloopcall listloopcall onlistloopcall)  #seealso#"))

(defmethod loop-gen-code ((self forloopcall) numout)
  (declare (ignore numout))
  (let ((step (when (third (inputs self)) (gen-code (third (inputs self)) 0))))
    (if step
      `(for ,(code self) from ,(gen-code (first (inputs self)) 0) to
            ,(gen-code (second (inputs self)) 0) by ,step)
      `(for ,(code self) from ,(gen-code (first (inputs self)) 0) to
            ,(gen-code (second (inputs self)) 0 )))))
                  

(defclass whileloopcall (loopIterators) ()
   (:documentation "This is the class for the WHILE 'condition' iterator.#enddoc#
#seealso# (forloopcall listloopcall onlistloopcall)  #seealso#"))

(defmethod loop-gen-code ((self whileloopcall) numout)
  (declare (ignore numout))
  `(while ,(gen-code (first (inputs self)) 0)))

(defclass listloopcall (loopIterators) ()
   (:documentation "This is the class for the FOR 'i' IN 'list' iterator.#enddoc#
#seealso# (forloopcall whileloopcall onlistloopcall)  #seealso#"))

(defmethod loop-gen-code ((self listloopcall) numout)
  (declare (ignore numout))
  (let ((step (when (second (inputs self)) (gen-code (second (inputs self)) 0))))
    (if step
      `(for ,(code self) in ,(gen-code (first (inputs self)) 0) by ,step)
      `(for ,(code self) in ,(gen-code (first (inputs self)) 0)))))

(defclass onlistloopcall (loopIterators) ()
   (:documentation "This is the class for the FOR 'i' ON 'list' iterator.#enddoc#
#seealso# (forloopcall whileloopcall listloopcall)  #seealso#"))

(defmethod loop-gen-code ((self onlistloopcall) numout)
  (declare (ignore numout))
  (let ((step (when (second (inputs self)) (gen-code (second (inputs self)) 0))))
    (if step
      `(for ,(code self) on ,(gen-code (first (inputs self)) 0) by ,step)
      `(for ,(code self) on ,(gen-code (first (inputs self)) 0)))))



(defmethod get-boxcallclass-fun ((self (eql 'forloop))) 'forloopcall)
(defmethod get-boxcallclass-fun ((self (eql 'whileloop))) 'whileloopcall)
(defmethod get-boxcallclass-fun ((self (eql 'listloop))) 'listloopcall)
(defmethod get-boxcallclass-fun ((self (eql 'onlistloop))) 'onlistloopcall)



(defmethod* forloop  ((from integer) (to integer) &optional by) 
  :icon 151  :initvals '(0 10 1) :indoc '("low val" "high val" "step")
  (declare (ignore by)) nil)

(defmethod* whileloop  ((expr t)) :numouts 0 :icon 152 :indoc '("invariant") nil)

(defmethod* listloop  ((list list)  &optional by) 
  :icon 153  :indoc '("list to iterate")
  (declare (ignore by)) nil)

(defmethod* onlistloop  ((list list)  &optional by) 
  :icon 158 :indoc '("list to iterate")  
  (declare (ignore by)) nil)

;======================ACCUMULATORS============================

(defclass acumboxes (OMBoxcall) 
   ((closure :initform nil :initarg :closure :accessor closure))
   (:documentation "This is the class for accumulator boxes in a loop.#enddoc#
#seealso# (OMBoxcall OMloop-box PatchForLoop) #seealso#
#closure# This slot store a lexical closure which define the binary operation performed
at each evaluation of the accumulator#closure#"))

(defmethod get-frame-name ((self acumboxes)) (name self))

(defmethod allow-lock-button ((self acumboxes))
   "Acumboxes do not allow lock buttom." nil)

(defmethod initialize-instance :after ((self acumboxes) &key controls)
   (declare (ignore controls))
   (setf (closure self) (gensym)))

(defmethod declare-closure ((self acumboxes))
  `(,(closure self) 
    (let (accum ret) 
       #'(lambda (cur init  fun flag &optional retard) 
           (cond ((= flag 0)  (when retard (setf ret retard))  (setf accum init))
                 ((= flag 1) (if ret (pop ret) (setf accum (funcall fun accum cur))))
                 ((= flag 2) (if ret (pop ret) accum)))))))

(defmethod omNG-box-value ((self  acumboxes) &optional (numout 0))
   (declare (ignore numout))
   (dialog-message "You can not evaluate this box")
   (om-abort))

(defmethod gen-code ((self acumboxes) numout)
   "(Case numout (1 trigger one accumulation and get the etat of self after the evaluation)
(2 get the etat of self without trigger the evaluation) (3 Initialize the accum)."
  (incf numout)
  (if (= numout 3)
    (gen-code self -1)
    `(funcall ,(closure self) ,(gen-code (first (inputs self)) 0) ,(gen-code (second (inputs self)) 0)
              ,(gen-code (third (inputs self)) 0) ,numout ,(gen-code (fourth (inputs self)) 0))))

(defmethod gen-code ((self acumboxes) (numout (eql 1)))
   `(funcall ,(closure self) nil nil nil 2 nil))
      
(defmethod gen-code ((self acumboxes) (numout (eql -1)))
   `(funcall ,(closure self) nil ,(gen-code (second (inputs self)) 0)
             nil 0 ,(gen-code (fourth (inputs self)) 0)))

(defmethod acumbox? ((self t)) nil)
(defmethod acumbox? ((self acumboxes)) t)

(defun find-acum-boxes (lis)
  (let* (rep)
    (mapc #'(lambda (box)
              (when (acumbox? box)
                (push box rep))) lis)
    (reverse rep)))
;------------------Specifiques accumulators---------------------

(defmethod* accumulator  ((accum t) init fun  &optional retard) :numouts 3 :icon 171  
  (declare (ignore init fun retard)) nil)

(defmethod get-boxcallclass-fun ((self (eql 'accumulator))) 'acumboxes)


(defclass funacumboxes (acumboxes) 
   ((fun-name :initform nil :initarg :fun-name :accessor fun-name)
    (init-val :initform nil :initarg :init-val :accessor init-val))
   (:documentation "There exist five specific accumulators (collect, max, min, count if and sum.
Each one is an instance af this class.#enddoc#
#seealso# (acumboxes OMloop-box) #seealso#
#fun-name# A name for the function called when you eval the accumuilator (i.e. + for sum, max for max, etc.)#fun-name#
#init-val# An initial value for the accumuilator (i.e. 0 for sum, nil for collect, etc.)#init-val# "))

(defmethod initialize-instance :after ((self funacumboxes) &key controls)
  (declare (ignore controls))
  (let ((acumfun (get-fun-acum (reference self))))
    (setf (fun-name self) (first acumfun))
    (setf (init-val self) (second acumfun))))


(defmethod gen-code ((self funacumboxes) numout)
  (incf numout)
  (if (= numout 3)
    (gen-code self -1)
  `(funcall ,(closure self) ,(gen-code (first (inputs self)) 0) ,(init-val self)
            ',(fun-name self) ,numout ,(gen-code (second (inputs self)) 0))))
        
(defmethod gen-code ((self funacumboxes) (numout (eql -1)))
  `(funcall ,(closure self) nil ,(init-val self) nil 0))

(defmethod gen-code ((self funacumboxes) (numout (eql 1)))
  `(funcall ,(closure self) nil nil nil 2 nil))

(defmethod* counter  ((what? t) &optional retard) :numouts 3 :icon 175  
  (declare (ignore retard)) nil)
(defmethod* sum  ((what? t) &optional retard) :numouts 3 :icon 172  
  (declare (ignore retard)) nil)
(defmethod* minim  ((what? t) &optional retard) :numouts 3 :icon 174
  (declare (ignore retard)) nil)
(defmethod* maxi  ((what? t) &optional retard) :numouts 3 :icon 173
  (declare (ignore retard)) nil)
(defmethod* listing  ((what? t) &optional retard) :numouts 3 :icon 206
  (declare (ignore retard)) nil)


(defmethod get-boxcallclass-fun ((self (eql 'counter))) 'funacumboxes)
(defmethod get-boxcallclass-fun ((self (eql 'sum))) 'funacumboxes)
(defmethod get-boxcallclass-fun ((self (eql 'minim))) 'funacumboxes)
(defmethod get-boxcallclass-fun ((self (eql 'maxi))) 'funacumboxes)


(defmethod get-fun-acum ((self (eql 'counter))) (list 'omcount 0))
(defmethod get-fun-acum ((self (eql 'sum))) (list '+ 0))
(defmethod get-fun-acum ((self (eql 'minim))) (list 'min (expt 2 32)))
(defmethod get-fun-acum ((self (eql 'maxi))) (list 'max (* -1 (expt 2 32))))
(defmethod get-fun-acum ((self (eql 'listing))) (list 'leftcons nil))

(defun omcount (x y)
  (if y (+ 1 x) x))

(defun leftcons (x y)
  (list+ x (list y)))


(defclass omcollect (acumboxes)  ())

(defmethod get-boxcallclass-fun ((self (eql 'listing))) 'omcollect)

(defmethod declare-closure ((self omcollect))
   `(,(closure self) 
     (let (accum ret tail) 
       #'(lambda (cur init  fun flag &optional retard)
           (declare (ignore fun init))
           (cond ((= flag 0)  
                  (when retard (setf ret retard))  
                  (setf accum (cons nil nil) tail accum))
                 ((= flag 1) 
                  (when ret 
                    (setf cur (pop ret)))
                  (setf (cdr tail) (cons cur nil))
                  (setf tail (cdr tail)))
                 ((= flag 2)  (rest accum)))))))




 


