;OpenMusic
;
;Copyright (C) 1997, 1998, 1999, 2000 by IRCAM-Centre Georges Pompidou, Paris, France.
; 
;This program is free software; you can redistribute it and/or
;modify it under the terms of the GNU General Public License
;as published by the Free Software Foundation; either version 2
;of the License, or (at your option) any later version.
;
;See file LICENSE for further informations on licensing terms.
;
;This program is distributed in the hope that it will be useful,
;but WITHOUT ANY WARRANTY; without even the implied warranty of
;MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;GNU General Public License for more details.
;
;You should have received a copy of the GNU General Public License
;along with this program; if not, write to the Free Software
;Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
;
;Authors: Gerard Assayag and Augusto Agon


;DocFile
;Red patches and red maquettes are defined in this file.
;Last Modifications :
;18/10/97 first date.
;DocFile
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package :om)

;==================THE BOX patch CLASS=======================
(defclass OMBoxAbsPatch (OMBoxPatch) ()
   (:documentation "Boxes having a OMPatchAbs as reference are instances of this class. #enddoc#
#seealso# (OMPatch OMPatchAbs patchboxabsFrame) #seealso#"))

(defmethod get-frame-class ((self OMBoxAbsPatch)) 'patchboxabsFrame)
(defmethod get-documentation ((self OMBoxAbsPatch)) "This patch is internal it is not a persistant object")
(defmethod get-object-insp-name ((self OMBoxAbsPatch)) "Internal patch")



;=====SAVE

(defmethod omNG-save ((self OMBoxAbsPatch) &optional (values? nil))
   "Save the box 'self' and its patch reference in the same file as the patch containing 'self'."
   (let* ((inputs (mapcar #'(lambda (input) (omNG-save input values?)) (inputs self)))
          (value (omNG-save (value self) values?)))
     `(om-load-boxcall 'abstraction ,(name self) ,(om-save (reference self))
                       ',inputs ,(om-save-point (frame-position self))
                       ,(om-save-point (frame-size self)) ,value ,(allow-lock self) ,(frame-name self))))

(defmethod om-load-boxcall ((self (eql 'abstraction)) name reference inputs position size value lock &optional fname numouts)
  (declare (ignore numoouts))
  (let ((newbox (omNG-make-new-boxcall reference (om-correct-point position) name)))
    (setf (frame-size newbox) (om-correct-point size))
    (setf (frame-name newbox) fname)
    (setf (inputs newbox) (mapcar #'(lambda (input) (eval input)) inputs))
    (setf (value newbox) value)
    (setf (allow-lock newbox) lock)
    newbox))

;=============COPY

(defmethod omNG-copy ((self OMBoxAbsPatch))
  `(let ((copy ,(omNG-make-new-boxcall (eval (omng-copy (reference self)))
                                       (frame-position self)
                                       (name self))))
     (setf copy (update-boxes ,self copy))
     copy))


;===========INTERNALIZE

(defmethod internalize-patch ((self t))
   "A blue patch or maquette  becomes a red one."
   nil)

(defmethod internalize-patch ((self patchboxframe))
   "A blue patch becomes a red patch."
   (when (equal (type-of (object self)) 'OMBoxPatch)
     (let* ((container (om-view-container self))
            (object (object self))
            (newpatch (patch2abs (reference object)))
            (newbox (omNG-make-new-boxcall newpatch
                                           (frame-position object)
                                           (mk-unique-name container (name self))))
            frame conec-to-me)
       (setf (frame-position newbox) (borne-position (frame-position object)))
       (setf (frame-size newbox) (frame-size object))
       (setf (frame-name newbox) (frame-name object))
       (setf (allow-lock newbox) (allow-lock object))
       (setf (value newbox) (eval (omNG-copy (value object))))
       (setf (inputs newbox) (eval (omNG-copy (inputs object))))
       (loop for input in (inputs object)
             for in in (inputs newbox) do
             (setf (connected? in) (connected? input)))
       (setf conec-to-me (get-conect-to-me object))
       (loop for item in conec-to-me do
             (change-conections  object item newbox))
       (omg-remove-element container self)
       (setf frame (make-frame-from-callobj newbox))
       (omg-add-element  container frame)
       (compile-patch newpatch)
       (update-graphic-connections frame (get-elements (object container))))))


(defmethod close-frame ((self patchboxframe))
   "Close the 'self' editor if 'self' is a red patch."
  (when (equal (type-of (object self)) 'OMBoxAbsPatch)
    (when (EditorFrame (reference (object self)))
      (om-close-window (window (EditorFrame (reference (object self)))))))
  (setf (frames (object self)) nil))

;===========EXTERNALIZE

(defmethod externalize ((self patchboxframe) newpatch) 
   "A red patch becomes a blue patch."
  (when (equal (type-of (object self)) 'OMBoxAbsPatch)
    (let* ((container (om-view-container self))
           (object (object self))
           (newbox (omNG-make-new-boxcall newpatch
                                          (frame-position object)
                                          (mk-unique-name container (name self))))
           frame conec-to-me)
      (setf (frame-position newbox) (borne-position (frame-position object)))
      (setf (frame-size newbox) (frame-size object))
      (setf (frame-name newbox) (frame-name object))
      (setf (allow-lock newbox) (allow-lock object))
      (setf (value newbox) (eval (omNG-copy (value object))))
      (setf (inputs newbox) (eval (omNG-copy (inputs object))))
      (loop for input in (inputs object)
            for in in (inputs newbox) do
            (setf (connected? in) (connected? input)))
      (setf conec-to-me (get-conect-to-me object))
      (loop for item in conec-to-me do
            (change-conections  object item newbox))
      (setf frame (make-frame-from-callobj newbox))
      (omg-remove-element  container self)
      (compile-patch newpatch)
      (omg-add-element  container frame)
      (update-graphic-connections frame (get-elements (object container)))
      (omng-save newpatch)
      )))



(defun change-conections  (self item newbox)
  (loop for inp in (inputs item) do 
        (let ((connec (connected? inp)))
          (when (and connec (equal (first connec) self))
            (setf (nth 0 (connected? inp)) newbox)))))


;==========================================
;MAQUETTE ABSTRACTION
;==========================================

       
;==================THE BOX MAQ CLASS=======================
(defclass OMBoxAbsmaq (OMBoxmaquette) ()
   (:documentation "Boxes having a OMMaqAbs as reference are instances of this class. #enddoc#
#seealso# (OMMaquette OMMaqAbs maquetteabsframe) #seealso#"))

(defmethod get-frame-class ((self OMBoxAbsmaq)) 'maquetteabsframe)
(defmethod get-documentation ((self OMBoxAbsmaq)) "cons or modify an internal maquette")
(defmethod get-object-insp-name ((self OMBoxAbsmaq)) "Internal maquette")


;=====SAVE


(defmethod omNG-save ((self OMBoxAbsmaq) &optional (values? nil))
  (let* ((inputs (mapcar #'(lambda (input) (omNG-save input values?)) (inputs self)))
         (value (omNG-save (value self) values?)))
    (if (= (mode self) 1)
      `(let ((box (om-load-boxcall 'maqabs ,(name self) ,(om-save (reference self))
                      ',inputs ,(om-save-point (frame-position self))
                      ,(om-save-point (frame-size self)) ,value ,(allow-lock self) ,(frame-name self))))
         (change-mode box 1 t)
         box)
      `(om-load-boxcall 'maqabs ,(name self) ,(om-save (reference self))
                      ',inputs ,(om-save-point (frame-position self))
                      ,(om-save-point (frame-size self)) ,value ,(allow-lock self) ,(frame-name self)))
    ))


(defun om-load-maq-abs1 (name boxes connections range markers 
                              &optional (colormaq nil) (metricparam '((4 60) ((4 4)) 16 t)) (show-connect t) (version nil) fond-ec)
   (let ((newmaquette (make-instance 'OMMaqAbs :name name :icon 265)))
     (setf (boxes newmaquette) nil)
     (mapc #'(lambda (box) (omNG-add-element newmaquette (eval box)))  boxes)
     (setf (boxes newmaquette) (reverse (boxes newmaquette)))
     (setf (connec newmaquette) connections)
     (remk-markers markers (boxes newmaquette))
     (when fond-ec (setf (pictu newmaquette) fond-ec))
     (when version
       (setf (omversion newmaquette) version))
     (setf (params newmaquette) (make-instance 'maquette-params 
                                  :range range
                                  :maq-color (om-correct-color colormaq)
                                  :metricparam metricparam
                                  :show-conect show-connect))
     (push newmaquette *loaading-stack*)
     newmaquette))

(defmethod om-load-boxcall ((self (eql 'maqabs)) name reference inputs position size value lock &optional fname numouts)
  (declare (ignore numouts))
  (let ((newbox (omNG-make-new-boxcall reference (om-correct-point position) name)))
    (setf (frame-size newbox) (om-correct-point size))
    (setf (frame-name newbox) fname)
    (setf (inputs newbox) (mapcar #'(lambda (input) (eval input)) inputs))
    (setf (value newbox) value)
    (setf (allow-lock newbox) lock)
    newbox))

;=============COPY

(defmethod omNG-copy ((self OMBoxAbsmaq))
  `(let ((copy ,(omNG-make-new-boxcall (eval (omng-copy (reference self)))
                                       (frame-position self)
                                       (name self))))
     (setf copy (update-boxes ,self copy))
     ,(if (= (mode self) 1) `(change-mode copy 1 t))
     copy))



;===========INTERNALIZE


(defmethod internalize-patch ((self maquetteframe))
  "A blue maquette becomes a red one"
  (when (equal (type-of (object self)) 'OMBoxMaquette)
    (let* ((container (om-view-container self))
           (object (object self))
           (newpatch (maq2abs (reference object)))
           (newbox (omNG-make-new-boxcall newpatch
                                          (frame-position object)
                                          (mk-unique-name container (name self))))
           frame conec-to-me)
      (setf (frame-position newbox) (borne-position (frame-position object)))
      (setf (frame-size newbox) (frame-size object))
      (setf (frame-name newbox) (frame-name object))
      (setf (allow-lock newbox) (allow-lock object))
      (setf (value newbox) (eval (omNG-copy (value object))))
      (setf (inputs newbox) (eval (omNG-copy (inputs object))))
      (loop for input in (inputs object)
            for in in (inputs newbox) do
            (setf (connected? in) (connected? input)))
      (setf conec-to-me (get-conect-to-me object))
      (loop for item in conec-to-me do
            (change-conections  object item newbox))
      (omg-remove-element  container self)
      (setf frame (make-frame-from-callobj newbox))
      (omg-add-element  container frame)
      (update-graphic-connections frame (get-elements (object container))))))


(defmethod close-frame ((self maquetteframe))
  "Close the 'self' editor if 'self' is a red maquette."
  (when (equal (type-of (object self)) 'OMBoxAbsmaq)
    (when (EditorFrame (reference (object self)))
      (om-close-window (window (EditorFrame (reference (object self)))))))
  (setf (frames (object self)) nil))


;===========EXTERNALIZE

(defmethod externalize ((self maquetteframe) newpatch)
  "A red maquette becomes a blue one." 
  (when (equal (type-of (object self)) 'OMBoxAbsmaq)
    (let* ((container (om-view-container self))
           (object (object self))
           (newbox (omNG-make-new-boxcall newpatch
                                          (frame-position object)
                                          (mk-unique-name container (name self))))
           frame conec-to-me)
      (setf (frame-position newbox) (borne-position (frame-position object)))
      (setf (frame-size newbox) (frame-size object))
      (setf (frame-name newbox) (frame-name object))
      (setf (allow-lock newbox) (allow-lock object))
      (setf (value newbox) (eval (omNG-copy (value object))))
      (setf (inputs newbox) (eval (omNG-copy (inputs object))))
      (loop for input in (inputs object)
            for in in (inputs newbox) do
            (setf (connected? in) (connected? input)))
      (setf conec-to-me (get-conect-to-me object))
      (loop for item in conec-to-me do
            (change-conections  object item newbox))
      (setf frame (make-frame-from-callobj newbox))
      (omg-remove-element  container self)
      ;(compile-patch newpatch)
      (omg-add-element  container frame)
      (update-graphic-connections frame (get-elements (object container)))
      (omng-save newpatch)
      )))



