;OpenMusic
;
;Copyright (C) 1997, 1998, 1999, 2000 by IRCAM-Centre Georges Pompidou, Paris, France.
; 
;This program is free software; you can redistribute it and/or
;modify it under the terms of the GNU General Public License
;as published by the Free Software Foundation; either version 2
;of the License, or (at your option) any later version.
;
;See file LICENSE for further informations on licensing terms.
;
;This program is distributed in the hope that it will be useful,
;but WITHOUT ANY WARRANTY; without even the implied warranty of
;MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;GNU General Public License for more details.
;
;You should have received a copy of the GNU General Public License
;along with this program; if not, write to the Free Software
;Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
;
;Authors: Gerard Assayag and Augusto Agon

(in-package :om)


(defclass internepicture ()
   ((thepict :initform nil :accessor thepict :initarg :thepict )
    (name :initform nil :accessor name :initarg :name)))


(defmethod omNG-save ((self internepicture) &optional (values? nil))
  (om-save-as-bmp self)
  `(let ((newpict (make-instance ',(type-of self)
                    :name ',(name self))))
     (setf (thepict newpict) (om-get-picture ',(name self)))
     (setf (name newpict) ',(name self))
     newpict))

(defclass! picture (internepicture) ()
   (:icon 491))

(defmethod! picture-hdl ((self picture))
   :icon 491
   :doc "get the picture handle"
   (thepict self))

(defmethod good-val-p? ((self picture))
   (thepict self))

(defmethod draw-obj-in-rect ((self  picture) x x1 y y1 edparams  view)
   (if (thepict self)
       (om-draw-picture view (thepict self) (om-make-point x y) (om-make-point (- x1 x) (- y1 y)))))


(defmethod update-miniview ((self t) (type picture))  (om-invalidate-view self t))

(defmethod get-initval ((self picture)) (make-instance 'picture))

(defmethod get-boxsize ((self picture)) (om-make-point 80 80))

(defmethod print-object ((self picture) x) (format x "Picture"))

(defmethod make-one-instance ((class picture) &rest slots-vals) 
   (get-picture-file))

(defmethod* get-picture-file () 
   :initvals nil
   :indoc nil
   :doc "load a pict file"
   :icon 148
  (let* ((pict-info (choose-resource-dialog :pict :kernel t :user t))
         name)
    (when (consp pict-info)
        (setf name (intern (string+ (string (third pict-info)) "-" (string (first pict-info)))))
        (make-picture name))))

(defun make-picture (name)
   (let ((rep (make-instance 'picture :name name)))
     (setf (thepict rep) (om-get-picture name))
     (setf (name rep) name)
     rep))


(defun load-pict (name)
   (om-load-if name 'load-picture))

(defmethod omNG-copy ((self picture))
  `(let ((newpict (make-instance ',(type-of self))))
     (setf (thepict newpict) ,(thepict self))
     (setf (name newpict) ',(name self))
     newpict))

;========



;====================================================
;BOX and frame

(defmethod get-type-of-ed-box ((self picture))  'OMpicturebox)

(defclass OMpicturebox (OMBoxEditCall) ())
(defmethod get-frame-class ((self OMpicturebox)) 'PictboxFrame)

;---------
;FRAME
(omg-defclass PictboxFrame (boxEditorFrame) ()
   (:documentation "Simple frame for OMBoxEditCall meta objects. #enddoc#
#seealso# (OMBoxEditCall) #seealso#"))

(defmethod pict2bkg ((self PictboxFrame))
   (let* ((thepict (thepict (value (object self))))
          (patchpanel (om-view-container self))
          newpict)
     (when thepict
       (setf newpict (make-instance 'patch-picture))
       ;;;(setf (thepict newpict) (om-copy-pict-handler thepict))
       (setf (thepict newpict)  thepict)
       (setf (pict-pos newpict) (om-view-position self))
       (setf (name newpict) (name (value (object self))))
       (setf (pict-size newpict) (om-view-size (iconview self)))
       (push newpict (pictu-list (object patchpanel)))
       (omg-remove-element patchpanel self)
       (om-invalidate-view patchpanel t))))

(defmethod reinit-size ((box PictboxFrame))
   (let ((thepict (thepict (value (object box)))) newsize)
     (when thepict
       (setf newsize (om-get-picture-size thepict))
       (setf (frame-size (object box)) (om-add-points newsize (om-make-point 0 16)))
       (box-draw-connections box nil)
       (omG-select (redraw-frame box)))))


;====================================================

(AddClass2Pack '(picture) *data-package*)
;;;(omNG-add-element *data-package* (omNG-protect-object (find-class 'picture)))

(AddGenFun2Pack '(picture-hdl) *data-package*)
;;;(omNG-add-element *data-package* (omNG-protect-object (fdefinition 'picture-hdl)))










