;OpenMusic
;
;Copyright (C) 1997, 1998, 1999, 2000 by IRCAM-Centre Georges Pompidou, Paris, France.
; 
;This program is free software; you can redistribute it and/or
;modify it under the terms of the GNU General Public License
;as published by the Free Software Foundation; either version 2
;of the License, or (at your option) any later version.
;
;See file LICENSE for further informations on licensing terms.
;
;This program is distributed in the hope that it will be useful,
;but WITHOUT ANY WARRANTY; without even the implied warranty of
;MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;GNU General Public License for more details.
;
;You should have received a copy of the GNU General Public License
;along with this program; if not, write to the Free Software
;Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
;
;Authors: Gerard Assayag and Augusto Agon


;DocFile
;Define a class to handle drag-and-drop operations.
;Last Modifications :
;18/10/97 first date.
;DocFile
(in-package :om)

(defvar *OM-drag&drop-handler* "An object that handles the drag and drop of views at the lisp level.")

(defclass omdrag-drop ()
   ((dragged-view :initform nil :accessor dragged-view)
    (dragged-list-objs :initform nil :accessor dragged-list-objs)
    (container-view :initform nil :accessor container-view)
    (target-view :initform nil :accessor target-view)
    (true-target-view :initform nil :accessor true-target-view)
    (true-dragged-view :initform nil :accessor true-dragged-view)
    (opt-key-p :accessor opt-key-p :initform nil)
    (initial-mouse-pos :initform nil :accessor initial-mouse-pos)
    (drop-mouse-pos :initform nil :accessor drop-mouse-pos)
    (drag-flavor :initform nil :accessor drag-flavor)))

(defun init-D&D-handler ()
  (setf *OM-drag&drop-handler* (make-instance 'omdrag-drop)))


(om-add-init-func 'init-D&D-handler) 


(defvar *receiving-in-drag* nil)


(defmethod om-drag-start ((view om-view-drag))
   (let* ((theview (get-drag-object view))
          (container (om-view-container theview))
          (regionall (om-new-region))
          (viewpos (om-view-position theview))
          (x0 (om-point-h viewpos))
          (y0 (om-point-v viewpos))
          (i 2)
            actives dragged-lisp-objects)
     (unless (active-mode theview) 
       (unless (om-shift-key-p) 
         (mapc #'(lambda (subframe)
                   (omG-unselect subframe)) (get-actives container))))
     (omG-select theview)
     (setf actives (get-actives container))
     (mapc #'(lambda (oneobject)
                       (let* ((region (om-new-region)))
                         (setf region (make-drag-region oneobject region x0 y0 view))
                         (push oneobject dragged-lisp-objects)
                         (setf regionall (om-union-region region regionall))
                         (om-dispose-region region))
                         ) 
       actives)
     (setf (dragged-view *OM-drag&drop-handler*)  theview
        (dragged-list-objs *OM-drag&drop-handler*)  dragged-lisp-objects
       (container-view *OM-drag&drop-handler*)  container
        (true-dragged-view *OM-drag&drop-handler*) view
        (drag-flavor *OM-drag&drop-handler*) :omvw)
     (unwind-protect
         (om-drag-create-item-bounds view regionall)
       (om-dispose-region regionall))
     ) 
   t)




(defmethod om-drag-receive ((view om-view-drop) 
                            (dragged-view t)) 
  (if (equal (drag-flavor *OM-drag&drop-handler*) :omsc)
    (score-drag-receive view dragged-view)
    (let* ((drop-pos (om-mouse-position view)) ;; (drag-mouse-drop-position view))
           (*receiving-in-drag* t) rep)
      (setf (opt-key-p *OM-drag&drop-handler*)         (om-option-key-p)
            (target-view  *OM-drag&drop-handler*)      (get-drag-object view)
            (initial-mouse-pos *OM-drag&drop-handler*)  (om-get-current-drag-initpos (dragged-view *OM-drag&drop-handler*))
            
            (true-target-view *OM-drag&drop-handler*)  view
            (drop-mouse-pos *OM-drag&drop-handler*)  (get-pos-in-object view drop-pos))
      (setf rep (finalize-drag&drop *OM-drag&drop-handler*))
      ;;; test
      (mapc #'(lambda (oneobject) (om-invalidate-view oneobject))
            (dragged-list-objs *OM-drag&drop-handler*))
      rep
      )))

 
(defun in-non-visible-icon (D&DHandler)
   (or (and (icon-finder-p  (true-target-view  D&DHandler)) (icon-finder-p (dragged-view  D&DHandler)))
       (and (boxframe-p  (true-target-view  D&DHandler)) (boxframe-p (dragged-view  D&DHandler)))))
;---------------------------------------------------------------------------------
(defmethod finalize-drag&drop ((D&DHandler omdrag-drop))
  (cond 
   ((opt-key-p *OM-drag&drop-handler*)
    (perform-duplicate-view D&DHandler))
   ((in-non-visible-icon D&DHandler)
    (if (equal (om-view-container (true-target-view  D&DHandler)) (om-view-container (dragged-view  D&DHandler)))
      (when (allow-move-element (om-view-container (target-view  D&DHandler)))
        (perform-move-view D&DHandler)
        (make-move-after (om-view-container (true-target-view  D&DHandler)) (dragged-list-objs D&DHandler)))
      (progn
        (setf (target-view  D&DHandler) (om-view-container (true-target-view  D&DHandler)))
        (perform-change-view D&DHandler)
        )))
    ((eq (target-view D&DHandler) (om-view-container (dragged-view  D&DHandler)))
     (when (allow-move-element (target-view  D&DHandler))
       (perform-move-view D&DHandler)
       (make-move-after (target-view  D&DHandler) (dragged-list-objs D&DHandler))))
    ((eq (target-view  D&DHandler) (dragged-view  D&DHandler))
     (when (allow-move-element (om-view-container (target-view  D&DHandler)))
       (perform-special-move-view D&DHandler)
       (make-move-after (om-view-container (target-view  D&DHandler)) (dragged-list-objs D&DHandler))))
    (t (perform-change-view D&DHandler)))
  )

;------DUPLICATE WITH OPT_KEY
(defmethod perform-duplicate-view ((D&DHandler omdrag-drop))
   (let ((target-frame (get-drag-object (target-view  D&DHandler)))
         (pos0 (get-relative-position (get-drag-object (dragged-view D&DHandler))))
         (some-item-used t)
         nameerr)
     (when (allow-move-element (target-view  D&DHandler))
       (mapc #'(lambda (dragged-frame)
                 (when  (or (no-allow-copy-p (object dragged-frame))
                            (not (drop-allow-p D&DHandler (object dragged-frame) (object target-frame)))
                            (null (can-copy (object target-frame) (object dragged-frame))))
                   (setf nameerr (name (object dragged-frame)))
                   (setf some-item-used nil))) (dragged-list-objs D&DHandler))
       (if some-item-used
         (let* ((subframes (mapcar #'(lambda (frame)
                                       (object frame)) (dragged-list-objs D&DHandler)))
                (copies (mapcar #'(lambda (frame)
                                    (eval (omNG-copy frame))) subframes)))
           (setf some-item-used (perform-duplicate-list D&DHandler (object target-frame) target-frame subframes copies pos0))
           )
         (om-beep-msg (string+ "The object " nameerr " can not be copied to this window")))
       )
     some-item-used))

                 
;------MOVE IN THE SAME WINDOW
(defmethod perform-move-view ((D&DHandler omdrag-drop))
  (let ((pos0 (get-position (dragged-view D&DHandler))))
    (mapc #'(lambda (oneobject)
              (OMGMoveObject oneobject
                             (om-add-points
                              (om-subtract-points (get-position oneobject) pos0) 
                              (om-subtract-points (drop-mouse-pos D&DHandler) (initial-mouse-pos D&DHandler)))))
      (dragged-list-objs D&DHandler)) t)
  )


;Move in the same object
(defmethod perform-special-move-view ((D&DHandler omdrag-drop))
   (mapc #'(lambda (oneobject)
             (OMGMoveObject oneobject
                            (om-add-points (get-position oneobject)
                                           (om-subtract-points (drop-mouse-pos D&DHandler) (initial-mouse-pos D&DHandler)))))
     (dragged-list-objs D&DHandler)) 
  t)


;----------WHEN THE TARGET AND THE SOURCES WINDOWS ARE DIFFERENTS

;;; si on drop plusieurs objets sur une mme cible, 
;; est ce qu'on renvoie la liste (t)
;; ou bien on fait des drops sur chaque objet (nil) 
(defmethod allow-drag-list ((self t)) nil)

(defmethod allow-drag-list ((self omgenericfunction)) 
  (function-allow-list self))

(defmethod perform-drop-list ((dd t) (dragged t) (target t) pos) nil)

(defmethod perform-change-view ((D&DHandler omdrag-drop))
  (let* ((target-frame (get-drag-object (target-view  D&DHandler)))
         (pos0 (get-position (get-drag-object (dragged-view D&DHandler))))
         (list-objs (mapcar #'(lambda (frame) (object frame)) (dragged-list-objs D&DHandler)))
         ;; c'etait en comment
         (connectlist (save-connections target-frame (container-view D&DHandler) list-objs))
         (some-item-used nil)
         (correctmove t))
    (loop for item in (dragged-list-objs D&DHandler)
          while correctmove do
          (setf correctmove (drop-allow-p D&DHandler (object item) (object target-frame))))
    (if correctmove
      (progn 
        ;; c'etait en comment
        (make-delete-before (container-view D&DHandler) (dragged-list-objs D&DHandler) target-frame)
        (if (allow-drag-list (object target-frame))
          (setf some-item-used (perform-drop-list D&DHandler (dragged-list-objs D&DHandler) target-frame
                                                  (om-add-points
                                                   (om-subtract-points (get-position (car (dragged-list-objs D&DHandler))) pos0) 
                                                   (om-subtract-points (drop-mouse-pos D&DHandler) (initial-mouse-pos D&DHandler)))))
          (mapc #'(lambda (dragged-frame)
                    (setf some-item-used (perform-drop D&DHandler dragged-frame target-frame
                                                       (om-add-points
                                                        (om-subtract-points (get-position dragged-frame) pos0) 
                                                        (om-subtract-points (drop-mouse-pos D&DHandler) (initial-mouse-pos D&DHandler))))))
                (dragged-list-objs D&DHandler))
          )
        ;; c'etait en comment
        (remake-draggeds-connections target-frame (container-view D&DHandler)  list-objs connectlist)
        ;;; new attention a tester...
        (when (patchpanel-p target-frame) (modify-patch target-frame))
        )
      (om-beep))
    (om-invalidate-view target-frame) 
    some-item-used))

  

; ---------------------------------------------------------------------
;To subclass
; ---------------------------------------------------------------------

(defmethod make-move-after           ((self t) list) 
   "This method is called after moving one or more elements of self."
   (declare (ignore list))
   t)

(defmethod receive-file              ((self t) (type t) path pos) (declare (ignore path pos))  nil)
(defmethod save-connections          ((source t) (target t) list) 
   "Save the connections beetwen boses of 'self' as a list."
   (declare (ignore list))
   nil)
(defmethod make-delete-before        ((source t) (target t) (cible t))
   "This method is called after deleting one or elements of source."
   nil)
(defmethod external-object-p         ((self t)) nil)


(defmethod get-drag-object           ((self t)) self)
(defmethod get-drag-object           ((self om-view)) self)
(defmethod get-position              ((self om-view)) (om-view-position self))
(defmethod get-relative-position     ((self om-view)) (om-view-position self))



(defmethod get-pos-in-object         ((self t) where) where)
(defmethod allow-move-element        ((self t))
   "T if you can change the position of a self subviews, the default is T"
   t)
(defmethod drop-allow-p              ((D&DHandler omdrag-drop) (dragged t) (target t)) nil)

(defmethod remake-draggeds-connections ((target t) (source t) listobj list) 
   "Used to remake connections when you drag and drop a subset of boxes from 'self' to 'source'."
  (declare (ignore list listobj)) nil)

(defmethod make-drag-region ((self om-view) region x0 y0 view) 
   (declare (ignore view))
   (om-set-rect-region region (- (x self) x0) (- (y self) y0) (- (x+w self) x0) (- (y+h self) y0))
   region)


(defmethod perform-drop ((D&DHandler omdrag-drop) (dragged t) (target t) position)
   (declare (ignore position)) nil)

;----------------------------------
;Icon + Drag&Drop Action
;----------------------------------
(defmethod do-default-action ((self t) icon)
  (declare (ignore icon)) t)

(defclass input-maker () ())

(defmethod do-default-action ((self input-maker) icon)
  (let* ((thescroller (om-view-container icon))
         (boxes (get-subframes thescroller)) 
         (i (length (find-class-boxes boxes 'TypedInFrame)))
         (pos (om-make-point (+ 5 (* i 50)) 45)))
    (omG-add-element thescroller
                     (make-frame-from-callobj 
                      (make-new-typed-input (mk-unique-name thescroller "input") 't (+ i 1) pos)))))

(defmethod drop-allow-p ((D&DHandler omdrag-drop) (dragged OMClass) (target input-maker)) t)

(defmethod drop-allow-p ((D&DHandler omdrag-drop) (dragged OMBasictype) (target input-maker)) t)

(defmethod drop-allow-p ((D&DHandler omdrag-drop) (dragged OMBoxEditCall) (target input-maker)) t)

(defmethod drop-allow-p ((D&DHandler omdrag-drop) (dragged OMBoxInstance) (target input-maker)) t)

(defmethod perform-drop ((D&DHandler omdrag-drop) (dragged class-icon-frame) (target unaire-fun-icon) position)
  (declare (ignore position))
  (funcall (action target) dragged))

(defmethod perform-drop ((D&DHandler omdrag-drop) (dragged type-icon-frame) 
                         (target unaire-fun-icon) position)
  (declare (ignore position))
  (funcall (action target) dragged))

(defmethod perform-drop ((D&DHandler omdrag-drop) (dragged instBoxframe) 
                         (target unaire-fun-icon) position)
  (declare (ignore position))
  (let* ((thetype (class-name (class-of (value (object dragged)))))
         (boxes (get-subframes (om-view-container target))) 
         (i (length (find-class-boxes boxes 'TypedInFrame)))
         (pos (om-make-point (+ 5 (* i 50)) 45))
         (new-input (make-new-typed-input (mk-unique-name (om-view-container target) "input")
                                          thetype (+ i 1) pos)))
    (setf (defval new-input) (omng-copy (value (object dragged))))
    (omG-add-element (om-view-container target)
                     (make-frame-from-callobj  new-input))))

(defmethod perform-drop ((D&DHandler omdrag-drop) (dragged boxEditorFrame) 
                         (target unaire-fun-icon) position)
  (declare (ignore position))
  (let* ((thetype (class-name (class-of (value (object dragged)))))
         (boxes (get-subframes (om-view-container target))) 
         (i (length (find-class-boxes boxes 'TypedInFrame)))
         (pos (om-make-point (+ 5 (* i 50)) 45))
         (new-input (make-new-typed-input (mk-unique-name (om-view-container target) "input")
                                          thetype (+ i 1) pos)))
    (setf (defval new-input) (omng-copy (value (object dragged))))
    (omG-add-element (om-view-container target)
                     (make-frame-from-callobj  new-input))))


;===================================================
;PERSISTANTS TO FINDER
;===================================================
;;; supprim 
