;OpenMusic
;
;Copyright (C) 1997, 1998, 1999, 2000 by IRCAM-Centre Georges Pompidou, Paris, France.
; 
;This program is free software; you can redistribute it and/or
;modify it under the terms of the GNU General Public License
;as published by the Free Software Foundation; either version 2
;of the License, or (at your option) any later version.
;
;See file LICENSE for further informations on licensing terms.
;
;This program is distributed in the hope that it will be useful,
;but WITHOUT ANY WARRANTY; without even the implied warranty of
;MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;GNU General Public License for more details.
;
;You should have received a copy of the GNU General Public License
;along with this program; if not, write to the Free Software
;Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
;
;Authors: Gerard Assayag and Augusto Agon


;DocFile
;This File contains functions and macros used for graphical porpuse.
;Documentations are not in the html doc.
;In particular this files implements QuickDraw high level functions
;and view methods.
;Last Modifications :
;18/10/97 first date.
;DocFile

(in-package :om)


;========= OM COLORS ===========

(defun color2list (color)
   (list (float (om-color-r color)) (float (om-color-g color)) (float (om-color-b color))))

(defun make-color-255 (r g b)
  (om-make-color (/ r 255.0) (/ g 255.0) (/ b 255.0)))

(defvar *om-white-color* nil)
(defvar *om-black-color* nil)
(defvar *om-gray-color* nil)
(defvar *om-red-color* nil)
(defvar *om-red2-color* nil) 
(defvar *om-blue-color* nil)
(defvar *om-green-color* nil)
(defvar *om-orange-color* nil)
(defvar *om-pink-color* nil) 
(defvar *om-wine-color* nil)
(defvar *om-dark-blue-color* nil)
(defvar *om-yellow-color* nil) 
(defvar *om-purple-color* nil)
(defvar *om-light-blue-color* nil) 
(defvar *om-cafe-color* nil)
(defvar *om-green2-color* nil)
(defvar *om-salmon-color* nil)
(defvar *om-steel-blue-color* nil)
(defvar *om-beige-color* nil)
(defvar *workspace-color* nil "the default color of the workspace")
(defvar *maq-color* nil "the default bg color of the maquette")
(defvar *azulito* nil)
(defvar *azulote* nil) 
(defvar *def-tempbox-color* nil)
(defvar *undefbox-color* nil)
(defvar *16-color-list* nil)
(defvar *controls-color* nil)
(defvar *controls-color+* nil)
(defvar *controls-color++* nil)
(defvar *controls-color-* nil)
(defvar *controls-color--* nil)
(defvar *editor-bar-color* nil)
(defvar *editor-bar-color++* nil)
(defvar *editor-bar-color+* nil)
(defvar *editor-bar-color--* nil)
(defvar *editor-bar-color-* nil)
(defvar *instboxframe-color* nil "def color for instance boxes")
(defvar *patch-box-color* nil)
(defvar *maquette-box-color* nil)
(defvar *obj-box-color* nil)
(defvar *scorepatch-color* nil "la couleur du score panel en mode patch")
(defvar *note-color* nil)
(defvar *system-color* nil)
(defvar *select-color* nil)
(defvar *ornement-color* nil)
(defvar *accident-color* nil)
(defvar *obje-color-list* nil)

(defun init-om-color-vars ()
  (setf *accident-color* (om-make-color 0.1 0.7 0.6))
  (setf *ornement-color* (om-make-color 0.9 0.2 0.2))
  (setf *scorepatch-color* (om-make-color 0.5 0.62 0.55))
  (setf *patch-box-color* (om-make-color 0.5 0.5 0.6))
  (setf *maquette-box-color* (om-make-color 0.6 0.5 0.5))
  (setf *obj-box-color* (om-make-color 0.5 0.6 0.5))
  (setf *instboxframe-color* (om-make-color 0.68 0.79 0.73))
  (setf *editor-bar-color* (make-color-255 210 210 210))
  (setf *editor-bar-color++* (make-color-255 170 170 170))
  (setf *editor-bar-color+* (make-color-255 235 235 235))
  (setf *editor-bar-color--* (make-color-255 110 110 110))
  (setf *editor-bar-color-* (make-color-255 175 175 175))
  (setf *om-white-color* (om-make-color 1 1 1))
  (setf *om-black-color* (om-make-color 0 0 0))
  (setf *om-gray-color* (om-make-color 0.5 0.5 0.5))
  (setf *om-red-color* (om-make-color 1 0 0))
  (setf *om-red2-color* (om-make-color 0.7 0.3 0.3)) 
  (setf *om-blue-color* (om-make-color 0 0 1))
  (setf *om-green-color* (om-make-color 0 1 0))
  (setf *om-orange-color* (om-make-color 1 0.5 0.25))
  (setf *om-pink-color* (om-make-color 1 0 1)) 
  (setf *om-wine-color* (om-make-color 0.50 0 0))
  (setf *om-dark-blue-color* (om-make-color 0.5 0 0))
  (setf *om-yellow-color* (om-make-color 1 1 0)) 
  (setf *om-purple-color* (om-make-color 0.75 0.61 0.98))
  (setf *om-light-blue-color* (om-make-color 0 1 1)) 
  (setf *om-cafe-color* (om-make-color 0.5 0.25 0.25))
  (setf *om-green2-color* (om-make-color 0.37 0.73 0.62))
  (setf *om-salmon-color* (om-make-color 0.99 0.62 0.45))
  (setf *om-steel-blue-color* (om-make-color 0.41 0.54 0.67))
  (setf *om-beige-color* (om-make-color 0.94 0.94 0.85))
  (setf *workspace-color*  (om-make-color 0.803 0.854 0.855))
  (setf *maq-color* (om-make-color 0.53333336 0.60784316 0.6))
  (setf *azulito* (om-make-color 0.772 0.855 0.788))
  (setf *azulote* (om-make-color 0.47 0.553 0.353))
  (setf *def-tempbox-color* (om-make-color 0.772 0.855 0.788))
  (setf *undefbox-color*(om-make-color 0.801 0.208 0.098))
  (setf *controls-color* (make-color-255 230 230 230)) 
  (setf *controls-color+* (make-color-255 240 240 240)) 
  (setf *controls-color++* (make-color-255 180 180 180)) 
  (setf *controls-color-* (make-color-255 155 155 155)) 
  (setf *controls-color--* (make-color-255 135 135 135))
  (setf *16-color-list* 
	(list *om-blue-color* *om-red-color* *om-green-color* *om-orange-color*
	      *om-pink-color* *om-wine-color* *om-dark-blue-color* *om-yellow-color*
	      *om-black-color* *om-gray-color* *om-purple-color* *om-light-blue-color*
	      *om-cafe-color* *om-green2-color* *om-salmon-color* *om-steel-blue-color*))
  (setf *obje-color-list* (list (om-make-color 0.525 0.769 0.549) 
				(om-make-color 0.51 0.49 0.788) 
				(om-make-color 0.851 0.851 0.475)
				(om-make-color 0.671 0.768 0.749)
				(om-make-color 0.812 1.0 1.0)
				(om-make-color 0.773 0.969 0.729)
				(om-make-color 0.812 1.0 1.0)))
  (setf *note-color* *om-black-color*)
  (setf *system-color* *om-black-color*)
  (setf *select-color* *om-select-color*))

(om-add-init-func 'init-om-color-vars)

(defun om-interpole-colors (begin end steps)
  (if (< steps 2)
    (list begin end)
  (let* ((difR (/ (- (om-color-r end) (om-color-r begin)) steps))
         (difG (/ (- (om-color-g end) (om-color-g begin)) steps))
         (difB (/ (- (om-color-b end) (om-color-b begin)) steps)))
    (loop for i from 0 to (- steps 1)
          collect (om-make-color
                   (float (+ (* i difR) (om-color-r begin)))
                   (float (+ (* i difG) (om-color-g begin)))
                   (float (+ (* i difB) (om-color-b begin))))))))


;=====FONTS

(defvar *controls-font* nil)

(defun init-om-fonts-vars ()
  (setf *controls-font* (om-make-font "LucidaGrande" 13)))

(om-add-init-func 'init-om-fonts-vars)

;;;;======================================
;;;; Graphic components utilities 
;;;; muste be called on graphic objects only (window, view,dialog-item,...)
;;;;======================================
(defmethod x   ((self t)) (om-point-h (om-view-position self)))
(defmethod y   ((self t)) (om-point-v (om-view-position self)))
(defmethod w   ((self t)) (om-point-h (om-view-size self)))
(defmethod h   ((self t)) (om-point-v (om-view-size self)))
(defmethod x+w ((self t)) (+ (x self)(w self)))
(defmethod y+h ((self t)) (+ (y self)(h self)))



;;;=== position in window ===
(defun view-position-win (view)
  (let ((rep (om-make-point 0 0))
        (init view)
        (win (om-view-window view)))
    (loop while (not (equal init win)) do
          (setf rep (om-add-points rep (om-view-position init)))
          (setf init (om-view-container init)))
    rep))


(defmethod window ((self t))
  (om-view-window self))


;;;====== POSITIONS AND BOXES ====

(defun point-max (point-1 point-2)
  (om-make-point (max (om-point-h point-1) (om-point-h point-2))
                                (max (om-point-v point-1) (om-point-v point-2))))

(defun rect-from-points (x y x1 y1)
   (om-pts-to-rect (om-make-point x y) (om-make-point x1 y1)))
   
(defun points-from-rect (rect)
  (list (om-rect-left rect)
         (om-rect-top rect)
         (om-rect-right rect) 
         (om-rect-bottom rect)))

;;;True if the point (x1 y1) is into the rectangle ((x,y) (x+w,y+h))
(defun inside-rectangle? (x1 y1 x y w h)
  (let* ((topleft (om-make-point x y))
           (rect (om-pts-to-rect topleft (om-add-points topleft (om-make-point w h))))
           (point (om-make-point x1 y1)))
          (om-point-in-rect-p point rect)))


;;;;-------select in x --------

(defun draw-h-rectangle (rec &optional fill mode)
  (when rec
    (if fill
      (om-draw-hilite-rect (car rec) (second rec) (- (third rec) (car rec) ) (- (fourth rec) (second rec)) mode)
      (om-with-fg-color nil *om-select-color*
        (om-draw-rect (car rec) (second rec) (- (third rec) (car rec) ) (- (fourth rec) (second rec)))))))



;;;===== SPECIAL DIALOG ITEMS =====

;;; size for names in dialog items
(defun get-name-size (name &optional (font *om-default-font2*))
  (+ 5 (om-string-size name font)))


;=======================================================================
;  invisible resize box
;=======================================================================
(omg-defclass c-resize-box (om-transparent-view) ()
   (:documentation "This subclass the view allow resize another view by click and draw within it.#enddoc#
#seealso# (box-frame) #seealso#"))

(defmethod om-view-cursor ((self c-resize-box))
   *om-resize-cursor*)

(defmethod om-view-click-handler ((self c-resize-box) where)
   (declare (ignore where))
   (let* ((boxframe (om-view-container self))
          (theeditor (editor (om-view-container boxframe)))
          (panel (om-view-container (om-view-container self)))
          (rx (x boxframe)) 
          (ry (y  boxframe)))
     (when (text-view theeditor)
       (exit-from-dialog (text-view theeditor) 
                                       (om-dialog-item-text (text-view theeditor))))
     (om-new-movable-object panel rx ry (w boxframe) (h boxframe) 'om-movable-rectangle)
     (om-init-motion-functions self 'resize-box-motion 'resize-box-release)
    ))
      

(defmethod resize-box-motion ((self c-resize-box) pos)
  (let* (  (panel (om-view-container (om-view-container self)))
           (initpoint (om-convert-coordinates pos self panel ))
           (rx (om-point-h initpoint))
           (ry (om-point-v initpoint))
           (rect  (om-init-point-movable-object panel)))
     (om-update-movable-object panel (first rect) (second rect) (max 4  (- rx (first rect) )) (max 4 (- ry (second rect) )))))

(defmethod resize-box-release ((view c-resize-box) pos) 
   (let* ((panel (om-view-container (om-view-container view)))
          (initpoint (om-convert-coordinates pos view panel ))
          (rx (om-point-h initpoint))
          (ry (om-point-v initpoint))
          (boxframe (om-view-container view))
          (rect  (om-init-point-movable-object panel))
          ctrl)
     (om-erase-movable-object panel) 
     (change-boxframe-size boxframe (om-make-point (- rx (first rect) ) (- ry (second rect) )))
     (om-set-view-position view (om-make-point (- (w boxframe) 8) (- (h boxframe) 8)))
       ))

(defmethod add-box-resize ((self om-view))
   "add a resize view to the wiew 'self'"
   (om-add-subviews self
     (om-make-view 'c-resize-box
       :size (om-make-point 8 8)
       :position (om-make-point (- (w self) 8) (- (h self) 8)))))



;=======================================================================
; A fred-window that does not ask for saving
;=======================================================================



;;; not-save-fred-window
;(defclass not-save-edit-window (om-text-edit-window) ()
;   (:default-initargs :ask-for-save nil)
;   )



;=========================================================================
; POP-UP MENUS
;========================================================================

(omg-defclass pair-pop-up-menu (om-pop-up-menu) 
  ((dialo :initform nil :accessor dialo)))

(defmethod update-dailo ((self pair-pop-up-menu) text)
  (when (dialo self)
    (om-set-dialog-item-text (dialo self) text)
    (om-invalidate-view (dialo self) t)))


(defun cons-pair-pop-menu (itemtext fun container list-vals default-item pos size)
  (let (menu itemlist default)
    (loop for item in list-vals 
        for i = 0 then (+ i 1) do
          (let ((title (car item))
                (arg (second item)))
            (if (equal arg default-item)
                (setf default i))
            (push (om-new-leafmenu  title #'(lambda ()
                                              (when (funcall fun container arg)  
                                                (update-dailo menu title))))
                  itemlist)))
    
    (setf menu (om-create-menu 'pair-pop-up-menu (reverse itemlist)))
    
    (when default
        (om-set-menu-default-item menu default))

    (when itemtext
        (setf (dialo menu) itemtext))
    
    (om-open-pop-up-menu menu container)
    
    ))



;;;===========================
;;; Resources
;;;===========================



;;;===================
;;; OM cursors 
;;;===================

(defvar *om-box-cursor* nil)
(defvar *om-contex-cursor* nil)
(defvar *om-loupe-cursor* nil)
(defvar *om-tree-cursor* nil)
(defvar *om-pack-cursor* nil)
(defvar *om-point-cursor* nil)
(defvar *om-hand-cursor* nil)
(defvar *om-hand-bpf-cursor* nil)



(defun init-curs ()
   (setf *om-box-cursor* (om-make-cursor "box-cursor"))
   (setf *om-tree-cursor* (om-make-cursor "tree-cursor"))
   (setf *om-pack-cursor* (om-make-cursor "pack-cursor"))
   (setf *om-contex-cursor* (om-make-cursor "contex-cursor"))
   (setf *om-loupe-cursor* (om-make-cursor "loupe-cursor" (om-make-point 6 6)))
   (setf *om-point-cursor* (om-make-cursor "point-cursor" (om-make-point 4 4)))
   (setf *om-hand-cursor* (om-make-cursor "main-cursor" (om-make-point 10 10)))
   (setf *om-hand-bpf-cursor* (om-make-cursor "main-bpf-cursor" (om-make-point 10 10)))
   )

(om-add-init-func 'init-curs)

;;;=======================
;;; d&d classes
;;;=======================

(defclass om-view-drop (om-drop-view) ()
   ;(:default-initargs nil
     ;;;:drag-allow-move-p t
     ;;;:drag-accepted-flavor-list (list :|OMVW| :|hfs | #$flavorTypePromiseHFS :|PICT| :|OMSC|))
     ;:drag-accepted-flavor-list (list :|OMVW| :|PICT| :|OMSC|)
   ;  )
   (:documentation "Abstract class, all view which wants to accept drops must inherit from this class.#enddoc#")
   )

(defclass om-view-drag (om-drag-view om-view-drop) () 
   (:documentation "Abstract class, all view which wants to be draggable and droppable must inherit from this class.#enddoc#")
   )

;=======================================================================
; Text dialog item which allow drop used for names in boxes and ws icons
; this class allow its text edition used in controls
;=======================================================================
(omg-defclass om-static-text-drag (om-view-drag om-static-text)
   ((selected-p :initform nil :accessor selected-p))
   (:documentation "This is the class for all graphic-text which allow drag and drop
i.e. boxes'names, default values in slots, etc. #enddoc#
It's posible also to edit this text by double click into it,
i.e. inputs edition, change the name for a folder, etc.
#seealso# (box-dialog-name ttybox icon-finder-name) #seealso#
#selected-p# used to hilite the text if this is selected.#selected-p#"))

(defmethod (setf selected-p) (selected-p (self om-static-text-drag)) 
   "Set the slot selected-p and hilite the text if selected-p equal True"
   (if selected-p
       (progn
         (om-set-part-color self :body *om-select-color*)
         (om-set-part-color self :text *om-white-color*))
     (let* ((panel (om-view-container self))
            (tmppan panel))
       (when panel
         ;;; a mettre dans l'api si transparent n'est pas gr par le systeme
         (loop while (and tmppan (not (or (metaobj-panel-p tmppan) (equal tmppan (window panel))))) do
               (setf tmppan (om-view-container tmppan))))
       (if (metaobj-panel-p tmppan) (setf panel tmppan))
       (om-set-part-color self :text *om-black-color*)
       (om-set-part-color self :body :transparent) ; (if panel  (om-get-bg-color panel) *om-white-color*))
       )))


(defmethod om-drag-selection-p ((self om-static-text-drag) mouse-position)
   "The Drag is allowed for om-static-text-drag if the control key is not down"
   (declare (ignore mouse-position))
   (not (om-control-key-p)))


;;;=======================
;;; drop view
;;;=======================


(omg-defclass drop-area (OMSimpleFrame om-view-drop)
  ())

(defmethod get-drag-object ((self drop-area)) self)

(defmethod om-drag-selection-p ((self drop-area) mouse-position)
  (declare (ignore mouse-position)) nil)

(defmethod om-view-drag-hilite-p ((self drop-area)) t)

(defmethod om-drag-enter-view ((self drop-area))
  (call-next-method))


;=== simple horizontal bar component ===
(omg-defclass bar-item (om-view) 
  ((fg-color :accessor fg-color :initarg :fg-color :initform *om-black-color*)))
(defmethod om-draw-contents ((self bar-item))
   (call-next-method)
   (om-with-focused-view  self 
     (om-with-fg-color self (fg-color self)
       (om-draw-line 0 0 (w self) 0)
       )
     ))




;;;=====================
;;; 3DBORDER VIEW
;;;=====================

(omg-defclass 3Dborder-view (om-view)
  ((c+ :accessor c+ :initform (om-make-color 0.835 0.835 0.843) :initarg :c+)
   (c++ :accessor c++ :initform (om-make-color 0.87 0.87 0.88) :initarg :c++)
   (c- :accessor c- :initform (om-make-color 0.604 0.604 0.604) :initarg :c-)
   (c-- :accessor c-- :initform (om-make-color 0.514 0.514 0.514) :initarg :c--))
  )


(defmethod om-draw-contents ((self 3Dborder-view))
  (call-next-method)
   (let ((x (om-h-scroll-position self))
        (y (om-v-scroll-position self))
        (w (om-point-h (om-interior-size self)))
        (h (om-point-v (om-interior-size self))))
    (let ((xx (+ x w))
          (yy (+ y h)))
      (om-with-focused-view self
                          (om-with-fg-color self (c++ self)
                                            (om-draw-line (+ x 1) y (- xx 1) y) 
                                            (om-draw-line x (+ y 1) x (- yy 1)))
                          (om-with-fg-color self (c+ self)
                                            (om-draw-line (+ x 2) (+ y 1) (- xx 2) (+ y 1)) 
                                            (om-draw-line (+ x 1) (+ y 2) (+ x 1) (- yy 2)))
                          (om-with-fg-color self (c-- self)
                                            (om-draw-line (+ x 1) (- yy 1) (- xx 1) (- yy 1)) 
                                            (om-draw-line (- xx 1) (+ y 1) (- xx 1) (- yy 1)))
                          (om-with-fg-color self (c- self)
                                            (om-draw-line (+ x 2) (- yy 2) (- xx 2) (- yy 2)) 
                                            (om-draw-line (- xx 2) (+ y 2) (- xx 2) (- yy 2)))
                          ))))


;;; ====================
;;; TITLE BAR 
;;;=====================

(defvar *titlebars-h* nil)
(setf *titlebars-h* 26)

(omg-defclass editor-titlebar (3dborder-view) ())


;;; ====================
;;; NEW SUPERCLASS 
;;;=====================

(omg-defclass object-editor (om-view) 
  ((title-bar :initarg :title-bar :accessor title-bar :initform nil)))

(defmethod get-control-h ((self object-editor)) 80)

(defmethod editor-minimum-size ((self object-editor)) (om-make-point 300 100))

(defmethod get-titlebar-class ((self object-editor)) 'editor-titlebar)

(defmethod init-titlebar ((self editor-titlebar))
  (let ((name (string (class-name (class-of (object (om-view-container self)))))))
    (om-add-subviews self
                     (om-make-dialog-item 'om-static-text (om-make-point 10 2) 
                                          (om-make-point (+ (om-string-size name *om-default-font2b*) 4)
                                                         18)
                                        name
                                        :bg-color *editor-bar-color*
                                        :fg-color *om-dark-gray-color*
                                        :font *om-default-font2b*
                                        ))))

(defmethod change-text ((self editor-titlebar) text)
  (om-set-dialog-item-text (car (om-subviews self)) text))

(defmethod initialize-instance :after ((self object-editor) &rest initargs)
  (om-add-subviews self
                   (setf (title-bar self) (om-make-view (get-titlebar-class self) :position (om-make-point 0 0)
                                                       :size (om-make-point (w self) *titlebars-h*)
                                                       :bg-color *editor-bar-color*
                                                       :c++ *editor-bar-color++*
                                                       :c+ *editor-bar-color+*
                                                       :c-- *editor-bar-color--*
                                                       :c- *editor-bar-color-*
                                                       )))
  (init-titlebar (title-bar self))
  )

(defmethod update-subviews ((self object-editor))
   (when (title-bar self)
     (om-set-view-size  (title-bar self) (om-make-point (w self) *titlebars-h*)))
   (om-invalidate-view self))



