;OpenMusic
;
;Copyright (C) 1997, 1998, 1999, 2000 by IRCAM-Centre Georges Pompidou, Paris, France.
; 
;This program is free software; you can redistribute it and/or
;modify it under the terms of the GNU General Public License
;as published by the Free Software Foundation; either version 2
;of the License, or (at your option) any later version.
;
;See file LICENSE for further informations on licensing terms.
;
;This program is distributed in the hope that it will be useful,
;but WITHOUT ANY WARRANTY; without even the implied warranty of
;MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;GNU General Public License for more details.
;
;You should have received a copy of the GNU General Public License
;along with this program; if not, write to the Free Software
;Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
;
;Authors: Gerard Assayag and Augusto Agon


;DocFile
;This File implements the class icon-view (a simple view which plots
;an icon in itself. 
;Last Modifications :
;18/10/97 first date.
;DocFile


(in-package :om)

;;;=== ICONS ======
(defvar *default-icon* 179 "Icon by default")

;icon pos if the icon is already loaded
(defun loaded-icon-p (icon)
   (declare (special *om-package-tree*))
   (position icon (icon-resources *om-package-tree*) :test #'(lambda (x y) (= x (car y)))))

;get the icon handler from the loaded list
(defun get-icon-loaded (icon)
   (declare (special *om-package-tree*))
   (let ((n (loaded-icon-p icon)))
     (when n
       (second (nth (loaded-icon-p icon) (icon-resources *om-package-tree*))))))


; load the icon list (id handle) for the om default-icon
(defun load-default-icon-list ()
   (let (list resource) 
     (without-interrupts
      (setq resource (om-load-icon *default-icon*))
      (when resource
        (push (list *default-icon* resource) list)))
     list))


;;;========================================
;;; ICONS ACCESSORS
;;;========================================

; returns list (iconID iconHandle) for an icon number
; correct ID/Handle if problem
(defmethod get&corrige-icon ((icon integer))
   (declare (special *om-package-tree*))
   (let (iconhdl)
     (unless (loaded-icon-p icon)
       (setq iconhdl (om-load-icon icon))
       (if iconhdl
         (push (list icon iconhdl) (icon-resources *om-package-tree*))
         (setf icon *default-icon*)))
     (list icon (get-icon-loaded icon))))

; if null returns default icon
(defmethod get&corrige-icon ((icon null))
   (get&corrige-icon *default-icon*))

; if list (id package) return (iconID iconHdler) from the package resources
(defmethod get&corrige-icon ((icon list))
   (let* ((iconId (first icon))
          (icon-list (icon-resources (second icon)))
          (pos (position iconid icon-list :test #'(lambda (x y) (= x (car y))))))
     (if pos 
       (list icon (second (nth pos icon-list)))
       (get&corrige-icon iconId))))


;;;=== ICON VIEW =======
(omg-defclass icon-view (om-transparent-view) 
;(defclass icon-view (om-view) 
   ((iconID :initform 1  :initarg :iconID :accessor iconID)
    (selected-p :initform nil :accessor selected-p))
   (:documentation "This class implement a simple view that draw one icon in it.
this class inherite from view.#enddoc#
#iconID# This slot stocks an ID for Cicn resources in the aplication. #iconID#
#selected-p# Used to draw the icon in selected or unselected mode.#selected-p#"))

(defmethod icon-view-p ((self t)) nil)
(defmethod icon-view-p ((self icon-view)) t)


(defmethod initialize-instance :after ((self icon-view) &key controls)
   (declare (ignore controls))
   (get&corrige-icon (iconID self)))


(defmethod plot-icon-view ((self icon-view) &optional (selected nil))
  "Plot the icon referenced by (iconID self) in the view self,
selected parameter is used to draw the icon in normal or selected mode"
  (without-interrupts
   (let* ((icon (iconID self))
          (iconparams (get&corrige-icon icon))
          (iconhdlr (second iconparams)))
     (setf (iconID self) (first iconparams))
     (om-with-focused-view self
       (om-draw-icon iconhdlr self (om-make-point 0 0) (om-view-size self) selected)
       )
     )
   )
  )


(defmethod om-draw-contents ((self icon-view))
   (call-next-method)
   (when (om-view-window self)
     (plot-icon-view self (selected-p self))
     ))
   
;get the point ( (min 64 sizex) , (min 64 sizey)) where size is the size of the icon
(defun icon-sizes (Icon)
   (without-interrupts
    (let (resource)
      (setf resource (second (get&corrige-icon icon)))
      (list (min 64 (om-pict-width resource))
            (min 64 (om-pict-height resource))))))

(defmethod om-view-cursor :around ((self icon-view))
   (if (and (om-control-key-p) (or (boxframe-p (om-view-container self)) (icon-finder-p (om-view-container self)))) 
       *om-contex-cursor*
       *om-arrow-cursor*
     ))

;==========================================================

(omg-defclass button-icon (icon-view) 
   ((action :initform nil :initarg :action :accessor action))
   (:documentation "The button-icon is a special icon-view which perform an action
when you click in it.#enddoc#
#seealso# (icon-view lock-button unaire-fun-icon) #seealso#
#action# This slot contains a function with one parameter (the icon-button), 
this function is executed each time that the user click into the button-icon.#action#"))

(defmethod om-view-click-handler ((self button-icon) where)
   "this function call the slot action of SELF with the parameter SELF"
   (declare (ignore where))
   (setf (selected-p self) t)
   (om-draw-contents self)
   (om-init-motion-functions self nil 'release-button-action))

(defmethod release-button-action ((Self button-icon) Where)
  (when (action self)
    (apply (action self) (list self)))
  (setf (selected-p self) nil)
  (om-invalidate-view self t)
  )

;==========================================================
(omg-defclass button-dbicon (button-icon) 
  ((icon2 :accessor icon2 :initform nil :initarg :icon2)
   (my-size :accessor my-size :initform nil :initarg :my-size)))

;add by jean Bresson 31/08/05 (bravo)
(defmethod initialize-instance :after ((self button-dbicon) &rest initargs)
  (if (my-size self)
    (om-set-view-size self (my-size self))
    (let ((pict (cadr (get&corrige-icon (iconID self))))
        (pict2 (cadr (get&corrige-icon (icon2 self)))))
    (om-set-view-size self (om-get-picture-size pict)))))


(defmethod om-view-click-handler ((self button-dbicon) where)
   "this function call the slot action of SELF with the parameter SELF"
   (declare (ignore where))
  (setf (selected-p self) t)
  (om-redraw-view self)
  (om-init-motion-functions self nil 'release-button-action))



  
(defmethod plot-icon-view ((self button-dbicon) &optional (selected nil))
  "Plot the icon referenced by (iconID self) in the view self,
selected parameter is used to draw the icon in normal or selected mode"
  (without-interrupts
    (let* ((icon (if (and selected (icon2 self)) (icon2 self) (iconID self)))
           (iconparams (get&corrige-icon icon))
           (iconhdlr (second iconparams)))
      (if selected 
          (setf (icon2 self) (first iconparams))
        (setf (iconID self) (first iconparams)))
      (om-with-focused-view self
                            (om-draw-picture self iconhdlr (om-make-point 0 0) (om-view-size self))
                            )
      )
    )
  )
;==========================================================


(omg-defclass unaire-fun-icon (icon-view OMSimpleframe om-view-drop) 
   ((action :initform nil :initarg :action :accessor action))
   (:documentation "The unaire-fun-icon is a special icon-view which perform an action
when you drag a OMFrame in it.
An unaire-fun-icon can also execute an action when you click in it, you must specifie 
the method do-default-action for the object of the unaire-fun-icon.#enddoc#
#seealso# (icon-view button-icon unaire-fun-view) #seealso#
#action# This slot contains a function with two parameters (the object of the OMFrame dragged)
and (unaire-fun-icon),this function is executed each time that the user drag an OMFrame 
into the  unaire-fun-icon.#action#"))

(defmethod om-view-click-handler ((self unaire-fun-icon) where)
   "Execute the method do-default-action specified for the object of SELF"
   (declare (ignore where))
   (do-default-action (object self) self))


;==========================================================

(omg-defclass unaire-fun-view (OMSimpleframe om-view-drop) 
   ((action :initform nil :initarg :action :accessor action))
   (:documentation "The unaire-fun-view is a special view which perform an action
when you drag a OMFrame in it.#enddoc#
#seealso# (unaire-fun-icon) #seealso#
#action# This slot contains a function with one parameter (the OMFrame dragged),
the object of the unaire-fun-view is itself,
this function is executed each time that the user drag an OMFrame 
into the unaire-fun-view.#action#"))

(defmethod om-draw-contents ((self unaire-fun-view))
   (call-next-method)
   (om-draw-view-outline self))

(defmethod initialize-instance :after ((self unaire-fun-view) &key controls)
   "Put the object of self to self"
   (declare (ignore controls))
   (setf (object self) self))



;==========================================================
;This code is used to get an icon from the icon resources
; in general it is an icon that has a special draw when it is select (un peu comme button-icon quand meme...)
;not in doc.

(omg-defclass choise-icon-button (button-icon) ())

(defmethod om-view-click-handler ((self choise-icon-button) where)
   (declare (ignore where))
   (when (action self)
     (apply (action self) (list self))))

(defmethod om-draw-contents ((self choise-icon-button))
   (when (selected-p self)
     (om-with-focused-view self
       (om-with-hilite-foreground
         (om-fill-rect 0 0 (w self) (h self)))))
   (call-next-method)) 


;;; a diolag for the user to choose an icon in kernel or user resources
;;; returns a icon number if it's a kernel icon
;;; returns a list (number package) if it's a user resource
;;; returns nil if no icon
(defun choise-icon (&optional  (kernel t) (user t))
  (let ((res (choose-resource-dialog :icon :kernel kernel :user user))
        (icn nil))
    (when (consp res)
        (cond 
         ((equal (third res) 'kernel)
          (setf icn (first (get&corrige-icon 
                                     (read-from-string (string (first res)))))))
          ((equal (third res) 'user) 
           (setf icn (first (get&corrige-icon 
                                     (list (read-from-string (string (first res))) *package-user*)))))
          (t nil))
      )
    icn))


;;; resizes an object to fit its container size without loosing proportions
(defun resize-to-fit (object-size container-size)
   (let (new-size new-pos fact fact2)
     (cond ((and (< (om-point-h object-size) (om-point-h container-size)) (< (om-point-v object-size) (om-point-v container-size)))
                  (setf new-size object-size)
                  (setf new-pos (om-make-point (- (round (om-point-h container-size) 2) (round (om-point-h object-size) 2))
                                                                           (- (round (om-point-v container-size) 2) (round (om-point-v object-size) 2)))))
                 ((and (>= (om-point-h object-size) (om-point-h container-size)) (< (om-point-v object-size) (om-point-v container-size)))
                  (setf fact (/ (om-point-h container-size) (om-point-h object-size)))
                  (setf new-size (om-make-point (round (* fact (om-point-h object-size))) (round (* fact (om-point-v object-size)))))
                  (setf new-pos (om-make-point 0 (- (round (om-point-v container-size) 2) (round (om-point-v new-size) 2)))))
                 ((and (< (om-point-h object-size) (om-point-h container-size)) (>= (om-point-v object-size) (om-point-v container-size)))
                  (setf fact (/ (om-point-v container-size) (om-point-v object-size)))
                  (setf new-size (om-make-point (round (* fact (om-point-h object-size))) (round (* fact (om-point-v object-size)))))
                  (setf new-pos (om-make-point (- (round (om-point-h container-size) 2) (round (om-point-h new-size) 2)) 0)))
                 ((and (>= (om-point-h object-size) (om-point-h container-size)) (>= (om-point-v object-size) (om-point-v container-size)))
                  (setf fact (/ (om-point-h container-size) (om-point-h object-size)))
                  (setf fact2 (/ (om-point-v container-size) (om-point-v object-size)))
                  (if (< fact fact2)
                      (progn 
                         (setf new-size (om-make-point (round (* fact (om-point-h object-size))) (round (* fact (om-point-v object-size)))))
                         (setf new-pos (om-make-point 0 (- (round (om-point-v container-size) 2) (round (om-point-v new-size) 2)))))
                    (progn 
                         (setf new-size (om-make-point (round (* fact2 (om-point-h object-size))) (round (* fact2 (om-point-v object-size)))))
                       (setf new-pos (om-make-point (- (round (om-point-h container-size) 2) (round (om-point-h new-size) 2)) 0)))
                    ))
                 )
     (list new-size new-pos)))

                          