;OpenMusic
;
;Copyright (C) 1997, 1998, 1999, 2000 by IRCAM-Centre Georges Pompidou, Paris, France.
; 
;This program is free software; you can redistribute it and/or
;modify it under the terms of the GNU General Public License
;as published by the Free Software Foundation; either version 2
;of the License, or (at your option) any later version.
;
;See file LICENSE for further informations on licensing terms.
;
;This program is distributed in the hope that it will be useful,
;but WITHOUT ANY WARRANTY; without even the implied warranty of
;MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;GNU General Public License for more details.
;
;You should have received a copy of the GNU General Public License
;along with this program; if not, write to the Free Software
;Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
;
;Authors: Gerard Assayag and Augusto Agon


;DocFile
;This File defines functions for picture manegement.
;All init pictures for OM are loaded in this file.
;Last Modifications :
;18/10/97 first date.
;DocFile

(in-package :om)

;Used by temporal objects in maquettes
(defclass om-picture (internepicture)
   ((selected-p :initform nil :accessor selected-p)))

(defun make-new-om-pict () (make-instance 'om-picture))


;;;=========================================

(defclass patch-picture (om-picture)
   ((pict-pos :initform (om-make-point 0 0) :accessor pict-pos)
    (pict-size :initform (om-make-point 10 10) :accessor pict-size)))

(defmethod draw-pict-patch ((self patch-picture) view)
   (let ((pictsize (pict-size self))
           (x0 (om-point-h (pict-pos self)))
           (y0 (om-point-v (pict-pos self))))
     (om-draw-picture view (thepict self) (om-make-point x0 y0) pictsize)
     (when (selected-p self)
       (om-with-line-size 2
         (om-draw-rect (+ 1 x0) (+ 1 y0) (- (om-point-h pictsize) 2) (- (om-point-v pictsize) 2))))))

(defmethod omNG-save ((self patch-picture) &optional (values? nil))
  `(let ((newpict (make-instance ',(type-of self)
                    :name ',(name self))))
     (setf (thepict newpict) (om-get-picture ',(name self)))
     (setf (name newpict) ',(name self))
     (setf (pict-pos newpict) ,(om-save-point (pict-pos self)))
     (setf (pict-size newpict) ,(om-save-point (pict-size self)))
     newpict))

(defmethod bgk2pict ((self patch-picture) patch)
   (let ((picture (make-instance 'picture))
         mynew-pict)
     (setf newcall (omNG-make-new-boxcall (find-class 'picture) (pict-pos self) (mk-unique-name patch "pict")))
     (setf (thepict picture) (thepict self))
     (setf (name picture) (name self))
     (setf (value newcall) picture)
     (setf (frame-size newcall) (om-add-points (pict-size self)  (om-make-point 0 16)))
     (setf (showpict newcall) t)
     (omG-add-element patch (make-frame-from-callobj newcall))
     (setf (pictu-list (object patch)) (remove self (pictu-list (object patch)) :test 'equal))
     (om-invalidate-view patch t)))


(omg-defclass picture-view (om-view)
   ((pict :initform nil :initarg :pict :accessor pict)))

(defmethod om-draw-contents ((self picture-view))
   (call-next-method)
   (when (pict self)
     (om-draw-picture self (pict self) (om-make-point 0 0) (om-make-point (w self) (h self)))))


(defmethod popup-pict-menu-context ((object patch-picture) patch)
   (let* ((itemlist (list (om-new-leafmenu "make-box" #'(lambda () (bgk2pict object patch)))))
            (menu (om-create-menu 'pair-pop-up-menu itemlist)))
     (om-open-pop-up-menu menu patch)))


;---------------------------------------------
;OM PICTURE MANAGEMENT
;---------------------------------------------
; name = symbol avec prefixe 'kernel ou 'user
(defun om-get-picture (name)
  (if name
   (let* ((namevals (multiple-value-list (string-until-char (string name) "-"))))
     (or (om-get-picture-loaded name)
           (om-load-and-store-picture (intern (second namevals)) (intern (first namevals)))
           )) *om-def-pict*))  
             
;;; name = symbol (sans prefixe)   -    folder = 'user ou 'kernel

(defun om-load-and-store-picture (name folder &optional full-folder-path)
  (or (om-get-picture-loaded (intern (string+ (string folder) "-" (string name))))
      (let* ((folderpath
              (if (equal folder 'full)
                  full-folder-path
                (if (equal folder 'user)
                    (make-pathname :directory (append (pathname-directory (mypathname *current-workspace*))
                                                      (list "resources" "pict")))
                  *om-pict-folder*)))
             (pict (om-load-pixmap (string name) *om-pict-type* (make-pathname :directory (pathname-directory folderpath)) nil)))
        (when pict
          (om-store-picture pict (intern (string+ (string folder) "-" (string name)))))
        pict)
      *om-def-pict*))

;---------------------------
;OM PICTURES
;---------------------------
(defvar *glass-pict* nil "The picture used in the instance boxes")
(defvar *impulsion-pict* nil "Used in temporal objects with dur = 0")
(defvar *ctrl-impulsion-pict* nil "Used in temporal objects with dur = 0")
(defvar *boxedit-pict* "The picture used in the editor boxes")
(defvar *graph-pres* "Presentation pict")
(defvar *play-controls-pict* nil "Playing controls pict resource.")
(defvar *maquette-pict* nil "Maquette palette pict.")
(defvar *om-def-pict* nil "Default pict when picture not found")


(defun init-all-pict ()
   (setf *play-controls-pict* (om-load-and-store-picture 'play-controls 'kernel))
   (setf *graph-pres* (om-load-and-store-picture 'initgraph 'kernel))
   (setf *glass-pict* (om-load-and-store-picture 'glass-pict 'kernel))
   (setf *impulsion-pict* (om-load-and-store-picture 'impulsion 'kernel))
   (setf *ctrl-impulsion-pict* (om-load-and-store-picture 'ctrlimpulsion 'kernel))
   (setf *maquette-pict* (om-load-and-store-picture 'maq-pal 'kernel))
   (setf *boxedit-pict* (om-load-and-store-picture 'editorpict 'kernel))
   (setf *om-def-pict* (om-load-and-store-picture 'def-pict 'kernel)))

(om-add-init-func 'init-all-pict)

