;OpenMusic
;
;Copyright (C) 1997, 1998, 1999, 2000 by IRCAM-Centre Georges Pompidou, Paris, France.
; 
;This program is free software; you can redistribute it and/or
;modify it under the terms of the GNU General Public License
;as published by the Free Software Foundation; either version 2
;of the License, or (at your option) any later version.
;
;See file LICENSE for further informations on licensing terms.
;
;This program is distributed in the hope that it will be useful,
;but WITHOUT ANY WARRANTY; without even the implied warranty of
;MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;GNU General Public License for more details.
;
;You should have received a copy of the GNU General Public License
;along with this program; if not, write to the Free Software
;Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
;
;Authors: Gerard Assayag and Augusto Agon

;DocFile
;This File contains high level functions and macros for Menubars management.
;Last Modifications :
;18/10/97 first date.
;DocFile

(in-package :om)


;-------------------high level menu functions (compatibility)--------------------------

;enable all menu items in the menu <menu>
(defun enable-all-menu-items (menu)
  (loop for item in  (om-menu-items menu) do
        (om-menuitem-enable item)))

;disable all menu items in the menu <menu>
(defun disable-all-menu-items (menu)
  (loop for item in  (om-menu-items menu) do
        (om-menuitem-disable item)))


;disable the menu items which name in the list <list> in the menu <menu>
(defun disable-this-menu-items (menu list)
  (loop for item in list do
        (let ((rep (om-find-menu-item menu item)))
          (when rep
            (om-menuitem-disable rep)))))

;disable the menu items which name in the list <list> in the menu <menu>
(defun enable-this-menu-items (menu list)
  (loop for item in list do
        (let ((rep (om-find-menu-item menu item)))
          (when rep
            (om-menuitem-enable rep)))))


;;;==========================
;;; MENUS CALLBACKS
;;;==========================
(defun show-about-win ()
  (om-select-window *presentation*))

(defun show-workspace-win ()
  (showobjecteditor *current-workspace*))

(defun show-packages-win () nil)

(defun show-globals-win () nil)

(defun show-resources-win () nil)

(defun show-preferences-win () nil)

(defun show-listener-win () 
  (om-make-new-listener))

;;;==========================
;;; MENUS
;;;==========================

(defmethod make-new-menu ((self t))
  (om-new-leafmenu "New" nil nil nil))

; disable is a list of disabled items
(defun make-om-menu (menuid &key disable editor)
  (cond ((equal menuid 'windows)
         (om-make-menu "Windows" (list 
                                  (list 
                                   (om-new-leafmenu "Workspace" 'show-workspace-win "W" (if (member "Workspace" disable :test 'string-equal) nil t))
                                   (om-new-leafmenu "Packages" 'show-packages-win nil (if (member "Packages" disable :test 'string-equal) nil t))
                                   (om-new-leafmenu "Global Variables" 'show-globals-win nil (if (member "Global Variables" disable :test 'string-equal) nil t))
                                   (om-new-leafmenu "Resources" 'show-resources-win nil (if (member "Resources" disable :test 'string-equal) nil t))
                                   )
                                  (list 
                                   (om-new-leafmenu "Lisp Listener" 'show-listener-win nil (if (member "Lisp Listener" disable :test 'string-equal) nil t))
                                   (om-new-leafmenu "Lisp Editor" 'om-open-new-text-editor nil (if (member "Lisp Editor" disable :test 'string-equal) nil t))
                                   )
                                  (mapcar #'(lambda (w) (om-new-leafmenu (om-window-title w) 
                                                                         #'(lambda () (om-select-window w))))
                                          (om-get-all-windows 'editorwindow))
                                  ())))
        ((equal menuid 'presentation)
         (om-make-menu "Presentation" (list
                                       (om-new-leafmenu "Name" #'(lambda ()(omG-change-presentation (om-front-window) 1)) nil 
                                                        (if (member "Name" disable :test 'string-equal) nil t))
                                       (om-new-leafmenu "Type" #'(lambda () (omG-change-presentation (om-front-window) 2)) nil 
                                                        (if (member "Type" disable :test 'string-equal) nil t))
                                       (om-new-leafmenu "Icons" #'(lambda () (omG-change-presentation (om-front-window) 0)) nil 
                                                        (if (member "Icons" disable :test 'string-equal) nil t))
                                       )))
        ((equal menuid 'file)
         (om-make-menu "File" (list
                               (make-new-menu editor)
                               (list 
                                (om-new-leafmenu "Save" #'(lambda () (window-save (om-front-window))) "s" 
                                                 (if (member "Save" disable :test 'string-equal) nil t))
                                (om-new-leafmenu "Last saved" #'(lambda () (window-last-saved (editor (om-front-window)))) nil 
                                                 (if (member "Last saved" disable :test 'string-equal) nil t))
                                )
                               (list 
                                (om-new-leafmenu "Close" #'(lambda () (om-close-window  (om-front-window))) "w" 
                                                 (if (member "Close" disable :test 'string-equal) nil t))
                                )
                               (list 
                                (om-new-leafmenu "Make Alias" #'(lambda () (alias-window  (om-front-window))) "m" 
                                                 (if (member "Make Alias" disable :test 'string-equal) nil t))
                                (om-new-leafmenu "Get Info" #'(lambda() (get-info-window  (om-front-window))) "i" 
                                                 (if (member "Get Info" disable :test 'string-equal) nil t))
                                )
                               (list 
                                (om-new-leafmenu "Page Setup" #'(lambda () (om-page-setup)) nil 
                                                 (if (member "Page Setup" disable :test 'string-equal) nil t))
                                (om-new-leafmenu "Print" #'(lambda () (om-print-window  (om-front-window))) "p" 
                                                 (if (member "Print" disable :test 'string-equal) nil t))
                                )
                               )))
        ((equal menuid 'edit)
         (om-make-menu "Edit" (list
                               (om-new-leafmenu "Undo" #'(lambda() (undo  (om-front-window))) "z" 
                                                (if (member "Undo" disable :test 'string-equal) nil t))
                               (list 
                                (om-new-leafmenu "Cut" #'(lambda () (cut  (om-front-window))) "x" 
                                                 (if (member "Cut" disable :test 'string-equal) nil t))
                                (om-new-leafmenu "Copy" #'(lambda () (copy  (om-front-window))) "c" 
                                                 (if (member "Copy" disable :test 'string-equal) nil t))
                                (om-new-leafmenu "Paste" #'(lambda () (paste  (om-front-window))) "v" 
                                                 (if (member "Paste" disable :test 'string-equal) nil t))
                                (om-new-leafmenu "Duplicate" #'(lambda () (duplicate-window  (om-front-window))) "d" 
                                                 (if (member "Duplicate" disable :test 'string-equal) nil t))
                                (om-new-leafmenu "Clear" #'(lambda () (clear  (om-front-window))) nil 
                                                 (if (member "Clear" disable :test 'string-equal) nil t))
                                )
                               (om-new-leafmenu "Select All" #'(lambda () (select-all  (om-front-window))) "a" 
                                                (if (member "Select All" disable :test 'string-equal) nil t))
                               )))
        ((equal menuid 'functions)
         (package-fun2menu *om-package-tree*))
        ((equal menuid 'classes)
         (package-classes2menu *om-package-tree*))
        ))


(defun package-fun2menu (package &optional name pos)
  (let ((subpack (copy-list (subpackages package))))
    (om-make-menu (or name "Functions")
                  (list  
                   (mapcar #'(lambda (f) (om-new-leafmenu (name f)
                                                          #'(lambda () (add-box-from-menu f pos))))
                           (functions package))
                   (remove nil 
                           (loop for item in subpack collect
                                 (unless (and (omlib-p item) (not (loaded? item)))
                                   (package-fun2menu item (name item) pos)))
                           )))))

(defun package-classes2menu (package &optional name pos)
  (let ((subpack (copy-list (subpackages package))))
    (om-make-menu (or name "Classes")
                  (list  
                   (mapcar #'(lambda (c) (om-new-leafmenu (name c)
                                                          #'(lambda () (add-box-from-menu c pos))))
                           (classes package))
                   (remove nil 
                           (loop for item in subpack collect
                                 (when (and (not (and (omlib-p item) (not (loaded? item))))
                                            (or (subpackages item) (classes item) (equal item *package-user*)))
                                   (package-classes2menu item (name item) pos)))
                           )))))


(defmethod get-menubar ((self t)) nil)
(defmethod update-menubar ((self t)) nil)

(defmethod get-menubar ((self EditorView)) (list (make-om-menu 'windows)))

(defmethod update-menubar ((self EditorView))
  (om-add-menu-to-win (om-view-window self)))

(defmethod get-menubar ((self NonRelationEditor)) (list (make-om-menu 'file :disable '("Last saved" "Page Setup" "Print") :editor self) 
                                                        (make-om-menu 'edit :disable '("Undo"))  
                                                        (make-om-menu 'presentation) 
                                                        (make-om-menu 'windows :disable '("Packages" "Global Variables"))))

(defmethod get-menubar ((self WorkspaceEditor)) (list (make-om-menu 'file :disable '("Last saved" "Make Alias" "Page Setup" "Print") :editor self) 
                                                      (make-om-menu 'edit :disable '("Undo"))  
                                                      (make-om-menu 'presentation) 
                                                      (make-om-menu 'windows :disable '("Workspace" "Packages" "Global Variables" "Resources"))))

(defmethod make-new-menu ((self WorkspaceEditor))
  (list 
   (om-new-leafmenu "New Patch" #'(lambda() (omG-make-new-icon-window (om-front-window) 'p)) "1")
   (om-new-leafmenu "New Maquete" #'(lambda() (omG-make-new-icon-window (om-front-window) 'm)) "2")
   ;;;(om-new-leafmenu "New Sheet" #'(lambda() (omG-make-new-icon-window (om-front-window) 'sh)) "5")
   (om-new-leafmenu "New Folder" #'(lambda() (omG-make-new-icon-window (om-front-window) 'f)) "n")))
  
(defmethod get-menubar ((self FolderEditor)) (list (make-om-menu 'file :disable '("Last saved" "Make Alias" "Page Setup" "Print") :editor self) 
                                                   (make-om-menu 'edit :disable '("Undo"))  
                                                   (make-om-menu 'presentation) 
                                                   (make-om-menu 'windows :disable '("Packages" "Global Variables" "Resources"))))

(defmethod get-menubar ((self PatchEditor)) (list (make-om-menu 'file :disable (list "Page Setup" "Print") :editor self)
                                                        (make-om-menu 'edit)  
                                                        (make-om-menu 'functions) (make-om-menu 'classes)
                                                        (make-om-menu 'windows :disable '("Packages" "Global Variables" "Resources"))))


;(defmethod get-menubar ((self metaobj-Editor)) *window-finder-menubar*)
;(defmethod get-menubar ((self patchEditor)) *patch-menubar*)
;(defmethod get-menubar ((self packageEditor)) *package-menubar*)
;(defmethod get-menubar ((self MaquetteEditor))  *simplest-menubar*)
;(defmethod get-menubar ((self inputEditor))  *simplest-menubar*)
;(defmethod get-menubar ((self classtreeEditor)) *package-menubar*)
;(defmethod get-menubar ((self classEditor)) *package-menubar*)
;(defmethod get-menubar ((self GenericFunEditor)) *package-menubar*)
;(defmethod get-menubar ((self methodEditor)) *patch-menubar*)
;(defmethod get-menubar ((self instanceEditor)) *simplest-menubar*)


;(defmethod extra-menu-info ((self t)) nil)

;(defmethod extra-menu-info ((self classEditor))
;  (disable-this-menu-items *window-menu-file*  '("save" "make alias" "duplicate")))

;(defmethod extra-menu-info ((self EditorView))
;  (disable-this-menu-items *window-menu-file*  '("Preferences" "Save Etat")))

;(defmethod extra-menu-info ((self classtreeEditor))
;  (disable-this-menu-items *window-menu-file*  '("Save" "duplicate"))
;  (disable-this-menu-items *window-menu-edit*  '("clear" "cut" "copy" "paste")))

;(defmethod extra-menu-info ((self relationEditor))
;  (enable-this-menu-items *window-menu-file*  '("Last saved"))
;  (enable-this-menu-items *window-menu-edit*  '("undo")))

;(defmethod extra-menu-info ((self MaquetteEditor))
;  (call-next-method)
;  (disable-this-menu-items *window-menu-file*  '("New")))

;(defmethod extra-menu-info ((self GenericFunEditor))
;  (disable-this-menu-items *window-menu-file*  '("save" "duplicate" "make alias"))
;  (disable-this-menu-items *window-menu-edit*  '( "cut" "copy" "paste")))

;(defmethod extra-menu-info ((self instanceEditor))
;  (disable-this-menu-items *window-menu-file*  '("New" "Save" "Duplicate" "Make alias"))
;  (disable-all-menu-items *window-menu-edit*))

;(defmethod extra-menu-info ((self methodEditor))
;  (disable-this-menu-items *window-menu-file*  '("new" "save")))

;(defmethod extra-menu-info ((self globalsFolderEditor))
;  (disable-this-menu-items *window-menu-file*  '("save" "new" "make alias")))

;(defmethod extra-menu-info ((self packageEditor))
;  (disable-this-menu-items *window-menu-file*  '("save" "make alias" "duplicate"))
;  (disable-this-menu-items *window-menu-edit*  '( "cut" "copy" "paste")))

;(defmethod extra-menu-info ((self Editorview))
;  (disable-this-menu-items *window-menu-file*  '("make alias" "save" "new")))


;  (setf *general-new-menu-item* (om-new-menu "New"
;                                             (om-new-leafmenu "New Patch" #'(lambda() (omG-make-new-icon-window (om-front-window) 'p)) #\1)
;                                             (om-new-leafmenu "New Maquete" #'(lambda() (omG-make-new-icon-window (om-front-window) 'm)) #\2)
;                                             (om-new-leafmenu "-" nil)
;                                             ;;;(om-new-leafmenu "New Score" #'(lambda() (omG-make-new-icon-window (om-front-window) 'sh)) #\5)
;                                             (om-new-leafmenu "New Folder" #'(lambda() (omG-make-new-icon-window (om-front-window) 'f)) #\n)))
;  
;  (setf *package-new-menu-item* (om-new-menu "New"
;                                             (om-new-leafmenu "Package" #'(lambda() (omG-make-new-icon-window  (om-front-window) 'p)))
;                                             (om-new-leafmenu "Remote User Library ..." #'(lambda() 
;                                                                                            (let ((path (om-choose-directory-dialog :directory (om-user-home))))
 ;                                                                                             (add-one-lib path))))
 ;                                            ))
 ;       
 ;       ;(om-new-leafmenu "New Package" #'(lambda() (omG-make-new-icon-window  (om-front-window) 'p)))
 ; 
 ; (setf *classtree-new-menu-item* (om-new-leafmenu "New Class" #'(lambda() (omG-make-new-icon-window  (om-front-window) 'c))))
 ; 
;  (setf *genfun-new-menu-item* (om-new-leafmenu "New Method" #'(lambda() (omG-make-new-icon-window  (om-front-window) 'm))))
;  
;  (setf *class-new-menu-item* (om-new-leafmenu "New Slot" #'(lambda() (omG-make-new-icon-window  (om-front-window) ))))
;  
 ; (setf *patch-new-menu-item* (om-new-leafmenu "New Generic Function" #'(lambda() (omG-make-new-icon-window  (om-front-window) 'f))))
 ; 
 ; (setf *null-new-menu-item* (om-new-leafmenu "New" #'(lambda () nil)))
 ; 
 ; (setf *globals-new-menu-item* (om-new-leafmenu "New var" #'(lambda () (omG-new-var (panel  (om-front-window))))))
  

;;; FILE et EDIT sont par defaut dans le LISTENER
(defmethod om-window-class-menubar ((self om-listener)) 
  (list (make-om-menu 'windows :disable '("Lisp Listener"))))
                             

;=============================
; SPECIAL CLASSES/FUNCTIONS MENU
;=============================

(defvar *adding-a-box* nil)

;(defun add-box-from-menu (item)
;  (let* ((window  (om-front-window))
;         (thepatch (panel window))
;         (newbox (omNG-make-new-boxcall item  (om-make-point 22 22) (mk-unique-name thepatch (name item))))
;         pospanel pos)
;    (om-wait-for-click window *om-box-cursor*)
;    (setf pospanel (om-mouse-position thepatch))
;    (setf pos (om-mouse-position window))
;   (when (om-point-in-rect-p pos (om-pts-to-rect (om-make-point 0 0) (om-view-size window)))
;      (setf (frame-position newbox) (borne-position pospanel))
;     (omG-add-element thepatch (make-frame-from-callobj newbox)))))

(defun add-box-from-menu (item &optional pos)
  (let* ((window  (om-front-window))
         (thepatch (panel window)))
    (setf *adding-a-box* (list item window))
    ;(set-cursor *om-box-cursor*)
    ; API !
    (when pos
      (om-view-click-handler thepatch (setf *new-obj-initial-pos* pos))
      )))


;======================================
;ACTIONS FOR CONTEX MENUS
;======================================

;; PB :
;; (defmethod om-view-click-handler ((self pckbrowser-icon) where)
;; (defmethod om-score-click-handler ((self scorePanel) where double-click-p)

;===========================input-menu
(defmethod om-get-menu-context ((object input-funboxframe))
  (items (thepopup (object object))))

;===========================loop
(defmethod om-get-menu-context ((object loopboxframe))
  (list+ (list (om-new-leafmenu "Update doc" #'(lambda () (apply-win (om-view-window object) 'update-doc)))
               (om-new-leafmenu  "-"#'(lambda () nil))) 
         (boxframe-default-list object)))

;==================== pict
(defmethod om-get-menu-context ((object PictboxFrame))
  (list (om-new-leafmenu  "Pict2bgk" #'(lambda () (pict2bkg object)))))

;===============WS and FOLDERS
(defmethod om-get-menu-context ((self metaobj-panel))
  (setf *new-obj-initial-pos* (om-mouse-position self))
  (list 
   (list 
    (om-new-menu "New..."
                 (om-new-leafmenu "New Patch" #'(lambda() (omG-make-new-icon-window (om-front-window) 'p)))                                             
                 (om-new-leafmenu "New Maquete" #'(lambda() (omG-make-new-icon-window (om-front-window) 'm)))                                  
                 (om-new-leafmenu "New Folder" #'(lambda() (omG-make-new-icon-window (om-front-window) 'f)))))
   (list 
    (om-new-leafmenu "Import File" #'(lambda () (import-file self)))
    (om-new-leafmenu "Import Folder" #'(lambda () (import-folder self)))
    ) ;(om-new-leafmenu "-" #'(lambda () nil))
   (om-new-menu "Presentation"
                (om-new-leafmenu "Icons" #'(lambda() (omG-change-presentation (om-view-window self) 0)))
                (om-new-leafmenu "List" #'(lambda() (omG-change-presentation (om-view-window self) 1))))))

(defmethod om-get-menu-context ((self globalsFolderPanel))
  (list (om-new-menu "View"
                     (om-new-leafmenu "Icons" #'(lambda() (omG-change-presentation (om-view-window self) 0)))
                     (om-new-leafmenu "List" #'(lambda() (omG-change-presentation (om-view-window self) 1))))))

(defmethod om-get-menu-context ((self PatchPanel))
  (let ((posi (om-mouse-position self)))
  (list 
   (om-new-leafmenu "Comment" #'(lambda () 
                                  (let ((newbox (omNG-make-new-boxcall 'comment posi  "comment" )))
                                    (when newbox
                                      (omG-add-element self (make-frame-from-callobj newbox))))))
   (list 
    (package-fun2menu *om-package-tree* nil posi)
    (package-classes2menu *om-package-tree* nil posi)
    )
   (list
    (om-new-leafmenu "Input" #'(lambda () (add-input self posi)))
    (om-new-leafmenu "Output" #'(lambda () (add-output self posi)))
    )
   (om-new-leafmenu "Last Saved"  #'(lambda () (window-last-saved (editor self))))
   )))


(defmethod om-get-menu-context ((self icon-finder-icon))
  (om-get-menu-context (om-view-container self)))

(defmethod om-get-menu-context ((self icon-finder))
  (list (list 
         (om-new-leafmenu "Open" #'(lambda () (apply-win (om-view-window self) 'open-icon-win (om-make-point 0 0))))
         ) ; (om-new-leafmenu "-" #'(lambda () nil))
        (om-new-leafmenu "Get Info" #'(lambda () (get-info-window (om-view-window self))))
        (om-new-leafmenu "Duplicate" #'(lambda () (duplicate-window (om-view-window self))))
        (om-new-leafmenu "Export" #'(lambda () (export-file (object self))))))

(defmethod om-get-menu-context ((self package-icon-frame))
   (list (om-new-leafmenu "Open Package" #'(lambda () (open-icon-win self (om-make-point 0 14))))
          (om-new-leafmenu "Open Class Tree" #'(lambda () (open-icon-win self (om-make-point 0 0))))
          (om-new-leafmenu "Show Package Functions" #'(lambda () (open-icon-scroller self)))
          (om-new-leafmenu "-" #'(lambda () nil))
          (om-new-leafmenu "Get info" #'(lambda () (get-info-window (om-view-window self))))
          ))

(defmethod om-get-menu-context ((self genfun-icon-frame))
   (list (om-new-leafmenu "Open" #'(lambda () (open-icon-win self (om-make-point 0 0))))
          (om-new-leafmenu "Show Function Methods" #'(lambda () (open-icon-scroller self)))
          (om-new-leafmenu "-" #'(lambda () nil))
          (om-new-leafmenu "Get info" #'(lambda () (get-info-window (om-view-window self))))
          ))

(defmethod om-get-menu-context ((self class-icon-frame))   
   (list (om-new-leafmenu "Open" #'(lambda () (open-icon-win self (om-make-point 0 0))))
          (om-new-leafmenu "Show Internal Method" #'(lambda () (open-icon-scroller self)))
          (om-new-leafmenu "-" #'(lambda () nil))
          (om-new-leafmenu "Get info" #'(lambda () (get-info-window (om-view-window self))))
          )) 

;==========================
;DONT HAVE MENU
;==========================
(defmethod simplest-menu-context ((self t))  
  (list (om-new-leafmenu "Open" #'(lambda () (apply-win (om-view-window self) 'open-icon-win (om-make-point 0 0))))
        (om-new-leafmenu "-"  #'(lambda () nil))
        (om-new-leafmenu "Get info" #'(lambda () (get-info-window (om-view-window self))))))

(defmethod om-get-menu-context ((self globals-folder-icon)) (simplest-menu-context self))
(defmethod om-get-menu-context ((self lispfun-icon-frame))   (simplest-menu-context self))
(defmethod om-get-menu-context ((self type-icon-frame))   (simplest-menu-context self))
(defmethod om-get-menu-context ((self instance-icon-frame))   (simplest-menu-context self))
(defmethod om-get-menu-context ((self slot-icon-frame))   (simplest-menu-context self))


(defmethod simplest-box-menu-context ((self t))  
  (list (om-new-leafmenu "Open" #'(lambda () (OpenObjectEditor (object  self))))
        (om-new-leafmenu "-"  #'(lambda () nil))
        (om-new-leafmenu "Get info" #'(lambda () (get-info-window (om-view-window self))))))

(defmethod om-get-menu-context ((self aliasboxframe))   (simplest-box-menu-context self))
(defmethod om-get-menu-context ((self classboxframe))   (simplest-box-menu-context self))
(defmethod om-get-menu-context ((self inframe))   (simplest-box-menu-context self))
(defmethod om-get-menu-context ((self TypedInFrame))   (simplest-box-menu-context self))


;==========================
;DONT HAVE MENU
;==========================
(defmethod om-get-menu-context ((self classPanel)) nil)
(defmethod om-get-menu-context ((self GenericFunPanel)) nil)
(defmethod om-get-menu-context ((self InstancePanel)) nil)
(defmethod om-get-menu-context ((self  PackagePanel)) nil)
(defmethod om-get-menu-context ((self  classTreePanel)) nil)
(defmethod om-get-menu-context ((self  outframe)) nil)
(defmethod om-get-menu-context ((self  tempOutFrame)) nil) 



;EDITOR BOXES


(defun boxframe-default-list (object)
  (list (list 
         (om-new-leafmenu  "Copy" #'(lambda () (copy (om-view-window object))))
         (om-new-leafmenu  "Cut" #'(lambda () (cut (om-view-window object))))
         (om-new-leafmenu  "Duplicate" #'(lambda () (duplicate-window (om-view-window object))))
         )
        (list 
         (om-new-leafmenu  "Get info" #'(lambda () (get-info-window (om-view-window object))))
         )
        (list
         (om-new-leafmenu "S/H Button" #'(lambda () (apply-win (om-view-window object) 'add-rem-lock-button)))
         (om-new-leafmenu "Eval" #'(lambda ()
                                    (om-eval-enqueue 
                                     `(progn
                                        (eval-box ,object)
                                        (clear-ev-once ,(om-view-container object)))))))))

(defmethod om-get-menu-context ((object boxframe))
  (boxframe-default-list object))

(defmethod om-get-menu-context ((object patchboxFrame))
  (list+  (list (list (om-new-leafmenu "Update Doc"
                                       #'(lambda () (apply-win (om-view-window object) 'update-doc))))
                ) 
          (boxframe-default-list object)))


(defmethod om-get-menu-context ((self maquetteframe))
  (list+ (call-next-method)
         (list (om-new-leafmenu "-" #'(lambda () nil))
               (if (= 0 (mode (object self)))
                 (om-new-leafmenu "Show maquette Ins/Outs" #'(lambda () (change-mode (object self) 1)))
                 (om-new-leafmenu "Hide maquette Ins/Outs" #'(lambda () (change-mode (object self) 0)))))
         ))


(defmethod om-get-menu-context ((self miniview))
  (om-get-menu-context (om-view-container self)))

(defmethod om-get-menu-context ((self commentview))
  (om-get-menu-context (om-view-container self)))
  
(defmethod om-get-menu-context ((object commentboxframe))
  (list (om-new-leafmenu "Text Color" #'(lambda () (color-comments (panel (om-view-window object)))))
        (om-new-leafmenu "Text Font" #'(lambda () (font-comments (panel (om-view-window object)))))))


(defmethod om-get-menu-context ((object boxeditorframe))
  (list+  ;(list (om-new-leafmenu (if (view-of-patch (object object)) "Free View of Patch" "Set View of Patch")
          ;                 #'(lambda () (change-view-of-patch object))))
          (list (list 
           (om-new-leafmenu "Import" #'(lambda () (import-box-editor (object object))))
           (om-new-leafmenu "Export" #'(lambda () (export-box-editor (object object))))
           ))
          (boxframe-default-list object)))


(defmethod om-get-menu-context ((self tempobjframe))
  (let (items)
    (setf items (list (om-new-leafmenu "Open" #'(lambda () (OpenObjectEditor (object  self))))
                      (om-new-leafmenu "Get info" #'(lambda () (get-info-window (om-view-window self))))
                      (om-new-leafmenu "-" #'(lambda () nil))))
    (if (thepict (pictu (object  self)))
      (setf items (list+ items (list (om-new-leafmenu "Kill Picture"
                                                      #'(lambda () 
                                                          (setf (thepict (pictu (object  self))) nil)
                                                          (setf (name (pictu (object  self))) nil)
                                                          (om-invalidate-view self t))))))
      (setf items (list+ items (list (om-new-leafmenu "Load Picture"
                                                      #'(lambda () 
                                                          (let ((newpict (get-picture-file)))
                                                            (setf (thepict (pictu (object self))) (thepict newpict))
                                                            (setf (name (pictu (object self))) (name newpict))
                                                           (om-invalidate-view self t))))))))
    (setf items (list+ items 
                        (list (om-new-leafmenu "-" #'(lambda () nil))
                              (if (mute (object self))
                                (om-new-leafmenu  "Mute Off" #'(lambda () 
                                                                 (setf (mute (object self)) nil)
                                                                 (om-invalidate-view self t)))
                                (om-new-leafmenu  "Mute" #'(lambda () 
                                                             (setf (mute (object self)) t)
                                                             (om-invalidate-view self t)))
                                )
                              (if (lock (object self))
                                (om-new-leafmenu  "Unlock" #'(lambda () 
                                                               (setf (lock (object self)) nil)
                                                               (om-invalidate-view self t)))
                                (om-new-leafmenu  "Lock" #'(lambda () 
                                                             (setf (lock (object self)) t)
                                                             (om-invalidate-view self t)))
                                ))))
    items))


;;;========================
;;; Test IN /OUT dans la maquette
;;;========================

(defmethod add-input ((self t) pos) nil)
(defmethod add-output ((self t) pos) nil)


(defmethod om-get-menu-context ((self MaquettePanel))
  (let (items
        (pos (om-mouse-position self)))
    (setf items (list (om-new-leafmenu "Eval" #'(lambda () (eval-maquette self)))
                      (om-new-leafmenu "Play" #'(lambda () (play-from-palette self)))
                      (om-new-leafmenu "-" #'(lambda () nil))
                      (om-new-leafmenu (if (show-conect (params (object self))) "hide connections" "show connections")
                                       #'(lambda () 
                                           (setf (show-conect (params (object self))) (not (show-conect (params (object self)))))
                                           (om-invalidate-view (om-view-window  self) t)))
                      ))
    (setf items (list+ items (list 
                              (om-new-leafmenu "-" #'(lambda () nil))
                              (om-new-leafmenu "Add INPUT"
                                                      #'(lambda () (add-input self pos)))
                              (om-new-leafmenu "Add OUTPUT"
                                                      #'(lambda () (add-output self pos)))
                              )))
    (if (thepict (pictu (object self)))
      (setf items (list+ items (list (om-new-leafmenu "kill picture"
                                                      #'(lambda () 
                                                          (setf (thepict (pictu (object self))) nil)
                                                          (setf (name (pictu (object  self))) nil)
                                                          (om-invalidate-view self t))))))
      (setf items (list+ items (list (om-new-leafmenu "load picture"
                                                      #'(lambda () 
                                                          (let ((pic (get-picture-file)))
                                                            (when pic
                                                              (setf (pictu (object self)) pic)
                                                              (om-invalidate-view self t)))))))))
    (unless (om-maquette-abs-p (object (editor self)))
      (setf items (list+ items (list 
                                (om-new-leafmenu "-" #'(lambda () nil))
                                (om-new-leafmenu "Last saved" #'(lambda () (window-last-saved (editor self)))))))
      )
    items))



