;OpenMusic
;
;Copyright (C) 1997, 1998, 1999, 2000 by IRCAM-Centre Georges Pompidou, Paris, France.
; 
;This program is free software; you can redistribute it and/or
;modify it under the terms of the GNU General Public License
;as published by the Free Software Foundation; either version 2
;of the License, or (at your option) any later version.
;
;See file LICENSE for further informations on licensing terms.
;
;This program is distributed in the hope that it will be useful,
;but WITHOUT ANY WARRANTY; without even the implied warranty of
;MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;GNU General Public License for more details.
;
;You should have received a copy of the GNU General Public License
;along with this program; if not, write to the Free Software
;Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
;
;Authors: Gerard Assayag and Augusto Agon


;DocFile
;A class to inspect graphic objects is defined here.
;Last Modifications :
;18/10/97 first date.
;DocFile

(in-package :om)

;==========================================================
;A palette window
;==========================================================


(defclass general-palette ()
   ((palette :initform nil :accessor palette)
    (pos-palette :initform (om-make-point 500 10) :accessor pos-palette)))

(defvar *global-palette* nil)
(defun init-globalpalette ()
   (setf *global-palette* (make-instance 'general-palette)))

(om-add-init-func 'init-globalpalette)

(omg-defclass palette-win (om-windoid)
   ((source-name :initform nil :initarg :source-name :accessor source-name)
    (thepict :initform nil :accessor thepict)
    (delta-inpix :initform (om-make-point 0 0) :initarg :delta-inpix :accessor delta-inpix)
    (editor-assoc :initform nil :initarg :editor-assoc :accessor editor-assoc)
    (hot-spot :initform nil :initarg :hot-spot :accessor hot-spot)))


(defmethod initialize-instance :after ((self palette-win) &rest initargs)
  (init-palette self))

;(defun palette-size (name)
;  (let ((pict (om-load-pixmap name *om-pict-type* *om-pict-folder*)))
;    (if pict (om-get-picture-size pict) (om-make-point 25 25))))

(defmethod init-palette ((self palette-win))
  (let* ((pict (if (source-name self) (om-load-and-store-picture (intern (source-name self)) 'kernel)))
         (pictsize (if pict (om-get-picture-size pict) (om-make-point 0 25))))
    (setf (thepict self) pict)
    (om-set-view-size self (om-add-points pictsize (delta-inpix self))))
  t)


;;;(defmethod update-themenubar ((win palette-win)) t)

(defmethod om-close-window ((self palette-win))
  (setf (pos-palette *global-palette*) (om-view-position self))
  (setf (palette *global-palette*) nil)
   (call-next-method))

(defmethod om-draw-contents :before ((self palette-win))
   (if (thepict self)
     (om-draw-picture  self (thepict self) (delta-inpix self) 
                    (om-get-picture-size (thepict self))))
  (when (and (editor-assoc self) (mode (editor-assoc self)))
    (om-with-focused-view self
     (push-button self (mode (editor-assoc self))))))

(defmethod om-view-click-handler ((self palette-win) where)
  (when (equal (call-next-method) self)
    (when (< (om-point-v where) 25)
      (let ((x (floor (om-point-h where) 25)))
        (setf x (click-button-inpal self x))))))

;;; pb pour les dialog items dans la palette
;;;(defmethod om-view-click-handler :after ((self palette-win) where)
;;;   (om-select-window (window (editor-assoc self))))

;(defmethod om-set-view-size ((self palette-win) size-point) (call-next-method))
;  (om-set-interior-size self size-point))


;====================================
;==========Buttons in palettes==========

(defmethod push-button ((view t) i)
   (om-with-focused-view view
       (om-draw-hilite-rect  (+ 1 (* 25 i)) 1 23 22)))

(defmethod unpush-button ((view t) i)
  (om-invalidate-corners view  (om-make-point (+ 1 (* 25 i)) 1) (om-make-point (+ 24 (* 25 i)) 23)))


(defun fill-button (i)
  (om-fill-rect (+ 1 (* 25 i)) 1 22 21))


(defvar *palette-mode* nil)

(defmethod click-button-inpal (self mode)
  (push-button self mode)
  (setf *palette-mode* mode)
  (om-init-motion-functions self 'make-push-button 'release-push-button)
  mode)

(defmethod make-push-button ((self t) where)
   (when  *palette-mode*
     (let ((r (om-pts-to-rect (om-make-point (+ 1 (* 25 *palette-mode*)) 0)  
                              (om-make-point (+ 1 (* 25 (+ 1 *palette-mode*))) 25)))) 
       (unless (om-point-in-rect-p where r)
         (unpush-button self *palette-mode*)
         (setf *palette-mode* nil)))))

(defmethod release-push-button ((self t) where)
   (when *palette-mode*
     (palette-act self *palette-mode*)
     (unpush-button self *palette-mode*)))








