;OpenMusic
;
;Copyright (C) 1997, 1998, 1999, 2000 by IRCAM-Centre Georges Pompidou, Paris, France.
; 
;This program is free software; you can redistribute it and/or
;modify it under the terms of the GNU General Public License
;as published by the Free Software Foundation; either version 2
;of the License, or (at your option) any later version.
;
;See file LICENSE for further informations on licensing terms.
;
;This program is distributed in the hope that it will be useful,
;but WITHOUT ANY WARRANTY; without even the implied warranty of
;MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;GNU General Public License for more details.
;
;You should have received a copy of the GNU General Public License
;along with this program; if not, write to the Free Software
;Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
;
;Authors: Gerard Assayag and Augusto Agon


;DocFile
;This file implementss the meta-object OMFolder.
;Folder are not computable objects in OM.
;Last Modifications :
;18/10/97 first date.
;DocFile
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(in-package :om)

;-----------------
;The Folder Class in OM
;-----------------
#|
(defclass OMFolder (OMPersistantFolder) 
   ((elements :initform nil :initarg :elements :accessor elements))
   (:documentation "The class of the OM folders. #enddoc#
#seealso# (OMPersistantObject OMWorkspace) #seealso#
#elements# A list of objects contained in the Folder i.e. patches, maquettes or anothers folders. #elements#"))
|#


;--------------------------------------------------
;Method redefinition of OMpersistantObject
;--------------------------------------------------

(defmethod get-elements ((self OMFolder))
   "Folder's elements are patches, maquettes or another folders."
   (elements self))

(defmethod get-class-icon ((self OMfolder)) 'folder-icon-frame)

(defmethod get-object-insp-name ((self OMFolder)) "folder")

(defmethod get-editor-class ((self OMFolder)) 'FolderEditor)

(defmethod OpenEditorframe ((self OMFolder))
   "Show the folder as a container."
   (or (editorframe self) 
       (panel (open-new-nonrelationFrame self (name self) (get-elements self)))))

;--------------------------------------------------
;Other Methods
;--------------------------------------------------
(defmethod folder-p ((self OMFolder)) t)
(defmethod folder-p ((self t)) nil)

(defmethod save-all-persistants ((self OMFolder))
   "Save a folder and its elements at the end of the OM session."
   (loop for item in (elements self) do
         (save-all-persistants item)))

(defmethod om-paste ((self t) (elem OMFolder))
   (push elem (elements self))
   (let ((path (make-pathname :directory (append (pathname-directory (elements-pathname self)) (list (name elem)))))
         (old-elemts (elements elem)))
     (setf (mypathname elem) path)
     (create-directory path)
     (setf (elements elem) nil)
     (mapc #'(lambda (el) (om-paste elem el)) old-elemts)))


;--------------------------------------------------
;Tools
;--------------------------------------------------
(defun omNG-make-new-folder (name &optional (posi (om-make-point 0 0)))
   (let ((rep (make-instance 'OMFolder :name name  :icon 186)))
     (set-icon-pos rep posi)
     rep))


;Used to prevent to put a folder in self.
(defmethod ancestor-p ((self t) (container t)) nil)

(defmethod ancestor-p ((self OMFolder) (container OMFolder))
   (cond
    ((eq self container) t)
    (t (let ((list (get-elements self))
             (rep nil))
         (loop while list do
               (when (ancestor-p (pop list) container)
                 (setf rep t)
                 (setf list nil)))
         rep))))


;;=========THIS IS THE SPECIAL FOLDER WITH global vars===================

(defvar *om-globalsfolder* nil "The globals folder in the workspace.")
(defvar *globals-list* nil "Contains the initial global variables, loaded when you start the session.")

(defclass OMglobalsFolder (OMFolder) ()
   (:documentation "This is the special folder which contains globals variables. #enddoc#
#seealso# (OMFolder OMinstance) #seealso#"))

;--------------------------------------------------
;INTERFACE
;--------------------------------------------------

(defmethod get-documentation ((self OMglobalsFolder)) 
  "Folder use this special folder to define your global instances")

(defmethod get-elements ((self OMglobalsFolder))
   "Global folder's elements are OMinstances." (elements self))

(defmethod get-class-icon ((self OMglobalsFolder)) 'globals-folder-icon)

(defmethod get-editor-class ((self OMglobalsFolder)) 'GlobalsFolderEditor)

(defmethod OpenEditorframe ((self OMglobalsFolder))
   "Show the globals folder 'self' as a container."
   (or (editorframe self)
       (panel (open-new-nonrelationFrame self (name self) (get-elements self)))))

;--------------------------------------------------
;Tools
;--------------------------------------------------
(defmethod mus-patch ((self t)) nil)


(defun init-globalsfolder ()
   "Make an instance of global folder and store it in the *om-globalsfolder* variable." 
   (setf *om-globalsfolder* (omNG-protect-object (make-instance 'OMglobalsFolder
                                                   :Name "globals"
                                                   :icon 201
                                                   :elements (copy-list *globals-list*)))))


(defun intiglobalslist ()
  "Called when you start OM, this function load all globals variables saved in the current workspace."
  (declare (special *current-workspace* *instance-to-load*))
  (let ((directories (om-directory (make-pathname :directory  (append (pathname-directory (mypathname *current-workSpace*))
                                                                      (list "globals"))) 
                                   :directories nil :files t)))
    (setf *globals-list*
          (loop for path in directories  
                when 
                (and  (om-persistant-p path)
                         (equal (file-type path) :INST))
                collect (let (newobj)
                          (setf *loaading-stack* nil)
                          (eval-non-text-file path)
                          (setf newobj *instance-to-load*)
                          (setf (mypathname newobj) path)
                          (setf *instance-to-load* nil)
                          (update-patches-pile)
                          (when (mus-patch (instance newobj))
                            (compat-get-old-picts (mypathname newobj) (mus-patch (instance newobj))))
                          (setf (loaded? newobj) t)
                          (setf (saved? newobj) t)
                          newobj)))))






