;OpenMusic
;
;Copyright (C) 1997, 1998, 1999, 2000 by IRCAM-Centre Georges Pompidou, Paris, France.
; 
;This program is free software; you can redistribute it and/or
;modify it under the terms of the GNU General Public License
;as published by the Free Software Foundation; either version 2
;of the License, or (at your option) any later version.
;
;See file LICENSE for further informations on licensing terms.
;
;This program is distributed in the hope that it will be useful,
;but WITHOUT ANY WARRANTY; without even the implied warranty of
;MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;GNU General Public License for more details.
;
;You should have received a copy of the GNU General Public License
;along with this program; if not, write to the Free Software
;Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
;
;Authors: Gerard Assayag and Augusto Agon



;DocFile
;This file contains the meta-object OMMethod.
;Last Modifications :
;18/10/97 first date.
;DocFile
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(in-package :om)

;-----------------
;The Methods in OM
;-----------------
#|
(defclass OMMethod (standard-method OMPersistantObject) 
   ((saved-connections :initform nil :accessor saved-connections)
    (graph-fun :initform nil :accessor graph-fun)
    (compiled? :initform t :accessor compiled?)
    (pictu-list :initform nil :accessor pictu-list)
    (class-method-p :initform nil :accessor class-method-p))
   (:documentation "The class of the OM method metaobject. #enddoc#
#seealso# (omgenericfunction) #seealso#
#saved-connections# Used to save connections beetween boxes in the method definition. #saved-connections#
#graph-fun# A list of the boxes which define the method. #graph-fun#
#compiled?# Nil if the method was modified and not compiled. #compiled?#
#class-method-p# T if the method is a class method. 
Class methods are the init-instance method and slot reader and writer. #class-method-p#")
   (:default-initargs :protected-p t)
   (:metaclass omstandardclass))
|#


;--------------------------------------------------
;Method redefinition of OMpersistantObject
;--------------------------------------------------
(defmethod allow-alias ((self OMMethod))
   "Methods do not allow alias" nil)

(defmethod get-documentation ((self OMMethod))
   "Methods have not documentation, see the generic fonction of the method.")

(defmethod get-elements ((self OMMethod))
   "Elements of a method are boxes."
   (graph-fun self))

(defmethod get-object-insp-name ((self OMMethod)) "Method")

(defmethod obj-file-type ((self OMMethod)) :METH)
(defmethod obj-file-extension ((self OMMethod)) "ome")

(defmethod omNG-add-element ((self OMMethod) elem)
   "Add the box 'elem' to the elements of 'self'."
   (setf (mycontainer elem) self)
   (push elem (graph-fun self)))

(defmethod omng-remove-element ((self OMMethod) elem)
   "Delete the box 'elem' from the elements of 'self'."
   (setf (graph-fun self) (remove elem (graph-fun self) :test 'equal)))


(defmethod method-name-ID ((self OMMethod))
  "Cons a unique name for a method : 'name-qualifier-spec1x...xspecn'."
  (let ((qualy (car (method-qualifiers self)))
        (spec (loop for item in (method-specializers  self)
                  collect (subseq (string (class-name item)) 0 1))))
    (if qualy
        (format nil "~D-~D~{x~D~}" (name self) (string qualy) spec)
      (format nil "~D~{x~D~}" (name self)  spec))))
       
        
(defmethod omNG-save ((self OMMethod) &optional (thepath nil))
  "Save the method 'self'."
  (setf *libs-to-load* nil)
  (let ((newpath (make-pathname :directory (pathname-directory thepath) 
                                :name (method-name-ID self)
                                :type "ome"))
        (method-code (om-save-methods self)) )
    (loop while (probe-file newpath)
          for i = 0 then (+ i 1) do
          (setf newpath (make-pathname :directory (pathname-directory newpath)
                                       :name (string+ (pathname-name newpath) (format nil "~D" newpath i))
                                       :type (pathname-type newpath))))
    (delete-file-protection newpath)
    (WITH-OPEN-FILE (out newpath :direction :output :if-does-not-exist :create 
                         :if-exists :supersede ) ;;;        :external-format :METH)
      (write-header self  out)
      (prin1 '(in-package :om) out)
      (prin1 `(load-lib-for-first ',(remove-duplicates *libs-to-load* :test 'string-equal)) out)
      (let ((*package* (find-package :om)))
        (prin1 method-code out))
      (om-set-mac-file-type out :METH))
    (om-set-file-format newpath))
  (setf *libs-to-load* nil)
  (setf (saved? self) t))



(defmethod get-editor-class ((self OMMethod)) 'MethodEditor)

(defmethod OpenEditorframe ((self OMMethod))
   "Open the method editor, The method is redefined when this window is close."
   (or (editorframe self)
       (cond
        ((class-method-p self)
         (open-class-ommethode self))
        (t (if (graph-fun self)
             (panel (open-new-RelationFrame self (name self) (graph-fun self)))
             (not (dialog-message (string+ "Compiled method"))))))))

;--------------------------------------------------
;Other Methods
;--------------------------------------------------
(defmethod ommethod-p ((self OMMethod)) t)
(defmethod ommethod-p ((self t)) nil)

;The icon of a method is the same as its generic function.
(defmethod icon ((self OMMethod))
   (icon (fdefinition (method-name self))))
 
(defmethod make-remove-method ((self OMMethod))
   "Delete the method 'self' from its generic function;"
   (when (mypathname self)
     (delete-file (mypathname self)))
   (remove-method (fdefinition (method-name self)) self))


(defmethod open-class-ommethode ((self OMMethod))
   "Open the editor of a special method (i.e. init-instance, slot's reader and writer)"
   (let* ((speci (method-specializers self))
          (theclass (car speci)))
     (when (editorframe theclass) (om-close-window (window (editorframe theclass))))
     (unless (graph-fun self)
       (init-graphfun-method 'init self))
     (if (graph-fun self)
       (panel (open-new-RelationFrame self (name self) (graph-fun self)))
       (not (dialog-message (string+ "I can not open this method, because it is empty !!"))))))

(defmethod method-equal ((self OMMethod) (obj OMMethod))
   "T if 'self' and 'obj' are the same method"
   (and (string-equal (name self) (name obj))
        (equal (method-specializers self) (method-specializers obj))
        (equal (method-qualifiers self) (method-qualifiers obj))))

(defmethod method-equal ((self t) (obj t))
   "T if 'self' and 'obj' are the same method" nil)

(defmethod qualifieurs ((self OMMethod))
  (let ((qual (method-qualifiers self)) rep)
    (when qual
      (cond 
       ((string-equal (string (car qual)) "BEFORE") (setf rep (list :before)))
       ((string-equal (string (car qual)) "AFTER") (setf rep (list :after)))
       ((string-equal (string (car qual)) "AROUND") (setf rep (list :around)))
       (t (setf rep (list (car qual)))))
      rep)))

(defmethod put-boxes-in-method ((self OMMethod) boxes)
   (let ((boxeslist (reverse boxes)))
     (setf (graph-fun self) nil)
     (loop for item in boxeslist do
          (when item (omng-add-element self item)))))


;-----------------
;Method Definition
;-----------------

;---------------------------------------------------
;select the optional keys in the function definition
;---------------------------------------------------
;(defun parse-defmethod* (name args)
;   (unless (function-spec-p name) (error "Illegal arg ~S" name))
;   (let* ((theargs args)
;          (body? nil)
;          qualy lambda-list icon numouts initvals doc menuins body indoc)
;     (when (or (equal (car theargs) :after) (equal (car theargs) :before) (equal (car theargs) :around))
;       (setf qualy (pop theargs)))
;     (setf lambda-list (pop theargs))
;     (loop while (and theargs (not body?))
;           do
;           (cond
;            ((equal (car theargs) :numouts)  (pop theargs) (setf numouts  (pop theargs)))
;            ((equal (car theargs) :initvals) (pop theargs) (setf initvals (pop theargs)))
;            ((equal (car theargs) :indoc)    (pop theargs) (setf indoc (pop theargs)))
;            ((equal (car theargs) :doc)      (pop theargs) (setf doc (pop theargs)))
;            ((equal (car theargs) :icon)     (pop theargs) (setf icon     (pop theargs)))
;            ((equal (car theargs) :menuins)  (pop theargs) (setf menuins   (pop theargs)))
;            ((stringp (car theargs))         (setf doc (pop theargs)))
;            (t                               (setf body theargs) (setf body? t))))
;     (unless numouts (setf numouts 1))
;     (unless doc (setf doc "no documentation for this function"))
;     (unless icon (setf icon 150))   ;an icon by default
;     (values qualy lambda-list numouts initvals icon indoc doc menuins body)))

(defun parse-defmethod* (name args)
   (unless (function-spec-p name) (error "Illegal arg ~S" name))
   (let* ((theargs args)
          (body? nil)
          qualy lambda-list icon numouts initvals doc menuins body indoc)
     (when (or (equal (car theargs) :after) (equal (car theargs) :before) (equal (car theargs) :around))
       (setf qualy (list (pop theargs))))
     (setf lambda-list (pop theargs))
     (loop while (and theargs (not body?))
           do
           (cond
            ((equal (car theargs) :numouts)  (pop theargs) (setf numouts  (pop theargs)))
            ((equal (car theargs) :initvals) (pop theargs) (setf initvals (pop theargs)))
            ((equal (car theargs) :indoc)    (pop theargs) (setf indoc (pop theargs)))
            ((equal (car theargs) :doc)      (pop theargs) (setf doc (pop theargs)))
            ((equal (car theargs) :icon)     (pop theargs) (setf icon     (pop theargs)))
            ((equal (car theargs) :menuins)  (pop theargs) (setf menuins   (pop theargs)))
            ((stringp (car theargs))         (setf doc (pop theargs)))
            (t                               (setf body theargs) (setf body? t))))
     (unless numouts (setf numouts 1))
     (unless doc (setf doc "no documentation for this function"))
     (unless icon (setf icon 150))   ;an icon by default
     (values qualy lambda-list numouts initvals icon indoc doc menuins body)))

;---------------------------------------------
;Take only the name parameter without the type
;---------------------------------------------
(defun get-lambda-var (list)
  (mapcar #'(lambda (nom) (first? nom)) list))


;-------------------------------------------------------------------------
;Define an OM method and an OM Generic function if this last doesn't exist
;-------------------------------------------------------------------------
;(defmacro defmethod* (name &rest args)
;   (multiple-value-bind (qualy lambda-list numouts initvals icon indoc doc menuins body)
;                        (parse-defmethod* name args)
;     (let ((lambda-var (get-lambda-var lambda-list)))
;       (unless initvals
;         (setf initvals `',(make-list (length lambda-var) :initial-element nil)))
;       (unless indoc
;         (setf indoc `',(make-list (length lambda-var) :initial-element "no documentation")))
;       `(progn
;          (unless (fboundp ',name)
;            (let* (gen-fun)
;              (setf gen-fun (defgeneric ,name ,lambda-var  
;                              (:documentation ,doc) 
;                              (:generic-function-class ,*def-metaclass-genfun*)
;                              (:method-class ,*def-metaclass-method*)))
;              (setf (numouts gen-fun) ,numouts)
;              (setf (inputs-default gen-fun) ,initvals)
;              (setf (inputs-doc gen-fun) ,indoc)
;              (setf (inputs-menus gen-fun) ,menuins)
;              (setf (icon gen-fun) ,icon)
;              (setf (name gen-fun) ,(string name))))
;          (let ((method (if ,qualy
;                          (defmethod ,name ,qualy ,lambda-list  ,.body)
;                          (defmethod ,name ,lambda-list  ,.body))))
;            (setf (name method) ,(string name))
;            method)))))

(defmacro defmethod* (name &rest args)
   (multiple-value-bind (qualy lambda-list numouts initvals icon indoc doc menuins body)
                        (parse-defmethod* name args)

     (let ((lambda-var (get-lambda-var lambda-list)))
       (unless initvals
         (setf initvals `',(make-list (length lambda-var) :initial-element nil)))
       (unless indoc
         (setf indoc `',(make-list (length lambda-var) :initial-element "no documentation")))
       `(progn
          (unless (fboundp ',name)
            (let* (gen-fun)
              (setf gen-fun (defgeneric ,name ,lambda-var  
                              (:documentation ,doc) 
                              (:generic-function-class ,*def-metaclass-genfun*)
                              (:method-class ,*def-metaclass-method*)))
              (setf (numouts gen-fun) ,numouts)
              (setf (inputs-default gen-fun) ,initvals)
              (setf (inputs-doc gen-fun) ,indoc)
              (setf (inputs-menus gen-fun) ,menuins)
              (setf (icon gen-fun) ,icon)
              (setf (name gen-fun) ,(string name))))     
	 (let ((method (if ,qualy
			  (defmethod ,name ,.qualy ,lambda-list  ,.body)
			  (defmethod ,name ,lambda-list  ,.body))))
            (setf (name method) ,(string name))
            method)))))

;--------------------------Tools

;-----------save methods as patches.
;When you delete a class, methods specialized by it, are deleted also.
;OM creates a folder with this methods saved as patches.
(defun save-methods-as-patch (methodlist class)
   (declare (special *current-workspace*))
   (let* ((scrollframe (editorframe *current-workSpace*))
          (foldername (mk-unique-name  scrollframe (string+ class "methods")))
          (new-object (omNG-make-new-folder foldername (om-make-point 10 10)))
          new-frame)
     (setf new-frame (make-icon-from-object  new-object 10 10 1 1))
     (omg-add-element scrollframe new-frame)
     (loop for met in methodlist do
           (let ((newpatch (make-patch-from-method met)))
             (omNG-add-element new-object newpatch)
             (make-remove-method met)))
     foldername))

(defmethod make-patch-from-method ((self OMMethod))
   (let ((new-object (omNG-make-new-patch (string (method-name self)) (om-make-point 10 10)))
         (newboxes (mapcar #'(lambda (box) (special-omNG-save box)) (graph-fun self)))
         (connections (mk-connection-list (graph-fun self))))
     (setf (boxes new-object) nil)
     (mapc #'(lambda (box) (omNG-add-element new-object (eval box)))  newboxes)
     (setf (boxes new-object) (reverse (boxes new-object)))
     (remk-connections (boxes new-object) (loop for con in connections collect (load-connection con)))
     (make-remove-method self)
     new-object))

(defmethod special-omNG-save ((self t))
   "Only OMtypedin must be changed to save methods as patches"
   (omNG-save self t))







