;OpenMusic
;
;Copyright (C) 1997, 1998, 1999, 2000 by IRCAM-Centre Georges Pompidou, Paris, France.
; 
;This program is free software; you can redistribute it and/or
;modify it under the terms of the GNU General Public License
;as published by the Free Software Foundation; either version 2
;of the License, or (at your option) any later version.
;
;See file LICENSE for further informations on licensing terms.
;
;This program is distributed in the hope that it will be useful,
;but WITHOUT ANY WARRANTY; without even the implied warranty of
;MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;GNU General Public License for more details.
;
;You should have received a copy of the GNU General Public License
;along with this program; if not, write to the Free Software
;Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
;
;Authors: Gerard Assayag and Augusto Agon


;DocFile
;Method Editor definition.
;Last Modifications :
;18/10/97 first date.
;DocFile
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package :om)


;---------------------------------------------------------------------
;EDITOR
;---------------------------------------------------------------------

(omg-defclass methodEditor (patchEditor)  
  ((win-mod :initform nil :initarg :win-mod :accessor win-mod)
   (quali-buttons :initform nil :initarg :quali-buttons :accessor quali-buttons)
    (iconID :initform 145 :accessor iconID)
    (compiled? :initform nil :accessor compiled?))
   (:documentation "xxx"))


(defmethod get-editor-panel-class ((self methodEditor))  'methodPanel)


(defvar *abort-definition* nil "This flag is used when there are problems in the method definition")

(defmethod close-editor-before ((self methodEditor))
   (call-next-method)
   (unless *abort-definition*
     (modify-genfun (panel self))))

(defmethod om-view-close-event ((self methodEditor))
   (unless *abort-definition*
     (cond
      ((equal (class-method-p (object (panel self))) 'init)
       (setf (editorFrame (find-class (internp (get-init-method-class-name (name (object (panel self))))
                                               (symbol-package (method-name (object (panel self))))))) nil))
      ((ommethod-p (object (panel self)))
       (setf (editorFrame (object (panel self))) nil))))
   (setf *abort-definition* nil))


;---------------------------------------------------------------------
;PANEL
;---------------------------------------------------------------------

(omg-defclass methodPanel (patchPanel) 
   ((docu :initform "non documentation" :accessor docu))
   (:documentation "Method editors are instance of this class.
Elements of methodPanels are instaces of the boxframe class.#enddoc#
#seealso# (OMMethod boxframe) #seealso#
#docu# Used to store the method's documentation. #docu#"))


(defmethod modify-patch ((self methodPanel))
   "Set the compiled flag of the window to nil."
   (setf (compiled? (editor self)) nil))

;-----------------------------------------------------
;Code Generation and Method Definition from the window
;-----------------------------------------------------

(defun make-method-lambda-list (lis)
   "Cons a lambda list from a list 'lis' of input boxes."
   (let (standard optional rest keywrds)
     (mapc #'(lambda (boxframe)
               (let ((box (object boxframe)))
                 (cond
                  ((equal (keys box) nil) (push (list (read-from-string (name box)) (reference box)) standard))
                  ((equal (keys box) '&optional) (push (list (read-from-string (name box)) (defval box)) optional))
                  ((equal (keys box) '&rest) (push (read-from-string (name box)) rest))
                  ((equal (keys box) '&key) (push (list (read-from-string (name box)) (defval box)) keywrds))))) lis)
     (setf standard (reverse standard))
     (when optional
       (setf standard (append standard (list '&optional) (reverse optional))))
     (when rest
       (setf standard (append standard (list '&rest) (reverse rest))))
     (when keywrds
       (setf standard (append standard (list '&key) (reverse keywrds))))
     (cond
      ((> (length rest) 1) "Only one &rest argument")
      (t standard))))

(defun unify-list (l1 l2)
   (loop for item in l1 do
         (when l2 (pop l2)))
   (append l1 l2))

(defvar *pretraitement-passing* nil)

(defun make-new-om-function (name args numouts initvals doc icon qualy out-box fundoc)
   "Define a method in Common Lisp."
   (let (old-function thebody thecode new-method oldflag)
     (setf *pretraitement-passing* nil)
     (setf old-function (fboundp name))
     (when old-function
       (setf oldflag (class-method-p (car (get-elements old-function)))))
     (setf thebody (append (list 'values)
                           (mapcar #'(lambda (out)
                                       (gen-code  (object out) 0)) out-box)))
     (setf thebody (list 'let* *let-list*  thebody))
     (cond
      
      ((equal oldflag 'init)
       (let ((class (find-class (second (car args)) nil))
             initbox)
         (when class
           (setf thebody `((call-next-method) (unless from-file ,thebody)))
           (eval  `(defmethod initialize-instance (,(car args) &rest args &key (from-file nil))
                            (declare(ignore args))
                     ,.thebody))
           (setf new-method (car (get-elements old-function)))
           (push/replace-init class new-method)
           (resave-class class)
           (setf initbox (car (find-class-boxes (graph-fun new-method) 'OMBoxCallNextInit)))
           (when initbox
             (setf *pretraitement-passing* t)
             (eval   `(defmethod pretraitement (,(car args) &rest args)
                        (setf args (call-next-method))
                        (unify-list (list ,.(cdr (decode initbox))) args)))
             (setf *pretraitement-passing* nil)))))
      ((equal oldflag 'get)
       (let ((class (find-class (second (car args)) nil)))
         (when class
           (eval `(defmethod ,(getsetfunname (string name))  ,args ,thebody ))
           (setf new-method (car (get-elements old-function)))
           (resave-class class))))
      ((equal oldflag 'set)
       (let ((class (find-class (second (car args)) nil)))
         (when class
           (eval `(defmethod (setf ,(getsetfunname (string name)))  ,(reverse args) ,thebody ))
           (setf new-method (car (get-elements old-function)))
           (resave-class class))))
      (t (setf thecode `(defmethod* ,name ,.qualy ,args :numouts ,numouts 
                          :initvals (list ,.initvals) :indoc (list ,.doc) :icon ,(car (list! icon)) :doc ,fundoc ,thebody ))
         (setf new-method (eval thecode))
         (when (and (not old-function)  (listp icon))
           (icon-for-user-package (fdefinition (method-name new-method)) (second icon)))
         (setf (protected-p new-method) nil)
         (when oldflag
           (setf (class-method-p new-method) oldflag))
         (when (and old-function (protected-p old-function))
           (funfromkernel  *package-user*  old-function))))
     new-method))


(defmethod modify-genfun ((self methodPanel))
   "This method is called when you close the window."
   (handler-bind 
           ((error #'(lambda (c) (declare (ignore c))
                             (om-select-window (window self))
                             (let ((rep 
                                      (om-y-or-n-dialog (format nil "Method Definition Error.~%Do you want to close the window?~%If yes no modification will be made to the method"))))
                              (when rep
                                  (setf *abort-definition* t)
                                  (om-close-window (window self)))
                               (om-abort)))))
      
      (let* ((*package* (find-package :om))
               (funname (if (or (equal (win-mod (editor self)) :abs)
                                            (equal (win-mod (editor self)) :new)) (interne (name self)) 
                                    (method-name (object self))))
               (controls (get-subframes self))
               (out-box (sort (find-class-boxes controls 'outFrame) #'< :key #'(lambda (item) (indice (object item)))))
               (in-boxes (sort (find-class-boxes controls 'TypedInFrame) #'< :key #'(lambda (item) (indice (object item)))))
               (lambda-list (make-method-lambda-list in-boxes))
               (initvals (mapcar #'(lambda (item) (get-default-input-val (object item))) in-boxes))
               (doc (mapcar #'(lambda (item) (docu (object item))) in-boxes))
               (fundoc (docu self))
               (iconid (iconid (editor self)))
               (*let-list* nil) thequaly themethod pictlist)
        
        (cond 
         
         ((equal (win-mod (editor self)) :abs)      
          (cond 
           
           ((not out-box) 
            (when (om-y-or-n-dialog (format nil "At least one output must be used!~%Do you want to close the window?~%If yes no function will be defined"))
             (setf *abort-definition* t)
             (om-close-window (window self)))
           (om-abort))
           
           ((not in-boxes) 
            (when (om-y-or-n-dialog (format nil "Bad Lambda List!~%Do you want to close the window?~%If yes no function will be defined"))
                (setf *abort-definition* t)
                (om-close-window (window self)))
            (om-abort))
           
          (t (if (stringp lambda-list)
                  (dialog-message lambda-list)
               (let (newGenFun)
                 (setf themethod (make-new-om-function funname lambda-list 
                                                                                             (length out-box) initvals doc iconid nil out-box fundoc))
                 (setf (EditorFrame themethod) self)
                 (setf pictlist (pictu-list (object self)))
                 (setf (object self) themethod)
                 (setf newGenFun (fdefinition funname))
                 (setf (protected-p newGenFun) nil)
                 (omNG-add-element *package-user* newGenFun) 
                 (upDateUsermenu)
                 )))
           ))
         
        ((equal (win-mod (editor self)) :new)
         (let ((old-methods (copy-list (get-elements (fdefinition funname)))))
           (if (om-checked-p (second (quali-buttons (editor self)))) (setf thequaly (list :before)))
           (if (om-checked-p (third (quali-buttons (editor self)))) (setf thequaly (list :after)))
           (if (om-checked-p (fourth (quali-buttons (editor self)))) (setf thequaly (list :around)))
           (setf themethod (make-new-om-function funname lambda-list (length out-box) 
                                                                                       initvals doc (icon (object self)) thequaly out-box nil))
           (loop for met in old-methods do
                     (when (method-equal met themethod)
                         (when (mypathname met)
                             (delete-file (mypathname met)))))
           (setf (EditorFrame themethod) self)
           (when (EditorFrame (fdefinition funname))
             (om-close-window (window (EditorFrame (fdefinition funname)))))
           (setf pictlist (pictu-list (object self)))
           (setf (object self) themethod)))
         
         (t
          (setf thequaly (qualifieurs (object self))) 
          (setf themethod (make-new-om-function funname lambda-list (length out-box) 
                                                                                      initvals doc (icon (object self)) thequaly out-box nil))
          (when (mypathname (object self))
              (delete-file (mypathname (object self))))
          (setf (EditorFrame themethod) self)
          (when (EditorFrame (fdefinition funname))
            (om-close-window (window (EditorFrame (fdefinition funname)))))
          (setf pictlist (pictu-list (object self)))
          (setf (object self) themethod)))
        (setf (pictu-list (object self)) pictlist)
        (put-boxes-in-method themethod (mapcar #'(lambda (item) (object item)) controls))
        (unless (class-method-p themethod)
           (if (get-real-container (fdefinition funname))
               (set-path-element (get-real-container (fdefinition funname)) themethod)
             (set-path-element *package-user* themethod))
          ))))

  
;--------------------------------------------------
;INITS
;--------------------------------------------------

(defmethod add-window-buttons ((self methodPanel)) 
   "There are not buttons in a method panel." nil)

;Editor to define a Generic Function for the first time
(defun make-new-genfunwin (self name iconid doc)
   (declare (ignore self))
   (let* ((new-win (make-editor-window 'methodEditor
                                       (omNG-make-new-patch "genfunpatch") name nil 
                                       :winsize (om-make-point 500 300) 
                                       :winpos (om-make-point 50 38)
                                       :winshow nil))
          (addin-button (om-make-view 'unaire-fun-icon
                          :iconID 157
                          :object (make-instance 'input-maker)
                          :position (om-make-point 30 5)
                          :size (om-make-point 24 24)
                          :action #'(lambda (item) 
                                      (let* ((thetype (class-name (object item)))   ;aaa*
                                             (boxes (get-subframes (panel new-win))) 
                                             (i (- (length (find-class-boxes boxes 'TypedInFrame)) 1))
                                             (pos (om-make-point (+ 5 (* (+ 1 i) 50)) 45)))
                                        (omG-add-element (panel new-win)
                                                         (make-frame-from-callobj 
                                                          (make-new-typed-input (mk-unique-name (panel new-win) "input")
                                                                                thetype (+ i 1) pos)))))))
          (addout-button (om-make-view 'button-icon
                           :iconID 156
                           :position (om-make-point 5 5)
                           :size (om-make-point 24 24)
                           :action
                           #'(lambda(item) (declare (ignore item))
                              (let* ((boxes (get-subframes (panel new-win))) 
                                     (i (- (length (find-class-boxes boxes 'outFrame)) 1))
                                     (pos (om-make-point (+ 5 (* (+ 1 i) 50)) 240)))
                                (omG-add-element (panel new-win)
                                                 (make-frame-from-callobj 
                                                  (make-new-output (mk-unique-name (panel new-win) "output")
                                                                                     (+ i 1) pos)))
                                ;;;new
                                (set-field-size (panel new-win))
                                ;;;
                                ))))
          
          (thescroll (panel new-win))
          (editor (editor new-win)))
     (setf (iconid editor) iconid)
     (setf (docu thescroll) doc)
     (setf (win-mod editor) :abs)
     (setf (name thescroll) name)
     (om-add-subviews thescroll  addin-button addout-button)
     (om-select-window new-win)))


;Editor to define a new method for a defined Generic Function
(defmethod make-new-method ((self OMgenericFunction))
   (let* ((outputs (numouts self))
          (lambda-lis (arglist self))
          (ind 0) intype thescroll
          (new-win (om-make-window 'EditorWindow 
                     :window-title (name self)
                     :position (om-make-point 50 38) 
                     :close t
                     :window-show nil
                     :size  (om-make-point 400 300)))
          (editor (om-make-view 'MethodEditor
                    :ref nil
                    :owner new-win
                    :object (omNG-make-new-patch "genfunpatch")
                    :position (om-make-point 0 0) 
                                :size (om-make-point 400 300)))
          (pb (om-make-dialog-item 'om-radio-button (om-make-point 40 5) (om-make-point 76 16) "primary" ))
          (bb (om-make-dialog-item  'om-radio-button  (om-make-point 120 5) (om-make-point 72 16) "before" ))
          (ab (om-make-dialog-item  'om-radio-button (om-make-point 200 5) (om-make-point 72 16) "after" ))
          (arb (om-make-dialog-item  'om-radio-button (om-make-point 280 5) (om-make-point 76 16) "around" )))
     (setf (editor new-win) editor)
     (setf thescroll (panel new-win))
     (setf (object thescroll) (omNG-make-new-patch "genfunpatch"))
     (setf (name thescroll) (name self))
     (setf (win-mod editor) :new)
     (om-add-subviews thescroll pb bb ab arb)
     (setf (quali-buttons editor) (list  pb bb ab arb))
     (loop for item in lambda-lis  do
           (if (not (member item lambda-list-keywords :test 'equal))
             (let* ((inbox (make-new-typed-input (string-downcase (string item))
                                                 't ind (om-make-point (+ 5 (* ind 50)) 45))))
               (setf (enable inbox) nil)
               (setf (keys inbox) intype)
               (omG-add-element thescroll (make-frame-from-callobj inbox))
               (incf ind))
             (setf intype item)))
     (loop for i from 0 to  (- outputs 1) do
           (omG-add-element thescroll
                            (make-frame-from-callobj 
                             (make-new-output (mk-unique-name thescroll "output")
                                              i (om-make-point (+ 5 (* i 50)) 240)))))
     (om-select-window new-win)
     ;;;new
     (set-field-size thescroll)
     ))