;OpenMusic
;
;Copyright (C) 1997, 1998, 1999, 2000 by IRCAM-Centre Georges Pompidou, Paris, France.
; 
;This program is free software; you can redistribute it and/or
;modify it under the terms of the GNU General Public License
;as published by the Free Software Foundation; either version 2
;of the License, or (at your option) any later version.
;
;See file LICENSE for further informations on licensing terms.
;
;This program is distributed in the hope that it will be useful,
;but WITHOUT ANY WARRANTY; without even the implied warranty of
;MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;GNU General Public License for more details.
;
;You should have received a copy of the GNU General Public License
;along with this program; if not, write to the Free Software
;Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
;
;Authors: Gerard Assayag and Augusto Agon


;DocFile
;Editors for OMinstances OMlistinstances and OMconstant.
;Last Modifications :
;18/10/97 first date.
;DocFile
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(in-package :om)


;---------------------

(omg-defclass InstanceEditor (nonrelationEditor) ())

(defmethod get-editor-panel-class ((self InstanceEditor))  'InstancePanel)

(defmethod get-clipboard ((self InstanceEditor)) nil)


(defmethod update-editor-after-eval ((self InstanceEditor) val)
   (let (slot-boxes)
     (setf (object self) val)
     (apply 'om-remove-subviews (cons (panel self) (om-subviews (panel self))))
     (setf slot-boxes (slots-inst-boxes (ref  self) val))
     (mapc #'(lambda (frame)
               (omG-add-element (panel self) frame)) slot-boxes)
     (om-invalidate-view self t)))

;---------------------
;PANEL
;---------------------
(omg-defclass InstancePanel (nonrelationPanel) ()
   (:documentation "This is the class for OMInstance's editors. 
Elements of these editors are instance-icon-frame instances.#enddoc#
#seealso# (OMInstance instance-icon-frame listPanel) #seealso#"))

(defmethod instsroller-p ((self InstancePanel)) t)
(defmethod instsroller-p ((self t)) nil)  

(defmethod allow-move-element ((self InstancePanel))
   "You can not slot-value boxes in an instance editor." nil)



;***********************************************
;Special list Editor
;***********************************************

;-------------------
;Window
;-------------------



(omg-defclass list-enter-view (edit-text-enter-view) 
   ((direction :initform 0 :initarg :direction :accessor direction)))

;; new : text-view sur self directement
(defmethod exit-from-dialog ((self list-enter-view) newtext)
   (let* ((newpos (read-from-string newtext))
         (container (om-view-container self))
         (instance (instance (object container))))
     (if (and (integerp newpos) (>= newpos 0))
       (change-list-ranges container (direction self) newpos)
       (om-beep))
     (setf (text-view container) nil)
     (om-remove-subviews container self)))

(omg-defclass control-list (om-view)  ())


(defmethod initialize-instance :after  ((self control-list) &key)
   (om-add-subviews self
     (om-make-view 'button-icon
       :iconID 188
       :position (om-make-point 14 5)
       :size (om-make-point 11 11)
       :action  #'(lambda (item) 
                    (advance-range (om-view-container (om-view-container item)) 0)))   
     (om-make-view 'button-icon
       :iconID 164
       :position (om-make-point 83 5)
       :size (om-make-point 11 11)
       :action  #'(lambda (item) 
                    (advance-range (om-view-container (om-view-container item)) 1)))
     
     (om-make-view 'button-icon
       :iconID 202
       :position (om-make-point 144 5)
       :size (om-make-point 11 11)
       :action  #'(lambda (item) 
                    (advance-range (om-view-container (om-view-container item)) 2)))   
     (om-make-view 'button-icon
       :iconID 165
       :position (om-make-point 213 5)
       :size (om-make-point 11 11)
       :action  #'(lambda (item) 
                    (advance-range (om-view-container (om-view-container item)) 3)))))
                
(defmethod om-draw-contents :before ((self control-list))
  (om-with-focused-view self
       (om-draw-line 0 19 (w self) 19)
       (om-draw-string 30 14  (format nil "~D" (car (ranges (om-view-container self)))))
    (om-draw-string 160 14 (format nil "~D" (second (ranges (om-view-container self)))))))

;;; textview on the panel
(defmethod om-view-click-handler ((self control-list) where)
   (let ((editor (om-view-container self)))
     (call-next-method)
     (when (text-view editor)
       (exit-from-dialog (text-view editor) (om-dialog-item-text (text-view editor))))
     (cond
      ((and (>= (om-point-h where) 29) (<= (om-point-h where) 80))
       (om-add-subviews editor
         (setf (text-view editor)
            (om-make-dialog-item 'list-enter-view (om-make-point 29 4)
                                                      (om-get-text-edit-size (om-make-point 50 10))
                                                      (format () "~D" (car (ranges editor)))
                                                      :allow-returns t
                                                      :object self
                                                      :font *om-default-font2*))))
      ((and (>= (om-point-h where) 159) (<= (om-point-h where) 210))
       (om-add-subviews editor
         (setf (text-view editor)
            (om-make-dialog-item 'list-enter-view (om-make-point 159 4)
                                                      (om-get-text-edit-size (om-make-point 50 10))
                                                      (format () "~D" (second (ranges editor)))
                                                      :direction 1
                                                      :allow-returns t
                                                      :object self
                                                      :font *om-default-font2*)))))))


;---------------------------
; Editor
;---------------------------

(omg-defclass listEditor (instanceEditor) 
   ((control :initform nil :accessor control)
    (ranges :initform '(0 0) :accessor ranges)))

(defmethod get-editor-panel-class ((self listEditor))  'listPanel)

(defmethod panel-position ((self listEditor)) (om-make-point 0 20))

;;; enlev les 15 du scroller
(defmethod panel-size ((self listEditor)) 
   ;(om-make-point (- (w self) 15) (- (h self) 35))
   (om-make-point (w self) (- (h self) 20))
   )

(defmethod initialize-instance :after ((self listEditor) &rest l)
   (declare (ignore l))
   (let* ((control (om-make-view 'control-list 
                     :owner self
                     :position (om-make-point 0 0) 
                     :size (om-make-point (w self) 20))))
     (setf (control self) control)
     ))

(defmethod update-subviews :after ((self listEditor))
   (om-set-view-size (control self) (om-make-point (w self) 20)))

(defmethod check-new-range ((self listEditor) range)
   (setf range (list (max  0 (car range)) (max 0 (second range))))
   (if (< (first range) (length (instance (object self))))
     range
     (om-beep)))

(defmethod  remake-boxes ((self listEditor) new-range)
   (let (newboxes)
     (setf (ranges self) new-range)
     (apply 'om-remove-subviews (cons (panel self) (om-subviews (panel self))))
     (setf newboxes (make-elemnts-list-boxes (object self) (car new-range) (second new-range)))
     (mapc #'(lambda (frame)
                       (omG-add-element (panel self) frame)) newboxes)
     (om-invalidate-view self t)))


(defmethod advance-range ((self listEditor) dir)
   (let* ((x (car (ranges self)))
          (y (second (ranges self))) new-range)
     (setf new-range (check-new-range self (case dir 
                                             (0 (list (- x *size-list-inx*) y))
                                             (1 (list (+ x *size-list-inx*) y))
                                             (2 (list x (- y *size-list-iny*)))
                                             (3 (list x (+ y *size-list-iny*))))))
     (when new-range
       (remake-boxes self new-range))))
     
         

(defmethod change-list-ranges ((self listEditor) dir newval)
   (let* ((x (car (ranges self)))
          (y (second (ranges self))) new-range)
     (setf new-range (check-new-range self (case dir 
                                             (0 (list newval y))
                                             (1 (list x newval)))))
     (when new-range
       (remake-boxes self new-range)))
   )



;-------------------
;Panel
;-------------------

(omg-defclass listPanel (InstancePanel) ()
   (:documentation "This is the class for OMListInstance's editors. 
Elements of these editors are instance-icon-frame instances.#enddoc#
#seealso# (OMListInstance instance-icon-frame InstancePanel) #seealso#"))

(defmethod om-view-cursor ((self InstancePanel))  *om-arrow-cursor*)
