;OpenMusic
;
;Copyright (C) 1997, 1998, 1999, 2000 by IRCAM-Centre Georges Pompidou, Paris, France.
; 
;This program is free software; you can redistribute it and/or
;modify it under the terms of the GNU General Public License
;as published by the Free Software Foundation; either version 2
;of the License, or (at your option) any later version.
;
;See file LICENSE for further informations on licensing terms.
;
;This program is distributed in the hope that it will be useful,
;but WITHOUT ANY WARRANTY; without even the implied warranty of
;MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;GNU General Public License for more details.
;
;You should have received a copy of the GNU General Public License
;along with this program; if not, write to the Free Software
;Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
;
;Authors: Gerard Assayag and Augusto Agon


;DocFile
;This file define the abstract class for metaobjects's editor having  connections 
;i.e. patches maquettes, etc.
;Last Modifications :
;18/10/97 first date.
;DocFile
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(in-package :om)


;----------------------------
;relationEditor
;----------------------------
(omg-defclass relationEditor (nonrelationeditor)  ()
   (:documentation "This is the general class for window containing a relationPanel, so windows for Patches, 
maquettes and hierarchical class editors.#enddoc#
#seealso# (relationPanel) #seealso#"))

(defmethod get-editor-panel-class ((self relationEditor))  'relationpanel)


;-------------Initialization
(defun open-new-RelationFrame (object name elements &optional ref)
   (let* (newwindow)
     (setf newwindow 
           (make-editor-window (get-editor-class object)
                               object name ref 
                               :winsize (get-win-size object) 
                               :winpos (get-win-position object)
                               :winshow nil))
     (mapc #'(lambda (elem)
                    (let ((newframe (make-frame-from-callobj elem)))
                      (om-add-subviews (panel newwindow) newframe)
                      (add-subview-extra newframe))) elements)
     (mapc #'(lambda (elem)
               (update-graphic-connections elem elements)) (get-subframes (panel newwindow)))
     (add-window-buttons (panel newwindow))
     ;;; new jb
     (set-field-size (panel newwindow))
     (om-select-window newwindow)
     newwindow))

;----------------------------------------------------
;MCL Events
;----------------------------------------------------

(defmethod close-editor-before ((self relationEditor))  
   (call-next-method)
   (mapc #'(lambda (frame) (close-frame frame)) (get-subframes (panel self))))

(defmethod close-editor-after ((self relationEditor)) 
   (call-next-method)
   (om-view-close-event self))

;;;(defmethod close-window ((self relationEditor))
(defmethod om-view-close-event ((self relationEditor))
   (compile-patch (object (panel self)))
   nil)

;----------------------------------------------------
;Methods redefinitions 
;----------------------------------------------------
(defmethod set-clipboard ((self relationEditor) val)
   (let* ((container (panel self)))
     (setf (scroll-scrap-pa container) val)))

(defmethod get-clipboard ((self relationEditor))
   (let* ((container (panel self)))
     (scroll-scrap-pa container)))  

(defmethod editor-paste ((self relationEditor)) 
   ; (print (eval (first (get-clipboard self)))))
   "Paste from the clipboard to 'self'."
   (let ((container (panel self)))
     (if (text-view self)
       (om-paste-command (text-view self))
       (let ((val (get-clipboard self)))
       (when val
         (let ((connections (second val))
               (new-frames (eval (first val))))
           (loop for item in (get-actives container) do
                 (omG-unselect item))
           (mapcar #'(lambda (object)
                       (setf (name object) (mk-unique-name container (name object)))
                       (setf (frame-position object) (paste-position object))
                       (let ((new-frame (make-frame-from-callobj object)))
                         (omG-add-element container new-frame)
                         (omG-select new-frame))) new-frames)
           (remake-draggeds-connections container container new-frames connections))
         (om-invalidate-view container)
         ))) 
     t))

(defmethod editor-copy ((self relationEditor))
   "Copy in the clipboard the selected icons in 'self'."
   (let ((container (panel self)))
     (if (text-view self)
       (om-copy-command (text-view self))
       (let* ((subframes (mapcar #'(lambda (frame)
                                     (object frame)) (get-actives container)))
              (cannotcopy nil) copies)
         (loop for item in subframes
               while (not cannotcopy) do
               (setf cannotcopy (no-allow-copy-p  item)))
         (if cannotcopy
           (om-beep-msg (string+ "Sorry I can't  copy " cannotcopy))
           (progn
             (setf copies (mapcar #'(lambda (frame)
                                      (omNG-copy frame)) subframes))
             (set-clipboard self (list `(list ,.copies) (save-connections container container subframes))))))) t)) 

(defmethod do-undo ((self relationeditor)) 
  (let ((type (car (undo self))))
    (cond ((equal type 'remove)
           (let (framelist)
             (mapc #'(lambda (elem)
                       (let ((newframe (make-frame-from-callobj elem)))
                         (push newframe framelist)
                         (om-add-subviews (panel self) newframe)
                         (add-subview-extra newframe))
                       ) (cdr (undo self)))
             (setf (undo self) (append (list 'add) framelist)) 
           ))
          ((equal type 'add)
           (let (boxlist)
             (mapc #'(lambda (frame)
                       (push (object frame) boxlist)
                       (omg-remove-element (panel self) frame)
                       ) (cdr (undo self)))
             (setf (undo self) (append (list 'remove) boxlist)) 
           ))
          (t (setf (undo self) nil))))
  )

(defmethod editor-clear ((self relationEditor))
   "Clear the selected icons in 'self'."
   (let* ((container (panel self))
          (subframes (get-actives container)))
     (when subframes
       (delete-general container))))

(defmethod editor-cut ((self relationEditor))
   "Cut the selected icons in 'self'."
   (let* ((container (panel self))
          (subframes (get-actives container)))
     (if (text-view (editor self))
       (om-cut-command (text-view (editor self)))
       (when subframes
         (editor-copy self)
         (delete-general container)))))

(defmethod editor-save ((self relationEditor))
   "Relation windows save the object associated to the scroller, not the selected icons."
   (let ((thepatch (object (panel self))))
     (omNG-save thepatch nil)))

(defmethod alias-editor ((self relationEditor))
   "Make aliases of the selected icons."
   (let* ((container (panel self))
          (subframes (get-actives container))
          object)
     (when subframes
       (mapc #'omG-unselect subframes)
       (mapcar #'(lambda (oldframe)
                   (setf object (omNG-make-alias (object oldframe)))
                   (when object
                     (let ((new-frame (make-frame-from-callobj object)))
                       (omG-add-element container new-frame)
                       (omG-select new-frame)))) subframes))))

;----------------------------------------------------
;Other Methods 
;----------------------------------------------------

(defmethod window-last-saved ((self t))  nil)
(defmethod window-last-saved ((self relationEditor))
   "Close the window 'self' and load the last save version of the patch associated to the scroller."
   (update-last-saved (object (panel self))))

(defmethod paste-position ((self t))
   "The position for a duplicate command is calculated by this method."
   (borne-position (om-add-points (frame-position self) (om-make-point 20 20))))



;----------------------------
;relationPanel
;----------------------------

(omg-defclass relationPanel (nonrelationpanel) 
   ((scroll-scrap-pa :initform nil :allocation :class :accessor scroll-scrap-pa))
   (:documentation "This is the general class of editors containing objects plus a relation beetwen them.
Patches, maquettes and hierarchical class editors are sub-classes of this class.#enddoc#
#seealso# (patchPanel maquettePanel ClasstreePanel) #seealso#
#scroll-scrap-pa# Used for Copy and paste. #scroll-scrap-pa#"))

;------------------------------------------------------
;Other Methods
;------------------------------------------------------
(defmethod add-window-buttons ((self relationPanel))
   "Add extra buttons at the top-button in 'self'." nil)

(defmethod omG-add-element ((self relationPanel) frame)
   "Add a boxframe to the scroller, this method call the 'omNG-add-element' method with the objects referenced by 'self' and 'frame'."
   (omNG-add-element (object self) (object frame))
   (om-add-subviews self frame)
   (add-subview-extra frame)
   ;; dans om-add-subviews
   ;;(set-field-size self)
   )

(defmethod omg-remove-element ((self relationPanel) frame)
   "Remove a boxframe from the scroller, this method call the 'omng-remove-element' method with the objects referenced by 'self' and 'frame'."
   (omng-remove-element (object self) (object frame))
   (close-frame frame)
   (om-remove-subviews self frame))


(defmethod get-subframes ((self relationPanel))
   "Return a list with the boxes (boxframe's instances) subviews of 'self'."
   (let* (rep)
     (mapc #'(lambda (icon)
               (if (boxframe-p icon)
                 (push icon rep))) (om-subviews self))
     rep))

(defmethod get-actives ((self relationPanel))
   "Return a list with the selected boxes (boxframe's instances) subviews of 'self'."
   (declare (ignore scroll))
   (let* (rep)
     (mapc #'(lambda (icon)
               (if (and (boxframe-p icon)
                        (active-mode icon)) 
                 (push icon rep))) (om-subviews self))
     (reverse rep)))

(defmethod get-connections ((self relationPanel))
   "Get a list with all connections in 'self'."
  (let* (rep)
    (mapc #'(lambda (box)
              (setf rep (list+ rep (connections box)))) (get-subframes self))
    rep))

(defmethod get-actives-connections ((self relationPanel))
   "Get a list with the selected connections in 'self'."
   (let* (rep)
     (mapc #'(lambda (connec)
               (when (selected? connec)
                 (setf rep (list+ rep (list connec))))) (get-connections self))
     rep))


(defmethod click-in-connection ((self relationPanel) where)
   "Called by 'view-click-event-handler' this method verify if you click on a connection. If this is the case
this method set the select flag of the connection to T and return a list with the selected connections."
   (let ((controls (get-subframes self))
         rep)
     (loop for box in  controls 
           while (not rep) do
           (let* ((connection (connections box)))
             (loop for oneconnection in connection 
                   while (not rep) do
                       (let* ((where-click (point-in-connection oneconnection self where)))
                     (cond
                      ((numberp where-click)
                       (draw-connection oneconnection nil)
                       (unless (selected? oneconnection)
                         (unless (om-shift-key-p)
                           (mapc #'(lambda (control) 
                                     (deactivate-connect control)) 
                                 (remove oneconnection (get-connections self) :test 'equal)))
                         (setf (selected? oneconnection) t))
                       (cond 
                        ((om-shift-key-p)
                         (if (member (nth where-click (points oneconnection)) (point-sel oneconnection))
                           (setf (point-sel oneconnection) 
                                 (remove (nth where-click (points oneconnection))
                                         (point-sel oneconnection)))
                           (progn
                             (push (nth where-click (points oneconnection)) (point-sel oneconnection))
                             (scroll-points oneconnection))))
                        (t (unless (member (nth where-click (points oneconnection)) (point-sel oneconnection))
                             (setf (point-sel oneconnection)  (list (nth where-click (points oneconnection)))))
                           (scroll-points oneconnection)))
                       (draw-connection oneconnection t)
                       (setf rep t))
                      (where-click
                       (select-connection oneconnection)
                       (unless (om-shift-key-p)
                         (mapc #'(lambda (control) 
                                   (deactivate-connect control)) 
                               (remove oneconnection (get-connections self) :test 'equal)))
                       (setf rep t))
                      (t nil)))))) rep))


(defmethod release-drag-conections ((Self relationPanel) Where)
  (let (finalrect)
  (setf (point-sel *cur-drag-connection*) 
        (loop for item in (point-sel *cur-drag-connection*) 
              collect (nth (position item *first-point*) (points *cur-drag-connection*))))
       (unless (equal (points *cur-drag-connection*) *first-point*)
         (setf (nth 2 (connected? (object (nth (index *cur-drag-connection*) (inputframes (thebox *cur-drag-connection*))))))
               (points *cur-drag-connection*)))
       (setf finalrect (get-rectangle-space *cur-drag-connection*))
       (om-invalidate-corners self 
                              (om-make-point (- (min (first *init-con-rect*) (first finalrect)) 4)
                                             (- (min (third *init-con-rect*) (third finalrect)) 4))
                              (om-make-point (+ 4 (max (second *init-con-rect*) (second finalrect)))
                                             (+ 4 (max (fourth *init-con-rect*) (fourth finalrect)))))))

(defmethod make-drag-conections ((Self relationPanel) Where)
  (let* ((old-mouse *con-last-click*)
         (first-mouse *con-first-click*)
         (new-mouse old-mouse) finalrect)
    (setq new-mouse where)
    (setf finalrect (get-rectangle-space *cur-drag-connection*))
    (om-invalidate-corners self 
                              (om-make-point (- (min (first *init-con-rect*) (first finalrect)) 8)
                                             (- (min (third *init-con-rect*) (third finalrect)) 8))
                              (om-make-point (+ 8 (max (second *init-con-rect*) (second finalrect)))
                                             (+ 8 (max (fourth *init-con-rect*) (fourth finalrect)))))
    (loop for point in (point-sel *cur-drag-connection*) do

          (let ((pos (position point *first-point*))) 
            (setf (nth pos (points *cur-drag-connection*))
                  (scr-point-con *cur-drag-connection* self (nth pos *first-point*) 
                                 (om-subtract-points new-mouse first-mouse))))) 
    (draw-connection *cur-drag-connection* t)
    (setq *con-last-click* where)))


(defmethod make-move-after ((self relationPanel) dragged)
   "This method is called after moving one or more boxes, it redraw the connections involving in the moving operation."
   (let* ((frames (get-subframes self))
          (rest-frame (set-difference frames dragged :test 'equal)))
     (mapc #'(lambda (oneobject)
                       (redraw-connections oneobject)
                       (mapc #'(lambda (source)
                                         (when (is-connected? (object oneobject) (object source))
                                             (redraw-connections source))) rest-frame)
                       (om-invalidate-view oneobject)
                       ) dragged)
     t))



(defmethod make-delete-before ((self relationPanel) deleted cible)
   "This method is called after deleting one or more boxes, 
it redraw the connections involving in the deleying operation."
   (if (move-and-not-action self cible)
     (real-make-delete-before  self deleted)))

;Sometimes you can not redraw connections, for exemple when you copy two editors from a patch
;to one maquette...
(defmethod move-and-not-action ((self t) (cible t)) t)

(defmethod real-make-delete-before ((self relationPanel) deleteds)
   (let* ((frames (get-subframes self))
          (rest-frame (set-difference frames deleteds :test 'equal))
          (boxes (get-elements (object self))))
     (mapc #'(lambda (oneobject)
               (mapc #'(lambda (conection) (draw-connection conection nil)) (connections oneobject))
               (mapc #'(lambda (source)
                         (when (is-connected? (object oneobject) (object source))
                           (mapc #'(lambda (conection) (draw-connection conection nil)) (connections source))
                           (unconnected (object oneobject) (object source))
                           (update-graphic-connections source boxes))) rest-frame)) deleteds)
     t))


;;;;------------------------------------------------------
;;;; Events
;;;;------------------------------------------------------
(defmethod priority-connecctions ((self relationPanel)) nil)

(defmethod om-draw-contents  ((self relationPanel))
 (if (priority-connecctions self)
       (progn
          (call-next-method)
          (mapc #'(lambda (frame)  (box-draw-connections frame t)) (get-subframes self))
          )
     (progn
       (mapc #'(lambda (frame) (box-draw-connections frame t)) (get-subframes self))
        (call-next-method))))

;;;;Click on the scroller not in a subview.
(defmethod control-actives ((view relationPanel) where)
  (when (text-view (editor view))
    (exit-from-dialog (text-view (editor view)) (om-dialog-item-text (text-view (editor view)))))
  (if (click-in-connection view where)
      (mapc #'(lambda (control) 
                (omG-unselect control)) (get-actives  view))
    
    (let* ((float (om-subtract-points  (om-mouse-position view) where)))
      (unless (om-shift-key-p)
        (mapc #'(lambda (control) 
                  (deactivate-connect control)) (get-connections view)))
      (cond 
       ((om-command-key-p) 
        (make-undefined-box view where))
       (t (call-next-method))))))

(defmethod do-select-items-in-rect ((self relationPanel) rect) 
 (let (user-rect scratch-rect-i scratch-rect-n i-rect n-rect)
    (when rect
      (setf user-rect (om-make-rect (first rect) (second rect) (+ (first rect) (third rect)) (+ (second rect) (fourth rect))))
       (dolist (item (get-subframes self))
              (setf i-rect (om-pts-to-rect (om-view-position item) 
                                           (om-add-points (om-view-position item) (om-view-size item))))
              (setf scratch-rect-i (om-sect-rect user-rect i-rect))
              (unless (om-rect-empty scratch-rect-i) 
                (omG-select item))))))
