;OpenMusic
;
;Copyright (C) 1997, 1998, 1999, 2000 by IRCAM-Centre Georges Pompidou, Paris, France.
; 
;This program is free software; you can redistribute it and/or
;modify it under the terms of the GNU General Public License
;as published by the Free Software Foundation; either version 2
;of the License, or (at your option) any later version.
;
;See file LICENSE for further informations on licensing terms.
;
;This program is distributed in the hope that it will be useful,
;but WITHOUT ANY WARRANTY; without even the implied warranty of
;MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;GNU General Public License for more details.
;
;You should have received a copy of the GNU General Public License
;along with this program; if not, write to the Free Software
;Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
;
;Authors: Gerard Assayag and Augusto Agon

;DocFile
;This File contains generals functions and macros.
;In particular this file implements a set of Le Lisp Macros.
;Last Modifications :
;17/11/97 first date.
;DocFile
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(in-package :om)


;----------------------
; VARS
;----------------------

(defvar *om-version* 6.0 "Contains the current version of OM")
(defvar *version-string* "6.0.beta")


(defvar *let-list* nil "This var is a list for the boxes in mode ev-once, it is used while code generation")

(defvar *om-current-file-version*) ;non used, only for correction with first version (before 2.03)

(defvar *music-font* "omicron")

(defvar *micron-font* "omicron")
(defvar *extras-font* "omextras")
(defvar *heads-font* "omheads")
(defvar *signs-font* "omsign")


(defvar *mouse-window-event* (om-make-point 0 0))
;(setf *mouse-window-event* (om-make-point 0 0))

;-----------------------
;UTILITIES
;-----------------------

(defmethod name ((self string)) self)
(defun first?  (thing) (if (listp thing) (first thing) thing))
(defun list!   (thing) (if (listp thing) thing (list thing)))
(defun string+ (&rest strings) (eval `(concatenate 'string ,.strings)))
(defun list+ (&rest lists) (apply 'concatenate (append (list 'list) lists)))

(defun firstn (list  n )
   (cond
    ((< (length list)  n) list )
    (t  (butlast  list (- (length list) n)))))

(defun replace-in-list (list elem pos)
   (let* ((first-list (subseq list 0 pos))
          (second-list (subseq list (+ pos 1))))
     (append first-list (list elem) second-list)))

(defun insert-in-list (list elem pos)
  (if (> pos (- (length list) 1))
    (append list (list elem))
    (let* ((first-list (subseq list 0 pos))
           (second-list (subseq list pos)))
      (append first-list (list elem) second-list))))

(defun erase-n (list pos)
   (loop for item in list
         for i = 0 then (+ i 1)
         when (not (= i pos)) collect item))
  
(defun omlistp (list)
  (and list (listp list) (or (not (symbolp (car list))) (not (fboundp (car list))))))

(defun valued-val (val)
   (if (or (symbolp val) (omlistp val)) val (eval val)))

;-----------------------
; POSITION
;-----------------------

(defun interval-intersec (int1 int2)
   (when (and int2 int1)
     (let ((x1 (max (car int1) (car int2)))
           (x2 (min (cadr int1) (cadr int2))))
       (if (<= x1 x2)
         (list x1 x2)))))

(defun point-in-interval (point interval)
   (and (<= point (second interval))
            (>= point (first interval))))


(defun point-in-rectangle-p (point top left bottom right)
   (let ((rect (om-make-rect left top right bottom ))
           rep)
       (setf rep (om-point-in-rect-p point rect))
     rep))


(defun borne-position (pos)
   (om-make-point (max 0 (om-point-h pos)) (max 0 (om-point-v pos))))

(defun point2list (pt)
  (list (om-point-h pt) (om-point-v pt)))

(defun print-point (pt)
  (print (point2list pt))
  pt)


;----------------------------------------------------
; PARAMETERS
;----------------------------------------------------

(defun get-fixed-par-num (lis funname)
   (let ((j 0))
     (loop while lis do
           (let ((var (pop lis)))
             (if (not (member var lambda-list-keywords :test 'equal))
               (incf j)))) 
     (- j (length (get-keywords-fun funname)))))

(defun min-inp-number-from-arglist (lis)
   (let ((i 0) (continue t))
     (when lis
       (loop while continue do
             (let ((var (nth i lis)))
               (if (member var lambda-list-keywords :test 'equal)
                 (setf continue nil)
                 (incf i))
               (if (= i (length lis))
                 (setf continue nil)))))
     i))

(defun min-inp-number (fun)
   (min-inp-number-from-arglist (arglist fun)))


(defun get-keywords-fun (fun)
  (let ((args (arglist fun))
        rep)
    (loop while (and args (not rep)) do
          (when (equal (pop args) '&key)
            (setf rep t)))
    (setf rep nil)
    (loop for item in args 
          while (not (member item lambda-list-keywords :test 'equal)) do
          (push item rep))
    rep))




;-------------------------------------
; OM files 
;-------------------------------------
;this function return a string that is not contained in the list list
;list is a list of object which have a slot name
(defun unique-name-from-list (name list)
   (let* ((nompot name)
          (i 0)
          (seguir (match-name1 list nompot)))
     (loop while seguir do
           (incf i)
           (setf nompot (string+ name (format nil "~D" i)))
           (setf seguir (match-name1 list nompot)))
     nompot))

(defun match-name1 (list name)
   (let* ((next t))
     (loop while (and list next) do
           (if (string-equal name (name (pop list)))
             (setf next nil)))
     (if next nil t)))

(defun names-from-list (list)
  (loop for item in list collect (name item)))

(defun uniqueNameinpatch (name patch)
   (unique-name-from-list name  (get-elements patch)))


; load a non text file : maquette, patch, ...
(defun eval-non-text-file (path)
   (declare (special *current-workSpace*))
   (if (probe-file path) (om-load-file path)
     (if (si-o-no (string+ "I don't find the file " (string (real-pathname-name path)) " Do you want look for it?") nil )
         (let ((name (om-choose-file-dialog :prompt "look for file"
                                                                         :button-string "look for"
                                                                         :directory (make-pathname :directory
                                                                                                   (append (pathname-directory (mypathname *current-workSpace*))
                                                                                                           (list "elements"))))))
           (when name
               (eval-non-text-file name))))))


;--------------------------------------------------------------
; SPECIAL MESSAGES
;--------------------------------------------------------------
(defun dialog-message (message) 
   (declare (special *receiving-in-drag*)) 
   (if  *receiving-in-drag*
     (not (om-beep-msg message))
     (om-message-dialog message)))

;special yes-or-not dialog, you can not show a dialog while a drag&drop operation
(defun si-o-no (message &optional (rep t))
   (declare (special *receiving-in-drag*))
   (if  *receiving-in-drag*
     (progn
       (om-beep-msg message) rep)
     (om-y-or-n-dialog message)))

(defun om-message-abort (message)
  (om-message-dialog message)
  (om-abort))

(defun om-beep-msg (string)
   (om-beep)
   (print string) nil)


;----------------------------------------------------
; Boxes info
;----------------------------------------------------
;Find all elements in the list <lis> which their class is <class>
(defun find-class-boxes (lis class)
   (let* (rep)
     (mapc #'(lambda (box)
               (if (equal (class-name (class-of box))  class)
                 (push box rep))) lis)
     (reverse rep)))

(defun find-condition-boxes (lis fun)
   (let* (rep)
     (mapc #'(lambda (box)
               (when (funcall fun box)
                 (push box rep))) lis)
     (reverse rep)))


(defun default-from-list (list)
   (cond
    ((null list) 0)
    ((= (length list) 1) (car list))
    (t (- (car (last list 1)) (car (last list 2))))))


;-----------------------------------------------------------
;SubStrings
;-----------------------------------------------------------
;Left prefix of a string upto a CR or the end if no
(defun string-until-CR (string)
  (let ((index (search (format nil "~%") string)))
    (if index (subseq string 0 index) string)))

(defun string-until-space (string)
   (if (or (string-equal string "OMsituation2.01") (string-equal string "OMsituation3.0"))
     "OMSituation"
     (let ((index (search " " string)))
       (if index (subseq string 0 index) string))))

(defun string-from-space (string)
   (let ((index (search " " string)))
     (if index (subseq string index) string)))

(defun delete-spaces (string)
   (let ((pos  (position-if #'(lambda (x) (and 
                                           (not (equal x #\Linefeed))   ;;; test jean
                                           (not (equal x #\Space))
                                           (not (equal x #\Tab)))) string)))
     (when pos
       (subseq string pos))))

;-----------------------------------------------------------
; folders and files 
;-----------------------------------------------------------

(defun string-until-char (string char)
  (let ((index (search char string)))
    (if index (values (subseq string 0 index) (subseq string (+ index 1)))
        (values string nil))))

(defun str2list-path (str)
  (let (list)
    (loop while str do
          (let ((rep (multiple-value-list (string-until-char str ";"))))
            (setf str (second rep))
            (when (first rep) (push (first rep) list))))
    (reverse list)))

;1 ---------add om root to str pathname
(defun OMroot (str)
  (let* ((list (str2list-path str))
         (name (car (last list))))
    (if (equal name "")
      (make-pathname :device (pathname-device *om-root*) :directory (append (pathname-directory *om-root* ) (butlast list)))
      (make-pathname :device (pathname-device *om-root*):directory (append (pathname-directory *om-root*) (butlast list)) :name name))))


;3 ---------change pathname file in pathne folder (if index is given rename the path with index)
(defun OMfile2folder (path &optional index)
  (if index
    (make-pathname :directory (append (pathname-directory path) (list (string+ (pathname-name path) (format nil "~D" index)))))
    (make-pathname :directory (append (pathname-directory path) (list (pathname-name path))))))

;4 ---------name + type and other special chars

;; --> om-namestring ajoute
(defun real-pathname-name (path)
  (let ((type (pathname-type path)))
    (if (or (equal type :UNSPECIFIC) (not type))
      (om-namestring (pathname-name path))
      (string+ (om-namestring (pathname-name path)) "." type)
      )))

(defun real-pathname-name (path)
  (om-namestring (pathname-name path)))


(defun copy-folder (srcpath targetpath)
  (create-directory targetpath :if-exists :supersede)
  (loop for item in (om-directory srcpath :files t :directories t) do
        (if (directoryp item)
          (copy-folder item (make-pathname :device (pathname-device targetpath) 
                                           :directory (append (pathname-directory targetpath) (last (pathname-directory item)))))
          (om-copy-file item (make-pathname :device (pathname-device targetpath) 
                                            :directory (pathname-directory targetpath)
                                            :name (pathname-name item) :type (pathname-type item))))))



;TOOLS
;---------name of a directory
(defun name-of-directory (path)
  (car (last (pathname-directory path))))

(defun path2str (path)
  (format nil "~D" path))


;;;===============================================================================

(defun copy-file-sp (source-path target-path)
   (om-copy-file source-path
                     target-path
                     :if-exists :supersede))


(defun delete-file-protection (path)
   (declare (special *delete-file*))
   (when (and *delete-file* (probe-file path))
     (delete-file path)))
 
(defun corrige-path-space (name)
  name)

(defvar *last-file-loaded-dir* nil)

(defvar *example-file-folder* nil)

(defun init-ex-folder ()
   (setf *example-file-folder* (omroot "exampleFiles;")))

(om-add-init-func 'init-ex-folder)

(defmethod restore-path ((self pathname))
  (let ((dir (pathname-directory self)))
    (if (equal :relative (car dir))
        (make-pathname 
         :device (pathname-device (mypathname *current-workspace*))
         :directory (append (pathname-directory (mypathname *current-workspace*)) (cdr dir))
                     :name (pathname-name self) :type (pathname-type self))
      self)))

(defmethod restore-path ((self string))
  (restore-path (pathname self)))

(defun om-load-if (pathname fun)
  (let ((name (restore-path pathname)))
   (if (probe-file name)
       (funcall fun name)
     (let (default-path)
       (if *last-file-loaded-dir*
           (setf default-path (make-pathname :directory (pathname-directory *last-file-loaded-dir*)
                                             :device (pathname-device *last-file-loaded-dir*)
                                             :name (pathname-name  name)
                                             :type (pathname-type name))))
       (if (and default-path (probe-file default-path))
           (funcall fun default-path)
         (progn
            (setf default-path (make-pathname :directory (pathname-directory *example-file-folder*)
                                              :device (pathname-device *example-file-folder*)
                                              :name (pathname-name name)
                                              :type (pathname-type name)))
            (if (and default-path (probe-file default-path))
                (funcall fun default-path)
              (if (si-o-no (string+ "I don't find the file " (namestring name)
                                                 " Do you want to look for it?") nil)
                  (let ((name (om-choose-file-dialog :prompt (string+ "looking for " (pathname-name name)) :button-string "OK")))
                    (when name
                        (progn
                           (setf *last-file-loaded-dir* (pathname-directory name))
                           (funcall fun name))))))))))))



;;;=========================
;;; SOURCES LOADING UTILS

;;;(defun get-current-pathname () *load-pathname*)

(defun make-local-path (folder-path relative-path)
  (if folder-path
    (make-pathname :directory (append (pathname-directory (truename folder-path)) (butlast (str2list-path relative-path)))
                   :name (car (last (str2list-path relative-path))))
    (pathname relative-path)))


;-----------------------------------------------------------
;symbol management
;-----------------------------------------------------------
(defun interne        (string)     (intern (string-upcase string) :om))
(defun internp        (string pck) (intern (string-upcase string) pck)) 
(defun string2initarg (str)        (read-from-string (string+ ":" str)))
(defun num2string     (num)        (format () "~D" num))

;;; return int or nil si error
(defun string2int (str)
   (if (equal str "") nil
     (let ((rep (read-from-string str)))
       (if (integerp rep) rep nil))))

(defun string2num (str)
   (if (equal str "") nil
     (let ((rep (read-from-string str)))
       (if (numberp rep) rep nil))))


(defun list-subtypep (list typeList)
   "Checks if every elt in 'list' belongs to one of the types in 'typelist'"
  (every #'(lambda (elt) (some #'(lambda (type) (subtypep (type-of elt) type)) (list! typelist))) list))

(defun xcons (x y) "Inverse cons" (cons y x))

(defun put-quote (obj) (if (or (symbolp obj) (listp obj)) `',obj obj))

;push item in place but in the last position
(defmacro pushr (item place)
  `(if ,place
     (setf (cdr (last ,place)) (cons ,item nil))
     (setf ,place (list ,item))))


(defmacro ignore-error-msg (msg &body body)
   `(ignore-errors
    (handler-bind ((error #'(lambda (c) (declare (ignore c))
                             (om-beep-msg ,msg))))
       (progn ,.body))))


;; =============================================================================-======
;; LELISP MACROS 
;; =============================================================================-======

(defmacro ifnot (testform elseform &body body)
  `(if (not ,testform) ,elseform (progn ,.body)))

(defmacro repeat (count &body body)
  `(dotimes (,(gensym) ,count)
     ,.body))

(defmacro for ((var begin step end) &body body)
  (let ((s2 (gensym)) (s3 (gensym)))
    `(let ((,var ,begin) (,s2 ,step) (,s3 ,end))
       (if (> ,s2 0)
         (loop
           (when (> ,var ,s3) (return))
           (progn ,.body)
           (incf ,var ,s2))
         (loop
           (when (< ,var ,s3) (return))
           (progn ,.body)
           (incf ,var ,s2))))))

;; =============================================================================-======

(defmacro newl (lst elem) `(push ,elem ,lst))

(defmacro nextl (lst &optional symb)
  (if symb
    `(setq ,symb (pop ,lst))
    `(pop ,lst) ))

(defmacro vref (vect index) `(svref ,vect ,index))
(defmacro vset (vect index val) `(setf (svref ,vect ,index) ,val))

;; =============================================================================-======

(defmacro tell (outlet fun &rest args)
  (let ((args-var (gensym "ARGS-")) (fun-var (gensym "FUN-")))
    (if args
      `(let ((,args-var (list ,@args)) (,fun-var ,fun))
         (mapc #'(lambda (x) (apply ,fun-var x ,args-var)) ,outlet))
      `(mapc ,fun ,outlet))))

(defmacro ask (outlet fun &rest args)
  (let ((args-var (gensym "ARGS-")) (fun-var (gensym "FUN-"))
        (out-var (gensym "OUT-")) (result-var (gensym "RESULT-")))
    `(let ((,args-var (list ,@args)) (,fun-var ,fun) (,result-var nil))
       (dolist (,out-var ,outlet)
         (when (setq ,result-var (apply ,fun-var ,out-var ,args-var))
           (return)))
       ,result-var)))

(defmacro ask-all (outlet fun &rest args)
  (let ((args-var (gensym "ARGS-")) (fun-var (gensym "FUN-")))
    (if args
      `(let ((,args-var (list ,@args)) (,fun-var ,fun))
         (mapcar #'(lambda (x) (apply ,fun-var x ,args-var)) ,outlet))
      `(mapcar ,fun ,outlet))))

;; =============================================================================-======
;; The syntaxe is different from the Le_Lisp "with"
(defmacro with (l-place-value &body body)
  "Changes locally the value of \"setf-able\" places (like a \"let\" where places
would not be restricted to variables)."
  (let ((places (mapcar #'first l-place-value))
        (values (mapcar #'second l-place-value))
        (vars (mapcar #'(lambda (pv) (declare (ignore pv)) (gensym "WITH-"))
                      l-place-value)))
    `(let ,(mapcar #'list vars places)
       (unwind-protect
         (progn
           ,.(mapcar #'(lambda (place value) `(setf ,place ,value)) places values)
           ,.body)
         ,.(mapcar #'(lambda (place var) `(setf ,place ,var)) places vars)))))

;(let ((l '(a . b))) (with (((car l) 1) ((cdr l) 2)) (print l)))



;==========================================================================
;     From Utilities.lisp - old in containers folder
;==========================================================================

(defmacro defclas (name inheritance &rest rest)
  "simpler defclass"
  (let ((slots (pop rest)))
    `(defclass
       ,name
       ,inheritance
       ,(mapcar #'(lambda (slot) 
                    `( ,(first slot) :accessor ,(first slot) :initarg 
                       ,(intern (symbol-name (first slot)) 'keyword)
                       ,. (rest slot)))
                slots)
       ,. rest)))

(defmacro mki (class &rest rest)
  "simpler make-instance"
  `(make-instance ,class ,. rest))

(defmethod copy-instance-to ((source standard-object) (target standard-object))
   "Shallow copies the slots of 'source' to the slots of 'target'. 
Source must be subclass of target"
   (loop for slot in (mapcar 'car (class-instance-slots (class-of source)))
          do (setf (slot-value target slot) (slot-value source slot)))
   target)


;;; -------------------------------

;; fullratios are either ratios or lists (num denum)
;; use fullratio function to cast a fullratio to a number
;; use fdenominator and fdenominator to access to a fullratio num and denum
;; obviously here to avoid MACL automatic simplification of ratios.

(defmethod fullratio ((self list)) (/  (first self)  (second self)))
(defmethod fullratio ((self number)) self)
(defmethod fullratio ((self float)) (round self))

(defmethod fdenominator ((self t)) (denominator (fullratio self)))
(defmethod fdenominator ((self list)) (second self))
(defmethod fnumerator ((self t)) (numerator (fullratio self)))
(defmethod fnumerator ((self list)) (first self))




;-----------------------------------------
;Mmmm 4.2 to 4.3
;-----------------------------------------

#|
(defun get-finder-comment (path)
   (ccl::get-finder-comment path))

(defun set-finder-comment (path str)
   (ccl::set-finder-comment path str))
|#


;=====================================
;   Files Types
;=====================================


;Used to save files
;(deftype :PATC (x) (equal x :TEST))
;(deftype :INST (x) (equal x :TEST))
;(deftype :MAQT (x) (equal x :TEST))
;(deftype :CLAS (x) (equal x :TEST))
;(deftype :METH (x) (equal x :TEST))


;=====================================
;   METACLASSES VARIABLES
;=====================================

(defvar *def-metaclass-class*  'omstandardclass "The meta-class for om classes")
(defvar *def-metaclass-genfun* 'omgenericfunction "The meta-class for om generic functions")
(defvar *def-metaclass-slot*   'omslot "The meta-class for om slots")
(defvar *def-metaclass-method* 'ommethod "The meta-class for om methods")
(defvar *def-metaclass-patch*  'ompatch "The meta-class for om patches")
(defvar *def-metaclass-maq*    'ommaquette "The meta-class for om maquettes")
(defvar *def-metaclass-box*    'OMBox "The meta-class for om boxes")
(defvar *def-metaclass-box-fun*    'omboxcall "The meta-class for om method boxes")
(defvar *def-metaclass-box-type*    'OMBoxTypeCall "The meta-class for om type boxes")
(defvar *def-metaclass-box-edit*    'OMBoxEditCall "The meta-class for om class editor boxes")
(defvar *def-metaclass-box-patch*    'OMBoxPatch "The meta-class for om patch boxes")
(defvar *def-metaclass-box-maq*    'OMBoxMaquette "The meta-class for om maquette boxes")
(defvar *def-metaclass-box-slot*    'OMSlotsBox "The meta-class for om slot boxes")

(defvar *def-metaclass-box-inst*    'OMBoxInstance "The meta-class for om instance boxes")
(defvar *def-metaclass-box-tempo*    'TemporalBox "The meta-class for om temporal boxes")


;=======================
; EXIT OM
;=======================
(defun om-quit ()  
   (when (om-y-or-n-dialog  "Quit OpenMusic?")
       (om-confirmed-quit)))

;=======================
; Print IN listener
;=======================

(defvar *om-verbose* t)
(defun om-print (str)  
  (when *om-verbose*
    (print str)))


;=======================
; Mmmm
;=======================

(defun str-check (string)
  (let ((pos nil))
    (cond 
     ((equal *om-os* :win)    
      ;; multiple tabs
      ;(setf string (remove-duplicates string :test #'(lambda (x y) (equal x #\Tab))))
      ;; 
      (loop while (setf pos (position (code-char 381) string)) do
            (replace string (string (code-char 233)) :start1 pos)) 
      ;; 
      (loop while (setf pos (position (code-char 136) string)) do
            (replace string (string (code-char 224)) :start1 pos))
      ;; 
      (loop while (setf pos (position (code-char 144) string)) do
            (replace string (string (code-char 234)) :start1 pos))
      ;; 
      (loop while (setf pos (position (code-char 143) string)) do
            (replace string (string (code-char 232)) :start1 pos))
      ;; 
      (loop while (setf pos (position (code-char 157) string)) do
            (replace string (string (code-char 249)) :start1 pos))
      ;; 
      (loop while (setf pos (position (code-char 153) string)) do
            (replace string (string (code-char 244)) :start1 pos))
      ;; 
      (loop while (setf pos (position (code-char 148) string)) do
            (replace string (string (code-char 238)) :start1 pos))
      ;; 
      (loop while (setf pos (position (code-char 0) string)) do
            (replace string (string (code-char 231)) :start1 pos))
      )
   ((equal *om-os* :mac)
    ;; 
      (loop while (setf pos (position (code-char 233) string)) do
            (replace string (string (code-char 142)) :start1 pos)) 
      ;; 
      (loop while (setf pos (position (code-char 224) string)) do
            (replace string (string (code-char 136)) :start1 pos))
      ;; 
      (loop while (setf pos (position (code-char 234) string)) do
            (replace string (string (code-char 144)) :start1 pos))
      ;; 
      (loop while (setf pos (position (code-char 232) string)) do
            (replace string (string (code-char 143)) :start1 pos))
      ;; 
      (loop while (setf pos (position (code-char 249) string)) do
            (replace string (string (code-char 157)) :start1 pos))
     ;; 
      (loop while (setf pos (position (code-char 244) string)) do
            (replace string (string (code-char 153)) :start1 pos))
      ;; 
      (loop while (setf pos (position (code-char 238) string)) do
            (replace string (string (code-char 148)) :start1 pos))
    ;; 
    (loop while (setf pos (position (code-char 231) string)) do
            (replace string (string (code-char 141)) :start1 pos))
    )     
     (t nil))
    string))
