
;OpenMusic
;
;Copyright (C) 1997, 1998, 1999, 2000 by IRCAM-Centre Georges Pompidou, Paris, France.
; 
;This program is free software; you can redistribute it and/or
;modify it under the terms of the GNU General Public License
;as published by the Free Software Foundation; either version 2
;of the License, or (at your option) any later version.
;
;See file LICENSE for further informations on licensing terms.
;
;This program is distributed in the hope that it will be useful,
;but WITHOUT ANY WARRANTY; without even the implied warranty of
;MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;GNU General Public License for more details.
;
;You should have received a copy of the GNU General Public License
;along with this program; if not, write to the Free Software
;Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
;
;Authors: Gerard Assayag and Augusto Agon


(in-package :om)

;-------------------------------
;

;;; corrig pour le cas ou num = 1 -- jb08/06
(defmethod verify-list-exacte ((self list) type num array)
   (declare (ignore  array))
   (cond 
    ((subtypep type 'list)
     (if (and (>= (length self) num) (list-subtypep self type))
       (make-instance 'exactlist :thelist (first-n self num))
       self))
    ((subtypep type 'number)
     (if (and (>= (length self) num) (list-subtypep self type))
       (make-instance 'exactbpf 
         :thebpf (simple-bpf-from-list (if (> num 1) '(0 1) '(0)) (first-n self num) 'bpf 6))
       self))
    (t self)))

(defmethod verify-list-exacte ((self t) type num array)
   (declare (ignore type num array))
   self)


;---------------------------------------------------
(defclass* exactvalue () ())

(defmethod get-ieme ((self exactvalue) i array type)
   (nth i (get-points self type array)))

;----
(defclass* exactlist (exactvalue)
   ((thelist :initform nil :initarg :thelist :accessor thelist)))

(defmethod get-points ((self exactlist)  type array)
   (declare (ignore type array)) 
   (thelist self))

(defmethod exact-list-p ((self exactlist)) t) 
(defmethod exact-list-p ((self t)) nil) 

(defmethod get-bfp-for-draw ((self exactlist) array preci type) 
   (get-bfp-for-draw (thelist self) array preci type))

;----
(defclass* exactbpf (exactvalue)
   ((thebpf :initform nil :initarg :thebpf :accessor thebpf)))

(defmethod exact-bpf-p ((self exactbpf)) t) 
(defmethod exact-bpf-p ((self t)) nil) 

(defmethod get-points ((self exactbpf)  type array)
   (declare (ignore type array)) 
   (y-points (thebpf self)))

(defmethod get-ieme ((self exactbpf) i array type)
   (declare (ignore type array))
   (let ((fun (if (zerop (decimals (thebpf self))) 
                #'point-v 
                #'(lambda (x) (/ (om-point-v x) (expt 10.0 (decimals (thebpf self))))))))
     (funcall fun (nth i (point-list (thebpf self))))))

(defmethod get-bfp-for-draw ((self exactbpf) array preci type) (thebpf self))


;--------------data conversion---------------------- 


;---------get a list of points
(defmethod get-points ((self null)  type array)
   (declare (ignore type array)) nil)


(defmethod get-points ((self t) type array)
  (cond 
    ((subtypep type 'list)
     (make-list  (numcols array) :initial-element (list! self)))
    ((subtypep (type-of self) type)
     (make-list  (numcols array) :initial-element self))
    (t (let ((obj (objfromobjs self (get-super-default-value type))))
         (if obj (make-list (numcols array) :initial-element obj)
             (om-beep-msg (format nil "Can not coerce a ~D into a ~D" (type-of self) type)))))))

(defmethod get-points ((self number)  type array)
   (cond 
    ((subtypep  type 'number)
     (make-list  (numcols array) :initial-element self))
    ((subtypep type 'list)
     (make-list  (numcols array) :initial-element (list self)))
    (t (call-next-method))))


(defmethod get-points ((self list)  type array)
   (let ((len (length self)))
     (if (subtypep  type 'number)
       (loop for i from 0 to (- (numcols array) 1)
             collect (nth (mod i len) self))
       (call-next-method))))
    
;dos optiones


(defmethod get-points ((self function) type array)
   (get-points (loop for i from 1 to (numcols array)
                     collect (funcall self i)) type array))

(defmethod get-points ((self symbol) type array)
   (if (fboundp  self)
     (get-points (fdefinition self) type array)
     (call-next-method)))


(defmethod get-points ((self bpf)  type array)
   (if (subtypep  type 'number)
     (let* ((ranges (real-bpf-range self))
            (prec 6) ;(if pfield? (get-attribute self num "precision") 6))
            (bpfinterpol  (bpf-sample  self (/ (first ranges) (expt 10 (decimals self))) 
                                       (/ (second ranges) (expt 10 (decimals self)))
                                       (numcols array) 1 prec)))
       bpfinterpol)
     (call-next-method)))


;-----------get the i-eme element

(defmethod get-ieme ((self null) i array type) nil)

(defmethod get-ieme ((self t) i array type)
   (cond 
    ((subtypep type 'list) (list! self))
    ((subtypep (type-of self) type) self)
    (t (let ((obj (objfromobjs self (get-super-default-value type))))
         (or obj  (om-beep-msg (format nil "Can not coerce a ~D into a ~D" (type-of self) type)))))))

(defmethod get-ieme ((self number) i  array type) self)

(defmethod get-ieme ((self list) i  array type)
    (if (subtypep type 'number)
     (nth (mod i (length self)) self)
     (call-next-method))) ;ca dpend du type

(defmethod get-ieme ((self function) i array type) (funcall self i))

(defmethod get-ieme ((self symbol) i array type)
   (if (fboundp  self)
     (get-ieme (fdefinition self) i array type)
     (call-next-method)))

(defmethod get-ieme ((self bpf)  i array type)
   (nth i (list! (get-points self type array))))

;---------get a bpf for dra in the mini view some times is economic
(defmethod get-bfp-for-draw ((self t) array preci type) nil)
(defmethod get-bfp-for-draw ((self null) array preci type) nil)

(defmethod get-bfp-for-draw ((self number) array preci type)
   (simple-bpf-from-list (list 0 (- (numcols array) 1)) (list self self) 'bpf preci))

(defmethod get-bfp-for-draw ((self list) array preci type)
   (if (list-subtypep self 'number)
     (simple-bpf-from-list '(0 1) (get-points self 'number array) 'bpf preci)
     nil))

(defmethod get-bfp-for-draw ((self function) array preci type) 
   (simple-bpf-from-list '(0 1) (get-points self 'number array) 'bpf preci))

(defmethod get-bfp-for-draw ((self symbol) array preci type)
   (if (fboundp  self)
     (get-bfp-for-draw (fdefinition self) array preci type)
     (call-next-method)))

(defmethod get-bfp-for-draw ((self bpf) array preci type) self)   

;---------get a bpf for draw in the editor------------

(defmethod get-bfp-all-points ((self t) array preci type) 
   (get-bfp-for-draw self array preci type))

(defmethod get-bfp-all-points ((self number) array preci type) 
   (simple-bpf-from-list '(0 1) (get-points self 'number array) 'bpf preci))


;============================================================
;============================================================
(defclass control-data ()
  ((data :accessor data :initform nil)
   (control :accessor control :initform nil)))

(defmethod set-data ((self t)) t)

;;; ROWS = les fields (slots)
;;; COLS = nbelements

(defclass OMArray () 
   ((work-vector :initform nil  :accessor work-vector)
    (numrows :initform 1 :initarg :numrows :accessor numrows)
    (numcols :initform 1 :initarg :numcols :accessor numcols)
    (Lcontrols :initform nil  :accessor Lcontrols)
    (Lprecision :initform nil :type list :accessor Lprecision))
   (:documentation "Internal OMArray class"))

(defmethod valor-from-index ((self OMArray) index)
   (if (< index (numrows self))
     (let* ((initargs (get-all-initargs-of-class (type-of self)))
            (theslot (nth index initargs)))
       (funcall (internp (name theslot) (slot-package theslot)) self))
     (let ((control (nth (- index (numrows self)) (Lcontrols self))))
       (if control
         (second control)
         (error (format nil "Bad index ~D" index))))))
     
(defmethod get-array-i-j ((self OMArray) i j)
   (get-ieme (valor-from-index self i) j self (row-type self i)))

(defmethod get-array-i-j-for-eval ((self OMArray) i j)
   (get-ieme (valor-from-index self i) j self (row-type self i)))

(defmethod get-array-col ((self OMArray) num)
   (declare (ignore l))
   (loop for i from 0 to (+ (length (Lcontrols self)) (- (numrows self) 1))
         collect (get-array-i-j self i num)))

(defmethod get-array-row ((self OMArray) num)
   (get-points (valor-from-index self num) (row-type self num) self))

(defmethod get-array-row-for-eval ((self OMArray) num)
     (get-points (valor-from-index self num) (row-type self num) self))

(defun get-the-slot-type (array slot)
  (second (find (internp (name slot) (slot-package slot))
                (all-init-args-types  (class-of array)) :test 'equal :key 'first)))

(defmethod row-type ((self OMArray) num)
   (let ((slots (get-all-initargs-of-class (type-of self))) theslot)
     (setf theslot (nth num slots))
     (if theslot (thetype theslot) 'number)))

(defmethod numberp-col ((self OMArray) num)
   (subtypep  (row-type self num) 'number ))

;==========================================================
;==========================================================
(defclass* class-array (omArray control-data) ()
   (:icon 324))

(defmethod set-data ((self class-array))
  (setf (data self) 
       (loop for row from 0 to (+ (numrows self) (length (Lcontrols self)) -1)
                          collect (if (<= (numcols self) 1)
                                    (get-array-i-j-for-eval self row 0)
                                    (get-bfp-for-draw  (get-array-row self row) self (get-precision self) (row-type self row))))
       ))


(defmethod get-editor-class ((self class-array)) 'ArrayEditor)

(defmethod omNG-copy ((self class-array))
  (let* ((theclass (class-name (class-of self)))
            (slots (mapcan #'(lambda (slot)
                               (list (string2initarg (string (first slot))) 
                                     (omNG-copy (slot-value self (first slot))))) 
                           (get-initargs-of-class theclass))))
     `(let ((copy (make-instance ',theclass ,.slots)))
        (setf (numrows copy) ,(numrows self))
        (setf (numcols copy) ,(numcols self))
        (setf (Lcontrols   copy) ,(omng-copy (Lcontrols self)))
        (setf (Lprecision copy) ',(Lprecision self))
        copy)))


(defmethod omNG-save ((self class-array) &optional (values? nil)) 
    (let* ((theclass (class-name (class-of self)))
              (i -1)
              (slots (mapcan #'(lambda (slot)
                                              (list `(setf (slot-value copy ',(first slot)) 
                                                               ,(save-slot-array self (slot-value self (first slot)) (incf i))))) 
                                         (get-init-slots-of-class theclass))))
      `(if (find-class ',theclass nil)
         (let ((copy (make-instance ',theclass
                                 :numrows ,(numrows self)
                                 :numcols ,(numcols self))))
           ,.slots
           (setf (Lcontrols   copy)
                    (list ,.(loop for con in (Lcontrols self)
                               for j = (numrows self) then (+ j 1)
                               collect (save-control self con j))))
           (setf (Lprecision copy) ,(omng-save (Lprecision self)))
           copy))))

(defmethod save-control ((self class-array) con j )
   `(list ,(omng-save (car con)) ,(save-slot-array self (second con) j)))

(defmethod save-slot-array ((self class-array) (val t) num )
    (declare (ignore num))
    (omNG-save val))

(defmethod save-slot-array ((self class-array) (val function) num ) 
    (omNG-save (get-array-row self num)))


;------------INIT
(defmethod initialize-instance :after ((self class-array) &rest l)
    (declare (ignore l))
    (let ((slot-list (get-all-initargs-of-class (type-of self))))
      (setf (numrows self)  (length slot-list))
      (setf (Lprecision self) (list 3))))


;------------Look for the slot position of a slot.

(defmethod index2label ((self class-array) index)
    (let ((names (slot-names (type-of self))))
     (or (nth index names)
          (string (car (nth (- index (length names)) (Lcontrols self)))))))

(defmethod label2index ((self class-array) label)
   (let ((pos (label2index (type-of self) label)))
     (unless pos
       (setf pos (position (string label) (Lcontrols self) :test 'string-equal :key #'(lambda (x) (string (car x)))))
       (when pos (setf pos (+ pos (numrows self)))))
     pos))

(defmethod label2index ((class omclass) label)
   (label2index (class-name class) label))

(defmethod label2index ((class symbol) label)
   "Return the position of the labeled slot."
   (position (string label) (get-all-initargs-of-class class) :test 'string-equal :key 'name))

;-----------------------------------------------
(defmethod put-precision ((self class-array) list)
   (put-precision self (car list)))

(defmethod put-precision ((self class-array) number)
    (setf (Lprecision self) (list number)))

(defmethod get-precision ((self class-array))
   (car (Lprecision self)) )


(defmethod array-special-keyword ((self class-array) key)
   (declare (ignore key))
   nil)

(defmethod array-special-keyword-action ((self class-array) key val)
   (declare (ignore key val))
   nil)

(defmethod cons-new-array ((self class-array) args argkeys objs class)
   (if objs
     (objFromObjs (first args) self)
     (let* ((rep (make-instance (type-of self) :numcols (second args)))
               (initargs (get-all-initargs-of-class (type-of rep)))
               (precision 3) slot?)
       (loop for slot in (cdr (fixed-slots-list rep))
             for i = 2 then (+ i 1) do
             (setf (slot-value rep slot) (nth i args)))
       (loop while argkeys do
             (cond
              ((setf slot? (car (member (string (car argkeys)) initargs :test 'string-equal :key 'name)))
               (pop argkeys)
               (setf (slot-value rep (internp (name slot?) (slot-package slot?))) (verify-list-exacte (pop argkeys) (thetype slot?) (numcols rep) rep)))
              ((string-equal (string (car argkeys)) "precision")
               (pop argkeys)
               (setf precision (pop argkeys)))
              ((array-special-keyword rep (car argkeys))
               (array-special-keyword-action rep (pop argkeys) (pop argkeys)))
              (t
               (pushr (list (pop argkeys) (verify-list-exacte (pop argkeys) 'number (numcols rep) rep)) (Lcontrols rep)))))
       (put-precision rep precision)
       rep)))


;-------------------------------------------------------

(defmethod get-min-key-list ((self class-array)) nil)

;*
(defmethod fixed-slots-list ((self class-array)) '(numcols))


;;;-----

(defmethod rep-editor ((self class-array) num)
    (if (<= num (length (fixed-slots-list self)))
      (cond 
        ( (= num -2) (parsing-fun self))
        ( (= num -1) (Lprecision self))
        ( (= num 0) self)
        (t (funcall (nth (- num 1) (fixed-slots-list self)) self)))
      (get-array-row-for-eval self (- num (+ 1 (length (fixed-slots-list self)))))))


(defmethod Class-has-editor-p ((self class-array)) t)

(defmethod get-type-of-ed-box ((self class-array))  'arrayBox)

(defmethod draw-mini-view ((self t) (value class-array))
   (draw-obj-in-rect value 0 (w self)  0 (h self) (view-get-ed-params self) self))

(defmethod get-draw-visibles-list ((self  class-array) edparams)
  (loop for item in (cdr (nth  (cur-group-ind edparams) (panel-list edparams)))
        when (second item) collect (car item)))

(defmethod draw-obj-in-rect ((self  class-array) x x1 y y1 edparams view)
  (let* ((visibles (get-draw-visibles-list self edparams))
         (numvis (length visibles)) deltay deltax)
    (cond ((or (<= numvis 0) (<= (numcols self) 0)) t)
          ((= (numcols self) 1)
           (setf deltay (max 15 (round (- y1 y)  numvis)))
           (loop for i in visibles
                 for posy = y then (+ posy deltay) do
                 (let ((obj 
                        (if (data self) 
                          (nth i (data self))
                          (get-array-i-j-for-eval self i 0)
                          )))
                    (when (bpf-p obj)
                      (setf (bpfcolor obj) (nth i (color-list edparams))))
                    (draw-obj-in-rect obj x x1  posy (+ posy deltay) (default-edition-params obj) view)
                    (om-draw-line x (+ posy deltay) x1  (+ posy deltay))
                    (om-draw-line x (+ 1 posy deltay) x1  (+ posy  1 deltay)))))
          (t
           (setf deltay (max 15 (round (- y1 y)  numvis))
                 deltax (max 10 (round (- x1 x) (numcols self))))
           (loop for i in visibles
                 for posy = y then (+ posy deltay) do
                 (let ((bpf? (if (data self) 
                               (nth i (data self))
                               (get-bfp-for-draw  (get-array-row self i) self (get-precision self) (row-type self i))))) 
                   (if bpf?
                     (progn
                       (setf (bpfcolor bpf?) (nth i (color-list edparams)))
                       (draw-obj-in-rect  bpf? x x1  posy (+ posy deltay) nil view))
                     (loop for di = 1 then (+ di 1)
                           for posx = (+ x (* di deltax))
                           while (< posx x1)  do
                           (om-draw-line posx posy posx (+ posy deltay))))
                   (om-draw-line x (+ posy deltay) x1  (+ posy deltay))
                   (om-draw-line x (+ 1 posy deltay) x1  (+ posy  1 deltay))))))))


(defmethod cons-work-vector ((self class-array))
   (setf (work-vector self) 
         (make-array (numcols self)
                     :initial-contents  
                     (mat-trans (loop for i from 0 to (+ (length (Lcontrols self)) (- (numrows self) 1))
                                      collect (get-array-row self i))))))


(defmethod free-work-vector ((self class-array))
   (setf (work-vector self) nil))

(defmethod get-array-ranges ((self  class-array))
   (if (= 1 (numcols self)) (list -1 1)  (list -1 (numcols self))))


(defmethod set-exact-bpf ((self  class-array) index bpf)
     (if (< index (numrows self))
       (let* ((initargs (get-all-initargs-of-class (type-of self)))
                 (theslot (nth index initargs)))
         (setf (slot-value self (internp (name theslot) (slot-package theslot)))
                  (make-instance 'exactbpf :thebpf bpf)))
     (let ((control (nth (- index (numrows self)) (Lcontrols self))))
       (if control
         (setf (nth 1(nth (- index (numrows self)) (Lcontrols self))) (make-instance 'exactbpf :thebpf bpf))
         (error (format nil "Bad index ~D" index))))))

(defmethod set-exact-list ((self  class-array) index list)
    (let ((copylist (loop for item in list collect (clone item))))
     (if (< index (numrows self))
       (let* ((initargs (get-all-initargs-of-class (type-of self)))
                 (theslot (nth index initargs)))
         (setf (slot-value self (internp (name theslot) (slot-package theslot)))
                  (make-instance 'exactlist :thelist
                       copylist)))
     (let ((control (nth (- index (numrows self)) (Lcontrols self))))
       (if control
         (setf (nth 1 (nth (- index (numrows self)) (Lcontrols self))) (make-instance 'exactlist :thelist
                       copylist))
         (error (format nil "Bad index ~D" index)))))))
;-------------------------------------------------------
;EDition parmaeters
;-------------------------------------------------------

(defun get-visibles-rows (list)
    (loop for item in list
               for i = 0 then (+ i 1)
               when (second item) collect i))
        
(defmethod init-panel-list ((self class-array))
   (list (cons "All open" (loop for i from 0  to (- (numrows self) 1)
                                collect (list i t)))     
         (cons "All close" (loop for i from 0  to (- (numrows self) 1)
                                                 collect (list i nil)))))

(defclass ed-par-array ()
   ((panel-list :initform nil :initarg :panel-list :accessor panel-list)
    (cur-group-ind :initform 0 :initarg :cur-group-ind :accessor cur-group-ind)
    (color-list :initform nil :initarg :color-list :accessor color-list)
    (params-list :initform nil :initarg :params-list :accessor params-list)))

(defmethod get-edit-param ((self ed-par-array) param)
   (slot-value self param))

(defmethod omNG-copy ((self ed-par-array))
   `(make-instance ',(type-of self)
      :panel-list ',(panel-list self)
      :cur-group-ind ,(cur-group-ind self)
      :color-list ,(omng-copy (color-list self))))


(defmethod set-edition-params ((self class-array) box)
   (setf (edition-params box) 
         (make-instance 'ed-par-array
           :panel-list (init-panel-list self)
           :cur-group-ind 0
           :color-list (loop for i from 0  to (- (numrows self) 1)
                             collect *om-black-color*))))

(defmethod default-edition-params ((self class-array))
   (make-instance 'ed-par-array
           :panel-list (init-panel-list self)
           :cur-group-ind 0
           :color-list (loop for i from 0  to (- (numrows self) 1)
                             collect *om-black-color*)))

(defmethod save-value-params ((self class-array) params)
   (omNG-copy params))

(defmethod corrige-edition-params ((self class-array) (params ed-par-array))
   params)

(defmethod corrige-edition-params ((self class-array) (params t))
    (default-edition-params self))



(defmethod copy-value-params ((self class-array) box)
   (omng-copy (edition-params box)))

;---



;==========================================================
;==========================================================
(defclass component () 
   ((event :initform nil :initarg :event :accessor event)
    (val-list :initform nil :initarg :val-list :accessor val-list)
    (index :initform 0 :initarg :index :accessor index)))

(defmethod omng-copy ((self component))
   `(let ((newcomp (make-instance ',(type-of self)
                     :event ,(event self)
                     :val-list ,(omng-copy (val-list self))
                     :index ,(index self))))
      newcomp))


;-----------Tools--------------

(defmethod omng-copy ((self component))
   `(let ((newcomp (make-instance ',(type-of self)
                     :event ,(event self)
                     :val-list ,(omng-copy (val-list self))
                     :index ,(index self))))
      newcomp))

(defmethod label2index ((self component) (LineId string))
   (label2index (event self) lineId))

;-----------Interface-----------

(defmethod* new-comp ((self class-array) vals)
   :doc "Create a new component..."
   :icon 323
   (make-instance 'component
                    :event self
                    :val-list vals))

(defmethod* get-comp ((self class-array) (Comp-number integer))
   :doc "Get  a component...."
   :icon 323
   (make-instance 'component
     :event self
     :val-list (if (work-vector self)
                 (aref (work-vector self) Comp-number )
                 (get-array-col self  Comp-number ))
     :index Comp-number))

(defmethod* set-comp ((self class-array) (Comp-number integer) (New-Comp component))
   :doc "Set a component...."
   :icon 323
   t)

(defmethod* comp-list ((self component)) 
   :icon 323
   (val-list self))


(defmethod* comp-field ((self component) (LineId integer) &optional val) 
   :icon 323
   (if val 
     (progn (setf (nth LineId (val-list self)) val) self)
     (nth LineId (val-list self))))

(defmethod* comp-field ((self component) (LineId string) &optional val)
   (comp-field self (label2index self LineId) val))

(defmethod* fill-comp ((self component) val) 
   :icon 323
   (let* ((list (copy-list (list! val)))
          (defval (car (last list))))
     (setf (val-list self)
           (loop for i from 0 to (- (length (val-list self)) 1)
                 collect (or (pop list) defval))))
   self)


;======================================================
;BOX
;======================================================

(defmethod get-slot-in-out-names ((self class-array))
   (values '("self" "numcols") '(nil 1)
           '("object or object list" "How many vertical components for init ?")
           '(nil nil )))

;-----------------------------------------
(defclass arrayBox (OMBoxEditCall) ())

(defmethod numouts ((self arrayBox)) (length (inputs self)))

(defmethod call-gen-code ((self arraybox) numout)
   `(rep-editor ,(gen-code-call self) ,(numout2label-num self numout)))

(defmethod gen-code-call ((self arrayBox))
  (matrix-gen-code self (value self)))
  

(defmethod matrix-gen-code ((self arrayBox) (val class-array))
   (let* ((params (decode self))
          (fixinputs (length (fixed-slots-list val)))
          (theval (if (connected? (first (inputs self)))
                    `(objFromObjs ,(gen-code (first (inputs self)) (second (inputs self))) ,val)
                    `(funcall 'cons-new-array ,val 
                              (list nil ,.(cdr (first-n params (+ fixinputs 1)))) 
                              (list ,.(nthcdr (+ fixinputs 1)  params)) nil ',(class-name (reference self))))))
     theval))

(defmethod numout2label-num ((self arrayBox) num) 
   (get-index-rep (nth num (inputs self)) num))

(defmethod omNG-box-value ((self arrayBox) &optional (numout 0))
   (handler-bind ((error #'(lambda (c)
                             (when *msg-error-label-on*
                               (om-message-dialog (format nil "Error while evaluating the ~D factory ~%~D" (class-name (reference self)) 
                                                       (om-report-condition c))
                                               :size (om-make-point 300 200))
                               (clear-after-error self)
                               (abort)))))
     (let ((editorclass (class-name (reference self))))
       (cond
        ((equal (allow-lock self) "l") (special-lambda-value self editorclass))
        ((equal (allow-lock self) "o") (reference self))
        ((and (equal (allow-lock self) "x") (value self))
         (set-data (value self))
         (rep-editor (value self) (numout2label-num self numout)))
        ((and (equal (allow-lock self) "&") (ev-once-p self)) (rep-editor (value self) (numout2label-num self numout)))
        (t (let* ((args  (loop for input in (inputs self)
                               when (not (keyword-input-p input)) collect (omNG-box-value  input)))
                  (argkeys (eval-keywords self))
                  rep)
             (setf rep (cons-new-array (value self) args argkeys (connected? (first (inputs self))) editorclass))
             (if (null rep)
               (progn
                 (om-beep-msg (string+ "I can not construct a " (string editorclass) " with these parameters"))
                 (abort))
               (progn
                 (setf (value self) rep)
                 (when (showpict self)
                   (update-miniview (iconview (car (frames self))) (value self)))
                 (when (editorFrame self)
                   (update-editor-after-eval (editorFrame self) rep))
                 (when (equal (allow-lock self) "&")
                   (setf (ev-once-p self) t))
                 (numout2label-num self numout)
                 (set-data (value self))
                 (rep-editor (value self) (numout2label-num self numout))))))))))



(defmethod omNG-copy ((self arrayBox))
  `(let ((copy ,(omNG-make-new-boxcall (reference self)
                                       (frame-position self)
                                       (name self))))
     (setf copy (update-boxes ,self copy))
     (setf (showpict copy) ,(showpict self))
     (setf (minieditor? copy) ,(minieditor? self))
     (setf (edition-params copy) ,(omng-copy (edition-params self)))
     copy))


;------------------------------Keywords-----------------------


(defmethod get-keywords-from-box ((self arrayBox))
   (loop for item in (get-all-initargs-of-class (type-of (value self))) 
         collect (string2initarg (string  (name item)))))


(defclass editor-keyword (input-keyword) 
   ((box-ref :initform nil :initarg :box-ref :accessor box-ref)))

(defmethod get-icon-input ((self editor-keyword)) 266)

(defmethod do-add-one-input ((self arrayBox))
   (let* ((keywordlist  (get-keywords-from-box self))
          (normalinputs (find-class-boxes  (inputs self) 'input-funbox))
          (controlinputs (find-class-boxes  (inputs self) 'control-keyword))
          (paraminputs (find-class-boxes  (inputs self) 'editor-keyword))
          (usedkeywords (loop for item in paraminputs collect (value item))))
     (if (< (length usedkeywords) (length keywordlist)) 
       (setf (inputs self) (list+ normalinputs
                                  (list+ paraminputs
                                         (list (make-instance 'editor-keyword
                                                 :name "Keyword input"
                                                 :box-ref self
                                                 :value (car (reverse (set-difference keywordlist usedkeywords :test 'equal))) 
                                                 :doc-string (string+ "Choose from this list " (format () "~S"  keywordlist)))))
                                  controlinputs))
       (om-beep-msg "All parameters are already used."))))


(defmethod do-delete-one-input ((self arrayBox))
  (let* ((usedkeywords (find-class-boxes (inputs self) 'editor-keyword))
         (minkeys (get-min-key-list  (value self))))
    (if (> (length usedkeywords) (length minkeys))
      (setf (inputs self) (remove (car (last usedkeywords)) (inputs self) :test 'equal))
      (om-beep-msg "No more erasable keywords."))))
  

(defmethod eval-not-connected ((self editor-keyword))
   (let ((slot (find-if #'(lambda (x) (string-equal (name x) (string (value self)))) 
                        (get-all-initargs-of-class (type-of (value (box-ref self)))))))
     (if slot (initformfromslot slot) 
         (om-beep-msg (format nil "~D is not a slot of the instance." (string (value self)))))))
   
(defmethod code-not-connected ((self editor-keyword))
   (let ((slot (find-if #'(lambda (x) (string-equal (name x) (string (value self)))) 
                        (get-all-initargs-of-class (value (box-ref self))))))
     (if slot (initformfromslot slot) 
         (om-beep-msg (format nil "~D is not a slot of the instance." (string (value self)))))))

(defmethod get-index-rep ((self editor-keyword) n)
   (+ (length (fixed-slots-list (value (box-ref self)))) 1 
      (label2index (value (box-ref self)) (string (value self)))))

(defmethod get-index-rep ((self input-funbox) n) n)


;------------------------------------------
(defclass control-keyword (input-keyword) 
   ((box-ref :initform nil :initarg :box-ref :accessor box-ref)))

(defmethod do-add-one-keyword ((self arrayBox))
   (let* ((usedkeywords (loop for item in (find-class-boxes  (inputs self) 'control-keyword)
                              collect (value item)))
          (newval "K0"))
     (loop for i = 1 then (+ i 1)
           while (member newval usedkeywords :test 'string-equal :key 'string) do
           (setf newval (format nil "K~D" i)))
     (setf (inputs self) (list+ (inputs self)
                                (list (make-instance 'control-keyword
                                        :name "Control input"
                                        :box-ref self
                                        :value (string2initarg newval) 
                                        :doc-string "Optional control keyword. There exists 2 special keywords :attributes and :parsingFun"))))))

(defmethod do-delete-one-keyword ((self arrayBox))
   (let* ((usedkeywords (find-class-boxes (inputs self) 'control-keyword)))
     (when usedkeywords
       (setf (inputs self) (remove (car (last (inputs self))) (inputs self) :test 'equal)))))

(defmethod get-index-rep ((self control-keyword) n)
  (cond
    ((equal (value self) :parsing-fun) -2)
    ((equal (value self) :precision) -1)
    (t (+ (length (fixed-slots-list (value (box-ref self)))) 1
          (label2index (value (box-ref self)) (string (value self)))))))

(defmethod set-new-keyword ((self control-keyword) newtext)
   (let ((keywords (get-keywords-from-box (box-ref self)))
         (usedkeywords (find-class-boxes (inputs (box-ref self)) 'control-keyword)) pos)
     (when (member (symbol-name (read-from-string newtext)) keywords :test 'equal :key 'symbol-name)
       (format () "~D is a parameter of the Instance. Use another control keyword."  newtext)
       (error "error"))
     (setf pos (position (read-from-string newtext) usedkeywords :test 'equal :key 'value))
     (when pos
       (format () "~D is already used by the ~D control input." newtext (incf pos))
       (error "error"))))

(defmethod code-not-connected ((self editor-keyword))
   (let ((slot (find-if #'(lambda (x) (string-equal (name x) (string (value self)))) 
                        (get-all-initargs-of-class (type-of (value (box-ref self)))))))
     (if slot (initformfromslot slot) 
         (om-beep-msg (format nil "~D is not a slot of the instance." (string (value self)))))))


;==============================================================


(defmethod update-miniview ((self t) (type class-array))  (om-invalidate-view self t))



(AddGenFun2Pack '(comp-field fill-comp get-comp new-comp comp-list) *data-package*)

(AddClass2Pack '(class-array) *data-package* :position (list (om-make-point 175 50)))