(in-package :om)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;      OM     by Assayag&Agon
;;      IRCAM                 Centre Georges Pompidou     
;;      file-box.lisp
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;DocFile
;The file-box boxes are defined in this file.
;There are classes for accumulaters, iteraters, but also the graphic editor.
;Last Modifications :
;18/10/97 first date.
;DocFile
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


        
;===========================================================================
;PATCH

(defclass openfilepatch (patchForLoop) ()
   (:documentation "This class implements the special patches for the file-box boxes.  #enddoc#
#seealso# (OMPatch file-box-box) #seealso#"))


(defmethod OpenEditorframe ((self openfilepatch))
   (or (editorframe self)
       (panel (open-new-RelationFrame self (name self) (get-elements self)))))


(defmethod compile-patch ((self openfilepatch))
   "Code generates by Loop patches is generate by this method."
   (let* ((boxes (boxes self))
          (oldletlist *let-list*)
          (*let-list* nil)
          (stream-box (car (find-class-boxes boxes 'Streambox)))
          (final-box (car (find-class-boxes boxes 'OMFinalDo)))
          (in-boxes (sort (find-class-boxes boxes 'OMin) '< :key 'indice))
          (symbols (mapcar #'(lambda (thein) (setf (in-symbol thein) (gensym))) in-boxes)))
     (eval  `(defun ,(car (code self)) ,symbols
                 (with-open-file ,(gen-stream-options stream-box)
                   (let* ,*let-list*
                     ,(gen-code final-box 0)))))
     (setf *let-list* oldletlist)))


(defmethod compile-patch ((self openfilepatch))
  "Code generates by Loop patches is generate by this method."
  (let* ((boxes (boxes self))
         (oldletlist *let-list*)
         (*let-list* nil)
         (do-box (car (find-class-boxes boxes 'OMLoopDo)))
         (final-box (car (find-class-boxes boxes 'OMFinalDo)))
         (init-box (car (find-class-boxes boxes 'OMinitDo)))
         (in-boxes (sort (find-class-boxes boxes 'OMin) '< :key 'indice))
         (symbols (mapcar #'(lambda (thein) (setf (in-symbol thein) (gensym))) in-boxes))
         (loop-boxes (find-loop-boxes boxes))
         (loop-code (mapcan #'(lambda (theout)
                                (loop-gen-code theout 0)) loop-boxes))
         (acum-boxes (find-acum-boxes boxes))
         (acum-declaration (mapcar #'(lambda (acumm)
                                       (declare-closure acumm)) acum-boxes))
         (acum-inits (mapcar #'(lambda (acumm)
                                 (gen-code acumm -1)) acum-boxes))
         (stream-box (car (find-class-boxes boxes 'Streambox)))
         body init)
    (setf body (gen-code do-box 0))
    (setf init (gen-code init-box 0))
    (cond
     (loop-boxes
      (eval `(defun ,(intern (string (first (code self))) :om)  (,.symbols)
                (with-open-loop-file ,stream-box
                  (let (,.acum-declaration) ,.acum-inits
                       ,init
                       (loop ,.loop-code 
                             finally (return (values ,.(loop for i from 0 to (- (length (inputs final-box)) 1)
                                                             collect  (gen-code final-box i)))) do
                             (let* ,*let-list* ,body)))))))
     (t 
      (eval `(defun ,(intern (string (first (code self))) :om)  (,.symbols) 
                (with-open-loop-file ,stream-box
                  (let (,.acum-declaration) ,.acum-inits
                       (let* ,*let-list* 
                         ,init
                         ,body
                         (values ,.(loop for i from 0 to (- (length (inputs final-box)) 1)
                                         collect  (gen-code final-box i))))))))))
    
    (setf *let-list* oldletlist)))
   



(defmacro with-open-loop-file (self &body body)
  `(let* ((pathname ,(car (decode self)))
          (pathname (or pathname (om-choose-file-dialog :prompt "choose a file to read/write")))
          rep)
     (cond 
      ((equal (get-stream-type ,self) 'text)
       (let ((stream (open pathname :if-does-not-exist :create :direction ,(direction self) :if-exists ,(if-ex self))))
         (setf rep ,@body)
         (close stream)))
      ((equal (get-stream-type ,self) 'sdif)
       (let* ((dir (cond ((equal ,(direction self) :input) :eReadFile)
                         ((equal ,(direction self) :output) :eWriteFile)
                         (t :eReadWriteFile)))
              (stream (sdif-open-file (om-path2cmdpath pathname) dir)))
         (setf rep ,@body)
         (sdif-close-file stream)))
      (t nil))
      rep))




(defmethod get-editor-class ((self openfilepatch)) 'openfileEditor)

;----------------
;EDITOR
;----------------
(omg-defclass openfileEditor (loopEditor) ())

(defmethod get-editor-panel-class ((self openfileEditor))  'openfilePanel)

;----------------
;PANEL
;----------------
(omg-defclass openfilePanel (loopPanel) ()
   (:documentation "This is the class for the editor panel which define a file-box box.  #enddoc#
#seealso# (file-box-box box-with-patch-frame) #seealso#
#patch# This slot keeps the patch associated to the box. #patch#"))

(defmethod omg-remove-element ((self openfilePanel) (box inframe))
   "The inputs of a loop patch can be removed only from the file-box box."
   (om-beep-msg "delete inputs in the loop box"))

(defmethod omg-remove-element ((self openfilePanel) (box t))
   "Finally and streamFile boxes can not be removed."
   (if (or (equal (class-name (class-of (object box))) 'Streambox)
           (equal (class-name (class-of (object box))) 'OMfinalDo))
     (om-beep-msg "These boxes can not be erased")
     (call-next-method)))

(defmethod add-window-buttons  ((self openfilePanel))
   "Add iterator and accumulator buttons to the patch."
   (call-next-method))



;--------------------
(defclass OMinitDo (box-seq-call) ()
   (:documentation "This is the class for the finally boxes in a loop.#enddoc#
#seealso# (omloop-box loopIterators acumboxes) #seealso#"))

(defmethod allow-lock-button ((self OMinitDo))
   "Finally boxes do not allow a lock button." nil)

(defmethod no-allow-copy-p ((self OMinitDo))
   "Finally boxes can not be directly copied" t)

(defmethod* initDo  ((patch t) &rest oppatch) :numouts 0 
  :initvals '(nil) :indoc '("patch") :doc "Connect in this box the result of the loop evaluation you can add multiple outputs" :icon 216
  (declare (ignore oppatch)))

(defmethod numouts ((self ominitDo)) 0)
(defmethod get-frame-name ((self OMfinalDo)) (name self))
(defmethod get-boxcallclass-fun ((self (eql 'initDo))) 'OMinitDo)


(defmethod do-delete-one-input-extra ((self ominitDo))
   "When you erase one input from the finally box you must delete one input from the loop box also."
   (call-next-method)
   ;(let ((box (box (object (om-view-container (car (frames self)))))))
   ;  (setf (numouts box) (- (numouts box) 1))
   ;  (redraw-frame (car (frames box))) t)
   )

(defmethod do-add-one-input-extra ((self ominitDo))
   "When you add one input to the finally box you must add one input to the loop box also."
   (call-next-method)
   ;(let ((box (box (object (om-view-container (car (frames self)))))))
   ;  (setf (numouts box) (+ (numouts box) 1))
   ;  (redraw-frame (car (frames box)))
   ;  t)
   )

;----------------
(defclass file-box-box (omloop-box) 
   ()
   (:documentation "This is the class for the file-box boxes.#enddoc#
#seealso# (box-with-editor loopIterators acumboxes) #seealso#"))


(defmethod initialize-instance :after ((self file-box-box) &key controls)
   (declare (ignore controls))
   (let ((thein (omNG-make-new-boxcall (fdefinition 'StreamFile) (om-make-point 50 30) "StreamFile"))
         (thefinal (omNG-make-new-boxcall (fdefinition 'initDo) (om-make-point 50 170) "init")))
     ;;(print (patch self))
     (omNG-add-element (patch self) thein)
     (omNG-add-element (patch self) thefinal)))

(defmethod openeditorframe ((self file-box-box))
   (openobjecteditor (patch self)) nil)

(defmethod get-patch-editor-class ((self file-box-box)) 'openfilepatch)


;;;=====================================================
;;; DEJA DEFINI DANS OMLOOP
#|
(defmethod call-gen-code ((self file-box-box) numout)
   (let ((in-list (mapcar #'(lambda (thein) (gen-code thein 0)) (inputs self))))
     (if (zerop numout) 
       `(,(first (code (patch self))) ,.in-list)
       `(nth ,numout (multiple-value-list (,(first (code (patch self))) ,.in-list))))))

(defmethod gen-code-call ((self file-box-box))
   (let ((in-list (mapcar #'(lambda (thein) (gen-code thein 0)) (inputs self))))
     `(,(first (code (patch self))) ,.in-list)))

(defmethod special-lambda-value ((self file-box-box) symbol)
   "Eval a loop box in lambda mode."
   (let* ((nesymbs nil)
          (args  (mapcar #'(lambda (input)
                             (if (connected? input)
                               `',(omNG-box-value  input)
                               (let ((newsymbol (gensym)))
                                 (push newsymbol nesymbs)
                                 newsymbol))) (inputs self))))
     (if (null nesymbs)
       symbol
       (eval `#'(lambda ,(reverse nesymbs)
                  (apply (fdefinition ',(first (code (patch self)))) (list ,.args)))))))

(defmethod curry-lambda-code ((self file-box-box) symbol)
   "Lisp code generation for a loop box in lambda mode."
  (let* ((nesymbs nil)
         (args  (mapcar #'(lambda (input)
                            (if (connected? input)
                              (gen-code input 0)
                              (let ((newsymbol (gensym)))
                                (push newsymbol nesymbs)
                                newsymbol))) (inputs self))))
    (if (null nesymbs)
      symbol
      `#'(lambda ,(reverse nesymbs)
                 (apply (fdefinition ',(first (code (patch self)))) (list ,.args))))))

(defmethod special-value ((self  file-box-box) &optional (args nil))
  (unless (compiled? (patch self)) 
    (compile-patch (patch self))
    (setf (compiled? (patch self)) t))
  (apply (fdefinition (intern (string (first (code (patch self)))) :om)) args))
|#
;;;
;;;=====================================================

(defmethod get-boxcallclass-fun ((self (eql 'file-box))) 'file-box-box)


(defmethod! file-box  (&rest oplist) 
   :numouts 1 
   :initvals '(nil) :indoc '("aditional input")
   :doc "General OM loop.
Add as many optional inputs as needed. Then dbl-click on the file-box icon.
Program your loop in the loop editor." 
   :icon 649 


   (declare (ignore oplist)) nil)

;=================================================
; STREAM BOX




(defclass Streambox (OMBoxcall) 
  ((filetype :accessor filetype :initarg :filetype :initform 'text)
   (direction :accessor direction :initarg :direction :initform :io)
   (if-ex :accessor if-ex :initarg :if-ex :initform :rename))
   (:documentation "This is the class for the method exception box OMOR.#enddoc#
#seealso# (OMBoxcall) #seealso#"))


(defmethod omNG-save ((self Streambox) &optional (values? nil))
  "Save a box"
  (let* ((inputs (mapcar #'(lambda (input) (omNG-save input values?)) (inputs self)))
         (value (saveValueinBox (value self))))
    `(let ((box (om-load-boxcall ',(saveBox? self) ,(name self) ',(save-reference self) ',inputs ,(om-save-point (frame-position self)) 
                      ,(om-save-point (frame-size self)) ,value ,(allow-lock self) ,(frame-name self) ,(numouts self))))
       (setf (filetype box) ',(filetype self)
             (direction box) ,(direction self)
             (if-ex box) ,(if-ex self))
       box)
    ))

(defmethod omNG-copy ((self Streambox))
  `(let* ((copy ,(omNG-make-new-boxcall (fdefinition 'StreamFile) (frame-position self) (name self))))
     (setf copy (update-boxes ,self copy))
     (setf (filetype copy) ',(filetype self))
     (setf (direction copy) ,(direction self))
     (setf (if-ex copy) ,(if-ex self))
     copy))


(defmethod no-allow-copy-p ((self Streambox))
   "Finally boxes can not be directly copied" t)




(defmethod get-boxcallclass-fun ((self (eql 'StreamFile))) 'Streambox)

(defmethod OpenEditorframe ((self Streambox))
  (or (editorframe self)
      (open-stream-dialog self)))

(defmethod! StreamFile  ((pathname t)) 
   :numouts 1 
   :initvals '(nil) :indoc '("pathname or string")
   :doc "file stream" 
   :icon 649
   t)


(defmethod omNG-box-value ((self Streambox) &optional (numout 0))
   (cond
    ((equal (allow-lock self) "l") (special-lambda-value self 'omor))
    ((and (equal (allow-lock self) "x") (value self)) (nth numout (value self)))
    ((and (equal (allow-lock self) "&") (ev-once-p self)) (nth numout (value self)))
    (t (let ((rep (omNG-box-value (first (inputs self)))))
         (loop while (not rep)
               for item in (cdr (inputs self)) do
               (setf rep (omNG-box-value item))) 
         (when (equal (allow-lock self) "&")
           (setf (ev-once-p self) t)
           (setf (value self) (list rep)))
         (when (equal (allow-lock self) "x")
           (setf (value self) (list rep)))
         rep))))

(defmethod get-stream-type ((self Streambox)) (filetype self))

(defmethod call-gen-code ((self Streambox) numout)
   (declare (ignore numout))
   'stream)

(defmethod gen-code-call ((self Streambox))
   (call-gen-code self 0))
  
(defmethod special-value ((self  Streambox) &optional (args nil))
   (declare (ignore numout))
   (om-beep-msg "Close this editor in order to read or write a file."))

(defmethod gen-stream-options ((self  Streambox))
   `(stream (let ((pathname ,.(decode self)))
              (or pathname (choose-file-dialog))) 
            :direction :io  
            :if-does-not-exist :create 
            :if-exists :overwrite
            :element-type 'character
            :external-format :TEXT))


;Editor
(omg-defclass streamwindow (editorwindow) ())


(defmethod get-editor-class ((self streamwindow)) 'streamEditor)



;;;(defmethod view-activate-event-handler :after ((self streamwindow))
;;;   (update-themenubar self))

(defmethod om-view-click-handler :before ((self streamwindow) where)
  (when (text-view (editor self))  ;;; and  (not (equal (find-clicked-subview self where) (text-view (editor self))))) 
    (exit-from-dialog (text-view (editor self)) (dialog-item-text (text-view (editor self))))))

(omg-defclass StreamEditor (editorView) ())

(defmethod handle-key-event ((self StreamEditor) char)  nil)


;(defmethod om-window-close-event  :after ((self streamwindow))
;   (setf (editorFrame (object (editor self))) nil))
(defmethod close-editorFrame ((self streamEditor))
  (setf (editorFrame (object self)) nil))


(defun open-stream-dialog (streambox)

   (let ((dialog (make-editor-window 'streameditor streambox (name streambox) nil
                                     :winpos :centered
                                     :winsize (om-make-point 300 150)
                                     :resize nil
                                     ))
         ifexists ifextext)
     (om-add-subviews (editor dialog)
                      (om-make-dialog-item 'om-radio-button (om-make-point 13 11) (om-make-point 95 15)
                                                  "Read/write"
                                                  :di-action (om-dialog-item-act item (setf (direction streambox) :io))
                                                  :checked-p (equal :io (direction streambox))
                                                  )
                      (om-make-dialog-item 'om-radio-button
                                           (om-make-point 115 11)  (om-make-point 70 16) "Write"
                                           :di-action (om-dialog-item-act item (setf (direction streambox) :output))
                                           :checked-p (equal :output (direction streambox))
                                           )
                      (om-make-dialog-item 'om-radio-button
                                           (om-make-point 199 11) (om-make-point 70 16) "Read"
                                           :di-action (om-dialog-item-act item (setf (direction streambox) :input))
                                           :checked-p (equal :input (direction streambox))
                                           )

                      (om-make-dialog-item 'om-static-text (om-make-point 10 50)
                                           (om-make-point 100 20)
                                           "File Type")
                      (om-make-dialog-item 'om-pop-up-dialog-item
                                  (om-make-point 100 50)
                                  (om-make-point 150 20)
                                  ""
                                  :range (list "TEXT" "SDIF")
                                  :value (string (filetype streambox)) 
                                  :di-action (om-dialog-item-act item 
                                               (setf (filetype streambox) (intern (om-get-selected-item item)))
                                               (om-enable-dialog-item ifexists (not (equal "SDIF" (om-get-selected-item item))))
                                               (om-set-part-color ifextext :text (if (equal "SDIF" (om-get-selected-item item)) *om-gray-color* *om-black-color*)))
                                                 
                                  )
                      
                      ;(om-make-dialog-item 'om-pop-up-dialog-item
                      ;                     (om-make-point 10 90)
                      ;                     (om-make-point 275 20)
                      ;                     "Element type "
                      ;                     :range (list "Character" "Signed-byte" "Bit" "Unsigned-byte")
                      ;                     )
                      

                      (setf ifextext (om-make-dialog-item 'om-static-text (om-make-point 10 80)
                                           (om-make-point 100 20)
                                           "If File Exists"
                                           :fg-color (if (equal 'sdif (filetype streambox)) *om-gray-color* *om-black-color*)))
                      
                      (setf ifexists (om-make-dialog-item 'om-pop-up-dialog-item
                                           (om-make-point 100 80)
                                           (om-make-point 150 20)
                                           ""
                                           :range (list "Rename" "Supersede" "Overwrite" "Append")
                                           :value (string (if-ex streambox))
                                           :di-action (om-dialog-item-act item (setf (if-ex streambox) 
                                                                                     (nth (om-get-selected-item-index item) '(:rename :supersede :overwrite :append)))))
                            )
                      )

     (om-enable-dialog-item ifexists (not (equal 'sdif (filetype streambox))))
     
     (editor dialog))
   )





;;; WRITE BOX

(defmethod! file-write-line ((line t) stream)
  :icon 908
  (om-print "write only numbers and strings"))

(defmethod! file-write-line ((line string) stream)
  :icon 908
  (write-line line stream))

(defmethod! file-write-line ((line symbol) stream)
  :icon 908
  (write-line (string line) stream))

(defmethod! file-write-line ((line number) stream)
  :icon 908
  (write-line (format nil "~D" line) stream))

(defmethod! file-write-line ((line character) stream)
  :icon 908
  (write-line (string line) stream))

;;============

(defmethod! file-write ((line t) stream)
  :icon 908
  (om-print "write only numbers symbols and strings"))

(defmethod! file-write ((line string) stream)
  :icon 908
  (write-string line stream))

(defmethod! file-write ((line symbol) stream)
  :icon 908
  (write-string (string line) stream))

(defmethod! file-write ((line number) stream)
  :icon 908
  (write-string (format nil "~D" line) stream))

(defmethod! file-write ((line character) stream)
  :icon 908
  (write-char line stream))
     


(addgenfun2pack 'file-box *data-package*)

;;; faire un sous package file writer
;(addgenfun2pack 'file-write *data-package*)
;(addgenfun2pack 'file-write-line *data-package*)

