;openmusic
;
;copyright (c) 1997, 1998, 1999, 2000 by ircam-centre georges pompidou, paris, france.
; 
;this program is free software; you can redistribute it and/or
;modify it under the terms of the gnu general public license
;as published by the Free Software Foundation; either version 2
;of the license, or (at your option) any later version.
;
;see file license for further informations on licensing terms.
;
;this program is distributed in the hope that it will be useful,
;but WITHOUT ANY WARRANTY; without even the implied warranty of
;merchantability or fitness for a particular purpose.  see the
;GNU General Public License for more details.
;
;You should have received a copy of the GNU General Public License
;along with this program; if not, write to the Free Software
;foundation, inc., 59 temple place - suite 330, boston, ma  02111-1307, usa.
;
;authors: gerard assayag and augusto agon


(in-package :om)


;===========================================================
;editor
;===========================================================
;==================================================
(omg-defclass comp-text-enter-view (edit-text-enter-view) ())

(defmethod exit-from-dialog ((self comp-text-enter-view) newtext)
   (let* ((container (editor (om-view-container self)))
           (comp-num (if (equal newtext "") 
                                      (select-comp container) 
                                    (read-from-string newtext)))
           )
     (when (and (integerp comp-num) (>= comp-num 0) (< comp-num (numcols (object container))))
         (setf (select-comp container) comp-num)
         (show-composant (panel container))
         (om-invalidate-view (control-view container) t))
     (setf (text-view container) nil)
     (om-remove-subviews container self)))

;=============The palette===========

(omg-defclass array-controls (om-view)  
   ((groupname :initform nil :accessor groupname)))

(defmethod om-draw-contents :before ((self array-controls))
   (om-with-focused-view self
     (om-draw-line 0 25 (w self) 25)
     (om-draw-string 150 8 "Composant")
     (om-draw-string 150 20 (format () "~D" (select-comp (om-view-container self))))))

(defmethod show-composant ((self array-controls))
   (om-with-focused-view self
     (om-with-fg-color self *om-white-color*
       (om-fill-rect 150 10 40 13))
     (om-draw-string 150 20 (format () "~D" (select-comp (om-view-container self))))))

(defmethod initialize-instance :after  ((self array-controls) &key show-cont (grupo "All open"))
  (om-add-subviews self
    (om-make-dialog-item 'om-check-box (om-make-point 25 0) (om-make-point 60 12) "lines"
                                              :font *om-default-font2*
                                              :checked-p t
                                              :di-action
                                              (om-dialog-item-act item
                                                                                   (change-lines  (om-view-container self) (om-checked-p item)))))
   (add-some-controls (om-view-container self) self show-cont grupo))

(defmethod om-view-click-handler ((self array-controls) where)
   (if (and (= (mode (panel (om-view-container self))) 6)
              (>= (om-point-h where) 50) (<= (om-point-h where) 180))
       (setf (text-view (editor (om-view-container self)))
          (om-make-dialog-item 'comp-text-enter-view (om-make-point 150 12) (om-get-text-edit-size (om-make-point 30 10))
                                                    (format () "~D" (select-comp (om-view-container self)))
                                                    :allow-returns t
                                                    :object (om-view-container self)
                                                    :container (om-view-container self)
                                                    :di-selected-p t
                                                    :font *om-default-font2*))
       (call-next-method)))


(omg-defclass array-palette (palette-win)  ()
   (:default-initargs :source-name "Array-palette"))

(defmethod palette-act ((self array-palette) x)
  (let* ((panel (panel (editor-assoc self)))
         (views (bpf-views panel)))
    (case x
      (2 (loop for item in views do
               (setf (grille-p item) (not (grille-p item)))
               (draw-grille item)))
      (3 (setf (rangex panel)
               (get-array-ranges (object (om-view-container panel))) ))
      (4 (om-print-window (window (editor-assoc self))))
      (5 (if (selected-index panel)
           (let ((new-color (om-choose-color-dialog)))
             (when new-color
               (loop for i in (selected-index panel) do 
                     (fill-bpf-with-color (nth i views) new-color)
                     (setf (nth i (color-list (edition-params (ref (editor-assoc self))))) new-color)
                     (om-invalidate-view (nth i views) t))))
           (om-beep-msg "Select the views to colorize." )))
      
      (otherwise
       (when x
         (setf (mode panel) x)
         (om-invalidate-view self t))))))

;=======================EDITOR====================
;=================================================

(defvar *size-h-min* 40)
(defvar *h-names-size* 15)

(omg-defclass arrayeditor (editorview) 
   ((select-comp :initform nil :accessor select-comp)
    (deltay :initform *size-h-min* :accessor deltay)
    (groups :initform nil :accessor groups)
    (cur-group-ind :initform 0 :accessor cur-group-ind)
    (control-view :initform nil :accessor control-view)
    (show-controls-p :initform nil :accessor show-controls-p)))

(defmethod editor-null-event-handler :after ((self arrayeditor))
  (do-editor-null-event self))

(defmethod metaobj-scrollbars-params ((self arrayeditor))  '(:v nil))

(defmethod editor-has-palette-p ((self arrayeditor)) 'array-palette)

(defmethod do-editor-null-event ((self arrayeditor))
   (when (and (om-view-contains-point-p (panel self) (om-mouse-position self)) (>= (om-point-h (om-mouse-position self)) 25))
     (setf (select-comp self) (om-point-h (pixel2point (panel self) (om-mouse-position self))))
     (setf (select-comp self) (min (max 0 (select-comp self)) (- (numcols (object self)) 1)))
     (unless (zerop (select-comp self))
       (show-composant (control-view self))
       (when (= (mode (panel self)) 6)
         (show-composant (panel self))))))

(defmethod om-view-click-handler ((self arrayeditor) where)
  (cond 
   ((text-view self)
    (exit-from-dialog (text-view self) (om-dialog-item-text (text-view self))))
   (t (call-next-method))))

(defmethod create-a-group ((self arrayeditor) name cdr)
   (cons name cdr))
   
(defmethod change-group ((self arrayeditor) val)
   (cond
    ((equal val -1)
     (let ((groupname (unique-name-from-list "group" (loop for item in (groups self) collect (car item)))))
       (setf (groups self) (cons  (create-a-group self groupname (cdr (panel-list (panel self))))
                                  (groups self)))
       (update-edition-params self 'panel-list (groups self))
       (setf (cur-group-ind self) 0)
       (update-edition-params self 'cur-group-ind 0)
       (setf (panel-list (panel self)) (copy-list (nth 0 (groups self))))
       (update-ins-panel (panel self) :removeall t)))
    ((equal val -2)
     (let ((groupname (car (nth (cur-group-ind self) (groups self)))))
       (if (or (string-equal "all open" groupname) (string-equal "all close" groupname))
         (om-beep-msg (format nil "Default group ~D can not be deleted." groupname))
         (progn
           (setf (groups self) (remove (nth (cur-group-ind self) (groups self)) (groups self) :test 'equal))
           (update-edition-params self 'panel-list (groups self))
           (setf (cur-group-ind self) 0)
           (update-edition-params self 'cur-group-ind 0)
           (setf (panel-list (panel self)) (copy-list (nth 0 (groups self))))
           (update-ins-panel (panel self) :removeall t)))))
    (t (let ((index (position val (groups self) :test 'equal)))
         (when index
           (setf (cur-group-ind self) index)
           (setf (panel-list (panel self)) (copy-list (nth (cur-group-ind self) (groups self))))
           (update-edition-params self 'cur-group-ind index)
           (update-ins-panel (panel self) :removeall t))))))

(defmethod mode ((self arrayeditor))
   (mode (panel self)))

(defmethod change-show-controls ((self arrayeditor))
   (setf (show-controls-p self) (not (show-controls-p self)))
   (update-ins-panel  (panel self) :removeall t)
   )

(defmethod change-lines ((self arrayeditor) val)
   (loop for item in (bpf-views (panel self)) do
         (setf (lines-p item) val)
         (om-invalidate-view item t)))


(defmethod load-edition-params ((self arrayeditor))
  (when (ref self)
    (setf (groups self) (copy-list (panel-list (edition-params (ref self)))))
    (setf (cur-group-ind self) (cur-group-ind (edition-params (ref self)))))
  )


(defmethod update-edition-params ((self arrayeditor) slot val)
  (when (ref self)
    (setf (slot-value (edition-params (ref self)) slot) val)))

(defmethod get-control-list ((self arrayeditor)) nil)

(defmethod add-some-controls ((self arrayeditor) view show-cont grupo)
  (let* ((groupitem (om-make-dialog-item 'om-static-text (om-make-point 300 7) (om-make-point 150 12) grupo
                                                                           :font *om-default-font2*))
         (groupbut (om-make-view 'button-icon
                                                          :iconid 149
                                                          :position (om-make-point 275 7)
                                                          :size  (om-make-point 20 12)
                                                          :action #'(lambda (item)
                                                                             ;(let ((pop-up 
                                                                             (cons-pair-pop-menu groupitem 'change-group self
                                                                                                                      (list+ '(("add-group" -1) ("delete-group" -2)
                                                                                                                                  ("-" -3))
                                                                                                                                (loop for item in (groups self)
                                                                                                                                       collect (list (car item) item)))
                                                                                                                      (+ 1 (cur-group-ind self))
                                                                                                                      (om-view-position item) (om-view-size item))
                                                                             
                                                                             ;(add-subviews self pop-up)
                                                                             ;(om-view-click-handler pop-up (om-make-point 5 5))
                                                                             ;(remove-subviews self pop-up)
                                                                             (om-set-dialog-item-text (groupname view) (car (nth (cur-group-ind self) (groups self))))
                                                                             ))))
    (setf (groupname view) groupitem)
    
    (om-add-subviews view
                                 (om-make-dialog-item 'om-check-box (om-make-point 25 13) (om-make-point 95 11) "Show Controls"
                                                                           :di-action (om-dialog-item-act item (declare (ignore item)) 
                                                                                                                                   (change-show-controls self))
                                                                           :font *om-default-font2*
                                                                           ) groupbut groupitem)))



(defmethod editor-array-panel-class ((self arrayeditor)) 'arraypanel)

(defmethod edition-params ((self arrayeditor)) 
  (default-edition-params (object self)))

(defmethod initialize-instance :after ((self arrayeditor) &rest l)
  (declare (ignore l))
  (load-edition-params self)
  (let* ((ed-view (om-make-view (editor-array-panel-class self)
                    :owner self
                    :position (om-make-point 0 25) 
                    :scrollbars (first (metaobj-scrollbars-params self))
                    :size (om-make-point (w self) (- (h self) 50))
                    :panel-list (copy-list (nth (cur-group-ind self) (groups self)))
                    :c-panel-list (get-control-list self)))
         (rulerx (om-make-view 'static-ruler 
                   :owner self
                   :axe 'x
                   :zoom 1
                   :minzoom 1
                   :position (om-make-point 25 (- (h self) 25)) 
                  :size (om-make-point (- (w self) 25) 25)))
         (control (om-make-view 'array-controls 
                    :owner self
                    :position (om-make-point 0 0) 
                    :size (om-make-point (w self) 25)
                    :grupo (car (nth (cur-group-ind self) (groups self))))))
    (setf (panel self) ed-view)
    (setf (rulerx (panel self)) rulerx)
    (setf (control-view self) control)
    (setf (rangex (panel self)) (get-array-ranges (object self)))
    (update-ins-panel ed-view :removeall t)
    ))

(defmethod update-subviews ((self arrayeditor))
   ;;;(om-set-view-size (panel self) (om-make-point (- (w self)  15) (- (h self) 50)))
   (om-set-view-size (panel self) (om-make-point (w self) (- (h self) 50)))
   (om-set-view-position (panel self) (om-make-point 0 25))
   (om-set-view-size (control-view self) (om-make-point (w self) 25))
   (change-delta-y self)
   (update-subviews (panel self))
   (om-invalidate-view self t))

(defmethod change-delta-y ((self arrayeditor))
   (let* ((panel (panel self))
          (sizebynames (* (length (name-views panel)) *h-names-size*))
          (resto (- (h panel) sizebynames))
          (bpfnum (length (bpf-views (panel self)))))
     (when (> bpfnum 0)
       (setf (deltay self) (max (round resto bpfnum) *size-h-min*)))))


;------------------------------------


(omg-defclass arraypanel (om-scroller view-with-ruler-x) 
  ((mode :initform 0 :accessor mode)
   (panel-list :initform nil :initarg :panel-list :accessor panel-list)
   (c-panel-list :initform nil :initarg :c-panel-list :accessor c-panel-list)
   (selected-index :initform nil :accessor selected-index)
   (name-views :initform nil :accessor name-views)
   (bpf-views :initform nil :accessor bpf-views))
  (:default-initargs :scrollbars :v :retain-scrollbars t))


(defmethod multibpf? ((self arraypanel)) nil)
(defmethod show-position ((self t)) nil)
(defmethod assoc-w ((self arraypanel)) (- (w self) 25))

(defmethod bpf-panel-class ((self arraypanel)) 'bpf-parameter-panel)
(defmethod list-panel-class ((self arraypanel)) 'list-parameter-panel)



(defmethod mode ((self arraypanel))
   (if (= 8 (slot-value self 'mode)) 9 (slot-value self 'mode)))

(defmethod fill-bpf-with-color ((self arraypanel) color)
   (loop for index in (selected-index self) do
         (let ((view (get-subview-by-index self 'bpf index)))
           (when view
           (setf (bpfcolor (currentbpf view)) color)
           (om-invalidate-view view t)))))


(defmethod show-composant ((self arraypanel))
   (loop for item in (name-views self) do
         (show-composant item)))

(defmethod update-ins-panel ((self arraypanel) &key removeall updateref)
  (when removeall
    (change-delta-y (om-view-container self))
    (loop for item in (om-subviews self) do
          (om-remove-subviews self item))
    (let* ((group (cdr (panel-list self)))
           (deltay (deltay (om-view-container self)))
           (cur-y 0)
           (rulerx (rulerx self))
           (thearray (object (om-view-container self)))
           (preci (get-precision thearray)))
      (setf (name-views self) nil)
      (setf (bpf-views self) nil)
      (setf (c-panel-list self) nil)
      (loop for pair-i in group collect
            (let ((i (car pair-i))
                  (open? (second pair-i)))
              (push (om-make-view 'bar-name 
                      :owner self
                      :index i
                      :position (om-make-point 0 cur-y)
                      :size (om-make-point (w self) *h-names-size*)
                      :open? open?) (name-views self))
              (incf cur-y *h-names-size*)
              (when open?
                (let ((newpanel (new-editor-from-val 
                                 (get-bfp-all-points (get-array-row thearray i) thearray preci (row-type thearray i))
                                 self i nil)))
                  (when newpanel
                    (om-set-view-position  newpanel (om-make-point 25 cur-y))
                    (om-set-view-size newpanel (om-make-point (- (w self) 25) deltay))
                    (incf cur-y deltay))))))
      (when (show-controls-p (om-view-container self))
        (let* ((controls (lcontrols thearray))
               (startind (numrows (object (om-view-container self)))))
          (loop for item in controls 
                 for i = startind then (+ i 1) do
                (let* ((newbpf (get-bfp-all-points (get-array-row thearray i) thearray preci (row-type thearray i)))
                       newpanel)
                  (push (om-make-view 'bar-name 
                          :owner self
                          :control-p t
                          :index i
                          :position (om-make-point 0 cur-y)
                          :size (om-make-point (w self) *h-names-size*)
                          :open? t) (name-views self))
                  (incf cur-y *h-names-size*)
                  (setf newpanel (new-editor-from-val newbpf  self i t))
                  (when newpanel 
                    (om-set-view-position newpanel (om-make-point 25 cur-y))
                    (om-set-view-size newpanel (om-make-point (- (w self) 25) deltay)))
                  (pushr (list i t) (c-panel-list self))
                  (incf cur-y deltay)))
          (om-set-field-size self (om-make-point (w self) cur-y))
          ))
      (setf (name-views self) (reverse (name-views self)))
      (setf (bpf-views self) (sort (bpf-views self) '< :key 'index))
      (setf (assoc-view rulerx) self)))
  (update-subviews (om-view-container self))
  )

(defmethod update-subviews ((self arraypanel))
   (let* ((group (append (cdr (panel-list self)) (c-panel-list self)))
          (deltay (deltay (om-view-container self)))  (cur-y 0))
     (om-set-view-size (rulerx self) (om-make-point (- (w (om-view-container (rulerx self))) 40) 25))
     (om-set-view-position (rulerx self) (om-make-point 25 (- (h (om-view-container (rulerx self))) 25)))
     (loop for pair-i in group do
           (let ((open? (second pair-i)))
             (om-set-view-position (get-subview-by-index self 'names (car pair-i))  (om-make-point 0 cur-y))
             (om-set-view-size (get-subview-by-index self 'names (car pair-i))  (om-make-point (w self) *h-names-size*))
             (incf cur-y *h-names-size*)
             (when open?
               (let* ((subview (get-subview-by-index self 'bpf (car pair-i))))
                 (update-panel-array subview self cur-y deltay)
                 (incf cur-y deltay)))))))

(defmethod get-subview-by-index ((self arraypanel) type index)
   (let ((list (if (equal type 'names) (name-views self) (bpf-views self))))
     (find-if  #'(lambda (x) (equal (index x) index)) list)))


(defmethod (setf rangex) ((ranges list) (self arraypanel))
   (setf (slot-value self 'rangex) ranges)
   (loop for item in (bpf-views self) do
         (init-coor-system item))
   (om-invalidate-view (rulerx self) t))

(defmethod handle-key-event ((self arraypanel) char)
   (loop for item in (bpf-views self) do
         (when (selected-p item)
           (handle-key-event item char))))

(defmethod open-internal-edit ((self arraypanel) where index)
  (let* ((select-comp (om-point-h (pixel2point self (om-mouse-position self))))
         (array (object (om-view-container self)))
         (valor (valor-from-index  array index)) rep)
    (unless (exact-list-p valor)
      (set-exact-list array index 
                      (get-points (valor-from-index array index) (row-type array index) array)))
    
    (setf rep (nth select-comp (thelist (valor-from-index array index)))) 
    (if (Class-has-editor-p rep)
      (push (make-editor-window (get-editor-class rep) rep (format nil "temp editor composant ~D" select-comp)
                                                           (om-view-container self)) 
                  (attached-editors (om-view-container self)))
      (om-beep))))


(defmethod special-set-ranges ((self arraypanel) new-pi-panel decimals rangebpf)
  (set-ranges new-pi-panel (om*  (rangex self) 
                                 (expt 10 decimals)) 
              (list (third rangebpf) (fourth rangebpf))))



;------------------------------------

(defvar *change-name-view* nil)


;---------------------------------------------------------------------
(omg-defclass bar-name (om-view) 
   ((index :initform 0 :initarg :index :accessor index)
    (control-p :initform nil :initarg :control-p :accessor control-p)
    (at-editor :initform nil  :accessor at-editor)))

(defmethod get-panel ((self bar-name))
   (om-view-container (om-view-container self)))

(defmethod initialize-instance :after  ((self bar-name) &key open?)
   (om-add-subviews self  
     (om-make-view 'button-icon
       :iconid (if open? 165 164)
       :action  #'(lambda (item) 
                    (setf (iconid item) (if (= (iconid item) 164) 165 164))
                    (show/hide-editor (om-view-container item)))
       :position (om-make-point 3 3)
       :size (om-make-point 11 11))))

(defmethod selected-p ((self bar-name)) (member (index self) (selected-index (om-view-container self))))

(defmethod get-fill-color ((self bar-name))
   (cond
    ((control-p self) *azulote*)
    ((selected-p self) *om-gray-color*)
    (t  *azulito*)))


(defmethod om-view-click-handler ((self bar-name) where) 
   (declare (ignore where))
   (unless (control-p self)
     (when (equal (call-next-method) self)
       (if (om-shift-key-p)
         (if (member (index self) (selected-index (om-view-container self)))
           (setf (selected-index (om-view-container self)) (remove  (index self) (selected-index (om-view-container self))))
           (push (index self) (selected-index (om-view-container self))))
         (setf (selected-index (om-view-container self))  (list (index self))))
         (update-bar-names self))))



(defmethod update-bar-names ((self bar-name))
   (loop for item in (name-views (om-view-container self)) do
         (om-invalidate-view item t)))

(defmethod get-name-view ((self bar-name))
  (index2label (object (get-panel self)) (index self)))
             
(defmethod om-draw-contents ((self bar-name))
  (let* ((panel (om-view-container self))
         (y0  (om-point-v (om-scroll-position panel))))
    (when (and (> (y+h  self) y0) (< (y self) (+ y0 (h panel)))) 
        (om-with-focused-view self
                                                    (om-with-font (om-get-font self)
        (om-with-fg-color self (get-fill-color self)
          (om-fill-rect 0 0 (w self) (h self)))
        (om-draw-string 20 10 (get-name-view self))
        (om-draw-line 0 (h self) (w self) (h self)))
                                                    (call-next-method)))))
  

(defmethod show-composant ((self bar-name))
  (let* ((array (object (get-panel self)))
         (comp (select-comp (get-panel self)))
         rep)
    (when comp
      (setf rep (get-array-i-j array  (index self) comp))
      (om-with-focused-view self
       (om-erase-rect-content 150 2 100 11)
        (om-draw-string 150 10 (cond
                             ((integerp rep) (format nil "~D" rep))
                             ((numberp rep) (format nil "~,6F" rep))
                             (t "*")))))))

(defmethod show/hide-editor ((self bar-name))
  (let* ((panel (om-view-container self))
         (editor (om-view-container panel))
         (thearray (object (get-panel self)))
         (preci (get-precision thearray))
         (oldlist (cons (car (panel-list panel)) 
                        (loop for item in (cdr (panel-list panel)) collect (copy-list item))))
         (old-c-list (loop for item in (c-panel-list panel) collect (copy-list item)))
         view-in-question)
    (if (< (index self) (length (cdr oldlist)))
      (setf open? (setf (nth 1 (nth (index self) (cdr oldlist)))
                        (not (nth 1 (nth (index self) (cdr oldlist))))))
      (setf open? (setf (nth 1 (nth (- (index self) (length (cdr oldlist))) old-c-list))
                        (not (nth 1 (nth (- (index self) (length (cdr oldlist))) old-c-list))))))
    (setf (panel-list panel) oldlist)
    (setf (c-panel-list panel) old-c-list)
    (if open?
      (without-interrupts  
       (new-editor-from-val  (get-bfp-all-points (get-array-row thearray (index self)) thearray 
                                                 preci (row-type thearray (index self)))
                             
                             panel (index self) nil)
       (setf (bpf-views panel) (sort (bpf-views panel) '< :key 'index)))
      (progn
        (setf view-in-question (find-if #'(lambda (x) (equal (index x) (index self))) (bpf-views panel)))
        (remove-extra-editor view-in-question)
        (setf (bpf-views panel) (remove view-in-question (bpf-views panel) :test 'equal)) 
        (om-remove-subviews panel view-in-question)))
    (om-set-view-size editor (om-view-size editor))
    ;(update-subviews panel)
    ))

(defmethod new-editor-from-val ((val bpf) panel index control-p)
  (let* ((newbpf val)
         (rangebpf (give-bpf-range newbpf))
         (new-pi-panel (om-make-view (bpf-panel-class panel)
                         :owner panel
                         :index index
                         :control-p control-p
                         :position (om-make-point 0 0)
                         :size (om-make-point 0 0))))
    (push new-pi-panel (bpf-views panel))
    (setf (rulerx new-pi-panel) 
          (om-make-view 'ruler 
            :assoc-view  new-pi-panel
            :hide-ruler? t
            :axe 'x
            :position (om-make-point 0 0)
            :size (om-make-point 0 0)))
    (setf (currentbpf new-pi-panel) newbpf)
    (setf (rulery new-pi-panel) 
          (om-make-view 'ruler 
            :owner panel
            :assoc-view  new-pi-panel
            :axe 'y
            :position (om-make-point 0 0)
            :size (om-make-point 0 0)))
    (special-set-ranges panel new-pi-panel (decimals newbpf) rangebpf)
    (setf (bpf-views panel) (sort (bpf-views panel) '< :key 'index))
    new-pi-panel))



(defmethod new-editor-from-val ((val null) panel index control-p)
   (let* ((new-panel (om-make-view (list-panel-class panel) 
                       :owner panel
                       :index index
                       :control-p control-p
                       :position (om-make-point 0 0)
                       :size (om-make-point 0 0))))
     (push new-panel (bpf-views panel))
     (setf (rulerx new-panel) (rulerx panel))
     (setf (rangex new-panel) (rangex panel))
     ;(pushr new-panel (assoc-view (rulerx panel)))
     new-panel))

;------------------------------------
(omg-defclass list-parameter-panel (om-view view-with-ruler-x) 
   ((index :initform 0 :initarg :index :accessor index)
    (lines-p :initform nil  :accessor lines-p)
    (control-p :initform nil :initarg :control-p :accessor control-p)
    (minipictslist :initform nil :accessor minipictslist)))

(defvar *minx-for-show* 30)
(defvar *miny-for-show* 30)

(defmethod get-panel ((self list-parameter-panel))    
   (om-view-container (om-view-container self)))

(defmethod selected-p ((self list-parameter-panel)) 
  (member (index self) (selected-index (om-view-container self))))

(defmethod om-draw-contents ((self list-parameter-panel))
  (let* ((panel (om-view-container self))
         (y0  (om-point-v (om-scroll-position panel))))
    (when (and (> (y+h  self) y0) (< (y self) (+ y0 (h panel)))) 
      (call-next-method)
      (om-draw-view-outline self)
      (when (grille-p self)
        (draw-grille self))
      (let* ((assoc-view (om-view-container self))
             (x (round (w self) (- (second (rangex assoc-view)) (first (rangex assoc-view))))))
        (if (and (>= x *minx-for-show*) (>= (h self) *miny-for-show*))
          (let* ((array (object (get-panel self)))
                 (delta-x (norme2pixel assoc-view 'x 1))
                 (cur-x 0))
            (loop while (< cur-x (w self))
                  for i = 0 then (+ i 1)
                  for j from 0 to (- (numcols array) 1) do
                  (let ((obj (get-array-i-j-for-eval  array (index self) (+ i (max 0 (car (rangex self))))))
                        (x0 (norme2pixel assoc-view 'x i)))
                    (cond
                     ((class-has-editor-p obj)
                      (setf cur-x (incf cur-x delta-x)) 
                      (draw-obj-in-rect obj x0 (+ x0 delta-x) 0 (h self) (default-edition-params obj) self))
                     (t 
                      (om-with-focused-view self
                                                                  (om-draw-string (+ cur-x 3) (round (h self) 2) (format nil "~D" obj))
                                                                  )
                      (incf cur-x delta-x))))))
          (om-with-focused-view self
             (om-with-fg-color self *azulote*
              (let ((bigfont (om-make-font *om-def-font-face* (round (h self) 1.5))))                      
                 (om-with-font bigfont
                                             (om-draw-string 0 (round (h self) 1.5) "{...")
                                             (om-draw-string (- (w self) (om-string-size "  ...}" bigfont))  (round (h self) 1.5) "...}")
               )
                                             ))))))))

(defmethod om-view-cursor ((Self list-parameter-panel))
   (declare (ignore where))
   (unless (om-command-key-p)
     (case (mode self)
       (1 *om-loupe-cursor*)
       (otherwise *om-arrow-cursor*))))

(defmethod om-view-click-handler ((self list-parameter-panel) where)
   self)


;;; XXX
(defmethod om-view-doubleclick-handler ((self list-parameter-panel) where)
   (open-internal-edit (om-view-container self) where (index self))
   self)

(defmethod remove-extra-editor ((self list-parameter-panel)) nil)

(defmethod init-coor-system ((self list-parameter-panel))
   (setf (rangex self) (rangex (om-view-container self)))
   (om-invalidate-view self t))


(defmethod update-panel-array ((subview list-parameter-panel) self cur-y deltay)
   (om-set-view-position subview  (om-make-point 25 cur-y))
   (om-set-view-size subview  (om-make-point (- (w self) 25) deltay)))

(defmethod fill-bpf-with-color ((self list-parameter-panel) color) t)

(defmethod get-mini-param ((self list-parameter-panel) param)
   (cdr (assoc param (default-edition-params (make-instance 'note)))))

(defmethod set-mini-param ((self list-parameter-panel) param val) nil)

(defmethod initx ((self list-parameter-panel)) 0) 

(defmethod inity ((self list-parameter-panel)) 0)
     
(defmethod get-mini-zoom  ((self list-parameter-panel)) 0.8)

(defmethod mode ((self list-parameter-panel))
   (mode (om-view-container self)))


;------------------------------------
(omg-defclass bpf-parameter-panel (bpfpanel) 
  ((index :initform 0 :initarg :index :accessor index)
   (control-p :initform nil :initarg :control-p :accessor control-p)))

(defmethod bpf-decimals ((self bpf-parameter-panel))
  (decimals (currentbpf self)))

(defmethod get-panel ((self bpf-parameter-panel))
  (om-view-container (om-view-container self)))

(defmethod editor ((self bpf-parameter-panel))
  (om-view-container (om-view-container self)))

;-----------draw
(defmethod draw-axis ((self bpf-parameter-panel))  nil)


;-----------Edit

(defmethod handle-key-event ((self bpf-parameter-panel) char)
   (if (or (equal char :om-key-up)
           (equal char :om-key-down)) (call-next-method)))

(defmethod scroll-point ((Self bpf-parameter-panel) where)
  (setf *bpf-last-click* (om-mouse-position self))
  (setf *bpf-first-click* *bpf-last-click*)
  (setf *bpf-offset-click* 0)
  (om-init-motion-functions self 'make-scroll-point 'release-scroll-point))

(defmethod release-scroll-point ((Self bpf-parameter-panel) Where) 
  (change2exact-bpf self)
  (om-invalidate-view self t))

(defmethod make-scroll-point ((Self bpf-parameter-panel) Where)
  (let* ((old-Mouse *bpf-last-click*)
         (first-Mouse *bpf-first-click*)
         (Inity *bpf-offset-click*)
         (new-mouse where)
         (Offy (pixel2norme self 'y (- (om-point-v first-mouse) (om-point-v new-mouse))))
         (moveds (move-points-in-bpf (currentbpf self) (selection? self)  0 (- offy inity ))))
    (if moveds (setf (selection? self) moveds))
    (om-invalidate-view self t)
    (setf *bpf-offset-click*  offy)
    (setq *bpf-last-click* where)))


(defmethod zoom-system ((self Bpf-Parameter-Panel) where)
  (om-init-motion-functions self 'zoom-system-motion 'zoom-system-release)
  (om-new-movable-object self (om-point-h where) 0 4 (h self) 'om-selection-rectangle))

(defmethod zoom-system-motion ((self Bpf-Parameter-Panel) pos)
  (let ((rect  (om-get-rect-movable-object self (om-point-h pos) (om-point-v pos))))
    (when rect
      (om-update-movable-object self (first rect) 0 (max 4 (third rect)) (h self)))))


(defmethod zoom-system-release ((self Bpf-Parameter-Panel) pos)
  (let ((editor (om-view-container self))
        (rect  (om-get-rect-movable-object self (om-point-h pos) (om-point-v pos)))
        user-rect )
    (when rect
      (om-erase-movable-object self)
      (setf user-rect (om-make-rect (first rect) (second rect) (+ (first rect) (third rect)) (+ (second rect) (fourth rect))))
      (let* ((x (om-rect-topleft user-rect))
             (x1 (om-rect-bottomright user-rect))
             (dif (om-subtract-points x1 x)))
        (when (and (> (abs  (om-point-h dif)) 10) (> (abs (om-point-v dif)) 10))
          (let ((new-point (pixel2point editor x))
                (new-point1 (pixel2point editor x1)))
            (when  (>  (om-point-h new-point1) (+  (om-point-h new-point) 1)) 
              (setf (rangex editor) (list (om-point-h new-point) (om-point-h new-point1))))))))))



(defmethod x-class-move-point ((self bpf-parameter-panel))
  'om-static-text)

(defmethod update-panel-array ((subview bpf-parameter-panel) self cur-y deltay)
   (om-set-view-position subview  (om-make-point 25 cur-y))
   (om-set-view-size subview  (om-make-point (- (w self) 25) deltay))
   (om-set-view-position (rulery subview)  (om-make-point 0 cur-y))
   (om-set-view-size (rulery subview)  (om-make-point 25 deltay)))

(defmethod init-coor-system ((self bpf-parameter-panel))
  (let* ((bpf (get-bpf-obj self))
         (rangebpf (give-bpf-range bpf))
         (panel (om-view-container self)))
    (special-set-ranges panel self (decimals bpf) rangebpf)
    (om-invalidate-view self)))


(defmethod set-ranges ((self bpf-parameter-panel) rangex rangey)
   (let* ((inty (- (second rangey) (first rangey)))
          (delta (round inty 5)))
     (setf (rangex self) rangex)
     (setf (rangey self) (list (- (first rangey) delta) (+ (second rangey) delta)))
     (set-units-ruler self (rulery self))
     (set-units-ruler self (rulerx self))))


(defmethod change2exact-bpf ((self bpf-parameter-panel))  
  (let* ((thearray (object (get-panel self)))
         (bpf (currentbpf self))
         (valor (valor-from-index  thearray (index self))))
    (if (exact-bpf-p valor)
      (if (not (equal (thebpf valor) bpf))
        (set-exact-bpf thearray (index self) bpf))
      (set-exact-bpf thearray (index self) bpf)
      )))

(defmethod selected-p ((self bpf-parameter-panel))
   (member (index self) (selected-index (om-view-container self))))

(defmethod get-bpf-obj ((self bpf-parameter-panel))  (currentbpf self) )

(defmethod get-string-nom  ((self bpf-parameter-panel)  num (axe (eql 'x)))  (format () "~D" num))

(defmethod remove-extra-editor ((Self Bpf-Parameter-Panel))
   (om-remove-subviews (om-view-container (rulery self)) (rulery self))
   (setf (rulery self) nil))

(defmethod om-view-click-handler ((Self Bpf-Parameter-Panel) Where)
   (cond 
    ((= 0 (mode self)) (select-system self where))
    ((= 1 (mode self)) (zoom-system self where))
   ))

(defmethod mode ((Self Bpf-Parameter-Panel))
   (mode (om-view-container self)))

(defmethod om-draw-contents ((self Bpf-Parameter-Panel))
  (let* ((panel (om-view-container self))
         (y0  (om-point-v (om-scroll-position panel))))
    (when (and (> (y+h  self) y0) (< (y self) (+ y0 (h panel)))) 
      (call-next-method))))
 

(defmethod draw-grille  ((self Bpf-Parameter-Panel))
  (om-with-focused-view self
    (om-with-dashline
     (let* ((assoc-view (om-view-container self))
            (rulerx (rulerx assoc-view))
            (system-etat (get-system-etat assoc-view))
            i j ppu unit )
       (setf unit (zoom rulerx))
       (setf i (first (rangex assoc-view)))
       (setf i (if (minusp i) (+ i (- unit (mod i unit))) (+ i (mod i unit))))
       (setf j (second (rangex assoc-view)))
       (setf ppu (norme2pixel assoc-view 'x unit)) 
       (loop while (< i j) do
             (let* ((pixel (point2pixel assoc-view (om-make-big-point i 0) system-etat)))
               (cond
                ((and  (> (* 10 ppu) 10) (zerop (mod i (* 10 unit))))
                 (draw-grille-linex assoc-view (om-point-h pixel) 0))
                ((>= ppu  20)
                 (draw-grille-linex assoc-view (om-point-h pixel) 0)))
               (incf i unit)))))))





