(in-package :om)

;=====================================================================
;OBJECTS PARAMS
;=====================================================================
(defclass object-edparams () 
  ((staff :accessor staff :initarg :staff :initform 'gf)
   (top-margin :accessor top-margin :initarg :top-margin :initform 1)
   (fontsize :accessor fontsize :initarg :fontsize :initform 24)
   (tone :accessor  tone :initarg :tone :initform 2)
   ))

(defmethod omNG-save ((self object-edparams) &optional (values? nil))
  `(make-instance 'object-edparams 
     :staff ',(staff self)
     :top-margin ,(top-margin self)
     :fontsize ,(fontsize self)
     :tone ,(tone self)
     ))

(defmethod copy-container ((self object-edparams) &optional pere)
  (make-instance (type-of self)
     :staff (staff self)
     :top-margin (top-margin self)
     :fontsize (fontsize self)
     :tone (tone self)))


;=====================================================================
;OBJECTS
;=====================================================================
(defclass* track-ev ()
  ((obj :initform nil :accessor obj :initarg :obj)
   (start-t :initform 20 :accessor start-t :initarg :start-t)
   (end-t :initform 200 :accessor end-t :initarg :end-t)
   (edparams :initform (make-instance 'object-edparams) :accessor edparams :initarg :edparams)
   (inputs :initarg :inputs :accessor inputs :initform (list (make-instance 'input-funbox)))
   (outs :initarg :outs :accessor outs :initform (list 0))
   (mycontainer :initform nil :accessor mycontainer)
   (frames :accessor frames :initform nil)
   ))

(defmethod track-ev-p ((self track-ev)) t)
(defmethod track-ev-p ((self t)) nil)

(defmethod name ((self track-ev)) "TRACK-EV")

(defmethod get-edit-param ((self track-ev) param)
  (slot-value (edparams self) param))

(defmethod set-edit-param ((self track-ev) param val)
  (setf (slot-value (edparams self) param) val))


(defmethod get-temp-pivots ((self track-ev))
  (append (list (list (start-t self) self))
          (get-obj-pivots (obj self))
          (list (list (start-t self) self))))

(defmethod get-obj-pivots ((self t)) nil)  
                


;=====================================================================
;TRACKS
;=====================================================================
(defclass* track (container named-score-object)
  ((minsize :initform 3 :accessor minsize :initarg :minsize)
   (size :initform 6 :accessor size :initarg :size)
   (objs :initform nil :accessor objs :initarg :objs)))

(defmethod track-p ((self track)) t)
(defmethod track-p ((self t)) nil)

(defmethod* PrepareToPlay ((player t) (self track) at &key  approx port interval voice)
  (loop for item in (objs self) do
        (PrepareToPlay player (obj item) (+ at (start-t item))
                       :approx approx
                       :port port
                       :interval interval
                       :voice voice)))

(defmethod* PrepareToPlay ((player t) (self ommaquette) at &key approx port interval voice)
  (PrepareToPlay player (value self) at
                 :approx approx
                 :port port
                 :interval interval
                 :voice voice))

;=====================================================================
;SHEETPARAMS
;=====================================================================

(defclass sheetparams () 
  ((nbvoices :initform 0 :accessor nbvoices :initarg :nbvoices)
   (font-size :initform 24 :accessor font-size :initarg :font-size)
   (inter :initform 24 :accessor inter :initarg :inter)
   (trackspos :initform nil :accessor trackspos :initarg :trackspos)
   (tstep :initform 1000 :accessor tstep :initarg :tstep))
  )

(defmethod initparams ((self sheetParams) (nbvoices integer))
  (setf (nbvoices self) nbvoices)
  (setf (trackspos self) (make-list nbvoices :initial-element 0)))


(defmethod omNG-save ((self sheetparams) &optional (values? nil))
  `(let ((params (make-instance 'sheetparams 
                   :nbvoices ,(nbvoices self)
                   :inter ,(inter self)
                   :font-size ,(font-size self)
                   :tstep ,(tstep self)
                   )))
     (setf (trackspos params) ',(trackspos self))
     params))

(defmethod copy-container ((self sheetparams) &optional pere)
  (make-instance (type-of self)
     :nbvoices (nbvoices self)
     :inter (inter self)
     :trackspos (trackspos self)
     :font-size (font-size self)
     :tstep (tstep self)
     ))

;=====================================================================
;LA CLASSE SHEET
;=====================================================================

(defclass* sheet (superposition OMPersistantObject) 
  ((voices :initform nil
           :accessor voices :initarg :voices)
   (params :initform (make-instance 'sheetparams :nbvoices 0)
           :accessor params :initarg :params))
  (:icon 127)
  (:documentation "a faire")
  )

(defmethod voices ((self sheet))
  (inside self))

(defmethod initialize-instance ((self sheet) &rest initargs &key (Empty nil))
  (declare (ignore initargs)) 
  (call-next-method)
  (unless (name self) (setf (name self) "Sheet"))
  (unless (icon self) (setf (icon self) 127))
  (setf (slot-value self 'inside) (remove nil (loop for item in (list! (slot-value self 'voices)) append (list! (collect-sheet-voice item)))))
  (setf (slot-value self 'voices) nil)
  (initparams (params self) (length (slot-value self 'inside)))
  (loop for tr in (voices self) do (loop for ev in (objs tr) do (setf (mycontainer ev) self)))
  self)


(defmethod get-obj-dur ((self sheet))
   (loop for track in (voices self) maximize 
         (if (objs track)
           (end-t (last-elem (objs track)))
           0)))

;============================================classes acepted in sheets

(defmethod allowed-in-sheet ((self t)) nil)
(defmethod allowed-in-sheet ((self poly)) t)
(defmethod allowed-in-sheet ((self voice)) t)
(defmethod allowed-in-sheet ((self chord-seq)) t)
(defmethod allowed-in-sheet ((self multi-seq)) t)
(defmethod allowed-in-sheet ((self chord)) t)

(defmethod allowed-in-sheet ((self sheet)) nil)
(defmethod allowed-in-sheet ((self sound)) t)
(defmethod allowed-in-sheet ((self midifile)) t)
(defmethod allowed-in-sheet ((self ommaquette)) t)

(defmethod allowed-in-sheet ((self bpf)) t)

(defmethod allowed-in-track ((self t)) (allowed-in-sheet self))
(defmethod allowed-in-track ((self multi-seq)) nil)
(defmethod allowed-in-track ((self poly)) nil)


;== Create tracks ==

(defmethod collect-sheet-voice ((self poly)) 
  (loop for v in (voices self) collect (collect-sheet-voice v)))

(defmethod collect-sheet-voice ((self multi-seq)) 
  (loop for v in (chord-seqs self) collect (collect-sheet-voice v)))

(defmethod collect-sheet-voice ((self track)) self)

(defmethod collect-sheet-voice ((self sheet)) 
  (loop for v in (voices self) collect (collect-sheet-voice v)))

(defmethod collect-sheet-voice ((self list))
  (let ((objs nil)
        (time 0))
    (loop for elt in self do
          (when (allowed-in-track elt)
            (pushr (make-instance 'track-ev
                     :obj (clone elt)
                     :start-t time
                     :end-t (+ time (get-obj-dur elt))
                     ) objs)
            (setf time (+ time (get-obj-dur elt) 10))))
    (make-instance 'track :objs objs)))

(defmethod collect-sheet-voice ((self t))
  (when (allowed-in-sheet self)
    (make-instance 'track
      :objs (list (make-instance 'track-ev
                    :obj (clone self)
                    :start-t 0
                    :end-t (get-obj-dur self)
                    ))
      )))

(defmethod get-obj-dur ((self bpf)) (last-elem (x-points self)))

;============================================minibox

(defmethod Class-has-editor-p ((self sheet)) t)

(defmethod get-boxsize ((self sheet)) (om-make-point 40 60))

(defmethod get-editor-class ((self sheet)) 'sheetEditor)

(defmethod cons-mini-obj ((self sheet) view sizefont size)
   (cons-multi-mini-obj self view sizefont size))

;============================================picts

(defmethod get-obj-pict ((self t))
  nil)

(defmethod get-obj-pict ((self sound))
  (or (pict-sound self) (om-sound-get-pict self)))

(defmethod get-obj-pict ((self ommaquette))
  (om-record-pict *om-default-font2* (om-make-point 1000 1000) 
    (draw-obj-in-rect self 0 1000 0 1000 nil nil)))


;=======================================
; add track
(defmethod add-one-track ((self sheet) &optional (pos 0)) 
  (setf (inside self) (insert-in-list (inside self) (make-instance 'track) pos))
  (setf (nbvoices (params self)) (+ (nbvoices (params self)) 1))
  (setf (trackspos (params self)) (insert-in-list (trackspos (params self)) 0 pos)))

;=======================================
; remove track
(defmethod remove-one-track ((self sheet) pos)
  (loop for trev in (objs (nth pos (inside self))) do (delete-connections self trev))
  (setf (inside self) (append (subseq (inside self) 0 pos) (subseq (inside self) (+ pos 1))))
  (setf (nbvoices (params self)) (- (nbvoices (params self)) 1))
  (setf (trackspos (params self)) (append (subseq (trackspos (params self)) 0 pos) (subseq (trackspos (params self)) (+ pos 1)))))


;=======================================
; switch tracks
(defun switch-up (list obj)
  (let ((pos (position obj list :test 'equal)))
    (if (<= pos 0)
      list
      (insert-in-list (remove obj list :test 'equal) obj (- pos 1)))))

(defun switch-down (list obj)
  (let ((pos (position obj list :test 'equal)))
    (if (>= pos (- (length list) 1))
      list
      (insert-in-list (remove obj list :test 'equal) obj (+ pos 1)))))

(defun nth-switch-up (list n)
  (if (or (<= n 0) (>= n (length list)))
    list
    (append (subseq list 0 (- n 1)) (list (nth n list) (nth (- n 1) list)) (subseq list (+ n 1)))))

(defun nth-switch-down (list n)
  (if (or (< n 0) (>= n (- (length list) 1)))
    list
    (append (subseq list 0 n) (list (nth (+ n 1) list) (nth n list)) (subseq list (+ n 2)))))

(defmethod switch-track ((self sheet) (track track) direction)
  (let ((pos (position track (inside self) :test 'equal)))
    (if (< direction 0)
      (progn 
        (setf (inside self) (switch-up (inside self) track))
        (setf (trackspos (params self)) (nth-switch-up (trackspos (params self)) pos))
        )
      (progn
        (setf (inside self) (switch-down (inside self) track))
        (setf (trackspos (params self)) (nth-switch-down (trackspos (params self)) pos))
        ))))

;=======================================
; move tracks
(defmethod move-track ((self sheet) (track track) (incr integer))
  (let ((pos (position track (inside self) :test 'equal)))
    (setf (nth pos (trackspos (params self))) (max 0 (+ (nth pos (trackspos (params self))) incr)))))


;=======================================
; resize tracks

(defmethod resize-track ((self track) (incr integer))
  (setf (size self) (max (minsize self) (+ (size self) incr))))



;=======================================
; add object
(defmethod add-one-object ((sheet sheet) (self track) object &optional (time-ms 0))
  (when (allowed-in-sheet object)
    (let* ((time (max 0 time-ms))
           (ev (make-instance 'track-ev
                 :obj object
                 :start-t time
                 :end-t (+ time (get-obj-dur object)))))
      (setf (mycontainer ev) sheet)
      (setf (objs self) (sort (append (objs self) (list ev)) '< :key 'start-t))
      )))
  
(defmethod remove-one-object  ((self sheet) (object track-ev)) 
  (let ((thetrack nil))
    (delete-connections self object)
    (loop for tr in (voices self) while (not thetrack) do
          (when (member object (objs tr) :test 'equal)
            (setf thetrack tr)))
    (setf (objs thetrack) (remove object (objs thetrack) :test 'equal))))




