;OpenMusic
;
;Copyright (C) 1997, 1998, 1999, 2000 by IRCAM-Centre Georges Pompidou, Paris, France.
; 
;This program is free software; you can redistribute it and/or
;modify it under the terms of the GNU General Public License
;as published by the Free Software Foundation; either version 2
;of the License, or (at your option) any later version.
;
;See file LICENSE for further informations on licensing terms.
;
;This program is distributed in the hope that it will be useful,
;but WITHOUT ANY WARRANTY; without even the implied warranty of
;MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;GNU General Public License for more details.
;
;You should have received a copy of the GNU General Public License
;along with this program; if not, write to the Free Software
;Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
;
;Authors: Gerard Assayag and Augusto Agon

(in-package :om)



;-------------------Class definition-----------------------------
(defclass internalsound (om-sound)   
  ((sound-offset :accessor sound-offset)
   (pict-sound :initform nil :accessor pict-sound)
   (pict-spectre :initform nil :accessor pict-spectre)
   (pict-spectre? :initform nil :accessor pict-spectre?)
   (time-started :initarg :time-started :initform nil :accessor time-started)
   (selec  :initform nil :accessor selec)))

(defclass* sound (simple-score-element internalsound) 
  ((tracknum :accessor tracknum :initarg :tracknum :initform 1)
   (markers :accessor markers :initarg :markers :initform nil)
   (vol :accessor vol :initform 100)  
   (pan :accessor pan :initform 0)
   )  
   (:icon 221)
   (:documentation "

Inputs (left to right)

a sound object

Outputs (left to right)

a sound object

Sound factories are generally obtained by dragging a aiff soundfile 
from the finder to a patch window. After doing that, lock immediatly 
the Sound icon, otherwise OM will complain if you try to evaluate it.

Play the Sound by selecting the icon and pressing 'p' on the keyboard.

You may also obtain a Sound factory in a patch by dragging the icon of
the Sound Class from the Package music:score to your patch. The Sound 
factory will come empty.

When you evaluate a Sound factory which is not locked and with nothing 
connected to its input, OM issues a file dialog that lets you choose 
an aiff Sound file to be loaded in the factory.

Dbl-clicking on a Sound icon opens a Sound displayer.

"))


(defmethod! set-channel ((self sound) chan)
  (setf (tracknum self) chan))
  

(defmethod get-type-of-ed-box ((self sound))  'OMaiffFilebox)

(defmethod real-dur ((self sound)) (round (extent->ms self)))

(defmethod real-duration ((self sound) time) 
  (values time (+ time (round (extent->ms self)))))


(defmethod copy-container ((self sound) &optional (pere ()))
   (let ((snd (if (om-sound-file-name self) 
                  (let ((copy (load-sound-file (om-sound-file-name self)))
                        (slots  (class-instance-slots (find-class 'simple-container))))
                    (setf (slot-value copy 'parent) pere)
                    (loop for slot in slots
                        when (not (eq (slot-definition-name slot) 'parent))
                        do (setf (slot-value  copy  (slot-definition-name slot))
                             (copy-container (slot-value self  (slot-definition-name slot)) copy)))
                    copy)
                (make-instance 'sound))))
     (setf (tracknum snd) (tracknum self))
     (setf (markers snd) (markers self))
     snd))


(defmethod execption-save-p ((self sound)) 'sound)

(defmethod save-exepcion ((self sound)) 
  (and (om-sound-file-name self)
       `(let ((thesound (load-sound ,(om-save-pathname-relative (om-sound-file-name self)) ,(tracknum self))))
          (when thesound
            (setf (markers thesound) ',(markers self)))
          thesound)))


(defmethod sound-name ((self sound) )
   (pathname-name (om-sound-file-name self)))

(defmethod sound-path ((self sound) )
   (om-sound-file-name self))


(defmethod set-sound-spectre ((self sound) (pict picture))
  (setf  (pict-spectre self) pict))



(defmethod cons-new-object ((self sound) args objs)
  (let ((rep (call-next-method)))
    (when rep
      (setf (tracknum rep) (if (integerp (nth 1 args)) (nth 1 args) 1))
      (when (consp (nth 2 args)) (setf (markers rep) (nth 2 args)))
    rep)))


;;;------------------Load aiff file method---------------------------            
 

(defmethod* get-sound () 
   :initvals nil
   :indoc nil
   :doc "load a Aiff file"
   :icon 148
  (let ((name (om-choose-file-dialog)))
    (when name
        (load-sound-file name))))


(defun load-sound-file (name &optional track)
  (when (probe-file name)
   (om-print (string+ "Loading sound file : " (om-namestring name)))
   (let ((thesound (om-make-sound 'sound name)) taille)         ; 
     (if (and thesound (not (equal 'oa::ERR (om-sound-format thesound))))
         (progn 
            ;;;(om-sound-get-info thesound)
            (setf (extent thesound)  (round (* 1000 (/ (om-sound-n-samples thesound) (om-sound-sample-rate  thesound)))))
            (setf (Qvalue thesound)  1000)
           (setf (pict-sound thesound) (om-sound-get-pict thesound))
            )
       (om-message-dialog (format nil "Cannot load the file : ~S" (namestring name)))
       )
     thesound)))


;;; copie : meme ptrs (pour le maquette play)
(defun copy-sound-file (sound)
  (let ((thesound (make-instance (type-of sound)
                    :filename (om-sound-file-name sound)
                    :audio-format (om-sound-format sound)
                    :number-of-samples (om-sound-n-samples sound)
                    :sample-rate (om-sound-sample-rate  sound)
                    :number-of-channels (om-sound-n-channels sound)
                    :data-position (om-sound-data-pos sound)
                    )))
    (if thesound
      (progn
        (setf (extent thesound)  (round (* 1000 (/ (om-sound-n-samples thesound) (om-sound-sample-rate  thesound)))))
        (setf (Qvalue thesound)  1000)
        (setf (pict-sound thesound) (pict-sound sound))
        (setf (tracknum thesound) (tracknum sound))
        (setf (markers thesound) (markers sound))
        (setf (vol thesound) (vol sound))
        (setf (pan thesound) (pan sound))
        )
      (om-message-dialog (format nil "Cannot copy the file : ~S" (namestring name)))
      )
    thesound))


(defun load-sound (name &optional track)
  (let ((snd (om-load-if name 'load-sound-file)))
    (when (and snd track) (setf (tracknum snd) track))
    snd))



(defmethod* objfromobjs ((self string) (type sound))
   (load-sound-file self))

(defmethod* objfromobjs ((self pathname) (type sound))
   (load-sound-file self))

(defmethod pic-to-draw ((self sound)) 
  (if (and (pict-spectre self) (pict-spectre? self))
    (thepict (pict-spectre self)) (pict-sound self)))


;;; copy for play in maq
(defmethod maq-copy-container ((self sound)  &optional (pere ())) 
  (let ((copy (copy-sound-file self))
        (slots  (class-instance-slots (find-class 'simple-container))))
    (setf (slot-value copy 'parent) pere)
    (loop for slot in slots
          when (not (eq (slot-definition-name slot) 'parent))
          do (setf (slot-value  copy  (slot-definition-name slot))
                   (copy-container (slot-value self  (slot-definition-name slot)) copy)))
    
    copy))


;=====================================================
;METHODS FOR EDITION
;=====================================================

(defmethod Class-has-editor-p ((self sound)) t)

(defmethod get-initval ((self sound)) (make-instance 'sound))

(defmethod get-boxsize ((self sound)) (om-make-point 80 50))

(defmethod get-editor-class ((self sound)) 'soundEditor)

(defmethod print-object ((self sound) x) (format x "AudioFile"))

(defmethod make-one-instance ((class sound) &rest slots-vals) 
  (let ((snd (get-sound)))
    (when (car slots-vals) (setf (tracknum snd) (car slots-vals)))
    snd))


;;;=====================
;;; OPTIMISER AVEC L'OUVERTURE DU FICHIER !!!
(defmethod* sound-points ((self sound) (num integer) &optional channel)
  :initvals '(nil 1000 1)
  :indoc '("sound file" "num of points")
  :doc "load the sound points"
  :icon 221
   (let ((sizedata (om-sound-sample-size self))
           (numdat (om-sound-n-samples  self))
           (numchan (om-sound-n-channels  self))
           (position 0)
           (old-pos 0)
           (rep nil))
     (setf channel (if channel channel 1))
     (setf position (+ (om-sound-data-pos self) (* (round sizedata 8) (- channel 1))))
     (setf old-pos position)
     (if (or (> channel numchan) (> num numdat)) 
       (om-message-dialog "Bad input values")
       (loop for i from 0 to numdat by (round numdat num) do
                 (let (new-point)
                   ;(cond
                   ; ((= sizedata 8) (setf new-point (om-read-sound-data self position 1)))
                   ; ((= sizedata 16) (setf new-point (round (om-read-sound-data self position 2) 256))))
                   (setf new-point (round (om-read-sound-data self position (round sizedata 8)) (ash 1 (- sizedata 8))))
                   (when (> new-point 128)
                       (setf new-point (- new-point 256)))
                   (push new-point rep)
                   )  
                 (setf position (+ old-pos (* numchan (* (/ sizedata 8) i))))
                 ))
     (reverse rep)))



(defmethod! sound-dur-ms ((s pathname))
  :initvals '(nil)
  :indoc '("sound file or pathname")
  :doc "return the sound duration in ms"
  :icon 221
  (let ((thesound (om-make-sound 'sound s)))
    (when thesound
      ;;;(om-sound-get-info thesound)   
      (round (* 1000 (/ (om-sound-n-samples  thesound) (om-sound-sample-rate  thesound)))))))

(defmethod! sound-dur-ms ((s sound))
  (round (* 1000 (/ (om-sound-n-samples s) (om-sound-sample-rate s)))))


 ;=======EDITOR 

(defmethod good-val-p? ((self sound))
   (pict-sound self))

(defclass OMaiffFilebox (OMBoxEditCall) ())

(defmethod get-frame-class ((self OMaiffFilebox)) 'boxsoundframe)

(omg-defclass boxsoundframe (boxEditorFrame) ())

(defmethod om-get-menu-context ((object boxsoundframe))
  (list+  (list (om-new-leafmenu (if (view-of-patch (object object)) "free view of patch" "set view of patch")
                                 #'(lambda () (change-view-of-patch object)))
                (om-new-leafmenu "-" #'(lambda () nil))
                ) 
          (boxframe-default-list object)))




;;;========= synthmaq

(defmethod cons-maq-mini-pict ((self sound) frame fontsize size)
  (om-sound-get-pict self))






  