;OpenMusic
;
;Copyright (C) 1997, 1998, 1999, 2000 by IRCAM-Centre Georges Pompidou, Paris, France.
; 
;This program is free software; you can redistribute it and/or
;modify it under the terms of the GNU General Public License
;as published by the Free Software Foundation; either version 2
;of the License, or (at your option) any later version.
;
;See file LICENSE for further informations on licensing terms.
;
;This program is distributed in the hope that it will be useful,
;but WITHOUT ANY WARRANTY; without even the implied warranty of
;MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;GNU General Public License for more details.
;
;You should have received a copy of the GNU General Public License
;along with this program; if not, write to the Free Software
;Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
;
;Authors: Gerard Assayag and Augusto Agon

(in-package :om)


;=================================================
;MIDI PREFERENCES MODULE  icon 212
;=================================================

;;;; old version saved ports as strings
(defun strig2numport (port &optional (out t))
  (if (integerp port) port
      (if out *Outmidiport* *Inmidiport*)))


(defun launch-midishare-setup ()
  (if (and *ms-setup-app* (om-find-process *ms-setup-app*))
    (om-select-process *ms-setup-app*)
    (om-launch-ms-setup :before 'om::close-ms-players :after 'om::open-ms-players)))

(defmethod put-preferences ((iconID (eql 212)))
   (when *MIDI*
     (let ((modulepref (get-pref-by-icon iconID)))
       (setf *Outmidiport* (strig2numport (first (defvals modulepref))))
       (setf *Inmidiport* (strig2numport (second (defvals modulepref)) nil)))))

(defmethod get-def-vals ((iconID (eql 212))) '(0 0))


(defmethod make-new-pref-scroll ((win preferenceEditor) (num (eql 212)))
   (let ((thescroll (om-make-view 'PreferencePanel
                                                           :size (om-make-point 500 500)
                                                           :position (om-make-point 66 0)
                                                           :scrollbars :v 
                                                           :retain-scrollbars t
                                                           ))
         (modulepref (get-pref-by-icon num (object win))))
     (om-add-subviews thescroll
       (om-make-dialog-item 'om-static-text (om-make-point 100 35) (om-make-point 160 15) "Midi Share SetUp")
       (om-make-view 'button-icon
                                    :position (om-make-point 135 65) 
                                    :size (om-make-point 32 32)
                                    :action #'(lambda (item) (declare (ignore item))
                                                       (launch-midishare-setup))
                                    :iconid 135)
       (om-make-dialog-item 'om-static-text (om-make-point 20 135) (om-make-point 150 15) "Default Output Port : ")
       (om-make-dialog-item 'om-static-text (om-make-point 20 165) (om-make-point 150 15) "Default  Input Port : ")
       (om-make-dialog-item 'om-editable-text (om-make-point 180 135) (om-get-text-edit-size (om-make-point 37 13)) (format nil "~D" *Outmidiport*) 
                                                 :di-action (om-dialog-item-act item
                                                  (let ((text (om-dialog-item-text item))
                                                          number)
                                                    (unless (string= "" text)
                                                       (setf number (read-from-string text))
                                                       (when (and (integerp number) (>= number 0) (<= number 255))
                                                           (setf (nth 0 (defvals modulepref)) number)))))
                                                 :font *om-default-font2*)
        (om-make-dialog-item 'om-editable-text (om-make-point 180 165) (om-get-text-edit-size (om-make-point 37 13)) (format nil "~D" *Inmidiport*) 
                                                  :di-action (om-dialog-item-act item
                                                   (let ((text (om-dialog-item-text item))
                                                           number)
                                                     (unless (string= "" text)
                                                        (setf number (read-from-string text))
                                                        (when (and (integerp number) (>= number 0) (<= number 255))
                                                            (setf (nth 1 (defvals modulepref)) number)))))
                                                  :font *om-default-font2*)) 
     thescroll))



;;;;=================================================
;;;;MUSIC PREFERENCES MODULE  icon 225
;;;;=================================================
(omg-defclass change-color-dialog-item (om-static-text) 
   ((object :initform 0 :initarg :object :accessor object)
    (i :initform 0 :initarg :i :accessor i))
   )

(defmethod om-dialog-item-action ((item change-color-dialog-item))
   (let ((newco (om-choose-color-dialog :color (nth (i item) (defvals (object item))))))
     (when newco
         (setf (nth (i item) (defvals (object item)))  newco))
     (om-set-part-color item :text newco)
     (om-invalidate-view item)))

(defmethod om-component-border ((self change-color-dialog-item))
   :line)

(defmethod om-view-click-handler ((self change-color-dialog-item) where)
  (declare (ignore where))
  (om-dialog-item-action self))


;;;--------
(defvar *page-reglages* '(1 15 2 2 1))
; (margenizq  ,  ancho ,  separation systemas , separation entre staffs, margen de arriba)
; Non editable 

(defvar *music-fontsize* 24)
(defvar *default-satff* 'g)
(defvar *global-deltachords* 100)
(defvar *quantify-def-params* '( 60 (4 4) 8  0 nil 0.5))


(defmethod put-preferences ((iconID (eql 225)))
   (let ((modulepref (get-pref-by-icon iconID)))
     (setf *global-midi-approx* (first (defvals modulepref)))
     (setf *music-fontsize* (second (defvals modulepref)))
     (setf *global-deltachords* (third (defvals modulepref)))
     (setf *default-satff* (fourth (defvals modulepref)))
     (setf *quantify-def-params* (fifth (defvals modulepref))
           *system-color* (or (sixth (defvals modulepref)) *om-black-color*)
           *note-color* (or (nth 6 (defvals modulepref)) *om-black-color*)
           *select-color* (or (nth 7 (defvals modulepref)) *om-light-gray-color*)
           *current-1/2-scale*  (or (nth 8 (defvals modulepref)) *2-tone-chromatic-scale*)
           *current-1/4-scale* (or (nth 9 (defvals modulepref)) *4-tone-chromatic-scale*)      
           *current-1/8-scale* (or (nth 10 (defvals modulepref)) *8-tone-chromatic-scale*)   
           )))

(defmethod corrige-preference-list ((iconID (eql 225)) list)
   (let (rep)
     (loop for i = 0 then (+ i 1)
           for newitem in (get-def-vals 225) do
           (if (car list)
             (if (> i 7) 
               (push (eval (pop list)) rep)
               (push (pop list) rep))
             (push newitem rep)))
     (reverse rep)))


(defmethod save-pref ((iconID (eql 225)) item)
   (list 225 `(list ,.(append (loop for i from 0 to 7 collect (omng-save (nth i (defvals item))))
                     (list  (omng-save *current-1/2-scale*)  (omng-save *current-1/4-scale*)
                            (omng-save *current-1/8-scale*))))))

(defmethod get-def-vals ((iconID (eql 225)))
   (list 2 24 100 'g '( 60 (4 4) 8  0 nil 0.5) 
     *om-black-color* *om-black-color* *om-light-gray-color* *2-tone-chromatic-scale* *4-tone-chromatic-scale* *8-tone-chromatic-scale*))
    
(defmethod make-new-pref-scroll ((win preferenceEditor) (num (eql 225)))
  (let ((thescroll (om-make-view 'PreferencePanel
                                 :size (om-make-point 500 500)
                                 :position (om-make-point 66 0)
                                 :scrollbars :v 
                                 :retain-scrollbars t))
        (modulepref (get-pref-by-icon num (object win))))
    (om-add-subviews thescroll
                     (om-make-dialog-item 'om-static-text  (om-make-point 10 20) (om-make-point 80 15) "Approx" )
                     (om-make-dialog-item 'om-radio-button (om-make-point 100 20) (om-make-point 40 15) "1/2" 
                                          :di-action (om-dialog-item-act item
                                                       (declare (ignore item))
                                                       (setf (nth 0 (defvals modulepref)) 2))
                                          :checked-p (= (first (defvals modulepref)) 2)  
                                          :radio-button-cluster 'approx
                                          :font *om-default-font2*)
                     
                     (om-make-dialog-item 'om-radio-button (om-make-point 160 20) (om-make-point 40 15) "1/4" 
                                          :di-action (om-dialog-item-act item
                                                       (declare (ignore item))
                                                       (setf (nth 0 (defvals modulepref)) 4))
                                          :checked-p (= (first (defvals modulepref)) 4)  
                                          :radio-button-cluster 'approx
                                          :font *om-default-font2*)
                     
                     (om-make-dialog-item 'om-radio-button (om-make-point 220 20) (om-make-point 40 15) "1/8" 
                                          :di-action (om-dialog-item-act item
                                                       (declare (ignore item))
                                                       (setf (nth 0 (defvals modulepref)) 8))
                                          :checked-p (= (first (defvals modulepref)) 8)  
                                          :radio-button-cluster 'approx
                                          :font *om-default-font2*)
                     
                     (om-make-dialog-item 'om-static-text (om-make-point 10 50) (om-make-point 80 15) "Scale"   )
                     (om-make-dialog-item 'om-static-text (om-make-point 10 50) (om-make-point 120 15) "Music Font Size" )
                     (om-make-dialog-item 'numbox (om-make-point 115 50) (om-make-point 40 15) 
                                          (format () "~D" (second (defvals modulepref)))
                                          :value (second (defvals modulepref))
                                          :min-val 10
                                          :max-val 80
                                          :afterfun #'(lambda (item) 
                                                        (unless (string= "" (om-dialog-item-text item))
                                                          (setf (nth 1 (defvals modulepref)) 
                                                                (read-from-string (om-dialog-item-text item)))))
                                          )
                     
                     (om-make-dialog-item 'om-static-text  (om-make-point 170 50) (om-make-point 120 15) "Delta Chords")
                     (om-make-dialog-item 'numbox (om-make-point 260 50) (om-make-point 40 15) 
                                          (format () "~D" (third (defvals modulepref)))                                           
                                          :value (third (defvals modulepref))
                                          :min-val 0
                                          :max-val 1000000
                                          :afterfun #'(lambda (item)
                                                        (unless (string= "" (om-dialog-item-text item))
                                                          (setf (nth 2 (defvals modulepref)) 
                                                                (read-from-string (om-dialog-item-text item)))))
                                          )
                     
                     (om-make-dialog-item 'om-static-text (om-make-point 10 110) (om-make-point 120 15) "Staff System")
                     
                     (om-make-dialog-item  'om-pop-up-dialog-item
                                           (om-make-point 100 110)
                                           (om-make-point 60 20)
                                           ""
                                           :range (loop for item in *all-satff-om* collect (string item))  ;;; (cons-staff-menu-items *all-satff-om* modulepref )
                                           :value (string (nth 3 (defvals modulepref)))   ;;; (position (nth 3 (defvals modulepref)) *all-satff-om* :test 'equal)
                                           :di-action (om-dialog-item-act item
                                                        (setf (nth 3 (defvals modulepref)) (nth (om-get-selected-item-index item) *all-satff-om*)))
                                           ;;; :auto-update-default t
                                           ;;; :item-display :selection 
                                           )
                     
                     (om-make-dialog-item 'om-static-text  (om-make-point 10 80) (om-make-point 80 15) "Staff Colors")
                     
                     (om-make-dialog-item 'change-color-dialog-item  (om-make-point 110 80) (om-make-point 60 15) "Normal"
                                          :fg-color (nth 5 (defvals modulepref))
                                          :font *om-default-font2b*
                                          :object  modulepref 
                                          :i 5)
                     
                     #|
(om-make-dialog-item 'change-color-dialog-item  (om-make-point 180 80) (om-make-point 40 15) "Note"
                                          :fg-color (nth 6 (defvals modulepref))
                                          :font *om-default-font2b*
                                          :object  modulepref 
                                          :i 6)
|#
                     
                     (om-make-dialog-item 'change-color-dialog-item  (om-make-point 200 80) (om-make-point 63 15) "Selected"
                                          :fg-color (nth 7 (defvals modulepref))
                                          :font *om-default-font2b*
                                          :object  modulepref 
                                          :i 7)
                     
                     (om-make-dialog-item 'om-static-text  (om-make-point 180 110) (om-make-point 100 15) "Quantify")
                     
                     (om-make-view  'button-icon
                                    :position (om-make-point 260 110) 
                                    :size (om-make-point 24 24)
                                    :action #'(lambda (item) (declare (ignore item))
                                               (let ((rep (show-quanty-dialog (nth 4 (defvals modulepref)))))
                                                 (when rep
                                                   (setf (nth 4 (defvals modulepref)) rep))))
                                    :iconid 252)
                     (om-make-dialog-item 'om-button (om-make-point 120 150) (om-make-point 120 22) "Dynamic Table" 
                                          :di-action (om-dialog-item-act item  (declare (ignore item))
                                                                         (open-dyn-ed *cur-dynamic-list*))
                                          :font *om-default-font1*)
                     ) 
    thescroll))


(defun show-quanty-dialog (preflist)
  (let* ((newwindow (om-make-window 'om-dialog
                      ;;;:grow-icon-p nil
                      :window-title "Quantify preferences" 
                      :close nil :resizable nil :maximize nil :minimize nil
                      :position (om-make-point 100 50)
                      :size (om-make-point 300 180)
                                                                        ))
         (tempi (om-make-dialog-item 'om-editable-text (om-make-point 140 5) (om-get-text-edit-size (om-make-point 80 15)) 
                                  (format () "~D" (first preflist))))
         (mes (om-make-dialog-item 'om-editable-text (om-make-point 140 25) (om-get-text-edit-size (om-make-point 80 15)) 
                                (format () "~D" (second preflist))))
         (maxi (om-make-dialog-item 'om-editable-text (om-make-point 140 45) (om-get-text-edit-size (om-make-point 80 15)) 
                                 (format () "~D" (third preflist))))
         (forb (om-make-dialog-item 'om-editable-text (om-make-point 140 65) (om-get-text-edit-size (om-make-point 80 15)) 
                                 (format () "~D" (fifth preflist)  )))
         (preci (om-make-dialog-item 'om-editable-text (om-make-point 140 85) (om-get-text-edit-size (om-make-point 80 15)) 
                                  (format () "~D" (sixth preflist)  )))
         (offset (om-make-dialog-item 'om-editable-text (om-make-point 140 105) (om-get-text-edit-size (om-make-point 80 15)) 
                                   (format () "~D" (fourth preflist)  ))))
    (om-add-subviews newwindow tempi mes maxi forb preci offset
                  (om-make-dialog-item 'om-static-text  (om-make-point 10 5) (om-make-point 120 15) "tempi" :bg-color *om-window-def-color*)
                  (om-make-dialog-item 'om-static-text  (om-make-point 10 25) (om-make-point 120 15) "measure" :bg-color *om-window-def-color*)
                  (om-make-dialog-item 'om-static-text  (om-make-point 10 45) (om-make-point 120 15) "maximum division" :bg-color *om-window-def-color*)
                  (om-make-dialog-item 'om-static-text  (om-make-point 10 65) (om-make-point 120 15) "forbidden divisions" :bg-color *om-window-def-color*)
                  (om-make-dialog-item 'om-static-text  (om-make-point 10 85) (om-make-point 120 15) "precision" :bg-color *om-window-def-color*)
                  (om-make-dialog-item 'om-static-text  (om-make-point 10 105) (om-make-point 120 15) "offset" :bg-color *om-window-def-color*)
                                        
                  (om-MAKE-DIALOG-ITEM 'om-button (om-make-point 60 145) (om-make-point 70 22) "cancel" 
                                    :di-action (om-dialog-item-act item (declare (ignore item))(om-return-from-modal-dialog newwindow ())))
                  (om-MAKE-DIALOG-ITEM 'om-button (om-make-point 160 145) (om-make-point 70 22) "Ok" 
                                    :di-action (om-dialog-item-act item (declare (ignore item)) 
                                                                                            (om-return-from-modal-dialog newwindow 
                                                                                                                                                     (list (read-from-string (om-dialog-item-text tempi))
                                                                                                                                                            (read-from-string (om-dialog-item-text mes))
                                                                                                                                                            (read-from-string (om-dialog-item-text maxi))
                                                                                                                                                            (read-from-string (om-dialog-item-text offset))
                                                                                                                                                            (read-from-string (om-dialog-item-text forb))
                                                                                                                                                            (read-from-string (om-dialog-item-text preci)))))))
    (om-modal-dialog newwindow)))


;;;(defun cons-staff-menu-items (list modulepref )
;;;  `(list ,.(loop for item in list
;;;                 collect `(make-instance 'menu-item
;;;                               :menu-item-title ,(string item)
;;;                               :menu-item-action #'(lambda () (setf (nth 3 (defvals ,modulepref)) ',item))))))


;;;==== ADD MIDI AND MUSIC PREFERENCE MODULES ========
(defun add-music/midi-preferences ()
   (pushr 
    (om-make-view 'pref-icon-view :iconID 212 :size (om-make-point 40 40) :position (om-make-point 5 75)
                                 :defvals '(0 0) 
                                 ;(Outportname Inportname)
                                 )
    *pref-item-list*)
   
   (pushr 
    (om-make-view 'pref-icon-view :iconID 225 :size (om-make-point 40 40) :position (om-make-point 5 135)
                                 :defvals (get-def-vals 225) 
                                 ;(approx fontsize deltachord default-staff quantiparams)
                                 )
    *pref-item-list*)
   
   
   )

(om-add-init-func 'add-music/midi-preferences)


;;;============================================
;;;;loading pictures

(defvar *aiff-palette* nil)
(defvar *midi-palette* nil)
(defvar *music-palette* nil)
(defvar *insertmodes* nil)
 
(defun init-music-pict ()
   (setf *midi-palette* (om-load-and-store-picture 'midi-palette 'kernel))
   (setf *aiff-palette* (om-load-and-store-picture 'aiff-palette 'kernel))
   (setf *music-palette* (om-load-and-store-picture 'musicpalette 'kernel))
   (setf *insertmodes* (om-load-and-store-picture 'insertmodes 'kernel)))

(om-add-init-func 'init-music-pict)


(defvar *add-cursor* nil)
(defvar *hand-up* nil)
(defvar *handleft* nil)
(defvar *staff-cursor* nil)
(defvar *move-chord* nil)
(defvar *c-nota* nil)
(defvar *c-chord* nil)
(defvar *c-measure* nil)
(defvar *c-voice* nil)
(defvar *c-group* nil)

(defun create-score-cursors ()
   (setf *add-cursor* (om-make-cursor "add-cursor" (om-make-point 7 0)))
   (setf *hand-up* (om-make-cursor "hand-up" (om-make-point 8 9)))
   (setf *handleft* (om-make-cursor "handleft" (om-make-point 8 9)))
   (setf *staff-cursor* (om-make-cursor "staff-cursor" (om-make-point 8 9)))
   (setf *move-chord* (om-make-cursor "move-chord" (om-make-point 7 0)))
   (setf *c-nota* (om-make-cursor "c-nota" (om-make-point 0 0)))
   (setf *c-chord* (om-make-cursor "c-chord" (om-make-point 0 0)))
   (setf *c-measure* (om-make-cursor "c-measure" (om-make-point 0 0)))
   (setf *c-voice* (om-make-cursor "c-voice" (om-make-point 0 0)))
   (setf *c-group* (om-make-cursor "c-group" (om-make-point 0 0)))
   )

(om-add-init-func 'create-score-cursors) 