;OpenMusic
;
;Copyright (C) 1997, 1998, 1999, 2000 by IRCAM-Centre Georges Pompidou,Paris, France.
;
;This program is free software; you can redistribute it and/or
;modify it under the terms of the GNU General Public License
;as published by the Free Software Foundation; either version 2
;of the License, or (at your option) any later version.
;
;See file LICENSE for further informations on licensing terms.
;
;This program is distributed in the hope that it will be useful,
;but WITHOUT ANY WARRANTY; without even the implied warranty of
;MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;GNU General Public License for more details.
;
;You should have received a copy of the GNU General Public License
;along with this program; if not, write to the Free Software
;Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
;
;Authors: Gerard Assayag and Augusto Agon

(in-package :om)
;===================================================
;popup + dialog used to change staff size port, etc.
;================================================
(defvar *all-satff-om*   '(f g g2 f2 gf gg ff gff ggf ggff empty))

(defvar *chord-satff-om* '(("F" f) ("G" g) ("GF" gf) ("GG" gg) ("FF" ff) ("GFF" gff) ("GGF" ggf) ("GGFF" ggff) ("EMPTY" empty)))
(defvar *voice-satff-om* '(("F" f) ("G" g) ("F2" f2) ("G2" g2) ("EMPTY" empty)))
(defvar *mus-font-size* '(("8" 8) ("12" 12) ("16" 16) ("20" 20) ("24" 24) ("28" 28)
                            ("36" 36) ("48" 48) ("60" 60) ("72" 72)))




;=================


(defclass* scale ()
    ((alteration-list :initform nil :initarg :alteration-list :accessor alteration-list)
     (lines-list :initform nil :initarg :lines-list :accessor lines-list)
     (approx-factor :initform 100 :initarg :approx-factor :accessor approx-factor))
   (:icon 262))

(defmethod Class-has-editor-p  ((self scale)) nil)

(defmethod give-alteration  ((self scale) midic)
    (let* ((vlength (length (lines-list self)))
           (int-cents (mod midic 1200))
           (index (round (/ int-cents (approx-factor self) )))
           (up-octave (truncate (/ index (length (lines-list self))))))
      (setq index (mod index (length (lines-list self))))
      (if (and (= (approx-factor self) 25) (= index (1- vlength))) (incf up-octave))
      (list (nth index  (lines-list self)) (nth  index (alteration-list self)) up-octave)))

;-------1
(defvar *1-tone-chromatic-scale*
    (make-instance 'scale
      :alteration-list (list  nil  nil nil  (diese)  (diese)  (diese))
      :lines-list (list  0 1 2 3 4 5 )
      :approx-factor 200))

(defvar *1#-tone-chromatic-scale*
    (make-instance 'scale
      :alteration-list (list  (diese)  (diese) nil  nil  nil  nil)
      :lines-list (list  0 1 3 4 5 6 )
      :approx-factor 200))

;-------2
(defvar *2-tone-chromatic-scale*
   (make-instance 'scale
     :alteration-list
     (list  nil (diese) nil (diese) nil nil (diese) nil (diese) nil (diese) nil)
     :lines-list
     (list  0 0 1 1 2 3 3 4 4 5 5 6)
     :approx-factor 100))

(defvar *c-major-scale*
   (make-instance 'scale
     :alteration-list
     (list  nil (diese)  nil (bemol)   nil   nil  (diese) nil (bemol) nil (bemol)  nil)
     :lines-list
     (list   0 0 1 2 2 3 3 4 5 5 6 6)
     :approx-factor 100))


;-------3
(defvar *3-tone-chromatic-scale*
    (make-instance 'scale
      :alteration-list (list  nil (t-4/12) (t-8/12)
                              nil (t-4/12) (t-8/12)
                              nil (t-4/12) (t-8/12)
                              #\# (d-1/3) (d-2/3)
                              #\# (d-1/3) (d-2/3)
                              #\# (d-1/3) (d-2/3))
      :lines-list (list  0 0 0 1 1 1 2 2 2 3 3 3 4 4 4 5 5 5)
      :approx-factor 200/3))

(defvar *3#-tone-chromatic-scale*
    (make-instance 'scale
      :alteration-list (list  #\# (d-1/3) (d-2/3)
                              #\# (d-1/3) (d-2/3)
                              nil (t-4/12) (t-8/12)
                              nil (t-4/12) (t-8/12)
                              nil (t-4/12) (t-8/12)
                              nil (t-4/12) (t-8/12))
      :lines-list (list  0 0 0 1 1 1 3 3 3  4 4 4 5 5 5 6 6 6 )
      :approx-factor 200/3))

;-------4
(defvar *4-tone-chromatic-scale*
   (make-instance 'scale
     :alteration-list
     (list  nil (t-1/4) (diese) (t-3/4) nil (t-1/4) (diese) (t-3/4) nil (t-1/4) nil (t-1/4) (diese)
            (t-3/4) nil (t-1/4) (diese) (t-3/4) nil (t-1/4) (diese) (t-3/4) nil (t-1/4))
     :lines-list
     (list  0 0 0 0 1 1 1 1 2 2 3 3 3 3 4 4 4 4 5 5 5 5 6 6)
     :approx-factor 50))

;-------5
(defvar *5-tone-chromatic-scale*
   (make-instance 'scale
     :alteration-list
     (list  nil (t-2/10) (t-4/10) (t-6/10) (t-8/10)
            nil (t-2/10) (t-4/10) (t-6/10) (t-8/10)
            nil (t-2/10) (t-4/10) (t-6/10) (t-8/10)
            #\# (d-1/5) (d-2/5) (d-3/5) (d-4/5)
            #\# (d-1/5) (d-2/5) (d-3/5) (d-4/5)
            #\# (d-1/5) (d-2/5) (d-3/5) (d-4/5))
     :lines-list
     (list  0 0 0 0 0
            1 1 1 1 1
            2 2 2 2 2
            3 3 3 3 3
            4 4 4 4 4
            5 5 5 5 5)
     :approx-factor 40))

(defvar *5#-tone-chromatic-scale*
    (make-instance 'scale
      :alteration-list (list  #\# (d-1/5) (d-2/5) (d-3/5) (d-4/5)
                              #\# (d-1/5) (d-2/5) (d-3/5) (d-4/5)
                              nil (t-2/10) (t-4/10) (t-6/10) (t-8/10)
                              nil (t-2/10) (t-4/10) (t-6/10) (t-8/10)
                              nil (t-2/10) (t-4/10) (t-6/10) (t-8/10)
                               nil (t-2/10) (t-4/10) (t-6/10) (t-8/10))
      :lines-list (list  0 0 0 0 0
                         1 1 1 1 1
                         3 3 3 3 3
                         4 4 4 4 4
                         5 5 5 5 5
                         6 6 6 6 6 )
      :approx-factor 40))

;-------6
(defvar *6-tone-chromatic-scale*
    (make-instance 'scale
      :alteration-list (list  nil (t-2/12) (t-4/12) (t-1/2) (t-8/12) (t-10/12)
                              nil (t-2/12) (t-4/12) (t-1/2) (t-8/12) (t-10/12)
                              nil (t-2/12) (t-4/12) nil (t-2/12) (t-4/12)
                              #\# (t-8/12) (t-10/12) nil (t-2/12) (t-4/12)
                              #\# (t-8/12) (t-10/12) nil (t-2/12) (t-4/12)
                              #\# (t-8/12) (t-10/12) nil (t-2/12) (t-4/12))
      :lines-list (list  0 0 0 0 0 0
                         1 1 1 1 1 1
                         2 2 2 3 3 3
                         3 3 3 4 4 4
                         4 4 4 5 5 5
                         5 5 5 6 6 6)
      :approx-factor 200/6))



;-------7
(defvar *7-tone-chromatic-scale*
   (make-instance 'scale
     :alteration-list
     (list  nil (t-2/14) (t-4/14) (t-6/14) (t-8/14) (t-10/14) (t-12/14)
            nil (t-2/14) (t-4/14) (t-6/14) (t-8/14) (t-10/14) (t-12/14)
            nil (t-2/14) (t-4/14) (t-6/14) (t-8/14) (t-10/14) (t-12/14)
            #\# (d-1/7) (d-2/7) (d-3/7) (d-4/7) (d-5/7) (d-6/7)
            #\# (d-1/7) (d-2/7) (d-3/7) (d-4/7) (d-5/7) (d-6/7)
            #\# (d-1/7) (d-2/7) (d-3/7) (d-4/7) (d-5/7) (d-6/7))
     :lines-list
     (list  0 0 0 0 0 0 0
            1 1 1 1 1 1 1
            2 2 2 2 2 2 2
            3 3 3 3 3 3 3
            4 4 4 4 4 4 4
            5 5 5 5 5 5 5)
     :approx-factor 200/7))

(defvar *7#-tone-chromatic-scale*
    (make-instance 'scale
      :alteration-list (list  #\# (d-1/7) (d-2/7) (d-3/7) (d-4/7) (d-5/7) (d-6/7)
                              #\# (d-1/7) (d-2/7) (d-3/7) (d-4/7) (d-5/7) (d-6/7)
                              nil (t-2/14) (t-4/14) (t-6/14) (t-8/14) (t-10/14) (t-12/14)
                              nil (t-2/14) (t-4/14) (t-6/14) (t-8/14) (t-10/14) (t-12/14)
                              nil (t-2/14) (t-4/14) (t-6/14) (t-8/14) (t-10/14) (t-12/14))
      :lines-list (list  0 0 0 0 0 0 0
                         1 1 1 1 1 1 1
                         3 3 3 3 3 3 3
                         4 4 4 4 4 4 4
                         5 5 5 5 5 5 5
                         6 6 6 6 6 6 6 )
      :approx-factor 200/7))

;-------8
(defvar *8-tone-chromatic-scale*
    (make-instance 'scale
      :alteration-list
      (list  nil (t-1/8) (t-1/4) (t-3/8) (t-1/2) (t-5/8) (t-3/4) (t-7/8)
             nil (t-1/8) (t-1/4) (t-3/8) (t-1/2) (t-5/8) (t-3/4) (t-7/8)
             nil (t-1/8) (t-1/4) (t-3/8)
             nil (t-1/8) (t-1/4) (t-3/8) (t-1/2) (t-5/8) (t-3/4) (t-7/8)
             nil (t-1/8) (t-1/4) (t-3/8) (t-1/2) (t-5/8) (t-3/4) (t-7/8)
             nil (t-1/8) (t-1/4) (t-3/8) (t-1/2) (t-5/8) (t-3/4) (t-7/8)
             nil (t-1/8) (t-1/4) (t-3/8) )
      :lines-list
      (list  0 0 0 0 0 0 0 0
             1 1 1 1 1 1 1 1
             2 2 2 2 3 3 3 3
             3 3 3 3 4 4 4 4
             4 4 4 4 5 5 5 5
             5 5 5 5 6 6 6 6)
      :approx-factor 25))

;-------10
(defvar *10-tone-chromatic-scale*
    (make-instance 'scale
      :alteration-list
      (list  nil (t-1/10) (t-2/10) (t-3/10) (t-4/10) (t-5/10) (t-6/10) (t-7/10) (t-8/10) (t-9/10)
             nil (t-1/10) (t-2/10) (t-3/10) (t-4/10) (t-5/10) (t-6/10) (t-7/10) (t-8/10) (t-9/10)
             nil (t-1/10) (t-2/10) (t-3/10) (t-4/10) nil (t-1/10) (t-2/10) (t-3/10) (t-4/10)
             #\# (t-6/10) (t-7/10) (t-8/10) (t-9/10)  nil (t-1/10) (t-2/10) (t-3/10) (t-4/10)
             #\# (t-6/10) (t-7/10) (t-8/10) (t-9/10) nil (t-1/10) (t-2/10) (t-3/10) (t-4/10)
             #\# (t-6/10) (t-7/10) (t-8/10) (t-9/10) nil (t-1/10) (t-2/10) (t-3/10) (t-4/10))
      :lines-list (list  0 0 0 0 0 0 0 0 0 0
                         1 1 1 1 1 1 1 1 1 1
                         2 2 2 2 2 3 3 3 3 3
                         3 3 3 3 3 4 4 4 4 4
                         4 4 4 4 4 5 5 5 5 5
                         5 5 5 5 5 6 6 6 6 6)
      :approx-factor 20))

;-------12
(defvar *12-tone-chromatic-scale*
    (make-instance 'scale
      :alteration-list
      (list  nil (t-1/12) (t-2/12) (t-3/12) (t-4/12) (t-5/12) (t-6/12) (t-7/12) (t-8/12) (t-9/12) (t-10/12) (t-11/12)
             nil (t-1/12) (t-2/12) (t-3/12) (t-4/12) (t-5/12) (t-6/12) (t-7/12) (t-8/12) (t-9/12) (t-10/12) (t-11/12)
             nil (t-1/12) (t-2/12) (t-3/12) (t-4/12) (t-5/12) nil (t-1/12)  (t-2/12) (t-3/12) (t-4/12) (t-5/12)
             #\# (t-7/12) (t-8/12) (t-9/12) (t-10/12) (t-11/12) nil (t-1/12)  (t-2/12) (t-3/12) (t-4/12) (t-5/12)
             #\# (t-7/12) (t-8/12) (t-9/12) (t-10/12) (t-11/12) nil (t-1/12)  (t-2/12) (t-3/12) (t-4/12) (t-5/12)
             #\# (t-7/12) (t-8/12) (t-9/12) (t-10/12) (t-11/12) nil (t-1/12)  (t-2/12) (t-3/12) (t-4/12) (t-5/12))
      :lines-list (list  0 0 0 0 0 0 0 0 0 0 0 0
                         1 1 1 1 1 1 1 1 1 1 1 1
                         2 2 2 2 2 2 3 3 3 3 3 3
                         3 3 3 3 3 3 4 4 4 4 4 4
                         4 4 4 4 4 4 5 5 5 5 5 5
                         5 5 5 5 5 5 6 6 6 6 6 6)
      :approx-factor 200/12))

;-------14
(defvar *14-tone-chromatic-scale*
    (make-instance 'scale
      :alteration-list
      (list  nil (t-1/14) (t-2/14) (t-3/14) (t-4/14) (t-5/14) (t-6/14) (t-7/14) (t-8/14) (t-9/14) (t-10/14) (t-11/14) (t-12/14) (t-13/14)
             nil (t-1/14) (t-2/14) (t-3/14) (t-4/14) (t-5/14) (t-6/14) (t-7/14) (t-8/14) (t-9/14) (t-10/14) (t-11/14) (t-12/14) (t-13/14)
             nil (t-1/14) (t-2/14) (t-3/14) (t-4/14) (t-5/14) (t-6/14) nil (t-1/14) (t-2/14) (t-3/14) (t-4/14) (t-5/14) (t-6/14)
             #\# (t-8/14) (t-9/14) (t-10/14) (t-11/14) (t-12/14) (t-13/14) nil (t-1/14) (t-2/14) (t-3/14) (t-4/14) (t-5/14) (t-6/14)
             #\# (t-8/14) (t-9/14) (t-10/14) (t-11/14) (t-12/14) (t-13/14) nil (t-1/14) (t-2/14) (t-3/14) (t-4/14) (t-5/14) (t-6/14)
             #\# (t-8/14) (t-9/14) (t-10/14) (t-11/14) (t-12/14) (t-13/14) nil (t-1/14) (t-2/14) (t-3/14) (t-4/14) (t-5/14) (t-6/14))
      :lines-list
      (list  0 0 0 0 0 0 0 0 0 0 0 0 0 0
             1 1 1 1 1 1 1 1 1 1 1 1 1 1
             2 2 2 2 2 2 2 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 6 6 6 6 6 6 6)
      :approx-factor 200/14))

;-------16
(defvar *16-tone-chromatic-scale*
    (make-instance 'scale
      :alteration-list
      (list  nil (t-1/16) (t-1/8) (t-3/16) (t-1/4) (t-5/16) (t-3/8) (t-7/16) (t-1/2) (t-9/16) (t-5/8) (t-11/16) (t-3/4) (t-13/16) (t-7/8) (t-15/16)
             nil (t-1/16) (t-1/8) (t-3/16) (t-1/4) (t-5/16) (t-3/8) (t-7/16) (t-1/2) (t-9/16) (t-5/8) (t-11/16) (t-3/4) (t-13/16) (t-7/8) (t-15/16)
             nil (t-1/16) (t-1/8) (t-3/16) (t-1/4) (t-5/16) (t-3/8) (t-7/16)
             nil (t-1/16) (t-1/8) (t-3/16) (t-1/4) (t-5/16) (t-3/8) (t-7/16) (t-1/2) (t-9/16) (t-5/8) (t-11/16) (t-3/4) (t-13/16) (t-7/8) (t-15/16)
             nil (t-1/16) (t-1/8) (t-3/16) (t-1/4) (t-5/16) (t-3/8) (t-7/16) (t-1/2) (t-9/16) (t-5/8) (t-11/16) (t-3/4) (t-13/16) (t-7/8) (t-15/16)
             nil (t-1/16) (t-1/8) (t-3/16) (t-1/4) (t-5/16) (t-3/8) (t-7/16) (t-1/2) (t-9/16) (t-5/8) (t-11/16) (t-3/4) (t-13/16) (t-7/8) (t-15/16)
             nil (t-1/16) (t-1/8) (t-3/16) (t-1/4) (t-5/16) (t-3/8) (t-7/16) )
      :lines-list
      (list  0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
             1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1
             2 2 2 2 2 2 2 2 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 6 6 6 6 6 6 6 6)
      :approx-factor 200/16))



(defvar *current-1/2-scale*  *2-tone-chromatic-scale*)
(defvar *current-1/4-scale* *4-tone-chromatic-scale*)
(defvar *current-1/8-scale* *8-tone-chromatic-scale*)

(setf *current-1/2-scale*  *2-tone-chromatic-scale*)
(setf *current-1/4-scale* *4-tone-chromatic-scale*)
(setf *current-1/8-scale* *8-tone-chromatic-scale*)



(defun get-current-scale (tone)
    (case tone
      (1   *1-tone-chromatic-scale*  )
      (1.0 *1#-tone-chromatic-scale*)
      (2   *current-1/2-scale*)
      (3   *3-tone-chromatic-scale*  )
      (3.0 *3#-tone-chromatic-scale*)
      (4   *current-1/4-scale*)
      (5   *5-tone-chromatic-scale*  )
      (5.0 *5#-tone-chromatic-scale*)
      (6   *6-tone-chromatic-scale*  )
      (7   *7-tone-chromatic-scale*  )
      (7.0 *7#-tone-chromatic-scale*)
      (8   *current-1/8-scale*)
      (10   *10-tone-chromatic-scale*  )
      (12   *12-tone-chromatic-scale*  )
      (14   *14-tone-chromatic-scale*  )
      (16   *16-tone-chromatic-scale*)
      (otherwise *current-1/2-scale*)))

(defun default-scale-p (scale)
    (or (equal *current-1/2-scale* scale)
        (equal *current-1/4-scale* scale)
        (equal *current-1/8-scale* scale)))

(defun approx2tone (app)
    (case app (25 8) (50 4) (100 2) (otherwise (* 2 (/ 100 app)))))

(defun get-new-scale (tone)
    (omng-copy (get-current-scale tone)))

(defmethod get-super-default-value ((self (eql 'scale)))
    (omng-copy (get-current-scale 2)))

(defmethod  omng-copy ((self scale))
    (make-instance 'scale
      :alteration-list (alteration-list self)
      :lines-list   (lines-list self)
      :approx-factor   (approx-factor self)))



;=======================================================
;Dynamic table
;======================================================
(defvar *cur-dynamic-list* '(-1 20 40 55 60 85 100 115 127))
(defvar *cur-dynamic-chars* 
    (list (dyn-ppp) (dyn-pp) (dyn-p) (dyn-mp) (dyn-mf) (dyn-f) (dyn-ff) (dyn-fff)))

(omg-defclass dynamic-win (om-window) 
  ((object :initform nil :initarg :object :accessor object)))



(defmethod om-draw-contents ((self dynamic-win))
    (call-next-method)
   (om-with-focused-view self
      (om-with-font *om-default-font2*
      (loop for dyn in (butlast (object self))
            for i = 0 then (+ 1 i) do
            (om-draw-string  60 (+ 28 (* i 30)) (string+ "from " (format () "~D" (+ dyn 1)) " to ")))))
   (om-with-focused-view self
        (om-with-font (om-get-font self)
                                  (loop for i from  0 to (- (length *cur-dynamic-chars*) 1) do
                                            (om-draw-string  10 (+ 28 (* i 30)) (nth i *cur-dynamic-chars*))))))

(defun open-dyn-ed (list)
    (let ((win (om-make-window 'dynamic-win
                                                         :object list
                                                         :font (om-make-font *extras-font* 24)
                                                         :resizable nil :maximize nil
                                                         :window-title "Dynamic editor"
                                                         :size (om-make-point 200 260)
                                                         :position :centered)))
      (loop for diain in (cdr list)
            for i = 0 then (+ 1 i)  do
            (om-add-subviews win
                          (om-make-dialog-item 'numbox (om-make-point 140 (+ 22 (* i 30))) (om-make-point 20 14) (format () "~D" diain)
                                                                    :value diain
                                                                    :min-val 1
                                                                    :max-val 127
                                                                    :name i
                                                                    :afterfun #'(lambda (item)
                                                                                          (change-in-dyn item))
                                                                    :font *om-default-font1*)))
      win))


(defun change-in-dyn (item)
   (let ((newnum (value item))
         (i (+ 1 (om-get-view-name item))))
     (if (and (integerp newnum) (< i (- (length *cur-dynamic-list*) 1))
              (> newnum (nth (- i 1) *cur-dynamic-list*))
              (< newnum (nth (+ i 1) *cur-dynamic-list*)))
       (progn
         (setf (nth i *cur-dynamic-list*) newnum)
         (om-invalidate-view (om-view-window item) t))
       (progn
         (set-value item (nth i *cur-dynamic-list*))
         (om-beep)))))

(defun get-dyn-from-vel (vel)
    (if (<= vel 0) (car *cur-dynamic-chars*)
        (let (rep)
          (loop for item in *cur-dynamic-list*
                for i = 0 then (+ i 1)
                while (not rep) do
                (when (> item vel)
                  (setf rep (nth (- i 1) *cur-dynamic-chars*))))
          (if rep rep (car (last *cur-dynamic-chars*))))))



;;;;=================================



