(in-package :om)

(defmethod get-connection-lines ((self boxframe) i)
  (let* ((ctrl (nth i (inputframes self)))
         (connection (connected? (object ctrl)))
         (boxsource (first connection))
         (possource (get-connect-pos-source (car (frames boxsource))))
         (sizesource (om-view-size (car (frames boxsource))))
         (x-self (x self))
         (y-self (y self))
         (in-xs (x ctrl))
         x1 y1 xi yi midx midy)
    (if (scorepanel-p (om-view-container self))
      (progn
        (setq x1 (round (+ (om-point-h possource)  
                           (- (* (+ (second connection) 1) (/ (om-point-h sizesource) (+ (numouts boxsource) 1))) 2))))
        (setq y1 (- (+ (om-point-v possource) (h (car (frames boxsource)))) 2))
        (setq xi (+ x-self in-xs 4))
        (setq yi  y-self)
        (list (om-make-point x1 y1) (om-make-point xi yi))
        )
      (progn                
        (setq x1 (+ (om-point-h possource)
                    (- (* (+ (second connection) 1) (round (om-point-h sizesource) (+ (numouts boxsource) 1))) 2)))
        (setq y1 (- (+ (om-point-v possource) (om-point-v sizesource)) 2)) 
        (setq xi (+ x-self in-xs 4))
        (setq yi  y-self)
        (setf midx (round (abs (- xi x1)) 2))
        (setf midy (round (abs (- yi y1)) 2))
        (cond
         ((< y1 yi)
          (list (om-make-point x1  y1)
                (om-make-point x1 (+ y1 midy))
                (om-make-point xi (+ y1 midy))
                (om-make-point xi yi)))
         ((>= x1 xi) 
          (list (om-make-point x1  y1)
                (om-make-point (- x1 midx) y1)
                (om-make-point (- x1 midx) (- yi 5))
                (om-make-point xi (- yi 5))
                (om-make-point xi yi)))
         (t
          (list (om-make-point x1  y1)
                (om-make-point (+ x1 midx) y1)
                (om-make-point (+ x1 midx) (- yi 5))
                (om-make-point xi (- yi 5))
                (om-make-point xi yi))))))))



;--------------------------------------------------
;BOXes in the score
;-------------------------------------------------- 

(defclas score-box (omboxcall) 
   ((connected? :initform nil
)))

(defmethod numouts ((self score-box)) 1)
(defmethod get-documentation ((self score-box)) "musical box")

(defun make-score-box (musobj class posi name size)
   (let* (rep)
     (setf rep (make-instance class
                 :name name
                 :reference musobj))
     (setf (frame-position rep) posi)
     (setf (frame-size rep) size)
     (setf (name rep) name) 
     rep))

(defmethod make-frame-from-callobj ((self score-box))
   (let* ((module (om-make-view (get-frame-class self) 
                    :position (frame-position self)
                    :help-spec (get-documentation self)
                    :size  (frame-size self)
                    :object self))
          (input (car (inputs self)))
          thenewout inputf)
     (setf thenewout (om-make-view (get-out-class self)
                       :position (om-make-point (- (round (om-point-h (frame-size self)) 2) 4) 
                                                  (- (h module) 8))
                       :size (om-make-point 8 8)
                       :help-spec "option-click to evalue or drag for connections"
                       :index 0))
     (push thenewout (outframes module))
     (om-add-subviews module thenewout )
     (setf (name module) (frame-name self))
     (setf (frames self) (list module))
     module))

(defmethod OpenEditorframe ((self score-box)) nil)

(defmethod* omNG-box-value ((self score-box) &optional (numout 0))
   (reference self))
                            
(defmethod change-music-reference ((self t) (new t)) 
   (om-beep-msg "Type mistmach!") (abort) )

(defmethod get-frame-class ((self score-box)) 'scoreboxframe)


;================================================
(omg-defclass scoreboxframe (boxframe) ())

(defmethod frame-in-score-p ((self scoreboxframe)) t)

(defmethod centre-icon  ((self scoreboxframe)) t) 

(defmethod draw-before-box ((self scoreboxframe))
  (box-revise-references self (om-view-container self))
  (om-with-focused-view self
     (om-with-fg-color nil *scorepatch-color*
       (if (active-mode self) 
         (om-with-pen (self 
                       ;:pattern *om-gray-pattern* 
                       :mode :xor)
           (om-fill-rect 0 0 (w self) (h self)))
       (om-draw-rect 0 0 (w self) (h self))))))


(defmethod omG-select ((self scoreboxframe))
   "Set the frame SELF and the object pointed for frame in selected mode"
   (when (not (active-mode self)) 
     (setf (active-mode self) t)
     (om-invalidate-view self t)))

(defmethod omG-unselect ((self scoreboxframe))
   "Set the frame SELF and the object pointed for frame in unselected mode"
   (when (active-mode self)
     (setf (active-mode self) nil)
     (om-invalidate-view self t)))


(defmethod add-lock-button ((self scoreboxframe) &optional (icon 167))
   t)

(defmethod remove-lock-button ((self scoreboxframe))
  t)





;========================

;CONTAINERS


(defmethod (setf mus-patch) ((val t) (self t)) t)
(defmethod mus-patch ((self t)) nil)

(defmethod remake-references ((self simple-container))
   (when (mus-patch self)
     (loop for item in (boxes (mus-patch self)) do
           (remake-reference item self))))

(defmethod remake-references ((self t)) t)

(defmethod remake-reference ((self t) obj) t)

(defmethod remake-reference ((self score-box) obj)
   (setf (reference self) (get-obj-from-container-path obj (reference self))))

;========================

;BOX IN SCORE


(defmethod save-reference ((self score-box))
   (cons-container-path (reference self)))


(defmethod omNG-save ((self score-box) &optional (values? nil))
   "Save a box"
   (let* ((inputs (mapcar #'(lambda (input) (omNG-save input values?)) (inputs self))))
     `(om-load-score-box ',(type-of self) ,(name self) ',(save-reference self) ',inputs ,(om-save-point (frame-position self)) 
                      ,(om-save-point (frame-size self))  ,(frame-name self))))

(defun om-load-score-box (class name reference inputs position size fname)
   (let* ((rep (make-instance class
                 :name name
                 :reference (reverse reference))))
     (setf (frame-position rep) (om-correct-point position))
     (setf (frame-size rep) (om-correct-point size))
     (setf (frame-name rep) fname)
     (setf (name rep) name)
     (setf (inputs rep) (mapcar #'(lambda (input) (eval input)) inputs))
     rep))



;========================

;CONSTR PATCH
(defmethod om-save ((self null) &optional (values? nil)) nil)





;======================================
;MUSBOXES IN THE SCORE
;======================================

(defmethod initialize-instance :after ((self scorePanel) &rest l ) 
   (declare (ignore l))
   (init-music-patch self))

(defmethod init-music-patch ((self scorePanel))
   (let ((object (object (editor self))))
     (unless (mus-patch object)
       (setf (mus-patch object) (make-instance 'OMPatchAbs :name "patch" :icon 210) ))
     (setf (object self) (mus-patch object))))

;=================

(defmethod change-size-and-pos ((self scorePanel))
   (mapc #'(lambda (box)
             (change-size-and-pos box)) (get-subframes self)))

(defmethod change-size-and-pos ((self t)) t)
(defmethod change-size-and-pos ((self scoreboxframe))
   (let* ((ref (reference (object self)))
          (grap-container (graphic-obj (om-view-container self)))
          (grap-ref (get-correspond-grap grap-container ref))
          (size (round (staff-size (om-view-container self)) 8))
          rec)
     (when grap-ref
       (setf rec (rectangle grap-ref))
       (om-set-view-position self (om-make-point (- (first rec) size) (- (second rec) size 8)))
       (om-set-view-size  self (om-make-point (+ (* 2 size) (- (third rec) (first rec))) 
                                       (+  16 (* 2 size) (- (fourth rec) (second rec))))))))

;--------------

(defmethod revise-references ((self scorePanel))
   (mapc #'(lambda (box) (box-revise-references box self)) (get-subframes self))
   (make-move-after self (get-subframes self)))

(defmethod box-revise-references ((self t) panel) t)
(defmethod box-revise-references ((self scoreboxframe) panel)
   (let* ((object (object (om-view-container panel)))
          (pere (get-root-parent object))
          (newref (get-obj-from-container-path pere (reverse (cons-container-path  (reference (object self)))))))
     (if newref
       (progn
         (setf (reference (object self)) newref)
         (change-size-and-pos self))
       (omg-remove-element panel self))))

;=================

(defmethod set-panel-boxes ((self scorePanel))
   (let ((elements (boxes (object self)))) 
     (mapc #'(lambda (elem)
               (let ((newframe (make-frame-from-callobj elem)))
                 (om-add-subviews self newframe)
                   (add-subview-extra newframe))) elements)
     (mapc #'(lambda (elem)
               (update-graphic-connections elem elements)) (get-subframes self))))


(defmethod remove-panel-boxes ((self scorePanel))
   (mapc #'(lambda (elem)
             (om-remove-subviews self elem)) (om-subviews self))
   (om-invalidate-view self t))

(defmethod scorepanel-p ((self scorePanel)) t)
(defmethod scorepanel-p ((self t)) nil)

(defmethod modify-patch ((self scorePanel)) t)


(defmethod clear-ev-once :after ((self scorePanel))
   "After one evaluation this methods set the ev-once flag of all boxes in ev-once mode to nil."
   (update-panel self t))

;
;=====================================
;CONSTRAINTS

(defmethod get-posi-score-box (item size)
   (om-make-point (- (first (rectangle item)) size) 
               (+ (- (second (rectangle item)) size) 8)))

(defmethod get-size-score-box (item size)
   (om-make-point (+ (* 2 size) (- (third (rectangle item)) (first (rectangle item))))
               (+ 8 (* 2 size) (- (fourth (rectangle item)) (second (rectangle item))))))

(defmethod mk-musobj-box ((self scorePanel))
   (let* ((mode-obj (grap-class-from-type  (obj-mode self)))
          (size (round (staff-size self) 8)))
     (list mode-obj (selection? self))
     (loop for item in (get-graph-selection? self mode-obj) do
           (let* ((newconst (make-score-box  (reference item) 'score-box 
                                             (get-posi-score-box item size)
                                             "musical box"
                                             (get-size-score-box item size)))
                  (frame (make-frame-from-callobj newconst)))
             (setf (selection? self) (remove  (reference item) (selection? self) :test 'equal))
             (setf (selected item) nil)
             (omG-add-element self frame)))))


(defun non-connected-box (self list)
   (let ((rep t))
     (loop for item in list 
           while list do
           (when (is-connected? self item)
             (setf rep nil)
             (setf list nil))) rep))

(defun get-score-to-value (list)
   (let (rep)
     (mapc #'(lambda (box)
               (when (non-connected-box box list)
                 (push box rep))) list) rep))

(defmethod patch-clear-ev-once ((self ompatch))
   (mapc #'(lambda (box)
             (clear-ev-once box)) (boxes self)))

(defmethod get-all-score-patches ((self simple-container))
  (when (mus-patch self)
    (list (mus-patch self))))

(defmethod get-all-score-patches ((self container))
  (append (loop for item in (inside self)
                append (get-all-score-patches item))
          (call-next-method)))

(defmethod eval-score ((self scorePanel))
   "Eval all boxes in 'self'."
   (let ((allpatches (get-all-score-patches (object (om-view-container self)))))
     (loop for patch in allpatches do
           (loop for item in (get-score-to-value (boxes patch)) do
                 (omNG-box-value item 0))
           (patch-clear-ev-once patch))
     (update-panel self t)))



;;=========================================
;; IN / OUTS


(defclass scorein (maq-omin) ())
(defclass scoreout (maq-omout) ())

(defmethod get-new-box-from-type ((type (eql 'in)) position (container scorepanel))
  (let* ((boxes (get-subframes container)) 
         (i (length (find-class-boxes boxes 'maqinFrame)))
         (newin (make-new-patch-input (mk-unique-name container "input") i position 165 'scorein)))
    (setf (offset newin) (pixel-toms container position))
    newin
    ))

(defmethod get-new-box-from-type ((type (eql 'out)) position container)
  (let* ((boxes (get-subframes container)) 
         (i (length (find-class-boxes boxes 'maqoutFrame)))
         (newout (make-new-output (mk-unique-name container "output") i position 165 'scoreout)))
    
    newout
    ))


(defmethod pixel2point ((self scorepanel) pixel)
   (om-make-point (pixel-toms self pixel) (om-point-v pixel)))

(defmethod get-system-etat ((self scorepanel)) nil)

(defmethod point2pixel ((self scorepanel) point sys-etat)
  (om-make-point (ms2pixel (om-point-h point)  (/ (staff-size self) 4) (staff-zoom self)) (om-point-v point))) 

(defmethod get-offset/posy-from-pixel ((container scorepanel) pointpixel)
  (let* ((posx (pixel-toms container pointpixel))
         (posy (om-point-v pointpixel)))
    (om-make-big-point posx posy)))


(defmethod get-offset/posy-in-pixel (tempobj (container scorepanel))
  (om-view-position tempobj))


