(in-package :om)

;****************************************************************************************************
;****************************************************************************************************
;DRAWING IN LINEAR MODE
;****************************************************************************************************
;****************************************************************************************************
(setf *om-tonalite* nil)
;============sheet============

(defclas grap-sheet (grap-poly) 
  ((page-list :initform nil)
   (timebpf :initform nil)))


;-----------BUILD GRAPHIC OBJ
(defmethod make-graph-form-obj ((self sheet)   x top linespace mode scale sel system stem)
  (let ((graph-obj (make-graph-ryth-obj  self top system linespace  scale sel nil nil nil))
        (size  (* linespace 4)))
    (space-objects (get-temporal-objects graph-obj) (* 4 linespace))
    (set-graph-rectangles graph-obj)
    graph-obj))

(defmethod make-pages-form-obj ((score sheetPanel) (self sheet)  x top linespace mode scale sel system stem)
  (let ((graph-obj (make-graph-ryth-obj self top system linespace  scale sel nil nil nil))
        (size  (* linespace 4)))
    (space-objects (get-temporal-objects graph-obj) (* 4 linespace))
    (set-graph-rectangles graph-obj)
    (cons-score-pages-list score self graph-obj system (* linespace 4))
    (make-score-pages-picts  score self graph-obj (+ x (* size (score-left-margin score))) (* (* linespace 4) (score-top-margin score)) 1  0 
                             (* linespace 4)  system nil)
    graph-obj))



(defmethod make-graph-ryth-obj ((self sheet) top staffsys linespace  scale sel pere durtot &optional ryth)
  (declare (ignore ryth durtot))
  (let* ((thesheet (make-instance 'grap-sheet :reference self :parent pere))
         (voicelist (loop for item in (inside self)
                          for i = 0 then (+ i 1) collect
                          (make-graph-ryth-obj  item (top-in-midi (nth i staffsys))  (nth i staffsys) linespace  scale sel thesheet nil))))
    (setf (inside thesheet) voicelist)
    (make-graphic-extras thesheet)
    thesheet))


;il faut le faire pour le chord-seq
(defmethod make-graph-ryth-obj ((self chord-seq) top staffsys linespace  scale sel pere durtot &optional ryth)
  (declare (ignore ryth))
  (let* ((newc-s (make-instance 'grap-chord-seq
                   :reference self
                   :parent pere))
         (inside (loop for item in (inside self) 
                       for i = 0 then (+ i 1)
                       collect 
                       (make-graph-ryth-prop-obj item  top staffsys linespace  scale sel newc-s))))
    (setf (inside newc-s) inside)
    (set-parent newc-s inside)
    (make-graphic-extras newc-s)
    newc-s))



;-----------DRAW GRAPHIC OBJ

(defun make-bpf-time (list)
  (let (lx ly)
    (setf list (sort (remove-duplicates list :key 'first) '< :key 'first))
    (loop for item in list do
          (push (first item) lx)
          (push (second item) ly))
    (simple-bpf-from-list (reverse lx) (reverse ly))))
  
(defmethod draw-object ((self grap-sheet) view x y zoom minx maxx miny maxy slot size modepage? staff grille-p chnote)
  (let* ((sheet (reference self))
         (timebpf (make-bpf-time (collect-bpftime-objects self 0 size)))
         (meas-list (cdr (get-aligne-measures sheet)))
         (posy y) 
         (positions nil))
    (setf (timebpf self) timebpf)
    (loop for item in (inside self)
          for i = 0 then (+ i 1) 
          for system in staff do
          (draw-object item view (if (grap-voice-p item) x (+ x size)) (- posy (round (* (posy (car (staff-list system))) (/ size 4)))) zoom minx maxx miny maxy slot size t (nth i staff) grille-p chnote)
          (setf posy (+ posy (get-delta-system system size view i)))
          (push posy positions))
    (setf positions (reverse positions))
    (collect-rectangles self)
    (draw-aligned-measures self meas-list staff size  positions)
    (draw-extras self view size staff)))



;=============track============


(defclas grap-track (grap-container) ())

(defclas grap-track-ev (simple-graph-container) 
   ((index :initform nil)
    (start? :initform nil)
    (segmentations :initform nil)
    (internal-rects :initform nil)))  ;a tenir en compte

;-----------
(defmethod make-graph-ryth-obj ((self track) top staffsys linespace  scale sel pere durtot &optional ryth)
  (declare (ignore ryth))
  (let* ((newtrack (make-instance 'grap-track
                   :reference self
                    :rectangle (list 0 0 0 0)
                   :parent pere))
         (inside (loop for item in (get-track-ev self) 
                       for i = 0 then (+ i 1)
                       append 
                       (make-graph-ryth-track-obj item  i top staffsys linespace  scale sel newtrack))))
    (setf (inside newtrack) inside)
    (make-graphic-extras newtrack) 
    newtrack))

(defmethod make-graph-ryth-track-obj ((self track-ev) i top staffsys linespace scale sel pere)
  (let* ((new-ev-s (make-instance 'grap-track-ev
                     :pict (get-obj-pict (obj self))
                     :reference self
                     :parent pere
                     :index i
                     :main-point (list 0 0)
                     :start? t
                     :rectangle (list 0 0 1 1)
                     :selected (member self sel :test 'equal)))
         (new-ev-e (make-instance 'grap-track-ev
                     :reference self
                     :index i
                     :start? nil
                     :main-point (list 0 0)
                     :parent pere
                     :rectangle (list 0 0 1 1)
                     :selected (member self sel :test 'equal))))
    (list new-ev-s new-ev-e)))


(defmethod draw-rectangle ((self grap-track) system size &optional fill )
   (when (rectangle self)
     (let* ((rec (rectangle (car (inside self))))
            (rec (list (- (car rec) 4) (- (second rec) 4) (+ (third rec) 4) (+ (fourth rec) 4))))
       (if fill
           (om-draw-hilite-rect (car rec) (second rec) (- (third rec) (car rec) ) (- (fourth rec) (second rec)) t)
         (om-with-fg-color nil *om-select-color*
                                          (om-draw-rect (car rec) (second rec) (- (third rec) (car rec) ) (- (fourth rec) (second rec))))
         ))))




(defmethod draw-object ((self grap-track) view x y zoom minx maxx miny maxy slot size modepage? staff grille-p chnote)
  (setf y (+ y (* 3 size)))
  (let ((sys-size (get-system-size staff size)))
    (loop for item in (get-track-ev (reference self))
          for grap-item in (inside self)
          for i = 0 then (+ i 1) do
          
          (let ((start (find-if #'(lambda (x) (and (= (index x) i) (start? x)))  (inside self)))
                (end (find-if #'(lambda (x) (and (= (index x) i) (not (start? x))))  (inside self))))
            (draw-track-event  (obj item) grap-item x 
                              (round (+ x (round size  3.5) (* zoom (car (main-point start))))) 
                              (round (+ x (round size  3.5) (* zoom (car (main-point end)))))
                              y  (- (+ y sys-size) 3) view zoom size)
            (setf (rectangle grap-item) 
                  (list (round (+ x (round size  3.5) (* zoom (car (main-point start))))) 
                        y 
                        (round (+ x (round size  3.5) (* zoom (car (main-point end))))) 
                        (- (+ y sys-size) 3)))))))


  
(defmethod draw-track-event ((self t) gobj deltax x0 x1 y0 y1 view zoom size sheetpanel)
  (draw-obj-in-rect self x0 x1 y0 y1 nil view))

;;(defun convert-bpf-ime (bpf time deltax zoom size)
;;  (round (+ deltax (round size 3.5) (* zoom (bpf-get-val bpf time )))))

(defun soundsec2pix (val size dur)
  (round (* val size) dur))

;--special cases

;------BPF
(defmethod draw-track-event ((bpf bpf) gobj deltax x w y h view zoom size sheetpanel)
  (om-with-fg-color view (bpfcolor bpf)
    (let ((ranges (give-bpf-range bpf))
          (pixels w)
          (points (length (point-list bpf))))
      (loop for point-list on (point-list bpf) do
            (let* ((thepoint (car point-list))
                   (pix-point (om-make-point (- (get-x-pos sheetpanel (+ (start-t (reference (reference gobj))) (om-point-h thepoint))) (sheet-offset view))
                                             (om-point-v (point-to-pixel-with-sizes ranges thepoint w h))))
                   (points-prev-next (give-prev+next-x bpf thepoint))
                   (prev-point (first points-prev-next))
                   (next-point (second points-prev-next)))
                   (when prev-point
                     (let ((prev-pixel (om-make-point (- (get-x-pos sheetpanel (+ (start-t (reference (reference gobj))) (om-point-h prev-point))) (sheet-offset view))
                                                      (om-point-v (point-to-pixel-with-sizes ranges prev-point w h)))))
                       (om-draw-line (+ x (om-point-h prev-pixel)) (+ y (om-point-v prev-pixel))
                                     (+ x (om-point-h pix-point)) (+ y (om-point-v pix-point)))))
                   (when (or (and next-point (null (second point-list)))
                             (and (second point-list) 
                                  (not (equal next-point (second point-list)))))
                     (let ((next-pixel (om-make-point (- (get-x-pos sheetpanel (+ (start-t (reference (reference gobj))) (om-point-h next-point))) (sheet-offset view))
                                                      (om-point-v (point-to-pixel-with-sizes ranges next-point w h)))))
                       (om-draw-line (+ x (om-point-h pix-point)) (+ y (om-point-v pix-point))
                                     (+ x (om-point-h next-pixel)) (+ y (om-point-v next-pixel))))))))))



                 
;------SOUND
(defmethod draw-track-event ((self sound) gobj deltax x0 x1 y0 y1 view zoom size sheetpanel)
  (when (pict-sound self)
    (let* ((bpftime (timebpf sheetpanel))
           (dur (/ (om-sound-n-samples  self) (om-sound-sample-rate  self)))
           (sizebox (- x1 x0))
           (picture (pic-to-draw self))
           (sizepict (om-pict-width picture))
           (hpict (om-pict-height picture))
           (start 0)
           (startpict 0))
      (if (markers self)
        (progn
          (loop for item in (markers self) do
                (let ((end (- (get-x-pos sheetpanel (round (+ (start-t (reference (reference gobj))) (* 1000 item)))) (sheet-offset view) x0))
                      (endpict (soundsec2pix item sizepict dur)))
                  (om-draw-picture-part (om-get-current-port) picture 
                                        (om-make-point startpict 0) (om-make-point (- endpict startpict) hpict)
                                        (om-make-point (+ x0 start) y0) (om-make-point (- end start) (- y1 y0)) )
                  (setf start end)
                  (setf startpict endpict)))          
          (loop for item in (markers self) do
                (let ((end (- (get-x-pos sheetpanel (round (+ (start-t (reference (reference gobj))) (* 1000 item)))) (sheet-offset view) x0)))
                  (om-with-fg-color nil *om-red-color*
                    (om-draw-line (+ x0 end) y0 (+ x0 end) y1))))
          (om-draw-picture-part (om-get-current-port) picture 
                                (om-make-point startpict 0) (om-make-point (- sizepict startpict) hpict)
                                (om-make-point (+ x0 start) y0) 
                                (om-make-point (- x1 (+ x0 start)) (- y1 y0)) ))
        (om-draw-picture (om-get-current-port) picture (om-make-point x0 y0) (om-make-point (- x1 x0) (- y1 y0)))))))

;------ MIDI FILE 
(defmethod draw-track-event ((self midifile) gobj deltax x0 x1 y0 y1 view zoom size sheetpanel)
  (let ((bpftime (timebpf sheetpanel)))
    (loop for track in (tracks self)
          for i = 0 then (+ i 1) do
          (om-with-fg-color nil (nth (mod i 15) *16-color-list*)
            (let* ((minx 0)
                   (maxx (real-dur self))
                   (miny 24)
                   (maxy 96)
                   (x-notes (give-notes-in-x-range track minx maxx))
                   (notes (sort (give-notes-iny-range x-notes miny maxy) '< :key 'second))
                   (ysize (round (- y1 y0) (- maxy miny)))
                   (ranges (list minx maxx miny maxy)))
              (loop for note in notes do
                    (let* ((topleft (point-to-pixel-with-sizes ranges (om-make-big-point (second note) (first note)) (- x1 x0) (- y1 y0)))
                           (realleft (- (get-x-pos sheetpanel (+ (start-t (reference (reference gobj))) (second note))) (sheet-offset view) x0))
                           (realright (- (get-x-pos sheetpanel (+ (start-t (reference (reference gobj))) (+ (second note) (third note)))) (sheet-offset view) x0)))
                      (om-fill-rect realleft (+ y0 (om-point-v topleft)) (- realright realleft) 1))))))))

;------MAQUETTE
(defmethod draw-track-event ((self ommaquette) gobj deltax x0 w y0 h view zoom size sheetpanel)
  (let* ((bpftime (timebpf sheetpanel))
         (durmaq (get-obj-dur self))
         (rangey (max-y-item self))
         (miny (second rangey))
         (maxy (first rangey))
         (rangey (- maxy miny))
         (delta (+ 5 (if (eval-func self) (round h 4) 0))))
    (om-with-fg-color nil (maq-color (params self))
      (om-fill-rect  x0  y0 w h )
      (when (boxes self)
        (loop for item in (boxes self) do
              (let ((x-0 (- (get-x-pos sheetpanel (+ (start-t (reference (reference gobj))) (slot-value item 'offset))) (sheet-offset view) x0))
                    (x-1 (- (get-x-pos sheetpanel (+ (start-t (reference (reference gobj))) (slot-value item 'extend) (slot-value item 'offset))) (sheet-offset view) x0))
                    (ry (- (- h delta) (round (* (- h delta) (- (posy item) miny)) rangey)))
                    (rh (round (* (- h delta) (sizey item)) rangey)))
                (om-with-fg-color nil (colorframe item)
                  (om-fill-rect x-0 ry (- x-1 x-0) rh))))
        (setf delta (- delta 10))
        (when (and (eval-func self) (value self))
          (om-with-fg-color nil *om-white-color*
            (om-fill-rect 0 (- h delta) w delta))
  
          ;(unless (maq-obj-p (value self))
          ;  (when (minipict view)
          ;    (om-draw-picture view (minipict view) (om-make-point 0 (- (h view) delta)) (om-make-point (w view) delta)))
          ;  )

          )))))

  
;----------------
(defmethod set-graph-rectangles ((self grap-track-ev) )
   (setf (rectangle self) (list (car (main-point self)) (second (main-point self)) 
                                (+ (car (main-point self)) (third (rectangle self))) 
                                (+ (second (main-point self)) (fourth (rectangle self))))) )
;----------------

(defmethod click-in-obj ((self grap-track-ev) type where)
   (let* ((rect (rectangle self)))
       (when (point-in-rectangle-p where (second rect) (first rect) (fourth rect) (third rect))
         self)))


;------

(defmethod collect-bpftime-objects ((self grap-container) at size)
   (loop for item in (inside self)
         append (collect-bpftime-objects item (+ at (offset->ms (reference item))) size)))

(defmethod collect-bpftime-objects ((self grap-rest) at size) 
   (list (list  at  ( + 0 (car (main-point self)) ))))


(defmethod collect-bpftime-objects ((self grap-group) at size)
   (loop for item in (inside self)
         append (collect-bpftime-objects item (+ at (offset->ms (reference item))) size)))

(defmethod collect-bpftime-objects ((self grap-ryth-chord) at size)
   (list (list  at  ( + 0 (car (main-point self)) ))))

(defmethod collect-bpftime-objects ((self grap-chord) at size)
   (list (list  at  ( + 0 (car (main-point self)) ))))

(defmethod collect-bpftime-objects ((self grap-track) at size)
   (loop for item in (inside self)
         append (collect-bpftime-objects item (+ at (if (start? item) (start-t (reference (reference item))) (end-t (reference (reference item))))) size)))

(defmethod collect-bpftime-objects ((self grap-track-ev) at size)
   (list (list  at  ( + 0 (car (main-point self)) ))))




;****************************************************************************************************
;****************************************************************************************************
;DRAWING IN PAGE MODE
;****************************************************************************************************
;****************************************************************************************************

(defclass page-pointer ()
  ((curpage :initform 0 :accessor curpage)
   (curline :initform 0 :accessor curline)
   (curelem :initform 0 :accessor curelem)
   (score :initform nil :accessor score :initarg :score)
   (voice :initform nil :accessor voice :initarg :voice)
   (view :initform nil :accessor view :initarg :view)
   (linesizex :initform nil :accessor linesizex :initarg :linesizex)
   (linesizey :initform nil :accessor linesizey :initarg :linesizey)))

  
;return the window
(defmethod move-page-pointer ((self page-pointer))
  (let* ((fdoc (score-fdoc (score self)))
         (curpage (nth (curpage self) (page-list fdoc)))
         (curline (nth (curline self) (line-list  curpage)))
         (hmlines (length (line-list curpage)))
         (hmelem (nth (voice self) (line-info curline))))
    (setf (curelem self) (+ (curelem self) 1))
    (if (= (curelem self) hmelem)
      (progn
        (setf (curelem self) 0)
        (setf (curline self) (+ (curline self) 1))
        (if (= (curline self) hmlines)
          (progn
            (print "change page")
            (setf (curline self) 0)
            (setf (curpage self) (+ (curpage self) 1))
            (setf (view self) (nth (curpage self) (score-picts-list (score self))))))))
    ))


(defmethod compute-delat-x-y ((self page-pointer))
  (let* ((score (score self))
         (fdoc (score-fdoc score))
         (linesizex  (linesizex self))
         (linesizey  (linesizey self))
         (repx 0) (repy 0))
    (loop for i from 0 to (- (curpage self) 1) do
          (setf repx (+ repx (* linesizex (length (line-list  (nth i (page-list fdoc))))))))
    (setf repx (+ repx (* linesizex (curline self))))
    (setf repy (* linesizey (curline self)))
    (list repx repy)))

;================================================================
;DRAWING SYSTEMS and STAFFS
;================================================================




  
;================================================================
;BREAK PAGE
;================================================================


(defmethod collect-objects-to-break ((self grap-sheet) n) 
  (loop for item in (inside self)
        for i = 0 then (+ i 1) 
        append (collect-objects-to-break item i)))


(defmethod collect-objects-to-break ((self grap-container) n)
  (loop for item in (inside self)
        append (collect-objects-to-break item n)))

(defmethod collect-objects-to-break ((self grap-rest) n) 
  (list (list n self (rectangle self))))

(defmethod collect-objects-to-break ((self grap-group) n) 
  (list (list n self (rectangle self))))

(defmethod collect-objects-to-break ((self grap-chord) n) 
  (list (list n self (rectangle self))))

(defmethod collect-objects-to-break ((self grap-track-ev) n) 
  (list (list n self (rectangle self))))

;------------------

(defmethod collect-onsets-to-break ((self grap-sheet) n) 
  (loop for item in (inside self)
        for i = 0 then (+ i 1) 
        append (collect-onsets-to-break item i)))

(defmethod collect-onsets-to-break ((self grap-container) n)
  (loop for item in (inside self)
        append (collect-onsets-to-break item n)))

(defmethod collect-onsets-to-break ((self grap-rest) n) 
  (list (list n (car (rectangle self)) (main-point self) (offset->ms (reference self)))))

(defmethod collect-onsets-to-break ((self grap-group) n) 
  (list (list n (car (rectangle self)) (main-point self) (offset->ms (reference self)))))

(defmethod collect-onsets-to-break ((self grap-chord) n) 
  (list (list n (car (rectangle self)) (main-point self) (offset->ms (reference self)))))

(defmethod collect-onsets-to-break ((self grap-track-ev) n)  
  (list (list n (car (rectangle self)) (main-point self) (if (start? self) (start-t (reference self)) (end-t (reference self))))))

;================================================================
;SPACING
;================================================================

(defmethod collect-temporal-objects ((self sheetpanel) at)
   (loop for item in (trackpanels  self)
         append (collect-temporal-objects item at)))

(defmethod collect-temporal-objects ((self trackpanel) at)
   (loop for item in (track-subviews self)
         append (collect-temporal-objects item at)))

;(defmethod collect-temporal-objects ((self sheet-objectframe) at)
;(list (list (+ at (start-t (reference self))) self))
;  (collect-temporal-objects (obj (reference self)) at)
;  )

(defmethod collect-temporal-objects ((self track-ev) at)
  (list (list (+ at (start-t self)) self)))

(defmethod collect-temporal-objects ((self grap-chord) at )
   (list (list  at  self )))

;(defmethod collect-temporal-objects ((self grap-track) at)
;   (loop for item in (inside self)
;         append (collect-temporal-objects item (+ at (if (start? item) (start-t (reference item)) (end-t (reference item)))))))

;(defmethod collect-temporal-objects ((self grap-track-ev) at )
;   (list (list  at  self)))


(defmethod collect-temporal-objects ((self grap-track) at)
   (loop for item in (inside self)
         append (collect-temporal-objects item (+ at (if (start? item) (start-t (reference (reference item))) (end-t (reference (reference item))))))))

(defmethod collect-temporal-objects ((self grap-track-ev) at )
   (list (list  at  self)))


;;;(defmethod measures-first ((a grap-ryth-chord) (b grap-chord))  t)

(defmethod space-size ((self grap-chord)) 
  (loop for item in (inside self) maximize  (third (rectangle item))))


;IL FAUT REVISER BIEN
;===============chord is i
;------last element
(defmethod space-girl ((obji+1 null) onseti+1 (obji grap-chord)  onseti  count size maxdur)
  (setf (main-point obji) (list (+ count  maxdur) (second (main-point obji))))
  (loop for item in (inside obji) do
        (setf (main-point item) (list (+ count (first (main-point item))  maxdur) (second (main-point item)))))
  (+ count  size ))

;------before a measure
(defmethod space-girl ((obji+1 grap-measure) (onseti+1 integer) (obji grap-chord) (onseti integer) count size maxdur) 
   (setf (main-point obji) (list (+ count  maxdur) (second (main-point obji))))
   (loop for item in (inside obji) do
         (setf (main-point item) (list (+ count (first (main-point item)) maxdur ) (second (main-point item))))
         )
   (if (zerop (- onseti+1 onseti))
     (+ count  (round size 2)) 
     (+ count (* maxdur 2)  (round (* size (ryhtm2pixels (- onseti+1 onseti)))) (round size 4))))

;------before a chord or a rest
(defmethod space-girl ((obji+1 t) (onseti+1 integer) (obji grap-chord) (onseti integer) count size maxdur) 
   (setf (main-point obji) (list (+ count  maxdur) (second (main-point obji))))
   (loop for item in (inside obji) do
         (setf (main-point item) (list (+ count (first (main-point item)) maxdur ) (second (main-point item)))))
   (if (zerop (- onseti+1 onseti))
     count
     (+ count (* maxdur 2) (round (* size (ryhtm2pixels (- onseti+1 onseti)))) (round size 4))))



;------last element
(defmethod space-girl ((obji+1 null) onseti+1 (obji grap-track-ev)  onseti  count size maxdur)
  (setf (main-point obji) (list (+ count  maxdur) (second (main-point obji))))
  (+ count  size ))

;------before a measure
(defmethod space-girl ((obji+1 grap-measure) (onseti+1 integer) (obji grap-track-ev) (onseti integer) count size maxdur) 
   (setf (main-point obji) (list (+ count  maxdur) (second (main-point obji))))
   (if (zerop (- onseti+1 onseti))
     (+ count  (round size 2)) 
     (+ count (* maxdur 2)  (round (* size (ryhtm2pixels (- onseti+1 onseti)))) (round size 4))))

;------before a chord or a rest
(defmethod space-girl ((obji+1 t) (onseti+1 integer) (obji grap-track-ev) (onseti integer) count size maxdur) 
   (setf (main-point obji) (list (+ count  maxdur) (second (main-point obji))))
   (if (zerop (- onseti+1 onseti))
     count
     (+ count (* maxdur 2) (round (* size (ryhtm2pixels (- onseti+1 onseti)))) (round size 4))))



;===========make a sheet

;==========

(defmethod make-graph-ryth-prop-obj ((self chord)  top staffsys linespace scale sel pere)
  (let* ((thenotes (sort (copy-list (inside self)) '< :key 'midic))
         (offsets (Loffset self))
         (zigzag-list (make-alt-zig-zag self scale))
         (note-head-list (make-chord-zig-zag self scale))
         (maxw 0)
         (grap-notes (loop for item in thenotes
                           for pos in note-head-list
                           for i = 0 then (+ i 1)
                           collect
                           (let (notegrap notew)
                             (setf notegrap (make-graph-ryth-prop-obj item  top staffsys linespace scale sel self))
                             (setf (delta-head notegrap) pos)
                             (when (natural-alt-char notegrap)
                               (setf (alteration notegrap) (correct-alteration notegrap (pop zigzag-list))))
                             (setf notew (round (* 2 (max (* linespace pos)
                                                          (* linespace (if (alteration notegrap) (* -1 (- (alteration notegrap) 1)) 0))))))
                             (setf (rectangle notegrap) (list 0 0 notew (round linespace)))
                             (setf (nth 0 (main-point notegrap)) (round (* linespace pos)))
                             (setf maxw (max maxw notew))
                             notegrap)))
         (newchord (make-instance 'grap-chord
                     :reference self
                     :parent pere
                     :rectangle (list 0 0 maxw 0)
                     :stem (round (* 3 linespace))
                     :selected (member self sel :test 'equal)
                     :inside grap-notes)) )
    (set-parent newchord grap-notes)
    (make-graphic-extras newchord)
    newchord))

(defmethod make-graph-ryth-prop-obj ((self note) top staffsys linespace scale sel  pere)
   (declare (ignore ryth))
   (let* ((ypos (get-graphic-pos self top linespace scale)) 
          thenote)
     (setf thenote
           (make-instance 'grap-note
             :reference self
             :parent pere
             :main-point (list 0 (- ypos (round linespace 2)))
             :selected (member self sel :test 'equal)
             :auxlines (get-aux-lines self staffsys top scale linespace ypos)))
     (make-graphic-extras thenote)
     thenote))







;==============

(defmethod get-aligne-measures ((self sheet))
   (let* ((ofsets (loop for item in (inside self)
                        when (voice-p item) collect (loop for mes in (inside item) collect (offset->ms mes))))
          (all (sort (remove-duplicates (flat ofsets) :test 'equal) '<)) rep)
     (loop for item in all do
           (let (resp )
             (loop for voice in ofsets
                   for i = 0 then (+ i 1) do
                   (let ((pos (position item voice)))
                     (when pos (push (list pos i) resp))))
              (push (reverse resp) rep)))
     (reverse rep)))


;a corriger
(defmethod draw-aligned-measures ((self grap-sheet) list staff size positions)
   t)

;---------------------------
;??????
;---------------------------

(defun make-pages-from-lines (list hpage line size score)
  (let* ((linesize (get-delta-line line size score ))
         (mod (max 1 (floor hpage linesize)))
         (rep (multiple-value-list (floor (length list) mod))))
    (if (zerop (second rep))
      (make-list (car rep) :initial-element (make-list mod :initial-element linesize))
    (append (make-list (car rep) :initial-element (make-list mod :initial-element linesize)) 
            (list (make-list (second rep) :initial-element linesize) )))))
    
       

(defun compute-line-info (list n)
  (loop for i from 0 to (- n 1)
        collect (count i list :key 'car)))
  
(defmethod cons-score-pages-list ((score sheetPanel) (self sheet) graph-obj line size)
  (let* ((onsets+voice (sort (collect-onsets-to-break graph-obj 0) '< :key 'second))
         (onsets (loop for item in onsets+voice collect (second item)))
         (wpage (score-widht score size))
         (hpage (score-height score size))
         (mixed-lines (make-page-break (x->dx onsets) wpage)) ; ojo los ultimos ceros deben pasar a la linea siguiente
         (pages (make-pages-from-lines mixed-lines hpage line size score))
         (fdoc (make-instance 'fdoc))
         (numvoices (length (inside self)))
         (linecount -1)
         )
    (loop for item in pages do
          (let ((newpage (make-instance 'fpage)))
            (loop for line in item do
                  (let ((newline (make-instance 'fline)))
                    (setf curline (nth (incf linecount) mixed-lines))
                    (setf (line-info newline) (compute-line-info (first-n onsets+voice (length curline)) numvoices))
                    (setf onsets+voice (nthcdr  (length curline) onsets+voice))
                    (push newline (line-list newpage))))
            (setf (line-list newpage) (reverse (line-list newpage)))
            (push newpage (page-list fdoc))))
    (setf (page-list fdoc) (reverse (page-list fdoc)))
    (score-fdoc score fdoc)))


(defun om-open-pict-window (size font)
  (let ((view (make-instance 'window 
                :window-show nil
                :window-title "jaja"
                :view-font (oa::font font)
                :view-size (oa::pt size))))
    (start-picture view)
    view))

(defun om-close-pict-window (view)
  (let ((thepict (get-picture view)))
     (window-close view)
     (oa::make-pict-gworld  thepict)))

(defmethod make-score-pages-picts ((score sheetPanel) self  grapobj x y zoom  slot size  staff chnote)
  (let ((newfont (om-make-font *heads-font* size))
        (page-size (score-paper-size score)))
    (score-picts-list score
                      (loop for i from 0 to (- (length (page-list (score-fdoc score))) 1) collect
                            (om-open-pict-window page-size newfont)))
    (loop for item in (score-picts-list score)
          for i = 0 then (+ i 1) do
          (om-with-focused-view item
            (om-with-fg-color nil *om-white-color* 
              (om-fill-rect 0 0 (om-point-h page-size) (om-point-v page-size))
              (draw-score-page  staff  (howmany-lines (score-fdoc score) i) score x y (score-widht score size) size nil nil))))
    (page-draw-object  grapobj score (car (score-picts-list score)) self x y zoom  slot size  staff  chnote)
    (score-picts-list score
                      (loop for item in  (score-picts-list score) collect
                            (om-close-pict-window item)))))




(defmethod page-draw-object ((self grap-track) score view sheet x y zoom  slot size  staff  chnote )
  (draw-object self view x y zoom 0 0 0 0 slot size t staff nil chnote))


         
(defmethod page-draw-object ((self grap-sheet) score view sheet x y zoom  slot size  staff  chnote )
  (let ((posy y)
        (timebpf (make-bpf-time (collect-bpftime-objects self 0 size))))
    (setf (timebpf self) timebpf)
    (om-with-focused-view view
      (loop for item in (inside self)
            for i = 0 then (+ i 1) 
            for system in staff do
            (page-draw-object item score view (make-instance 'page-pointer 
                                          :score score
                                          :view view
                                          :voice i
                                          :linesizex (score-widht  score size)
                                          :linesizey (get-delta-line staff size score )) 
                              x (- posy (round (* size (/ 23 8)))) zoom  slot size (nth i staff)  chnote )
            (setf posy (+ posy (get-delta-system system size score i))))
      (collect-rectangles self)
      ;(draw-aligned-measures self meas-list staff size y positions)
      (draw-extras self view size staff))))


(defmethod page-draw-object ((self grap-poly) score view pointer x y zoom  slot size  staff  chnote )
  (with-focused-view (view pointer)
    (let* ((def-prefer (third (copy-list (second modepage?))))
         (pos&size (get-staff-pos&size staff (fourth def-prefer)))
         (positions (cdr pos&size))
         (meas-list (cdr (get-aligne-measures (reference self))))
         )
    (loop for item in (inside self)
          for i = 0 then (+ i 1)
          for posy in positions do
          (page-draw-object item score view pointer x (+ (round (* size posy)) y) zoom  slot size (nth i staff)  chnote ))
    (collect-rectangles self)
    (draw-aligned-measures self meas-list staff size y positions)
    (draw-extras self view size staff)
    )))



(defmethod page-draw-object ((self grap-chord) score view pointer x y zoom  slot size  staff  chnote )
  (with-focused-view (view pointer)
    (let ((deltaxy (compute-delat-x-y pointer)))
      (setf x (- x (car deltaxy))
            y (+ y (second deltaxy)))
      (move-page-pointer pointer)
      (loop for item in (inside self) do
            (page-draw-object item score view pointer x y zoom  slot size  staff  chnote ))
      
      (when *om-tonalite*
        (draw-chiffrage self x y zoom size))
      ;(collect-rectangles self)
      (draw-extras self view size staff)
      )))


(defmethod page-draw-object ((self grap-container) score view pointer x y zoom slot size  staff  chnote )
  (with-focused-view (view pointer)
   (loop for item in (inside self) do
         (page-draw-object item score view pointer x y zoom  slot size  staff  chnote ))
   (draw-extras self view size staff)))

(defmethod page-draw-object ((self grap-chord-seq) score view pointer x y zoom  slot size  staff  chnote )
 (with-focused-view (view pointer)
    (loop for item in (inside self) do 
          (page-draw-object item score view pointer x y zoom  slot size staff  chnote )
              (when *om-tonalite*
                (draw-modulation self item previous x y zoom size)
                ))
              
    (collect-rectangles self)
    (draw-extras self view size staff)
    ))

(defmethod page-draw-object ((self grap-multiseq) score view pointer x y zoom  slot size  staff  chnote )
  (with-focused-view (view pointer)
(let* ((def-prefer (third (copy-list (second modepage?))))
         (pos&size (get-staff-pos&size staff (fourth def-prefer)))
         (positions (cdr pos&size)))
    (loop for item in (inside self)
          for i = 0 then (+ i 1)
          for posy in positions do
          (page-draw-object item score view pointer x (+ (round (* size posy)) y) zoom  slot size (nth i staff)  chnote ))
    (collect-rectangles self)
    (draw-extras self view size staff)
    )
  ))

(defmethod page-draw-object ((self grap-poly) score view pointer x y zoom  slot size  staff  chnote )
  (with-focused-view (view pointer)
    (let* ((def-prefer (third (copy-list (second modepage?))))
         (pos&size (get-staff-pos&size staff (fourth def-prefer)))
         (positions (cdr pos&size))
         (meas-list (cdr (get-aligne-measures (reference self))))
         )
    (loop for item in (inside self)
          for i = 0 then (+ i 1)
          for posy in positions do
          (page-draw-object item score view pointer x (+ (round (* size posy)) y) zoom  slot size (nth i staff)  chnote ))
    (collect-rectangles self)
    (draw-aligned-measures self meas-list staff size y positions)
    (draw-extras self view size staff)
    )))

(defmethod page-draw-object ((self grap-voice) score view pointer x y zoom  slot size  staff  chnote )
  (with-focused-view (view pointer)
    (loop for i from 0 to (- (length (inside self)) 1) 
        do (let ((cur-mes (nth i (inside self)))
                 (next-mes (nth (+ i 1) (inside self)))) 
             (page-draw-object-ryth cur-mes score view pointer x y zoom  slot size  staff chnote )))
  (collect-rectangles self)
  (remake-measures (inside self))
  (draw-extras self view size staff)
  ))

(defmethod page-draw-object ((self grap-note) score view pointer x y zoom  slot size  staff  chnote )
  (declare (ignore minx maxx miny maxy grille-p))
  (with-focused-view (view pointer)
    (let* ((realrealpos (+ 1 x (* (/ size 4) (delta-head self)) (* zoom (- (x self) (* (/ size 4) (delta-head self))))))
         (realpos (round realrealpos))
         (altpos (if (alteration self) 
                     (round (- (+ realrealpos (* (- (alteration self) 1) (/ size 4))) (* (/ size 4) (delta-head self))))
                   realpos))
         (extra-head (get-extra-head (reference self)))
         (str (or extra-head (headchar self)))
         (headsizex (get-name-size str (om-make-font *heads-font* size)))
         (note (reference self))
         (note-color (get-mus-color note))
         (altstr (string (alt-char self)))  
         tie)
    
    (om-with-fg-color nil (if chnote (nth (chan (reference self)) *16-color-list*) note-color)
      
                      (om-draw-string  realpos (+ y (y self)) str)
                      
                      (write-note-points self  (+ headsizex realpos) (+ y (y self)) size) 
                      
                      (when (alteration self)
                        (om-draw-string altpos (+ y (y self)) altstr))
                      
                      (setf (rectangle self) (list altpos (+ y (- (y self) (round size 8)))
                                                   (+ realpos (round size 3)) (+ y (round size 8) (y self))))
                      
                      ;si on veut voir le degre des notes
                      ;(when *om-tonalite*
                      ;  (draw-degre self realpos y size))
                      
                      (setf tie (tie (reference self))) 
                      (when tie
                        (let* ((direction (get-tie-direction self))
                               (toptie (if (string-equal direction "down") (+ y (y self)) (- (+ y (y self)) (round size 3))))
                               (bottie (if (string-equal direction "down") (+ y (y self) (round size 3)) (+ y (y self))))
                               othernote)
                          (when (or (equal tie 'begin) (equal tie 'continue))
                            (setf othernote (get-next-tied-noted self))
                            (when othernote
                              (draw-liason-begin (+ realpos (round size 3)) toptie 
                                                 (round (+ realpos (* zoom (abs (- (x othernote) (x self))))))
                                                 bottie direction)))
                          (when (or (equal tie 'end) (equal tie 'continue))
                            (setf othernote (get-last-tied-noted self))
                            (when othernote
                              (draw-liason-end (round (- (+ realpos (round size 3))  (* zoom (abs (- (x self) (x othernote))))))
                                               toptie realpos bottie direction)))))
                      )
    (when (auxlines self)
      (om-with-fg-color nil *system-color* 
                        (let ((dir (car (auxlines self)))
                              (topy (+ (- y (round size 8)) (second (auxlines self))))        
                              (limy (+ (- y (round size 8)) (third (auxlines self)))))
                          (if (equal dir 'dw)
                              (progn
                                (setf topy (+ topy (round size 4))) 
                                (loop while (<= topy limy) do
                                      (om-draw-line (- realpos (round size 8)) topy 
                                                    (+  headsizex  realpos) topy)
                                      (setf topy (+ topy (round size 4)))))
                            (progn
                              (setf topy (- topy (round size 4))) 
                              (loop while (>= topy limy) do
                                    (om-draw-line (- realpos (round size 8)) topy 
                                                  (+  headsizex  realpos) topy)
                                    (setf topy (- topy (round size 4)))))))))
    (write-note-slot self realpos (+ y (y self)) slot size zoom)
    (draw-extras self view size staff)
    )))


;r

(defmethod page-draw-object-ryth ((self grap-container) score view pointer x y zoom  slot size  staff chnote )
   (with-focused-view (view pointer)
     (loop for item in (inside self) do
         (page-draw-object-ryth item score view pointer x y zoom  slot size  staff chnote ))))

(defmethod page-draw-object-ryth ((self grap-group) score view pointer x y zoom  slot size   staff chnote )
  (with-focused-view (view pointer)
  (without-interrupts  
   (let (deltaxy ))
   (when (figure-? self)
     (setf deltaxy (compute-delat-x-y pointer))
     (setf x (- x (car deltaxy))
           y (+ y (second deltaxy)))
     (move-page-pointer pointer)
     )
     (loop for item in (inside self) do
           (page-draw-object-ryth item score view pointer x y zoom  slot size  staff chnote ))
     (collect-rectangles self))
   (om-with-fg-color nil (mus-color (reference self))
     (let ((dire (dirgroup self)))
       (when (figure-? self)
         (group-draw-stems self dire x y (rectangle self) zoom size)
         (draw-beams-note-in-group self dire x -1 (rectangle self) zoom size)
         (draw-num-denom-s self x size (rectangle self)))))
   (draw-extras self view size staff)))

(defmethod page-draw-object-ryth ((self grap-measure) score view pointer x y zoom  slot size   staff chnote )
  (with-focused-view (view pointer)
    (let ((space (get-chifrage-space self size))
          (deltachiff (round size 5))
          (poly-p (poly-p (GET-ROOT-PARENT (reference self)))))
      
      (let ((previous nil))
        (loop for item in (inside self) do
              (page-draw-object-ryth item score view pointer (+ x space) y zoom  slot size  staff chnote )
              ;;;=================
              (when *om-tonalite*
                (setf previous (draw-modulation (parent self) item previous x y zoom size)))
              ;;;=================
              ))
      (collect-rectangles self)
      (setf (nth 0 (rectangle self)) (- (nth 0 (rectangle self)) space))
      (om-with-fg-color nil *system-color*
        (loop for thestaff in (staff-list staff) do
              (let ((ys (+ y (line2pixel (posy thestaff) nil (/ size 4)))))
                (setf (nth 1 (rectangle self)) (min (nth 1 (rectangle self)) ys))
                (setf (nth 3 (rectangle self)) (max (nth 3 (rectangle self)) (+ ys (system-size-in-pix staff size))))
                (cond 
                 ((first-of-group? self)
                  (om-draw-string  (+ (car (rectangle self)) deltachiff) (+ (round size 4) ys) (format () "~D" (caar (tree (reference self)))))
                  (om-draw-string  (+ (car (rectangle self)) deltachiff) (+ ys (round size 4) (round size 2)) (format () "~D" (cadar (tree (reference self)))))
                  )
                 ((show-chifrage self) 
                  (om-draw-string  (+ (car (rectangle self)) deltachiff) (+ (round size 4) ys) (format () "~D" (caar (tree (reference self)))))
                  (om-draw-string  (+ (car (rectangle self)) deltachiff) (+ ys (round size 4) (round size 2)) (format () "~D" (cadar (tree (reference self)))))
                  )
                 (t (let ((previous-mes (nth (- (position self (inside (parent self)) :test 'equal) 1) (inside (parent self)))))
                      (unless (equal (metric self) (metric previous-mes))
                        (om-draw-string  (+ (car (rectangle self)) deltachiff) (+ (round size 4) ys) (format () "~D" (caar (tree (reference self)))))
                        (om-draw-string  (+ (car (rectangle self)) deltachiff) (+ ys (round size 4) (round size 2)) (format () "~D" (cadar (tree (reference self)))))
                        ))))
                (unless (or (first-of-group? self) poly-p)
                  (om-draw-line  (car (rectangle self))  ys 
                                 (car (rectangle self)) (+ ys size))))))
      (draw-extras self view size staff)
      )))

(defmethod page-draw-object-ryth ((self grap-rest) score view pointer x y zoom  slot size   staff chnote )
   (declare (ignore  maxy slot staff miny minx maxx))
   (with-focused-view (view pointer)
     (let* ((realpos (round (+ x  (* zoom (x self)))))
          (str (headchar self))
          (line (case (elt str 0) ((code-char 183)  (round size 4)) ((code-char 238) 0)))
          (strsizex (get-name-size str (om-make-font *heads-font* size)))
          delta-y delta-h )
     (if (group-p (reference (parent self)))
       (progn 
         (if (string-equal (stemdir self) "up")
           (progn
             (case (elt str 0)
               ((code-char 183)  (setf delta-y (round size 4)
                           delta-h 0))
               ((code-char 238) (setf delta-y (round size 8)
                          delta-h (round size 8)))
               ((code-char 206) (setf delta-y (round size 2) 
                          delta-h (round size 2)))
               (otherwise (setf delta-y (* (round size 4) (delta-beams-rest (beams-num self)))
                                delta-h (if (= (beams-num self) 1) (round size 4) (round size 2)))))
             (setf (rectangle self) (list realpos (+ y (y self) (* (delta-beams-up (beams-num self))  (round size 4)))
                                          (+ realpos strsizex) 
                                          (+ y (y self) delta-y delta-h))))
           (progn
             (case (elt str 0)
               ((code-char 183)  (setf delta-y (round size -4)
                           delta-h 0))
               ((code-char 238) (setf delta-y (round size -8)
                          delta-h (round size 8)))
               ((code-char 206) (setf delta-y (round size -2) 
                          delta-h (round size )))
               (otherwise (setf delta-y (* (round size -4)  (delta-beams-dwn (beams-num self))  )
                                delta-h (if (= (beams-num self) 1) (round size 4) (round size 2)))))
             (setf (rectangle self) (list realpos (+ y (y self) delta-y (* (round size -4) (- (min 5 (delta-beams-rest (beams-num self))) 1)))
                                          (+ realpos strsizex) 
                                          (+ y (y self))))))
         (om-with-fg-color nil  (mus-color (reference self)) 
           (if (bigrest self)
             (draw-long-silence size realpos (+ y (y self) delta-y) (bigrest self) (points  self))
             
             (progn
                (om-draw-string  realpos (+ y (y self) delta-y) str)
               (write-note-points self  (+ (round size 2) realpos) (+ y (y self) delta-y) size)
               (when line
                 (om-draw-line (- realpos (round size 8)) (- (+ y (y self) delta-y) line) (+ realpos (round size 12/5)) (- (+ y (y self) delta-y) line))))
             )))
       (progn
         (case (elt str 0)
           ((code-char 183)  (setf delta-y (round size 4)
                       delta-h 0))
           ((code-char 238) (setf delta-y (round size 8)
                      delta-h (round size 8)))
           ((code-char 206) (setf delta-y (round size 2) 
                      delta-h (round size 2)))
           (otherwise (setf delta-y (* (round size 4) (min 4 (- (delta-beams-rest (beams-num self)) 1)))
                            delta-h (if (= (beams-num self) 1) (round size 4) (round size 2)))))
         (draw-one-rest self str strsizex size realpos (+ y (y self)) line delta-y delta-h)
         ))
     (draw-extras self view size staff))))

(defmethod page-draw-object-ryth ((self grap-ryth-chord) score view pointer x y zoom  slot size   staff chnote )
  (with-focused-view (view pointer)
    (om-with-fg-color nil (mus-color (reference self))
                    (let* ((dir (stemdir self))
                           (thenotes (copy-list (inside self))))
                      (loop for item in thenotes do
                            (page-draw-object-ryth item score view pointer x y zoom  slot size  staff chnote ))
                      (collect-rectangles self)
                      (when (bigchord self) 
                        (om-with-font (om-make-font *heads-font* (round size 1.6))
                                      (om-draw-string  (+ (car (rectangle self)) (round size 5)) 
                                                      (+ (round size 7) (fourth (rectangle self))) 
                                                      (format nil "~D" (bigchord self)))
                                      ))
                      (when (figure-? self) 
                        (when (stem self)
                          (draw-chord-beams self x y zoom (beams-num self) dir size)))
                      (draw-extras self view size staff))
                    
                    (when *om-tonalite*
                      ;(draw-degre self (- (round (+  x (/ size 3.5) (* zoom (x self)))) 4) y size))
                      (draw-chiffrage self x y zoom size))
                    )))

(defmethod page-draw-object-ryth ((self t) view score pointer x y zoom  slot size   staff chnote )
   (with-focused-view (view pointer)
     (page-draw-object self score view pointer x y zoom  slot size  staff  chnote )
  (draw-extras self view size staff)
  ))

(defmethod page-draw-object-ryth ((self voice) score view pointer x y zoom  slot size  staff chnote )
   (with-focused-view (view pointer)
     (call-next-method)
   (remake-measures (inside self))
  (draw-extras self view size staff)
  ))

