(in-package :om)


;;;===================
;;; EDITOR CLASSES

(omg-defclass sheetEditor (scoreEditor) ())

(defmethod get-score-class-panel ((self sheetEditor)) 'sheetPanel)

(omg-defclass sheetPanel (scorepanel)  
  ((timebpf :accessor timebpf :initarg :timebpf :initform nil)
   (trackpanels :accessor trackpanels :initarg :trackpanels :initform nil)
   (show-tracks :accessor show-tracks :initarg :show-tracks :initform t)
   (show-time :accessor show-time :initarg :show-time :initform nil)
   (grille-p :initform nil :accessor grille-p)
   ))


(defmethod reference ((self sheetpanel)) (object (editor self)))
(defmethod om-drag-selection-p ((self sheetpanel) position) nil)

(defmethod close-editor-after ((self sheeteditor))
  (when (reference (panel self))
    (setf (Editorframe (reference (panel self))) nil)))


(defmethod update-editor-after-eval ((self sheetEditor) val)
  (om-with-compiler-warnings nil
    (setf (object self) val)
    (init-music-patch (panel self))
    (remove-panel-boxes (panel self))
    (init-tracks (panel self))
    (update-panel (panel self))))


(defmethod draw-measure-cursor ((self sheetpanel) poly (events null)) t)

(defmethod draw-measure-cursor ((self sheetpanel) poly events) 
  (draw-sheet-cursor self poly events nil))

(defun draw-sheet-cursor (self voice events last-evt)  ;a faire
  (let* ((cur-evt (second (car events)))
        (cur-time (first (car events)))
        (cur-pixel (get-x-pos self cur-time))
        (y0 (om-v-scroll-position self)))
     (when (> cur-pixel (+ (om-h-scroll-position self) (w self)))
       (om-set-scroll-position self (om-make-point cur-pixel (om-v-scroll-position self))))
     (om-with-focused-view self
       (om-with-dashline
         (when last-evt
           (om-draw-line (get-x-pos self last-evt) y0 (get-x-pos self last-evt) (+ y0 (h self)))) 
         (om-draw-line  cur-pixel y0 cur-pixel (+ y0 (h self)))
         (let ((next (second events)))
           (if (or (Idle-p *general-player*) (not next))
             (let ((palette (palette *global-palette*)))
               (when palette
                 (om-draw-line cur-pixel y0 cur-pixel (+ y0 (h self)))
                 (setf (whoplay palette) nil)
                 (om-invalidate-view palette t)))
             (dfuncall (- (- (car next) (car (car events)) ) (- (- (clock-time) *metric-strat-time*) (car (car events))))   
                       'draw-sheet-cursor self voice (cdr events) (first (car events)))))))))


(defmethod draw-interval-cursor ((self sheetpanel)) (rythm-draw-interval-cursor self))

(defmethod convert-interval ((self sheetpanel)) (rhythm-convert-interval self))

(defmethod graphic-obj ((self sheetpanel)) self)

(defmethod get-graph-type-obj ((self sheetpanel) type)
  (let ((rep nil))
    (setf (selection? self) nil)
    (loop for tr in (trackpanels self) do
          (loop for obj in (om-subviews tr)
                collect obj))))

(defmethod om-view-cursor ((self sheetpanel))
  (cond 
    ((om-option-key-p) *add-cursor*)
    (t *om-arrow-cursor*)))

(defvar *sheet-new-menu* nil)
(defun init-sheet-menu ()
  (setf *sheet-new-menu* (om-new-menu "New" (om-new-leafmenu "New Track" #'(lambda() (add-one-graptrack (panel (om-front-window)) (om-make-point 0 0)))))))
(om-add-init-func 'init-sheet-menu)


(defmethod get-new-menu ((self sheetEditor)) *sheet-new-menu*)

;;;==========================
;;; SCORE PANEL METHODS

(defmethod update-slot-edit ((self sheetpanel)) t)

(defmethod record-undo ((self sheeteditor)) nil)

(defmethod get-score-class-ctrls ((self sheetEditor)) 'sheet-controls)

(defmethod buttons-list ((self sheetPanel)) nil)

(defmethod play-from-palette ((self sheetpanel))
  (play-from-ryth-palette self))

(defmethod get-obj-to-play ((self sheetPanel))
   (list (object (om-view-container self)) 
         :port (strig2numport (get-edit-param (om-view-container self) 'outport))
         :approx (get-edit-param (om-view-container self) 'approx)))


(defmethod editor-save ((self sheetEditor))
   "Relation windows save the object associated to the scroller, not the selected icons."
   (let ((thesheet (reference (panel self))))
     (if (mypathname thesheet)
       (omNG-save thesheet nil)
       (om-beep-msg "This sheet has no attached file")
       )))

(defmethod modify-patch ((self sheetPanel))
   "Set differents flag ON to say that the patch was modified."
   ;;(setf (compiled? (reference self)) nil)
   (setf (saved? (reference self)) nil)
   (when (mypathname (reference self))
     (om-set-window-title (window self) (string+ "^" (name (reference self))))))


;;;==========================
;;; menu actions

(defmethod extra-menu-info ((self sheeteditor))
  (disable-this-menu-items *window-menu-file*  '("make alias" "duplicate" "print" "page setup" "new" "get info"))
  (enable-this-menu-items *window-menu-file*  '("last saved"))
  (disable-this-menu-items *window-menu-edit*  '( "cut" "copy" "paste" "clear"))
  (enable-this-menu-items *window-menu-edit*  '("undo")))

(defmethod window-last-saved ((self sheetEditor))
   "Close the window 'self' and load the last save version of the patch associated to the scroller."
   (update-last-saved (object self)))

(defmethod update-last-saved ((self sheet))
   "Close the patch and load the last save version of the patch."
   (when (editorframe self)
     (om-close-window (window (editorframe self))))
   (setf (loaded? self) nil)
   (OpenobjectEditor self))

;;;===================
;;; TRACK

(omg-defclass trackpanel (om-view om-view-drop)
  ((selected :accessor selected :initform nil)
   (reference :accessor reference :initarg :reference :initform nil)
   ))

(defmethod object ((self trackpanel)) (reference self))

(defmethod track-subviews ((self trackpanel))
  (append 
   (find-class-boxes (om-subviews self) 'sheet-linobjectframe)
   (find-class-boxes (om-subviews self) 'sheet-scoreobjectframe)))


;;;===================
;;; OBJECT FRAMESs

(omg-defclass sheet-objectframe (om-scroller om-view-drag) 
  ((selected :accessor selected :initform nil)
   (reference :initarg :reference :accessor reference :initform nil)
   (inputframes :initarg :inputframes :accessor inputframes :initform nil)
   (outframes :initform nil :accessor outframes)
   (connections :initform nil :accessor connections)
   ))


  

;;; scoreobjects
(omg-defclass sheet-scoreobjectframe (scorepanel sheet-objectframe) ())


;;; maquette, sound, bpf, midifile
(omg-defclass sheet-linobjectframe (sheet-objectframe) ((graphic-obj :initform nil :accessor graphic-obj)))

(defmethod get-sheetframe-class ((self t)) 'sheet-linobjectframe)
(defmethod get-sheetframe-class ((self score-element)) 'sheet-scoreobjectframe)
(defmethod get-sheetframe-class ((self midicontrol)) 'sheet-linobjectframe)


(defmethod editor ((self sheet-objectframe)) self)

(defmethod object ((self sheet-objectframe)) 
   (reference self))

(defmethod objectfromeditor ((self sheet-objectframe)) 
   (obj (reference self)))

;(defmethod rectangle ((self sheet-objectframe)) 
;  (list (x self) (y self) (+ (x self) (w self)) (+ (y self) (h self))))

; ...
(defmethod rectangle ((self sheet-objectframe)) 
  (list (x self) (+ (y (om-view-container self)) (y self))
        (+ (x self) (w self)) (+ (y (om-view-container self)) (y self) (w self))))

(defmethod sheet-container ((self sheet-objectframe)) 
   (om-view-container (om-view-container self)))

(defmethod sheet-container ((self trackpanel)) 
   (om-view-container self))

(defmethod track-container ((self sheet-objectframe)) 
  (om-view-container self))


(defmethod do-initialize-instance ((self sheet-objectframe))
  (let* ((track (om-view-container self))
        (panel (when track (om-view-container track)))
        (obj (reference self)))
    (when panel
      (om-set-view-size self (om-make-point (- (get-x-pos panel (end-t obj)) (get-x-pos panel (start-t obj))) (h self)))
    nil)))

(defmethod update-panel ((self sheet-objectframe) &optional (updateref nil)) nil)


;;;==================
;;; LIN OBJS

(defmethod collect-bpftime-objects ((self sheet-linobjectframe) at size)
  (loop for item in (get-temp-pivots (reference self))
        collect (list (+ at (car item)) 0)))

(defmethod collect-temporal-objects ((self sheet-linobjectframe) at)
  (collect-temporal-objects (graphic-obj self) (start-t (reference self))))

;;;===================
;;; SCORE OBJ

(defmethod objectfromeditor ((self sheet-scoreobjectframe)) 
   (obj (reference self)))

(defmethod object ((self sheet-scoreobjectframe)) 
   (reference self))

(defmethod do-initialize-instance ((self sheet-scoreobjectframe))
  (call-next-method)
  (setf staff (correct-staff self (get-edit-param (reference self) 'staff)))
  (setf (staff-sys self) (get-staff-system staff))
  (setf (staff-size self) (get-edit-param (reference self) 'fontsize))
  (setf (staff-tone self) (get-edit-param (reference self) 'tone))
  )


(defmethod correct-staff ((self sheet-scoreobjectframe) staff)
  (let ((staff (if (member staff *chord-satff-om* :test 'equal :key 'cadr) staff
                   (second (cadr *chord-satff-om*)))))
    (set-edit-param (reference self) 'staff  staff)
    staff))

(defmethod editor ((self sheet-scoreobjectframe)) self)

(defmethod set-edit-param ((self sheet-objectframe) param val) 
  (set-edit-param (reference self) param val))

(defmethod get-edit-param ((self sheet-objectframe) param) 
  (get-edit-param (reference self) param))

(defmethod score-top-margin ((self sheet-scoreobjectframe) &optional val) 
(if val
  (setf (top-margin (edparams (reference self))) val)
  (top-margin (edparams (reference self)))))

(defmethod get-key-space ((self sheet-scoreobjectframe))
   (+ (*  (staff-size self) 2) (if (grap-voice-p (graphic-obj self)) 0 (staff-size self))))

(defmethod collect-temporal-objects ((self sheet-scoreobjectframe) at)
  (collect-temporal-objects  (graphic-obj self) (start-t (reference self))))


;;;===================
;;; TRACKS SETTINGS

;;; creation des tracks et des objets dans les tracks
;;; appelee quand le panel est cree ou quand la structure du sheet a change
(defmethod init-tracks ((self sheetpanel))
  (loop for trackp in (trackpanels self) do 
        (om-remove-subviews self trackp))
  (let ((allevts (sheet-get-allobjs-evts self))
        (tmppanellist nil))
    (loop for track in (inside (object (editor self)))
          for pos in (trackspos (params (reference self)))
          for i = 0 then (+ i 1) do
          (let ((trackp (om-make-view 'trackpanel 
                                      :position (om-make-point 0 0) 
                                      :size (om-make-point (om-point-h (om-field-size self)) (* (font-size (params (reference self))) (size track)))
                                      :reference track)))
            (loop for object in (objs track) do
                  (let ((objview (om-make-view (get-sheetframe-class (obj object))
                                         :reference object
                                         :position (om-make-point (get-x-pos self (start-t object)) 3)
                                         :scrollbars nil
                                         :size (om-make-point (- (get-x-pos self (end-t object)) (get-x-pos self (start-t object))) (- (h trackp) 6))
                                         :field-size (om-make-point (- (get-x-pos self (end-t object)) (get-x-pos self (start-t object))) (- (h trackp) 6))
                                         )))
                    ;(update-graphic-connections objview (remove objview (sheet-get-allobjs self)))
                    (setf (frames object) (list objview))
                    ;(om-add-subviews objview (om-make-view 'om-view :size (om-make-point 30 30) :bg-color *om-red-color*))
                    (set-edit-param objview 'fontsize (font-size (params (reference self))))
                    (do-initialize-instance objview)
                    (update-panel objview)
                    (om-add-subviews trackp objview)
                    ))
            (when (in-patch-mode? self) 
              (add-patch-objs trackp)
              )
            ;(om-set-bg-color trackp (om-make-color 0.95 0.95 0.95))
            (om-add-subviews self trackp)
            (pushr trackp tmppanellist)))
    (setf (trackpanels self) tmppanellist)
    (loop for obj in (sheet-get-allobjs self) do
          ;(update-graphic-connections obj (remove (reference obj) (sheet-get-allobjs-evts self)))
          (update-graphic-connections obj (append allevts (boxes (object self))))
          )
    ))

(defmethod initialize-instance :after ((self sheetPanel) &rest initargs)
  ;;;(setf (obj-mode self) "voice")
  (init-tracks self)
  (update-panel self)
  )


;;;==========================
;;; EVENTS

(defmethod handle-key-event ((self sheetpanel) char)
  (case char
    (#\SPACE (if (Idle-p *general-player*)
                 (palette-act (palette *global-palette*) 0)
                 (palette-act (palette *global-palette*) 1)))
    (#\v 
     (om-eval-enqueue 
      `(progn
         (mapc 'eval-box ',(get-actives self))
         (mapc 'eval-box ',(get-actives-track-ev self))
         (clear-ev-once ,self))))
    (otherwise
     (if (selection? self)
       (case char
         
         (#\t (set-score-page-mode self (not (score-page-mode self )))
          (update-panel self t))
         (#\h (show-help-window "Score Help" (get-help-list self)))
         (:om-key-up  (cond 
                       ((track-p (car (selection? self)))
                        (cond ((om-shift-key-p) (resize-staff self -1))
                              ((om-control-key-p) (switch-staff self -1))
                              (t (move-selection self 0))))
                       ((track-ev-p (car (selection? self))) 
                        (marginmove-object self -1)
                        )
                       (t nil)))   
         (:om-key-down  (cond ((track-p (car (selection? self)))
                               (cond ((om-shift-key-p) (resize-staff self 1))
                                     ((om-control-key-p) (switch-staff self 1))
                                     (t (move-selection self 1))))
                              ((track-ev-p (car (selection? self))) 
                               (marginmove-object self 1))
                              (t nil)))
         (:om-key-right (cond 
                         ((track-ev-p (car (selection? self)))
                          (tmove-selection self (if (om-shift-key-p) (* 10 (tstep (params (reference self)))) (tstep (params (reference self))))))
                         (t nil)))      
         (:om-key-left (cond 
                        ((track-ev-p (car (selection? self)))
                         (tmove-selection self (if (om-shift-key-p) (* -10 (tstep (params (reference self)))) (- (tstep (params (reference self)))))))
                        (t nil)))
         (otherwise (when (print (selection? self))
                      (case char
                        (:om-key-delete (delete-selection self))
                        (#\o (open-internal-editor self))
                        (otherwise (om-beep))))))
       (call-next-method))
     )
  ))

(defmethod reference (grap-obj) nil)

(defmethod om-score-click-handler ((self sheetPanel) where double-click-p)
  (cond 
   ;((score-page-mode self)
   ; (page-click-handler self where selection double-click-p))
   (t 
    (if (cursor-p self)
      (om-with-focused-view self (control-actives self where))
      (let ((selection (om-find-view-containing-point self where)))
        (cond ((equal self selection)
               (when (in-patch-mode? self)
                 (mapc #'(lambda (control) 
                           (omG-unselect control)) (get-actives self))
                 (click-in-connection self where))
               (off-selection self)
               (setf (selection? self) nil)
               (cond ((om-option-key-p) (add-one-graptrack self where))
                     (t (om-with-focused-view self (control-actives self where))))
               (om-invalidate-view self))
              ((track-p (reference selection))
               (unless (track-p (car (selection? self)))
                 (off-selection self))
               (if (om-shift-key-p) 
                 (omselect-staff-with-shift self selection)
                 (when (not (member (reference selection) (selection? self) :test 'equal))
                   (off-selection self)
                   (setf (selected selection) t)
                   (setf (selection? self) (list (reference selection)))
                   ))
               (om-invalidate-view self))
              ((track-ev-p (reference selection))
               (unless (track-ev-p (car (selection? self)))
                 (off-selection self))
               (if double-click-p
                 (open-sheet-internal-editor self selection) ;; march pas
               (if (om-shift-key-p) 
                 (omselect-staff-with-shift self selection)
                 (when (not (member (reference selection) (selection? self) :test 'equal))
                   (off-selection self)
                   (setf (selected selection) t)
                   (setf (selection? self) (list (reference selection)))
                   )))
               (om-invalidate-view self)
               )
              ((in-patch-mode? self)
               ;; renvoyer :call-next-method pour appeler le click handler du patchpanel
               (scorepatch-click-handler self selection where double-click-p))
              )
        ))
    )))


;;;=====================
;;; ACTIONS

(defmethod select-all ((self sheetPanel))
  (setf (selection? self) (inside (object (editor self))))
  (loop for item in (trackpanels self) do (setf (selected item) t))
  (om-invalidate-view self))

(defmethod off-selection ((self sheetPanel))
  (loop for tr in (trackpanels self) do
        (setf (selected tr) nil)
        (loop for obj in (om-subviews tr)
              do (setf (selected obj) nil))))

(defmethod omselect-staff-with-shift ((self sheetPanel) graph-obj)
  (if (member (reference graph-obj) (selection? self) :test 'equal)
    (setf (selection? self) (remove  (reference graph-obj) (selection? self) :test 'equal)
          (selected graph-obj) nil)
    (progn (push (reference graph-obj) (selection? self))
           (setf (selected graph-obj) t))))

(defmethod general-delete ((view sheetPanel) (self track))
  (let ((pos (position self (inside (reference view)) :test 'equal)))
    (remove-one-track (reference view) pos))
  (init-tracks view))

(defmethod general-delete ((view sheetPanel) (self track-ev))
  (remove-one-object (reference view) self)
  (init-tracks view))

(defmethod delete-selection ((self sheetPanel))
  (loop for obj in (selection? self) do
        (general-delete self obj))
  (setf (selection? self) nil)
  (update-panel self))

(defmethod add-one-graptrack ((self sheetpanel) pos)
  (let ((y (om-point-v pos))
         (p 0))
    (loop for tr in (trackpanels self) 
          for i = 1 then (+ i 1) do
          (when (> y (y tr)) (setf p i)))
    (add-one-track (reference self) p)
    (init-tracks self)
    (update-panel self)
    ))
                         

(defmethod score-move-a  ((self track) panel trans)
  (move-track (reference panel) self (if (>= 0 trans) 1 -1))
  (update-panel panel))

(defmethod resize-staff ((self sheetPanel) dir)
  (let ((sheet (reference self)))
    (loop for item in (selection? self) do
          (when (track-p item)
            (resize-track item dir)))
    (update-panel self)))

(defmethod switch-staff ((self sheetPanel) dir)
  (let* ((sheet (reference self))
        (tmpselection (let (ll)
                        (loop for item in (inside sheet) do 
                              (when (member item (selection? self) :test 'equal) 
                                (pushr item ll)))
                        ll)))
     
    (loop for item in (if (= dir 1) (reverse tmpselection) tmpselection) do
          (switch-track sheet item dir))
    (init-tracks self)
    (setf (selection? self) tmpselection)
    (loop for item in (trackpanels self) do 
          (when (member (reference item) (selection? self) :test 'equal)
            (setf (selected item) t)))
    (update-panel self)))


(defmethod tmove-selection ((self sheetpanel) val)
  (loop for obj in (selection? self) do
        (when (track-ev-p obj)
          (let* ((new-start (max 0 (+ (start-t obj) val)))
                (new-end (+ new-start (- (end-t obj) (start-t obj)))))
            (setf (start-t obj) new-start)
            (setf (end-t obj) new-end)
            )))
  (update-panel self))


(defmethod marginmove-object ((self sheetpanel) val)
  (loop for obj in (selection? self) do
        (when (track-ev-p obj)
          (setf (top-margin (edparams obj)) (max 0 (+ (top-margin (edparams obj)) val)))
          ))
  (update-panel self))


;;;========================
;;; PAGE

(defmethod page-click-handler ((self sheetPanel) where selection double-click-p)
  (print "page clic"))


;;;============================================
;;; Sheet Controls

(omg-defclass sheet-controls (3dBorder-view) 
  ((stepitem :accessor stepitem :initform nil :initarg :stepitem)
   (stepedit :accessor stepedit :initform nil :initarg :stepedit))
  (:default-initargs 
    :c++ *controls-color++* :c+ *controls-color+* 
    :c-- *controls-color--* :c- *controls-color-*))

(defmethod initialize-instance :after ((self sheet-controls) &rest l 
                                       &key (tone "1/2") (staff "ffgg") (font-size "24") (zoom 1) (mode 0))
  (declare (ignore l))
  (let* ((bgcol *controls-color*)
         (l1 230)
         (l2 380)
         (sizeitem (om-make-dialog-item 'om-static-text 
                                        (om-make-point 60 26) 
                                        (om-make-point 90 20) "Font size"
                                        :font *controls-font*
                                        :bg-color *controls-color*))
         (sizebut (om-make-dialog-item 'om-pop-up-dialog-item 
                                       (om-make-point (+ 60 (om-string-size "Font size" *controls-font*) 8) 26) 
                                       (om-make-point 56 20)
                                       ""
                                       :di-action (om-dialog-item-act item
                                                    (let ((newsize (cadr (nth (om-get-selected-item-index item) *mus-font-size*))))
                                                      (change-editor-size (panel (om-view-container self)) newsize)))
                                       :font *controls-font* ;*om-default-font1*
                                       :range (loop for item in *mus-font-size* collect (car item)) 
                                       :value font-size
                                       :bg-color *controls-color*
                                       ))
         (tracksbox (om-make-dialog-item 'om-check-box (om-make-point 200 2) (om-make-point 100 20) "Show tracks"
                                         :checked-p (show-tracks (panel (om-view-container self)))
                                         :font *controls-font*
                                         :bg-color *controls-color*
                                         :di-action (om-dialog-item-act item 
                                                      (setf (show-tracks (panel (om-view-container self))) (om-checked-p item))
                                                      (om-invalidate-view (panel (om-view-container self))))))
         (timebox (om-make-dialog-item 'om-check-box (om-make-point 200 26) (om-make-point 100 20) "Show times"
                                       :checked-p (show-time (panel (om-view-container self)))
                                       :font *controls-font*
                                       :bg-color *controls-color*
                                       :di-action (om-dialog-item-act item 
                                                    (setf (show-time (panel (om-view-container self))) (om-checked-p item))
                                                    (om-invalidate-view (panel (om-view-container self))))))
         
         (grilletxt (om-make-dialog-item 'om-static-text 
                              (om-make-point 350 6) 
                              (om-make-point 60 14)
                               "t Step"
                               :font *controls-font*
                               :bg-color *controls-color*))
         (stepitem (om-make-dialog-item 'grillettybox (om-make-point 400 6) (om-make-point 40 14) 
                                        (format nil " ~D" (tstep (params (object (om-view-container self)))))
                                        :font *controls-font*
                                        :bg-color *om-white-color*))
         (grillebut (om-make-dialog-item 'om-check-box 
                               (om-make-point 350 26) 
                               (om-make-point 100 20)
                               "Show"
                               :checked-p (grille-p (panel (om-view-container self)))
                               :di-action (om-dialog-item-act item
                                            (setf (grille-p (panel (om-view-container self))) (om-checked-p item))
                                            (update-panel (panel (om-view-container self)))
                                            t)
                               :font *controls-font*
                               :bg-color *controls-color*))
         
         )
    
    (setf (stepitem self) stepitem)
    (om-set-bg-color self *controls-color*)
    (om-add-subviews self sizeitem sizebut tracksbox timebox grilletxt grillebut stepitem)
    (add-zoom2control self zoom)
    ))

(defmethod change-editor-size ((self sheetPanel) newsize)
  (call-next-method)
  (setf (font-size (params (reference self))) newsize)
  (loop for tr in (trackpanels self) do
        (loop for scoreobj in (find-class-boxes (track-subviews tr) 'sheet-scoreobjectframe) do
              (change-editor-size scoreobj newsize))) 
  (update-panel self))


;;;---------
;;; special ttybox from mlmaquette

(omg-defclass grillettybox (ttybox) ())

(defmethod om-component-border ((self grillettybox)) :line)

(defmethod om-view-click-handler ((self grillettybox) where)
   (declare (ignore where))
   nil)

(defmethod initial-text-ttybox ((self grillettybox))
  (om-dialog-item-text self))

(defmethod open-ttybox ((self grillettybox))
   (let* ((thetext (initial-text-ttybox self))
          (container (om-view-container self)))
     (setf (selection? (panel (om-view-container (om-view-container self)))) nil)
     (om-remove-subviews container (stepitem container))
     (setf (stepitem container) 
       (om-make-dialog-item 'om-editable-text
                            (om-view-position self) (om-get-text-edit-size (om-make-point 40 16))
                            thetext
                            :allow-returns nil
                            :di-selected-p t
                            :font *controls-font*))
     (om-add-subviews container (stepitem container))
     (setf (text-view (om-view-container container)) (stepitem container))
     (setf (stepedit container) t)
     ))

(defmethod om-view-click-handler :before ((self sheeteditor) where) 
  (when (stepedit (ctr-view self))
    (let* ((text (om-dialog-item-text (stepitem (ctr-view self))))
           (val (unless (equal "" text) (read-from-string text)))
           (pos (om-view-position (stepitem (ctr-view self)))))
      (when (and val (integerp val))
        (if (< val 1) (setf val 1) (if (> val 30000) (setf val 30000)))
        (setf (tstep (params (object self))) val)
        (update-panel (panel self)))
      (om-remove-subviews (ctr-view self) (stepitem (ctr-view self)))
      (setf (stepitem (ctr-view self)) 
            (om-make-dialog-item 'grillettybox pos (om-make-point 40 14) 
                                 (format nil " ~D" val)
                                 :font *controls-font*
                                 :bg-color *om-white-color*))
      (om-add-subviews (ctr-view self) (stepitem (ctr-view self)))
      (setf (stepedit (ctr-view self)) nil)
      (setf (text-view self) nil)
      (om-invalidate-view (ctr-view self)))))

(defmethod om-drag-selection-p ((self grillettybox) mouse-position) nil)

;;;========================
;;; OPEN INTERNAL

(defmethod open-sheet-internal-editor ((self sheetPanel) objframe)
  (let* ((object (obj (reference objframe)))
        (int-info (obj-for-internal-editor object)))
    (when int-info
      (if (maquette-p object) 
        (openobjecteditor object)
        (push 
         (let ((edwin (make-editor-window (first int-info) object (second int-info) objframe)))
           (set-editor edwin objframe)
           edwin)
         (attached-editors (om-view-container self)))))))


(defmethod set-editor ((self editorwindow) (objframe sheet-objectframe)) t)

(defmethod set-editor ((self editorwindow) (objframe sheet-scoreobjectframe)) 
  (setf (staff-sys (panel self)) (get-staff-system (get-edit-param objframe 'staff)))
  (setf (staff-tone (panel self)) (get-edit-param objframe 'tone))
  (update-panel (panel self))
  )

(defmethod obj-for-internal-editor ((self ommaquette))
  (list 'maquetteeditor "sheet track"))

(defmethod obj-for-internal-editor ((self sound))
  (list 'soundeditor "sheet track"))

(defmethod obj-for-internal-editor ((self bpf))
  (list 'bpfeditor "sheet track"))


(defmethod real-internal-editor ((self track)) (car (objs self)))

(defmethod change-in-int-editor ((self sheetPanel) (internal scorePanel) newobject lastobj)
  (update-panel self t))

(defmethod change-in-int-editor ((self sheet-scoreobjectframe) (internal scorePanel) newobject lastobj)
  (set-edit-param (reference self) 'staff (sysname (staff-sys internal)))
  (set-edit-param (reference self) 'tone (staff-tone internal))
  (setf (end-t (reference self)) (+ (start-t (reference self)) (get-obj-dur (obj (reference self)))))
  (do-initialize-instance self)
  (update-panel (sheet-container self) nil))


;;;=======================
;;; D&D

(defmethod om-drag-selection-p ((view sheet-objectframe) local-mouse-position)
  (declare (ignore view local-mouse-position)) t)

(defmethod drop-allow-p ((D&DHandler omdrag-drop) (dragged omboxeditcall) (target track)) t)

(defmethod drop-allow-p ((D&DHandler omdrag-drop) (dragged track-ev) (target track)) t)

(defmethod drop-allow-p ((D&DHandler omdrag-drop) (dragged track-ev) (target track-ev)) nil)
(defmethod drop-allow-p ((D&DHandler omdrag-drop) (dragged track-ev) (target sheet)) nil)

(defmethod drop-allow-p ((D&DHandler omdrag-drop) (dragged track-ev) (target ompatch)) t)


(defmethod move-and-not-action ((self t) (cible trackpanel)) nil)

(defmethod get-drag-object ((self sheet-objectframe)) self)
(defmethod get-drag-object ((self sheet-scoreobjectframe)) self)

(defmethod save-connections ((source sheet-scoreobjectframe) (target sheetpanel) list) 
  (declare (ignore list))
   nil)

(defmethod save-connections ((source sheet-objectframe) (target t) list) 
  (declare (ignore list))
   nil)

(defmethod om-drag-start ((view sheet-scoreobjectframe))
  (sheet-startdrag-object view))

(defmethod om-drag-start ((view sheet-objectframe))
  (sheet-startdrag-object view))

(defmethod sheet-startdrag-object ((self sheet-objectframe))
  (let* ((container (om-view-container (om-view-container self)))
         (region (om-new-region))
         (viewpos (om-view-position self)))
    (setf region (om-set-rect-region region 0 0 (w self) (h self)))
    (setf (dragged-view *OM-drag&drop-handler*)  self
          (dragged-list-objs *OM-drag&drop-handler*)  (list self)
          (container-view *OM-drag&drop-handler*)  container
          (true-dragged-view *OM-drag&drop-handler*) self
          (drag-flavor *OM-drag&drop-handler*) :omvw)
    (unwind-protect
      (om-drag-create-item-bounds self region)
      (om-dispose-region region))
    )
  )


(defun stepfloor (val step)
  (* step (floor val step)))

(defmethod perform-drop ((D&DHandler omdrag-drop) (dragged boxeditorframe) 
                                                  (target trackPanel) position)
  (let ((spanel (om-view-container target))
        (val (clone (value (object dragged)))))
    (add-one-object (reference spanel) (reference target) val (stepfloor (get-ms-pos spanel (om-point-h position)) (tstep (params (reference spanel)))))
    (init-tracks spanel)
    (update-panel spanel)
    t
    ))

(defmethod perform-drop ((D&DHandler omdrag-drop) (dragged sheet-objectframe) 
                                                  (target trackPanel) 
                                                  position)
  (let* ((spanel (om-view-container target))
         (trev (reference dragged)))
    (if (equal spanel (sheet-container dragged))
      (progn
        (remove-one-object (reference spanel) trev)
        (add-one-object (reference spanel) (reference target) (obj trev) (stepfloor (get-ms-pos spanel (om-point-h position)) (tstep (params (reference spanel)))))
        )
      (add-one-object (reference spanel) (reference target) (clone (obj trev)) (stepfloor (get-ms-pos spanel (om-point-h position)) (tstep (params (reference spanel)))))
      )
    (init-tracks spanel)
    (update-panel spanel)
    t))

(defmethod perform-drop ((D&DHandler omdrag-drop) (dragged sheet-objectframe) 
                                                  (target patchPanel) 
                                                  position)
  (omG-add-element target (make-frame-from-callobj 
                           (omNG-make-new-boxcall (omNG-make-new-instance (clone (obj (reference dragged)))
                                                                          (mk-unique-name target (name dragged)))
                                                  (om-mouse-position target) (mk-unique-name target (name dragged)))))
  t
  )


(defmethod perform-duplicate-list ((D&DHandler omdrag-drop) (targetobj track) (target-frame trackpanel) subframes copies pos0)
   (declare (ignore subframes))
   (let ((spanel (om-view-container target-frame))
         (pos (om-add-points
               (om-subtract-points (get-position (car (dragged-list-objs D&DHandler))) 
                                   (get-position (get-drag-object (dragged-view D&DHandler)))) 
               (om-subtract-points (drop-mouse-pos D&DHandler) (initial-mouse-pos D&DHandler))))
         )
     (add-one-object (reference spanel) targetobj (obj (car copies)) (stepfloor (get-ms-pos spanel (om-point-h pos)) (tstep (params (reference spanel)))))
     (init-tracks spanel)
     (update-panel spanel)
     t))




(defmethod OMGMoveObject ((self sheet-objectframe) new-position)
  (let* ((sheetpanel (om-view-container (om-view-container self)))
         (time (stepfloor (get-ms-pos sheetpanel (max 0 (om-point-h new-position))) (tstep (params (reference sheetpanel)))))
         (dur (- (end-t (reference self)) (start-t (reference self))))
         (newx (get-x-pos sheetpanel time)))
    (setf (start-t (reference self)) time)
    (setf (end-t (reference self)) (+ time dur))
    (om-change-view-position self (om-make-point newx (y self)))
    (update-panel (sheet-container self))
    ))



;;;======================
;;; ALIGNEMENT


(defclas grap-start-event (simple-graph-container) ())

(defclas grap-end-event (simple-graph-container) ())


(defmethod get-x-pos ((self sheetpanel) (time-ms integer))
  (if (timebpf self)
    (let ((max (car (last (x-points (timebpf self)))))
          (min (car (x-points (timebpf self)))))
      (cond
       ((> time-ms max) (round (+ (bpf-get-val (timebpf self) max) (round (- time-ms max) (staff-size self)))))
       ((< time-ms min) (round (- (bpf-get-val (timebpf self) min) (round (- min time-ms) (staff-size self)))))
       (t (round (bpf-get-val (timebpf self) time-ms)))))
    (round time-ms (staff-size self))))

(defmethod get-ms-pos ((self sheetpanel) (x-pos integer))
  (if (timebpf self)
    (let ((max (car (last (y-points (timebpf self)))))
          (min (car (y-points (timebpf self)))))
      (cond
       ((> x-pos max) (round (+ (car (y-transfer (timebpf self) max 0)) (* (- x-pos max) (staff-size self)))))
       ((< x-pos min) (list min (car (y-transfer (timebpf self) min 0))
                                   (round (- min x-pos) (staff-size self)))
        (round (- (car (y-transfer (timebpf self) min 0)) (round (- min x-pos) (staff-size self)))))
       (t (round (car (y-transfer (timebpf self) x-pos 0))))))
    (round (* x-pos (staff-size self)))))



(defmethod get-sheet-event-pos ((self sheetPanel) (panel sheet-scoreobjectframe) time)
   (- (get-x-pos self time) (sheet-key-space self)))

(defmethod get-sheet-event-pos ((self sheetPanel) (panel t) time)
   (get-x-pos self time) )

(defmethod get-sheet-temporal-objects (temp-list)
  (let ((maxdur 0)
        serie? rep)
    (loop for item in temp-list do
          (if serie?
            (if (= (first item) (first (car serie?)))
              (progn
                (push item serie?)
                (setf maxdur (max maxdur (space-size (second item)))))
              (progn
                (setf rep (append rep (put-maxdur (sort serie? 'measures-first :key 'second) maxdur)))
                (setf serie? (list item))
                (setf maxdur (space-size (second item) ))))
            (progn 
              (setf serie? (list item))
              (setf maxdur (max maxdur (space-size (second item)))))))
    (setf rep (append rep (put-maxdur (sort serie? 'measures-first :key 'second) maxdur)))
    rep))

(defmethod space-sheet-tracks ((self sheetPanel)) 
  (let (temp-list gobjs rep)
    (setf gobjs (loop for event in (sheet-get-event-panels self)
                      collect (list (graphic-obj event) (start-t (reference event)) (end-t (reference event)))))
    (setf temp-list (sort  (loop for item in gobjs  append 
                                 (append nil
                                         (collect-temporal-objects  (car item) (second item)))) '< :key 'car ) )
    (setf rep (get-sheet-temporal-objects temp-list))
    (space-objects rep (staff-size self))
    
    (setf (timebpf self) (make-sheet-bpf-time  self (append '((0 0)) (loop for item in gobjs append 
                                                                           (collect-bpftime-objects (car item) (second item) (staff-size self))))))
    
    (loop for item in gobjs do
          (set-graph-rectangles (car item)))
    (setf *a* (timebpf self))))


(defmethod sheet-key-space ((self sheetPanel))
   (*  (staff-size self) 3))

(defmethod make-sheet-bpf-time ((self sheetPanel) list)
  (let (lx ly)
    (setf list (sort (remove-duplicates list :key 'first) '< :key 'first))
    (loop for item in list do
          (push (first item) lx)
          (push (second item) ly))
    (simple-bpf-from-list (reverse (cons (+ 1000 (car lx)) lx)) (om+ (sheet-key-space self) (reverse (cons (+ 20 (car ly)) ly))))))

;=======

(defmethod measures-first ((a grap-measure) (b grap-start-event))  nil)
(defmethod measures-first ((a grap-start-event) b)  t)

(defmethod space-size ((self grap-start-event) ) 24)
(defmethod space-size ((self grap-end-event) ) 0)


;end
(defmethod space-girl ((obji+1 t) (onseti+1 integer) (obji grap-end-event) (onseti integer) count size maxdur) 
  (setf (main-point obji) (list (+ count  maxdur) (second (main-point obji))))
  (if (zerop (- onseti+1 onseti))
    count
    (+ count maxdur   )))

(defmethod space-girl ((obji+1 grap-end-event) (onseti+1 integer) (obji t) (onseti integer) count size maxdur)
  (setf (main-point obji) (list (+ count  (*  size 3) ) (second (main-point obji))))
  (+ count  (*  size 3) ))

(defmethod space-girl ((obji+1 null) onseti+1 (obji grap-end-event)  onseti  count size maxdur)
  (+ count  (*  size 3) ))

;=================
; UPDATE PANEL

;;; met a jour les positions et tailles des tracks et des objets
(defmethod update-panel ((self sheetPanel) &optional (updateref nil))
  ;(print "UPDATE")
  ;pas tout le temps
  (loop for panel in (sheet-get-event-panels self) do
        (update-the-panel panel))
  ;pas tout le temps
  (space-sheet-tracks self)
  (let ((inter (inter (params (reference self))))
        (spacelist (trackspos (params (reference self))))
        (pos (inter (params (reference self)))))
    (loop for trp in (trackpanels self)
          for sp in spacelist do
          (setf pos (+ pos (* sp inter)))
          (om-set-view-position trp (om-make-point 0 pos))
          (om-set-view-size trp (om-make-point (w trp) (* (font-size (params (reference self))) (size (reference trp)))))
          (setf pos (+ pos inter (h trp)))
          (loop for ov in (track-subviews trp) do 
                (om-set-view-position ov (om-make-point (get-sheet-event-pos self ov (start-t (reference ov))) 3))
                (om-set-view-size ov (om-make-point (- (get-x-pos self (end-t (reference ov))) (get-sheet-event-pos self ov (start-t (reference ov)))) (- (h trp) 6)) )
                (when (in-patch-mode? self) (redraw-connections ov)) 
                ))
    (when (in-patch-mode? self) 
      (loop for box in (get-subframes self) do
            (redraw-connections box)
            (update-boxes-connections self (get-subframes self)))
      ;(draw-sheet-connections self)
      )
    (om-invalidate-view self t)))
    
(defmethod sheet-get-event-panels ((self sheetPanel))
  (loop for item in (trackpanels self) append (om-subviews item)))

(defmethod update-the-panel ((self sheet-scoreobjectframe) &optional (updateref nil))
  (set-editor-tonality (panel self))
  (let ((linespace (/ (staff-size self) 4)))
    (setf (graphic-obj self) (make-graph-ryth-obj  (objectfromeditor self) (top-in-midi (staff-sys self)) (staff-sys self) linespace 
                                                  (get-approx-scale self)
                                                  (selection? self)  nil nil nil
                                                  ))))

(defmethod update-the-panel ((self sheet-linobjectframe) &optional (updateref nil)) 
  (setf (graphic-obj self) (make-graph-linear-obj self)))

(defmethod make-graph-linear-obj ((self sheet-linobjectframe))
  (declare (ignore ryth))
  (let* ((newtrack (make-instance 'grap-track
                   :reference self
                    :rectangle (list 0 0 0 0)))
         (inside (make-graph-linear-bounds self )))
    (setf (inside newtrack) inside)
    newtrack))

(defmethod make-graph-linear-bounds ((self sheet-linobjectframe))
  (let* ((new-ev-s (make-instance 'grap-track-ev
                     :reference self
                     :main-point (list 0 0)
                     :start? t
                     :rectangle (list 0 0 1 1)))
         (new-ev-e (make-instance 'grap-track-ev
                     :reference self
                     :start? nil
                     :main-point (list 0 0)
                     :rectangle (list 0 0 1 1))))
    (list new-ev-s new-ev-e)))


;;;==========================
;;; DRAW

(defmethod draw-view-contents ((self sheetPanel)) 
  (when (grille-p self)
    (let ((step (tstep (params (reference self))))
          (width (w self))
          (posx (om-h-scroll-position self))
          (miny (om-v-scroll-position self))
          (maxy (+ (om-v-scroll-position self) (h self))))
      (loop while (< (get-x-pos self step) 10) do (setf step (* step 10)))
      (om-with-focused-view self
      (om-with-dashline
        (loop for i = 0 then (+ i step) while (< posx width) do
              (setf posx (get-x-pos self i))
              (om-draw-line posx miny posx maxy)))))))

(defmethod om-draw-contents ((self trackpanel))
  (om-with-focused-view self
    (if (selected self)
      (om-with-fg-color self *select-color*
        (om-fill-rect 0 0 (w self) (h self)))
      (when (show-tracks (om-view-container self))
        (om-with-fg-color self  (om-make-color 0.9 0.9 0.9)
          (om-fill-rect 0 0 (w self) (h self))))
      )
    )
  (call-next-method))
 

(defmethod sheetobject-draw ((self sheet-objectframe))
  (om-with-focused-view self
    (when (selected self)
      (om-draw-rect-outline 0 0 (w self) (h self) 2))
    (when (show-time (sheet-container self))
      (om-with-font (om-make-font *om-verdana-font* (round (staff-size (sheet-container self)) 2))
                    (om-draw-string 4 (- (h self) 2) (integer-to-string (start-t (reference self))))))
    ))

(defmethod om-draw-contents ((self sheet-scoreobjectframe))
  (call-next-method)
  (sheetobject-draw self))

(defmethod om-draw-contents ((self sheet-linobjectframe))
  (let ((sheetpanel (om-view-container (om-view-container self))))
    (om-with-focused-view self
      (when (graphic-obj self)
        (draw-track-event (objectfromeditor self) (graphic-obj self) (* -1  (sheet-offset self)) 0 (w self) 0 (h self) self 
                          (staff-zoom sheetpanel) (staff-size sheetpanel) sheetpanel)))
    (sheetobject-draw self)))

(defmethod draw-mini-view ((self sheet-linobjectframe) (value MidiFile))
   (if (midifilename value)
     (om-with-focused-view self
       (draw-mini-midi self 5000 value)) (call-next-method)))

(defmethod draw-view-contents ((self sheet-scoreobjectframe))
  (let* ((x0  (om-h-scroll-position self))
         (y0  (om-v-scroll-position self))
         (deltax (get-key-space self))
         (size (staff-size self))
         (deltay (round (* size (score-top-margin self)))))
    (om-with-focused-view self
      (om-with-font (om-make-font *heads-font* size)
                    (draw-system-only self)
                    (when (graphic-obj self)
                      (draw-object (graphic-obj self) self (- deltax (sheet-offset self))
                                   (- deltay (round (* (posy (car (staff-list (staff-sys self)))) (/ size 4))))
                                   (staff-zoom self) x0 (+ x0 (w self)) y0 (+ y0 (h self)) 
                                   (slots-mode self) size (linear? self) 
                                   (staff-sys self) nil (noteaschan? self)))))))

(defmethod sheet-offset ((self scorepanel)) 0)

(defmethod sheet-offset ((self sheet-linobjectframe)) 
  (let ((sheetpanel (om-view-container (om-view-container self))))
    (get-sheet-event-pos sheetpanel self (start-t (reference self)))))

(defmethod sheet-offset ((self sheet-scoreobjectframe)) 
  (let ((sheetpanel (om-view-container (om-view-container self))))
    (get-sheet-event-pos sheetpanel self (start-t (reference self)))))



;;;===============================

(defmethod sheet-get-allobjs ((self sheetpanel))
  (remove nil 
          (flat 
           (loop for tr in (trackpanels self) collect
                 (loop for objv in (track-subviews tr) collect objv)))))


(defmethod sheet-get-allobjs-evts ((self sheetpanel))
  (remove nil 
          (flat 
           (loop for tr in (inside (reference self)) collect
                 (loop for objv in (objs tr) collect objv)))))