(in-package :om)

(defvar  *midirealtime* "si les modifs sont envoyees au player en temps reel")
(setf *midirealtime* t)

;================================================
;=== CHANNEL CONTROLLER                       ===
;=== a single track of the general controller ===
;================================================
(defclass* Channel-Ctrl () 
  ((midiport :initform 0 :initarg :midiport :accessor midiport :type integer)
   (midichannel :initform 1 :initarg :midichannel :accessor midichannel :type integer)
   (program :initform 0 :initarg :program :accessor program :type integer)
   (pan-ctrl :initform 64 :initarg :pan-ctrl  :accessor pan-ctrl :type integer)
   (control1-num :initform 1 :initarg :control1-num :accessor control1-num :type integer)
   (control2-num :initform 2 :initarg :control2-num :accessor control2-num :type integer)
   (control1-val :initform 0 :initarg :control1-val :accessor control1-val :type integer)
   (control2-val :initform 0 :initarg :control2-val :accessor control2-val :type integer)
   (vol-ctrl :initform 100 :initarg :vol-ctrl :accessor vol-ctrl :type integer)
   (pitch-ctrl :initform 64 :initarg :pitch-ctrl :accessor pitch-ctrl :type integer)))

(defmethod channel-ctrl-p ((self channel-ctrl))  t)
(defmethod channel-ctrl-p ((self t)) nil)


;====================================
;=== SETTINGS CONTROLLER          ===
;=== a set of channel controllers ===
;====================================
(defclass* settings-ctrl (midi-score-element)
   ((midiport :initform 0 :initarg :midiport :accessor midiport :type integer)
    (nbtracks :initform 1 :initarg :nbtracks :accessor nbtracks :type integer)
    (channels-ctrl :initform nil :accessor channels-ctrl :type t))
  (:icon 918))

(defclass* midi-mix-console (settings-ctrl) ()
  (:icon 918))

(defmethod get-channel-ctrl-class ((self t)) 'channel-ctrl)

(defmethod initialize-instance :after ((self settings-ctrl) &rest l)
   (declare (ignore l))
   (if (< (nbtracks self) 1) (setf (nbtracks self) 1))
   (if (> (nbtracks self) 16) (setf (nbtracks self) 16))

   (setf (channels-ctrl self) 
         (loop for i from 1 to (nbtracks self) collect (make-instance (get-channel-ctrl-class self) 
                                                         :midiport (midiport self)
                                                         :midichannel i)))
)

(defmethod settings-ctrl-p ((self settings-ctrl))  t)
(defmethod settings-ctrl-p ((self t)) nil)

(defmethod allowed-in-maq-p ((self settings-ctrl))  t)

(defmethod get-impulsion-pict ((self settings-ctrl)) *ctrl-impulsion-pict*)

(defmethod Class-has-editor-p  ((self settings-ctrl)) t )
(defmethod get-editor-class ((self settings-ctrl)) 'controllerEditor)

(defmethod draw-mini-view  ((self t) (value settings-ctrl)) 
   (draw-obj-in-rect value 0 (w self) 0 (h self) (view-get-ed-params self) self))

(defmethod update-miniview ((self t) (value settings-ctrl)) 
   (om-invalidate-view self t))

(defmethod draw-obj-in-rect ((self settings-ctrl) x x1 y y1 edparams view)
   (om-with-focused-view view
     nil))


(defmethod omNG-copy ((self settings-ctrl))
  "Cons a Lisp expression that return a copy of self when it is valuated."
  `(let ((rep (make-instance ',(type-of self)
                :midiport ,(midiport self)
                :nbtracks ,(nbtracks self))))
     (setf (channels-ctrl rep) (list ,.(loop for ctrl in (channels-ctrl self) collect
                                             (omNG-copy ctrl))))
     rep
     ))

(defmethod copy-container  ((self settings-ctrl) &optional (pere nil))
  "Cons a Lisp expression that return a copy of self when it is valuated."
  (let ((rep (make-instance (type-of self)
                :midiport (midiport self)
                :nbtracks (nbtracks self))))
    (setf (channels-ctrl rep) (loop for ctrl in (channels-ctrl self) collect
                                    (eval (omNG-copy ctrl))))
    rep
    ))

(defmethod omNG-save ((self settings-ctrl) &optional (values? nil))
  "Cons a Lisp expression that retunr a copy of self when it is valuated."
  `(let ((rep (make-instance ',(type-of self) 
                :midiport ,(midiport self) 
                :nbtracks ,(nbtracks self))))
     (setf (channels-ctrl rep) (list ,.(loop for ctrl in (channels-ctrl self) collect
                                  (omNG-save ctrl))))
     rep
     ))


(defmethod get-obj-dur ((self settings-ctrl)) 0)


;=========== CONTROLLER EDITOR ===================
;=== The Editor will be a scrollable editor with fixed size 
;(defmethod get-win-ed-size ((self settings-ctrl)) (om-make-point (+ 6 (min (* *channel-w* 6) (* *channel-w* (nbtracks self)))) 560))
(defmethod get-win-ed-size ((self settings-ctrl)) (om-make-point (min (* *channel-w* 10) (* *channel-w* (nbtracks self))) 560))

(defmethod make-editor-window ((class (eql 'controllerEditor)) object name ref &key 
                                 winsize winpos (close-p t) (winshow t) 
                                 (resize nil) (maximize nil))
   (let ((win (call-next-method class object name ref :winsize winsize :winpos winpos :resize nil 
                                                      :close-p t :winshow t
                                                      )))
    win))


(omg-defclass controllerEditor (EditorView) 
  ((ch-panels :initform nil :accessor ch-panels :type list)))

(defmethod editor-has-palette-p ((self controllerEditor)) nil)

(defmethod get-panel-class ((Self controllerEditor)) 'controllerPanel)

(defmethod update-subviews ((Self controllereditor))
   (om-set-view-size (panel self ) (om-make-point (w self) (h self)))
   ;;;(om-update-scroll-bar-limits (panel self )) ;;; mettre dans om-set-field-size pour mac
   ;(om-invalidate-view self t)
   )


;=== MAIN PANEL ===
(omg-defclass controllerPanel (om-scroller) ()
  ;;;(:default-initargs :scrollbars :h :retain-scrollbars t)
   )

(defmethod get-object ((Self controllerPanel))
   (object (om-view-container self)))

(defmethod report-modifications ((self controllerPanel))
  (report-modifications (om-view-container self)))
                                   

;======== CHaNNeL ConTrOllEr PAneL =====

(omg-defclass channelPanel (om-View) 
  ((channelctr :initform nil :initarg :channelctr :accessor channelctr)
   (channelText :initform nil :accessor channelText :type t)
   (channelBox :initform nil :accessor channelBox :type t)
   (programMenu :initform nil :accessor programMenu :type t)
   (volumeText :initform nil :accessor volumeText :type t)
   (volumeVal :initform nil :accessor volumeVal :type t)
   (volumpeSlider :initform nil :accessor volumeSlider :type t)
   (pitchText :initform nil :accessor pitchText :type t)
   (pitchVal :initform nil :accessor pitchVal :type t)
   (pitchSlider :initform nil :accessor pitchSlider :type t)
   (panText :initform nil :accessor panText :type t)
   (panVal :initform nil :accessor panVal :type t)
   (panSlider :initform nil :accessor panSlider :type t)
   (ctrl1menu :initform nil :accessor ctrl1menu :type t)
   (ctrl1Val :initform nil :accessor ctrl1Val :type t)
   (ctrl1Slider :initform nil :accessor ctrl1Slider :type t)
   (ctrl2menu :initform nil :accessor ctrl2menu :type t)
   (ctrl2Val :initform nil :accessor ctrl2Val :type t)
   (ctrl2Slider :initform nil :accessor ctrl2Slider :type t)
   (resetbutton :initform nil :accessor resetbutton :type t)
))

(defmethod update-subviews ((Self channelpanel))
   (om-set-view-size (panel self ) (om-make-point (w self) (h self)))
   (om-invalidate-view self t))

(defmethod om-draw-contents ((self channelPanel))
   (call-next-method))


(defmethod get-object ((Self channelPanel))
   (get-object (om-view-container self)))

(defmethod report-modifications ((self channelPanel))
  (report-modifications (om-view-container self)))

(defmethod get-channelpanel-class ((self controllerpanel)) 'channelpanel)

(defvar *channel-w* 140)
(setf *channel-w* 80)

;=======================
;=== INITIALIZATIONS ===
;=======================

(defmethod metaobj-scrollbars-params ((self controllerEditor))  '(:h t))

(defmethod initialize-instance :after ((self controllerEditor) &rest l)
   (declare (ignore l))
   
   (setf (panel self) (om-make-view (get-panel-class self) 
                                                     :owner self
                                                     :position (om-make-point 0 0) 
                                                     :scrollbars (first (metaobj-scrollbars-params self))
                                                     :retain-scrollbars (second (metaobj-scrollbars-params self))
                                                     :field-size  (om-make-point (* *channel-w* (nbtracks (object self))) 540)
                                                     ;;;:size (om-make-point (w self) (- (h self) 15)))
                                                     :size (om-make-point (w self) (h self)))
      )
   
   (setf (ch-panels self) 
      (loop for chctrl in (channels-ctrl (object self))
             for i = 0 then (+ i 1) collect
                (om-make-view (get-channelpanel-class (panel self))
                 :channelctr chctrl
                 :owner (panel self)
                 :position (om-make-point (* i *channel-w*) 0) 
                 :size (om-make-point *channel-w* (h self)))))
   
)


(defmethod update-editor-after-eval ((self controllereditor) val)
  (setf (object self) val)
   (om-set-view-size (window self) (get-win-ed-size (object self)))
   (om-set-field-size (panel self) (om-make-point (* *channel-w* (nbtracks (object self))) 540))
   (loop for chpan in (ch-panels self) do 
             (om-remove-subviews (panel self) chpan))
   (setf (ch-panels self) 
      (loop for chctrl in (channels-ctrl (object self))
             for i = 0 then (+ i 1) collect
                (om-make-view (get-channelpanel-class (panel self))
                                             :channelctr chctrl
                                             :owner (panel self)
                                             :position (om-make-point (* i *channel-w*) 0) 
                                             :size (om-make-point *channel-w* (- (h self) 15)))))
)


(defmethod simple-controls-only ((self channelpanel)) nil)


(defmethod initialize-instance :after ((self channelPanel) &rest l)
  (declare (ignore l))
  (do-initialize-channel self))

(defmethod do-initialize-channel ((self channelPanel))  
  (let* ((progList *midi-programs*)
         (ctrlList *midi-ctrl-chge*) bar1 bar2 bar3 bar4
         (pos 0)
         ctrlchgText)
    
    (setf pos (+ pos 10))
    (setf (channelText self) (om-make-dialog-item 'om-static-text
                                                  (om-make-point 5 pos) 
                                                  (om-make-point 60 16) "Channel" 
                                                  :font *om-default-font2*))
    
    (setf (channelBox self) (om-make-dialog-item 'numBox (om-make-point 60 pos) (om-make-point 16 16) 
                                                 (format nil "~D" (midichannel (channelctr self)))
                                                 :min-val 1
                                                 :max-val 16
                                                 :value (midichannel (channelctr self))
                                                 :afterfun #'(lambda (item)
                                                               (change-channel self (value item)))
                                                 :font *om-default-font2*
                                                 ))
    
    (om-set-part-color (channelBox self) :body *om-white-color*)
    
    (setf pos (+ pos 40))
    (setf (programMenu self) 
          (om-make-dialog-item 'om-pop-up-dialog-item 
                               (om-make-point 5 pos) 
                               (om-make-point 70 12)
                               ""
                               :di-action (om-dialog-item-act item
                                            (change-program self (second (nth (om-get-selected-item-index item) progList))))
                               :font *om-default-font1*
                               :range (loop for item in progList collect (first item))
                               :value (first (nth (program (channelctr self)) progList))
                               )
          )
    
    ;;;(set-part-color (programMenu self) :frame *om-gray-color*)
    
    (setf pos (+ pos 30))
    (setf (panText self) (om-make-dialog-item 'om-static-text
                                              (om-make-point 16 pos) 
                                              (om-make-point 40 16)
                                              "Pan"
                                              :font *om-default-font2*
                                              ))
    
    (setf (panVal self) (om-make-dialog-item 'om-static-text
                                             (om-make-point 45 pos) 
                                             (om-make-point 30 16)
                                             (pan2str (pan-ctrl (channelctr self)))
                                             :font *om-default-font2*
                                             ))
    
    (setf pos (+ pos 20))
    (setf (panSlider self) 
          (om-make-view 'graphic-numbox :position (om-make-point 30 pos) 
                        :size (om-make-point 20 20) ;;; (format nil "~D" zoom)
                        :pict (om-load-and-store-picture 'dial 'kernel)
                        :nbpict 65
                        :pict-size (om-make-point 24 24)
                        :di-action (om-dialog-item-act item
                                     (change-pan self (value item)))
                        :font *om-default-font2*
                        :value (pan-ctrl (channelctr self))
                        :min-val 0
                        :max-val 127))
    
    (setf pos (+ pos 26))
    (setf bar1 (om-make-view 'bar-item 
                             :position (om-make-point 4 pos) 
                             :size (om-make-point 72 10)
                             :bg-color (om-get-bg-color self)))
    
    (setf pos (+ pos 5))
    (setf (volumeText self) (om-make-dialog-item 'om-static-text 
                                                 (om-make-point 16 pos) 
                                                 (om-make-point 40 16)
                                                 "Vol"
                                                 :font *om-default-font2*
                                                 ))
    
    (setf (volumeVal self) (om-make-dialog-item 'om-static-text 
                                                (om-make-point 45 pos) 
                                                (om-make-point 30 16)
                                                (format nil "~D" (vol-ctrl (channelctr self)) )
                                                :font *om-default-font2*
                                                ))
    
    (setf pos (+ pos 20))
    (setf (volumeSlider self) (om-make-dialog-item 'om-slider  
                                                   (om-make-point 30 pos) 
                                                   (om-make-point 20 100) ""
                                                   :di-action (om-dialog-item-act item
                                                                
                                                                (change-volume self (om-slider-value item)))
                                                   :increment 1
                                                   :range '(0 127)
                                                   :value (vol-ctrl (channelctr self))
                                                   :direction :vertical
                                                   :tick-side :none
                                                   ))
    
    
    
    (unless (simple-controls-only self)

      (setf pos (+ pos 110))
      (setf bar2 (om-make-view 'bar-item 
                               :position (om-make-point 4 pos) 
                               :size (om-make-point 72 10)
                               :bg-color (om-get-bg-color self)
                               ))
      
      (setf pos (+ pos 4))
      
      (setf (pitchText self) (om-make-dialog-item 'om-static-text
                                                  (om-make-point 16 pos) 
                                                  (om-make-point 40 16)
                                                  "Pitch"
                                                  :font *om-default-font2*
                                                  ))
      
      (setf (pitchVal self) (om-make-dialog-item 'om-static-text 
                                                 (om-make-point 50 pos) 
                                                 (om-make-point 30 16)
                                                 (format nil "~D" (- (pitch-ctrl (channelctr self)) 64) )
                                                 :font *om-default-font2*
                                                 ))
      
      (setf pos (+ pos 26))
      (setf (pitchSlider self) (om-make-dialog-item 'om-slider  
                                                    (om-make-point 5 pos) 
                                                    (om-make-point 70 20)
                                                    ""
                                                    :di-action (om-dialog-item-act item
                                                                 (change-pitchbend self (om-slider-value item)))
                                                    :increment 1
                                                    :range '(0 127)
                                                    :value (pitch-ctrl (channelctr self))
                                                    :direction :horizontal
                                                    :tick-side :none
                                                    ))
      
      (setf pos (+ pos 40))
      (setf bar3 (om-make-view 'bar-item 
                               :position (om-make-point 4 pos) 
                               :size (om-make-point 72 10)
                               :bg-color (om-get-bg-color self)
                               ))
      
      (setf pos (+ pos 10))
      
      (setf ctrlchgText (om-make-dialog-item 'om-static-text
                                             (om-make-point 10 pos) 
                                             (om-make-point 70 16)
                                             "CtrlChange"
                                             :font *om-default-font2*
                                             ))
      
      (setf pos (+ pos 20))
      
      
      (setf (ctrl1Menu self) (om-make-dialog-item 'om-pop-up-dialog-item 
                                                  (om-make-point 5 pos) 
                                                  (om-make-point 70 12) 
                                                  ""
                                                  :di-action (om-dialog-item-act item
                                                               (change-ctrl1-num self (second (nth (om-get-selected-item-index item) ctrlList))))
                                                  :font *om-default-font1*
                                                  :range (loop for item in ctrlList collect (first item))
                                                  :value (first (nth (control1-num (channelctr self)) ctrlList))
                                                  ))
      
      (setf pos (+ pos 25))
      (setf (ctrl1Slider self) (om-make-dialog-item 'om-slider  
                                                    (om-make-point 5 pos) 
                                                    (om-make-point 70 20) 
                                                    ""
                                                    :di-action (om-dialog-item-act item
                                                                 (change-ctrl1-val self (om-slider-value item)))
                                                    :increment 1
                                                    :range '(0 127)
                                                    :value (control1-val (channelctr self))
                                                    :direction :horizontal
                                                    :tick-side :none
                                                    ))
      
      (setf pos (+ pos 20))
      (setf (ctrl1Val self) (om-make-dialog-item 'om-static-text 
                                                 (om-make-point 10 pos) 
                                                 (om-make-point 30 16)
                                                 (format nil "~D" (control1-val (channelctr self)))
                                                 :font *om-default-font2*
                                                 ))
      
      (setf pos (+ pos 20))
      (setf (ctrl2Menu self) (om-make-dialog-item 'om-pop-up-dialog-item 
                                                  (om-make-point 5 pos) 
                                                  (om-make-point 70 12)
                                                  ""
                                                  :di-action (om-dialog-item-act item
                                                               (change-ctrl2-num self (second (nth (om-get-selected-item-index item) ctrlList))))
                                                  :font *om-default-font1*
                                                  :range (loop for item in ctrlList collect (first item))
                                                  :value (first (nth (control2-num (channelctr self)) ctrlList))
                                                  ))
      
      (setf pos (+ pos 30))
      (setf (ctrl2Slider self) (om-make-dialog-item 'om-slider  
                                                    (om-make-point 5 pos) 
                                                    (om-make-point 70 20)
                                                    ""
                                                    :di-action (om-dialog-item-act item
                                                                 (change-ctrl2-val self (om-slider-value item)))
                                                    :increment 1
                                                    :range '(0 127)
                                                    :value (control2-val (channelctr self))
                                                    :direction :horizontal
                                                    :tick-side :none
                                                    ))
      
      (setf pos (+ pos 20))
      (setf (ctrl2Val self) (om-make-dialog-item 'om-static-text 
                                                 (om-make-point 10 pos) 
                                                 (om-make-point 30 16)
                                                 (format nil "~D" (control2-val (channelctr self)))
                                                 :font *om-default-font2*
                                                 ))
      
      (setf pos (+ pos 20))
      (setf bar4 (om-make-view 'bar-item 
                               :position (om-make-point 4 pos) 
                               :size (om-make-point 72 10)
                               :bg-color (om-get-bg-color self)
                               ))

      (setf pos (+ pos 10))
    
    
      (setf (resetbutton self) (om-make-dialog-item 'om-button
                                                    (om-make-point 15 pos) 
                                                    (om-make-point 50 16)
                                                    "Reset"
                                                    :di-action (om-dialog-item-act item
                                                                 (reset-all-values self))
                                                    :font *om-default-font1*)  
            
            )   ;;; fin unless simple-controls-only
      
      
      )
    
    (om-add-subviews self (channelText self) (channelBox self) 
                     (programMenu self) 
                     bar1 
                     (volumeSlider self) 
                     (volumeText self) (volumeVal self)
                     (panSlider self) 
                     (panText self) 
                     (panVal self)
                     )
    
    (unless (simple-controls-only self)
      (om-add-subviews self bar2 bar3 bar4
                       (pitchSlider self) 
                       (pitchText self) (pitchVal self)
                       (ctrl1Slider self) 
                       (ctrl1Menu self) 
                       (ctrl1Val self)
                       (ctrl2Slider self) 
                       (ctrl2Menu self) 
                       (ctrl2Val self)
                       ctrlchgText
                       (resetbutton self)
                       ))
    
    ))



;==============================
;=== ACTIONS ON CHANNEL PANELS:
;==============================

(defmethod change-channel ((self channelPanel) value)
  (setf (midichannel (channelctr self)) value)
  (report-modifications self))

(defmethod change-program ((self channelPanel) value)
  (setf (program (channelctr self)) value)
  (when *midirealtime* (channel-send-prog (channelctr self)))
  (report-modifications self))

(defmethod change-volume ((self channelPanel) value)
  (setf (vol-ctrl (channelctr self)) value)
  (when *midirealtime* (channel-send-vol (channelctr self)))
  (let ((new-str (integer-to-string value))
        (target (volumeVal self)))
    (unless (string= new-str (om-dialog-item-text target))
      (om-set-dialog-item-text target new-str)
      (om-redraw-view target)
      ))
  (report-modifications self))

(defun pan2str (panvalue)
  (let* ((value (- panvalue 64))
        (new-str (cond ((= value 0) (integer-to-string value))
                        ((< value 0) (format nil "L~D" (- value)))
                        ((> value 0) (format nil "R~D" value)))))
    new-str))

(defmethod change-pan ((self channelPanel) value)
  (setf (pan-ctrl (channelctr self)) value)
  (when *midirealtime* (channel-send-pan (channelctr self)))
  (let* ((target (panVal self))
         (new-str (pan2str value)))
    (unless (string= new-str (om-dialog-item-text target))
      (om-set-dialog-item-text target new-str)
      (om-redraw-view target)))
  (report-modifications self))

(defmethod change-ctrl1-val ((self channelPanel) value)
  (setf (control1-val (channelctr self)) value)
  (when *midirealtime* (channel-send-ct1 (channelctr self)))
  (let ((new-str (integer-to-string value))
        (target (ctrl1Val self)))
    (unless (string= new-str (om-dialog-item-text target))
      (om-set-dialog-item-text target new-str)
      (om-redraw-view target)))
  (report-modifications self))

(defmethod change-ctrl1-num ((self channelPanel) value)
  (setf (control1-num (channelctr self)) value)
  (report-modifications self))

(defmethod change-ctrl2-val ((self channelPanel) value)
  (setf (control2-val (channelctr self)) value)
  (when *midirealtime* (channel-send-ct2 (channelctr self)))
  (let ((new-str (integer-to-string value))
        (target (ctrl2Val self)))
    (unless (string= new-str (om-dialog-item-text target))
      (om-set-dialog-item-text target new-str)
      (om-redraw-view target)))
  (report-modifications self))

(defmethod change-ctrl2-num ((self channelPanel) value)
  (setf (control2-num (channelctr self)) value)
  (report-modifications self))

(defmethod change-pitchbend ((self channelPanel) value)
  (setf (pitch-ctrl (channelctr self)) value)
  (when *midirealtime* (channel-send-pitch (channelctr self)))
  (let ((new-str (integer-to-string (- value 64)))
        (target (pitchVal self)))
    (unless (string= new-str (om-dialog-item-text target))
      (om-set-dialog-item-text target new-str)
      (om-redraw-view target)))
  (report-modifications self))


(defmethod reset-all-values ((self channelPanel))
  (change-pan self 64)
  (set-value (panSlider self) 64)
  (change-volume self 100)
  (om-set-slider-value (volumeSlider self) 100)
  (change-pitchbend self 64)
  (om-set-slider-value (pitchSlider self) 64)
  (change-ctrl1-val self 0)
  (om-set-slider-value (ctrl1Slider self) 0)
  (change-ctrl2-val self 0)
  (om-set-slider-value (ctrl2Slider self) 0)
)
  
 

;=================================
;=== SENDING CONTROLLER SETTINGS :
;=================================
;(defmethod handle-key-event ((self ControllerPanel) char)
;  (case char
;    (#\space 
;     ;(play (object (view-container self)) :port (midiport (object (view-container self))))
;     (send-midi-settings (object (om-view-container self)))
;     (print "done"))
;    (otherwise (om-beep))))


(defmethod send-midi-settings ((self settings-ctrl))
  (loop for chan-ctrl in (channels-ctrl self) do
        (send-midi-settings chan-ctrl)))

(defmethod channel-send-prog ((self channel-ctrl))
  (let ((event (om-midi-new-evt (om-midi-get-num-from-type "ProgChange") 
                                :chan (- (midichannel self) 1) :port (midiport self)
                                :vals (program self))))
    (when event   
      (om-midi-send-evt event *player*)
      t)
    ))
    
(defmethod channel-send-vol ((self channel-ctrl))
  (let ((event (om-midi-new-evt  (om-midi-get-num-from-type "CtrlChange") 
                                 :chan (- (midichannel self) 1) :port (midiport self)
                                 :vals (list 7  (vol-ctrl self)))))
    (when event   
      (om-midi-send-evt event *player*)
      t)
    ))

(defmethod channel-send-pan ((self channel-ctrl))
  (let ((event (om-midi-new-evt  (om-midi-get-num-from-type "CtrlChange") 
                                 :chan (- (midichannel self) 1) :port (midiport self)
                                 :vals (list 10 (pan-ctrl self)))))
    (when event   
      (om-midi-send-evt event *player*)
      t)
    ))
    
(defmethod channel-send-ct1 ((self channel-ctrl))
  (let ((event (om-midi-new-evt  (om-midi-get-num-from-type "CtrlChange") 
                                 :chan (- (midichannel self) 1) :port (midiport self)
                                 :vals (list (control1-num self) (control1-val self)))))
    (when event   
      (om-midi-send-evt event *player*)
      t)
    ))
    
(defmethod channel-send-ct2 ((self channel-ctrl))
  (let ((event (om-midi-new-evt  (om-midi-get-num-from-type "CtrlChange")
                                 :chan (- (midichannel self) 1) :port (midiport self)
                                 :vals (list (control2-num self) (control2-val self)))))    
    (when event   
      (om-midi-send-evt event *player*)
      t)
    ))

(defmethod channel-send-pitch ((self channel-ctrl))
  (let ((event (om-midi-new-evt  (om-midi-get-num-from-type "PitchBend") 
                                 :chan (- (midichannel self) 1) :port (midiport self)
                                 :vals (list 0  (pitch-ctrl self)))))
    (when event   
      (om-midi-send-evt event *player*)
      t)
    ))

(defmethod send-midi-settings ((self channel-ctrl))
  (channel-send-prog self) 
  (channel-send-vol self) 
  (channel-send-pan self) 
  (channel-send-ct1 self) 
  (channel-send-ct2 self) 
  (channel-send-pitch self) 
  ) 


(defmethod* PrepareToPlay ((player t) (self settings-ctrl) at &key approx port interval voice)
  (declare (ignore approx))
  (when *player*
    (let ((evtseq (objfromobjs self (make-instance 'eventmidi-seq)))) 
      (PrepareToPlay player evtseq at 
                     :approx approx 
                     :port port
                     :interval interval
                     :voice voice)
      )))


;======================
;=== GET MIDIEVENTS ===
;======================

(defmethod! get-midievents ((self settings-ctrl) &optional test)
  (declare (ignore test))
  (let ((evtList nil))
  (loop for chan-ctrl in (channels-ctrl self) do 
        (push (get-midievents chan-ctrl) evtList))
  (flat (reverse evtList))))

(defmethod! get-midievents ((self channel-ctrl) &optional test)
  (declare (ignore test))
  (let ((evtList nil) 
        (progevt (make-instance 'MidiEvent
                       :ev-date 0
                       :ev-type (om-midi-get-num-from-type "ProgChange")
                       :ev-chan (midichannel self)
                       :ev-port (midiport self)
                       :ev-fields (program self)))
        
        (volevt (make-instance 'MidiEvent
                       :ev-date 0
                       :ev-type (om-midi-get-num-from-type "CtrlChange")
                       :ev-chan (midichannel self)
                       :ev-port (midiport self)
                       :ev-fields (list 7 (vol-ctrl self))))
        
        (panevt (make-instance 'MidiEvent
                       :ev-date 0
                       :ev-type (om-midi-get-num-from-type "CtrlChange")
                       :ev-chan (midichannel self)
                       :ev-port (midiport self)
                       :ev-fields (list 10 (pan-ctrl self))))
        
        (pitchevt (make-instance 'MidiEvent
                       :ev-date 0
                       :ev-type (om-midi-get-num-from-type "PitchBend")
                       :ev-chan (midichannel self)
                       :ev-port (midiport self)
                       :ev-fields (list 0 (pitch-ctrl self))))
        
        (ctr1evt (make-instance 'MidiEvent
                       :ev-date 0
                       :ev-type (om-midi-get-num-from-type "CtrlChange")
                       :ev-chan (midichannel self)
                       :ev-port (midiport self)
                       :ev-fields (list (control1-num self) (control1-val self))))
        
        (ctr2evt (make-instance 'MidiEvent
                       :ev-date 0
                       :ev-type (om-midi-get-num-from-type "CtrlChange")
                       :ev-chan (midichannel self)
                       :ev-port (midiport self)
                       :ev-fields (list (control2-num self) (control2-val self))))
        )
        
    (setf evtlist (list progevt volevt panevt pitchevt ctr1evt ctr2evt))

    evtList))




;;;;;====================
;;; SIMPLE MIDI MIXER
;;;;;====================

(defclass* simple-midi-mix-console (midi-mix-console) ()
  (:icon 918))

(defmethod get-channel-ctrl-class ((self simple-midi-mix-console)) 'simple-channel-ctrl)

(defclass* simple-Channel-Ctrl (channel-ctrl) ())

(defmethod send-midi-settings ((self simple-channel-ctrl))
  (channel-send-prog self) 
  (channel-send-vol self) 
  (channel-send-pan self) 
  )



(defmethod! get-midievents ((self simple-channel-ctrl) &optional test)
  (declare (ignore test))
  (let ((evtList nil) 
        (progevt (make-instance 'MidiEvent
                       :ev-date 0
                       :ev-type (om-midi-get-num-from-type "ProgChange")
                       :ev-chan (midichannel self)
                       :ev-port (midiport self)
                       :ev-fields (program self)))
        
        (volevt (make-instance 'MidiEvent
                       :ev-date 0
                       :ev-type (om-midi-get-num-from-type "CtrlChange")
                       :ev-chan (midichannel self)
                       :ev-port (midiport self)
                       :ev-fields (list 7 (vol-ctrl self))))
        
        (panevt (make-instance 'MidiEvent
                       :ev-date 0
                       :ev-type (om-midi-get-num-from-type "CtrlChange")
                       :ev-chan (midichannel self)
                       :ev-port (midiport self)
                       :ev-fields (list 10 (pan-ctrl self))))
        
        )
        
    (setf evtlist (list progevt volevt panevt))

    evtList))

;;;=== EDITOR ===

(defmethod get-win-ed-size ((self simple-midi-mix-console)) (om-make-point (+ 6 (min (* *channel-w* 6) (* 140 (nbtracks self)))) 340))

(defmethod make-editor-window ((class (eql 'simplecontrollerEditor)) object name ref &key 
                                 winsize winpos (close-p t) (winshow t) 
                                 (resize nil) (maximize nil))
   (let ((win (call-next-method class object name ref :winsize winsize :winpos winpos :resize nil 
                                                      :close-p t :winshow t
                                                      )))
    win))

(omg-defclass simplecontrollerEditor (controllerEditor) ())
(omg-defclass simplecontrollerPanel (controllerPanel) ())
(omg-defclass simplechannelpanel (channelpanel) ())

(defmethod get-editor-class ((self simple-midi-mix-console)) 'simplecontrollerEditor)
(defmethod get-panel-class ((Self simplecontrollerEditor)) 'simplecontrollerPanel)
(defmethod get-channelpanel-class ((self simplecontrollerPanel)) 'simplechannelpanel)


(defmethod simple-controls-only ((self simplechannelpanel)) t)


(defmethod reset-all-values ((self simplechannelPanel))
  (change-pan self 64)
  (set-value (panSlider self) 64)
  (change-volume self 100)
  (om-set-slider-value (volumeSlider self) 100)
)
;;;===================