(in-package :om)


(defmethod get-score-player ((self scorepanel)) 
  (if (micro-player? self) 'microplayer t))

(when *osc*
  (add-assoc-player *general-player* 'microplayer))


;=====================================================
;MICROPLAYER

(defvar *microosc-packets* nil)
(defvar *index-packets* 0)

(defvar *microplayer-port* nil)
(setf *microplayer-port* 3000)

(defvar *microplayer-host* nil)
(setf *microplayer-host* "127.0.0.1")

;; (defvar *max-buf-size* nil)
;; definie dans l'api
(setf *om-udp-max-buf-size* 500)



;================
(defun micro-reset ()
  (om-send-osc-bundle *microplayer-port* *microplayer-host*  '(("/play.t/reset")))
  (setf *microosc-packets* nil))

;================
(defun micro-start ()
    (om-send-osc-bundle *microplayer-port* *microplayer-host*  '(("/play.t/start"))))

;================
(defun sort-micro-events ()
  (setf *microosc-packets* (sort *microosc-packets* '< :key 'second)))

;================


(defun send-200 ()
  (loop for i from 1 to 200
        while (< *index-packets* (length *microosc-packets*)) do
        (let ((event (copy-list (nth  *index-packets* *microosc-packets*))))
          (unless (zerop *index-packets*)
            (setf (nth 1 event) (- (nth 1 event) (second (nth  (- *index-packets* 1) *microosc-packets*)))))
          (setf *index-packets* (+ *index-packets* 1))
          (om-send-osc-bundle *microplayer-port* *microplayer-host*  (list event))))
  (when (= *index-packets* (length *microosc-packets*))
    ( kill-micro-receive)))

;================
(defvar *micro-listen-process* nil)
(defvar *micro-listen-buff* nil)
(defvar *micro-listen-ep* nil)

;================
;(micro-listen)
(defun micro-listen ()
  (setf *micro-listen-buff* (om-make-pointer *om-udp-max-buf-size*))
  (setf *micro-listen-ep* (om-make-receive-connection 3010))
  (setf *micro-listen-process* (om-run-process 
                                "micro-listen" 
                                #'start-micro-receive 
                                (list *micro-listen-ep* *micro-listen-buff*))))


;(micro-listen)
;================
;( kill-micro-receive)
(defun kill-micro-receive ()
  (when *micro-listen-process*
  (om-kill-process *micro-listen-process*)
  (setf *micro-listen-process* nil)
  (om-close-udp-connection *micro-listen-ep*)
  (setf *micro-listen-ep* nil)
  (om-free-pointer *micro-listen-buff*)))

;================
(defun start-micro-receive (endpoint buffer)
  (let ((result nil))
    (loop while t do
          (let ((bufsize (om-receive-udp-packet endpoint buffer *om-udp-max-buf-size*)) bundle)
            (setf bundle (parse-buffer buffer 0 bufsize))
            (send-200)))))


(defmethod PlayANy ((player (Eql 'microplayer)) (object t) &key (approx 2) (port nil) (interval nil))
  (let ((objdur (get-obj-dur object))
        (avance (real-duration  object 0)))
    (when (second interval)
      (setf objdur (- (second interval) (first interval))))
    (setf avance (if (< avance 0) (abs avance) 0))
    (reset-player *general-player*)
    (setf (playedobjdur *general-player*) objdur)
    (micro-reset)
    (setf (playedobjdur *general-player*) objdur)
    (setf *microosc-packets* nil)
    (setf *MidiShare-start-time* 1)
    (when *player*
      (om-midi-set-player *player* (om-midi-new-seq) 1000))
    (PrepareToPlay 'microplayer object avance 
                   :approx approx 
                   :port port
                   :interval interval)
    (sort-micro-events)
    (play-player *general-player*)))


(defmethod* PrepareToPlay ((player (Eql 'microplayer)) (self t) at &key  approx port interval voice)
   (declare (ignore approx))
   (call-next-method))

(defmethod* PrepareToPlay ((player (Eql 'microplayer)) (self measure) at &key approx port interval voice)
   (setf port (verify-port port))
   (loop for sub in (inside self) do
         (let ((objstart (+ at (offset->ms sub))))
           (if interval
             (let ((newinterval (interval-intersec interval 
                                                   (list objstart (+ objstart (get-obj-dur sub))))))
               (when newinterval
                 (PrepareToPlay player sub objstart 
                                :approx approx 
                                :port port
                                :interval interval
                                :voice voice)))
             (PrepareToPlay player sub objstart 
                            :approx approx 
                            :port port
                            :voice voice)))))

(defmethod* PrepareToPlay ((player (Eql 'microplayer)) (self note) at &key  approx port interval voice)
   (when (not (memq (tie self) '(continue end)))
     (let ((chan (chan self))
           (pitch (/ (approx-m (midic self) approx) 100.0))
           (vel (vel self))
           (dur (- (real-dur self) 2))
           (date (+  *MidiShare-start-time* at)))
       (if interval
         (let ((newinterval (interval-intersec interval (list at (+ at (- (real-dur self) 1)))))) 
           (when newinterval
             (playoscnote chan pitch vel (- (second newinterval) (first newinterval) 1) 
                       (- (+  *MidiShare-start-time* (first newinterval)) 
                          (first interval)))))
         (playoscnote chan pitch vel dur date)))))



(defun playoscnote (chan pitch vel dur date)
   (push (list "/play.t/fifos" date  pitch vel  dur chan) *microosc-packets*))


(defmethod Play-player ((self (eql 'microplayer)))
   (micro-listen) 
   (setf *index-packets* 0) 
   (send-200) 
   (micro-start) ) 


(defmethod Continue-Player ((self (eql 'microplayer)))
    (om-send-osc-bundle *microplayer-port* *microplayer-host*  '(("/play.t/continue"))))

(defmethod Pause-Player ((self (eql 'microplayer)))
  (om-send-osc-bundle *microplayer-port* *microplayer-host*  '(("/play.t/pause"))))

(defmethod Stop-Player ((self (eql 'microplayer)) &optional view)
   (declare (ignore view))
   (micro-reset) (kill-micro-receive))

(defmethod Reset-Player ((self (eql 'microplayer)) &optional view)
   (declare (ignore view))
   (micro-reset) (kill-micro-receive))





;=====

(defvar *micro-player-app* "the id of the midishare-setup application")
(defvar *micro-player-path*)

(defun init-microplayer ()
  (setf *micro-player-path* (om-external-app "microplayer")))

(om-add-init-func 'init-microplayer)

;BUGGGGG
(defun launch-microplayer-app ()
  (unless (and *micro-player-app* (om-find-process *micro-player-app*))
    (let ((path *micro-player-path*))
      (setf *micro-player-app* (om-run-application path ))
      ))
  )

