(in-package :om)


(defmethod! chant-synth ((self sdiffile) outfile &optional (channels 1) (sr 44100))
  :icon 607
  (let* ((path-sdif (filepathname self))
        (outpath (or (pathname outfile) (om-choose-new-file-dialog)))
        (path-temp (make-pathname :device (pathname-device outpath) :directory (pathname-directory outpath)
                                    :name (pathname-name outpath) :type "tmp")))
  (if (probe-file (om-external-app "chant"))
    (if (probe-file path-sdif) 
      (progn
        (print (format nil "Running CHANT with ~D" (om-path2cmdpath path-sdif)))
        (om-cmd-line (print (format nil "~s -v -i~S | ~s -of -c~D -R~D ~S" 
                             (om-path2cmdpath (om-external-app "chant"))
                             (om-path2cmdpath path-sdif) 
                             (om-path2cmdpath (om-external-app "tosf"))
                             channels sr
                             (om-path2cmdpath path-temp))))
        (print "End synthesis")
        (svp-rescale path-temp outpath)
        (print (om-namestring outpath))
        )
      (print "error : bad pathname"))
    (print "CHANT is not available"))))



(defun numfromsymb (symb)
  (cond 
   ((string-equal symb "FOB") 0)
   ((string-equal symb "REB") 1)
   ((string-equal symb "NOI") 2)))

(defun chant-write-one-IDS (file sid patch type nbelt size a-time time)
  (write-1IDS-table file (numfromsymb type) "Chant" 
                    (format nil "Patch~D/~D/~D/~D/~D/~F/~F"
                            sid
                            patch
                            type
                            nbelt
                            size
                            a-time
                            time )))

(defmethod! chant-write-sdif-NVT (stream synthtime &optional (ch-num 1) (sr 44100))
  :icon 908
  (write-1nvt-table stream
                    (list "BufferSize" "NumberOfChannels" "EndTime" "SdifTypesVersion" "ChantLibraryVersion" "Author" "SamplingRate")
                    (list "512" (format nil "~D" ch-num) (format nil "~D" synthtime) "1.01.ppc" "1.01.ppc" "OM" (format nil "~F" sr)))
  )

(defmethod! chant-write-sdif-IDS (file streamid patch synthtime 
                                 &key 
                                 (nbfob 0) (nbreb 0) (nbnoi 0)
                                 (sizefob 0) (sizereb 0) (sizenoi 0)
                                 (tfob 0.0) (treb 0.0) (tnoi 0.0)
                                 (sound ""))
  :icon 908
  (cond
      ((= patch 0) 
       (chant-write-one-IDS file streamid patch "FOB" nbfob sizefob tfob synthtime))
      ((or (= patch 2) (= patch 4)) 
       (chant-write-one-IDS file streamid patch "REB" nbreb sizereb treb synthtime)
       (chant-write-one-IDS file streamid patch "NOI" nbnoi sizenoi tnoi synthtime))
      ((or (= patch 6) (= patch 7) (= patch 10)) 
       (chant-write-one-IDS file streamid patch "FOB" nbfob sizefob tfob synthtime)
       (chant-write-one-IDS file streamid patch "REB" nbreb sizereb treb synthtime))
      ((or (= patch 1) (= patch 3) (= patch 8) (= patch 9)) 
       (chant-write-one-IDS file streamid patch "FOB" nbfob sizefob tfob synthtime)
       (chant-write-one-IDS file streamid patch "REB" nbreb sizereb treb synthtime)
       (chant-write-one-IDS file streamid patch "NOI" nbnoi sizenoi tnoi synthtime))
      ((= patch 5) 
       (chant-write-one-IDS file streamid patch "REB" nbreb sizereb treb synthtime)))
  (when (or (= patch 2) (= patch 3) (= patch 5) (= patch 7) (= patch 8) (= patch 10))
    (write-1IDS-table file 3 "Chant" (format nil "Patch~D/streamid/SND/1/0/0.0/~F/~S"
                                             patch synthtime sound)))
  )



;;;=========================================

(defmethod! make-chant-sdif-file ((self sdif-buffer) patch synthtime &optional pathname
                                  &key (streamid 0) 
                                 (nbfob 0) (nbreb 0) (nbnoi 0)
                                 (sizefob 0) (sizereb 0) (sizenoi 0)
                                 (tfob 0.0) (treb 0.0) (tnoi 0.0)
                                 (sound ""))
   :icon 639
   :doc "save as SDIF file"
   (let ((outfile (or pathname (om-choose-new-file-dialog))))
     (let ((thefile (sdif-open-file (om-path2cmdpath outfile) :eWriteFile)))
         (sdif::SdifFWriteGeneralHeader thefile)
         (chant-write-sdif-NVT thefile synthtime)
         (chant-write-sdif-ids thefile streamid patch synthtime 
                               :nbfob nbfob :nbreb nbreb :nbnoi nbnoi 
                               :sizefob sizefob :sizereb sizereb :sizenoi sizenoi
                               :tfob tfob :treb treb :tnoi tnoi
                               :sound sound)
         (write-types-table thefile (list! (types self)))
         (sdif::SdifFWriteAllASCIIChunks thefile)
                               
           (loop for item in (LFrames self) do
                 (save-sdif item thefile))
           (sdif-close-file thefile))
     outfile))

;;;=========================================

(defclass! soft-sdifmatrix ()
  ((signature :initform "" :initarg :signature  :accessor signature)
   (num-elts :initform 1 :initarg :num-elts  :accessor num-elts)
   (num-fields :initform 1 :initarg :num-fields  :accessor num-fields)
   (data :initform nil :initarg :data  :accessor data)))
   


(defmethod save-sdif ((matrix soft-sdifmatrix) ptr)
  (let* ((datatype 4)
         (values (om-make-pointer (* datatype (num-fields matrix) (num-elts matrix)))))
    (loop for val in (data matrix) 
          for i = 0 then (+ i 1) do
          (om-write-ptr values (* i datatype) :single-float (coerce val 'single-float)))
    (sdif::SdifFWriteMatrix ptr (sdif-string-to-signature (signature matrix)) datatype
                            (num-elts matrix) 
                            (num-fields matrix) values)))


(defmethod calcul-sdif-size ((self soft-sdifmatrix))
   (+ (* 4 (num-elts self) (num-fields self)) 16))

;=================================================


(defmethod! transitions ((tmpobjlist list) mkev mkinter modifbeg modifend)
  (let ((sortedlist (sort tmpobjlist '< :key 'offset)))
    (flat (loop for curr-ev in sortedlist
                for prev-ev in (x-append (car sortedlist) sortedlist)
                for next-ev in (x-append (cdr sortedlist) (last-elem sortedlist))
                collect 
                (if 
                          (and (< (offset prev-ev) (offset curr-ev)) ; secu begin
                               (> (+ (offset prev-ev) (* (extend prev-ev) (strech-fact prev-ev))) (offset curr-ev)))
                          ;;; p2 > e1
                          (if (and (< (offset curr-ev) (offset next-ev)) ; secu end
                                   (> (+ (offset curr-ev) (* (extend curr-ev) (strech-fact curr-ev))) (offset next-ev)))
                            ;;; e2 > n1
                            ;;;'double-switch
                            (let ((evt (funcall mkev curr-ev)))
                              (funcall modifend 
                                       (funcall modifbeg evt (+ (offset prev-ev) (* (extend prev-ev) (strech-fact prev-ev))))
                                       (offset next-ev))
                              )
                            (if (and (< (offset curr-ev) (offset next-ev)) ; secu end
                                     (< (+ (offset curr-ev) (* (extend curr-ev) (strech-fact curr-ev))) (offset next-ev)))
                              ;;; e2 < n1
                              ;;;'begin-switch+silence
                              (let ((evt (funcall mkev curr-ev))
                                    (nex (funcall mkev next-ev)))
                                (list 
                                 (funcall modifbeg evt (+ (offset prev-ev) (* (extend prev-ev) (strech-fact prev-ev)))) 
                                 (funcall mkinter evt nex)))
                              ;;;'begin-switch
                              (let ((evt (funcall mkev curr-ev)))
                                (funcall modifbeg evt (+ (offset prev-ev) (* (extend prev-ev) (strech-fact prev-ev))))
                                )
                              )
                            )
                          (if (and (< (offset curr-ev) (offset next-ev)) ; secu end
                                   (> (+ (offset curr-ev) (* (extend curr-ev) (strech-fact curr-ev))) (offset next-ev)))
                            ;;; e2 > n1
                            ;;;'end-switch
                            (let ((evt (funcall mkev curr-ev)))
                              (funcall modifend evt (offset next-ev))
                              )
                            (if (and (< (offset curr-ev) (offset next-ev)) ; secu end
                                     (< (+ (offset curr-ev) (* (extend curr-ev) (strech-fact curr-ev))) (offset next-ev)))
                              ;;; e2 < n1
                              ;;; 'evt+silence
                              (let ((evt (funcall mkev curr-ev))
                                    (nex (funcall mkev next-ev)))
                                (list evt (funcall mkinter evt nex))
                                )
                              ;;; 'evt
                              (let ((evt (funcall mkev curr-ev)))
                                evt)
                              )
                            )
                          )
                ))))
  



(defun mk-ev2 (ev1 dur)
  (let ((new (copy-instance ev1)))
    (setf (action-time new) (+ (action-time new) dur))
    new))