;*****************************************************************************
;-------| DG SYSTEM
;-------| This file is: $LLdg/DF.ll
;-------| Version V1.0: Jan 22, 1990, V2.0, Aug 2000
;-------| By Marco Stroppa
;-------| Copyright 1990 IRCAM
;*****************************************************************************

; THIS FILES CONTAINS ALL THE FUNCTIONS THAT GENERATE DATA PASSED AS TABLE-
;    -LOOKUP FUNCTIONS THAT WILL BE SAMPLED A CERTAIN AMOUNT OF TIMES

; ADDED AUGUST 2000: VARIOUS KINDS OF INTERPOLATION (SEE FUN.LISP)

(in-package chroma)

;-----------------------------------------------------------------------------
; GLOBALS
;-----------------------------------------------------------------------------
;-----------------------------------------------------------------------------
; AVAILABLE FUNCTIONS (in alphabetical order):
;	lkp / fix-lkp	: linear table-lookup (fix-point output)
;	lkpr / fix-lkp	: linear table-lookup with rescaling of y values
;-----------------------------------------------------------------------------

;-----------------------------------------------------------------------------
; (lkp NEV FUN) / (fix-lkp NEV FUN)
; 	sample FUN NEV equally spaced times
;	FUN must be of type FUN
;	the input format is NO LONGER compatible with lkpr
; EX: (setf fun (make_fun '(0 0 1 1))) (setf a '(0 0 1 1)),
; (lkp 11 fun) ===> (0 .1 .2 .3 .4 .5 .6 .7 .8 .9 1)
;ADDED 0008, Marco: generalize other forms of interpolation

(defun lkp (nev fun &optional (exp 0.0))
  (let ((fun (copy-list fun))
        (nev (if (< nev 2) 2 nev))
        )
    (x-resc_fun fun 0 (1- nev))
    (let ((result ())
          (step 1)
          (acc 0) )
      (loop while (>= (decf nev) 0)
            do (newl result (y-val_fun fun acc exp))
            do (incf acc step))
      (nreverse result)) ) )

(defun fix-lkp (nev fun &optional (exp 0.0))
  (let ((fun (copy-list fun))
         (nev (if (< nev 2) 2 nev))
        )
    (x-resc_fun fun 0 (1- nev))
    (let ((result ())
          (step 1)
          (acc 0) )
      (loop while (>= (decf nev) 0)
            do
            (let* ((val (y-val_fun fun acc exp))
                   (sig (/ val (abs val))) )
              (newl result (truncate (+ val (* sig 0.5))) ))
            do (incf acc step))
      (nreverse result)) ) )

;-----------------------------------------------------------------------------
; (lkpr NEV FUN YMIN YMAX [OFFS]) / (fix-lkpr NEV FUN YMIN YMAX [OFFS])
; 	sample FUN NEV equally spaced times
;	Y values will be rescaled to from YMIN and YMAX and offset by OFFS
;	FUN must be of type FUN
;	to make the format compatible with lkp, if YMIN and YMAX are not
;	   specified, lkpr behaves like lkp
; EX: (setf fun (make_fun '(0 0 1 1))), (lkpr 5 fun -10 10) ===> (-10 -5 0 5 10)
(defun lkpr (nev fun ymin ymax &optional (exp 0.0) &key (offset 0.0))
  (let ((fun (copy-list fun)) )
    (y-resc_fun fun (+ ymin offset) (+ ymax offset) exp)
    (lkp nev fun exp)) )

(defun fix-lkpr (nev fun ymin ymax &optional (exp 0.0) &key (offset 0.0))
  (let ((fun (copy-list fun)) )
    (y-resc_fun fun (+ ymin offset) (+ ymax offset))
    (fix-lkp nev fun exp)) )
;-----------------------------------------------------------------------------
