#|

Further debugged and modified (defun->defmethod) by Marco, May 1998

AVAILABLE PITCH CONVERSIONS


 ---->      fq pch midi ratio itvl semitones  pch-class midic
fq          \  *    *     *    0      0           0         *
pch         *  \    *     0    *      *           *         *
midi        0  *    \     0    0      *           *         *
ratio       *  0    0     \    *      *           0         0
itvl        *  *    *     *    \      *           0         *
semitones   0  0    0     *    *      \           0         0
pch-class   0  0    *     0    0      0           \         0
midic   0  *    \     0    0      *           *         \
                                                   (* = yes , 0 = no)

|#

(in-package chroma)


;FREQUENCY
(defmethod fq->pch ((f number) &optional (approx 0))
;(om::defmethod! fq->pch ((f number) &optional (approx 0))
;  :icon '(141)
  (fq-to-pch f approx))

(defmethod fq->pch ((f list) &optional (approx 0))
;(om::defmethod! fq->pch ((f list) &optional (approx 0))
    (cond ((pitch-with-octave-p f) f)
         ((or (symbolp f) (stringp f))
          (error "CANNOT CALCULATE THE FREQ OF ~a"f))
         (t (mapcar (lambda (x) (fq-to-pch x approx)) f))))

(defmethod fq->pch ((f t) &optional (approx 0))
;(om::defmethod! fq->pch ((f t) &optional (approx 0))
  (declare (ignore approx))
    (if (pitch-with-octave-p f)
      f
      (error "MYSTERIOUS ARGUMENT, SIR ~a" f)))

(defmethod fq-to-pch ((f t) &optional (approx 0))
  (if (pitch-with-octave-p f)
    f
    (midi->pch (fq->midi f ) approx)))


(defmethod fq->midi ((l list))
;(om::defmethod! fq->midi ((l list))
;  :icon '(141)
 (mapcar #'fq-to-midi l))

(defmethod fq->midi (l)
;(om::defmethod! fq->midi (l)
      (fq-to-midi l))

(defmethod fq-to-midi (f)
  (+ 69 (/(log (/ f (get-gbl 'diapason)))(log(expt 2 (/ 1 12))))))


(defmethod fq->midic ((l list))
;(om::defmethod! fq->midic ((l list))
;  :icon '(141)
 (mapcar #'fq->midic l))

(defmethod fq->midic (l)
;(om::defmethod! fq->midi (l)
      (midi->midic (fq-to-midi l)))


(defmethod fq->ratio ((l list))
;(om::defmethod! fq->ratio ((l list))
;  :icon '(141)
  (cond ((null (second l))())
        (t ( cons (/(second l)(first l))(fq->ratio(cdr l))))))

(defmethod fq->ratio (l)
;(om::defmethod! fq->ratio (l)
  (error "I NEED A LIST OF AT LEAST TWO ARGUMENTS, SIR AND YOU GAVE ME ~a" l))



;PITCH
(defmethod pch->fq ((note number) &rest diap)        ;nouvelle version pour pch (cf : LLdg-pitch.lisp) 
;(om::defmethod! pch->fq ((note number) &rest diap)
;  :icon '(141)
  (declare (ignore diap))
  note)

(defmethod pch->fq ((note list) &rest diap)        ;nouvelle version pour pch (cf : LLdg-pitch.lisp) 
;(om::defmethod! pch->fq ((note list) &rest diap)
  (let ((diap (ifn diap (get-gbl 'DIAPASON) (car diap))) )
    (if(or(memberp (cdr note) *DEVIATIONS*) (numberp (cdr note)))
      (freq (make-instance 'symbolic-pitch :pitch (string (car note)) 
                           :diapason diap 
                           :deviation (cdr note)))
      (mapcar (lambda (n) (pch->fq n diap)) note))))

(defmethod pch->fq ((note t) &rest diap)        ;nouvelle version pour pch (cf : LLdg-pitch.lisp) 
;(om::defmethod! pch->fq ((note t) &rest diap)        ;nouvelle version pour pch (cf : LLdg-pitch.lisp) 
  (let ((diap (ifn diap (get-gbl 'DIAPASON) (car diap))) )
    (freq (make-instance 'symbolic-pitch :pitch (string note) 
                         :diapason diap))))

(defmethod pch->midi (note &rest diap)
;(om::defmethod! pch->midi (note &rest diap)
;  :icon '(141)
  (let ((diap (ifn diap (get-gbl 'DIAPASON) (car diap))) )
    (cond ((numberp note) note)
          ((listp note)
           (if(or (memberp (cdr note) *DEVIATIONS* ) (numberp (cdr note)))
             (midi-note (make-instance 'symbolic-pitch :pitch (string (car note)) 
                                       :diapason diap 
                                       :deviation (cdr note)))
             (mapcar (lambda (n) (pch->midi n diap)) note)))
          (t
           (midi-note (make-instance 'symbolic-pitch :pitch  (string note) 
                                     :diapason diap))))))


(defmethod pch->midic (note &rest diap)
;(om::defmethod! pch->midic (note &rest diap)
;  :icon '(141)
  (ifn diap
       (midi->midic (pch->midi note))
    (midi->midic (pch->midi note diap))))


(defmethod pch->itvl (note)
;(om::defmethod! pch->itvl (note)
;  :icon '(141)
  (semitones->itvl(pch->semitones note)))

(defmethod pch->semitones (note)
;(om::defmethod! pch->semitones (note)
;  :icon '(141)
  (midi->semitones(pch->midi note)))


(defmethod pch->pch-class ((note cons)) ; if a cons it is a pitch with deviation
;(om::defmethod! pch->pch-class ((note cons)) ; if a cons it is a pitch with deviation
;  :icon '(141)
  (if (numberp (cdr note))                ; IF it is a cons with deviation
    (if (pitch-without-octave-p note)             ; if there is no octave, add 48 so as
      (+ 48 (midi->pch-class (pch->midi note)))   ;    to start the list at 0 (DO4)
      (midi->pch-class (pch->midi note)))
    (mapcar #'pch->pch-class note)))      ; ELSE it is a list of notes

(defmethod pch->pch-class ((note t)) ; default: single pitch without deviation
;(om::defmethod! pch->pch-class ((note t)) ; default: single pitch without deviation
  (if (pitch-without-octave-p note)             ; if there is no octave, add 48 so as
    (+ 48 (midi->pch-class (pch->midi note)))   ;    to start the list at 0 (DO4)
    (midi->pch-class (pch->midi note))))


;MIDI
(defmethod midi->pch ((p number) &optional (approx 0))
;(om::defmethod! midi->pch ((p number) &optional (approx 0))
; :icon '(141)
 (setf p (from-midic p))
 (let ((*package* (find-package :chroma))
       (octave (floor (+ p .5) 12))
       (fnote (mod p 12)))
    (multiple-value-bind(note deviation)(round fnote)
      (if (eq approx 0)
        (cons (format () "~a~a"(cdr(assoc note *SCPITCH-ALIST*)) (- octave 1)) (* deviation 100))
        (if (numberp approx)
          (cons(format () "~a~a"(cdr(assoc note *SCPITCH-ALIST*)) (- octave 1))
               (float(*(round(/(* 100 deviation) approx))approx)))
          (format () "~a~a"(cdr(assoc note *SCPITCH-ALIST*)) (- octave 1)))
        ))))

(defmethod midi->pch ((p list) &optional (approx 0))
;(om::defmethod! midi->pch ((p list) &optional (approx 0))
  (mapcar #'(lambda (x) (midi->pch x approx)) p))


(defmethod midi->semitones ((l list))
;(om::defmethod! midi->semitones ((l list))
;  :icon '(141)
 (cond((null (second l))())
       (t( cons (- (from-midic (second l)) (from-midic (first  l))) (midi->semitones (cdr l))))))

(defmethod midi->semitones ((l t))
;(om::defmethod! midi->semitones ((l t))
  (error "I NEED A LIST OF AT LEAST TWO ARGUMENTS, SIR AND YOU GAVE ME ~a" l))


(defmethod midi->pch-class ((note number))
;(om::defmethod! midi->pch-class ((note number))
;  :icon '(141)
 (- (from-midic note) 60))

(defmethod midi->pch-class ((l list))
;(om::defmethod! midi->pch-class ((l list))
  (mapcar #'midi->pch-class l))


(defmethod from-midic ((midi number))
;(om::defmethod! from-midic ((note number))
;  :icon '(141)
  (if (< midi 127) midi
      (/ midi 100.0)))
;    (round (/ midi 100.0)))) ; WRONG VERSION, corrected by Marco, Feb 21, 2000

(defmethod from-midic ((midi list))
  (mapcar #'from-midic midi))


(defmethod midi->midic ((l number))
;(om::defmethod! midi-midic ((l list))
;  :icon '(141)
  (round (* (from-midic l) 100)))

(defmethod midi->midic ((l list))
  (mapcar #'midi->midic l))



;midic
(defmethod midic->midi ((l number))
;(om::defmethod! midic-midi ((l list))
;  :icon '(141)
  (/ l 100.0))

(defmethod midic->midi ((l list))
  (mapcar #'midic->midi l))



;RATIO
(defmethod ratio->fq ((int list) (ref number))
;(om::defmethod! ratio->fq ((int list) (ref number))
;    :icon '(141)
 (if (null int)(list ref)  
               (cons ref (ratio->fq (cdr int) (* ref  (car int))))))

(defmethod ratio->fq ((int number) (ref number))
;(om::defmethod! ratio->fq ((int number) (ref number))
  (ratio->fq (list int) ref))

(defmethod ratio->fq (int ref)
;(om::defmethod! ratio->fq (int ref)
  (error "MYSTERIOUS ARGUMENTS, SIR: ~a - ~a" int ref))

(defmethod ratio->itvl (int)
;(om::defmethod! ratio->itvl (int)
;  :icon '(141)
 (semitones->itvl(ratio->semitones int)))


(defmethod ratio->semitones ((l list))
;(om::defmethod! ratio->semitones ((l list))
;  :icon '(141)
 (mapcar #'ratio-to-semitones l))

(defmethod ratio->semitones ((l number))
;(om::defmethod! ratio->semitones ((l number))
     (ratio-to-semitones l))

(defmethod ratio->semitones (l)
;(om::defmethod! ratio->semitones (l)
  (error "MYSTERIOUS ARGUMENT, SIR: ~a" l))

(defmethod ratio-to-semitones (ratio)
;(om::defmethod! ratio-to-semitones (ratio)
  (* 12 (log ratio 2)))


;INTERVAL
(defmethod itvl->fq (int (ref number))
;(om::defmethod! itvl->fq (int (ref number))
;  :icon '(141)
  (let ((deviation 0)(octave 0)(intervalle)(freq)
        (int (if (listp int) int (list int))))
    (if (null int) (list ref)
        (progn (if (memberp (car int) *INTERVALLES*) 
                 (progn (setf intervalle (car int))
                        (setf deviation 0)
                        (setf octave 0))
                 (progn (if (listp (car int))
                          (setf intervalle (caar int))
                          (error "UNREALISTIC INTERVAL, SIR: ~a" (car int)))
                        (unless (memberp intervalle *INTERVALLES*)
                          (error "CAN'T BELIEVE 'THIS IS AN INTERVAL: ~a" intervalle))
                        (if (null (third (car int)))
                          (progn (setf deviation 0)
                                 (setf octave (second (car int))))
                          (progn (setf octave (second (car int)))
                                 (setf deviation (third (car int)))))))
               (when (symbolp intervalle)
                 (setf intervalle (internc intervalle)))
               (setf freq 
                     (* ref (expt 2 (/ (+ (cdr (assoc intervalle *INTERVALLES-ALIST*))
                                          (/ deviation 100) 
                                          (* 12  octave)) 12))))
               (cons ref (itvl->fq (cdr int) freq)
                     )))))

(defmethod itvl->fq (int (ref t))
;(om::defmethod! itvl->fq (int (ref t))
  (declare (ignore int))
          (error "CAN'T UNDERSTAND YOUR REFERENCE FREQUENCY: ~a" ref))

 
(defmethod itvl->midi (int (ref number))
;(om::defmethod! itvl->midi (int (ref number))
;  :icon '(141)
 (let ((deviation 0)(octave 0)(intervalle)(pitch)
        (int (if (listp int) int (list int))))
    (if (null int) (list ref)
        (progn (if (memberp (car int) *INTERVALLES*) 
                 (progn (setf intervalle (car int))
                        (setf deviation 0)
                        (setf octave 0))
                 (progn (if (listp (car int))
                          (setf intervalle (caar int))
                          (error "UNREALISTIC INTERVAL, SIR: ~a" (car int)))
                        (unless (memberp intervalle *INTERVALLES*)
                          (error "CAN'T BELIEVE 'THIS IS AN INTERVAL: ~a" intervalle))
                        (if(null (third (car int)))
                          (progn (setf deviation 0)
                                 (setf octave (second (car int))))
                          (progn (setf octave (second (car int)))
                                 (setf deviation (third (car int)))))))
               (when (symbolp intervalle)
                 (setf intervalle (internc intervalle)))
               (setf pitch  (+ ref 
                               (cdr(assoc intervalle *INTERVALLES-ALIST*))
                               (/ deviation 100) 
                               (* 12  octave )))
               (cons ref (itvl->midi (cdr int) pitch)
                     )))))

(defmethod itvl->midi (int (ref t))
;(om::defmethod! itvl->midi (int (ref t))
  (itvl->midi int (pch->midi ref)))


(defmethod itvl->midic (int (ref number))
;(om::defmethod! itvl->midic (int (ref number))
;  :icon '(141)
 (midi->midic (itvl->midi int ref)))

(defmethod itvl->midic (int (ref t))
;(om::defmethod! itvl->midic (int (ref t))
  (itvl->midic int (pch->midi ref)))


(defmethod itvl->ratio (int)
;(om::defmethod! itvl->ratio (int)
;  :icon '(141)
 (semitones->ratio (itvl->semitones int)))

(defmethod itvl->pch (int note &optional (approx 0))
;(om::defmethod! itvl->pch (int note &optional (approx 0))
;  :icon '(141)
; if approx is needed, go through a frequency conversion (less precise)
;    otherwise use a midi conversion
 (if (and (numberp approx) (= approx 0))
    (midi->pch (itvl->midi int (pch->midi note)))
    (fq->pch (itvl->fq int (pch->fq note)) approx)))


(defmethod itvl->semitones (int)
;(om::defmethod! itvl->semitones (int)
;  :icon '(141)
 (let ((deviation 0)(octave 0)(intervalle)(semitones)
        (int (if (listp int) int (list int))))
    (if (null int) ()
        (progn (if (memberp (car int) *INTERVALLES*) 
                 (setf intervalle (car int) deviation 0 octave 0)
                 (progn (if (listp (car int))
                          (setf intervalle (caar int))
                          (error "UNREALISTIC INTERVAL, SIR: ~a" (car int)))
                        (unless (memberp intervalle *INTERVALLES*)
                          (error "CAN'T BELIEVE 'THIS IS AN INTERVAL: ~a" intervalle))
                        (if (null (third (car int)))
                          (setf deviation 0 octave (second (car int)))
                          (setf octave (second (car int)) deviation (third (car int))))))
               (when (symbolp intervalle)
                 (setf intervalle (internc intervalle)))
               (setf semitones (+
                                (cdr(assoc intervalle *INTERVALLES-ALIST*))
                                (/ deviation 100) 
                                (* 12  octave )))
               (cons semitones (itvl->semitones (cdr int))
                     )))))

;SEMITONES
(defmethod semitones->ratio ((l list))
;(om::defmethod! semitones->ratio ((l list))
;  :icon '(141)
 (mapcar #'semitones-to-ratio l))

(defmethod semitones->ratio ((l number))
;(om::defmethod! semitones->ratio ((l number))
  (semitones-to-ratio l))

(defmethod semitones->ratio (l)
;(om::defmethod! semitones->ratio (l)
  (error "CANNOT TRUST YOUR ARGUMENT: ~a" l))

(defmethod semitones-to-ratio (ratio)
     (expt 2 (/ ratio 12)))

(defmethod semitones->itvl ((l list))
;(om::defmethod! semitones->itvl ((l list))
;  :icon '(141)
 (mapcar #'semitones-to-itvl l))

(defmethod semitones->itvl ((l number))
;(om::defmethod! semitones->itvl ((l number))
  (semitones-to-itvl l))

(defmethod semitones->itvl (l)
;(om::defmethod! semitones->itvl (l)
  (error "MIND YOUR ARGUMENT, PLEASE: ~a" l))

(defmethod semitones-to-itvl (i)
  (multiple-value-bind (int deviation)(round i)
    (multiple-value-bind (octave interval)(floor int 12)
      (if (<(abs deviation) 0.00001)
        (list(car(rassoc interval *INTERVALLES-ALIST*)) octave)
      (list(car(rassoc interval *INTERVALLES-ALIST*))
           octave
           (round(* 100 deviation)))))))

;PITCH CLASSES
(defmethod pch-class->pch ((l list) &optional (approx 0))
;(om::defmethod! pch-class->pch ((l list) &optional (approx 0))
;  :icon '(141)
 (mapcar #'(lambda (x) (pch-class->pch x approx)) l))

(defmethod pch-class->pch ((l number) &optional (approx 0))
;(om::defmethod! pch-class->pch ((l number) &optional (approx 0))
  (midi->pch (pch-class->midi l) approx))

(defmethod pch-class->pch ((l t) &optional (approx 0))
  (declare (ignore approx))
  (error "I HAVE SOME PROBLEMS WITH YOUR ARGUMENT: ~a~%" l))


(defmethod pch-class->midi ((l list))
;(om::defmethod! pch-class->midi ((l list))
;  :icon '(141)
 (mapcar #'pch-class->midi l))

(defmethod pch-class->midi ((l number))
;(om::defmethod! pch-class->midi ((l number))
  (+ l 60))

(defmethod pch-class->midi ((l t))
;(om::defmethod! pch-class->midi ((l t))
  (error "MIND YOUR ARGUMENT, PLEASE: ~a" l))


(defmethod pch-class->fq ((l list))
;(om::defmethod! pch-class->fq ((l list))
;  :icon '(141)
 (mapcar #'pch-class->fq l))

(defmethod pch-class->fq ((l number))
;(om::defmethod! pch-class->fq ((l number))
  (pch->fq (pch-class->pch l)))

(defmethod pch-class->fq ((l t))
;(om::defmethod! pch-class->fq ((l t))
  (error "I HAVE SOME PROBLEMS WITH YOUR ARGUMENT: ~a~%" l))


#|
(fq->pch 881 )
(fq->pch 129.677 )
(fq->pch 881 t)
(fq->pch 881 10)
(fq->pch '(200 201 204 205 440))
(fq->pch '(200 201 204 205 440) t)
(fq->pch '(200 201 204 205 440) 100)
(loop for i from 220 to 440
     do (print (fq->pch i)))
(fq->midi 440)
(fq->midi '(200 201 204 205 440))
(fq->ratio '(100 200 220))

(pch->fq '(la4 (12 34 re3) (211 323 (323)) la4))
(pch->fq '("REd4" "MIb4"))
(pch->fq "LA4" )
(pch->fq 'DO4 )
(pch->fq '(ut2 do2))
(pch->fq '((red4 . 10) mib4) 443.0)
(pch->fq '( red4 (red4 . 10) red4))
(pch->fq '((sol2 . -50) (fad2 . 50)))
(pch->fq 100 )
(pch->fq '(REb2 . q))
(pch->fq 'REb2 )
(pch->fq '((DO . 12) (LAb . -5) 100 RE1 SOL1 (REb2 . -q)))
(pch->itvl '(DO MI))
(pch->itvl '((DO . 12) (LAb . -5) 100 RE1 SOL1 (REb2 . -q)))

(ratio->fq  '(2 2 2) 100)
(ratio->itvl 2)
(ratio->itvl '(2 2 1 3/2))
(ratio->semitones '(2 2 1 3/2))

(itvl->fq '((1 1)(2- 0)) 100)
(itvl->ratio '((1 0) (2- 0 2) (1 1) 5))
(itvl->semitones '((1 0) (2- 0 2) (1 1) 5))
(itvl->semitones '(2- ))
(itvl->semitones '((2- 1)))

(itvl->semitones '((2- 0) (1 0) (5 0)))             ;;>> identique a
(itvl->semitones '(2- 1 5))             ;;>> accord different de 
(itvl->semitones '((2- 1 5)))           ;;>intervalle
(itvl->semitones '2-)                   ; ;donc illegal !
(itvl->semitones '(-2- (3+ 0)(2- 1 5)))

(semitones->itvl 5 )
(semitones->itvl '(5 6))
|#