
(in-package :om)

(defclass! Chant-Evt (SynthesisEvt) ()
   (:icon 600))

(defmethod collect-chant-events ((self t) num) nil)
(defmethod collect-chant-events ((self list) num) 
   (remove-if 'null (loop for object in self collect (collect-chant-events object num))))

(defmethod collect-chant-events ((self container) num) 
   (remove-if 'null (loop for object in (inside self) collect (collect-chant-events object num))))
                          
(defmethod collect-chant-events ((self Chant-Evt) num)
   (when (in-chant-patch-p self num) self))

(defmethod get-obj-dur ((self Chant-Evt)) 0)

(defmethod extra-channels ((self Chant-Evt))
   (let (rep)
     (loop for item in (Lcontrols self) do
           (let ((ctlname (string (car item))))
           (when (and (> (length ctlname) 6)
                      (string-equal "channel" (subseq ctlname 0 7)))
             (push (label2index self ctlname) rep))))
     (reverse rep)))

(defmethod get-channels-num ((self Chant-Evt)) 
   (+ 1 (length (extra-channels self))))

(defmethod parsing-fun ((self Chant-Evt)) 
   (slot-value self 'parsing-fun))


;=================================
;BANK DE FOF
;=================================

(defclass Chant-Evt-info ()
   ((FQ0 :initform nil :accessor FQ0)))
    

(defclass! Ch-Fob-Evt (Chant-Evt Chant-Evt-info) 
   ((Frequency :initform 440.0 :initarg :Frequency :type number :accessor Frequency)
    (Amplitude :initform 0 :initarg :Amplitude :type number :accessor Amplitude)
    (BandWidth :initform 0 :initarg :BandWidth :type number :accessor  BandWidth)
    (Tex :initform 0.002 :initarg :Tex :type number :accessor  Tex)
    (DebAtt :initform 0.05 :initarg :DebAtt :type number :accessor  DebAtt)
    (Atten :initform 0.004 :initarg :Atten :type number :accessor  Atten)
    (FofPhase :initform 0 :initarg :FofPhase :type number :accessor FofPhase)
    (Channel1 :initform 1 :initarg :Channel1 :type number :accessor Channel1)) 
   (:icon 607))

(defmethod fob-p ((self Ch-Fob-Evt)) t)
(defmethod fob-p ((self t)) nil)



(defmethod get-obj-dur ((self Ch-Fob-Evt)) 
   (+ (apply 'max (om+ (get-array-row self 4) (get-array-row self 5))) (action-time self)))



(defmethod in-chant-patch-p ((self Ch-Fob-Evt) num) 
   (or (= num 0) (= num 1) (= num 3) (= num 6) (= num 7) (= num 8) (= num 9) (= num 10)))
    
(defmethod fixed-slots-list ((self Ch-Fob-Evt)) '(numcols action-time fq0))

(defmethod get-slot-in-out-names ((self Ch-Fob-Evt))
   (values '("self" "numcols" "start time" "fq0") 
           '(nil 1  0 nil)
           '("object or object list" "How many vertical components for init ?" 
             "Start time in seconds" "Impulsion")
           '(nil nil nil nil)))


(defmethod matrix-gen-code ((self arrayBox) (val Ch-Fob-Evt))
   (let* ((params (decode self))
          (theval (if (connected? (first (inputs self)))
                    `(objFromObjs ,(gen-code (first (inputs self)) (second (inputs self))) ,val)
                    `(funcall 'cons-new-array ,val (list nil ,(second params) ,(third params) ,(fourth params)) 
                              (list ,.(nthcdr 4 params)) nil ',(class-name (reference self))))))
     theval))


(defmethod save-sdif-chant ((self Ch-Fob-Evt) ptr)
   (without-interrupts  
    (let* ((datatype 4)
           (channels (append '(7) (extra-channels self)))
           (framesize (+ 68 
                         datatype 
                         (* 7 datatype  (numcols self))
                         (* (length channels) datatype  (numcols self))))
           (fq0vals (om-make-pointer datatype))
           (values (om-make-pointer (* 7 datatype  (numcols self))))
           (chanval (om-make-pointer (* datatype  (numcols self) (length channels)))))
      (sdif::SdifFSetCurrFrameHeader ptr (sdif-string-to-signature "1FOB") framesize 3  0 (coerce (action-time self) 'double-float))
      (sdif::SdifFWriteFrameHeader ptr)
      (om-write-ptr fq0vals 0 :single-float (coerce (fq0 self) 'single-float))
      (sdif::SdifFWriteMatrix ptr (sdif-string-to-signature "1FQ0")  datatype 1 1 fq0vals)
      (loop for i from 0 to (- (numcols self) 1) do
            (let* ((sdifrow (aref (work-vector self) i)))  ;(sdifrow (first-n (get-array-col self i) 7))
              (loop for j from 0 to 6
                    for val in sdifrow do
                    (om-write-ptr values (+ (* j datatype) (* i 7 datatype)) :single-float (coerce val 'single-float) ))
              
              (loop for k in channels
                    for delta = 0 then (+ delta 1) do
                    (let ((chval (nth k sdifrow)))
                      (om-write-ptr chanval (+ (* delta datatype) (* i (length channels) datatype)) :single-float (coerce chval 'single-float) )))
              ))
      (sdif::SdifFWriteMatrix ptr (sdif-string-to-signature "1FOF") datatype (numcols self)  7 values)
      (sdif::SdifFWriteMatrix ptr (sdif-string-to-signature "1CHA") datatype (numcols self)  (length channels) chanval))))


;copy and save

(defmethod omNG-copy ((self Ch-Fob-Evt))
   `(let ((copy ,(call-next-method)))
        (setf (FQ0 copy) ,(FQ0 self))
        copy))

(defmethod omNG-save ((self Ch-Fob-Evt) &optional (values? nil)) 
    `(let ((copy ,(call-next-method)))
        (setf (FQ0 copy) ,(FQ0 self))
       copy))


;  0 Chant:Patch0/1/FOB/1/21/0.000000/1.800000;
; (format nil "Chant:Patch~D/1/FOB/~D/~D/~D" patchnum (numcols self) starttime endtime) 

;=================================
;BANK DE Filtres
;=================================

(defclass! Ch-Filt-Evt (Chant-Evt) 
   ((Frequency :initform 440.0 :initarg :Frequency :type number :accessor Frequency)
    (Amplitude :initform 0.8 :initarg :Amplitude :type number :accessor Amplitude)
    (BandWidth :initform 0.5 :initarg :BandWidth :type number :accessor  BandWidth)
    (saliance :initform 1 :initarg :saliance :type number :accessor  saliance)
    (correction :initform 0 :initarg :correction :type number :accessor  correction)
    (Channel1 :initform 1 :initarg :Channel1 :type number :accessor Channel1)) 
   (:icon 608))

(defmethod reb-p ((self Ch-Filt-Evt)) t)
(defmethod reb-p ((self t)) nil)

(defvar *filt-limit* 1000)

(defmethod get-obj-dur ((self Ch-Filt-Evt)) 
   (+ (/ (log *filt-limit*) (* pi (apply 'min (get-array-row self 2) ))) (action-time self)))

(defmethod get-num-of-channel ((self Ch-Filt-Evt)) 
   2)


(defmethod fixed-slots-list ((self Ch-Filt-Evt)) '(numcols action-time))

(defmethod get-slot-in-out-names ((self Ch-Filt-Evt))
   (values '("self" "numcols" "start time" ) 
           '(nil 1  0 nil)
           '("object or object list" "How many vertical components for init ?" 
             "Start time in seconds" )
           '(nil nil nil )))

(defmethod in-chant-patch-p ((self Ch-Filt-Evt) num) 
   (or  (= num 1) (= num 2) (= num 3) (= num 4) (= num 5) (= num 6) (= num 7) (= num 8) (= num 9) (= num 10)))


(defmethod save-sdif-chant ((self Ch-Filt-Evt) ptr)
   (let* ((datatype 4)
          
          (framesize (+ 48 datatype (* 7 datatype  (numcols self)) (* datatype  (numcols self)) ))
          
          (values (om-make-pointer (* 5 datatype  (numcols self))))
          
          (chanval (om-make-pointer (* datatype  (numcols self)))))
     (sdif::SdifFSetCurrFrameHeader ptr (sdif-string-to-signature "1REB") framesize 2  1 (coerce (action-time self) 'double-float))
     (sdif::SdifFWriteFrameHeader ptr)
     
     (loop for i from 0 to (- (numcols self) 1) do
           (let* ((sdifrow (aref (work-vector self) i)))
             (loop for j from 0 to 4
                   for val in sdifrow do
                   (om-write-ptr values (+ (* j datatype) (* i 5 datatype)) :single-float (coerce val 'single-float) ))
             (om-write-ptr chanval (* i datatype) :single-float (coerce (car (last sdifrow)) 'single-float) ) ))
     (sdif::SdifFWriteMatrix ptr (sdif-string-to-signature "1RES") datatype (numcols self)  5 values)
     (sdif::SdifFWriteMatrix ptr (sdif-string-to-signature "1CHA") datatype (numcols self)  1 chanval)))

;=================================
;NOISE
;=================================

(defclass! Ch-Noise-evt (Chant-Evt)
   ((Distribution :initform 1 :initarg :Distribution :type number :accessor Distribution)
    (Amplitude :initform 1 :initarg :Amplitude :type number :accessor Amplitude))
   (:icon 604))

(defmethod noi-p ((self Ch-Noise-evt)) t)
(defmethod noi-p ((self t)) nil)

(defmethod fixed-slots-list ((self Ch-Noise-evt)) '(numcols action-time))

(defmethod get-slot-in-out-names ((self Ch-Noise-evt))
   (values '("self" "numcols" "start time" ) 
           '(nil 1  0 nil)
           '("object or object list" "How many vertical components for init ?" 
             "Start time in seconds" )
           '(nil nil nil )))


(defmethod in-chant-patch-p ((self Ch-Noise-evt) num) 
   (or  (= num 1) (= num 2) (= num 3) (= num 4) (= num 8) (= num 9)))

(defmethod save-sdif-chant ((self Ch-Noise-evt) ptr)
   (let* ((datatype 4)
          (framesize (+ 32 datatype (* 2 datatype  (numcols self)) (* datatype  (numcols self)) ))
          (values (om-make-pointer (* 2 datatype  (numcols self)))))
     (sdif::SdifFSetCurrFrameHeader ptr (string2initarg (string :1NOI)) framesize 1  2 (coerce (action-time self) 'double-float))
     (sdif::SdifFWriteFrameHeader ptr)
     
     
     (loop for i from 0 to (- (numcols self) 1) do
           (let* ((sdifrow (aref (work-vector self) i)))
             (loop for j from 0 to 1
                   for val in sdifrow do
                   (om-write-ptr values (+ (* j datatype) (* i 2 datatype)) :single-float (coerce val 'single-float) ))))
     (sdif::SdifFWriteMatrix ptr (string2initarg (string :1DIS)) datatype (numcols self)  2 values)
     
     
     ))



;=================================
;Synthesize
;=================================

(defun split-list (list)
   (let (lfob lreb lnoi)
     (loop for item in list do
           (cond
            ((reb-p item) (push item lreb))
            ((fob-p item) (push item lfob))
            ((noi-p item) (push item lnoi))))
     (list (reverse lfob) (reverse lreb) (reverse lnoi))))


(defun symbfromnum (num)
  (case num
    (0 "FOB") (1 "REB") (2 "NOI")))

(defun write-one-IDS (file num patch events time)
  (write-1IDS-table file num "Chant" (format nil "Patch~D/1/~D/~D/~D/~F/~F"
                                             patch
                                             (symbfromnum num)
                                             (length events)
                                             (numcols (car (last events)))
                                             (action-time (car events))
                                             time )))


(defun write-this-lines (file patch list splitlist time)
  (loop for item in list do
        (write-one-IDS file item patch (nth item splitlist) time)))

(defun writeIDStable (file patch events time sound)
   (let ((split-list (split-list events)))
     (cond
      ((= patch 0) (write-this-lines file patch '(0) split-list time))
      ((or (= patch 2) (= patch 4)   ) (write-this-lines file patch '(1 2) split-list time))
      ((or (= patch 6) (= patch 7) (= patch 10) ) (write-this-lines file patch '(0 1) split-list time))
      ((or (= patch 1) (= patch 3) (= patch 8) (= patch 9)) (write-this-lines file patch '(0 1 2) split-list time))
      ((= patch 5) (write-this-lines file patch '(1) split-list time)))
     (when (or (= patch 2) (= patch 3) (= patch 5) (= patch 7)
               (= patch 8) (= patch 10))
       (write-1IDS-table file 3 "Chant" (format nil "Patch~D/1/SND/1/0/0.0/~F/~S"
                                                patch time sound)))))

(defmethod get-str-for-sound ((self t)) nil)

(defmethod get-str-for-sound ((self string)) self)

(defmethod get-str-for-sound ((self pathname)) self)



(defmethod ch-Synthesize ((events t)  &key 
                          (patch 0) 
                          (name "my_synthesize") (resolution 16)
                          (rescale  0.0) (sr 44100) (run t) sound duration)
  
  (setf sound (get-str-for-sound sound))
  
  (setf events (collect-chant-events (list! events) patch))
  
  (setf *ae-error-p* nil)
  
  (let* ((continue? (if *automatic-rename* t (replace-if-exist name '(".sdif"))))
         path-sdif path-aiff path-temp aiff-temp)
    (when continue?
      (if *automatic-rename*
        (setf path-sdif (om-namestring (unique-pathname *om-tmpfiles-folder* "name" ".sdif"))
              path-temp (om-namestring (unique-pathname *om-outfiles-folder* name ".temp"))
              aiff-temp (om-namestring (unique-pathname *om-outfiles-folder* name ".temp2")))
        (setf path-sdif (om-namestring (make-pathname :directory (mac-namestring *om-tmpfiles-folder*) :name (string+ name ".sdif")))
              path-temp (om-namestring (make-pathname :directory (mac-namestring *om-outfiles-folder*) :name (string+ name ".temp")))
              aiff-temp (om-namestring (make-pathname :directory (mac-namestring *om-outfiles-folder*) :name (string+ name ".temp2")))))
      (setf path-aiff (om-namestring (make-pathname :directory (mac-namestring *om-outfiles-folder*) :name (string+ name ".aiff"))))
      (unless events (beep-msg "Not synthesis events were found."))       
      (when events
        (setf events (sort events '< :key 'action-time))
        (let ((thefile (sdif-open-file (om-path2cmdpath (mac-namestring path-sdif)) 1))
              (ch-num (loop for item in events  maximize (get-channels-num item)))
              (synthestime (or duration (loop for item in events  maximize (get-obj-dur item)))))
          (sdif::SdifFWriteGeneralHeader thefile)
          
          (write-1nvt-table thefile
                            (list "BufferSize" "NumberOfChannels" "EndTime" "SdifTypesVersion" "ChantLibraryVersion" "Author" "SamplingRate")
                            (list "512" (format nil "~D" ch-num) (format nil "~D" synthestime) "1.01.ppc" "1.01.ppc" "OM" (format nil "~F" sr)))
          (writeIDStable thefile patch events synthestime sound)
          (sdif::SdifFWriteAllASCIIChunks thefile)
          (loop for item in events 
                for i = 0 then (+ i 1) do
                (cons-work-vector item)
                (save-sdif-chant item thefile)
                (free-work-vector item)
                (print i))
          (sdif-close-file thefile)
          
          (when run
            (if (probe-file (om-external-app "chant"))
              (progn
                (print path-temp)
                (print path-aiff)
                (print path-sdif)
                (print (format nil "Running Chant... writing ~D" (om-path2cmdpath path-sdif)))
                (om-cmd-line (print (format nil "~s -v -i~S | ~s -of -c~D -R~D ~S" 
                                            (om-path2cmdpath (om-external-app "chant"))
                                            (om-path2cmdpath path-sdif) 
                                            (om-path2cmdpath (om-external-app "tosf"))
                                            ch-num sr
                                            (om-path2cmdpath path-temp))))
                
                (if (probe-file (om-external-app "supervp"))
                  (progn 
                    (print (format nil "Rescaling ~D to ~D dB and converting to ~D bits" (om-path2cmdpath path-temp) rescale resolution))
                    (print "==========")
                    
                    (om-cmd-line (print (format nil "~s -S~S -norm ~D -Osa~D ~S" 
                                                (om-path2cmdpath (om-external-app "supervp"))
                                                (om-path2cmdpath path-temp) 
                                                (* -1 rescale) resolution
                                                (om-path2cmdpath path-aiff))))
                    )
                  (progn
                    (print "SVP no available for rescale...")
                    (rename-file path-temp path-aiff))
                  )
            
                (print "End synthesis")
                
                (print (om-namestring path-aiff)))
              (print "CHANT is not available"))
            )
          
          (when *delete-inter-file*
              (delete-intermediates (list path-sdif path-temp aiff-temp)))

          (om-namestring path-aiff)
          
          )))))



(defmethod! Synthesize ((synthetiser (eql 'chant)) elements &key
                          (name "my_synt")
                          (sr 44100)
                          (rescale 0.0)
                          (run t) (evt-test  nil) (resolution 16)
                          ;For Csound
                          (kr  4410) tables (nchnls  1)
                          ;For Chant
                          (patch 0) sound  duration)
   ;(verify-def-folders)
   (ch-Synthesize elements :name name :sr sr :patch patch :sound sound :duration duration :rescale rescale :run run :resolution resolution))


;=============================
;HELP



#|
(defmethod! chant-patches ()
  :icon 600
  (om-select-window *ch-patches-window*))

|#

(defmethod! sdif-write-frame (stream (self chant-evt))
  :icon 908
  (cons-work-vector self)
  (save-sdif-chant self stream)
  (free-work-vector self))

