;=============================================================================
;		ADD-2.ORC
; SIMPLE ADDITIVE SYNTHESIS (FROM ACCCI, 02_41_1.ORC) / MONO
; AMPLITUDE ENVELOPE WITH LINEN
;=============================================================================

; Timbre:    Brass
; Synthesis: Additive, same building blocks units
;            Basic instrument with added random frequency variation
; Source:    #200, Brass-like Sounds through Independent Control of
;            Harmonics, Risset (1969)
; Coded:     jpg 8/93, modified ms 9/02

; NB 1: this instrument works better with short notes.
; NB 2: the apparently arbitrary amplitude range (0-1000, rather than 0-1)
;         avoids Lisp printing small values with exponential notation
;-----------------------------------------------------------------------------

;	p1	= instrument number
;	p2	= action time [sec]
;	p3	= duration [sec]
;	p4	= maximum amplitude [linear, 0.0 -> 1000.0]
;	p5	= fundamental frequency [Hz]
;	p6	= attack time of the amplitude envelope [sec]
;	p7	= decay time of the amplitude envelope [sec]
;	p8	= spectral scaler for the audio frequency
;	p9	= % of jitter (low-frequency random frequency variation) [0-1]

;-----------------------------------------------------------------------------
; COMPULSORY GEN FUNCTIONS :
;	f1	audio wave
;_____________________________________________________________________________

; CLASS: ADD2

;  GLOBAL KEYWORDS (default values within parentheses):
;	NUMROWS		: amount of rows (components) in the event (1)
;	ACTION-TIME	: start time of the whole event [sec] (0.0)
;	USER-FUN	: user-defined parsing function (nil)

;  LOCAL KEYWORDS:
;   E-DELS	:	entry delays [sec] (0.0)
;	DURS	:	duration [sec] (1.0)
;	AMP	:	amplitude [sec] (500.0)
;	FREQ	:	frequency [Hz] (440.0)
;	ATK	:	attack time [sec] (0.003)
;	DEC	:	decay time [sec] (0.01)
;	SCAL	:	spectral scaler for the audio frequency (1.0)
;	JTF	:	frequency jitter [0-1, % of fundamental frequency] (0.06)

;*****************************************************************************

sr	= 44100
kr	= 44100
ksmps	= 1
nchnls	= 1


gimaxamp init 32767 			; 16 bits


instr 1 ; -------------------------------------------------------------

idur    =  p3
iamp	= p4 * gimaxamp * 0.001	; amplitude range 0-1000
ifund   =  p5
ifq		=  ifund * p8
irise   =  p6					; steep rise  (.5 to 3 ms)
idec    =  p7					; steep decay (4 to 11 ms)
iran    =  p9
iampran =  ifund * iran

        afqr   randi   iampran, 10 		; random frequency = 10 Hz
        aenv   linen   iamp, irise, idur, idec
        a1     oscili  aenv, ifq + afqr, 1
               out     a1
endin
