;******************************************************************
;		     CLASS BUZZ-1
;******************************************************************

(in-package om)

; LISP-DEFINED CLASSES SHOULD RESIDE IN THE LIBRARY'S PACKAGE AND 
;   NOT IN THE USER PACKAGE, WHICH CONTAINS ALL THE CLASSES
;   GRAPHICALLY DEFINED


(defclass! buzz-1
  (cs-evt) 			; INHERIT FROM CS-EVT
  (

; GLOBAL SLOTS (LIGHT BLUE, ON THE LEFT OF THE CLASS):
;    THE METHOD BELOW TRANSFORMS THEM INTO GLOBAL SLOTS ("SHOW" UNCHEKED)
;    ATTENTION: A GLOBAL SLOT SHOULD NOT HAVE AN INITARG

   ( source-code :initform
                 (load-buffer-textfile
                  (get-orc-source (get-orc "buzz-1"))
                  'textfile "append")
                 :allocation :class
                 :type textfile
                 :accessor source-code)
   ( numchan :initform (or (get-orc-channels (get-orc "buzz-1")) 1)
             :allocation :class  :accessor numchan)
   ( globals-list :initform 
                  (get-orc-globals
                   (get-orc "buzz-1"))
                  :allocation :class
                  :type list
                  :accessor globals-list)
   (InstID :initform 1  :allocation :class  :accessor InstID)

; LOCAL SLOTS (RED, CORRESPONDING TO THE P-FIELDS)
;    ATTENTION: A GLOBAL SLOT SHOULD HAVE AN INITARG
  ( amp		:type number
		:initarg :amp 
  		:initform 500.0
		:accessor amp)
   ( freq	:type number
        	:initarg :freq 
        	:initform 220.0
        	:accessor freq)
   ( aenv	:type cs-table
		:initarg :aenv 
                       ; x-points y-points decimals number size
  		:initform (make-cs-table 'Gen07  '(0 256 513)
                                         '(0 1 0) 1 "?" 513)
		:accessor aenv)
   ( bzlh	:type number
        	:initarg :bzlh 
        	:initform 1
        	:accessor bzlh)
   ( bzhh	:type number
        	:initarg :bzhh 
        	:initform 1.0
        	:accessor bzhh)
   ( bzenv	:type cs-table
		:initarg :bzenv 
                       ; x-points y-points decimals number size
  		:initform (make-cs-table 'Gen07  '(0 256 513)
                                         '(0 1 0) 1 "?" 513)
		:accessor bzenv)
   )
  (:documentation
   "
;=============================================================================
;				BUZZ-1.ORC
; DYNAMIC SPECTRUM OSCILLATOR (FROM ACCSI, 43_21_1.ORC) / MONO
; AMPLITUDE ENVELOPE WITH OSCILI
;=============================================================================

; Timbre:    Various controlled noise spectra
; Synthesis: (g)buzz
;            OSCILI envelopeS
; Coded:     jpg 8/92, modified ms 9/02

; NB: the apparently arbitrary amplitude range (0-1000, rather than 0-1)
;         avoids Lisp printing small values with exponential notation
;-----------------------------------------------------------------------------

;	p1	= instrument number
;	p2	= action time [sec]
;	p3	= duration [sec]
;	p4	= maximum amplitude [linear, 0.0 -> 1000.0]
;	p5	= frequency [Hz]
;	p6	= amplitude envelope [GEN number]
;	p7	= lowest harmonic present in the buzz
;	p8	= % of maximum possible harmonic present [0-1]
;	p9	= envelope for the multiplier in the series of amplitude coefficients
;			of the buzz

;-----------------------------------------------------------------------------
; COMPULSORY GEN FUNCTIONS
;	f5	large cosine
;_____________________________________________________________________________

; CLASS: BUZZ-1

;  GLOBAL KEYWORDS (default values within parentheses):
;	NUMROWS		: amount of rows (components) in the event (1)
;	ACTION-TIME	: start time of the whole event [sec] (0.0)
;	USER-FUN	: user-defined parsing function (nil)

;  LOCAL KEYWORDS:
;   E-DELS	:	entry delays [sec] (0.0)
;	DURS	:	duration [sec] (1.0)
;	AMP	:	amplitude [sec] (500.0)
;	FREQ	:	frequency [Hz] (220.0)
;	AENV	:	function number for the amplitude envelope [GEN] (triangle)
;	BZLH	:	lowest harmonic present in the buzz [integer] (1)
;	BZHH	:	% of maximum possible harmonic present [0-1] (1)
;	BZENV	:	function number for the buzz envelope [GEN] (triangle)

;*****************************************************************************

"
   )
  (:icon 618)
  )

;------------------------------------------------------------------
;	METHODS FOR THIS CLASS
;------------------------------------------------------------------
; INCLUDE THE COMPULSORY FUNCTIONS AT THE BEGINNING OF A SCORE FILE
;     THIS METHOD MUST BE DEFINED FOR THIS CLASS

(defmethod! orc-header ((self buzz-1))
  (list

   "; GEN functions **********************************************************"
   "; single cosine(!) wave for the buzz"
   "f5 0 65536 11 1 1"

   ))

;******************************************************************
