;******************************************************************
;		     CLASS FOF-2
;******************************************************************

(in-package om)

; LISP-DEFINED CLASSES SHOULD RESIDE IN THE LIBRARY'S PACKAGE AND 
;   NOT IN THE USER PACKAGE, WHICH CONTAINS ALL THE CLASSES
;   GRAPHICALLY DEFINED


(defclass! fof-2
  (cs-evt) 			; INHERIT FROM CS-EVT
  (

; GLOBAL SLOTS (LIGHT BLUE, ON THE LEFT OF THE CLASS):
;    THE METHOD BELOW TRANSFORMS THEM INTO GLOBAL SLOTS ("SHOW" UNCHEKED)
;    ATTENTION: A GLOBAL SLOT SHOULD NOT HAVE AN INITARG

   ( source-code :initform
                 (load-buffer-textfile
                  (get-orc-source (get-orc "fof-2"))
                  'textfile "append")
                 :allocation :class
                 :type textfile
                 :accessor source-code)
   ( numchan :initform (or (get-orc-channels (get-orc "fof-2")) 1)
             :allocation :class  :accessor numchan)
   ( globals-list :initform 
                  (get-orc-globals
                   (get-orc "fof-2"))
                  :allocation :class
                  :type list
                  :accessor globals-list)
   (InstID :initform 1  :allocation :class  :accessor InstID)
; LOCAL SLOTS (RED, CORRESPONDING TO THE P-FIELDS)
;    ATTENTION: A GLOBAL SLOT SHOULD HAVE AN INITARG
  ( amp		:type number
		:initarg :amp 
  		:initform 500.0
		:accessor amp)
   ( f0		:type number
        	:initarg :f0 
        	:initform 100.0
        	:accessor f0)
   ( freq	:type number
        	:initarg :freq 
        	:initform 609
        	:accessor freq)
   ( bw		:type number
        	:initarg :bw 
        	:initform 77
        	:accessor bw)
   ( aenv	:type cs-table
		:initarg :aenv 
                       ; x-points y-points decimals number size
  		:initform (make-cs-table 'Gen07  '(0 1 512 513)
                                         '(0 1 1 0) 1 "?" 513)
                :accessor aenv)
   ( tex	:type number
        	:initarg :tex 
        	:initform 0.003
        	:accessor tex)
   ( debatt	:type number
        	:initarg :debatt 
        	:initform 0.02
        	:accessor debatt)
   ( atten	:type number
        	:initarg :atten 
        	:initform 0.007
        	:accessor atten)

   )

  (:documentation
   "
;=============================================================================
;		FOF-2.ORC
; FORMANTIC WAVEFORM (FROM ACCCI, 45_01_2.ORC) / MONO
; AMPLITUDE ENVELOPE WITH OSCILI
; CONTROL OF THE DETAILS OF THE FOF FROM THE SCORE
;=============================================================================

; Timbre:    Granular synthesis with fof module, voice-like tones
; Synthesis: FOF (Forme d'Onde Formatique)
;            OSCILI envelopes
; Coded:     jpg 8/92, modified ms 9/04

; NB1: this implementation works better with audio and sub-audio f0's
;         and allows for an independent control of tex, debatt and atten
; NB2: the apparently arbitrary amplitude range (0-1000, rather than 0-1)
;         avoids Lisp printing small values with exponential notation
;-----------------------------------------------------------------------------

;	p1	= instrument number
;	p2	= action time [sec]
;	p3	= duration [sec]
;	p4	= maximum amplitude [linear, 0.0 -> 1000.0]
;	p5	= fundamental frequency [Hz]
;	p6	= formant frequency [Hz]
;	p7	= bandwidth [Hz]
;	p8	= amplitude envelope [GEN number]
;	p9	= tex [sec]
;	p10	= krise [sec]
;	p11	= kdur [sec]

;-----------------------------------------------------------------------------
; COMPULSORY GEN FUNCTIONS
;	f1	large sine tone
;	f19	sigmoid rise/decay shape
;_____________________________________________________________________________

; CLASS: FOF-2

;  GLOBAL KEYWORDS (default values within parentheses):
;	NUMROWS		: amount of rows (components) in the event (1)
;	ACTION-TIME	: start time of the whole event [sec] (0.0)
;	USER-FUN	: user-defined parsing function (nil)

;  LOCAL KEYWORDS:
;	E-DEL		: entry delays [sec] (0.0)
;	DURS		: duration [sec] (1.0)
;	AMP		: amplitude [absolute 0-1000] (500.0)
;	F0		: fundamental frequency [Hz] (100.0)
;	FREQ		: center frequency [Hz] (609.0)
;	BW		: bandwidth [Hz] (77.0)
;	AENV		: fun number for the amplitude envelope [GEN] (fast up/ slow down)
;	TEX		: excitation (rise) time of the FOF [sec]
;	DEBATT		: total duration of the FOF [sec]
;	ATTEN		: decay of the FOF [sec]
;
;*****************************************************************************
"
   )
  (:icon 618)
  )

;------------------------------------------------------------------
;	METHODS FOR THIS CLASS
;------------------------------------------------------------------
; INCLUDE THE COMPULSORY FUNCTIONS AT THE BEGINNING OF A SCORE FILE
;     THIS METHOD MUST BE DEFINED FOR THIS CLASS

(defmethod! orc-header ((self fof-2))
  (list

   "; large sine tone"
   "f1  0 65536  10  1"
   "; sigmoid rise/decay"
   "f19 0  8192  19 .5 .5 270 .5"

   ))

;******************************************************************
