;******************************************************************
;		     CLASS FOG-1
;******************************************************************

(in-package om)

; LISP-DEFINED CLASSES SHOULD RESIDE IN THE LIBRARY'S PACKAGE AND 
;   NOT IN THE USER PACKAGE, WHICH CONTAINS ALL THE CLASSES
;   GRAPHICALLY DEFINED


(defclass! fog-1
  (cs-evt) 			; INHERIT FROM CS-EVT
  (

; GLOBAL SLOTS (LIGHT BLUE, ON THE LEFT OF THE CLASS):
;    THE METHOD BELOW TRANSFORMS THEM INTO GLOBAL SLOTS ("SHOW" UNCHEKED)
;    ATTENTION: A GLOBAL SLOT SHOULD NOT HAVE AN INITARG

   ( source-code :initform
                 (load-buffer-textfile
                  (get-orc-source (get-orc "fog-1"))
                  'textfile "append")
                 :allocation :class
                 :type textfile
                 :accessor source-code)
   ( numchan :initform (or (get-orc-channels (get-orc "fog-1")) 1)
             :allocation :class  :accessor numchan)
   ( globals-list :initform 
                  (get-orc-globals
                   (get-orc "fog-1"))
                  :allocation :class
                  :type list
                  :accessor globals-list)
   (InstID :initform 1  :allocation :class  :accessor InstID)

; LOCAL SLOTS (RED, CORRESPONDING TO THE P-FIELDS)
;    ATTENTION: A GLOBAL SLOT SHOULD HAVE AN INITARG
   ( amp	:type number
		:initarg :amp 
  		:initform 500.0
		:accessor amp)
   ( xpf	:type number
        	:initarg :xpf 
        	:initform 1.0
        	:accessor xpf)
   ( spd	:type number
        	:initarg :spd 
        	:initform 1.0
        	:accessor spd)
   ( dens	:type number
        	:initarg :dens 
        	:initform 100.0
        	:accessor dens)
   ( afun	:type number
        	:initarg :afun 
        	:initform 101.0
        	:accessor afun)
   ( aenv	:type cs-table
		:initarg :aenv 
                       ; x-points y-points decimals number size
  		:initform (make-cs-table 'Gen07  '(0 20 110 512 513)
                                         '(0 1 1 0 0) 1 "?" 513)
                :accessor aenv)
   ( fris	:type number
        	:initarg :fris 
        	:initform 0.01
        	:accessor fris)
   ( fdur	:type number
        	:initarg :fdur 
        	:initform 0.1
        	:accessor fdur)
   ( fdec	:type number
        	:initarg :fdec 
        	:initform 0.05
        	:accessor fdec)
   )


  (:documentation
   "
;=============================================================================
;			FOG-1.ORC
; GRANULAR SYNTHESIS (NEW INSTRUMENT) / MONO
; AMPLITUDE ENVELOPE WITH OSCILI
;=============================================================================

; Timbre:    Granular synthesis with fog module, voice-like tones
; Synthesis: FOG (Forme d'Onde Granulaire)
;            OSCILI envelopes
; Coded:     ms 9/02

; NB1: this instrument works AT BEST with samples whose length is a power of 2
;      those which are shorter than the immediately superior power of 2 accepted
;		 by the GEN 1, will produce silence when looking up the unsued portion
;		  of the table

; NB2: the apparently arbitrary amplitude range (0-1000, rather than 0-1)
;         avoids Lisp printing small values with exponential notation

;-----------------------------------------------------------------------------

;	p1	= instrument number
;	p2	= action time [sec]
;	p3	= duration [sec]
;	p4	= maximum amplitude [linear, 0.0 -> 1000.0]
;	p5	= transposition factor [1 = no transposition]
;	p6	= speed [1 = original]
;	p7	= density of the grains [Hz]
;	p8	= audio file [GEN1 number]
;	p9	= amplitude envelope [GEN number]
;	p10	= rise time of the excitation envelope [sec]
;	p11	= duration of the excitation envelope [sec]
;	p12	= decay time of the excitation envelope [sec]

;-----------------------------------------------------------------------------
; COMPULSORY GEN FUNCTIONS
;	f1	audio file
;	f19	sigmoid rise/decay shape
;_____________________________________________________________________________

; CLASS: FOG-1

;  GLOBAL KEYWORDS (default values within parentheses):
;	NUMROWS		: amount of rows (components) in the event (1)
;	ACTION-TIME	: start time of the whole event [sec] (0.0)
;	USER-FUN	: user-defined parsing function (nil)

;  LOCAL KEYWORDS:
;   E-DELS	:	entry delays [sec] (0.0)
;	DURS	:	duration [sec] (1.0)
;	AMP	:	amplitude [sec] (500.0)
;	XPF	:	transposition factor [1 = same pitch as original) (1.0)
;	SPD	:	speed (rate at which successive grains progress through the
;                  	   stored audio table [1 = original speed] (1.0)
;	DENS	:	density of the grains [Hz] (100.0)
;	AFUN	:	audio function number [GEN] (veena.aiff)
;	AENV	:	function number for the amplitude envelope [GEN] (fast < / slow >)
;	FRIS	:	rise time of the excitation envelope [sec] (0.01)
;	FDUR	:	duration of the excitation envelope [sec] (0.1)
;	FDEC	:	decay time of the excitation envelope [sec] (0.05)

;*****************************************************************************


"
   )
  (:icon 618)
  )

;------------------------------------------------------------------
;	METHODS FOR THIS CLASS
;------------------------------------------------------------------
; INCLUDE THE COMPULSORY FUNCTIONS AT THE BEGINNING OF A SCORE FILE
;     THIS METHOD MUST BE DEFINED FOR THIS CLASS

(defmethod! orc-header ((self fog-1))
  (list

; to write a full path in the score file
   "; audio file"
   (format nil "f101  0  524288  1 ~S  0    0  0" (om-path2cmdpath "cl:chroma;example-files;veena.aif"))
   "; sigmoid rise/decay"
   "f19 0  8192  19 .5 .5 270 .5"

   ))

;******************************************************************



