;=============================================================================
;			FOG-1.ORC
;	GRANULAR SYNTHESIS (NEW INSTRUMENT) / MONO
;	AMPLITUDE ENVELOPE WITH OSCILI
;=============================================================================

; Timbre:    Granular synthesis with fog module, voice-like tones
; Synthesis: FOG (Forme d'Onde Granulaire)
;            OSCILI envelopes
; Coded:     ms 9/02

; NB1: this instrument works AT BEST with samples whose length is a power of 2
;      those which are shorter than the immediately superior power of 2 accepted
;		 by the GEN 1, will produce silence when looking up the unsued portion
;		  of the table

; NB2: the apparently arbitrary amplitude range (0-1000, rather than 0-1)
;         avoids Lisp printing small values with exponential notation

;-----------------------------------------------------------------------------

;	p1	= instrument number
;	p2	= action time [sec]
;	p3	= duration [sec]
;	p4	= maximum amplitude [linear, 0.0 -> 1000.0]
;	p5	= transposition factor [1 = no transposition]
;	p6	= speed [1 = original]
;	p7	= density of the grains [Hz]
;	p8	= audio file [GEN1 number]
;	p9	= amplitude envelope [GEN number]
;	p10	= rise time of the excitation envelope [sec]
;	p11	= duration of the excitation envelope [sec]
;	p12	= decay time of the excitation envelope [sec]

;-----------------------------------------------------------------------------
; COMPULSORY GEN FUNCTIONS
;	f1	audio file
;	f19	sigmoid rise/decay shape
;_____________________________________________________________________________

; CLASS: FOG-1

;  GLOBAL KEYWORDS (default values within parentheses):
;	NUMROWS		: amount of rows (components) in the event (1)
;	ACTION-TIME	: start time of the whole event [sec] (0.0)
;	USER-FUN	: user-defined parsing function (nil)

;  LOCAL KEYWORDS:
;   E-DELS	:	entry delays [sec] (0.0)
;	DURS	:	duration [sec] (1.0)
;	AMP	:	amplitude [sec] (500.0)
;	XPF	:	transposition factor [1 = same pitch as original) (1.0)
;	SPD	:	speed (rate at which successive grains progress through the
;				   stored audio table [1 = original speed] (1.0)
;	DENS	:	density of the grains [Hz] (100.0)
;	AFUN	:	audio function number [GEN] (veena.aiff)
;	AENV	:	function number for the amplitude envelope [GEN] (fast < / slow >)
;	FRIS	:	rise time of the excitation envelope [sec] (0.01)
;	FDUR	:	duration of the excitation envelope [sec] (0.1)
;	FDEC	:	decay time of the excitation envelope [sec] (0.05)

;*****************************************************************************


sr	= 44100
kr	= 44100
ksmps	= 1
nchnls	= 1

gimaxamp init 32767 			; 16 bits


instr 1; *****************************************************************
idur   	= p3
idurosc	= 1/p3
iamp	= p4 * gimaxamp * 0.001	; amplitude range 0-1000
ixpf	= p5
ispeed	= p6
idens	= p7
ifun	= p8
iaenv	= p9

ikris	= p10
ikdur	= p11
ikdec	= p12

; amplitude envelope
	amp	oscili	iamp, idurosc, iaenv


; scaling to reflect sample rate and table length
	i1	=	sr/ftlen (ifun)
	a2	phasor	i1*ispeed

;						  	 xspd, kband,                   iolaps,    ifnb,    iphs
;		   xamp, xdens, xptch,	 koct,  kris,  kdur,  kdec,       ifna,  itotdur, ifmodel
	a1	fog	amp, idens, ixpf, a2, 0, 0, ikris, ikdur, ikdec, 1000, ifun, 19, idur, 0, 1

      out     a1

endin
