;=============================================================================
;			RAN-1.ORC
; RANDOM NUMBER GENERATION MODULATING THE AMPLITUDE OF AN OSCILLATOR
;    (FROM ACCCI, 02_41_1.ORC) / MONO
; AMPLITUDE ENVELOPE WITH LINEN, CONTROL OF AUDIO FUN FROM THE SCORE
;=============================================================================

; Timbre:       Noise spectra, with control of bandwidth and center freq
; Synthesis:    Random Number Generation
;               RANDI(02)
;               LINEN envelope on RANDI ring modulates an oscillator (2)
; Source:       Dodge(1985), p.92
; Coded:        jpg 8/92. modified ms 9/02

; NB 1: this instrument works better with short notes.
; NB 2: the apparently arbitrary amplitude range (0-1000, rather than 0-1)
;         avoids Lisp printing small values with exponential notation
;-----------------------------------------------------------------------------

;	p1	= instrument number
;	p2	= action time [sec]
;	p3	= duration [sec]
;	p4	= maximum amplitude [linear, 0.0 -> 1000.0]
;	p5	= frequency [Hz]
;	p6	= frequency of the noise module [Hz]
;	p7	= attack time of the amplitude envelope [sec]
;	p8	= decay time of the amplitude envelope [sec]
;	p9	= spectral scaler for the audio frequency [GEN number]

;-----------------------------------------------------------------------------
; COMPULSORY GEN FUNCTIONS
;	f1	sine tone (default)
;_____________________________________________________________________________

; CLASS: RAN-1

;  GLOBAL KEYWORDS (default values within parentheses):
;	NUMROWS		: amount of rows (components) in the event (1)
;	ACTION-TIME	: start time of the whole event [sec] (0.0)
;	USER-FUN	: user-defined parsing function (nil)

;  LOCAL KEYWORDS:
;	E-DELS	:	entry delays [sec] (0.0)
;	DURS	:	duration [sec] (1.0)
;	AMP	:	amplitude [sec] (500.0)
;	FREQ	:	frequency [Hz] (440.0)
;	RANF	:	frequency of the noise [Hz] (50.0)
;	ATK	:	attack time [sec] (0.2)
;	DEC	:	decay time [sec] (0.3)
;	AFUN	:	number of the audio function [GEN number] (1 = sine)


;*****************************************************************************


sr	= 44100
kr	= 44100
ksmps 	= 1
nchnls	= 1

gimaxamp init 32767 				; 16 bits


instr 1; *****************************************************************

idur		= p3
iamp		= p4 * gimaxamp * 0.001	; amplitude range 0-1000
ifq			= p5
iran_fq		= p6
iatk		= p7
idec		= p8
iaudiofun	= p9

   kenv  linen    iamp, iatk, idur, idec		; envelope
   kran  randi    kenv, iran_fq				; random numbers
   a1    oscili   kran, ifq, iaudiofun			; applied to amplitude slot
         out      a1

endin

