;******************************************************************
;		     CLASS WSHP-1
;******************************************************************

(in-package om)

; LISP-DEFINED CLASSES SHOULD RESIDE IN THE LIBRARY'S PACKAGE AND 
;   NOT IN THE USER PACKAGE, WHICH CONTAINS ALL THE CLASSES
;   GRAPHICALLY DEFINED


(defclass! wshp-1
  (cs-evt) 			; INHERIT FROM CS-EVT
  (

; GLOBAL SLOTS (LIGHT BLUE, ON THE LEFT OF THE CLASS):
;    THE METHOD BELOW TRANSFORMS THEM INTO GLOBAL SLOTS ("SHOW" UNCHEKED)
;    ATTENTION: A GLOBAL SLOT SHOULD NOT HAVE AN INITARG

   ( source-code :initform
                 (load-buffer-textfile
                  (get-orc-source (get-orc "wshp-1"))
                  'textfile "append")
                 :allocation :class
                 :type textfile
                 :accessor source-code)
   ( numchan :initform (or (get-orc-channels (get-orc "wshp-1")) 1)
             :allocation :class  :accessor numchan)
   ( globals-list :initform 
                  (get-orc-globals
                   (get-orc "wshp-1"))
                  :allocation :class
                  :type list
                  :accessor globals-list)
   (InstID :initform 1  :allocation :class  :accessor InstID)

; LOCAL SLOTS (RED, CORRESPONDING TO THE P-FIELDS)
;    ATTENTION: A GLOBAL SLOT SHOULD HAVE AN INITARG
   ( amp	:type number
		:initarg :amp 
  		:initform 500.0
		:accessor amp)
   ( freq	:type number
        	:initarg :freq 
        	:initform 220.0
        	:accessor freq)
   )


  (:documentation
   "
;=============================================================================
;				WSHP-1.ORC
; 		WAVESHAPING (FROM ACCCI, 40_02_1.ORC) / MONO
; 		AMPLITUDE ENVELOPE WITH LINEN
;=============================================================================

; Timbre:    Clarinet-like
; Synthesis: Waveshaping
;            Basic instrument with duration dependent envelope
; Source:    Risset(1969)
;            #150, Serial Excerpt with Clarinet-like Sounds by Nonlinearity
; Coded:     jpg 8/92, modified ms 9/02

; NB: the apparently arbitrary amplitude range (0-1000, rather than 0-1)
;         avoids Lisp printing small values with exponential notation
;-----------------------------------------------------------------------------

;	p1	= instrument number
;	p2	= action time [sec]
;	p3	= duration [sec]
;	p4	= maximum amplitude [linear, 0.0 -> 1000.0]
;	p5	= frequency [Hz]

;-----------------------------------------------------------------------------
; COMPULSORY GEN FUNCTIONS
;	f1	sine tone
;	f31	transfer function for the waveshaper
;_____________________________________________________________________________

; CLASS: WSHP-1

;  GLOBAL KEYWORDS (default values within parentheses):
;	NUMROWS		: amount of rows (components) in the event (1)
;	ACTION-TIME	: start time of the whole event [sec] (0.0)
;	USER-FUN	: user-defined parsing function (nil)

;  LOCAL KEYWORDS:
;	E-DELS		:	entry delays [sec] (0.0)
;	DURS		:	duration [sec] (1.0)
;	AMP		:	amplitude [sec] (500.0)
;	FREQ		:	frequency [Hz] (220.0)

;*****************************************************************************
"
   )
  (:icon 618)
  )

;------------------------------------------------------------------
;	METHODS FOR THIS CLASS
;------------------------------------------------------------------
; INCLUDE THE COMPULSORY FUNCTIONS AT THE BEGINNING OF A SCORE FILE
;     THIS METHOD MUST BE DEFINED FOR THIS CLASS

(defmethod! orc-header ((self wshp-1))
  (list

   "; GEN functions **********************************************************"
   "f1   0 4097 10 1                           ; sinus"
   ""
   "; transfer function waveshaper"
   "f31  0  4097  7 -1 1600 -.5 897 .5 1600 1"

   ))

;******************************************************************
