;=============================================================================
;				WSHP-1.ORC
; 		WAVESHAPING (FROM ACCCI, 40_02_1.ORC) / MONO
; 		AMPLITUDE ENVELOPE WITH LINEN
;=============================================================================

; Timbre:    Clarinet-like
; Synthesis: Waveshaping
;            Basic instrument with duration dependent envelope
; Source:    Risset(1969)
;            #150, Serial Excerpt with Clarinet-like Sounds by Nonlinearity
; Coded:     jpg 8/92, modified ms 9/02

; NB: the apparently arbitrary amplitude range (0-1000, rather than 0-1)
;         avoids Lisp printing small values with exponential notation
;-----------------------------------------------------------------------------

;	p1	= instrument number
;	p2	= action time [sec]
;	p3	= duration [sec]
;	p4	= maximum amplitude [linear, 0.0 -> 1000.0]
;	p5	= frequency [Hz]

;-----------------------------------------------------------------------------
; COMPULSORY GEN FUNCTIONS
;	f1	sine tone
;	f31	transfer function for the waveshaper
;_____________________________________________________________________________

; CLASS: WSHP-1

;  GLOBAL KEYWORDS (default values within parentheses):
;	NUMROWS		: amount of rows (components) in the event (1)
;	ACTION-TIME	: start time of the whole event [sec] (0.0)
;	USER-FUN	: user-defined parsing function (nil)

;  LOCAL KEYWORDS:
;	E-DELS		:	entry delays [sec] (0.0)
;	DURS		:	duration [sec] (1.0)
;	AMP		:	amplitude [sec] (500.0)
;	FREQ		:	frequency [Hz] (220.0)

;*****************************************************************************

sr	= 44100
kr	= 44100
ksmps	= 1
nchnls	= 1

gimaxamp init 32767 			; 16 bits

; *****************************************************************

instr 1
idur	= p3
iamp	= p4 * gimaxamp * 0.001			; amplitude range 0-1000
ifq		= p5

iaudiofun	= 1
itbloffset	= 2048				; must be the half of the table size
ixferfun	= 31


iatk	= 0.085					; constant attack time
idec  =  0.64					; idec for idur > 0.75
if idur > 0.75 igoto start
idec  =  idur - 0.085				; idec for idur <= 0.75

start:
        aenv    linen    itbloffset-1, iatk, idur, idec		; envelope
        a1      oscili   aenv, ifq, iaudiofun			; sinus
        a1      tablei   a1 + itbloffset, ixferfun		; transfer function
                out      a1 * iamp				; scale to amplitude
endin
