;******************************************************************
;		     CLASS AD1S-ROOT
; 	      Root class for the class AD1S
;******************************************************************

(in-package om)

; LISP-DEFINED CLASSES SHOULD RESIDE IN THE LIBRARY'S PACKAGE AND 
;   NOT IN THE USER PACKAGE, WHICH CONTAINS ALL THE CLASSES
;   GRAPHICALLY DEFINED

(defclass! ad1s-root
  (cs-evt)
  (
   ( source-code :initform
                 (load-buffer-textfile
                  (get-orc-source (get-orc "ad1s") )
                  'textfile "append")
                 :allocation :class
                 :type textfile
                 :accessor source-code)

   ( numchan :initform 2
             :allocation :class  :accessor numchan)
   ( globals-list :initform 
                  (get-orc-globals
                   (get-orc "ad1s"))
                  :allocation :class
                  :type list
                  :accessor globals-list)
   ( amp	:type number
		:initarg :amp 
  		:initform 500.0
		:accessor amp)
   ( freq	:type number
        	:initarg :freq 
        	:initform 440.0
        	:accessor freq)
   ( bal	:type number
		:initarg :bal 
  		:initform 0.5
		:accessor bal)
   ( jta	:type number
		:initarg :jta 
  		:initform 0.0
		:accessor jta)
   ( tra	:type number
		:initarg :tra 
  		:initform 0.0
		:accessor tra)
   ( trf	:type number
		:initarg :trf 
  		:initform 5.0
		:accessor trf)
   ( atk	:type number
		:initarg :atk 
  		:initform 0.01
		:accessor atk)
   ( dec	:type number
		:initarg :dec 
  		:initform 0.05
		:accessor dec)
   ( aenv	:type cs-table
		:initarg :aenv 
                                               ; x-pts y-pts decimals number size
  		:initform (make-cs-table  'Gen07  '(0 256 513) '(0 1 0) 1 "?" 513)
		:accessor aenv)
   ( jtv	:type number
		:initarg :jtv 
  		:initform 0.5
		:accessor jtv)
   ( vfq	:type number
		:initarg :vfq 
  		:initform 6.0
		:accessor vfq)
   ( fdev	:type number
		:initarg :fdev 
  		:initform 0.0
		:accessor fdev)
   ( fenv	:type cs-table
		:initarg :fenv 
                                               ; x-pts y-pts decimals number size
  		:initform (make-cs-table  'Gen07  '(0 256 513) '(0 1 0) 1 "?" 513)
		:accessor fenv)
   ( plow	:type number
		:initarg :plow 
  		:initform 0.0
		:accessor plow)
   ( pup 	:type number
		:initarg :pup 
  		:initform 0.0
		:accessor pup)
   ( penv	:type cs-table
		:initarg :penv 
                                               ; x-pts y-pts decimals number size
  		:initform (make-cs-table  'Gen07  '(0 256 513) '(0 1 0) 1 "?" 513)
		:accessor penv)
   ( pdur	:type number
		:initarg :pdur 
  		:initform 0.1
		:accessor pdur)
   )
  (:documentation
   "
Instrument AD1s.orc from CSorc.

;*****************************************************************************
;	ADDITIVE SYNTHESIS INSTRUMENT N. 11
;		TYPE 1 : COMPLETE INSTRUMENT STEREO (BALANCE MODULE WITH SINE)
;-----------------------------------------------------------------------------
;                      NEW VERSION VALID FROM CHROMA2001
;-----------------------------------------------------------------------------
; SIMPLIFIED PERCEPTUAL COMPENSATION: SINCE THE COMPUTATION IS FLOATING POINT
;    NO NEED TO REDUCE THE AMPLITUDES IF MAXIMUM IS > 1 AFTER THE COMPENSATION
; ALSO NO MODIFICATION OF THE GLOBAL AMPLITUDE TO A RELATIVE AMPLITUDE

;-----------------------------------------------------------------------------
;	p1	instrument number [11]
;	p2	action time [sec]
;	p3	duration [sec]
;	p4	global amplitude [0:1000]
;	p5	frequency [Hz]
;	p6	stereo pan   [0 = left / 1 = right]
;	p7	jitter amplitude, with internal perceptual compensation [0:1]
;	p8	tremolo amplitude [0:1]
;	p9	tremolo frequency [Hz]
;	p10	attack duration [sec]
;	p11	decay duration [sec], with internal readjustment if needed
;	p12 	function number for amplitude envelope
;	p13	balance freq.jitter/vibrato [1 = jitter / 0 = vibrato]
;	p14	vibrato frequency [Hz]
;	p15	frequency deviation [semitones]
;	p16	function number for frequency envelope
;	p17	lower interval for portamento [semitones]
;	p18	upper interval for portamento [semitones]
;	p19	function number for portamento envelope
;	p20	portamento duration [sec]

;-----------------------------------------------------------------------------
; KONTROL KEYWORDS FOR CHROMA2001 (CTL1)

;  GLOBAL KEYWORDS:
;	ACTION-TIME	: start time of the whole event (0.0)
;	DURTOT		: total duration of the event (longer components are
;                    		automatically clipped) (1.0)
;	AMPTOT		: amplitude scaler (1.0)
;	NUMROWS		: amount of rows (components) in the event (1)

;  LOCAL KEYWORDS:
;  GENERAL
;   E-DELS		:	entry delays [sec] (0.0)
;	DURS		:	duration [0-1, scaler of DURTOT] (1.0)
;	AMP		:	amplitude [0-1000, scaler of AMPTOT] (500.0)
;	FREQ		:	frequency [Hz] (440.0)
;	BAL		:	panpot [0=left, 1=right] (STEREO VERSION ONLY) (0.5)

;  AMPLITUDE ENVELOPE
;	ATK		:	attack time [sec] (0.01)
;	DEC		:	decay time [sec] (0.05)
;	AENV		:	amplitude envelope [GEN]

;  AMPLITUDE MODULATION
;	JTA		:	jitter amplitude [0-1, % of amplitude + compensation] (0.0)
;	TRA		:	tremolo amplitude [0-1, % of amplitude] (0.0)
;	TRF		:	tremolo frequency [Hz] (5.0)

;  FREQUENCY MODULATION
;	JTV		:	vibrato/jitter panpot [0 = all vibrato, 1 = all jitter] (0.5)
;	VFQ		:	vibrato frequency [Hz] (6.0)
;	FDEV		:	frequency deviation [semitones] (0.0)
;	FENV		:	frequency envelope [GEN]

;  PORTAMENTO
;	PLOW		:	lowest (minimum) portamento amplitude [semitones] (0.0)
;	PUP		:	upper (maximum) portamento amplitude [semitones] (0.0)
;	PENV		:	portamento envelope [GEN]
;	PDUR		:	portamento duration [sec[ (0.1)

;  SUB-COMPONENTS
;	NPART		:	number of sub-components (0)
;	STON		:	aleatoric frequency distribution for each sub-component (0.06)
;				POSITIVE = distribution is linear
;				NEGATIVE = distribution is logarithmic
;	ED2		:	entry delay of each sub-compoment [sec] (0.01)
;				   (cumulative effect)
;	DUR2		:	duration of each sub-component [0-1, nil = use the duration
;				  of the main component] (1.0)
;	AMP2		:	amplitude of each sub-component [0-1, nil = use the amplitude
;				  of the main component] (1.0)
;-----------------------------------------------------------------------------
; COMPULSORY FUNCTIONS :
;	f1	audio sine wave
;	f2	vibrato (sine wave)
;	f3	tremolo (sine wave)
; FUNCTIONS USED BY OTHER INSTRUMENTS
;   f4	function for janusing
;	f5	large non-interpolating sine for the FOF
; OTHER USEFUL FUNCTIONS (CHROMA2001)
;   f6	constant 1
;   f7	constant 0
;   f8	ascending line (0->1) for reading into a table
;   f9	simple triangle (0-1-0) for default envelopes
;************************************************************************

   "
   )
  (:icon 610)
  )


;******************************************************************
;			SUB-CLASSES
;******************************************************************
; AD1S
;------------------------------------------------------------------

(defclass! ad1s
  (ad1s-root)
  (
   ; GLOBAL SLOTS: THE METHOD BELOW TRANSFORMS THEM INTO GLOBAL SLOTS
   ;    ATTENTION: A GLOBAL SLOT SHOULD NOT HAVE AN INITARG
   ( durtot :type number
            :initform 1.0
            :accessor durtot)
   ( amptot :type number
            :initform 1.0
            :accessor amptot)
   ( npart :type number
           :initarg :npart 
           :initform 0
           :accessor npart)
   ( ston :type number
          :initarg :ston
          :initform 0.06
          :accessor ston)
   ( ed2 :type number
         :initarg :ed2
         :initform 0.01
         :accessor ed2)
   ( dur2 :type number
          :initarg :dur2
          :initform -1
          :accessor dur2)
   ( amp2 :type number
          :initarg :amp2
          :initform -1
          :accessor amp2))
  (:documentation
   
   "Extension of the instrument ADD1s, class ad1s-root.

Added global slots:
  durtot [sec] = total duration of the whole event
  amptot [0-1] = global amplitude scaler

Added local slots:
  npart [int] = number of sub-components
  ston [0-1]  = maximum frequency deviation for the sub-components
		positive = linear distribution (value in %)
		negative = logarithmic distribution (value in semitones)
  ed2 [sec]   = entry delay of each sub-component (cumulative effect) 
  dur2 [0-1]  = duration of each sub-component (scaler of DURTOT) (nil)
                   (if nil, use the value of the main component)
  amp2 [0-1]  = amplitude of each sub-component (scaler of AMPTOT) (nil)
                   (if nil, use the value of the main component)
")
  (:icon 610)
  )


;------------------------------------------------------------------
;	METHODS FOR THIS CLASS
;------------------------------------------------------------------
; ADD SOME GLOBAL SLOTS (THE FIRST ONES ON THE LEFT OF A CLASS BOX)
;     THIS METHOD MUST BE DEFINED FOR THIS CLASS

(defmethod get-slot-in-out-names ((self ad1s))
; MEANING OF THE FIELDS:
;   SLOT NAMES (THE FIRST 3 ARE THE DEFAULT ONES)
;   DEFAULT VALUES
;   DOCUMENTATION
;   T IF IT IS A MENU
   (values '("self" "numcols" "action-time" "user-fun" "durtot" "amptot")
           '(nil 1 0 nil 1.0 1.0)
           '("object or object list"
             "How many vertical components for init ?"
             "local event offset (in seconds)"
             "user-defined function"
             "Total duration for the whole event [sec]"
             "Global amplitude scaler [0-1]")
           '(nil nil nil nil nil nil)))


; SPECIFY WHICH ARE THE GLOBAL SLOTS
(defmethod fixed-slots-list ((x om::ad1s))
  "Specifies the fixed (global) slots for this class"
  '(numcols action-time user-fun durtot amptot))


; ELIMINATE THE LAST 5 SLOTS IN THE INSTRUMENT (NOT NEEDED IN THE SCORE)
(defmethod nullslots ((x om::ad1s))
  "Eliminate the last 5 slots from the score"
  5)

;------------------------------------------------------------------
; INCLUDE THE COMPULSORY FUNCTIONS AT THE BEGINNING OF A SCORE FILE
;     THIS METHOD MUST BE DEFINED FOR THIS CLASS

(defmethod! orc-header ((self ad1s))
  (list

   "; Marco's GEN functions *********************************************************"
   "; AUDIO WAVE"
   "f1  0   4097   9  1  1  0"

   "; FREQUENCY VIBRATO"
   "f2  0   4097   9  1  1  0"

   "; AMPLITUDE VIBRATO"
   "f3  0   4097   9  1  1  0"

   ))

(defmethod! orc-header ((self ad1s-root))
  (list

   "; Marco's GEN functions *********************************************************"
   "; AUDIO WAVE"
   "f1  0   4097   9  1  1  0"

   "; FREQUENCY VIBRATO"
   "f2  0   4097   9  1  1  0"

   "; AMPLITUDE VIBRATO"
   "f3  0   4097   9  1  1  0"

   ))

;******************************************************************




;;; ...
(defmethod get-obj-dur ((self ad1s-root))
  (round (* 1000 (durtot self))))