;;;================================================================================================================
;;; SuperVP Sound Analysis Tools
;;; jb - 2005 
;;;================================================================================================================


(in-package :om)



;;;================================================================================================================
;;; FFT
;;;================================================================================================================

(defmethod! fft ((fileName string) begin-time end-time &optional windowsize fftsize step windowtype gabarit outfile)
  :initvals '(nil nil nil 4096 4096 256 "hanning" 1 "fft.sdif")
  :menuins (list (list 6 '(("Blackman" "blackman") ("Hanning" "hanning") ("Hamming" "hamming")))
                 (list 7 '(("log amplitude" 0) ("amplitude" 1) ("phase" 2) ("phase / amplitude" 3) ("frequency / amplitude" 4) ("real / imaginary" 5))))
  :icon 952
  :doc "Calculates the FFT of a Sound file using SVP and returns the resulting SDIF file pathname

- FileName : the pathname of the sound file (Aiff) to be analysed

- begin-time : the begin time of the analysis

- end-time : the end time of the analysis

- windowsize : the number of samples of the analysis window

- fftsize : the number of points in fft

- step : the number of samples between two successive analysis windows

- windowtype : the shape of the analysis window

- gabarit : the output fft format
            0 = logarithmic amplitudes
            1 = amplitude
            2 = phase
            3 = phase, amplitude
            4 = frequency, amplitude
            5 = real, imaginary 

Note that only sdiffiles at gabarit 3 and 5 can be used to recalculate a sound file using IFFT."
  (if (and *SVP-PATH* (probe-file *SVP-PATH*))
    (let* ((outname (if outfile 
                      (if (pathnamep outfile) outfile (outfile outfile))
                      (om-CHOOSE-NEW-FILE-DIALOG :prompt "Choose a new SDIF FFT file")))
           (unix-outname (om-path2cmdpath outname))
           (beginstr (if begin-time (format nil "-B~D " begin-time) ""))
           (endstr (if end-time (format nil "-E~D " end-time) ""))
           (fftstr (format nil "-N~D -M~D -W~D -I~D " 
                           (if fftsize fftsize 4096) (if windowsize windowsize 4000) 
                           (if windowtype windowtype "hanning") (if step step 250)))
           (ostr (format nil "-OS~D " (if (and gabarit (>= gabarit 0) (<= gabarit 5)) gabarit 
                                          (progn (om-message-dialog (format nil "FFT gabarit ~D does not exist.~% FFT will be formated at gabarit 1" gabarit))
                                                 1)))) 
           
           ;(cmd (string+ (om-path2cmdpath *SVP-PATH*) 
           ;              " -v -t -S" (om-path2cmdpath fileName) " -Afft -U "
           ;              fftstr beginstr endstr ostr unix-outname))
           (cmd (format nil "~s -v -t -S~s -Afft -U ~A ~A ~A ~s ~s" 
                        (om-path2cmdpath *SVP-PATH*)
                        (om-path2cmdpath fileName)
                        fftstr beginstr endstr ostr unix-outname))
           
           )
      (print (string+ "SVP PROCESS : " cmd))
      (om-cmd-line cmd *svp-console*)
      outname)
    (om-beep-msg "SUPERVP not found!"))
  )

(defmethod! fft ((fileName pathName) begin-time end-time &optional windowsize fftsize step windowtype gabarit outfile)
  :initvals '(nil nil nil 4096 4000 250 "hanning" 1)
  :menuins (list (list 6 '(("Hamming" "hamming")("Hanning" "hanning")("Blackman" "blackman"))))
  (fft (om-namestring filename) begin-time end-time fftsize windowsize step windowtype gabarit outfile))

(defmethod! fft ((self sound) begin-time end-time &optional windowsize fftsize step windowtype gabarit outfile)
  :initvals '(nil nil nil 4096 4000 250 "hanning" 1)
  :menuins (list (list 6 '(("Hamming" "hamming")("Hanning" "hanning")("Blackman" "blackman"))))
  (fft (sound-path self) begin-time end-time fftsize windowsize step windowtype gabarit outfile))


;;;================================================================================================================
;;; FFT TOOLS
;;;================================================================================================================


;;; "sous echantilloner" une fft...
(defmethod! format-fft ((self SdifFile) streamNum f-scale f-min f-max f-step)
  :initvals '(nil 0 "lin" 0 22000 100) 
  (let ((fftp nil))
    (loop for item in (framedesc self) while (not fftp) do
          (when (string-equal (string (first item)) "IGBG")
            (setf fftp t)))
    (if (not fftp) (om-message-dialog "Sorry this SDIF file does not contain a SVP fft")
        (let* ((error nil) time (onemat nil)
               (out-path (om-CHOOSE-new-FILE-DIALOG))
               (outfile (sdif-open-file (om-path2cmdpath out-path) 1))
               val (newMat nil) (data nil) str sstr)
          (sdif::SdifFWriteGeneralHeader outfile)
          (write-1nvt-table outfile
                            (list "Author" )
                            (list (string+ "OM " *version-string*)))
          (setf str (format nil "{1MTD 1GB4 {frequency, amplitude} 1FTD XFFT {1GB4  fftata;}}"))
          (setf sstr (sdif::SdifStringNew)) 
          (Sdif-String-Append sstr str)
          (sdif::SdifStringGetC sstr)
          (sdif::SdifFGetAllTypefromSdifString outfile sstr)
          (sdif::SdifFWriteAllASCIIChunks outfile)
          (loop for item in (framesdesc self)
                while (not error)
                for i = 0 then (+ i 1) do
                (setf onemat nil)
                (setf newmat nil)
                (setf time (second item))
                (if (and (= streamNum (third item)) (string-equal (string (first item)) "1GB4")) 
                  (loop for mat in (fifth item)
                        for j = 0 then (+ j 1) do
                        (if (string-equal (first mat) "1GB4")
                          (let ((ptrfile (goto-sdif-frame-mat self i j)) col row val r1 r2)
                            (when ptrfile
                              (progn
                                (sdif::SdifFReadMatrixHeader ptrfile)
                                (setf col (sdif::SdifFCurrNbCol ptrfile))
                                (setf row (sdif::SdifFCurrNbRow ptrfile))
                                (setf onemat (loop for k from 1 to row 
                                                   collect (progn
                                                             (sdif::SdifFReadOneRow ptrfile)
                                                             (loop for c from 1 to col collect
                                                                   (sdif::SdifFCurrOneRowCol ptrfile c)))))
                                (dynamic-close self ptrfile)
                                (loop for f from f-min to f-max by f-step do
                                      (setf val (find-amplitude f onemat))
                                      (push (list f val) newmat))
                                (print (format nil "writing frame t=~D ..." time))
                                (write-1GB4-frame outfile (reverse newMat) time)
                                ;(push (list time (reverse newmat)) data)
                                ))
                            )))))
          (sdif-close-file outfile)
          (om-namestring out-path)
          ;(reverse data)
          ))))


(defmethod! fftdata2sdif ((fftdata list) &optional timelist)
  :initvals '(nil 0 "lin" 0 22000 100) 
  (let* ((error nil) time
    (out-path (om-CHOOSE-new-FILE-DIALOG))
    (outfile (sdif-open-file (om-path2cmdpath out-path) 1))
    val (data nil) str sstr)
  (sdif::SdifFWriteGeneralHeader outfile)
  (write-1nvt-table outfile
                    (list "Author" )
                    (list (string+ "OM " *version-string*)))
  (setf str (format nil "{1MTD 1GB1 {amplitude} 1FTD XFFT {1GB1  fftata;}}"))
  (setf sstr (sdif::SdifStringNew)) 
  (sdif::SdifStringAppend sstr str)
  (sdif::SdifStringGetC sstr)
  (sdif::SdifFGetAllTypefromSdifString outfile sstr)
  (sdif::SdifFWriteAllASCIIChunks outfile)
  (loop for framedata in fftdata
        while (not error)
        for i = 0 then (+ i 1) do
        (setf time (if timelist (nth i timeList) i))
        ;(print (format nil "writing frame t=~D ..." time))
        (write-1GB1-frame outfile framedata time)
        ;(push (list time framedata) data)
        )
  (sdif-close-file outfile)
  (om-namestring out-path)
))   


(defun find-amplitude (freq f-a-list)
  (let ((min-dist 44000) (val 0) dist)
    (loop for f-a in f-a-list do
          (setf dist (abs (- freq (first f-a))))
          (if (<= dist min-dist) (setf min-dist dist val (second f-a))))
    val))


(defun write-1GB1-frame (ptr points time)
   (without-interrupts  
    (let* ((datatype 4)
           (numcols 1)
           (numlines (length points))
           (framesize (+ 32 (calc-pad (* numcols datatype numlines))))
           (values (om-make-pointer (* numcols datatype numlines))))
      (sdif::SdifFSetCurrFrameHeader ptr (sdif-string-to-signature "XFFT") framesize 1 0 (coerce time 'double-float))
      (sdif::SdifFWriteFrameHeader ptr)
      (loop for i from 0 to (- numlines 1)
            for val in points do
            (om-write-ptr values (* i datatype) :single-float (coerce val 'single-float)))
      (sdif::SdifFWriteMatrix ptr (sdif-string-to-signature "1GB1") datatype numlines numcols values)
      ;(print (format nil "FRAME ~D -- ~D" time values))
)))


(defun write-1GB4-frame (ptr points time)
   (without-interrupts  
    (let* ((datatype 4)
           (numcols 2)
           ;(mat-points (mat-trans points))
           (numlines (length points))
           (framesize (+ 32 (calc-pad (* numcols datatype numlines))))
           (values (om-make-pointer (* numcols datatype  numlines))))
      (sdif::SdifFSetCurrFrameHeader ptr (sdif-string-to-signature "XFFT") framesize 1  0 (coerce time 'double-float))
      (sdif::SdifFWriteFrameHeader ptr)
      (loop for i from 0 to (- numlines 1)
            for sdifrow in points do
            (loop for j from 0 to (- numcols 1)
                  for val in sdifrow do
                  (om-write-ptr values (+ (* j datatype) (* i numcols datatype)) :single-float (coerce val 'single-float))))
      (sdif::SdifFWriteMatrix ptr (sdif-string-to-signature "1GB4") datatype numlines numcols values))))




;;;================================================================================================================
;;; FUNDAMENTAL FREQUENCY ESTIMATE
;;;================================================================================================================

(defmethod! f0-estimate ((fileName string) begin-time end-time 
                         fund-minfreq fund-maxfreq spectrum-maxfreq noise-threshold smooth-order
                         &optional windowsize fftsize step windowtype outfile)
  :initvals '(nil nil nil 50.0 1000.0 4000.0 50.0 3 4096 4096 256 "hanning" "f0.sdif")
  :menuins (list (list 11 '(("Blackman" "blackman")("Hanning" "hanning")("Hamming" "hamming"))))
  :icon 952
  :doc "Calculates the Fundamental Fequency Estimation of a Sound file using SVP and returns the resulting SDIF file pathname

- FileName : the pathname of the sound file (Aiff) to be analysed

- begin-time : the begin time of the analysis (s)
 
- end-time : the end time of the analysis (s)

- fund-minfreq : Fundamental minimal frequency (Hz)

- fund-maxfreq : Fundamental maximal frequency (Hz)

- spectrum-maxfreq : Maximal frequency in spectrum (Hz)

- noise-threshold : Noise threshold (dB)

- smooth-order : Smooth Order (int)

- windowsize : the number of samples of the analysis window

- fftsize : the number of points of fft

- step : the number of samples between two successive analysis windows

- windowtype : the shape of the analysis window

"
  (if (and *SVP-PATH* (probe-file *SVP-PATH*))
    (let* ((outname (if outfile 
                      (if (pathnamep outfile) outfile (outfile outfile))
                      (om-CHOOSE-NEW-FILE-DIALOG :prompt "Choose a new SDIF F0 file")))
           (unix-outname (om-path2cmdpath outname))
           (beginstr (if begin-time (format nil "-B~D " begin-time) ""))
           (endstr (if end-time (format nil "-E~D " end-time) ""))
           (fftstr (format nil "-N~D -M~D -W~D -I~D " 
                           (if fftsize fftsize 4096) (if windowsize windowsize 4000) 
                           (if windowtype windowtype "hanning") (if step step 250)))
           (f0params (format nil "fm~D fM~D F~D sn~D smooth~D " 
                             fund-minfreq fund-maxfreq spectrum-maxfreq noise-threshold smooth-order))
           (cmd (format nil "~s -v -t -ns -U -S~s -Af0 ~s ~A ~A ~A -OS0 ~s" 
                        (om-path2cmdpath *SVP-PATH*)
                        (om-path2cmdpath fileName)
                        f0params
                        fftstr beginstr endstr unix-outname))
           
           )
      (print (string+ "SVP PROCESS : " cmd))
      (om-cmd-line cmd *svp-console*)
      outname)
    (om-beep-msg "SUPERVP not found!"))
  )


(defmethod! f0-estimate ((fileName pathName) begin-time end-time 
                         fund-minfreq fund-maxfreq spectrum-maxfreq noise-threshold smooth-order
                         &optional windowsize fftsize step windowtype outfile)
  :initvals '(nil nil nil 50.0 1000.0 4000.0 50.0 3 4096 4096 256 "hanning" nil)
  :menuins (list (list 11 '(("Hamming" "hamming")("Hanning" "hanning")("Blackman" "blackman"))))
  :icon 952
  (f0-estimate (om-namestring filename) begin-time end-time 
               fund-minfreq fund-maxfreq spectrum-maxfreq noise-threshold smooth-order
               windowsize fftsize step windowtype outfile))

(defmethod! f0-estimate ((self sound) begin-time end-time 
                         fund-minfreq fund-maxfreq spectrum-maxfreq noise-threshold smooth-order
                         &optional windowsize fftsize step windowtype outfile)
  :initvals '(nil nil nil 50.0 1000.0 4000.0 50.0 3 4096 4096 256 "hanning" nil)
  :menuins (list (list 11 '(("Hamming" "hamming")("Hanning" "hanning")("Blackman" "blackman"))))
  :icon 952
  (f0-estimate (sound-path self) begin-time end-time 
               fund-minfreq fund-maxfreq spectrum-maxfreq noise-threshold smooth-order
               windowsize fftsize step windowtype outfile))

;;;================================================================================================================
;;; F0 TOOLS
;;;================================================================================================================

(defmethod! f0->bpf ((self sdiffile))
  :icon 608
  (let ((streamid 0) streams bpf)
    (loop for f in (framesdesc self) do
          (when (string-equal "1FQ0" (car f))
            (unless (member (nth 2 f) streams) (push (nth 2 f) streams))))
    (when streams
      (setf streamid (car streams))
      (multiple-value-bind (y x) (getsdifdata self streamid "1FQ0" "1FQ0" 0 nil nil nil nil)
        (setf bpf (simple-bpf-from-list x (flat y) 'bpf 3)))
        )
    bpf))


;;;================================================================================================================
;;; TRANSIENT DETECTION : GENERATE MARKERS
;;;================================================================================================================

(defmethod! transient-detection ((fileName string)
                                 treshold minfreq maxfreq
                                 &optional windowsize fftsize step-oversamp windowtype outfile)
  :initvals '(nil 1.4 0.0 22050.0 4096 4096 8 "hanning" "markers.sdif")
  :menuins (list (list 6 '(("1/4" 4) ("1/8" 8) ("1/16" 16) ("1/32" 32)))
                 (list 7 '(("Blackman" "blackman")("Hanning" "hanning")("Hamming" "hamming"))))
  :icon 952
  :doc "Generates markers using transient detection

- FileName : the pathname of the sound file (Aiff) to be analysed

- treshold : transient detection treshold (1-10)

- minfreq : minimal detection frequency (Hz)

- maxfreq : maximal detection frequency (Hz)

- windowsize : the number of samples of the analysis window

- fftsize : the number of points of fft

- step-oversamp : the oversampling proportion between two successive analysis windows

- windowtype : the shape of the analysis window

"
  
  ;;; supervp -t -U  -S"/Applications/AudioSculpt 2.5.2/Sounds/africa.aiff" -Afft  -N2048 -M2048 -oversamp 8 -Wblackman  
  ;;; -OT -td_thresh 1.4 -td_G 2.5 -td_band 0.0,22050.0 -td_nument 10.0 "/Applications/AudioSculpt 2.5.2/Markers/tempmark-2Kae.sdif"
  (if (and *SVP-PATH* (probe-file *SVP-PATH*))
    (let* ((outname (if outfile 
                      (if (pathnamep outfile) outfile (outfile outfile))
                      (om-CHOOSE-NEW-FILE-DIALOG :prompt "Choose a new SDIF markers file")))
           (unix-outname (om-path2cmdpath outname))
           (fftstr (format nil "-N~D -M~D -W~D -oversamp ~D " 
                           (if fftsize fftsize 4096) (if windowsize windowsize 4000) 
                           (if windowtype windowtype "hanning") (if step-oversamp step-oversamp 8)))
           (transient-params (format nil "-OT -td_thresh ~D -td_G 2.5 -td_band ~D,~D -td_nument 10.0"
                                     treshold minfreq maxfreq))
           (cmd (format nil "~s -v -t -ns -S~s -Afft ~A ~A ~s" 
                        (om-path2cmdpath *SVP-PATH*)
                        (om-path2cmdpath fileName)
                        fftstr transient-params
                        unix-outname))
           
           )
      (print (string+ "SVP PROCESS : " cmd))
      (om-cmd-line cmd *svp-console*)
      outname)
    (om-beep-msg "SUPERVP not found!"))
  )


(defmethod! transient-detection ((fileName pathName) 
                         treshold minfreq maxfreq
                         &optional windowsize fftsize step windowtype outfile)
  (transient-detection (om-namestring filename) 
               treshold minfreq maxfreq
               windowsize fftsize step windowtype outfile))

(defmethod! transient-detection ((self sound)  
                         treshold minfreq maxfreq
                         &optional windowsize fftsize step windowtype outfile)
  (transient-detection (sound-path self) 
               treshold minfreq maxfreq
               windowsize fftsize step windowtype outfile))

;;;================================================================================================================
;;; MARKERS TOOLS
;;;================================================================================================================

(defmethod! get-mrk-onsets ((self sdiffile) &optional (streamid 0))
  :icon 608
  (get-times self streamid "1MRK" "1BEG" nil nil))




;;;================================================================================================================
;;; PACK
;;;================================================================================================================

(defvar *svp-analyse-package* (omNG-protect-object (omNG-make-new-package "supervp")))
(addPackage2Pack *svp-analyse-package* *analysispackage*)
(AddGenFun2Pack  '(fft f0-estimate transient-detection) *svp-analyse-package*)

