(in-package :om)


(defvar *SVP-PATH* "path to svp")
(defvar *param-files-folder* nil)

(defvar *delete-param-files* t)

(defvar *svp-console* nil)


;;;== PARAM FILES ===

(defmethod! set-paramfiles-folder ((path string))
  :icon 186
  (setf *param-files-folder* path))

(defmethod! set-paramfiles-folder ((path pathname))
  :icon 186
  (setf *param-files-folder* (namestring path)))

(defmethod! paramfile ((path string) &optional (unix nil))
  :icon 186
  (let ((pa (string+ (namestring *param-files-folder*) path)))
    (if unix (om-path2cmdpath pa) pa)))

(defmethod! paramfile ((path null) &optional (unix nil))
  :icon 186
  (let ((pa (namestring *param-files-folder*)))
    (if unix (om-path2cmdpath pa) pa)))


(AddGenFun2Pack '(paramfile) *soundfolderspackage*)

;;;=== SVP PATH ===

(defmethod! set-svp-path ((path string))
  :icon 186
  (setf *SVP-PATH* path))

(defmethod! set-svp-path ((path pathname))
  :icon 186
  (setf *SVP-PATH* (namestring path)))

(defmethod! svp-path (&optional (unix nil))
  :icon 186
  (let ((pa (namestring *SVP-PATH*)))
    (if unix (om-path2cmdpath pa) 
        pa)))



(defun init-svp ()
  (setf *SVP-PATH* (om-external-app "supervp"))
  (setf *param-files-folder* *om-infiles-folder*)
  )

(add-init-user-func 'init-svp)    

(defmethod get-external-def-vals ((module (eql 'svp))) 
  (list (om-external-app "supervp")
        *om-infiles-folder*
        nil
        *delete-param-files*))

(defmethod save-external-prefs ((module (eql 'svp))) 
  `(list ,(om-save-pathname *SVP-PATH*)
         ,(om-save-pathname *param-files-folder*)
         ,*svp-console*
         ,*delete-param-files*))

(defmethod put-external-preferences ((module (eql 'svp)) list-prefs)
  (if (and (nth 0 list-prefs) (probe-file (nth 0 list-prefs)))
    (setf *SVP-PATH* (nth 0 list-prefs))
    (push "SVP" *restore-defaults*))
  (if (and (nth 1 list-prefs) (probe-file (nth 1 list-prefs)))
    (setf *param-files-folder* (nth 1 list-prefs))
    (push "params" *restore-defaults*))
  (when (>= 3 (length list-prefs)) 
    (setf *svp-console* (nth 2 list-prefs)))
  (setf *delete-param-files* (nth 3 list-prefs))
  )

(defmethod show-external-prefs-dialog ((module (eql 'svp)) prefs)
  (let* ((rep-list (list (nth 0 prefs) (nth 1 prefs) (nth 2 prefs) (nth 3 prefs)))
         (dialog (om-make-window 'om-dialog
                                 :window-title "SVP options"
                                 :size (om-make-point 340 220)
                                 :position :centered
                                 :resizable nil :maximize nil :close nil))
         pathtext paramstext consolebox parambox)
    (om-add-subviews dialog
         
         (om-make-dialog-item 'om-static-text  (om-make-point 10 20) (om-make-point 120 15) "SupeVP Location :" :font *om-default-font2b*)
         (setf pathtext (om-make-dialog-item 'om-static-text  (om-make-point 10 40) (om-make-point 200 45) (om-namestring *SVP-PATH*)
                                           :font *om-default-font1*))
         (om-make-dialog-item 'om-button (om-make-point 250 40) (om-make-point 70 15) "Browse" 
                              :di-action (om-dialog-item-act item
                                           (declare (ignore item))
                                           (let ((newpath (om-choose-file-dialog)))
                                             (when newpath 
                                               (om-set-dialog-item-text pathtext (om-namestring newpath))
                                               ))))

         (om-make-dialog-item 'om-static-text  (om-make-point 10 80) (om-make-point 120 15) "Params files folder :" :font *om-default-font2b*)
         (setf paramstext (om-make-dialog-item 'om-static-text  (om-make-point 10 100) (om-make-point 200 45) (om-namestring *param-files-folder*)
                                           :font *om-default-font1*))
         (om-make-dialog-item 'om-button (om-make-point 250 90) (om-make-point 70 15) "Browse" 
                              :di-action (om-dialog-item-act item
                                           (declare (ignore item))
                                           (let ((newpath (om-choose-directory-dialog :directory *param-files-folder*)))
                                             (when newpath 
                                               (om-set-dialog-item-text paramstext (om-namestring newpath))
                                               ))))

         (setf consolebox (om-make-dialog-item 'om-check-box (om-make-point 10 140) (om-make-point 120 15) " Show Console" 
                              :checked-p *svp-console*
                              :font *om-default-font2b*
                              ))
         
         (setf parambox (om-make-dialog-item 'om-check-box (om-make-point 10 160) (om-make-point 160 15) " Delete parameter files" 
                              :checked-p *delete-param-files*
                              :font *om-default-font2b*
                              ))
         
         (om-make-dialog-item 'om-button (om-make-point 140 185) (om-make-point 60 20) "OK"
                           :di-action (om-dialog-item-act item
                                        (let ((argerror nil)
                                              (svptxt (om-dialog-item-text pathtext))
                                              (partxt (om-dialog-item-text paramstext)))
                                          (if (probe-file svptxt) 
                                            (setf (nth 0 rep-list) svptxt)
                                            (setf argerror t))
                                          (if (probe-file partxt) 
                                            (setf (nth 1 rep-list) partxt)
                                            (setf argerror t))
                                          (setf (nth 2 rep-list) (om-checked-p consolebox))
                                          (if argerror
                                            (om-message-dialog (format nil "Error in an SVP option.~% Preferences could not be recorded.")) 
                                            (om-return-from-modal-dialog dialog rep-list))
                                          ))
                           )

      (om-make-dialog-item 'om-button (om-make-point 230 185) (om-make-point 60 20) "Cancel"
                           :di-action (om-dialog-item-act item
                                        (om-return-from-modal-dialog dialog nil)))
      
      )
    (om-modal-dialog dialog)))


