;;;================================================================================================================
;;; SuperVP Sound Processing
;;; jb 2005
;;;================================================================================================================

(in-package :om)
  
;;; CALL SVP
(defmethod! supervp-command ((cmd string))
  :icon 950  
  (if (probe-file (om-external-app "supervp"))
    (om-cmd-line (format nil "~s ~S" (om-path2cmdpath (om-external-app "supervp")) cmd) *svp-console*) 
    (print "SVP is not available.")
    ))


;;;=================================================================================================
;;; TOOLS
;;;=================================================================================================

(defvar *tmpparfiles* nil)

(defun clean-tmp-files ()
  (loop for file in *tmpparfiles* do (delete-file file))
  (setf *tmpparfiles* nil))


;;; format un textfile pour faire des params
(defmethod get-lists-from-buffer ((self TextFile))
  (ignore-errors 
   (let* ((fred-buff (buffer-text self))
          (size (om-buffer-size fred-buff))
          (pos (om-buffer-skip-forward fred-buff 0 size))
          exp)
     (om-with-cursor *om-wait-cursor*
       (loop while (< pos size) collect
             (let ((rep nil) (cr nil))
               (loop while (and (not cr) (< pos size)) do 
                     (multiple-value-bind (exp length) (read-from-string (om-buffer-text fred-buff) nil nil :start pos)
                       (setq pos  (or (om-buffer-skip-forward fred-buff length size)
                                      (1+ size)))
                       (setf cr (equal (om-buffer-char fred-buff (- pos 1)) #\Newline))
                       (pushr exp rep)
                       ))
               rep))))))



(defmethod save-params ((self bpf) (file pathname))
  (with-open-file (out file :direction :output 
                         :if-does-not-exist :create :if-exists :supersede)
        (loop for x in (x-points self) 
              for y in (y-points self) do
              (format out "~D ~D~%" x y))
        )
  file)

;;; ((x1 y1 ... yn) (x2 y2 ... yn) ...)
(defmethod save-params ((self list) (file pathname))
  (with-open-file (out file :direction :output 
                       :if-does-not-exist :create :if-exists :supersede)
    (loop for elt in self do
          (loop for p in (list! elt) do (format out "~D " p))
          (format out "~%"))
    )
  file)


(defmethod save-params ((self textfile) (file pathname))
  (let ((parlist (get-lists-from-buffer self)))
    (with-open-file (out file :direction :output 
                         :if-does-not-exist :create :if-exists :supersede)
      (loop for line in parlist do
            (loop for p in line do (format out "~D " p))
            (Format Out "~%"))
      ))
  file)

(defmethod save-params ((self t) (file string))
  (save-params self (pathname file)))


;;;=================================================================================================
;;; TIME STRETCH
;;;=================================================================================================

;;; TIME STRETCH DIRECT
(defmethod svp-t-stretch ((self sound) param-file out-name fftsize windowsize)
  (let* ((out (if out-name out-name "snd-out"))
         (outpath (unique-pathname *om-outfiles-folder* out "aiff")))
    (if (pathnamep param-file) 
      (om-cmd-line (format nil "~s -t -v -Z -A -S~s -D~s -M~D -N~D ~s" 
                           (om-path2cmdpath *SVP-PATH*)
                           (om-path2cmdpath (sound-path self))
                           (om-path2cmdpath param-file)
                           windowsize fftsize 
                         (om-path2cmdpath outpath)) *svp-console*)
      (om-cmd-line (format nil "~s -t -v -Z -A -S~s -D~D -M~D -N~D ~s" 
                           (om-path2cmdpath *SVP-PATH*)
                           (om-path2cmdpath (sound-path self))
                           param-file
                           windowsize fftsize 
                         (om-path2cmdpath outpath)) *svp-console*))
    (om-namestring outpath)))


(defmethod! svp-time-stretch ((self sound) (param string) &optional out-name (fftsize 4096) (windowsize 2000))
  :icon 950  
  (svp-t-stretch self param out-name windowsize fftsize))

(defmethod! svp-time-stretch ((self sound) (param number) &optional out-name (fftsize 4096) (windowsize 2000))
  :icon 950  
  (svp-t-stretch self param out-name windowsize fftsize))



;;;=================================================================================================
;;; TRANSPOSITION
;;;=================================================================================================


;;; TRANSPOSITION DIRECT 
(defmethod svp-transp ((self sound) param-file out-name fftsize windowsize)
  (let* ((out (if out-name out-name "snd-out"))
         (outpath (unique-pathname *om-outfiles-folder* out "aiff")))
    (if (pathnamep param-file) 
      (om-cmd-line (format nil "~s -t -v -Z -A -S~s -trans ~s -M~D -N~D ~s" 
                         (om-path2cmdpath *SVP-PATH*)
                         (om-path2cmdpath (sound-path self))
                         (om-path2cmdpath param-file)
                         windowsize fftsize 
                         (om-path2cmdpath outpath)) *svp-console*)
      (om-cmd-line (format nil "~s -t -v -Z -A -S~s -trans ~D -M~D -N~D ~s" 
                         (om-path2cmdpath *SVP-PATH*)
                         (om-path2cmdpath (sound-path self))
                         param-file
                         windowsize fftsize 
                         (om-path2cmdpath outpath)) *svp-console*))
    (om-namestring outpath)))


(defmethod! svp-transposition ((self sound) (param pathname) &optional out-name (fftsize 4096) (windowsize 4000))
  :icon 950  
  (svp-transp self param out-name windowsize fftsize))

(defmethod! svp-transposition ((self sound) (param integer) &optional out-name (fftsize 4096) (windowsize 4000))
  :icon 950  
  (svp-transp self param out-name windowsize fftsize))


;;;=================================================================================================
;;; NORMALIZE
;;;=================================================================================================

(defun svp-rescale (path newpath &optional (rescale 0.0) (resolution 16))
  (if (probe-file (om-external-app "supervp"))
    (let ((path-temp (make-pathname :device (pathname-device path) :directory (pathname-directory path)
                                    :name (pathname-name path) :type "tmp")))
          (print (format nil "Rescaling ~D to ~D dB and converting to ~D bits" (om-path2cmdpath path) rescale resolution))
          (om-cmd-line (print (format nil "~s -S~S -norm ~D -Osa~D ~S" 
                                      (om-path2cmdpath (om-external-app "supervp"))
                                      (om-path2cmdpath path) 
                                      (* -1 rescale) resolution
                                      (om-path2cmdpath newpath))) *svp-console*)
          ;(when (probe-file path-temp)
          ;  ;(delete-file path)
          ;  (rename-file path-temp path :if-exists :supersede))
          )
    (progn 
      (print "SVP no available for rescale...")
      (rename-file path newpath)
      ))
  newpath
  )




