(in-package :om)

;;; Midi basic events are 5-uple
;;; MEPitch, MEOnset, Medur, MEVel, MEChannel, &optional audio phrase number
;;; Raw events have no dur and keyon-keyoff is encoded by 0 vel

(defun new-midi-event (pitch onset dur vel chan &optional (phrase 0) (audio-onset 0))
  (list pitch onset dur vel chan phrase audio-onset))

(defmacro MEPitch (event)
  `(first ,event))

(defmacro MEOnset (event)
  `(second ,event))

(defmacro MEDur (event)
  `(third ,event))

(defmacro MEVel (event)
  `(fourth ,event))

(defmacro MEChannel (event)
  `(fifth ,event))

(defmacro MEPhrase (event)
  `(sixth ,event))

(defmacro MEAudioOnset (event)
  ` (seventh ,event))


(defmacro IsKeyOff? (event)
  `(zerop (MEVel ,event)))


;;; Midishare utils


#|
(defun send-note-at (MEnote refnum at)
   (let ((event (ms:MidiNewEv ms:typeNote)))	; ask for a new note event
     (unless (ccl:%null-ptr-p event)	; if the allocation was succesfull
       (ms:chan event (1- (MEChannel MEnote)))			; set the midi channel to 0 (means channel 1)
       (ms:port event 0)			; set the destination port to 0
       (ms:pitch event (MEPitch MEnote))		; set the pitch field
       (ms:vel event  (MEVel MEnote))		; set the velocity field
       (ms:dur  event (MEdur MEnote))		; set the duration field to 1 second
       (ms:MidiSendAt refnum event at))	; send the note at
     ))
|#


#|
(defun send-note-at (MEnote refnum at)
  (let ((event (ms:MidiNewEv ms:typeNote)) copy)	; ask for a new note event
    (unless (ccl:%null-ptr-p event)	; if the allocation was succesfull
      (ms:chan event (1- (MEChannel MEnote)))			; set the midi channel to 0 (means channel 1)
      (ms:port event 0)			; set the destination port to 0
      (ms:pitch event (MEPitch MEnote))		; set the pitch field
      (ms:vel event  (smoother (MEVel MEnote))	)	; set the velocity field
      (ms:dur  event (- (MEdur MEnote) 5))		; set the duration field to 1 second
      (setf copy (ms:MidiCopyEv event))
      (ms:MidiSendAt refnum event at)
      (ms:port copy 1)			        
      (ms:chan copy 1)
      (ms:MidiSendAt refnum copy at)           ; send the note at
      )	
    ))
|#

(defun smoother (vel)
   (if (<=  1 vel 30) vel (round (+ 30 (/ (- vel 30)  3)))))


#|
(defun smoother (vel)
  vel)
|#
