(in-package :om)

;------------------------------------------------------------------------------------------------


(defmacro timeOrigin () `(time-origin *supervizer*))


(defmacro a-oracle (n) `(aref (oracles *supervizer*) ,n))
(defmacro a-learner (n) `(aref (learners *supervizer*) ,n))
(defmacro a-improvizer (n) `(aref (improvisers *supervizer*) ,n))
(defmacro a-slicer (n) `(aref (slicers *supervizer*) ,n))
(defmacro a-unslicer (n) `(aref (unslicers *supervizer*) ,n))
(defmacro a-player (n) `(aref (players *supervizer*) ,n))
(defmacro a-listener (n) `(aref (listeners *supervizer*) ,n))

(defmacro osc-in-host () `(slot-value *supervizer* 'osc-in-host))
(defmacro osc-out-host () `(slot-value *supervizer* 'osc-out-host))
(defmacro osc-out-time-port () `(slot-value *supervizer* 'osc-out-time-port))
(defmacro osc-out-midi-port () `(slot-value *supervizer* 'osc-out-midi-port))


(defmacro send-osc-midi (bundle)
  `(send-osc-bundle *osc-midi-end-point*  ,bundle))

(defmacro send-osc-time (bundle)
  `(send-osc *osc-time-end-point* (first ,bundle)))

;; version 1.2 : send as a bundle, not as a message !
(defmacro send-Bosc-midi (bundle)
  `(send-osc-bundle *Bosc-midi-end-point*  ,bundle))

(defmacro dprint (format &rest args)
  `(when *OMaxDebug* (format *stdout* ,format ,. args)))

(defmacro dprint2 (format &rest args)
  `(when *OMaxDebug2* (format *stdout* ,format ,. args)))



;------------------------------------------------------------------------------------------------

(defclass MS-Listener (agent)
  ((name :initform "ms-listener" :accessor name :initarg :name :type string)
   (refnum :initform -1 :accessor refnum :initarg :refnum :type integer)
   (port :initform 0 :accessor port :initarg :port :type integer)
   (start-time :initform 0 :accessor start-time :initarg :start-time :type integer)))



#|
(defmethod clean-up ((self ms-listener))
  (ms:MidiFlushEvs (refnum self))
  self)

|#
(defmethod clean-up ((self ms-listener))
  self)

#|
(defmethod task ((listener ms-listener))
  (let ((event) copyev)
    (setf event (ms:MidiGetEv (refnum listener)))
    (cond
     ((%null-ptr-p event) 
      (process-allow-schedule))
     (T
      (case  (ms:type event)
        (#.ms:typeActiveSens)
        ((#.ms:typeKeyOn #.ms:typeKeyOff)
         (setf copyev (ms:MidiCopyEv event))
         (ms:port copyEv 1)
         (ms:MidiSendIM (refnum listener) copyEv)
         (talk listener (new-midi-event (ms:pitch event) (- (ms:date event) (start-time listener)) 
                                        (ms:dur event) (if (eq (ms:type event) #.ms:typeKeyOff) 0 (ms:vel event) )
                                        (1+ (ms:chan event)) 0 (- (ms:date event) (start-time listener))))))
      (ms:MidiFreeEv event)))))
|#


; this is for the purpose of june 4 ....
(defmethod task ((listener ms-listener))
  (process-allow-schedule))



#|
(defmethod initialize-instance  ((listener ms-listener) &rest initargs)
  (declare (ignore initargs))
  (call-next-method)
  (setf (refnum listener) (ms:midiopen (name listener)))
  (ms:MidiConnect (port listener) (refnum listener) -1)
  (ms:MidiConnect (refnum listener) (port listener)  -1)
  (ms:MidiFlushEvs (refnum listener))
  (setf (start-time listener) (TimeOrigin))
  listener)

(defun new-ms-listener (name ms-port)
  (make-instance 'MS-Listener :name name :port ms-port))

(defmethod kill ((self ms-listener))
  (call-next-method)
  (ms:MidiConnect (refnum self) (port self) 0)
  (ms:MidiConnect  (port self) (refnum self) 0)
  (ms:midiclose (refnum self)))
|#


(defmethod initialize-instance  ((listener ms-listener) &rest initargs)
  (declare (ignore initargs))
  (call-next-method)
  (setf (start-time listener) (TimeOrigin))
  listener)

(defun new-ms-listener (name ms-port)
  (make-instance 'MS-Listener :name name :port ms-port))

(defmethod kill ((self ms-listener))
  (call-next-method))


;------------------------------------------------------------------------------------------------


(defclass osc-midi-listener (agent)
  ((name :initform "ms-listener" :accessor name :initarg :name :type string)))

(defvar *osc-midi-listener*)

(defun new-osc-midi-listener (name)
  (make-instance 'osc-midi-listener :name name))

(defmethod task ((listener osc-midi-listener))
  (sleep 1000)
  (process-allow-schedule))

(setf *count* 0)


(defun handle-osc-midi-evt (midi-data) ; (pitch vel nchan nphrase date)
  ;(format t "Pitch ~a vel ~a np ~a date ~a ~%~%" 
  ;        (first midi-data) (second midi-data) (fourth midi-data) (fifth midi-data))
  ;(format *stdout* "~S..."  (round (fifth midi-data) 1000))
  (if (boundp '*osc-midi-listener*)
      (osc-midi-arrived *osc-midi-listener* midi-data)
    (format *stdout* "< ")))


(defmethod osc-midi-arrived  ((self osc-midi-listener) midi-data)
  ; osc internal midi event have (pitch date dur vel chan  nphrase audio-date)
  (talk self (new-midi-event (first midi-data) (fifth midi-data) 0 (second midi-data)
                             (third midi-data) (fourth midi-data) (fifth midi-data))))

(defun handle-osc-midi-phrase (midi-data)
  ;; not handled yet
  (format *stdout* "PHRASE n ~a ~a date ~a ~%~%" (first midi-data) (if (= 0 (second midi-data)) "FIN" "DEBUT")  (third midi-data)))

;;; ============================================================================================================

(defclass MS-Player (agent)
  ((name :initform "ms-player" :accessor name :initarg :name :type string)
   (refnum :initform -1 :accessor refnum :initarg :refnum :type integer)
   (port :initform 0 :accessor port :initarg :port :type integer)
   (segdur :initform 600 :accessor segdur :initarg :segdur :type integer)
   (avance :initform 300 :accessor avance :initarg :avance :type integer)
   (start-play :initform nil :accessor start-play :initarg :start-play :type boolean)
   (next-time :initform 0 :accessor next-time :initarg :next-time :type integer)
   (cursegdur :accessor cursegdur :initform 0 :initarg :cursegdur)
   (segment :accessor segment :initform () :initarg :segment)
   (seg-onset :accessor seg-onset :initform 0 :initarg :seg-onset) ))


#|
(defmethod initialize-instance ((self ms-player) &rest initargs)
  (call-next-method)
  (setf (refnum self) (ms:midiopen (name self)))
  ;(ms:MidiConnect ms-port (refnum self) -1)
  (ms:MidiConnect (refnum self) (port self)  -1)
  (ms:MidiConnect (refnum self) (1+ (port self))  -1)
  (ms:MidiFlushEvs (refnum self))
  self)
|#

(defmethod initialize-instance ((self ms-player) &rest initargs)
  (call-next-method)
  self)


(defmethod task ((self ms-player))
  (let ((event nil))
    ;(process-wait "wait start" #'start-play self)
    ;(process-wait "wait nextseg" #'ready-for-next-seg self)
;(print (look-time-ahead self))
    (if (not (ready-for-next-seg self))
      (process-allow-schedule)
      (if (null (setf event (listens self)))
        (process-allow-schedule)
        (progn
          (incf (cursegdur self) (duration event))
          (push event (segment self))
          (when (> (cursegdur self) (segdur self)) ; time to play
            (setf (seg-onset self)
                  (max (next-time self) (+ (get-internal-real-time) 100)))
            (loop for evt in (nreverse (segment self))
                  with delay = 0
                  do (loop for note in (MidiSet evt)
                           do (if (pause-event-p  event)
                                T ; do nothing
                                (send-note-at note (refnum self) (+ (seg-onset self) delay))))
                  (incf delay (duration evt)  ))
            (setf (next-time self) (+ (seg-onset  self) (cursegdur self))
                  (cursegdur self) 0
                  (segment self) nil)
            ))))))

(defun new-ms-player (name ms-port )
  (make-instance 'MS-player :name name :port ms-port :cursegdur 0 :segment 0 :seg-onset 0))

(defmethod clean-up ((self ms-player))
  (empty-queue (agent-listening-queue self))
  (reinitialize-instance self  :cursegdur 0 :segment 0 :seg-onset 0 :next-time 0)
  self)

#|
(defmethod kill ((self ms-player))
  (call-next-method)
  (ms:MidiConnect (refnum self) (port self) 0)
  ;(ms:MidiConnect  (port self) (refnum self) 0)
  (ms:midiclose (refnum self)))

(defmethod silent ((self ms-player))
  (ms:MidiConnect (refnum self) (port self) 0))

(defmethod audible ((self ms-player))
  (ms:MidiConnect (refnum self) (port self) -1))

|#


(defmethod kill ((self ms-player))
  (call-next-method))
 
(defmethod silent ((self ms-player))
 )

(defmethod audible ((self ms-player))
  )



(defmethod ready-for-next-seg ((self MS-Player))
  (or (zerop (Next-time self))
      (<= (- (Next-time self) (get-internal-real-time))
          (avance self))))

(defmethod look-time-ahead ((self ms-player))
  (let ((queue (agent-listening-queue self)))
    (if (null queue) 0 (queue-duration queue))
    ))


;;; ============================================================================================================
;;; OSC (augmented) midi  player


(defclass osc-midi-player (agent)
  ((name :initform "ms-player" :accessor name :initarg :name :type string)
   (segdur :initform 600 :accessor segdur :initarg :segdur :type integer)
   (avance :initform 300 :accessor avance :initarg :avance :type integer)
   (start-play :initform nil :accessor start-play :initarg :start-play :type boolean)
   (next-time :initform 0 :accessor next-time :initarg :next-time :type integer)
   (cursegdur :accessor cursegdur :initform 0 :initarg :cursegdur)
   (segment :accessor segment :initform () :initarg :segment)
   (seg-onset :accessor seg-onset :initform 0 :initarg :seg-onset) ))

;; this one if for bloc send
(defclass Bosc-midi-player (osc-midi-player)
  ((flush-p :accessor flush-p :initform nil)
   (durmax :accessor durmax :initform 3000))
  )


(defmethod initialize-instance ((self osc-midi-player) &rest initargs)
  (call-next-method)
  self)

#|
(defmethod task ((self osc-midi-player))
  (let ((event nil))
    (if (not (ready-for-next-seg self))
      (process-allow-schedule)
      (if (null (setf event (listens self)))
        (process-allow-schedule)
        (progn
          (incf (cursegdur self) (duration event))
          (push event (segment self))
          (when (>= (cursegdur self) (segdur self)) 
(print " play")
            ; time to play
            (setf (seg-onset self)
                  (max (next-time self) (+ (ms:MidiGetTime) 50)))
            (osc-send (list (list '/ma 'start)) (osc-out-host) (osc-out-midi-port))
            (osc-send  (loop for evt in (unslice (nreverse (segment self)) (seg-onset self))
                             collect (osc-note-at evt))
                       (osc-out-host) (osc-out-midi-port))
            (osc-send (list (list '/ma 'stop)) (osc-out-host) (osc-out-midi-port))
            (setf (next-time self) (+ (seg-onset  self) (cursegdur self))
                  (cursegdur self) 0
                  (segment self) nil)
            ))))))
|#


(defmethod task ((self osc-midi-player))
  (let ((event nil) delay bundle)
    (if (not (ready-for-next-seg self))
      (process-allow-schedule)
      (if (null (setf event (listens self)))
        (process-allow-schedule)
        (progn
          (incf (cursegdur self) (duration event))
          (push event (segment self))
          (when (>= (cursegdur self) (segdur self)) 
            ;(print "time to play") 
            (setf delay (get-internal-real-time))
            ; time to play
            (let ((av (- (next-time self)  (get-internal-real-time)))) (when (< av 100) (dprint  "~%short avance : ~a~%" av)))
            (setf (seg-onset self)
                  (max (next-time self) (+ (get-internal-real-time) 100)))
            (send-osc-midi (list (list '/ma 'start)))
            ;(dprint "SENDING BUNDLE ~%")
            (setf bundle (loop for evt in (unslice (nreverse (segment self)) (seg-onset self))
                               collect (osc-note-at evt)))
            (send-osc-midi  bundle)
            (send-osc-midi (list (list '/ma 'stop)))
            (setf (next-time self) (+ (seg-onset  self) (cursegdur self))
                  (cursegdur self) 0
                  (segment self) nil)
            ;(print (list "time to play :" (- (get-internal-real-time) delay)))
            ;(print "end  play")
            ))))))

#|
(defmethod task ((self osc-midi-player))
  (let ((event nil))
    (if (>= (cursegdur self) (segdur self))
      (if  (ready-for-next-seg self)
        (progn 
          ; time to play
(print "time to play")
          (setf (seg-onset self)
                (max (next-time self) (+ (get-internal-real-time) 50)))
          (osc-send (list (list '/ma 'start)) (osc-out-host) (osc-out-midi-port))
          (osc-send  (loop for evt in (unslice (nreverse (segment self)) (seg-onset self))
                           collect (osc-note-at evt))
                     (osc-out-host) (osc-out-midi-port))
          (osc-send (list (list '/ma 'stop)) (osc-out-host) (osc-out-midi-port))
          (setf (next-time self) (+ (seg-onset  self) (cursegdur self))
                (cursegdur self) 0
                (segment self) nil) )
        (process-allow-schedule))
      (if (null (setf event (listens self)))
        (process-allow-schedule)
        (progn
          (incf (cursegdur self) (duration event))
          (push event (segment self)))))))
|#

(defmethod unslice ((self list) start-time)
  (loop for evt in self
        with delay = 0
        ;do (analyse-slice evt)
        if (not (pause-event-p  evt))
        append (loop for note in (MidiSet evt)
                     do ;(setf (MEChannel note) 1)
                     (setf (MEOnset note) (+ start-time delay ))
                     (unless (MEPhrase note) ;; old oracle without audio dates
                       (let ((last (last note)))
                         (setf (cdr last) (list 0 0))))
                     
                     collect (let ((noteoff (copy-list note)))
                               (setf (MEVel noteoff) 0)
                               (setf (MEOnset noteoff) (+ (MeOnset note) (MEDur note)  -5))
                               ;(setf (MEOnset noteoff) (+ (MeOnset note) (MEDur note) ))
                               (setf (MEAUdioOnset noteoff) (+  (MeAudioOnset note) (MEDur note)))
                               noteoff))
        into note-offs
        do (incf delay (duration evt)) 
        finally return (sort (append (loop for event in self append (Midiset event))
                                     note-offs)
                             #'(lambda (x y) (or (< (MeOnset x) (MEOnset y))
                                                 (and (= (MeOnset x) (MEOnset y))
                                                      (= (MePitch x) (MePitch y))
                                                      (IsKeyOff? x))))
                                                 
                              )))

(defun new-osc-midi-player (name)
  (make-instance 'osc-midi-player :name name  :cursegdur 0 :segment 0 :seg-onset 0))

(defmethod clean-up ((self osc-midi-player))
  (empty-queue (agent-listening-queue self))
  (reinitialize-instance self  :cursegdur 0 :segment 0 :seg-onset 0 :next-time 0)
  self)

(defmethod kill ((self osc-midi-player))
  (call-next-method)
  )

(defmethod ready-for-next-seg ((self osc-midi-player))
  (or (zerop (Next-time self))
      (<= (- (Next-time self) (get-internal-real-time))
          (avance self))))


(defmethod look-time-ahead ((self osc-midi-player))
  (let ((queue (agent-listening-queue self)))
    (if (null queue) 0 (queue-duration queue))
    ))

(defmethod osc-note-at (note)
  (let ((bundle (list '/ma (MEOnset note) (MEPitch note) (MEvel note) 
                      (MEChannel note) (MEPhrase note) (MEAudioOnset note) (MEDur note))))
    (dprint2 "~a~%" bundle)
    bundle))



;--------------------------------------------------------------------------------
; BOSC-midi-player bloc mode



(defmethod initialize-instance ((self Bosc-midi-player) &rest initargs)
  (call-next-method)
  self)

(defmethod clean-up ((self Bosc-midi-player))
  (empty-queue (agent-listening-queue self))
  (reinitialize-instance self  :cursegdur 0 :segment 0 :seg-onset 0 :next-time 0 :segdur 1000)
  self)

(defun new-Bosc-midi-player (name)
  (make-instance 'Bosc-midi-player :name name  :cursegdur 0 :segment 0 :seg-onset 0 :segdur 1000))


#|
(defmethod task ((self Bosc-midi-player))
  (let ((event nil) delay)
    (if (null (setf event (listens self)))
      (process-allow-schedule)
      (progn
        (incf (cursegdur self) (duration event))
        (push event (segment self))
        (when (or (flush-p self) (>= (cursegdur self) (segdur self)))
(print (cursegdur self))
          (setf delay (get-internal-real-time))
          ; time to play
          (setf (seg-onset self) (next-time self))
          ;(send-Bosc-midi (list (list '/ma 'start)))
(print "-------")
          (send-Bosc-midi    (loop for evt in (unslice (nreverse (segment self)) (seg-onset self))
                                   collect (osc-note-at evt)))
          ;(send-Bosc-midi (list (list '/ma 'stop)))
          (setf (next-time self) (+ (seg-onset  self) (cursegdur self))
                (cursegdur self) 0
                (segment self) nil)
          (when (flush-p self) (setf (flush-p self) 'done))
          )))))
|#



(defmethod task ((self Bosc-midi-player))
  (if (>=  (cursegdur self)  (durmax self))
    (progn ;(print (list (cursegdur self)(durmax self)))
           (terminate-bloc-generation *supervizer*) 
           (process-allow-schedule) )  
    (let ((event (listens self)))
      (if (null event)
        (process-allow-schedule)
        (progn
          (incf (cursegdur self) (duration event))
;(print (cursegdur self))
          (unless (pause-event-p  event)
            (send-Bosc-midi
             (loop for evt in (unslice-1 event (next-time self) (durmax self))
                   collect (osc-note-at evt))))
          (setf (next-time self) (+ (next-time self) (duration event)) ))))))

;;; added timemax for bloc play : resonances shouldn't be longer then bloc dur
;;; ga 8/10/2006

(defmethod unslice-1 ((self crossevent) start-time timemax)
  ;(analyse-slice self)
  (loop for note in (MidiSet self)
        do 
        (setf (MEOnset note) start-time)
        (let ((noteoff-date
               (if (<= (+ (MeOnset note) (MEDur note)) timemax)
                 (+ (MeOnset note) (MEDur note))
                 (if (>= (- timemax (MEOnset note)) 500)
                   timemax
                   (+ (MEOnset note) 500)))))
          (setf (MEdur note) (- noteoff-date (MEonset note))))
        collect
        (progn
          (unless (MEPhrase note) ;; old oracle without audio dates
            (let ((last (last note)))
              (setf (cdr last) (list 0 0))))
          note) 
        into notes
        collect 
        (let ((noteoff (copy-list note)))
          (setf (MEVel noteoff) 0
                (MEOnset noteoff) (+ (MeOnset note) (MEdur note))
                (MEAUdioOnset noteoff) (+  (MeAudioOnset note) (MEDur note)))
          noteoff)
        into noteoffs
        finally return (nconc notes noteoffs)))

#|

3008 (stream) 3009 (Bloc mode)

/ma date-mid pitch vel chan nphrase date-buffer-audio
/ma start
/ma stop

|#


;;; ============================================================================================================
;;; A class and process that slices a stream of MidiEvents (5uples) into a list of crossevents
;;; typical call : (new-polySlicer  "Slicer-0" feeder)
;;; where feeder is an object with a slot 'fifo-out, from which the MidiEvents will be popped.


(defclass PolySlicer (agent)
  ((name :accessor name :initarg :name)
   (magnet1 :initform 0 :accessor magnet1 :initarg :magnet1 :type integer)
   (magnet2 :initform 0 :accessor magnet2 :initarg :magnet2 :type integer) 
   (delta-t :initform 50 :accessor delta-t :initarg :delta-t :type integer)
   (running-notes :initform nil :accessor running-notes :initarg :running-notes :type list)
   (complete-notes :initform nil :accessor complete-notes :initarg :complete-notes :type list)
   (mindur :initform 50 :accessor mindur :initarg :mindur :type integer)
   (prev-phrase-nb :initform -1 :accessor prev-phrase-nb :initarg :prev-phrase-nb :type integer)
   (last-offset :initform 0 :accessor last-offset :initarg :last-offset :type integer)))


(defun new-PolySlicer (name)
  (make-instance 'PolySlicer :name name))

#|
(defmethod task ((self polyslicer))
  (let ((event) )
    (setf event (listens self))
    (if (null event) 
      (process-allow-schedule)
      (progn
        ;(print (list "eater:" event))
        (eat-note-evt self event)))))
|#


(defmethod task ((self polyslicer))
  (loop for event =  (listens self)
        while event
        do
        ;(print (list "eater:" event))
        (eat-note-evt self event)))


(defmethod eat-note-evt ((self polySlicer) (MEnote list))
  (if (IsKeyOff? MeNote) (eat-note-off self MENote) (eat-note-on self MENote)))

(defmethod eat-note-on ((self polySlicer) (MEnoteon list))
  (cond
   ((in-time-window1 self  MENoteOn)
    (setf (MEOnset MeNoteOn) (magnet1 self))
    (add-to-run-notes self MeNoteOn))
   ((start-phrase-p self MEnoteon)
    (setf (running-notes self) nil (complete-notes self) nil
          (magnet1 self) (MEOnset MeNoteOn)
          (magnet2 self) (magnet1 self))
    (add-to-run-notes self MeNoteOn))
   (t (setf (magnet2 self) (MEOnset MeNoteOn))
      (check-note-off self)
      (when (running-notes self) ;(format t "~d " (length (running-notes self)))
        (output-slice1 self))
      (reset-note-off self)       
      (setf (magnet1 self) (magnet2 self))
      (remove-old-notes self)
      (add-to-run-notes self MENoteOn)))
  self)


(defmethod start-phrase-p ((self polyslicer) (MeNoteon list))
  (and 
   (every #'(lambda (note) (member note (complete-notes self) :key #'first :test #'eq)) (running-notes self))
   (> (- (MEOnset MeNoteon) (last-offset self)) 2000)))


#|
(defmethod eat-note-off ((self polySlicer) (MEnoteoff list))
   (cond
    ((in-time-window1 self  MENoteOff)
     (setf (MEOnset MeNoteOff) (magnet1 self))
     (remove-from-run-notes self MeNoteOff))
   (t (setf (magnet2 self) (MEOnset MeNoteOff))
      (when (running-notes self) (output-slice1 self))
      (setf (magnet1 self) (magnet2 self))
      (remove-from-run-notes self MENoteOff)))
  self)
|#

(defmethod remove-old-notes ((self polyslicer))
  (setf (running-notes self)
        (loop for note in (running-notes self)
              if (> (- (magnet2 self) (MEOnset note)) 5000)
              collect note into old
              finally return (if (null old)
                               (running-notes self)
                               (loop for note in (running-notes self)
                                     if (not (memq note old)) collect note)))))
                      


(defmethod eat-note-off ((self polySlicer) (MEnoteoff list))
  (let ((running (find-running-note self MENoteoff)))
    (setf (last-offset self) (MEOnset MENoteOff))
    (when running (push (cons running MENoteOff) (complete-notes self))))
  self)


(defmethod in-time-window1 ((self polySlicer) (MEnoteon list))
  (<= (magnet1 self) (MEOnset MENoteOn) (+ (magnet1 self) (delta-t self))))
 

(defmethod add-to-run-notes ((self polySlicer) (MEnoteon list))
  (when (member MeNoteOn (running-notes self) :key #'first) (format *stdout*  "note ~d twice in running-notes list~%" MENoteOn))
  (push  MeNoteOn (running-notes self)))

(defmethod remove-from-run-notes ((self polySlicer) (MEnoteoff list))
  (setf (running-notes self)
        (delete-if #'(lambda (note) (and (= (MEPitch note)  (MEPitch MENoteOff))
                                         (= (MEChannel note)  (MEChannel MENoteOff))))
                   (running-notes self)))
  self)


(defmethod find-running-note ((self polySlicer) (MEnoteoff list))
  (find-if #'(lambda (note) (and (< (MEOnset note) (MEOnset MENoteOff))
                                 (= (MEPitch note)  (MEPitch MENoteOff))
                                 (= (MEChannel note)  (MEChannel MENoteOff))))
           (running-notes self)))

#|
(defmethod check-note-off ((self polySlicer))
  (loop for (noteon . noteoff) in (complete-notes self)
        if (and (not (>= (MEOnset noteon) (magnet1 self)))
                (< (MEOnset noteoff) 
                   (+ (magnet1 self)
                      (* 0.33 (- (magnet2 self) (magnet1 self))))))
        do (setf (running-notes self)
                 (delete noteon (running-notes self) :test #'eq))))
|#

(defmethod check-note-off ((self polySlicer))
  (loop for (noteon . noteoff) in (complete-notes self)
        if (and (not (>= (MEOnset noteon) (magnet1 self)))
                (< (MEOnset noteoff)  (Magnet2 self)))
        do (setf (running-notes self)
                 (delete noteon (running-notes self) :test #'eq))))

(defmethod reset-note-off ((self polySlicer))
  (loop for (noteon . noteoff) in (complete-notes self)
        do (setf (running-notes self)
                 (delete noteon (running-notes self) :test #'eq))
        finally do (setf (complete-notes self) nil)))


(defmethod output-slice1 ((self polySlicer))
  (let ((cross (mki 'crossevent)) (phrasenb (or (MEPhrase (first (running-notes self))) 0))
        (dur (- (magnet2 self) (magnet1 self))))
    (setf (duration cross) dur
          (MidiSet cross)
          (loop for note in (running-notes self)
                collect 
                (new-midi-event  
                 (if (< (MEOnset note) (magnet1 self))
                   (- (MEPitch note))
                   (MEPitch note))
                 0 
                 dur
                 (MEVel note) 
                 (MEChannel note)
                 (or (MEPhrase note) 0)
                 (or (MEAUdioOnset note) 0)))
          (startPhrase cross)
          (/= phrasenb (prev-phrase-nb self))
          (phraseNumber cross)
          phrasenb)
    (setf (prev-phrase-nb self) phrasenb)
    ;(format t "~%Slice : ~D ~D~%" (duration cross) (MidiSet cross))
    ;(analyse-slice cross)
    (when (>= dur (mindur self)) (talk self cross))
    cross))

(defun analyse-slice (slice)
  (let ((ms (MidiSet slice)))
    (when (> (length ms) 1)
      (dprint  "~%~%ACCORD : ~d ~d ~%~%" (length ms) (-  (MEAudioOnset (first ms))(MEAudioOnset (second ms)))))))
#|

      (loop for m in ms do
            (format T "Midi: ~d Audio ~d~%" (MEOnset m) (MEAudioOnset m))))))

|#
(defmethod clean-up ((self polyslicer))
  (empty-queue (agent-listening-queue self))
  (reinitialize-instance self  :running-notes nil :complete-notes nil :last-offset 0)
  self)


;;; ============================================================================================================

(defclass UnSlicer (agent)
  ((name :initform "ms-player" :accessor name :initarg :name :type string)
   (waiting-list :accessor waiting-list :initarg :waiting-list :initform ())
   (new-waiting-list :accessor new-waiting-list :initarg :new-waiting-list :initform ())
   ))


;------ INTERFACE -------------------------------

(defun new-UnSlicer (name )
  (make-instance 'UnSlicer :name name))


(defmethod clean-up ((self unslicer))
  (empty-queue (agent-listening-queue self))
  (reinitialize-instance self :waiting-list nil :new-waiting-list nil))


;------ INTERFACE -------------------------------



(defmethod task ((self unslicer))
  (let ((event (listens self)))
    (if (null event) 
      (process-allow-schedule) 
      (progn
        (setf event (clone-object event))
        (loop for ME in (midiset event)
              with new-midiset = nil
              initially (setf (new-waiting-list self) nil)
              do (cond ((plusp (MEPitch ME))
                        (push ME (new-waiting-list self))
                        (push ME new-midiset))
                       ((minusp (MEPitch ME))
                        (let ((waiting (find ME (waiting-list self)
                                             :test #'(lambda (m1 m2) 
                                                       (and (= (abs (MEPitch m1)) (abs (MEPitch m2))) 
                                                            (= (MEChannel m1) (MEChannel m2)))))))
                          ;(print (list ME waiting))
                          (cond 
                           (waiting
                            (setf (MEDur waiting) (+ (MEDur waiting) (MEDur ME)))
                            (push waiting (new-waiting-list self)))
                           (t (setf (MEPitch ME) (abs (MEPitch ME)))
                              (push ME (new-waiting-list self))
                              (push ME new-midiset)))))) ; might be empty if only running notes
              finally (progn (setf (Midiset event) new-midiSet) 
 ;(analyse-slice event)
                             (talk self event)
                             ;(pushr event *crosslist*)
                             ))))
    (setf (waiting-list self) (new-waiting-list self))))


;;; ============================================================================================================

(defclass* coracle (improvizer)
  ((lock  :initform (make-process-lock :name 'impro-lock)  :accessor lock :initarg :lock)))

(defclass improviser (agent)
  ( (oracle :accessor oracle :initarg :oracle)
    (continuity :initform 0 :initarg :continuity :accessor continuity)
    (max-continuity :initform 8 :initarg :max-continuity :accessor max-continuity)
    (start-play :initform nil :accessor start-play :initarg :start-play )
    (stop :initform nil :accessor stop :initarg :stop)
    (gindex :initform 0  :accessor gindex :initarg :gindex :type integer)
    (durseg :initform 0  :accessor durseg :initarg :durseg :type integer)
    (blocmode :initform nil  :accessor blocmode :initarg :blocmode)
    (startOnPhrase :initform nil  :accessor startOnPhrase :initarg :startOnPhrase)
    (blocStartPhrase :initform -1  :accessor blocStartPhrase :initarg :blocStartPhrase)
    )
  )

(defclass learner (agent)
  ((oracle :accessor oracle :initarg :oracle)
   (last-learned :initform nil  :accessor last-learned :initarg :last-learned)))


(defun New-Improvizer (oracle)
  (make-instance 'improviser :oracle oracle :state :asleep ))

(defun New-Learner (oracle)
  (make-instance 'learner :oracle oracle :state :asleep))

(defun NewOracle () (make-instance 'coracle :comparateur 'CompareEvents :lrsMode t ))

#|
(defmethod clone-oracle ((self learner))
  (clone (oracle self)))
|#

(defmethod clean-up ((self improviser))
  (empty-queue (agent-listening-queue self))
  (reinitialize-instance self :gindex 0 :durseg 0 :stop nil :continuity 0))

(defmethod clean-up ((self learner))
  (empty-queue (agent-listening-queue self))
  (reinitialize-instance self :last-learned nil))

(defmethod task  ((self learner))
  (let ((event))
    (setf event  (listens self))
    (if (null event) 
      (process-allow-schedule) 
      (progn
        (setf *current-improvizer* (oracle self)) ;; this doesn't work concurrently !
        (with-lock-grabbed ((lock (oracle self)))
          (setf (last-learned self) event)
          (ajouter-objet (oracle self) event))))))
 
(defmethod too-fast ((self improviser)) (setf (stop self) T))
(defmethod too-slow ((self improviser)) (setf (stop self) nil))

(defmethod task  ((self improviser))
  (if (stop self) 
    (process-allow-schedule)
    (let ((previndex (gindex self)) (index2) (links) (mode 'continuity) (tocollect nil) (harmgrid nil) (label nil))
      (when (null (veclrs (oracle self))) (setf (bestSuffixMode (oracle self)) nil)
            (format *stdout* "Old Oracle, setting bestSuffix mode to nil~%"))
      
      ;(process-wait "silence" #'start-play self)
      ;(process-wait "empty" #'(lambda (s) (> (maxetat s) 0))  (oracle self))
      
      (if (<= (maxetat (oracle self)) 0)
        (process-allow-schedule)
        (progn
          (with-lock-grabbed ((lock (oracle self)))
            (setf previndex (gindex  self))
            (cond
             ((zerop (gindex self)) 
              (if (or (not  (blocmode self)) (< (blocStartPhrase self) 0))
                (setf index2 (find-beat-label-match1 (oracle self) label (StartOnPhrase self)))
                (setf index2 (find-start-phrase (oracle self) (blocStartPhrase self))))
              (format *stdout* "Starting point : ~a in phrase ~a~% " index2 (PhraseNumber (otext (oracle self) index2)))
              (setf tocollect (otext (oracle self) index2)
                    (gindex self) index2))
             (t (if (check-continuity  self)
                  (setf mode 'continuity)  
                  (if (available-suppleance (oracle self) (gindex self)) 
                    (setf mode 'suppleance)
                    (setf mode 'continuity)))
                (when (eq mode 'continuity)
                  ;(when (= index (maxetat (oracle self))) (setf index 0)
                  ;      (format t "zero~%" ))
                  (setf index2 (next-factor-state (oracle self) (gindex self) 0))
                  ;(setf links (flink (oracle self) (gindex self))
                  ;      index2 (choose-factor-link (oracle self) links label))
                  (if index2 
                    (progn
                      ;(dprint  "c : ~a " index2)
                      (setf (gindex self) index2
                            tocollect  (otext (oracle self) index2)))
                    (if (available-suppleance (oracle self) (gindex self)) 
                      (setf mode 'suppleance)
                      (setf mode 'nothing))))
                (when (eq mode 'suppleance)
                  (setf index2 (continuations-by-suppleance (oracle self) (gindex self) label))
                  (if (and index2 (/= index2  previndex) (/= index2 (+ previndex 1)))
                    (progn 
                      ;(dprint  "---->s : ~a " index2)
                      (setf (gindex self) index2
                            tocollect (otext (oracle self) index2) 
                            (continuity  self) 0))
                    (setf mode 'nothing)))
                (when (eq mode 'nothing)
                  (setf index2 (find-beat-label-match1 (oracle self) label nil))
                  (if index2 
                    (progn (setf tocollect (otext (oracle self) index2)
                                 (gindex self) index2) (continuity self) 0)
                    (progn               
                      (print 'empty)
                      (setf (gindex self) 0
                            tocollect nil ;(null-beat)
                            )))))))
          (when tocollect  
            #|
(when (and (< (gindex self) (maxetat (oracle self)) )
                       (startPhrase (otext (oracle self) (1+ (gindex self)))))
              (print "pause 1000")
              (setf (continuity self) 0)
              (talk self (mk-pause-event 1000)))
|#
            (talk self tocollect)
            (incf (durseg self) (duration tocollect)))
          ;(unless (blocmode self) (process-allow-schedule))
          (when (> (durseg self) 5000) (setf (durseg self) 0) (process-allow-schedule) )
          )))))

(defun mk-pause-event (dur)
  (make-instance 'crossevent :duration dur))

(defmethod pause-event-p ((event crossevent)) (null (MidiSet event)))


(defmethod find-beat-label-match1 ((self improvizer) (label list) &optional StartOnPhrase)    
  (let* ((left  (round (* (/ (first (start-region self)) 100) (1- (maxetat self)))))
         (width (round (* (/ (- (second (start-region self)) (first (start-region self))) 100)  (1- (maxetat self)))))
         (start (max 1 (min (+ (random width) left) (1- (maxetat self))))))
    (catch
      'tag-fblm
      (when (= (maxetat self) 1) (throw 'tag-fblm 1))
      (when StartOnPhrase
        (loop for i from (max 1 (min left (1- (MaxEtat self)))) to (1- (MaxEtat self))
              do (when (startPhrase (otext self i)) (throw 'tag-fblm i))))
      (loop for i from start to (1- (maxetat self))
            ;do (print (list i ( eligible-beat? (otext self i) label) label (RefHarmLabel (otext self i))))
            if ( eligible-beat? (otext self i) label)  
            do (throw 'tag-fblm i))
      (setf  left 1 width (1- (maxetat self))
             start (min (+ (random width) left) (1- (maxetat self))) )
      (loop for i from start to (1- (maxetat self))
            if ( eligible-beat? (otext self i) label)  
            do (throw 'tag-fblm i))
      1)))


(setf *last-suppleance* nil)
(defmethod continuations-by-suppleance ((self improvizer) (index integer) (label list))
   (let (back-cont forw-cont (best-suffix-state -1) (max-suffix 0))
     (when (bwSuffix self)
       (setf back-cont
             (loop for supp = (suppleance self index) then (suppleance self supp)
                   with previous = index
                   with suffix-length = 0
                   while (and supp (> supp 0))
                   do 
                   ;(print (lrs self previous)) 
                   (when (bestSuffixMode self) (setf suffix-length (lrs self previous)  previous supp))
                   append (loop for cont in (flink self  supp)
                                if (and (not (memq  supp *last-suppleance*))  (eligible-beat? (otext self cont) label))
                                do (when (> suffix-length max-suffix) (setf max-suffix suffix-length))
                                and collect (list cont suffix-length supp )))))
     (when (fwSuffix self)
       (setf forw-cont 
             (loop for supp = (suppleance-> self index) then (suppleance-> self supp)
                   with suffix-length = 0
                   while supp
                   do
                   (when (bestSuffixMode self) (setf suffix-length (lrs self supp)))
                   ;(print (list '-> (lrs self supp)))
                   ;(when (> suffix-length max-suffix) (setf best-suffix-state supp max-suffix suffix-length))
                   append (loop for cont in (flink self supp)
                                if (and (not (memq  supp *last-suppleance*) ) (eligible-beat? (otext self cont) label))
                                do (when (> suffix-length max-suffix) (setf max-suffix suffix-length))
                                and collect (list cont suffix-length supp )))))
     (let ((cont (if (bestSuffixMode self) 
                   (nth-random (loop for cont in  (append back-cont forw-cont)
                                     if (= (second cont) max-suffix) collect  cont))
                   (nth-random (append back-cont forw-cont)))))
       ; dont use one of the 3 previously used suppleances
       (dprint "Jumping from ~d to ~d then ~d with context length ~d~%"  index (third cont) (first cont) (second cont))
       ;(unless cont (setf cont (list (max (1- (Maxetat self)) (1+ index)) 0 *last-suppleance*)))
       (unless cont (setf cont (list 1 0 *last-suppleance*)))
       (push  (third cont) *last-suppleance* )
       (when (> (length *last-suppleance*) 3) (setf  (cdddr *last-suppleance*) nil))
       (first cont))))



(defmethod find-start-phrase ((self improvizer) StartPhrase)
  (catch
    'tag-fblm
    (loop for i from 1 to (1- (maxetat self))
          if (and (startPhrase (otext self i))
                  (= (phraseNumber (otext self i)) StartPhrase))
          do (progn (format *stdout* "Bloc by phrase ~d~%" StartPhrase) (throw 'tag-fblm i)))
    1))

(defmethod check-continuity ((self improviser))
  (if (<= (continuity self) (max-continuity self))
    (true (incf (continuity self)))
    (false (setf (continuity self) 0))))

;; modulo 12, intersection, dures similaires
(defmethod CompareEvents ((Event1 CrossEvent) (Event2 CrossEvent))
  
  (let* ((m1 (delete-duplicates  
              (mapcar #'(lambda (x) (mod (abs (MEPitch x)) 12)) (MidiSet Event1))))
         (m2 (delete-duplicates  
              (mapcar #'(lambda (x) (mod (abs (MEPitch x)) 12)) (MidiSet Event2))))
         (l1 (length m1))
         (l2 (length m2))
         (l3 (length (intersection m1 m2))))
    (and
     (<= (abs (log (/ (duration event1) (duration event2)) 0.5)))
     (or 
      (= l1 l2 l3)
      (>= l3 (/ (+ l1 l2) 3))))) )

#|

;; modulo 12, identit, dures similaires

(defmethod CompareEvents ((Event1 CrossEvent) (Event2 CrossEvent))
  
  (let* ((m1 (delete-duplicates  
              (mapcar #'(lambda (x) (mod (abs (MEPitch x)) 12)) (MidiSet Event1))))
         (m2 (delete-duplicates  
              (mapcar #'(lambda (x) (mod (abs (MEPitch x)) 12)) (MidiSet Event2))))
         (l1 (length m1))
         (l2 (length m2))
         (l3 (length (intersection m1 m2))))
    (and (= l1 l2 l3) 
         (<= (abs (log (/ (duration event1) (duration event2)) 0.5))))))

;; pas modulo 12, galit, dures similaires

(defmethod CompareEvents ((Event1 CrossEvent) (Event2 CrossEvent))
  (let ((p1 (MEPitch (first (Midiset event1))))
        (p2 (MEPitch (first (Midiset event2))))
        (d1 (duration event1))
        (d2 (duration event2)))
    (if (or (null p1) (null p2))
      (and (null p1) (null p2))
      (and (= p1 p2) 
           (<= (abs (log (/ d1 d2) 0.5)))))))
  
;; dures uniquement
(defmethod CompareEvents ((Event1 CrossEvent) (Event2 CrossEvent))
  (let ((p1 (MEPitch (first (Midiset event1))))
        (p2 (MEPitch (first (Midiset event2))))
        (d1 (duration event1))
        (d2 (duration event2)))
    (if (or (null p1) (null p2))
      (and (null p1) (null p2))
      (<= (abs (log (/ d1 d2) 0.5))))))

|#  

(defmethod choose-factor-link ((self improvizer) (indexes list) (label list) )
  (if (null label)
    (nth-random indexes)
     (let ((succeed (loop for index in indexes 
                          if (eligible-beat? (otext self index) label)  collect index)))
       ;(setf succeed (sort succeed #'(lambda (i j) (< (length (midiset (otext self i))) (length (midiset (otext self j)))))))
       ;(first succeed) )))

       (nth-random succeed))))

(defmethod next-factor-state ((self improvizer) (state number) (proba number))
  (let ((tirage  (< (random 100) (1- proba))))
    (if tirage 
      (choose-factor-link self (flink  self state) nil)
      (if (< state (maxetat self)) 
        (1+ state)
        nil))))




;;; ============================================================================================================

(defclass ImproSupervizer (agent)
  ((time-origin :accessor time-origin)
   (osc-out-time-port :initarg :osc-out-time-port)
   (osc-out-midi-port :initarg :osc-out-midi-port)
   (osc-in-host :initarg :osc-in-host)
   (osc-out-host :initarg :osc-out-host)
   (bloc-impro-state :initform 'ready ::accessor bloc-impro-state) 
   (oracles  :accessor oracles :initform (make-array 4 :initial-element nil))
   (learners :accessor learners :initform (make-array 4 :initial-element nil))
   (improvisers :accessor improvisers :initform (make-array 4 :initial-element nil))
   (slicers :accessor slicers :initform (make-array 4 :initial-element nil))
   (unslicers :accessor unslicers :initform (make-array 4 :initial-element nil))
   (players :accessor players :initform (make-array 4 :initial-element nil))
   (listeners :accessor listeners :initform (make-array 4 :initial-element nil))))


(defvar *supervizer* nil)

(defvar *osc-in-host* "127.0.0.1")
(defvar *osc-out-host* "127.0.0.1")


(defvar *osc-out-time-port* 3007)
(defvar *osc-out-midi-port* 3008)
(defvar *Bosc-out-midi-port* 3009)


(defvar *OMaxDebug* nil)
(defvar *OMaxDebug2* nil)

(defvar *nb-clones* 10)
(defvar *clone-array* (make-array *nb-clones* :initial-element nil))


(defmethod task ((self improsupervizer))
  (let ((playahead (look-time-ahead (a-player 1))))
    (cond
     ((>= playahead 2500) ;(print playahead)
      (too-fast (a-improvizer 0)) )
     (t (too-slow (a-improvizer 0)))))
  (when (and (a-improvizer 1) ; ongoing bloc improvizer
             (eq (bloc-impro-state self) 'terminated)
             )
    (kill-bloc-impro self)
    (print "BLOC SENT")
    )
  ;(sleep .050)
  (process-allow-schedule)                
  )

(defmethod kill :before ((self ImproSupervizer))
  (loop for i from 0 to 3 do (and (a-learner i) (kill (a-learner i))))
  (loop for i from 0 to 3 do (and (a-improvizer i) (kill (a-improvizer i))))
  (loop for i from 0 to 3 do (and (a-slicer i) (kill (a-slicer i))))
  (loop for i from 0 to 3 do (and (a-unslicer i) (kill (a-unslicer i))))
  (loop for i from 0 to 3 do (and (a-player i) (kill (a-player i))))
  (loop for i from 0 to 3 do (and (a-listener i) (kill (a-listener i))))
  ;(kill *osc-midi-listener*)
  )


;---------------------------------------------------------------------------
; Bloc mode improvizer


; intantiate : 1 improviser, 1 unslicer, 1 Bosc-midi-player and connect them

(defmethod New-bloc-impro ((self ImproSupervizer) startphrase)
  (setf (a-improvizer 1)
        (new-improvizer (oracle (a-improvizer 0)))
        (a-unslicer 1)
        (new-unslicer "unslicer-1")
        (a-player 2)
        (new-Bosc-midi-player "Bosc-player"))
  (setf (blocmode (a-improvizer 1)) T 
        (blocStartPhrase (a-improvizer 1)) startphrase
        (max-continuity (a-improvizer 1)) (max-continuity (a-improvizer 0)))
  (connect (a-improvizer 1) :toport 4)
  (connect (a-unslicer 1) :fromport 4)
  (connect (a-unslicer 1) :toport 5)
  (connect (a-player 2) :fromport 5))

(defmethod kill-bloc-impro ((self ImproSupervizer))
  ;(disconnect (a-improvizer 1) :toport 4)
  ;(disconnect (a-unslicer 1) :fromport 4)
  ;(disconnect (a-unslicer 1) :toport 5)
  ;(disconnect (a-player 2) :fromport 5)
  ;kill does disconnect ..
  (send-Bosc-midi (list (list '/ma 'stop)))
  (kill (a-improvizer 1)) (setf (a-improvizer 1) nil)
  (kill (a-unslicer 1))
  (kill (a-player 2))
  (setf (bloc-impro-state self) 'ready)
  (setf nc::*sleep* nc::*save-sleep*))

(defmethod terminate-bloc-generation ((self ImproSupervizer) )
  (setf (bloc-impro-state self) 'terminated)
  )


(DefAgentMessage send-bloc-impro :now ((self ImproSupervizer) (duration number) startPhrase)
  (if (or (a-improvizer 1) (not (eq (bloc-impro-state self) 'ready)))
    (print "WAIT END OF PREVIOUS BLOC")
    (if (< (maxetat (oracle (a-improvizer 0))) 2)
      (print "Oracle too short. Not sending bloc")
      (progn (setf *last-suppleance* nil)
             (setf (bloc-impro-state self) 'running)
             (print "SENDING BLOC..........")
             (new-bloc-impro self startphrase)
             (setf (durmax (a-player 2)) duration)
             (setf (bloc-impro-state self) 'starting)
             (send-Bosc-midi (list (list '/ma 'start)))
             (resume (a-improvizer 1))
             ))))

;;; ============================================================================================================
;;; OSC messages for controlling processes


#|
(defun omaxon () 
  (setf *osc-listener-msg* (new-osc-listener 3005 *max-buf-size* "osc-listener-3005"))
  (setf *osc-listener-midi* (new-osc-listener 3006 *max-buf-size* "osc-listener-3006"))
  (setf *osc-time-end-point* (new-osc-end-point *osc-out-host* *osc-out-time-port* 1911))
  (setf   *osc-midi-end-point* (new-osc-end-point *osc-out-host* *osc-out-midi-port* 1912))
  (setf   *Bosc-midi-end-point* (new-osc-end-point *osc-out-host* *Bosc-out-midi-port* 1913))
  )

(defun OMaxOff()
  (and *supervizer* (OmaxKillSession))
  (loop for process in *all-processes* 
        do (unless (some #'(lambda (critical) (string= (process-name process) critical))
                         '("osc-listener-3006"  "osc-listener-3005"  "Listener"  "Initial" ))
             (process-kill process)))
  (kill-osc-listener *osc-listener-msg*)
  (kill-osc-listener *osc-listener-midi*)
  (close-osc-end-point *osc-midi-end-point*)
  (close-osc-end-point *osc-time-end-point*)
  (close-osc-end-point *Bosc-midi-end-point*)
  )
|#


(defun synchronize-max-om-time () 
  (send-osc-time (list (list '/mt (get-internal-real-time)))))

;;; init session with MS Listener/player plus osc listener and player


(defun OMaxInitSession ()
  (print "Initializing Impro Session")
  
  (setf *supervizer* (make-instance 'improSupervizer :state :asleep))
  (setf (time-origin *supervizer*) (get-internal-real-time))
  
  (setf (osc-in-host) *osc-in-host*
        (osc-out-host) *osc-out-host*
        (osc-out-time-port) *osc-out-time-port*
        (osc-out-midi-port) *osc-out-midi-port* )
  
  
  
  (let* (
         ;(ls (new-ms-listener "ms-listener" 0))
         (ol (new-osc-midi-listener "osc-listener"))
         (sl (new-polySlicer  "Slicer-0"))
         (or (neworacle))
         (ln (new-learner  or))
         (im (new-improvizer or))
         (un (new-unslicer "Unslicer-0"))
         ;(pl (new-ms-player "MS-Play-0" 0))
         (op (new-osc-midi-player "osc-play")))
    
    (setf (a-oracle 0) or
          (a-learner 0) ln
          (a-improvizer 0) im
          (a-slicer 0) sl
          (a-unslicer 0) un
          ;(a-player 0) pl
          (a-player 1) op
          ;(a-listener 0) ls
          )
    
    (setf *osc-midi-listener* ol (a-listener 1) ol)
    
    ;(connect ls :toport 0)
    (connect ol :toport 0)
    (connect sl :fromport 0)
    (connect sl :toport 1)
    (connect ln :fromport 1)
    (connect im :toport 2)
    (connect un :fromport 2)
    (connect un :toport 3)
    ;(connect pl :fromport 3)
    (connect op :fromport 3)
    )
  
  (synchronize-max-om-time)
  
  ;(send-osc-time (list (list '/mt (ms:MidiGetTime))))
  
  ;(osc-send (list (list '/mt (timeOrigin))) (osc-out-host) (osc-out-time-port))
  
  ;(silent (a-player 0))
  
  (setf *clone-array* (make-array *nb-clones* :initial-element nil))
  
  (resume *supervizer*)

  (print "OMAX INITIALISATION OK")
  )

(defun OMaxKillSession ()
  (kill *supervizer*)
  ;(kill *osc-midi-listener*)
  ;(setf *supervizer* nil)
  (print "OMAX QUITTING OK")
  )

(defAgentMessage Start-Learning :now ((self improsupervizer))
  ;(reset (a-listener 0))
  (reset (a-listener 1))
  (reset (a-slicer 0))
  (reset (a-learner 0))
  (resume (a-learner 0))
  (resume (a-slicer 0)))


(defun PutCloneToMain (nclone)
  (if (null  (aref *clone-array* (1- nclone)))
    (Format *stdout* "Clone nb ~d is empty !~%" nclone)
    (progn (stop-learning *supervizer*)
           (setf (oracle (a-learner 0)) (aref *clone-array* (1- nclone)))
           (start-learning *supervizer*))))


(nc:defAgentMessage Stop-Learning :now ((self improsupervizer))
  (reset (a-slicer 0))
  (reset (a-learner 0))
  (pause (a-slicer 0))
  (pause (a-learner 0)))

(DefAgentMessage Start-playing :now ((self improsupervizer))
  (reset (a-improvizer 0))
  (reset (a-unslicer 0))
  ;(reset (a-player 0))
  (reset (a-player 1))
  (synchronize-max-om-time)
  ;(resume (a-player 0))
  (resume (a-player 1)))

(DefAgentMessage Stop-Playing :now ((self improsupervizer))
  (reset (a-improvizer 0))
  (reset (a-unslicer 0))
  ;(reset (a-player 0))
  (reset (a-player 1))
  ;(pause (a-player 0))
  (pause (a-player 1)))


(DefAgentMessage start-impro :now ((self improsupervizer))
  (setf *last-suppleance* nil)
  ;(reset (a-improvizer 0))
  ;(reset (a-unslicer 0))
  ;(reset (a-player 1))
  (synchronize-max-om-time)
  (resume (a-improvizer 0))
  (resume (a-player 1)))


(DefAgentMessage stop-impro :now ((self improsupervizer))
  (pause (a-improvizer 0))
  (reset (a-improvizer 0))
  (reset (a-unslicer 0))
  (reset (a-player 1))
)



(DefAgentMessage reset-main :now ((self improsupervizer))
  (let ((awake-learn (awake-p (a-learner 0))) (awake-impro (awake-p (a-improvizer 0))))
    (stop-impro self)
    (stop-learning self)
    (let ((oracle (make-instance 'coracle :comparateur 'CompareEvents :lrsMode t)))
      (setf (oracle (a-improvizer 0)) oracle
            (oracle (a-learner 0)) oracle))
    (synchronize-max-om-time)
    (when awake-learn (start-learning self))
    (when awake-impro (start-impro self))))

(DefAgentMessage reset-learner :now ((self improsupervizer))
  (stop-learning self)
  (let ((oracle (make-instance 'coracle :comparateur 'CompareEvents :lrsMode t)))
    (setf (oracle (a-learner 0)) oracle))
  (synchronize-max-om-time)
  (start-learning self))

 
(DefAgentMessage wake-clone :now ((self improsupervizer) n)
  (if (or (> n *nb-clones*) (null  (aref *clone-array* (1- n))))
    (Format *stdout* "Clone nb ~d is empty !~%" n)
    (let ((awake-impro (awake-p (a-improvizer 0))))
      (stop-impro self)
      (setf (oracle (a-improvizer 0)) (aref *clone-array* (1- n)))
      (reset (a-improvizer 0))
      (reset (a-unslicer 0))
      ;(reset (a-player 0))
      (reset (a-player 1))
      (synchronize-max-om-time)
      (when awake-impro (start-impro self)))))

(DefAgentMessage wake-main :now ((self improsupervizer))
  (let ((awake-impro (awake-p (a-improvizer 0))))
    (stop-impro self)
    (setf (oracle (a-improvizer 0)) (oracle (a-learner 0)))
    (reset (a-improvizer 0))
    (reset (a-unslicer 0))
    ;(reset (a-player 0))
    (reset (a-player 1))
    (synchronize-max-om-time)
    (when awake-impro (start-impro self))))


(DefAgentMessage clone-oracle :now ((self improsupervizer) n)
  (when (<= n *nb-clones*)
(Print "CLONING...........")
    (with-lock-grabbed ((lock (oracle (a-learner 0))))
      (setf (aref *clone-array* (1- n))
            (clone (oracle (a-learner 0))))
      (reinitialize-instance (aref *clone-array* (1- n))
                             :lock (make-process-lock :name 'impro-lock)))
(Print "CLONED...........")
))

(defun load-oracle-direct (file)
  (handler-case
    (WITH-OPEN-ORACLE-FILE (in file :direction :input :if-does-not-exist :error)
      (eval (read in)) (eval (read in)))
    (error (any) (format *stdout* "Error while loading file : ~S~%" any) )))


#|
(DefAgentMessage load-oracle :now ((self improsupervizer) filename)
  (let*((archive (load-oracle-direct filename)))
(print archive)
    (when (and archive (or (typep archive 'coracle) (typep archive 'orchestre)))
      (let* ((oracle (if (typep archive 'coracle) 
                       archive 
                       (change-class (aref (improvizer-array archive) 0) 'coracle)))
             (awake-learn (awake-p (a-learner 0))) 
             (awake-impro (awake-p (a-improvizer 0))))
        (stop-impro self)
        (stop-learning self)
        (reinitialize-instance oracle :lock (make-lock 'impro-lock)) 
        ; the lock is not omng-saved
        (setf (oracle (a-improvizer 0)) oracle
              (oracle (a-learner 0)) oracle)
        (when awake-learn (start-learning self))
        (when awake-impro (start-impro self))))))

|#



(DefAgentMessage load-oracle-n :now ((self improsupervizer) filename clone-nb)
  (print "LOADING ORACLE ............")
  (let*((archive (load-oracle-direct filename)))
    (when (and archive (or (typep archive 'coracle) (typep archive 'orchestre)))
      (let* ((oracle (if (typep archive 'coracle) 
                       archive 
                       (change-class (aref (improvizer-array archive) 0) 'coracle))))
        (reinitialize-instance oracle :lock (make-process-lock :name 'impro-lock)) 
        ; the lock is not omng-saved
        (setf (aref *clone-array* (1- clone-nb)) oracle)))
    (print "ORACLE LOADED ............")
    ))

#|
(DefAgentMessage save-oracle :now ((self improsupervizer) filename)
  (let ((clone ))
    (print "First, CLONING ...........")
    (with-lock-grabbed ((lock (oracle (a-learner 0))))
      (setf clone (clone (oracle (a-learner 0)))))
    (print "SAVING ...................")
    (WITH-OPEN-FILE (out filename :direction :output  :if-does-not-exist :create :if-exists :supersede
                         :external-format :|TEXT|)
      (prin1 '(in-package :om) out)
      (prin1 (omng-save  clone) out))
    (print "SAVED")
    T))

|#

(DefAgentMessage save-oracle :now ((self improsupervizer) filename)
  (let ((clone ))
    (print "First, CLONING ...........")
    (with-lock-grabbed ((lock (oracle (a-learner 0))))
      (setf clone (clone (oracle (a-learner 0)))))
    (process-run-function "SavingFile"
                          #'(lambda (oracle)
                              (print "SAVING ...................")
                              (WITH-OPEN-FILE (out filename :direction :output  :if-does-not-exist :create :if-exists :supersede
                                                  )
                                (prin1 '(in-package :om) out)
                                (prin1 (omng-save  oracle) out))
                              (Print "SAVED")
                              (process-kill *current-process*)
                              )
                          clone)
    T))





#|
(DefAgentMessage save-oracle :now ((self improsupervizer) filename)
  (print filename)
  (with-lock-grabbed ((lock (oracle (a-learner 0))))
    (prin1 '(in-package :om) )
    (prin1 (omng-save  (oracle (a-learner 0))) ))
  T)

|#

(DefAgentMessage start-on-phrase :now ((self improsupervizer) onoff)
  (setf (StartOnPhrase (a-improvizer 0)) (= onoff 1)))


(defun improOn () (start-impro *supervizer*))
(defun improOff () (stop-impro *supervizer*))
(defun learnOn () (start-learning *supervizer*))
(defun learnOff () (stop-learning *supervizer*))
(defun ResetAgent () (reset-main *supervizer*))
(defun ResetLearner () (reset-learner *supervizer*))
(defun playOn () (start-Playing *supervizer*))
(defun playOff () (stop-Playing *supervizer*))
(defun WakeClone (n) (wake-clone *supervizer* n))
(defun wakemain () (wake-main *supervizer*))
(defun cloneOracle (n) (clone-oracle *supervizer* n))
(defun LoadOracleDirect (filename clone-nb) (load-oracle-n *supervizer* filename clone-nb))
(defun LoadOracle (clone-nb) 
  (catch-cancel
    (let ((name (choose-file-dialog)))
      (when name (loadOracleDirect name clone-nb)))))
(defun OMaxImpRange (min max)
  ;;; should start impro ???
  (setf (start-region (oracle (a-improvizer 0) )) (list min max)))
(defun SaveOracle (name) (save-oracle *supervizer* name))
(defun SaveOracleAs ()
  (catch-cancel
    (let ((name (choose-new-file-dialog :prompt "Save Oracle As...")))
      (when name
        (save-oracle *supervizer* name)))))

(defun OMaxSetContinuity (n)
  (setf (max-continuity (a-improvizer 0)) n))
(defun SendBlocImpro (dur) (send-bloc-impro *supervizer* dur -1))

(defun StartOnAPhrase (onoff) (start-on-phrase *supervizer* onoff))

(defun SendBlocImproOnPhrase (dur phrasenb) (send-bloc-impro *supervizer* dur phrasenb))

(defun OMaxDebug (onoff)  (setf *OmaxDebug* (= onoff 1)))
(defun OMaxDebug2 (onoff)  (setf *OmaxDebug2* (= onoff 1)))



#|

(loadoracle)

(saveoracleas)
(resetagent)


(progn 
(omaxinitsession)

(startlearn (a-learner 0))
(startImpro (a-improvizer 0))
(startPlayer (a-player 0))
)

(silent (a-player 0))
(audible (a-player 0))


(connect (a-player 0) :fromport 3)
(disconnect (a-player 0) :fromport 3)



(OMaxKillSession)

(stopPlayer (a-player 0))

(inspect *supervizer*)
(players *supervizer*)

(setf nc::*iac-central* (make-instance 'nc::iac-central))

(inspect nc::*iac-central*)

(inspect (a-improvizer 0))
(inspect (a-learner 0))
(inspect (a-listener 1))
(inspect (a-slicer 0))

(inspect *all-processes* )
(mapc #'process-kill (setf x (first-n  *all-processes* 50)))
(process-kill (setf x  (first *all-processes*)))
(process-disable (first *all-processes*))
(process-enable (first *all-processes*))
(process-whostate (first *all-processes*))
(process-whostate x)
kill

(setf osclistener (new-osc-listener 3006 *max-buf-size* "osc-listener-3006"))
(kill-osc-listener osclistener)


(osc-send '((/toto 1 2 3)) "169.254.190.131" 3007)

(osc-send (list (list '/mt (ms:midigettime))) "127.0.0.1" 3101)
(osc-send (list (list '/mt (ms:midigettime))) "127.0.0.1" 3102)
(send-osc *osc-time-end-point*  (list (list '/mt  (ms:midigettime))) )

2008

/ma date-mid pitch vel chan nphrase date-buffer-audio
/ma start
/ma stop

(loop for process in *all-processes* 
      do (unless (some #'(lambda (critical) (string= (process-name process) critical))
                       '("osc-listener-3006"  "osc-listener-3005"  "Listener"  "Initial" ))
           (process-kill process)))


(omaxoff)


(loop for i from 1 to (maxetat (oracle (a-improvizer 0)))
      do (format t "~d " (length (midiset (otext (oracle (a-improvizer 0)) i)))))



|#