(in-package :om)



#|
(defun osc-route-bundle (bundle)
  (handler-case
    (loop for item in bundle collect
          (cond
           ((stringp (car item)) 
            (cond ((string-equal (car item) "/mo") 
                   ;(print item) 
                   (funcall #'handle-osc-midi-evt (cdr item)))
                  ((string-equal (car item) "/ph") (funcall #'handle-osc-midi-phrase (cdr item)))
                  ((string-equal (car item) "/omsend") (osc-method-invoke item))
                  ((string-equal (car item) "/lisp") (print item) 
                   (execute-message item)
                   )
                  ((string-equal (car item) "/lisp-string") (execute-message-from-string item))
                  (t (format t "~D~" item))))
           ((consp (car item)) (osc-route-bundle item))
           (t nil))
          )
    (error (any) (print "Error trying to route OSC message") (print any) NIL))
  )
|#


;;; in save.lisp
;;; this is because using let has the 8192K barrier

(defvar *hashtable-tmp*)

(defmethod omNG-save ((self list) &optional (values? nil))
  (if  (null (nthcdr 25 self))
      `(list ,.(mapcar #'(lambda (x) (omNG-save x values?)) self))
    `(Biglist ,.(mapcar #'(lambda (x) (omNG-save x values?)) self))))

(defmethod omNG-save ((self array) &optional (values? nil))
   `(make-array ,(length self) :initial-contents (BigList ,.(map 'list #'(lambda (elt) (omNG-save elt values?)) self))))


(defmethod omNG-save ((self hash-table) &optional (values? nil)) 
  (declare (ignore values?))
  (let (keylist vallist)
    (maphash #'(lambda (key val)
                 (push key keylist)
                 (push val vallist)) self)
    (setf keylist (reverse keylist)
          vallist (reverse vallist))
    `(progn 
       (setf  *hashtable-tmp* (make-hash-table))
       (mapc #'(lambda (key val)
                 (sethash  *hashtable-tmp* key val))
             ,(omNG-save keylist)
             ,(omNG-save vallist)
             )
       *hashtable-tmp*)))


;; this is for the new udpsend max object
;; il n'envoie pas de bundle


(defun parse-buffer (buf i size)
  (let* ( rep mssg elemsize)
    (setf elemsize size)
    (loop while (< i size) do
          (setf mssg (read-osc-string buf i))
          ;(print mssg)
          (setf i (advance-str mssg i))
          (push  (parse-mssg buf i mssg elemsize) rep)
          (setf i (+ i elemsize))
          (setf elemsize (read-int32 buf i))
          (setf i (+ i 4)))
    (reverse rep)
    )
  )

;--------------


(defvar *max-play-ahead* 1000)
(defvar *max-impro-ahead* 100)

(defmethod task ((self improsupervizer))
  (let ((playahead (look-time-ahead (a-player 1))))
    (cond
     ((>= playahead *max-play-ahead*) ;(print playahead)
      (too-fast (a-improvizer 0)) )
     (t (too-slow (a-improvizer 0)))))
  (when (and (a-improvizer 1) ; ongoing bloc improvizer
             (eq (bloc-impro-state self) 'terminated)
             )
    (kill-bloc-impro self)
    (print "BLOC SENT")
    )
  ;(sleep .050)
  (process-allow-schedule)                
  )



;;;----------------------------------------------------------------------------------------------------------

;;; Nouvelle gestion des suffixes avant, nouvelle gestion
;;; de la navigation dans l'oracle



(defclass* oracle ()
  (
   (vectext :initform () :accessor vectext :initarg :vectext)
   (hashtransition :initform (make-hash-table :test 'equal) :accessor hashtransition :initarg :hashtransition)
   (hashsuppl :initform (make-hash-table :test 'equal) :accessor hashsuppl :initarg :hashsuppl)
   (hashsuppl-> :initform (make-hash-table :test 'equal) :accessor hashsuppl-> :initarg :hashsuppl->)
   (hashsuppl<-> :initform (make-hash-table :test 'equal) :accessor hashsuppl<-> :initarg :hashsuppl<->)
   (veclrs :initform () :accessor veclrs :initarg :veclrs)
   (maxetat :initform 0 :accessor maxetat :initarg :maxetat)
   (hashrepeats :accessor hashrepeats)
   (lrsmode :initform nil :accessor lrsmode :initarg :lrsmode)
   (nbresauts :accessor nbresauts :initform 0)
   (ptitsaut :accessor ptitsaut :initform 0)
   ))



(defmethod (setf suppleance) ((etat2 integer) (oracle oracle) (etat integer) )
  (setf (gethash etat (hashsuppl oracle))
        etat2)
  (if (null (gethash etat2 (hashsuppl-> oracle)))
    (setf (gethash etat2 (hashsuppl-> oracle))
          etat)
    (loop 
      for state1 = (gethash etat2 (hashsuppl-> oracle)) then (gethash state1 (hashsuppl<-> oracle))
      for state2 = (gethash state1 (hashsuppl-> oracle)) then (gethash state2 (hashsuppl<-> oracle))
          while state2
          finally (setf (gethash state1 (hashsuppl<-> oracle)) etat))))


(defmethod suppleance<-> ((oracle oracle) (etat integer))
  (loop for state = (gethash etat (hashsuppl-> oracle)) then (gethash state (hashsuppl<-> oracle))
        while state
        collect  state ))





;;;--------------------

(defvar *impro-tabou-states* nil)

(defvar *suffix-quality-thresh* 0.25)
(defvar *min-suffix-length* 3)
(defvar *max-nb-jump-candidates* 6)
(defvar *scanregion-divisor* 5)
(setf *SLength-distribution* #(1 2 3 25 50 80 130 210 340 550 900 1450 2000))
(setf *Tabou-list-length* 8)
(defvar *Close-edge-dist* 15)
(defvar *proba-if-1-sol* 0.5)

(defvar *regional-impro* NIL)




(defclass improviser (agent)
  ( (oracle :accessor oracle :initarg :oracle)
    (continuity :initform 0 :initarg :continuity :accessor continuity)
    (max-continuity :initform 8 :initarg :max-continuity :accessor max-continuity)
    (start-play :initform nil :accessor start-play :initarg :start-play )
    (stop :initform nil :accessor stop :initarg :stop)
    (gindex :initform 0  :accessor gindex :initarg :gindex :type integer)
    (durseg :initform 0  :accessor durseg :initarg :durseg :type integer)
    (blocmode :initform nil  :accessor blocmode :initarg :blocmode)
    (startOnPhrase :initform nil  :accessor startOnPhrase :initarg :startOnPhrase)
    (blocStartPhrase :initform -1  :accessor blocStartPhrase :initarg :blocStartPhrase)        
    )
  )



(defmethod maxetat ((self improviser) )
  (maxetat (oracle self) ))


(defmethod suppleance ((self improviser) (etat integer))
  (suppleance (oracle self) etat))

(defmethod suppleance<-> ((self improviser) (etat integer))
  (suppleance<-> (oracle self) etat))

(defmethod suppleance-> ((self improviser) (etat integer))
  (suppleance-> (oracle self) etat))


(defmethod lrs ((self improviser) (etat integer))
  (lrs (oracle self) etat))

(defmethod CompareEvents ((Event1 CrossEvent) (Event2 CrossEvent))
  
  (let* ((m1 (delete-duplicates  
              (mapcar #'(lambda (x) (mod (abs (MEPitch x)) 12)) (MidiSet Event1))))
         (m2 (delete-duplicates  
              (mapcar #'(lambda (x) (mod (abs (MEPitch x)) 12)) (MidiSet Event2))))
         (l1 (length m1))
         (l2 (length m2))
         (l3 (length (intersection m1 m2))))
    (and
     ;(similar-durations (duration event1) (duration event2))
     (or 
      (= l1 l2 l3)
      (>= l3 (/ (+ l1 l2) 3))))) )




(defmethod task  ((self improviser))
  (if (or (stop self) (<= (maxetat (oracle self)) 0))
    (process-allow-schedule)
    (cond
     ;; new impro
     ((zerop (gindex self))
      (setf (continuity self) 0)
      ;(setf   *impro-tabou-states*  nil)
      (cond 
       ((and (blocmode self) (>= (blocStartPhrase self) 0))
        (start-new-impro-on-phrase-nb self))
       
       ((StartOnPhrase self)
        (start-new-impro-on-next-phrase self))
       
       (t
        (start-new-impro-in-region self))))
     
     ;; continuing   
     (t ; (in-continuity self)
      (continue-impro self))))
  
  (when (> (durseg self) *max-impro-ahead* ) (setf (durseg self) 0) (process-allow-schedule) )
  )



(defmethod start-new-impro-on-phrase-nb  ((self improviser))
  (catch
    'tag-fblm
    (loop for i from 1 to (1- (maxetat (oracle self)))
          if (and (startPhrase (otext (oracle self) i))
                  (= (phraseNumber (otext (oracle self) i)) (blocStartPhrase self)))
          do (progn (format t "Bloc starting at phrase ~d~%" (phraseNumber (otext (oracle self) i))) 
                    (impro-output-1-event self i)
                    (throw 'tag-fblm i)))
    (impro-failure self)))


(defmethod impro-failure ((self improviser))
  (dprint "Impro failure at ~a ~%" (gindex self))
  (setf (gindex self) 0 (continuity self) 0)
  (process-allow-schedule))
  

(defmethod start-new-impro-on-next-phrase  ((self improviser))
  (let ((left  (round (* (/ (first (start-region (oracle self))) 100) (1- (maxetat (oracle self)))))))
    (catch
      'tag-fblm
      (loop for i from (max 1 (min left (1- (MaxEtat (oracle self))))) to (1- (MaxEtat (oracle self)))
            do (when (startPhrase (otext (oracle self) i))
                 (impro-output-1-event self i)
                 (format t "~%Starting on phrase ~a~%" (phraseNumber (otext (oracle self) i)))
                 (throw 'tag-fblm i)))
      (loop for i from (max 1 (min left (1- (MaxEtat (oracle self))))) downto 1
            do (when (startPhrase (otext (oracle self) i))
                 (impro-output-1-event self i)
                 (format t "~%Starting on phrase ~a~%" (phraseNumber (otext (oracle self) i)))
                 (throw 'tag-fblm i)))
      (impro-failure self))))


(defmethod start-new-impro-in-region  ((self improviser))
  (let* ((left  (round (* (/ (first (start-region (oracle self))) 100) (1- (maxetat (oracle self))))))
         (width (round (* (/ (- (second (start-region (oracle self))) (first (start-region (oracle self)))) 100)  (1- (maxetat (oracle self))))))
         (start (max 1 (min (+ (random width) left) (1- (maxetat (oracle self)))))))
    (dprint "Starting at : ~a~%" start)
    (impro-output-1-event self start) ))


(defmethod impro-output-1-event ((self improviser) (index integer))
  (talk self (otext (oracle self) index))
  (setf (gindex self) index)
  (incf  (continuity self))
  (dprint "~a(~a)~%" index (continuity self))
  (incf (durseg self) (duration (otext (oracle self) index))))



(defmacro with-proba (n &body body)
  `(if (<= (random 1.0) ,n)
     ,(car body)
     ,(cadr body)))



(defun SQuality (s) (first s))
(defun SIndex (s) (second s))
(defun SSuffix (s) (third s))
(defun SLength (s) (fourth s))
(defun SDurRef (s) (fifth s))
(defun SDurSuffix (s) (sixth s))


;; all-suffixes-xx -> (... (quality index suffix length dur-ref dur-suffix ) ...)
;; scanregion improviser -> (indexleft indexright)
;; time-to-scan-ahead
;; scan-ahead-length

(defmethod continue-impro ((self improviser))
  (if (not (time-to-scan-ahead self))
    (move-linearly self)
    (let* ((bwsuffixes (all-suffixes-bw self (scanregion self)))
           (fwsuffixes (all-suffixes-fw self (scanregion self)))
           (bwfwsuffixes (all-suffixes-bw-fw self (scanregion self)))
           (choice (filter-all-suffixes self bwsuffixes fwsuffixes bwfwsuffixes)))
      (if (null choice)
        ;(progn (setf (continuity self) 0) (move-linearly self))
        (move-to-end-scan self)
        (move-to-suffix self choice)))))


(defmethod move-to-suffix  ((self improviser) suffix)
  (loop for i from (1+ (gindex self)) to (SIndex suffix)
        do (impro-output-1-event self i))
  (setf (continuity self) 0)
  (dprint "~%~%Jumping from ~a to ~a length ~a qual ~,2f ~%~%" 
          (Sindex suffix) (1+ (SSuffix suffix)) (Slength suffix) (Squality suffix))
  (impro-output-1-event self (1+ (SSuffix suffix)))
  (push (SSuffix suffix) *impro-tabou-states*)
  (when (> (length *impro-tabou-states*) *Tabou-list-length*) 
    (setf  (nthcdr *Tabou-list-length* *impro-tabou-states*)  nil))
  )


(defmethod move-to-end-scan  ((self improviser))
(print "--")
  (loop for i from (gindex self) to (second (scanregion self))
        do (move-linearly self)))
  


#|
      (with-proba 0.5
        (jump-to-some-suffix self fwsuffixes)
        (with-proba 0.5
          (jump-to-some-suffix self bwsuffixes)
          (bw-fw-suffix-indirection self bwsuffixes))))))
|#


;; this to be redefined for actual scan regions 

#|
(defmethod scanregion ((self improviser))
  (list (gindex self) (gindex self)))

|#



#|
(defmethod scanregion ((self improviser))
  (if (< (continuity self) (- (max-continuity self) 3))
   (list  (gindex self) (gindex self))
   (list (gindex self) (min (maxetat self) (+ (gindex self) 3)))))
|#



(defmethod scanregion ((self improviser))
  (if (close-to-the-edge? self)
    (list (gindex self) (maxetat self))
    (list (gindex self) 
          (min (maxetat self)
               (+ (gindex self) 
                  (* 2 (round (max-continuity self) 
                              *scanregion-divisor* )))))))



#|
(defmethod time-to-scan-ahead ((self improviser))
  (>=  (continuity self) (max-continuity self)))

|#

#|
(defmethod time-to-scan-ahead ((self improviser))
  (>=  (continuity self) (- (max-continuity self) 3)))
|#


(defmethod time-to-scan-ahead ((self improviser))
  (or (and (close-to-the-edge? self) 
           (true (setf *impro-tabou-states* nil) (dprint "~%Close to the edge~%")))
      (>=  (continuity self) 
           (- (max-continuity self) 
              (round (max-continuity self) *scanregion-divisor* )))))



(defmethod close-to-the-edge? ((self improviser))
  (>=
   (gindex self) 
   (- (maxetat self)
      (max *Close-edge-dist*  (round (max-continuity self) *scanregion-divisor* )))))



;; this to be redefined for actual scan regions 


(defmethod move-linearly ((self improviser))
  (if (< (gindex self) (maxetat (oracle self)))
    (impro-output-1-event self (1+ (gindex self)))
    (impro-failure self)))


(defmethod all-suffixes-bw ((self improviser) (scanregion list))
  (loop for index from (first scanregion) to (second scanregion)
        nconc
        (loop for supp = (suppleance self index) then (suppleance self supp)
              for previous = index then (suppleance self previous)
              while (and supp (> supp 0))
              for dur-ref = (suffixduration self index (lrs  self previous))
              for dur-suffix = (suffixduration self supp (lrs  self previous))
              if (not (memq supp *impro-tabou-states*))
              collect (list (suffix-quality self index supp (lrs self previous) dur-ref dur-suffix)
                            index
                            supp
                            (lrs self previous)
                            dur-ref
                            dur-suffix)
              )))

(defmethod all-suffixes-fw ((self improviser) (scanregion list))
  (loop for index from (first scanregion) to (second scanregion)
        nconc 
        (loop for supp in (suppleance<-> (oracle self) index)
              for dur-ref = (suffixduration self index (lrs self supp))
              for dur-suffix = (suffixduration self supp (lrs self supp))
              if (not (memq supp *impro-tabou-states*))
              collect (list (suffix-quality self index supp (lrs self supp) dur-ref dur-suffix)
                            index
                            supp
                            (lrs self supp)
                            dur-ref
                            dur-suffix))))

(defmethod all-suffixes-bw-fw ((self improviser) (scanregion list))
  (loop for index2 from (first scanregion) to (second scanregion)
        for index = (suppleance self index2)
        nconc
        (and index 
             (> index 0)
             (loop for supp in (suppleance<-> (oracle self) index)
                   for dur-ref = (suffixduration self index2 (min (lrs self supp)(lrs self index2)))
                   for dur-suffix = (suffixduration self supp (min (lrs self supp)(lrs self index2)))
                   if (not (memq supp *impro-tabou-states*))
                   collect (list (suffix-quality self index2 supp (lrs self supp) dur-ref dur-suffix)
                                 index2
                                 supp
                                 (min (lrs self supp)(lrs self index2))
                                 dur-ref
                                 dur-suffix)))))


       


(defmethod suffixduration ((self improviser) (index integer) (sufflength integer))
  (loop for i from 0 to (1- sufflength)
        for j = (- index i)
        while (> j 0)
        sum (duration (otext  (oracle self) j))))



(defmethod suffix-quality ((self improviser) index suffix lrs dur-ref dur-suffix)
  (declare (ignore  index suffix lrs))
  (abs (log (/ dur-ref dur-suffix))))

(defun stoch-choose-suffix (sufflist)
  (and sufflist
       (stochastic-choice  
        (mapcar 
         #'(lambda (l)
             (aref *SLength-distribution* (min (1- (Slength l)) (1- (length *SLength-distribution*)))))
         sufflist)
        sufflist)))

(defun stochastic-choice (ProbaVector Values)
  (loop with random-value = (random (float (apply #'+ probavector)))
        for i from 0 to (1- (length ProbaVector))
        sum (nth i ProbaVector) into SumProba
        if (> sumproba random-value) return (nth i values)
        finally return (first (last values))))


(defmethod filter-all-suffixes ((self improviser) bwsuffixes fwsuffixes bw-fwsuffixes)
  ;(print (list bwsuffixes fwsuffixes bw-fwsuffixes)) 
  (let ((best-candidates 
         (firstn 
          (sort (loop for s in (nconc  bwsuffixes fwsuffixes bw-fwsuffixes)
                      if (and (< (SQuality s) *suffix-quality-thresh*) 
                              (not (= (Sindex s) (SSuffix s)))
                              (not (< (- (maxetat self) (SSuffix s)) *Close-edge-dist*))
                              (if (close-to-the-edge? self) 
                                (and (< (SSuffix s) (SIndex s))
                                     (>= (Slength s) 2))
                                (not (< (Slength s) *min-suffix-length*)))
                              )
                      collect s) 
                #'> :key #'SLength)
          *max-nb-jump-candidates*)))
    (dprint "best Candidate NA : ~a~%" (mapcar #'Slength best-candidates))
    (if (and best-candidates (null (rest best-candidates))) ; only one ?
      (with-proba *proba-if-1-sol*
        (progn (dprint "--------1 solution, I take it ~%")
               (first best-candidates))
        (progn (dprint "--------1 solution, I DON'T take it ~%")
               nil))
      (stoch-choose-suffix   best-candidates)
      )))


#|
(defmethod check-not-too-far ((self improviser) (suffix integer))
  (or
   (> (- (maxetat self) suffix) *Close-edge-dist*)
   (loop for i from (1+ suffix) to (maxetat self)
         for s = (suppleance self i)
         when s return (true (setf *impro-tabou-states*  nil))
         finally return nil )))


(defmethod filter-all-suffixes ((self improviser) bwsuffixes fwsuffixes bw-fwsuffixes)
  ;(print (list bwsuffixes fwsuffixes bw-fwsuffixes)) 
  (let ((best-candidates 
         (firstn 
          (sort (loop for s in (nconc  bwsuffixes fwsuffixes bw-fwsuffixes)
                      if (and (< (SQuality s) *suffix-quality-thresh*) 
                              (not (= (Sindex s) (SSuffix s)))
                              (or (not (close-to-the-edge? self) )
                                  (< (SSuffix s) (SIndex s)))
                              (>= (Slength s) *min-suffix-length*)
                              (check-not-too-far self s)
                              )
                      collect s) 
                #'> :key #'SLength)
          *max-nb-jump-candidates*)))
    (dprint "best Candidate NA : ~a~%" (mapcar #'Slength best-candidates))
    (stoch-choose-suffix   best-candidates)
    ))
    

|#
        


(defmethod filter-all-suffixes ((self improviser) bwsuffixes fwsuffixes bw-fwsuffixes)
  ;(print (list bwsuffixes fwsuffixes bw-fwsuffixes)) 
  (let ((best-candidates 
         (firstn 
          (sort (loop for s in (nconc  bwsuffixes fwsuffixes bw-fwsuffixes)
                      if (and (suffix-in-region self (SSuffix s))
                              (< (SQuality s) *suffix-quality-thresh*) 
                              (not (= (Sindex s) (SSuffix s)))
                              (not (< (- (maxetat self) (SSuffix s)) *Close-edge-dist*))
                              (if (close-to-the-edge? self) 
                                (and (< (SSuffix s) (SIndex s))
                                     (>= (Slength s) 2))
                                (not (< (Slength s) *min-suffix-length*)))
                              )
                      collect s) 
                #'> :key #'SLength)
          *max-nb-jump-candidates*)))
    (dprint "best Candidate NA : ~a~%" (mapcar #'Slength best-candidates))
    (if (and best-candidates (null (rest best-candidates))) ; only one ?
      (with-proba *proba-if-1-sol*
        (progn (dprint "--------1 solution, I take it ~%")
               (first best-candidates))
        (progn (dprint "--------1 solution, I DON'T take it ~%")
               nil))
      (stoch-choose-suffix   best-candidates)
      )))

(defmethod suffix-in-region ((self improviser) (suffix integer))
  (if (not  *regional-impro*)
    T
    (let* ((left  (round (* (/ (first (start-region (oracle self))) 100) (maxetat (oracle self)))))
           (right (round (* (/ (second (start-region (oracle self))) 100) (maxetat (oracle self))))))
      ;(print (list left right))
      (<= (max 1 (1- left)) suffix (1- right)))))


(defmethod move-linearly ((self improviser))
  (let ((left  (if *regional-impro*
                 (round (* (/ (first (start-region (oracle self))) 100) (maxetat (oracle self))))
                 0))
        (right (if *regional-impro*
                 (round (* (/ (second (start-region (oracle self))) 100) (maxetat (oracle self))))
                 (maxetat self))))
    (if (< (1- left) (gindex self) right)
      (impro-output-1-event self (1+ (gindex self)))
      (impro-failure self))))


(defmethod move-to-end-scan  ((self improviser))
  (loop for i from (gindex self) to 
        (if *regional-impro*
          (min (second (scanregion self)) (round (* (/ (second (start-region (oracle self))) 100) (maxetat (oracle self)))))
          (second (scanregion self)))
        do (move-linearly self)))


#|
(defmethod move-linearly ((self improviser))
  (if (< (gindex self) (maxetat (oracle self)))
    (impro-output-1-event self (1+ (gindex self)))
    (impro-failure self)))

|#
