(in-package :om)


;;; OSC receive process
;;; waits for messages in the form of
;;;           /lisp func-name arg1 arg2 ...
;;; or        /lisp-string "<S-expr>"
;;; or        /omsend obj-name method-name arg1 arg2 ...
;;; if method-name is a slot name, performs (setf (slot-value obj method-name) arg1) where obj is the object referred
;;; by obj-name. All names are packaged in :om.
;;; 

(defun execute-message (message)
  (apply (if (symbolp (cadr message))
             (cadr message)
           (intern (read-from-string (cadr message)) :om))
         (cddr message)))


(defun handle-omax-osc-message (msg host)
  (let ((message (osc:decode-message msg)))
    (handler-case 
        (route-omax-osc-message message)            
      (error (any) (error any)))
    nil))

(defun  route-omax-osc-message (message)
    (handler-case
;(progn (format *stdout* "~S ~%" message) 
        (cond
           ((stringp (car message)) 
            (cond ((string-equal (car message) "/mo") (funcall #'handle-osc-midi-evt (cdr message)))
                  ((string-equal (car message) "/ph") (funcall #'handle-osc-midi-phrase (cdr message)))
                  ((string-equal (car message) "/sp") (funcall #'handle-osc-midi-stop-evt (cdr message)))
                  ((string-equal (car message) "/cc") (funcall #'handle-osc-cc-evt (cdr message)))
                  ((string-equal (car message) "/omsend") (osc-method-invoke message))
                  ((string-equal (car message) "/lisp") 
                   (format *stdout* "~s~%" message) 
                   (execute-message message))
                  ((string-equal (car message) "/lisp-string") 
                   (format *stdout* "~s~%" message)
                   (execute-message-from-string message))
                  (t (format *stdout* "~D~%" message))))
           ((consp (car message)) (osc-route-bundle message))
           (t nil))
    (simple-condition (any) (apply #'format *stdout* (concatenate 'string "Error trying to route OSC message :" 
                                                                  (simple-condition-format-control any)
                                                                  "~%")
                                   (simple-condition-format-arguments any)))
    (cell-error (any)  (format *stdout* "Error trying to route OSC message ~S : ~S~%" any (cell-error-name any)))
    (error (any)  (format *stdout* "Error trying to route OSC message ~S : ~S~%" any message))))



(defun osc-method-invoke (invocation)
  (let ((object (symbol-value (intern  (string-upcase  (second invocation)) :om)))
        (method  (intern  (string-upcase  (third invocation)) :om)))
    (if (slot-exists-p  object method)
      (setf (slot-value object method) (fourth invocation))
      (apply  (symbol-function method) object (cdddr invocation)))))



(defun new-osc-listener (host port name)
  (osc:start-osc-server host port #'handle-omax-osc-message))


(defun kill-osc-listener (osc-server)
  (osc:stop-osc-server osc-server))


(defun send-osc (osc-stream message)
  (handler-case
      (osc:write-osc-msg message osc-stream)
    (error (any) (format *stdout* "> "))))

(defun send-osc-bundle (osc-stream bundle)
  (handler-case
      (osc:write-osc-bundle bundle osc-stream)
    (error (any) (format *stdout* "> "))))



