
(in-package :om)


(defvar *diapason-freq* 440.0)
(defvar *diapason-midic* 6900)


(defun midics->freqs (midics)
  (loop for midic in midics
        collect (* *diapason-freq*
                   (expt 2.0 (/ (- midic *diapason-midic*) 1200.0)) )))


(defun approx-midic (midic approx)
  (if (zerop approx) 
    (round midic)
    (round (* (floor (+ (* midic approx) 100) 200) 200) approx)))

(defun freq->midic (freq)
  (+ (* (log (max 0.0000000001 (/ freq *diapason-freq*))) #.(/ 1200 (log 2.0)))
     *diapason-midic*))


(defun virtual-fundamental (midis approx)
  (approx-midic (freq->midic (floating-gcd (midics->freqs  (mapcar #'(lambda (x) (* 100 x)) midis))
                                          (1- (expt 2.0 (/ approx 1200.0)))))
                (/ 200 approx)))


(defun floating-gcd (values grid-ratio)
  "floating gcd with tolerance grid-ratio around the values."
  (labels ((grid-above (val) (* val (1+ grid-ratio)))
           (grid-below (val) (/ val (1+ grid-ratio)))
           (gcd-try (values gcd-min gcd-max)
             (when (<= gcd-min gcd-max)
               (ifnot values
                 (/ (+ gcd-min gcd-max) 2.0)
                 (let* ((val-below (grid-below (first values)))
                        (val-above (grid-above (first values)))
                        (quo-min (ceiling (/ val-below gcd-max)))
                        (quo-max (floor (/ val-above gcd-min))))
                   (do* ((quotient quo-min (1+ quotient)) (gcd-interval))
                        ((> quotient quo-max) nil)
                     (setf gcd-interval
                           (gcd-try (rest values)
                                         (max gcd-min (/ val-below quotient))
                                         (min gcd-max (/ val-above quotient))))
                     (when gcd-interval
                       (return-from gcd-try gcd-interval))))))))
    (gcd-try values .1 (grid-above (apply 'min values)))))





; le deuxieme parametre (approx) est la precision spectrale dans le matching des notes avec les partiels
; thoriques
; 100 : au 1/2 ton 50: au 1/4 de ton 25: au 1/8 de ton
; on utilise le plus souvent 50
; le rsultat est un midicent arrondi selon approx ce qui permet la comparaison exacte
; ainsi avec approx = 50 (resp 100, 25) les resultats sont toujours un 1/4 (resp 1/2ton, 1/8ton) de ton juste 
; soit un multiple de 50 (resp 100, 25) en midicents



;(virtual-fundamental '(60 67 72 76) 100)
;(virtual-fundamental '(60 67 72 76 77) 100)
;(virtual-fundamental '(60 67 72 76) 50)
;(virtual-fundamental '(60 67 72 76 77) 25)

#|
(defmethod CompareEvents ((Event1 CrossEvent) (Event2 CrossEvent))
  
  (let* ((m1 (virtual-fundamental 
              (mapcar #'(lambda (x) (abs (MEPitch x))) (MidiSet Event1)) 50))
         (m2 (virtual-fundamental 
              (mapcar #'(lambda (x) (abs (MEPitch x))) (MidiSet Event2)) 50)))
          (= (mod m1 1200) (mod m2 1200))) )

(defmethod CompareEvents ((Event1 CrossEvent) (Event2 CrossEvent))
  
  (let* ((m1 (mapcar #'(lambda (x) (abs (MEPitch x))) (MidiSet Event1)) )
         (m2 (mapcar #'(lambda (x) (abs (MEPitch x))) (MidiSet Event2)) ))
         (and
          ( < (dist-estrada (om* m1 100) (om* m2 100)) 3)
          (= (mod (virtual-fundamental m1 50) 1200) (mod (virtual-fundamental m2 50) 1200)))))

(defmethod CompareEvents ((Event1 CrossEvent) (Event2 CrossEvent))
  
  (let* ((m1 (mapcar #'(lambda (x) (abs (MEPitch x))) (MidiSet Event1)) )
         (m2 (mapcar #'(lambda (x) (abs (MEPitch x))) (MidiSet Event2)) ))
         (and
          ( = (dist-estrada (om* m1 100) (om* m2 100)) 0)
          (= (mod (virtual-fundamental m1 50) 1200) (mod (virtual-fundamental m2 50) 1200)))))


(defmethod CompareEvents ((Event1 CrossEvent) (Event2 CrossEvent))
  
  (let* ((m1 (mapcar #'(lambda (x) (abs (MEPitch x))) (MidiSet Event1)) )
         (m2 (mapcar #'(lambda (x) (abs (MEPitch x))) (MidiSet Event2)) ))
         (and
          ( = (floor (log (length m1) 2)) (floor (log (length m2) 2)))
          (= (mod (virtual-fundamental m1 50) 1200) (mod (virtual-fundamental m2 50) 1200))
          )))


|#


